/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   @return 
     - "true" is x1 and x2 are the same clock:
          - the formula are in the same class
	  -  the formula are in the same class looking at the suspended synchronisations.
     - "false" otherwise.
*/
  boolean assertions2events::satisfiesClockEq (ptinsth x1, ptinsth x2)
{
  ptinsth vh1, vh2;
  ptlinsth lh1, lh2;

  x1 = CKM->getRepresentativeFormula (x1);
  x2 = CKM->getRepresentativeFormula (x2);
  if (x1 == x2)
    return pk_true;
  CKM->getSuspendedSynchros (&lh1, &lh2);
  while (ptlinsth_get_next (&lh1, &vh1))
    {
      ptlinsth_get_next (&lh2, &vh2);
      vh1 = CKM->getRepresentativeFormula (vh1);
      vh2 = CKM->getRepresentativeFormula (vh2);
      if ((vh1 == x1) && (x2 == vh2))
	return pk_true;
      if ((vh1 == x2) && (x1 == vh2))
	return pk_true;
      // to be completed
    }
  return pk_false;
}


/**
   It finalizes the functionality by adding new assumptions in the system (induced by the rewriting) and
   updates the system by adding suspended synchronisations (suspended during the functionality).
 */
void
assertions2events::finalize ()
{
  ptinsth vh1, vh2;

  CKM->setModeSynchroSolving (pk_true);
  while (ptlinsth_get_next (&lass1, &vh1))
    {
      ptlinsth_get_next (&lass2, &vh2);
      CKM->makeAssumption (vh1, vh2);
    }
  // make suspended Synchros
  CKM->hclmklequiv (0);
}

/**
   For two extractions (vh1 = when C1, vh2 = when C2), with the same clock (vhc) for the referenced expression (C1,C2), it
   builds an extraction (C1 op C2) and builds the induced extractions (nvhpos, nvhneg).
   @return "true" if the operation succeeds, "false" otherwise (not extractions,....).
*/
boolean
assertions2events::mkOpAssertion (noperators op, ptinsth vh1, ptinsth vh2, ptinsth * nvhpos, ptinsth * nvhneg, ptinsth * vhc)
{
  ptexpcal e1, e2, vexpcond;
  ptinsth vhcond1, vhcond2;

  if (!CKM->satisfiesSub (vh1))
    return pk_false;
  if (!CKM->satisfiesSub (vh2))
    return pk_false;
  vhcond1 = CKM->getClockSub (vh1);
  vhcond2 = CKM->getClockSub (vh2);
  if (!satisfiesClockEq (vhcond1, vhcond2))
    return pk_false;

  e1 = CKM->getConditionSub (vh1);
  if (CKM->getOperator (vh1) == nosub)
    e1 = EM->mkUnaryExp (opnon, e1, EM->getBackAnnotation (e1));

  e2 = CKM->getConditionSub (vh2);
  if (CKM->getOperator (vh2) == nosub)
    e2 = EM->mkUnaryExp (opnon, e2, EM->getBackAnnotation (e2));

  vexpcond = EM->mkBinaryExp (op, e1, e2, NILBACKANNOTATION);
  *nvhpos = CKM->makeSub (vhcond1, vexpcond, EM->getPredecessors (vexpcond));
  *nvhneg = CKM->makeNoSub (vexpcond);
  *vhc = vhcond1;
  return pk_true;
}

/**
   It adds new assertions (internal attributes lass1, lass2): the i-th element of lass1 is equivalent to the i-th element of lass2
   (the equivalence is assumed satisfied).
   These assumptions will be added in the system at the end of the functionality (finalize() method).
*/
boolean
assertions2events::addAssert (ptinsth vh1, ptinsth vh2, ptinsth vhnote1, ptinsth vhnote2, boolean b)
{
  ptinsth t1, t2;

  if (b)
    {
      t1 = vh2;
      t2 = vhnote2;
    }
  else
    {
      t1 = vhnote2;
      t2 = vh2;
    }
  lass1 = ptlinsth_insert_elem (lass1, vh1);
  lass2 = ptlinsth_insert_elem (lass2, t1);
  lass1 = ptlinsth_insert_elem (lass1, vhnote1);
  lass2 = ptlinsth_insert_elem (lass2, t2);
  return pk_true;
}

/**
   when b is
     - pk_true, it adds the assumptions vh1 = vh2, and vhnote1=vhnote2 (these equivalences are assumed satisfied),
     - pk_false, it adds the assumptions vh1 = vhnote2, and vhnote1=vh2 (these equivalences are assumed satisfied).
*/
void
assertions2events::replaceAssert (ptinsth vh1, ptinsth vh2, ptinsth vhnote1, ptinsth vhnote2, boolean b)
{
  ptinsth t1, t2, vh1c;

  if (b)
    {
      t1 = vh2;
      t2 = vhnote2;
    }
  else
    {
      t1 = vhnote2;
      t2 = vh2;
    }

  if (CKM->satisfiesSub (vh1))
    {
      if (!CKM->satisfiesSub (vhnote1))
	{
	  CKM->getComplExtraction (vh1, &vh1c);
	  CKM->makeAssumption (vh1c, CKM->makeMinus (CKM->getClockSub (vh1), t1));
	}
    }
  else
    {
      if (CKM->satisfiesSub (vhnote1))
	{
	  CKM->getComplExtraction (vhnote1, &vh1c);
	  CKM->makeAssumption (vh1c, CKM->makeMinus (CKM->getClockSub (vhnote1), t2));
	}
    }
  CKM->makeAssumption (vh1, t1);
  CKM->makeAssumption (vhnote1, t2);
}



/**
   It rewrites a formula vh = when C, with assert(C), ^C = when C. 
    - C defined by (b1=b2) --> when b1 = when b2, when not b1 = when not b2
    - C defined by (b1!=b2) --> when b1 = when not b2 and when not b1 = when b2
    othewise, it applies the rules defined in isFreeCondition() method.
   @param vh  (in) a formula, assumed to be a positive extraction.
*/
boolean
assertions2events::rewAssertion (ptinsth vh)
{
  ptexpcal e, e1, e2;
  noperators op;
  boolean b, x, b1, b2;
  ptinsth vh1, vh2, vhcond, vhnote1, vhnote2;
  ptinsth nvhpos, nvhneg, vhncond;
  ptinsth vhsubn1, vhnosubn1, vhsubn2, vhnosubn2;

  if (!CKM->satisfiesIsAnAssumption (vh))
    return pk_false;
  if (!CKM->satisfiesSub (vh))
    return pk_false;
  e = CKM->getConditionSub (vh);
  if (EM->isMarked (e, vmark))
    return pk_false;
  EM->addMark (e, vmark);
  vhcond = CKM->getClockSub (vh);
  op = EM->getOperator (e);
  b = satisfiesClockEq (vhcond, vh);
  switch (op)
    {
    case opegal:
    case opdiff:
      e1 = EM->getBinaryArgstAt (e, 1);
      if (!EM->IsBoolean (e1))
	return pk_false;
      e2 = EM->getBinaryArgstAt (e, 2);

      vh1 = mksubnosub (e1, vhcond, &vhnote1);
      vh2 = mksubnosub (e2, vhcond, &vhnote2);

      b1 = isFreeCondition (e1, vhcond, &vhsubn1, &vhnosubn1);
      b2 = isFreeCondition (e2, vhcond, &vhsubn2, &vhnosubn2);
      CKM->removeSub (vh, pk_true, &x);
      if (op == opegal)
	addAssert (vh1, vh2, vhnote1, vhnote2, b);
      else
	addAssert (vh1, vhnote2, vhnote1, vh2, b);
      if (b1)
	{
	  vhsubn1 = vh1;
	  vhnosubn1 = vhnote1;
	}
      else
	replaceAssert (vh1, vhsubn1, vhnote1, vhnosubn1, pk_true);
      if (b2)
	{
	  vhsubn2 = vh2;
	  vhnosubn2 = vhnote2;
	}
      else
	replaceAssert (vh2, vhsubn2, vhnote2, vhnosubn2, pk_true);

      if (b1 && b2)
	return pk_false;

      if (mkOpAssertion (op, vhsubn1, vhsubn2, &nvhpos, &nvhneg, &vhncond))
	replaceAssert (nvhpos, vhncond, nvhneg, the_nullClock_B2E_copy, b);
      else
	replaceAssert (vhsubn1, vhsubn2, vhnosubn1, vhnosubn2, b);
      return pk_true;

    default:
      if (isFreeCondition (e, vhcond, &vhsubn1, &vhnosubn1))
	return pk_false;
      CKM->getComplExtraction (vh, &vhnote1);
      replaceSubs (vh, vhsubn1, vhnote1, vhnosubn1);
      return pk_true;
    }

}

/**
   Local attribute management: it adds a mark (pKbool2eventMark) to a clock (vh).
   It is used to mark the expressions not implicated in assertions.
*/
void
assertions2events::markClock (ptinsth vh)
{
  if (!CKM->satisfiesMarked (vh, pKbool2eventMark))
    CKM->addMark (vh, pKbool2eventMark);
}

/**
   Local attribute management: it adds a mark (m) to the expression of definition of a signal (vsig), its clocks.
   It is used to mark the expressions not implicated in assertions (makeSynchrosNotImplicatedInAssertions()).
*/
void
assertions2events::markSignal (ptexpcal vsig, typeMark m)
{
  ptINode vinstr;
  ptexpcal vhe;
  ptexpcal exp;
  ptinsth h, vhs, vhns;

  if (!EM->IsSignal (vsig))
    return;
  if (EM->isStateVar (vsig))
    return;
  if (EM->IsBoolean (vsig))
    {
      h = EM->getClock (vsig);
      markClock (h);
      vhs = mksubnosub (vsig, h, &vhns);
      markClock (vhs);
      markClock (vhns);
    }

  vinstr = EM->getNode (vsig);
  if (NM->is_equation (vinstr))
    {
      if (NM->get_rightOperand (vinstr, &vhe))
	{
	  if (EM->IsCellExpr (vhe))
	    {			// vhe ::= (E when h) cell k or vhe ::= (e1 default e2) cell k
	      exp = EM->getWhenOrCellArg (vhe);
	      if (EM->getOperator (exp) == opextr)
		vhe = exp;
	    }
	  else
	    exp = EM->getWhenOrCellArg (vhe);
	  if (EM->getOperator (exp) == opmele)
	    markExpression (exp, m);
	  else
	    markExpression (vhe, m);
	}
    }
}

/**
   Local attribute management: it marks (m) an expression (vexp) and recursively its operands.
   It is used to mark the expressions not implicated in assertions (makeSynchrosNotImplicatedInAssertions()).
*/
void
assertions2events::markExpression (ptexpcal vexp, typeMark m)
{
  ptexpcal e1, e2;
  ptident id;
  ptinsth h;
  if (vexp != NILEXPCAL)
    if (!EM->isMarked (vexp, m))
      {
	EM->addMark (vexp, m);
	switch ((int) (EM->getOperatorClass (vexp)))
	  {

	  case variabl:
	    if (EM->IsSignal (vexp))
	      markSignal (vexp, m);
	    break;

	  case oper1:
	  case oper1tp:
	    markExpression (EM->getUnaryArg (vexp), m);
	    break;
	  case oper2:
	  case oper2tp:
	    markExpression (EM->getBinaryArgstAt (vexp, 1), m);
	    markExpression (EM->getBinaryArgstAt (vexp, 2), m);
	    break;
	  case oper3:
	  case oper3tp:
	    markExpression (EM->getTernaryArgstAt (vexp, 1), m);
	    markExpression (EM->getTernaryArgstAt (vexp, 2), m);
	    markExpression (EM->getTernaryArgstAt (vexp, 3), m);
	    break;
	  case call:		// function call
	    EM->getFunctionCall (vexp, &id, &e1, &e2);
	    markExpression (e1, m);
	    markExpression (e2, m);
	    break;

	  case constant:
	  case opertyp:
	    break;

	  case opereh:
	    e1 = EM->getWhenOrCellArg (vexp);
	    h = EM->getContributionClock (vexp);
	    if (!CKM->satisfiesInitClock (h))
	      markClock (h);
	    markExpression (e1, m);
	    break;
	  }
      }
}




/**
   It marks all the expressions and clocks implicated in assertions.
   Then the synchronisations not marked are solved.
   At the end, the marks are removed.   
 */
void
assertions2events::makeSynchrosNotImplicatedInAssertions ()
{
  ptinsth vh;
  ptlinsth list;
  typeMark m;
  ptexpcal e;

  // CKM->booleanReducing(3);
  // CKM->booleanReducing(1);

  list = CKM->getFormulaList (sub);
  if (!ptlinsth_isEmptyList (list))
    {
      m = getMark ();
      while (ptlinsth_get_next (&list, &vh))
	if (CKM->satisfiesIsAnAssumption (vh))
	  {
	    e = CKM->getConditionSub (vh);
	    markExpression (e, m);
	  }
      CKM->setModeSynchroSolving (pk_true);
      CKM->makeSynchrosUnmarked (pKbool2eventMark);
      CKM->setModeSynchroSolving (pk_false);

      // Cleaning.
      CKM->resetMarkAll (m);
      CKM->resetMarkAll (pKbool2eventMark);
      EM->resetMarkAll (m);
    }
}
