/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   It adds to the list of \b cycles, the entry point of the cycle is the top element of the global \b stack.
*/
  void
CircuitCalculus::cycleDetected ()
{
  ptStack currentStack;
  ptlPointer aCycle;
  ptElemStack curelem, cycleDetectedOn;
  boolean found;

  aCycle = NILLPOINTER;
  currentStack = stack;
  stackTop (currentStack, &cycleDetectedOn);
  aCycle = ptlPointer_add (aCycle, (pointer) cycleDetectedOn);
  found = pk_false;
  while (!found && stackPrevious (currentStack, &currentStack))
    {
      stackTop (currentStack, &curelem);
      aCycle = ptlPointer_add (aCycle, (pointer) curelem);
      found = (curelem == cycleDetectedOn);
    }
  cycles = ptlPointer_add (cycles, (pointer) aCycle);
}


/**
   @return "false" if a cycle is detected (the node is already in the stack), "true" otherwise.
   If a node (node) is already in the stack, a cycle is detected. It calls the cycleDetected() method
   to produce and add the cycle in the global list (cycles).

   Otherwise, the node is pushed and marked (vmarkpushed).
*/
boolean CircuitCalculus::push (ptINode node)
{
  stack = stackPush (stack, (ptElemStack) node);
  if (NM->isMarked (node, vmarkpushed))
    {
      cycleDetected ();
      return pk_false;
    }
  NM->addMark (node, vmarkpushed);
  return pk_true;
}

/**
   It pops a node (node), assumed to be the top element of the stack, from the stack.
*/
void
CircuitCalculus::pop (ptINode node)
{
  stack = stackPop (stack);
  NM->removeMark (node, vmarkpushed);
}


/**
   If a node (node) has not been explored, 
   - it is pushed (push() method): if it is in the stack, a cycle is detected.
   - its successors are visited (recursively)
   - then, it is popped, but it is marked as explored (markexplored mark).       
*/

_VIRTUAL_ void
CircuitCalculus::visitNode (ptINode node)
{
  ptINode imoti;

  if (!NM->isMarked (node, vmarkexplored))
    {
      if (push (node))		// else cycle.
	{
	  visitDependences (NM->get_successors (node));
	  visitClocks (NM->get_clockSuccessors (node));
	  NM->addMark (node, vmarkexplored);

	  if (NM->is_partialDefinition (node))
	    {
	      imoti = EM->getNode (NM->get_leftOperand (node));
	      visitNode (imoti);
	    }
	  else
	    imoti = node;
	  while (NM->get_inComposedNode (imoti, &imoti))
	    visitNode (imoti);
	}
      pop (node);
    }
}

/**
   @return "false" if a cycle is detected (the clock is already in the stack), "true" otherwise.
   If a clock (clk) is already in the stack, a cycle is detected. It calls the cycleDetected() method
   to produce and add the cycle in the global list (cycles).

   Otherwise, the clock is pushed and marked (vmarkpushed).
*/
boolean CircuitCalculus::push (ptinsth clk)
{
  stack = stackPush (stack, (ptElemStack) clk);
  if (CKM->satisfiesMarked (clk, vmarkpushed))
    {
      cycleDetected ();
      return pk_false;
    }
  CKM->addMark (clk, vmarkpushed);
  return pk_true;
}

/**
   It pops a clock (clk), assumed to be the top element of the stack, from the stack.
*/
void
CircuitCalculus::pop (ptinsth clk)
{
  stack = stackPop (stack);
  CKM->removeMark (clk, vmarkpushed);
}


/**
   It applies visitNode() for each element of a list of nodes (lnode).
*/
void
CircuitCalculus::visitNodes (ptlINode lnode)
{
  ptINode node;
  while (ptlINode_get_elem (&lnode, &node))
    visitNode (node);
}

/**
   For a dependence X->Y of a list of dependences (ld), it applies visitNode() on Y.
*/
void
CircuitCalculus::visitDependences (ptldependence ld)
{
  ptdependence aDep;
  ptinsth h1;
  ptINode vfrom, vto;

  while (ld->get_next (&aDep, &ld))
    {
      h1 = aDep->getDependence (&vfrom, &vto);
      visitNode (vto);
    }
}

/**
   It applies visitClock() for each element of a list of clocks (lh).
*/
void
CircuitCalculus::visitClocks (ptlinsth lh)
{
  ptinsth h;

  while (ptlinsth_get_next (&lh, &h))
    visitClock (h);
}

/**
   If the clock (clk) has not been explored, 
   - it is pushed (push() method): if it is in the stack, a cycle is detected.
   - its successors are visited (recursively)
   - then, it is popped but it is marked as explored (markexplored mark).       
*/
_VIRTUAL_ void
CircuitCalculus::visitClock (ptinsth clk)
{
  if (clk == CKM->getRepresentativeFormula (clk))
    if (!CKM->satisfiesMarked (clk, vmarkexplored))
      {
	if (push (clk))
	  {
	    visitClocks (CKM->getClockReferencesTo (clk));
	    visitNodes (CKM->getNodeReferencesTo (clk));
	    CKM->addMark (clk, vmarkexplored);
	  }
	pop (clk);
      }
}

/**
   When the graph contains cycles, at least one of the list  \b lireached,  \b lhreached is not empty.
   These lists give the entry points of the cycles. A node/clock is in the previous list if 
   it has been reached by at least one of its predecessor but not by all of them.

   This method produces the cycles in a list called \b cycles.
   For the computing, it uses a stack (\b stack) and some marks (\b vmarkpushed, \b vmarkexplored).\n
   It is called by the getCycles() method.

   <b> lireached, lhreached, cycles, stack,vmarkpushed, vmarkexplored </b>  are attribute members of the class.
*/
_VIRTUAL_ void
CircuitCalculus::visitReached ()
{
  ptlINode ln;
  ptINode node;
  ptlinsth lh;
  ptinsth h;

  vmarkpushed = getMark ();
  vmarkexplored = getMark ();
  ln = lireached;
  while (ptlINode_get_elem (&ln, &node))
    {
      stack = NILSTACK;
      visitNode (node);
    }
  lh = lhreached;
  while (ptlinsth_get_next (&lh, &h))
    {
      stack = NILSTACK;
      visitClock (h);
    }
  NM->resetMarkAll (vmarkexplored);
  CKM->resetMarkAll (vmarkexplored);
}
