/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PROTECTED_
/**
   A clock (h) is added to the list of nodes referenced during the greedy algorithm
   but not selected (some of its predecessors are not selected).
*/
  void
CircuitCalculus::incrnbEntryPoint (ptinsth h)
{
  if (lptent.find (h) == lptent.end ())
    lptent[h] = 1;
  else
    lptent[h]++;
}

/**
  A clock/node  has been selected by the greedy algorithm.
  The associative lptent is updated for a clock (h). The counter is decreased and if the number of treferences
  becomes equal to 0 the entry is removed from the table.
*/
void
CircuitCalculus::decrnbEntryPoint (ptinsth h)
{
  if (lptent.find (h) != lptent.end ())
    if (lptent[h] == 1)
      lptent.erase (h);
    else
      lptent[h]--;
}





/** It applies incrnbEntryPoint for each clock of a list of clocks (lh)
 */
void
CircuitCalculus::addlinsthfront (ptlinsth lh)
{
  ptinsth h;

  while (ptlinsth_get_next (&lh, &h))
    if (CKM->getRepresentativeFormula (h) == h)
      if (!PARC->isReached (h) && PARC->isClockReachable (h))
	{
	  incrnbEntryPoint (CKM->getFather (h));
	  PARC->setReached (h, pk_true);
	}
}


/** It applies incrnbEntryPoint for a node, if it verifies the criteria.
 */
void
CircuitCalculus::addinstrfront (ptINode i)
{
  if (!PARC->isReached (i) && PARC->isNodeReachable (i))
    {
      incrnbEntryPoint (NM->get_nodeClock (i));
      PARC->setReached (i, pk_true);
    }
}


/** It applies addinstrfront for each node of a list of nodes (li)
 */
void
CircuitCalculus::addlinstrfront (ptlINode li)
{
  ptINode i;
  while (ptlINode_get_elem (&li, &i))
    addinstrfront (i);
}



/**
   It applies, for each dependence from -> to of a list (ld) the addinstrfront method with "to" as parameter.
*/
void
CircuitCalculus::addlinstrfront (ptldependence ld)
{
  ptdependence aDep;
  ptinsth h1;
  ptINode vfrom, vto;
  while (ld->get_next (&aDep, &ld))
    {
      h1 = aDep->getDependence (&vfrom, &vto);
      addinstrfront (vto);
    }
}


/**
   For a clock (h) 
   - it adds as entry points (referenced node) all successors of the clock,
   - it removes the clock(h) from this set of entry points (it has been visited).
*/
void
CircuitCalculus::clockAction (ptinsth h)
{
  addlinsthfront (CKM->getClockReferencesTo (h));
  addlinstrfront (CKM->getNodeReferencesTo (h));
  addlinstrfront (ptgfd_get_nodes (CKM->getClockSubGraph (h)));
  if (PARC->isReached (h))
    {
      decrnbEntryPoint (CKM->getFather (h));
      PARC->setReached (h, pk_false);
    }
}

/**
   For a node (i)
   - it adds as entry points (referenced node) all successors of the node,
   - it removes the node from this set of entry points (it has been visited).
*/
void
CircuitCalculus::nodeAction (ptINode i)
{
  addlinsthfront (NM->get_clockSuccessors (i));
  addlinstrfront (NM->get_successors (i));

  if (PARC->isReached (i))
    {
      decrnbEntryPoint (NM->get_nodeClock (i));
      PARC->setReached (i, pk_false);
    }
}


/**
   It adds a node (i) in a list (lireached) if the node has not been visited 
   during the greedy algorithm.
*/
void
CircuitCalculus::searchLockedNode (ptINode i, ptlINode * lireached)
{
  if (PARC->isReached (i))
    ptlINode_insert_elem (lireached, i);
}


/**
   It applies searchLockedNode() for each node of a list of nodes (li).
*/
void
CircuitCalculus::searchLockedNodes (ptlINode li, ptlINode * lireached)
{
  ptINode i;
  ptlINode ln;
  ptlinsth lh;

  while (ptlINode_get_elem (&li, &i))
    {
      searchLockedNode (i, lireached);
      if (NM->is_subgraph (i))
	{
	  NM->get_subGraphDefinitions (i, &ln, &lh);
	  searchLockedNodes (ln, lireached);
	}
      else if (NM->is_array (i))
	{
	  NM->get_array_bodyDefinitions (i, &ln, &lh);
	  searchLockedNodes (ln, lireached);
	  NM->get_array_initDefinitions (i, &ln, &lh);
	  searchLockedNodes (ln, lireached);
	}
    }
}

/**
   It adds a clock (h) in a list (lhreached) if the clock has not been visited 
   during the greedy algorithm.
*/
void
CircuitCalculus::searchLockedClock (ptinsth h, ptlinsth * lhreached)
{
  boolean b;
  if (PARC->isReached (h) && PARC->isClockReachable (h))
    *lhreached = ptlinsth_sorted_insert_elem (*lhreached, h, &b);
}

/**
   It applies searchLockedClock() for each clock of a list of clocks (lh).
*/
void
CircuitCalculus::searchLockedClocks (ptlinsth lh, ptlinsth * lhreached)
{
  ptinsth h;
  while (ptlinsth_get_next (&lh, &h))
    searchLockedClock (h, lhreached);
}

/**
   @return "true" if there are some nodes/clocks not visited during the travel of the graph (ie graph contains cycles), 
   "false" otherwise.
   It classifies the list of entry points of cycles (lpent) in two lists: one for the signals (lireached),
   another for clocks (lhreached). \n
   lpent, lireached and lhreached are attribute members of the class.
*/
boolean CircuitCalculus::existLockedNodes ()
{
  ptlinsth
    lhfils;
  ptinsth
    h;
  ptlINode
    li;
  entrypointType::iterator i;


  lireached = NILLINODE;
  lhreached = NILLINSTH;
  i = lptent.begin ();
  while (i != lptent.end ())
    {
      if ((*i).second > 0)
	{
	  h = (*i).first;
	  lhfils = CKM->getChildren (h);
	  searchLockedClocks (lhfils, &lhreached);
	  li = ptgfd_get_nodes (CKM->getClockSubGraph (h));
	  searchLockedNodes (li, &lireached);
	}
      i++;
    }
  return ((lireached != NILLINODE) || (lhreached != NILLINSTH));
}

/**
   @return true if all the nodes of a subgraph (sg) are already visited, "false" otherwise.
*/
boolean CircuitCalculus::isVisitedSubGraph (ptINode sg)
{
  ptlinsth
    lh;
  ptlINode
    ln;
  ptinsth
    h;
  ptINode
    n;

  if (!NM->get_subGraphDefinitions (sg, &ln, &lh))
    return pk_false;
  while (ptlinsth_get_next (&lh, &h))
    if (!PARC->isClockAlreadySelected (h))
      return pk_false;
  while (ptlINode_get_elem (&ln, &n))
    if (!PARC->isNodeAlreadySelected (n))
      return pk_false;
  return pk_true;
}

/** Initial travel of the nodes assigned to the "initial clock". */
void
CircuitCalculus::graphTravellingInitialClock (void)
{
  ptINode sg, i;
  ptlsucc linst;
  ptlINode laux, lsg;
  boolean bContinue;

  if (DEBUG_MODE)
    printf (" graphTravellingInitialClock BEGIN \n");
  PARC->initTravelForInitialClock ();
  graphTravelling (NILINODE);
  if (AGRAPH->getNodesAtInitClock (&linst))
    {
      while (ptlINode_get_elem (&linst, &i))
	if (!PARC->isNodeAlreadySelected (i))
	  {
	    if (DEBUG_MODE)
	      printf (" Selected node %d \n", NM->get_num (i));
	    PARC->ptINodeAttr_setSelected (i);
	    PARC->signalSelectedNode (i);
	    nodeAction (i);
	  }
      // Added Oct 2014: If all the nodes of a subgraph have been selected, the sub-graph is marked selected.
      if (NM->getSubGraphs (&lsg))
	do
	  {
	    bContinue = pk_false;
	    laux = lsg;
	    while (ptlINode_get_elem (&lsg, &sg))
	      if (isVisitedSubGraph (sg))
		{
		  if (DEBUG_MODE)
		    printf (" Selected node %d \n", NM->get_num (sg));
		  PARC->ptINodeAttr_setSelected (sg);
		  PARC->signalSelectedNode (sg);
		  nodeAction (sg);
		  bContinue = pk_true;
		}
	  }
	while (bContinue);
    }
  if (DEBUG_MODE)
    printf (" graphTravellingInitialClock END \n");
}


/**
  @return "true" if the graph contains cycles, "false" otherwise.
  It initializes the functionality by creating a "traveller of graph" (PARC attribute member of the class) and 
  computes nodes assigned to the "initial clock" (graphTravellingInitialClock() method).
  This traveller object allows to apply to the graph a greedy algorithm that takes into account the dependences of the graph.
  When cycles are detected, the "locked" nodes (resp. clocks) are available in the list lireached (resp. lhreached).\n
  lireached and lhreached are attribute members of the class.
*/
boolean CircuitCalculus::circuitDetection ()
{
  boolean
    b;

  if (CKM->isAnEmptySystem ())
    return pk_false;

  PARC->initialize ();
  graphTravellingInitialClock ();
  if (PARC->initTravelForClock (AGRAPH->getRootClock ()))
    {
      graphTravelling (NILINODE);
      b = existLockedNodes ();
    }
  else
    {
      pK_messenger->error (errrecsystem);
      b = pk_true;
    }


  PARC->PRINTDEBUG (lireached, lhreached);

  lptent.clear ();		// reset of the table.
  PARC->finalize ();
  return b;
}


/**
   Travelling the nodes/clocks of the current graph in the composed node context (composedNode).
   It calls for each selected element the adequate method nodeProcessing() for a node,  and clockProcessing()
   for a clock.
*/
void
CircuitCalculus::graphTravelling (ptINode composedNode)
{
  ptINode node;
  ptinsth vclock;
  boolean bIsNode;

  if (composedNode != NILINODE)
    {
      PARC->openBlock (composedNode);
      if (DEBUG_MODE)
	printf ("Opening block node %d \n", NM->get_num (composedNode));
    }
  while (PARC->getNextNode (&bIsNode, &node, &vclock))
    if (bIsNode)
      {
	if (DEBUG_MODE)
	  printf (" Selected node %d \n", NM->get_num (node));
	nodeProcessing (node);
      }
    else
      {
	if (DEBUG_MODE)
	  printf (" Selected clock %s \n", ident2string (CKM->getName (vclock)));
	clockProcessing (vclock);
      }

  if (composedNode != NILINODE)
    {
      PARC->closeBlock ();
      if (DEBUG_MODE)
	printf ("Closing block node %d \n", NM->get_num (composedNode));
    }
}



/**
   For a node (node):
    - it signals to its successors that it has been selected by the greedy algorithm
    - it applies the specific calculus (nodeAction() method)

    Note that it recalls (recursively) the graphTravelling() method for composed nodes (array of processes, sub-graphs).
*/
_VIRTUAL_ void
CircuitCalculus::nodeProcessing (ptINode node)
{
  ptINode n;

  switch (NM->get_oper (node))
    {
    case opmoti:
      PARC->signalArraySelectedNode (node);
      graphTravelling (node);
      break;

    case opret:
      break;

    case opappe:
      break;
    case opdefpart:
    case opdefsig:
      break;

    case opsubgraph:
      PARC->signalSubGraphSelectedNode (node);
      graphTravelling (node);
      break;

    case opldefpart:
    case opbund:		// struct/bundle node.
    case opnodecall:		// Nothing to do: done on the outputs of the node call.
      break;

    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
  PARC->signalSelectedNode (node);
  nodeAction (node);
}

/**
   For a clock (vclock):
    - it signals to its successors that it has been selected by the greedy algorithm
    - it applies the specific calculus (clockAction method)
*/
_VIRTUAL_ void
CircuitCalculus::clockProcessing (ptinsth vclock)
{
  PARC->signalSelectedClock (vclock);
  clockAction (vclock);
}
