/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   @return the clock (dependence clock) assigned to a dependence iclos.
*/
  ptinsth ClosureCalculus::iclosureGetClock (pticlosure iclos)
{
  return CKM->getRepresentativeFormula (pticlosure_get_hdep (iclos));
}

/**
   It sets the dependence clock (h) to a dependence (iclos).
*/
void
ClosureCalculus::iclosureSetClock (pticlosure iclos, ptinsth h)
{
  pticlosure_set_hdep (iclos, CKM->getRepresentativeFormula (h));
}

/**
  @return the clock (dependence clock) assigned to a dependence hclos
*/
ptinsth
ClosureCalculus::hclosureGetClock (pthclosure hclos)
{
  return CKM->getRepresentativeFormula (pthclosure_get_hdep (hclos));
}

/**
   It sets the dependence clock (h) to a dependence (hclos).
*/
void
ClosureCalculus::hclosureSetClock (pthclosure hclos, ptinsth h)
{
  pthclosure_set_hdep (hclos, CKM->getRepresentativeFormula (h));
}

/**
  @return for a signal (exp) the list of the dependences of the form 
  <b> ( Y --> exp when H)  where Y is a signal or a clock </b> produced by the closure calculus
  and previously set by ptinsthAttr_set_lclos.
*/
ptlclosure ClosureCalculus::signal_get_lclosures (ptexpcal exp)
{
  ptlclosure
    lclos;

  if (EM->getAttribute (exp, CLOSUREATTRIBUTE, (attributeValue *) & lclos))
    return lclos;
  return NILLCLOSURE;
}

/**
  It assigns for a signal (exp) the list of the dependences of the form 
  <b> ( Y --> exp when H) where Y is a signal or a clock </b> produced by the closure calculus.
*/
void
ClosureCalculus::signal_set_lclosures (ptexpcal exp, ptlclosure lclos)
{
  EM->addAttribute (exp, CLOSUREATTRIBUTE, lclos);
}

/**
  It assigns for a clock (h) the list of the dependences of the form 
  <b> ( Y --> exp when H) where Y is a signal or a clock </b> produced by the closure calculus.
*/
void
ClosureCalculus::ptinsthAttr_set_lclos (ptinsth h, ptlclosure lclos)
{
  h = CKM->getRepresentativeFormula (h);
  CKM->addAttribute (h, CLOSUREATTRIBUTE, lclos);
}

/**
  @return for a clock (h) the list of the dependences produced by the closure calculus 
  and previously set by ptinsthAttr_set_lclos.
*/
ptlclosure ClosureCalculus::ptinsthAttr_get_lclos (ptinsth h)
{
  ptlclosure
    lclos;
  h = CKM->getRepresentativeFormula (h);
  if (CKM->getAttribute (h, CLOSUREATTRIBUTE, (attributeValue *) & lclos))
    return lclos;
  return NILLCLOSURE;
}

/**
   @return 
   - a list of dependences initialized with (exp -->exp when clockOf(exp)), if a signal (exp) 
   is element on which the closure calculus is asked (in lexpclos list),
   - an empty list, otherwise.
*/
ptlclosure ClosureCalculus::getexpcallclosureinit (ptexpcal exp)
{
  ptlclosure
    lclos;
  if (!ptlinstances_isin_exp (lexpclos, exp))
    return NILLCLOSURE;
  lclos = NILLCLOSURE;
  return catlclosureiclos (lclos, exp, EM->getClock (exp));
}

/**
   It initialises the closure calculus for a signal/constant/list of signals (exp) 
   The closure informations assigned to a clock are propapated to the closure informations of exp with 
   a clock (to be completed).
*/
ptlclosure ClosureCalculus::initClosure (ptexpcal exp, ptinsth h)
{
  ptlclosure
    lclos;
  ptlinstances
    ls,
    lsave;
  ptexpcal
    exp1;

  switch ((int) (EM->getOperatorClass (exp)))
    {
    case variabl:
      lclos = getexpcallclosureinit (exp);
      lclos = subfusionlclosure (lclos, ptinsthAttr_get_lclos (h), h);
      signal_set_lclosures (exp, lclos);
      break;

    case constant:
      lclos = NILLCLOSURE;
      break;

    default:
      ls = EM->getSignalsReferencedIn (exp);
      lsave = ls;
      lclos = NILLCLOSURE;
      while (ptlinstances_get_next (&ls, &exp1))
	lclos = fusionlclosure (lclos, initClosure (exp1, h));
      ptlinstances_free (lsave);
    }
  return lclos;
}

/** 
    Closure calculus for a clock (vclk) and the signals
    assigned to this clock and evaluated at the initial clock (constant signal).
*/
void
ClosureCalculus::clockClosure (ptinsth vclk)
{
  ptinsth hg, hd;
  ptlsucc liapred;
  ptINode iapred;
  ptexpcal expg, expd;
  ptlinstances lexp, l;
  ptlclosure lclos;

  lclos = NILLCLOSURE;
  if (ptlinsth_sorted_isin_elem (lhclos, vclk))
    lclos = catlclosurehclos (lclos, vclk, vclk);
  switch ((int) (CKM->getOperator (vclk)))
    {
    case infer:
      if (!CKM->getMinusArgs (vclk, &hg, &hd))
	CKM->getInfArgs (vclk, &hg, &hd);
      lclos = unionlclosure (lclos, ptinsthAttr_get_lclos (hg), vclk);
      lclos = unionlclosure (lclos, ptinsthAttr_get_lclos (hd), vclk);
      ptinsthAttr_set_lclos (vclk, lclos);
      break;
    case super:
      CKM->getSupArgs (vclk, &hg, &hd);
      lclos = unionlclosure (lclos, ptinsthAttr_get_lclos (hg), vclk);
      lclos = unionlclosure (lclos, ptinsthAttr_get_lclos (hd), vclk);
      ptinsthAttr_set_lclos (vclk, lclos);
      break;
    case sub:
    case nosub:
      // lexp = CKM->getSignalsOfCondition(vclk);
      lexp = EM->getSignalsReferencedIn (CKM->getConditionSub (vclk));
      l = lexp;
      while (ptlinstances_get_next (&lexp, &expg))
	lclos = unionlclosure (lclos, signal_get_lclosures (expg), vclk);
      ptinsthAttr_set_lclos (vclk, lclos);
      ptlinstances_free (l);
      break;
    case hcomp:
      hg = CKM->getFather (vclk);
      CKM->getUnaryArg (vclk, &hd);
      lclos = unionlclosure (lclos, ptinsthAttr_get_lclos (hg), vclk);
      lclos = unionlclosure (lclos, ptinsthAttr_get_lclos (hd), vclk);
      ptinsthAttr_set_lclos (vclk, lclos);
      break;
    case hnot:
    case vari:
    case predef:
      break;
    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }

  /* traitement des expressions (affectations) à l'horloge d'initialisation. */
  liapred = ptgfd_get_nodes (CKM->getClockSubGraph (vclk));
  while (ptlINode_get_elem (&liapred, &iapred))
    {
      if (NM->get_oper (iapred) == opdefsig)
	if (CKM->satisfiesInitClock (NM->get_contributionClockEquation (iapred)))
	  {
	    NM->get_equation_arg (iapred, &expg, &expd);
	    initClosure (expg, vclk);
	  }
    }
}

/**
   It propagates the closure dependences for the node n, initialized by lclos.
   
   \b Remind: The following rules are applied (closure calculus)
   - For n2 -> n1 with H as dependence clock, and n1 -> n with K as dependence clock
   then n2-> n with H*K as dependence clock.
   - if it exists two path from n1 to n, it is reduced to one path with the upper bound of the clock of each path.
*/
ptlclosure ClosureCalculus::nodeClosure (ptINode n, ptlclosure lclos)
{
  ptldependence
    lpred;
  ptdependence
    aDep;
  ptinsth
    hdep;
  ptlinstances
    lsig;
  ptexpcal
    vsig;
  ptINode
    n1,
    n2;

  lpred = NM->get_predecessors (n);
  while (lpred->get_next (&aDep, &lpred))
    {
      hdep = aDep->getDependence (&n1, &n2);
      lsig = NM->getInodelOutputFlows (n1);
      while (ptlinstances_get_next (&lsig, &vsig))
	lclos = inffusionlclosure (lclos, signal_get_lclosures (vsig), hdep);
    }
  return lclos;
}


/**
   Closure calculus for an equation (inst)
*/
void
ClosureCalculus::equClosure (ptINode inst)
{
  ptinsth h, hc, hdep;
  ptexpcal expg, expd;
  ptlclosure lclos;
  ptlinsth lh;

  h = NM->get_nodeClock (inst);
  hc = NM->get_contributionClockEquation (inst);
  NM->get_equation_arg (inst, &expg, &expd);
  lclos = initClosure (expg, hc);
  lclos = nodeClosure (inst, lclos);
  if (NM->get_rightOperand (inst, &expd))
    {
      lh = EM->getClocksReferencedIn (expd);
      while (ptlinsth_get_next (&lh, &hdep))
	lclos = subfusionlclosure (lclos, ptinsthAttr_get_lclos (hdep), hdep);
    }
  if (h != hc)
    lclos = subfusionlclosure (lclos, ptinsthAttr_get_lclos (h), h);
  signal_set_lclosures (expg, lclos);
}



/**
   Closure calculus for a delay (inst)
*/
void
ClosureCalculus::delayClosure (ptINode inst)
{
  ptlclosure lclos;
  ptexpcal expg, expd, nulexp, expret, vinit;
  ptinsth h;

  h = NM->get_nodeClock (inst);
  NM->get_delay_arg (inst, &expg, &expd, &expret, &nulexp, &vinit);
  lclos = getexpcallclosureinit (expg);
  if ((expret == NILEXPCAL) && EM->IsSignal (expd))
    lclos = fusionlclosure (lclos, signal_get_lclosures (expd));
  lclos = subfusionlclosure (lclos, ptinsthAttr_get_lclos (h), h);
  signal_set_lclosures (expg, lclos);
}

/**
   Closure calculus for a signal (exp) with n as node and h the clock of n.
*/
void
ClosureCalculus::signalClosure (ptexpcal exp, ptINode n, ptinsth h)
{
  ptlclosure lclos;
  lclos = initClosure (exp, h);
  lclos = nodeClosure (n, lclos);
  signal_set_lclosures (exp, lclos);
}


/** 
    Closure calculus for a procedure call (inst).
*/
void
ClosureCalculus::procCallClosure (ptINode inst)
{
  ptexpcal expg, expd, exp, par;
  ptinsth h;
  ptident nom;
  ptINode nc;

  if (NM->is_writeNode (inst))
    return;

  h = NM->get_nodeClock (inst);
  if (NM->is_outputNodeCall (inst, &nc))
    {
      exp = NM->get_io_arg (inst);
      signalClosure (exp, inst, h);
    }
  else if (NM->is_readNode (inst))
    {
      exp = NM->get_io_arg (inst);
      initClosure (exp, h);
    }
  else
    {
      NM->get_procCall_arg (inst, &nom, &par, &expg, &expd);
      while (expd != NILEXPCAL)	// outputs
	{
	  exp = EM->getNaryArgs (&expd);
	  signalClosure (exp, inst, h);
	}
    }
}


/**
   Closure calculus for an array of processes.
*/
void
ClosureCalculus::arrayClosure (ptINode arrayNode)
{
  ptexpcal vind;
  ptinsth h;
  ptINode n;

  n = NM->get_indexNode (arrayNode);
  vind = NM->get_index (n);
  h = NM->get_nodeClock (EM->getNode (vind));
  initClosure (vind, h);
}


/**
   @return a copy of a dependence structure (iclos)
*/
pticlosure ClosureCalculus::dupiclosure (pticlosure iclos)
{
  pticlosure
    vresult;
  pticlosure
    icloscour;

  if (iclos == NILICLOSURE)
    return NILICLOSURE;
  icloscour = pticlosure_RIEN_create (iclosureGetClock (iclos), pticlosure_get_from (iclos), NILICLOSURE);
  vresult = icloscour;
  iclos = pticlosure_get_next (iclos);
  while (iclos != NILICLOSURE)
    {
      pticlosure_set_next (icloscour, pticlosure_RIEN_create (iclosureGetClock (iclos), pticlosure_get_from (iclos), NILICLOSURE));
      icloscour = pticlosure_get_next (icloscour);
      iclos = pticlosure_get_next (iclos);
    }
  return vresult;
}



/**
   @return a copy of a dependence structure (hclos)
*/
pthclosure ClosureCalculus::duphclosure (pthclosure hclos)
{
  register pthclosure
    vresult;
  pthclosure
    hcloscour;

  if (hclos == NILHCLOSURE)
    return NILHCLOSURE;
  hcloscour = pthclosure_RIEN_create (hclosureGetClock (hclos), pthclosure_get_from (hclos), NILHCLOSURE);
  vresult = hcloscour;
  hclos = pthclosure_get_next (hclos);
  while (hclos != NILHCLOSURE)
    {
      pthclosure_set_next (hcloscour, pthclosure_RIEN_create (hclosureGetClock (hclos), pthclosure_get_from (hclos), NILHCLOSURE));
      hcloscour = pthclosure_get_next (hcloscour);
      hclos = pthclosure_get_next (hclos);
    }
  return vresult;
}


/**
   @return "true" if a new element is added in the list (deblit), "false" otherwise.
   It inserts if it is not the list (deblist) a new element (exp, NILINSTH).
   In all the cases, the element referencing  the signal (exp) is returned in curlist.
*/
boolean ClosureCalculus::inserticlos (pticlosure * deblist, pticlosure * curlist, ptexpcal exp)
{
  boolean
    bres;
  pticlosure
    nextlist;
  integer
    num,
    numcour;

  bres = pk_false;
  if ((*deblist) == NILICLOSURE)
    {
      *deblist = pticlosure_RIEN_create (NILINSTH, exp, NILICLOSURE);
      *curlist = *deblist;
      bres = pk_true;
    }
  else
    {
      num = EM->getNum (exp);
      numcour = EM->getNum (pticlosure_get_from (*deblist));
      if (num < numcour)
	{
	  *deblist = pticlosure_RIEN_create (NILINSTH, exp, *deblist);
	  *curlist = *deblist;
	  bres = pk_true;
	}
      else
	{
	  *curlist = *deblist;
	  while (numcour > 0)
	    if (num == numcour)
	      {
		numcour = 0;
		bres = pk_false;
	      }
	    else
	      {
		nextlist = pticlosure_get_next (*curlist);
		if (nextlist == NILICLOSURE)
		  {
		    nextlist = pticlosure_RIEN_create (NILINSTH, exp, NILICLOSURE);
		    pticlosure_set_next (*curlist, nextlist);
		    numcour = 0;
		    bres = pk_true;
		  }
		else
		  {
		    numcour = EM->getNum (pticlosure_get_from (nextlist));
		    if (numcour > num)
		      {
			nextlist = pticlosure_RIEN_create (NILINSTH, exp, nextlist);
			pticlosure_set_next (*curlist, nextlist);
			numcour = 0;
			bres = pk_true;
		      }
		  }
		*curlist = nextlist;
	      }
	}
    }
  return bres;
}


/**
   @return "true" if a new element is added in the list (deblit), "false" otherwise.
   It inserts if it is not the list (deblist) a new element (h, NILINSTH).
   In all the cases, the element referencing  the clock (h) is returned in curlist.
*/
boolean ClosureCalculus::inserthclos (pthclosure * deblist, pthclosure * curlist, ptinsth h)
{
  boolean
    bres;
  pthclosure
    nextlist;
  integer
    num,
    numcour;

  if ((*deblist) == NILHCLOSURE)
    {
      *deblist = pthclosure_RIEN_create (NILINSTH, h, NILHCLOSURE);
      *curlist = *deblist;
      bres = pk_true;
    }
  else
    {
      num = CKM->getNum (h);
      numcour = CKM->getNum (pthclosure_get_from (*deblist));
      if (num < numcour)
	{
	  *deblist = pthclosure_RIEN_create (NILINSTH, h, *deblist);
	  *curlist = *deblist;
	  bres = pk_true;
	}
      else
	{
	  *curlist = *deblist;
	  while (numcour > 0)
	    if (num == numcour)
	      {
		numcour = 0;
		bres = pk_false;
	      }
	    else
	      {
		nextlist = pthclosure_get_next (*curlist);
		if (nextlist == NILHCLOSURE)
		  {
		    nextlist = pthclosure_RIEN_create (NILINSTH, h, NILHCLOSURE);
		    pthclosure_set_next (*curlist, nextlist);
		    numcour = 0;
		    bres = pk_true;
		  }
		else
		  {
		    numcour = CKM->getNum (pthclosure_get_from (nextlist));
		    if (numcour > num)
		      {
			nextlist = pthclosure_RIEN_create (NILINSTH, h, nextlist);
			pthclosure_set_next (*curlist, nextlist);
			numcour = 0;
			bres = pk_true;
		      }
		  }
		*curlist = nextlist;
	      }
	}
    }
  return bres;
}

/**
   @return the list obtained by the merging of two lists (iclosmain, iclosinsert) with modification of iclosmain.
   Moreover, the dependences of iclosinsert are filtered by a clock (h).
*/
pticlosure ClosureCalculus::inffusioniclosure (pticlosure iclosmain, pticlosure iclosinsert, ptinsth h)
{
  pticlosure
    iclos;
  ptinsth
    hinf;

  if (iclosinsert != NILICLOSURE)
    {
      hinf = CKM->makeInf (iclosureGetClock (iclosinsert), h);
      if (inserticlos (&iclosmain, &iclos, pticlosure_get_from (pticlosure_removehead_elem (&iclosinsert))))
	iclosureSetClock (iclos, hinf);
      else
	iclosureSetClock (iclos, CKM->makeSup (iclosureGetClock (iclos), hinf));

      while (iclosinsert != NILICLOSURE)
	{
	  hinf = CKM->makeInf (iclosureGetClock (iclosinsert), h);
	  if (inserticlos (&iclos, &iclos, pticlosure_get_from (pticlosure_removehead_elem (&iclosinsert))))
	    iclosureSetClock (iclos, hinf);
	  else
	    iclosureSetClock (iclos, CKM->makeSup (iclosureGetClock (iclos), hinf));
	}
    }
  return iclosmain;
}



/**
   @return the list obtained by the merging of two lists (hclosmain, hclosinsert) with modification of hclosmain.
   Moreover, the dependences of hclosinsert are filtered by a clock (h).
*/
pthclosure ClosureCalculus::inffusionhclosure (pthclosure hclosmain, pthclosure hclosinsert, ptinsth h)
{
  pthclosure
    hclos;
  ptinsth
    hinf;

  if (hclosinsert != NILHCLOSURE)
    {
      hinf = CKM->makeInf (hclosureGetClock (hclosinsert), h);
      if (inserthclos (&hclosmain, &hclos, pthclosure_get_from (pthclosure_removehead_elem (&hclosinsert))))
	hclosureSetClock (hclos, hinf);
      else
	hclosureSetClock (hclos, CKM->makeSup (hclosureGetClock (hclos), hinf));
      while (hclosinsert != NILHCLOSURE)
	{
	  hinf = CKM->makeInf (hclosureGetClock (hclosinsert), h);
	  if (inserthclos (&hclos, &hclos, pthclosure_get_from (pthclosure_removehead_elem (&hclosinsert))))
	    hclosureSetClock (hclos, hinf);
	  else
	    hclosureSetClock (hclos, CKM->makeSup (hclosureGetClock (hclos), hinf));
	}
    }
  return hclosmain;
}

/**
   @return the list obtained by the merging of two lists (iclosmain, iclosinsert) with modification of iclosmain.
*/
pticlosure ClosureCalculus::fusioniclosure (pticlosure iclosmain, pticlosure iclosinsert)
{
  pticlosure
    iclos;
  ptinsth
    h;

  if (iclosinsert != NILICLOSURE)
    {
      h = iclosureGetClock (iclosinsert);
      if (inserticlos (&iclosmain, &iclos, pticlosure_get_from (pticlosure_removehead_elem (&iclosinsert))))
	iclosureSetClock (iclos, h);
      else
	iclosureSetClock (iclos, CKM->makeSup (iclosureGetClock (iclos), h));
      while (iclosinsert != NILICLOSURE)
	{
	  h = iclosureGetClock (iclosinsert);
	  if (inserticlos (&iclos, &iclos, pticlosure_get_from (pticlosure_removehead_elem (&iclosinsert))))
	    iclosureSetClock (iclos, h);
	  else
	    iclosureSetClock (iclos, CKM->makeSup (iclosureGetClock (iclos), h));
	}
    }
  return iclosmain;
}


/**
    @return the list obtained by the merging of two lists (hclosmain, hclosinsert) with modification of hclosmain.
*/
pthclosure ClosureCalculus::fusionhclosure (pthclosure hclosmain, pthclosure hclosinsert)
{
  pthclosure
    hclos;
  ptinsth
    h;

  if (hclosinsert != NILHCLOSURE)
    {
      h = hclosureGetClock (hclosinsert);
      if (inserthclos (&hclosmain, &hclos, pthclosure_get_from (pthclosure_removehead_elem (&hclosinsert))))
	hclosureSetClock (hclos, h);
      else
	hclosureSetClock (hclos, CKM->makeSup (hclosureGetClock (hclos), h));
      while (hclosinsert != NILHCLOSURE)
	{
	  h = hclosureGetClock (hclosinsert);
	  if (inserthclos (&hclos, &hclos, pthclosure_get_from (pthclosure_removehead_elem (&hclosinsert))))
	    hclosureSetClock (hclos, h);
	  else
	    hclosureSetClock (hclos, CKM->makeSup (hclosureGetClock (hclos), h));
	}
    }
  return hclosmain;
}

/**
   @return the list obtained by the merging of two lists (lclosmain, lclosinsert) with modification of lclosmain.
   It induces the merging of the two sublists (iclosure,hclosure).
   Here we known that if (x,K) is in lclosinsert, and (x,h) is in lclosmain, then h <= K.
*/
ptlclosure ClosureCalculus::unionlclosure (ptlclosure lclosmain, ptlclosure lclosinsert, ptinsth h)
{
  pticlosure
    iclosmain,
    iclos,
    iclosinsert;
  pthclosure
    hclosmain,
    hclos,
    hclosinsert;

  if (lclosinsert != NILLCLOSURE)
    {
      iclosinsert = ptlclosure_get_iclos (lclosinsert);
      iclosmain = ptlclosure_get_iclos (lclosmain);
      if (iclosinsert != NILICLOSURE)
	{
	  inserticlos (&iclosmain, &iclos, pticlosure_get_from (pticlosure_removehead_elem (&iclosinsert)));
	  iclosureSetClock (iclos, h);
	  while (iclosinsert != NILICLOSURE)
	    {
	      inserticlos (&iclos, &iclos, pticlosure_get_from (pticlosure_removehead_elem (&iclosinsert)));
	      iclosureSetClock (iclos, h);
	    }
	  ptlclosure_set_iclos (&lclosmain, iclosmain);
	}
      hclosinsert = ptlclosure_get_hclos (lclosinsert);
      hclosmain = ptlclosure_get_hclos (lclosmain);
      if (hclosinsert != NILHCLOSURE)
	{
	  inserthclos (&hclosmain, &hclos, pthclosure_get_from (pthclosure_removehead_elem (&hclosinsert)));
	  hclosureSetClock (hclos, h);
	  while (hclosinsert != NILHCLOSURE)
	    {
	      inserthclos (&hclos, &hclos, pthclosure_get_from (pthclosure_removehead_elem (&hclosinsert)));
	      hclosureSetClock (hclos, h);
	    }
	  ptlclosure_set_hclos (&lclosmain, hclosmain);
	}
    }
  return lclosmain;
}



/**
   @return the list obtained by the merging of two lists (lclosmain, lclosinsert) with modification of lclosmain.
   It induces the merging of the two sublists (iclosure,hclosure).
   The dependences of the inserted list (lclosinsert) are replaced by the clock operand (h).
*/
ptlclosure ClosureCalculus::subfusionlclosure (ptlclosure lclosmain, ptlclosure lclosinsert, ptinsth h)
{
  pticlosure
    iclosmain,
    iclos,
    iclosinsert;
  pthclosure
    hclosmain,
    hclos,
    hclosinsert;

  if (lclosinsert != NILLCLOSURE)
    {
      iclosinsert = ptlclosure_get_iclos (lclosinsert);
      iclosmain = ptlclosure_get_iclos (lclosmain);
      if (iclosinsert != NILICLOSURE)
	{
	  if (inserticlos (&iclosmain, &iclos, pticlosure_get_from (pticlosure_removehead_elem (&iclosinsert))))
	    iclosureSetClock (iclos, h);
	  else
	    iclosureSetClock (iclos, CKM->makeSup (iclosureGetClock (iclos), h));
	  while (iclosinsert != NILICLOSURE)
	    if (inserticlos (&iclos, &iclos, pticlosure_get_from (pticlosure_removehead_elem (&iclosinsert))))
	      iclosureSetClock (iclos, h);
	    else
	      iclosureSetClock (iclos, CKM->makeSup (iclosureGetClock (iclos), h));
	  ptlclosure_set_iclos (&lclosmain, iclosmain);
	}
      hclosinsert = ptlclosure_get_hclos (lclosinsert);
      hclosmain = ptlclosure_get_hclos (lclosmain);
      if (hclosinsert != NILHCLOSURE)
	{
	  if (inserthclos (&hclosmain, &hclos, pthclosure_get_from (pthclosure_removehead_elem (&hclosinsert))))
	    hclosureSetClock (hclos, h);
	  else
	    hclosureSetClock (hclos, CKM->makeSup (hclosureGetClock (hclos), h));
	  while (hclosinsert != NILHCLOSURE)
	    if (inserthclos (&hclos, &hclos, pthclosure_get_from (pthclosure_removehead_elem (&hclosinsert))))
	      hclosureSetClock (hclos, h);
	    else
	      hclosureSetClock (hclos, CKM->makeSup (hclosureGetClock (hclos), h));
	  ptlclosure_set_hclos (&lclosmain, hclosmain);
	}
    }
  return lclosmain;
}

/**
   @return the list obtained by the filtered-merging of two lists (lclosmain, lclosinsert) with modification of lclosmain.
   It induces the merging of the two sublists by the methods inffusioniclosure() and inffusionhclosure().
*/
ptlclosure ClosureCalculus::inffusionlclosure (ptlclosure lclosmain, ptlclosure lclosinsert, ptinsth h)
{
  pticlosure
    iclos;
  pthclosure
    hclos;

  if (lclosinsert == NILLCLOSURE)
    return lclosmain;
  iclos = inffusioniclosure (ptlclosure_get_iclos (lclosmain), ptlclosure_get_iclos (lclosinsert), h);
  hclos = inffusionhclosure (ptlclosure_get_hclos (lclosmain), ptlclosure_get_hclos (lclosinsert), h);
  ptlclosure_set_iclos (&lclosmain, iclos);
  ptlclosure_set_hclos (&lclosmain, hclos);
  return lclosmain;
}


/**
   @return the list obtained by the merging of two lists (lclosmain, lclosinsert) with modification of lclosmain.
   It induces the merging of the two sublists by the methods fusioniclosure() and fusionhclosure().
*/
ptlclosure ClosureCalculus::fusionlclosure (ptlclosure lclosmain, ptlclosure lclosinsert)
{
  pticlosure
    iclos;
  pthclosure
    hclos;

  if (lclosinsert == NILLCLOSURE)
    return lclosmain;
  iclos = fusioniclosure (ptlclosure_get_iclos (lclosmain), ptlclosure_get_iclos (lclosinsert));
  hclos = fusionhclosure (ptlclosure_get_hclos (lclosmain), ptlclosure_get_hclos (lclosinsert));
  ptlclosure_set_iclos (&lclosmain, iclos);
  ptlclosure_set_hclos (&lclosmain, hclos);
  return lclosmain;
}



/** It adds a new dependence (exp,hdep) in lclos.
 */
ptlclosure ClosureCalculus::catlclosureiclos (ptlclosure lclos, ptexpcal exp, ptinsth hdep)
{
  pticlosure
    iclos,
    curiclos,
    newiclos;

  iclos = ptlclosure_get_iclos (lclos);
  if (inserticlos (&iclos, &curiclos, exp))
    iclosureSetClock (curiclos, hdep);
  else
    {
      newiclos = pticlosure_RIEN_create (hdep, exp, pticlosure_get_next (curiclos));
      pticlosure_set_next (curiclos, newiclos);
    }
  ptlclosure_set_iclos (&lclos, iclos);
  return lclos;
}



/** It adds a new dependence (h, hdep) in lclos.
 */
ptlclosure ClosureCalculus::catlclosurehclos (ptlclosure lclos, ptinsth h, ptinsth hdep)
{
  pthclosure
    hclos,
    curhclos,
    newhclos;

  hclos = ptlclosure_get_hclos (lclos);
  if (inserthclos (&hclos, &curhclos, h))
    hclosureSetClock (curhclos, hdep);
  else
    {
      newhclos = pthclosure_RIEN_create (hdep, h, pthclosure_get_next (curhclos));
      pthclosure_set_next (curhclos, newhclos);
    }
  ptlclosure_set_hclos (&lclos, hclos);
  return lclos;
}

/**
   @return the list obtained by concataning element of lcloscat to lclosmain (modified). 
*/
ptlclosure ClosureCalculus::catlclosure (ptlclosure lclosmain, ptlclosure lcloscat)
{
  pticlosure
    iclosmain,
    curiclos,
    icloscat,
    tmpiclos;
  pthclosure
    hclosmain,
    curhclos,
    hcloscat,
    tmphclos;
  boolean
    insert;

  icloscat = ptlclosure_get_iclos (lcloscat);
  if (icloscat != NILICLOSURE)
    {
      iclosmain = ptlclosure_get_iclos (lclosmain);
      curiclos = iclosmain;
      while (icloscat != NILICLOSURE)
	{
	  if (iclosmain == curiclos)
	    insert = inserticlos (&iclosmain, &curiclos, pticlosure_get_from (icloscat));
	  else
	    insert = inserticlos (&curiclos, &curiclos, pticlosure_get_from (icloscat));
	  if (insert)
	    {
	      iclosureSetClock (curiclos, iclosureGetClock (icloscat));
	      icloscat = pticlosure_get_next (icloscat);
	    }
	  else
	    {
	      tmpiclos = pticlosure_get_next (curiclos);
	      pticlosure_set_next (curiclos, icloscat);
	      curiclos = icloscat;
	      icloscat = pticlosure_get_next (icloscat);
	      pticlosure_set_next (curiclos, tmpiclos);
	    }
	}
      ptlclosure_set_iclos (&lclosmain, iclosmain);
    }
  hcloscat = ptlclosure_get_hclos (lcloscat);
  if (hcloscat != NILHCLOSURE)
    {
      hclosmain = ptlclosure_get_hclos (lclosmain);
      curhclos = hclosmain;
      while (hcloscat != NILHCLOSURE)
	{
	  if (curhclos == hclosmain)
	    insert = inserthclos (&hclosmain, &curhclos, pthclosure_get_from (hcloscat));
	  else
	    insert = inserthclos (&curhclos, &curhclos, pthclosure_get_from (hcloscat));
	  if (insert)
	    {
	      hclosureSetClock (curhclos, hclosureGetClock (hcloscat));
	      hcloscat = pthclosure_get_next (hcloscat);
	    }
	  else
	    {
	      tmphclos = pthclosure_get_next (curhclos);
	      pthclosure_set_next (curhclos, hcloscat);
	      curhclos = hcloscat;
	      hcloscat = pthclosure_get_next (hcloscat);
	      pthclosure_set_next (curhclos, tmphclos);
	    }
	}
      ptlclosure_set_hclos (&lclosmain, hclosmain);
    }
  return lclosmain;
}


/**
   Closure calculus initialize.
*/
void
ClosureCalculus::closureInitialize ()
{
}

/** Parcours initial des instructions de clock hinit */
void
ClosureCalculus::graphTravellingInitialClock ()
{
  ptINode i;
  ptlsucc linst;

  PARC->initTravelForInitialClock ();
  graphTravelling (NILINODE);
  if (AGRAPH->getNodesAtInitClock (&linst))
    while (ptlINode_get_elem (&linst, &i))
      if (!PARC->isNodeAlreadySelected (i))
	{
	  PARC->ptINodeAttr_setSelected (i);
	  PARC->signalSelectedNode (i);
	  nodeAction (i);
	}
}

/**
   For a node (node):
    - it signals to its successors that it has been selected by the greedy algorithm
    - it applies the specific calculus ( equClosure(), delayClosure(), procCallClosure(), arrayClosure, ...)

    Note that
     - it recalls (recursively) the graphTravelling() method for composed nodes (array of processes, sub-graphs).
     - for a node call, the calculus is reported on the outputs of the node call.
*/
_VIRTUAL_ void
ClosureCalculus::nodeProcessing (ptINode node)
{
  //  ptINode n;

  switch (NM->get_oper (node))
    {
    case opdefpart:
    case opdefsig:
      equClosure (node);
      break;

    case opmoti:
      arrayClosure (node);
      PARC->signalArraySelectedNode (node);
      // nodeAction(n);
      graphTravelling (node);
      break;

    case opret:
      delayClosure (node);
      break;

    case opappe:
      procCallClosure (node);
      break;

    case opsubgraph:
      PARC->signalSubGraphSelectedNode (node);
      graphTravelling (node);
      break;

    case opldefpart:
    case opbund:
    case opnodecall:		// Nothing to do: done on the outputs of the node call.
      break;

    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
  PARC->signalSelectedNode (node);
  nodeAction (node);
}

/**
   For a clock (vclock):
    - it signals to its successors that it has been selected by the greedy algorithm
    - it applies the specific calculus (clockClosure method)
*/
_VIRTUAL_ void
ClosureCalculus::clockProcessing (ptinsth vclock)
{
  clockClosure (vclock);
  PARC->signalSelectedClock (vclock);
  clockAction (vclock);
}



/**
   It imports the result of the transitive closure into the graph.
   Informations stored in specific structures are translated into dependences.
   \see computeIOTransitiveClosure
*/
void
ClosureCalculus::importTransitiveClosure (ptlinstances lsig)
{
  ptlinsth lhclos;
  ptinsth h, hundef, hexp, hdep;
  ptlinstances lexpclos, levents;
  ptexpcal exp, expFrom;
  ptlclosure lclos;
  pticlosure iclos;

  lexpclos = lsig;
  while (ptlinstances_get_next (&lexpclos, &exp))
    if (EM->getAttribute (exp, CLOSUREATTRIBUTE, (attributeValue *) & lclos))
      {
	iclos = ptlclosure_get_iclos (lclos);
	while (iclos != NILICLOSURE)
	  {
	    expFrom = pticlosure_get_from (iclos);
	    if (exp != expFrom)
	      {
		hdep = CKM->getRepresentativeFormula (pticlosure_get_hdep (iclos));
		if (!CKM->satisfiesNullClock (hdep))
		  NM->addDependence (expFrom, exp, hdep, pk_true);
	      }
	    iclos = pticlosure_get_next (iclos);
	  }
      }
  // lhclos is NILLINSTH : dead code...
  lhclos = NILLINSTH;
  while (ptlinsth_get_next (&lhclos, &h))
    {
      h = CKM->getRepresentativeFormula (h);
      if (CKM->getAttribute (h, CLOSUREATTRIBUTE, (attributeValue *) & lclos))
	{
	  iclos = ptlclosure_get_iclos (lclos);
	  while (iclos != NILICLOSURE)
	    {
	      expFrom = pticlosure_get_from (iclos);
	      hdep = CKM->getRepresentativeFormula (pticlosure_get_hdep (iclos));
	      if (!CKM->satisfiesNullClock (hdep))
		NM->addDependence (EM->getNode (expFrom), h);
	      iclos = pticlosure_get_next (iclos);
	    }
	}
    }
}
