/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   It manages the nbMaxIndex attribute member that defines the max number of indices
   used at the same time in the program: nbMaxIndex=max(n,nbMaxIndex).
*/
  void
cGenerator::setMaxIndex (int n)
{
  if (n > nbMaxIndex)
    nbMaxIndex = n;
}

/**
   @return "true" if a node has a list of associated indexes (ls), "false" otherwise.
   The list is synthetized from operators provided for spatial processing.
*/
boolean
cGenerator::getNodeTravelIndex (ptINode node, ptlinstances * ls)
{
  ptexpcal vdef;

  if (NM->is_equation (node) || NM->is_partialDefinition (node))
    if (NM->get_rightOperand (node, &vdef))
      return EM->getTravelIndexes (vdef, ls);
  return pk_false;
}


/**
   @return the term of a type for a named type (t) or t otherwise.
*/
_VIRTUAL_ TypeType cGenerator::GetUnamedType (TypeType t)
{
  if (TypeTestNamedClass (t))
    return GetUnamedType (NamedTypeGetTerm (t));
  return t;
}

/**
   It generates the language key-word for static accessibility.
   Whithout effect when SEPARATE mode is activated or INTERNAL_BLOCK is activated.   
*/
_VIRTUAL_ void
cGenerator::genAccessibility ()
{
  if (!SEPARATE && !INTERNAL_BLOCK)
    TextForm->printkwstatic ();
}


/** 
    It signals that the generator begins a set of declarations.
 */
_VIRTUAL_ void
cGenerator::cdebdecl ()
{
  debutdcl = pk_true;
}


/** 
    It signals that the generator finalizes a set of declarations. 
*/
_VIRTUAL_ void
cGenerator::cfindecl ()
{
  if (!debutdcl)
    TextForm->printreleasedcl ();
}

/**  
     It generates
     - the "local" scope key-word (cGenerator::genAccessibility())
     - the type t without dimensions (cGenerator::genTypeBase())
     - store t as the current type (tprec attribute member)
 */
_VIRTUAL_ void
cGenerator::genTypeBeginDeclaration (TypeType t)
{
  cGenerator::genAccessibility ();
  cGenerator::genTypeBase (t);
  TextForm->mark ();
  tprec = t;
}

/** 
    Generation of the type t without its dimensions.
    - genTypeBase() when functionDeclaration mode is active.
    - genTypeBeginDeclaration() when the generator is initialized for a new declaration or the current type (tprec) is not t
    - the separator of elements for a list of declaration otherwise.
*/
_VIRTUAL_ void
cGenerator::gentype (TypeType t)
{

  if (functionDeclaration)
    genTypeBase (t);
  else if (debutdcl)
    {
      genTypeBeginDeclaration (t);
      debutdcl = pk_false;
    }
  else if (t != tprec)
    {
      TextForm->printreleasedcl ();
      genTypeBeginDeclaration (t);
    }
  else
    {
      TextForm->printComma ();
      TextForm->printSpace ();
    }
}


/**
   It generates the expression (aDim).
   Isolated to allow the specialisation of the generation of dimensions in the declarations.
   Some languages (C) uses the syntax such <em> int x[3] </em>, other ones (Java) uses <em> [3]int x </em>.
*/
_VIRTUAL_ void
cGenerator::genDimensionInDeclaration (ptexpcal aDim)
{
  genexpression (aDim, opnullary);
}

/** 
   Generation of the dimensions of a type (t).
   It calls 
   - genDimensionInDeclaration() for the generation of a dimension.
   - setMaxIndex() to produce the max-number of internal indices useful for array manipulations.
*/
_VIRTUAL_ void
cGenerator::genDclDimensions (TypeType TT)
{
  ptexpcal vdim, aDim;
  int n;

  if (TypeTestArrayClass (TT))
    {
      vdim = ArrayMatrixTypeGetDimLst (TT);
      while (EM->getNaryArgs (&vdim, &aDim))
	{
	  TextForm->printLeftSquareBracket ();
	  genDimensionInDeclaration (aDim);
	  TextForm->printRightSquareBracket ();
	}

      // computing the max of required indexes of the program.
      n = getNbDimFlattenedType (TT);
      setMaxIndex (n);
    }
}

/** 
   Generation of the declaration of the signal exp when it is a node mn of a delay tree.
*/
_VIRTUAL_ void
cGenerator::genDclSignalDelayTree (ptexpcal exp, ptmemnode mn)
{
  TypeType t1;

  t1 = EM->getType (exp);
  gentype (t1);
  ecrnomrootinddelay (mn);
  TextForm->printLeftSquareBracket ();
  TextForm->printint (ptmemnode_get_size (mn));
  TextForm->printRightSquareBracket ();
  genDclDimensions (EM->getType (ptmemnode_get_signal (mn)));
}


/**
  Generation of the declaration of the signal when it is a node NM of a delay tree.
  In this case, the signal is seen as an index on an area.
*/
_VIRTUAL_ void
cGenerator::genDclSignalAsIndex (ptmemnode mn)
{
  gentype (get_IntegerType ());
  ecrnominddelay (mn);
}

/**
  Generation of the declaration of the identifier id whose type is t.
*/
_VIRTUAL_ void
cGenerator::genIdentDeclaration (ptident id, TypeType t)
{
  gentype (t);
  TextForm->printident (id);
  genDclDimensions (t);
}

/**
  Generation of the declaration of the signal exp when it is not a node of a delay tree.
*/
_VIRTUAL_ void
cGenerator::genSignalDeclaration (ptexpcal exp)
{
  genIdentDeclaration (EM->getName (exp), EM->getType (exp));
}


/**
   @return "true" if a signal (exp) is a node of a delay tree, "false" otherwise.
   This method declares an index for the signal (genDclSignalAsIndex()) and declares also the array when it is
   the root of a delay tree (genDclSignalDelayTree()).
   
   \note  when a signal is a node of a delay tree, it will be replaced for the code generation by an access to 
   a global area (array that allows to represent all the signals of the same tree of delays) using a 
   indice to this array.
*/
_VIRTUAL_ boolean
cGenerator::genDclTreeDelaySignal (ptexpcal exp)
{
  ptmemnode mn;
  ptlmemnode lmn;

  exp = EM->getCloneClassHead (exp);
  if (!ptexpcal_satisfies_node_tree_delay (exp, &mn))
    return pk_false;
  lmn = ptmemnode_get_equiv (mn);
  do
    {
      genDclSignalAsIndex (mn);
      if (ptmemnode_satisfies_root (mn))
	genDclSignalDelayTree (exp, mn);
      if (!ptlmemnode_getNaryArgs (&lmn, &mn))
	mn = NILMEMNODE;
    }
  while (mn != NILMEMNODE);
  if (copyForExternal (exp))
    genSignalDeclaration (exp);
  return pk_true;
}

/**
   Generation of the declaration of a signal (exp) when it is useful.
   It calls genDclTreeDelaySignal() method for a node of a delay tree, genSignalDeclaration() otherwise.
*/
_VIRTUAL_ void
cGenerator::genDclSignal (ptexpcal exp)
{
  ptINode n;
  ptlinstances ls;
  ptexpcal s;

  n = EM->getNode (exp);
  if (EM->IsOutputFlow (exp) || EM->IsInputFlow (exp) || mustBeGenerated (n))
    {
      if (getNodeTravelIndex (n, &ls))
	setMaxIndex (ptlinstances_get_size (ls));
      if (!genDclTreeDelaySignal (exp))
	genSignalDeclaration (exp);
      // Struct/bundle --> indexes in tree delays
      if (NM->is_lfields (n))
	{
	  ls = NM->get_lfields_arg (n);
	  while (ptlinstances_get_next (&ls, &s))
	    genDclTreeDelaySignal (s);
	}
    }
}

/**
   Generation of the declarations of the signals of a list of signals (lsig).
*/
_VIRTUAL_ void
cGenerator::genDclSignals (ptlinstances lsig)
{
  ptexpcal sig;

  cdebdecl ();
  while (ptlinstances_get_next (&lsig, &sig))
    genDclSignal (sig);
  cfindecl ();
}

/**
   Generation of the declaration of a constant (exp) when it is useful 
   and when its type is not a predefined one 
   (for a constant  with a predefined type, a constant is replaced by its value).
*/
_VIRTUAL_ void
cGenerator::genConstantDeclaration (ptexpcal exp)
{
  if (mustBeGenerated (exp))
    if (!EM->IsExternalConstant (exp))
      if (!TypeTestPredef (EM->getUnamedType (exp)))
	genSignalDeclaration (exp);
}

/** 
    Parameters are not generated here for C generator, but in an external file (cGeneratorExtern_* file).
    This method  must be kept for specialisation.
*/
_VIRTUAL_ void
cGenerator::genParameters ()
{
}

/** 
    Generation of nb elements of a list of declarations of signals (lst).
    When nb is negative value, all the elements of the list are declared.
 */
_VIRTUAL_ void
cGenerator::genIndexLstDeclaration (ptlinstances lst, int nb)
{
  ptexpcal exp;
  int i;

  if (nb != 0)
    {
      cdebdecl ();
      if (nb < 0)
	while (ptlinstances_get_next (&lst, &exp))
	  genSignalDeclaration (exp);
      else
	{
	  i = 0;
	  while ((i < nb) && ptlinstances_get_next (&lst, &exp))
	    {
	      genSignalDeclaration (exp);
	      i++;
	    }
	}
      cfindecl ();
    }
}

/**
   Generation of the declarations of the non-substituted parameters and indexes of the graph. 
*/
_VIRTUAL_ void
cGenerator::genDclParametersAndIndexes ()
{
  ptlinstances liste;
  ptexpcal exp;

  cdebdecl ();
  genParameters ();
  liste = Constants;
  while (ptlinstances_get_next (&liste, &exp))
    genConstantDeclaration (exp);
  cfindecl ();

  genIndexLstDeclaration (Indexes, -1);

}

/** 
    Generation of identifiers for previous value of state variables. 
*/
_VIRTUAL_ void
cGenerator::genStateVariables ()
{
  ptlinstances lexpr;
  ptexpcal exp;
  ptident idPrevious;

  if (!EM->getExprList (opvvarin, &lexpr))
    return;
  // lexpr = Sets of x expressions
  cdebdecl ();
  while (ptlinstances_get_next (&lexpr, &exp))
    if (EM->getAttribute (exp, AttributeNameGencode, (attributeValue *) & idPrevious))
      genIdentDeclaration (idPrevious, EM->getType (EM->getUnaryArg (exp)));

  cfindecl ();
}

/**
   It declares a Signal external object (ie separated compiling) taking into account the implementation language (vlang).
   externalObj gives the type of the declared object and vdata the name of the object.
*/
_VIRTUAL_ void
cGenerator::genExternalSignalObjectDeclaration (ptrepstr externalObj, ptrepstr vdata, ntrees vlang)
{
  ptrepstr vl;

  vl = getStrLanguageExternalDef (vlang);
  genAccessibility ();
  if (vl == nilstr)
    TextForm->printCExternalSignalObjectDeclaration (externalObj, vdata);
  else if (streq (vl, string2repstr ("C")))
    TextForm->printCExternalSignalObjectDeclaration (externalObj, vdata);
  else if (streq (vl, string2repstr ("C++")))
    TextForm->printCExternalSignalObjectDeclaration (externalObj, vdata);
  else
    {
      pK_messenger->errorStrStr (errInterfaceToLanguage, string2repstr ("C"), vl);
      TextForm->printkwnotimpl ();
      TextForm->printCExternalSignalObjectDeclaration (externalObj, vdata);
      setSomeProblem ();
    }
}

/** 
    Generation of identifiers for external Signal objects (Black boxes, Grey Boxes).
    For a reference (node) to an external model, the compiler has associated to it a pragma 
    specifying the name of the entry point and a number to have multi references to the same model in an application.
    These informations are given by getClassAndObjectExternalCall().

    It calls genExternalSignalObjectDeclaration().
    
*/
_VIRTUAL_ void
cGenerator::genExternalSignalObjects (ptINode node, ptlrepstr * lstr)
{
  ptrepstr externalObj, vdata;
  ptpragma valPragma;
  ntrees vlang;

  if (isAnnotatedGreyBox (node, &valPragma) || isAnnotatedBlackBoxGraph (node, &valPragma))
    {
      externalObj = getClassAndObjectExternalCall (valPragma, &vdata);
      if (ptlMark_insert_elem (lstr, vdata))
	{
	  if (!NM->getAttribute (node, EXTERN_LANGUAGE_IMPL, (attributeValue *) & vlang))
	    SET_NILTREE (vlang);
	  genExternalSignalObjectDeclaration (externalObj, vdata, vlang);
	}
    }

}

/** 
    This method generates the declaration of external Signal objects of the graph. 
    These declarations are induced by the use of models (Black boxes, Grey Boxes) that have been 
    separated compiled.
*/
_VIRTUAL_ void
cGenerator::genExternalSignalObjects ()
{
  ptlINode lNodeCall;
  ptINode aNodeCall;
  ptlrepstr lstr;

  /* Declaration of external Black and Grey boxes */
  lstr = NILLREPSTR;
  if (NM->getNodeCalls (&lNodeCall))
    while (ptlINode_get_elem (&lNodeCall, &aNodeCall))
      genExternalSignalObjects (aNodeCall, &lstr);

  if (NM->getProcCalls (&lNodeCall))
    while (ptlINode_get_elem (&lNodeCall, &aNodeCall))
      genExternalSignalObjects (aNodeCall, &lstr);
}

/**
   It generates the declaration of the "micro step" clock (reserved name) if 
   the graph is an endochronous with an internal clock root.
*/
_VIRTUAL_ void
cGenerator::genMicrostepDecl ()
{
  if (bEndochronousWithInternalClockRoot)
    {
      genIdentDeclaration (getMicroStepName (), BooleanType);
      cfindecl ();
    }
}

/**
   Generation of the declarations of the signals and non-substituted parameters of the graph
   (body file).
*/
_VIRTUAL_ void
cGenerator::gendeclarationsBody ()
{
  ptrepstr savePrefix;

  savePrefix = getprefix ();
  setPrefix (nilstr);
  genComment (comm_dclpar);
  genDclParametersAndIndexes ();
  genComment (comm_dclent);
  genDclSignals (Inputs);
  genComment (comm_dclsort);
  genDclSignals (Outputs);
  genComment (comm_dclloc);
  genDclSignals (Locals);
  genStateVariables ();
  genExternalSignalObjects ();
  genInternalIndexes ();
  genAffineSampleCounterDeclaration ();
  genMicrostepDecl ();
  TextForm->printLine ();
  genCmpOperators ();
  setPrefix (savePrefix);
}

/**
   Generation of the identifier _id_ where id denotes the identifier of the signal (sig). 
   When star is "true", the indirection operator is generated before the identifier.
   When id cannot be fixed from sig, a name is created with n (sig is the n-th formal paramater).
*/
_VIRTUAL_ void
cGenerator::genFormalParameterIdent (ptexpcal sig, boolean star, int n)
{
  ptident id;

  if (star)
    TextForm->printstar ();
  if (EM->getOperatorClass (sig) == variabl)
    {
      id = EM->getName (sig);
      //      if (EM->isField(sig)) xxxx;
      TextForm->printidentunderscore (id);
    }
  else				// an unnamed constant,...
    TextForm->printident (makeIdent_ident_int (string2ident ("fp"), n));
}

/**
   Generation of the declaration of a signal (sig) as a formal parameter. 
   @param sig (in) a signal.
   @param isoutput (in) must be "true" for an output parameter, "false" otherwise.
   @param bprintIdent (in) when it is "true", a name is printed.
*/
_VIRTUAL_ void
cGenerator::genDclFormalParameter (ptexpcal sig, boolean isoutput, boolean bprintIdent, int n)
{
  boolean star;
  TypeType t;
  t = EM->getType (sig);
  gentype (t);
  star = isoutput && (EM->getNumberOfDimensions (sig) == 0);
  if (bprintIdent)
    genFormalParameterIdent (sig, star, n);
  else if (star)
    TextForm->printstar ();
  genDclDimensions (t);
}

/**
   Generation of the declaration of the signals of a list (lsig) as formal parameters.
   It calls genDclFormalParameter() for each of element.
   @param lsig (in) a list of signals.
   @param isoutput (in) must be "true" for an output parameter, "false" otherwise.
   @param bprintIdent (in) when it is "true", a name is printed.
   @param bvirg (in) if "true" the last declaration is completed by the separator of a parameter list.
   @param n (in) the current number of parameters. It is used for creating identifiers (if needed).
   @return n + the size of the list.
*/
_VIRTUAL_ int
cGenerator::genDclFormalParameters (ptexpcal lsig, boolean bvirg, boolean isoutput, boolean bprintIdent, int n)
{
  ptexpcal sig;

  if (lsig == NILEXPCAL)
    return n;
  while (lsig != NILEXPCAL)
    {
      EM->getBinaryArgs (lsig, &sig, &lsig);
      n++;
      genDclFormalParameter (sig, isoutput, bprintIdent, n);
      if (lsig != NILEXPCAL)
	TextForm->printComma ();
    }
  if (bvirg)
    TextForm->printComma ();
  return n;
}

/** 
   Generation of the hearder of the global/external procedure (or function if asFunc) name with inputs and outputs as formal parameters.
   When ClusterEntry is "false", the name of the formal parameters are the names of the signals of the inputs/outputs lists, otherwise 
   for a signal X of inputs/outputs, the name of the formal parameter is _X_.
*/
_VIRTUAL_ boolean
cGenerator::genExternalFunction (ptident nom, ptexpcal linputs, ptexpcal loutputs, boolean asFunc, boolean ClusterEntry)
{
  ptexpcal saveloutputs, s;
  TypeType t;
  boolean b;
  int n;

  if (TextForm->isCastFunction (nom))
    return pk_false;

  functionDeclaration = pk_true;
  b = pk_false;
  saveloutputs = loutputs;
  if (asFunc)
    {
      if (loutputs != NILEXPCAL)
	{
	  s = EM->getNaryArgs (&loutputs);
	  if (loutputs == NILEXPCAL)
	    {
	      t = EM->getType (s);
	      b = (EM->getNumberOfDimensions (s) == 0);
	    }
	}
    }

  if (b)
    TextForm->printdclreftypeextfunc (nom, t);
  else
    {
      TextForm->printdclrefvoidextCfunc (nom);
      loutputs = saveloutputs;
    }

  if (ClusterEntry && SEPARATE)
    {
      TextForm->printClassDataArea ();
      if ((linputs != NILEXPCAL) || (loutputs != NILEXPCAL))
	TextForm->printComma ();
    }

  n = genDclFormalParameters (linputs, (loutputs != NILEXPCAL), pk_false, ClusterEntry, 0);
  n = genDclFormalParameters (loutputs, pk_false, pk_true, ClusterEntry, n);

  functionDeclaration = pk_false;
  // printf("DEBUG  exiting cGenerator::genExternalFunction *\n"); fflush(stdout);
  return pk_true;
}

/** Generation of the hearder of the global/external procedure or function.
   @return true if the header is generated (false for cast function)
   @param name (in) name of the procedure/function
   @param asFunc (in) at pk_true for a function, pk_false for a procedure.
   @param inputs (in) input formal parameters.
   @param outputs (in) output formal parameters.
   @param ClusterEntry (in) at pk_false, the name of formal parameters are the names of the signals of the inputs/outputs lists,
           otherwise for a signal X of inputs/outputs, the name of the formal parameter is _X_.
   @param declaration (in) at pk_true, the header is completed by the separator of declarations.
*/
_VIRTUAL_ boolean
cGenerator::genEntryPoint (ptident name, ptexpcal inputs, ptexpcal outputs, boolean asFunc, boolean ClusterEntry, boolean declaration)
{
  boolean b;

  b = genExternalFunction (name, inputs, outputs, asFunc, ClusterEntry);
  if (declaration)
    TextForm->printRPSemiColonLine ();
  else
    TextForm->printRPline ();
  return b;
}


/** Generation of the hearder of the procedure name with inputs and outputs as parameters.
   This procedure is an external one if bExternal is "true" or a static one othewise. 
   The header is completed by the separator of declarations if declaration.
*/
_VIRTUAL_ void
cGenerator::genEntryPointDeclaration (ptident name, ptexpcal inputs, ptexpcal outputs, boolean declaration, boolean bExternal)
{
  if (bExternal)
    genEntryPoint (name, inputs, outputs, pk_false, pk_true, declaration);
  else
    {
      TextForm->printstaticfunction (name, !declaration, pk_false);
      if (declaration)
	TextForm->printSemiColonLine ();
    }
}

/**
   It generates the list of internal indices. 
   The number of indices (nbMaxIndex) has been produced by the calls to setMaxIndex().
   - It produces a list of identifiers ( newIndex()) 
   - and calls genIndexLstDeclaration() for the generation of this list.
*/
_VIRTUAL_ void
cGenerator::genInternalIndexes ()
{
  int i;
  ptlinstances lexpr;
  ptexpcal exp;

  if (EM->getExprList (oppcar, &lexpr))
    while (ptlinstances_get_next (&lexpr, &exp))
      {
	// Cartesian product needs a specific variable (for the definition of the selector)
	i = EM->getSize (EM->getUnaryArg (exp)) + 1;
	setMaxIndex (i);
      }
  for (i = 1; i <= nbMaxIndex; i++)
    InternalIndexLst = ptlinstances_insert (InternalIndexLst, newIndex ());

  genIndexLstDeclaration (InternalIndexLst, -1);
}

/** 
    Generation of n elements of a list of declarations of signals (l) using genIndexLstDeclaration() 
    as a local declaration in a specific context:
    - INTERNAL_BLOCK is set to "true" to inhibit the generation  of the accessibility attribute (push)
    - genIndexLstDeclaration() is called to generate the n elements of l
    - INTERNAL_BLOCK is reset to its previous value (pop).
*/
_VIRTUAL_ void
cGenerator::genLocalIndexLstDeclaration (ptlinstances l, int n)
{
  boolean b;
  if (n != 0)
    {
      b = INTERNAL_BLOCK;
      INTERNAL_BLOCK = pk_true;
      genIndexLstDeclaration (l, n);
      INTERNAL_BLOCK = b;
    }
}

/** 
    It assigns a counter for each affine_sample undersampling.
    The identifier is assigned as an attribute of the output signal of the node 
    B := affine_sample{pahse, d} (C).
*/
_VIRTUAL_ void
cGenerator::genAffineSampleCounterDeclaration ()
{
  ptlINode ln;
  ptINode node;
  ptexpcal linputs, loutputs, par, boolClock;
  ptident nom, id, idCounter;

  if (NM->getProcCalls (&ln))
    {
      cdebdecl ();
      while (ptlINode_get_elem (&ln, &node))
	if (mustBeGenerated (node))
	  if (NM->is_intrinsicfunc (node))
	    {
	      NM->get_procCall_arg (node, &nom, &par, &linputs, &loutputs);
	      if (isSampleId (nom))	// xxxxxx
		{
		  lAffineSample = ptlINode_append_head (node, lAffineSample);
		  EM->getNaryArgs (&loutputs, &boolClock);
		  id = EM->getName (boolClock);
		  idCounter = nameManager->mkIdentName (id);
		  genIdentDeclaration (idCounter, get_IntegerType ());
		  EM->addAttribute (boolClock, affineSampleAttribute, (attributeValue) idCounter);
		}
	    }
      cfindecl ();
    }
}

_VIRTUAL_ void
cGenerator::genUserPragma (ptlpragmas lpragmas, ptrepstr selectedPragma)
{
  ptpragma aPragma;
  ptrepstr nPragma;
  ntrees vt;
  while (ptlpragmas_get_next (&lpragmas, &aPragma))
    {
      nPragma = getPragmaName (aPragma);
      if (streq (nPragma, selectedPragma))
	{
	  vt = getPragmaStatement (aPragma);
	  genPragmaCode (NILIDENT, vt, NILEXPCAL, NILEXPCAL, NILEXPCAL);
	}
    }
}

/**
   It generates the user imports using the PRAGMA_CODEGEN_<LANGUAGE>_IMPORT pragma.
*/
_VIRTUAL_ void
cGenerator::genUserImports ()
{
  ptlINode lNode;
  ptINode aNode;
  ptDCG aGraph;
  ptexpcal OrderedParams, OrderedInputs, OrderedOutputs;

  if (NM->getNodeCalls (&lNode))
    while (ptlINode_get_elem (&lNode, &aNode))
      {
	NM->get_nodeCall_arg (aNode, &aGraph, &OrderedParams, &OrderedInputs, &OrderedOutputs);
	genUserPragma (aGraph->getPragmas (), PragmaImportLanguage);
      }

  if (NM->getProcCalls (&lNode))
    while (ptlINode_get_elem (&lNode, &aNode))
      genUserPragma (NM->getPragmas (aNode), PragmaImportLanguage);
}


/**
   It generates the user imports using the PRAGMA_CODEGEN_<LANGUAGE>_DECLARATION pragma.
*/
_VIRTUAL_ void
cGenerator::genUserDeclarations ()
{
  ptlINode lNode;
  ptINode aNode;
  ptDCG aGraph;
  ptexpcal OrderedParams, OrderedInputs, OrderedOutputs;

  if (NM->getNodeCalls (&lNode))
    while (ptlINode_get_elem (&lNode, &aNode))
      {
	NM->get_nodeCall_arg (aNode, &aGraph, &OrderedParams, &OrderedInputs, &OrderedOutputs);
	genUserPragma (aGraph->getPragmas (), PragmaDeclareLanguage);
      }

  if (NM->getProcCalls (&lNode))
    while (ptlINode_get_elem (&lNode, &aNode))
      genUserPragma (NM->getPragmas (aNode), PragmaDeclareLanguage);
}
