/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_

/** 
    @return "true" if the node node must be generated, "false" otherwise.
    @param node (in) can be annotated by one of the attribute  : GencodeMark, substituteGenCodeMark
*/
_VIRTUAL_ boolean cGenerator::mustBeGenerated (ptINode node)
{
  return NM->mustBeGenerated(node);
}

/** 
    @return "true" if the opernad signal (vsig) must be generated, "false" otherwise.
    @param node (in) can be annotated by one of the attribute  : GencodeMark, substituteGenCodeMark
*/
_VIRTUAL_ boolean cGenerator::mustBeGenerated (ptexpcal vsig)
{
  return mustBeGenerated (EM->getNode(vsig));
}

/**
   It generates the contextual prefix (genPrefix()) before the identifier of a reference to the identifier if bref (printidentref()).
*/
void
cGenerator::printPrefixedIdentRef (ptident id, boolean bref, boolean bfield)
{
  genPrefix ();
  TextForm->printidentref (id, bref, bfield);
}


/**
   It generates the contextual prefix (genPrefix()) before the name of an assumed identifier (s) (printident()).
*/
void
cGenerator::printPrefixedSignalIdentifier (ptexpcal s)
{
  printPrefixedIdentRef (EM->getName (s), pk_false, EM->IsField (s));
}

/**
   @return "true" if expg and expd are the same variable (ie x = x or x = x[]) when the current operation is a sequential definition (next), "false" oherwise.
   The current operation is managed by a attribute member of the class (contextOp).

   Example: T := (T$ init [1,2,3]) next K:A \n
   If T and T$ are in the same memory (by optimisation), it is not necessary to generate the copy of T in T.
*/
boolean cGenerator::IsOptimizedSequentialDefinition (ptexpcal expg, ptexpcal expd)
{
  noperators ope;
  if (contextOp == opsequ)
    return pk_false;
  if (expg == expd)
    return pk_true;
  ope = EM->getOperator (expd);
  if (ope == opsdec)
    return (EM->getUnaryArg (expd) == expg);
  return pk_false;
}

/**
   It finalizes a statement with n back indentations.
*/
void
cGenerator::endStatement (int n, boolean addSemiColon)
{
  if (!sameLine)
    printptvlinebi (n);
  else
    {
      TextForm->backindent (n);
      if (addSemiColon)
	TextForm->printSemiColon ();
      TextForm->printSpace ();
    }
}


/**
   @return "true" if the node i defines a constant signal, "false" otherwise.
   @param i (in) a node.
   @param expd (out) the expression of the constant.
*/
_VIRTUAL_ boolean
cGenerator::isNodeAtInitialClock (ptINode i, ptexpcal * expd)
{
  // Doesn't reference textform object
  ptexpcal expg;

  if (!NM->is_equation (i))
    return pk_false;
  if (!CKM->satisfiesInitClock (NM->get_contributionClockEquation (i)))
    return pk_false;
  NM->get_equation_arg (i, &expg, expd);
  return (EM->getNumberOfDimensions (expg) == 0);
}


//----- Node ASSIGNMENT,  X:= E


/** Code generation for an equation (not a delay), such that X:=E.
    It realizes a switch depending on the operator of E. 
    It calls one of the following methods: 
    - genCounterModulo() for X := E1 count M
    - genAssignDefaultCell(), for X := (E1 default E2) cell C
    - genAssignDefault(),  for X := (E1 default E2)
    - genAssignCell(),  for X := E1 cell C
    - genAssignSynchronous(),  for X := f(E1,....,E2) f is a synchronous operator.

    @param node (in) assumed to be an equation.
*/
_VIRTUAL_ void
cGenerator::genEquNode (ptINode node)
{
  ptexpcal expg, expd;
  boolean bcell;
  ptinsth h;
  noperators ope;

  expg = NM->get_leftOperand (node);
  genIncrIndexDelayTreeNode (expg);

  NM->get_rightOperand (node, &expd);
  bcell = EM->IsCellExpr (expd);
  expd = EM->getWhenOrCellArg (expd);
  if (EM->IsWhenOrCellExpr (expd))
    expd = EM->getWhenOrCellArg (expd);
  if (!IsOptimizedSequentialDefinition (expg, expd))
    {
      if (bcell)
	h = NM->get_nodeClock (node);
      else
	{
	  h = NM->get_contributionClockEquation (node);
	  if (CKM->satisfiesInitClock (h))
	    h = NM->get_nodeClock (node);
	}
      // genBlockOpeningOf(h);
      ope = EM->getOperator (expd);
      if (ope == opnombm)
	{
	  genBlockOpeningOf (h);
	  genCounterModulo (expg, EM->getBinaryArgstAt (expd, 2));
	}
      else if (ope == opmele)
	{
	  genBlockOpeningOf (h);
	  if (bcell)
	    genAssignDefaultCell (expg, expd);
	  else
	    {
	      // b = (h != currentClock);  xxxx
	      genAssignDefault (expg, expd);
	    }
	}
      else if (bcell)
	{
	  h = NM->get_contributionClockEquation (node);
	  genBlockOpeningOf (h);
	  genAssignCell (h, expg, expd);
	}
      else
	{
	  genBlockOpeningOf (h);
	  genAssignSynchronous (expg, expd);
	}
    }
  genAssignEquivalence (expg);
}

/** 
    Code generation for an equation of a counter modulo.
*/
_VIRTUAL_ void
cGenerator::genCounterModulo (ptexpcal expg, ptexpcal vmodulo)
{
  gensig (expg, pk_false);
  TextForm->printassign ();
  TextForm->printLP ();
  gensig (expg, pk_false);
  TextForm->printplusOne ();
  TextForm->printRP ();
  TextForm->printkwmod ();
  genexpression (vmodulo, opnullary);
  printptvlinebi (0);
}

/**
   It generates the assignment expg := expd when the operator of expd is an vectorial definition such that
   { i from 1 to n, j ..., k...}:expression (see Signal Grammar for more details).
  
   @param expg (in) left expression of the assignment
   @param expd (in) right expression of the assignment ;  the operator of expd must be "opensvec".
   @param ajustdim (in) when it is "true", the first dimension of expd is ignored (this case happends only for the generation of the initialisations of delays).
*/
_VIRTUAL_ void
cGenerator::genAssignVectSet (ptexpcal expg, ptexpcal expd, boolean ajustdim)
{
  ptexpcal lparc, savelparc, expr;
  ptexpcal ident, borne, binf, bsup, pas;
  int i;

  EM->getBinaryArgs (expd, &lparc, &expr);
  if (ajustdim && (EM->getOperator (expr) != opelevec))
    expd = EM->getNaryArgs (&lparc);

  savelparc = lparc;
  i = 0;
  while (lparc != NILEXPCAL)
    {
      EM->getBinaryArgs (EM->getNaryArgs (&lparc), &ident, &borne);
      EM->getTernaryArgs (borne, &binf, &bsup, &pas);
      if (genLoopBegin (ident, binf, bsup, pas))
	i = i + 1;
    }
  if (EM->getOperator (expr) == opelevec)
    genAssignVectElem (expg, expr, ajustdim);
  else
    genAssignElemParc (expg, savelparc, expr);
  TextForm->backindent (i);
}

/**
   @param expg (in) left part of a definition.
   @param expd (in) right part of a definition, assumed to be a list of vectorial definitions
   - sequential enumeration given by an iterative enumeration (see genAssignVectSet())
   - sequential enumeration given by elements (see genAssignVectElem()).
   @param ajustDim (in) when it is "true", the first dimension of expd is ignored 
   (this case happends only for the generation of the initialisations of delays).
*/
_VIRTUAL_ void
cGenerator::genListVectorElem (ptexpcal expg, ptexpcal expd, boolean ajustDim)
{
  ptexpcal vexp;

  while (expd != NILEXPCAL)
    {
      vexp = EM->getNaryArgs (&expd);
      switch ((int) (EM->getOperator (vexp)))
	{
	case opelevec:
	  genAssignVectElem (expg, vexp, ajustDim);
	  break;
	case opensvec:
	  genAssignVectSet (expg, vexp, ajustDim);
	  break;
	default:
	  pK_messenger->Caseerror (currentLine, currentFile);
	}
    }
}


/** 
    It generates the assignment expg := expg where expg denotes a multiplication operation on matrix/vectors.
*/
_VIRTUAL_ void
cGenerator::genMatrixMult (ptexpcal expg, ptexpcal expd)
{
  TypeType TT, t;
  ptexpcal exp1, exp2, d1, d2, d11, d12, d21, d22, ins, outs;
  int n1, n2;
  ptrepstr codeop, strSuff;
  ptident name;

  t = EM->getUnamedType (expg);
  if (TypeTestArrayClass (t))
    TT = ArrayMatrixTypeGetTerm (t);
  else
    TT = t;
  EM->getBinaryArgs (expd, &exp1, &exp2);
  d1 = ArrayMatrixTypeGetDimLst (EM->getType (exp1));
  d2 = ArrayMatrixTypeGetDimLst (EM->getType (exp2));
  n1 = EM->getSize (d1);
  n2 = EM->getSize (d2);
  d11 = EM->getNaryArgs (&d1);
  d21 = EM->getNaryArgs (&d2);
  outs = EM->mkNaryExp (oplesign, expg, NILEXPCAL, NILBACKANNOTATION);
  ins = NILEXPCAL;
  name = getIdentNil ();
  if (TypeTestComplexNumber (TT))
    strSuff = string2repstr ("Complex");
  else
    strSuff = nilstr;
  if (n1 == 2)
    {
      d12 = EM->getNaryArgs (&d1);
      if (n2 == 2)
	{
	  ins = EM->mkNaryExp (oplesign, EM->getNaryArgs (&d2), ins, NILBACKANNOTATION);
	  ins = EM->mkNaryExp (oplesign, d12, ins, NILBACKANNOTATION);
	  ins = EM->mkNaryExp (oplesign, d11, ins, NILBACKANNOTATION);
	  ins = EM->mkNaryExp (oplesign, exp2, ins, NILBACKANNOTATION);
	  ins = EM->mkNaryExp (oplesign, exp1, ins, NILBACKANNOTATION);
	  strconc2str (string2repstr ("MMMult"), strSuff, pk_false, &codeop);
	  genPragmaCode (name, getEntryCodeOp (tableOpCode, codeop), NILEXPCAL, ins, outs);
	}
      else
	{
	  ins = EM->mkNaryExp (oplesign, d12, ins, NILBACKANNOTATION);
	  ins = EM->mkNaryExp (oplesign, d11, ins, NILBACKANNOTATION);
	  ins = EM->mkNaryExp (oplesign, exp2, ins, NILBACKANNOTATION);
	  ins = EM->mkNaryExp (oplesign, exp1, ins, NILBACKANNOTATION);
	  strconc2str (string2repstr ("MVMult"), strSuff, pk_false, &codeop);
	  genPragmaCode (name, getEntryCodeOp (tableOpCode, codeop), NILEXPCAL, ins, outs);
	}
    }
  else if (n2 == 2)
    {
      d22 = EM->getNaryArgs (&d2);
      ins = EM->mkNaryExp (oplesign, d22, ins, NILBACKANNOTATION);
      ins = EM->mkNaryExp (oplesign, d21, ins, NILBACKANNOTATION);
      ins = EM->mkNaryExp (oplesign, exp2, ins, NILBACKANNOTATION);
      ins = EM->mkNaryExp (oplesign, exp1, ins, NILBACKANNOTATION);
      codeop = string2repstr ("VMMult");
      strconc2str (codeop, strSuff, pk_false, &codeop);
      genPragmaCode (name, getEntryCodeOp (tableOpCode, codeop), NILEXPCAL, ins, outs);
    }
  else
    {
      ins = EM->mkNaryExp (oplesign, d11, ins, NILBACKANNOTATION);
      ins = EM->mkNaryExp (oplesign, exp2, ins, NILBACKANNOTATION);
      ins = EM->mkNaryExp (oplesign, exp1, ins, NILBACKANNOTATION);
      codeop = string2repstr ("VVMult");
      strconc2str (codeop, strSuff, pk_false, &codeop);
      genPragmaCode (name, getEntryCodeOp (tableOpCode, codeop), NILEXPCAL, ins, outs);
    }
}



/** It generates the assignment expg := expg where expg denotes a matrix transposition expression.
 */
_VIRTUAL_ void
cGenerator::genMatrixTransposition (ptexpcal expg, ptexpcal expd)
{
  ptexpcal exp1, d1, d11, d12, outs, ins;
  TypeType t;

  exp1 = EM->getUnaryArg (expd);
  t = EM->getUnamedType (expg);
  d1 = ArrayMatrixTypeGetDimLst (t);
  d11 = EM->getNaryArgs (&d1);
  d12 = EM->getNaryArgs (&d1);
  ins = EM->mkNaryExp (oplesign, d12, NILEXPCAL, NILBACKANNOTATION);
  ins = EM->mkNaryExp (oplesign, d11, ins, NILBACKANNOTATION);
  ins = EM->mkNaryExp (oplesign, exp1, ins, NILBACKANNOTATION);
  outs = EM->mkNaryExp (oplesign, expg, NILEXPCAL, NILBACKANNOTATION);
  genPragmaCode (getIdentNil (), getEntryCodeOp (tableOpCode, string2repstr ("optran")), NILEXPCAL, ins, outs);
}



/** It generates the assignment expg := expd for expd a synchronous expression.
    It calls one of the following methods when IsOptimizedSequentialDefinition() is not applied, depending on the operator of expd:
    - genListVectorElem()  for a list of vectorial definitions
    - genConcatenate() for a concatenation.
    - genMatrixMult() for a multiplication of matrix/vectors.
    - genMatrixTransposition() for a transposition of matrix.
    - genEnumVector()   for a definition of an index (E1..E2 step E3).
    - genRepeat() for a repetition expression ( E1 |* E2 )
    - genNext()  for a next operation (E1 next E2), used mainly for redefinition of elements of arrays.
    - genAssign() otherwise.
   
    @param expg (in) left expression of the assignment.
    @param expd (in) right expression of the assignment.
*/
_VIRTUAL_ void
cGenerator::genAssignSynchronous (ptexpcal expg, ptexpcal expd)
{
  ptexpcal x;
  if (!EM->isNILEXPCAL (expd))
    {
      if (EM->getAttribute (expg, CONSTANT_ATTRIBUTE, (attributeValue *) & x))
	expd = x;
      if (!IsOptimizedSequentialDefinition (expg, expd))
	switch ((int) EM->getOperator (expd))
	  {
	  case oplelevec:
	    genListVectorElem (expg, expd, pk_false);
	    break;
	  case opconc:
	    genConcatenate (expg, expd);
	    break;
	  case opmmul:
	    genMatrixMult (expg, expd);
	    break;
	  case optran:
	    genMatrixTransposition (expg, expd);
	    break;
	  case openum:
	    genEnumVector (expg, expd);
	    break;
	  case oprepe:
	    genRepeat (expg, expd);
	    break;
	  case opsequ:
	    genNext (expg, expd);
	    break;
	  default:
	    genAssign (expg, expd);
	  }
      else
	endStatement (0, pk_true);
    }
}

/**
   When expg 
   - is a list (U1,..., Un), vdef is a list <<I1,... In>> , it call genVectorCartesianProduct() for each element (Uk, Ik)
   - otherwise, it generates expg[ vexprIndex ] = e, where e is the associated index with vdef. This index is known using the 
   IteratorAttribute attribute; it is set in genLoopForList()).
*/
_VIRTUAL_ void
cGenerator::genVectorCartesianProduct (ptexpcal expg, ptexpcal vdef, ptexpcal vexprIndex)
{
  ptexpcal exp1, exp2;

  if (EM->satisfiesList (expg))
    while (EM->getNaryArgs (&expg, &exp1))
      {
	EM->getNaryArgs (&vdef, &exp2);
	genVectorCartesianProduct (exp1, exp2, vexprIndex);
      }
  else
    {
      exp2 = EM->mkNaryExp (opleflot, vexprIndex, NILEXPCAL, NILBACKANNOTATION);
      genelemdim (expg, exp2, NILEXPCAL);
      TextForm->printassign ();
      vdef = getExpressionToGenerate (vdef);
      if (isOptimizedIndex (vdef))
	{
	  if (EM->getAttribute (vdef, IteratorAttribute, (attributeValue *) & exp1))
	    vdef = exp1;
	  genexpression (vdef, opnullary);
	}
      else if (EM->getAttribute (vdef, IteratorAttribute, (attributeValue *) & exp1))
	{
	  genPrefixedIdentifier (vdef);
	  printIndexAccess (exp1);
	}
      else
	genexpression (vdef, opnullary);	// error 
      TextForm->printSemiColonLine ();
    }
}

/**
   Code generation expg:=expd for expd a cartesian product expression (<<I1,... In>> operator).
   <<I1,... In>> defines a tuple of n vectors (II1,....IIn). These vectors II1,....IIn are such that
   the tuples obtained by their elements of same index describe successively the successive values of the elements
   I1,...,In in embedded loops such that the most external one enumerates the values of I1 and the most internal one enumerates the
   values of In.
   
   It calls 
   - genLoopForList() for the opening of the loops.
   - genDefIndex()
   - genVectorCartesianProduct()

   \code
   Example: (I,J):= <<0..4, 0..2>> produces the following (pseudo) code
   for L = 0, 4;
   for U = 0, 2;
   { 
   I[ 3L + U] = L;
   J[ 3L + U] = U;
   }
   \endcode
*/
_VIRTUAL_ void
cGenerator::genCartesianProduct (ptexpcal expg, ptexpcal expd)
{
  ptexpcal vdef, vexprIndex;
  int n;

  setOptimizedIndexes (expd);
  vdef = EM->getUnaryArg (expd);
  n = genLoopForList (vdef);
  TextForm->printblockopening ();
  vexprIndex = genDefIndex (vdef);
  if (!EM->satisfiesList (expg))
    EM->getNaryArgs (&vdef, &vdef);
  genVectorCartesianProduct (expg, vdef, vexprIndex);
  TextForm->printblockclosing ();
  printptvlinebi (n);
  genCloseIteration (n);
  resetOptimizedIndexes (expd);
}

/**  
     Code generation expg:=expd for expd an expression E1 next E2, used mainly for redefinition of elements of arrays.
     It generates expg:= E1 followed by expg:= E2.
*/
_VIRTUAL_ void
cGenerator::genNext (ptexpcal expg, ptexpcal expd)
{
  ptexpcal exp1, exp2;
  noperators s;
  boolean b;

  TextForm->printLBracket ();
  EM->getBinaryArgs (expd, &exp1, &exp2);
  s = contextOp;
  if (s == opsequ)
    b = pk_true;
  else
    b = !IsOptimizedSequentialDefinition (expg, exp1);
  if (b)
    genAssignSynchronous (expg, exp1);
  contextOp = opsequ;
  genAssignSynchronous (expg, exp2);
  contextOp = s;
  TextForm->printRBracket ();
}


/**  
     Code generation expg:=expd for expd an repetition expression E1 |* E2 (E1 repeated E2 times).
     e1 is recopied e2 times.
*/
_VIRTUAL_ void
cGenerator::genRepeat (ptexpcal expg, ptexpcal expd)
{
  ptexpcal exp1, exp2;
  EM->getBinaryArgs (expd, &exp1, &exp2);
  genAssign (expg, exp1);
}

/**
   Code generation expg:=expd for expd an expression E1..E2 step E3 (definition of an index).
*/
_VIRTUAL_ void
cGenerator::genEnumVector (ptexpcal expg, ptexpcal expd)
{
  ptexpcal outs, ins, lb, ub, step, d1, d11;
  ptrepstr codeop;
  long int n;
  boolean optimized;
  TypeType t;

  EM->getTernaryArgs (expd, &lb, &ub, &step);
  t = EM->getType (expd);
  if (TypeTestNamedClass (t))
    t = NamedTypeGetTerm (t);
  d1 = ArrayMatrixTypeGetDimLst (t);
  d11 = EM->getNaryArgs (&d1);
  optimized = pk_false;
  if (getValue (lb, &n))
    if (n == 0)
      if (getValue (step, &n))
	optimized = (n == 1);

  ins = EM->mkNaryExp (oplesign, d11, NILEXPCAL, NILBACKANNOTATION);
  if (optimized)
    codeop = string2repstr ("openumOpt");
  else
    {
      ins = EM->mkNaryExp (oplesign, step, ins, NILBACKANNOTATION);
      ins = EM->mkNaryExp (oplesign, ub, ins, NILBACKANNOTATION);
      ins = EM->mkNaryExp (oplesign, lb, ins, NILBACKANNOTATION);
      codeop = string2repstr ("openum");
    }
  outs = EM->mkNaryExp (oplesign, expg, NILEXPCAL, NILBACKANNOTATION);
  genPragmaCode (getIdentNil (), getEntryCodeOp (tableOpCode, codeop), NILEXPCAL, ins, outs);

}


/**
   It generates expg := expd  for expd a synchronous expression of type String. It is generalized for
   duplication of string variable of structures: expg.e1....eN = expd.f1.....fm where 
   e1,...,en is a reference to a field of expg (lfields)
   f1,...,fm is a reference to a field of expd (lfieldsdef)
   For C code, 
   It generates  expg:=expd for strcopy( x.e1....eN, y.f1.....fm).
   It calls genFieldsList() for field generation.

*/
_VIRTUAL_ void
cGenerator::genAssignString (ptexpcal expg, ptexpcal lfields, ptexpcal expd, ptexpcal lfieldsdef)
{
  TextForm->printkwstrcopy ();
  TextForm->printLP ();
  gensig (expg, pk_false);
  genFieldsList (lfields);
  TextForm->printComma ();
  genexpression (expd, opnullary);
  genFieldsList (lfieldsdef);
  TextForm->printRP ();
}

/** 
    It generates the assignment expg := expd  for expd a synchronous expression of type "string" (or array of string)
    @return "true" if operation is done, "false" otherwise.
    @param expg (in) left expression of the assignment.
    @param expd (in) right expression of the assignment; synchronous expression.
*/
_VIRTUAL_ boolean
cGenerator::isGenAssignString (ptexpcal expg, ptexpcal expd)
{
  TypeType t;

  t = EM->getUnamedType (expg);
  if (TypeTestArrayClass (t))
    t = ArrayMatrixTypeGetTerm (t);
  if (t != StringType)
    return pk_false;
  genAssignString (expg, NILEXPCAL, expd, NILEXPCAL);
  return pk_true;
}


/**
   It generates the expansion of the lpart:= rpart in lpart.e1....eN = rpart.f1.....fm  for each field of lpart 
   for a signal of structure type, where
   - e1....eN are given by lfields (optional) 
   - f1.....fm are given by lfieldsdef (optional).
   - t is the the type of lpart.e1....eN.
   It calls genAssignString() for the string type.
*/
_VIRTUAL_ void
cGenerator::genFieldsTupleStruct (ptexpcal lpart, ptexpcal lpartAdded, TypeType t, ptexpcal rpart, ptexpcal rpartAdded)
{
  ptexpcal lpartAdded1, rpartAdded1, e, velem;
  NameList NL;
  TypeList fieldsType;
  TypeType TT;
  TypeSubList NSubLstType;
  boolean Goon, b;
  NameSubList NSubLst;
  ptrepstr fieldName;
  int n;

  // urgence : copie de tableau (copie globale non admise dans C++ ansi , application EXPRESSO )
  if (TypeTestNamedClass (t))
    t = NamedTypeGetTerm (t);
  if (TypeTestStructureClass (t))
    {
      NL = StructureTypeGetNameLst (t);	// field Name
      NSubLst = NameListCast2SubList (NL);
      fieldsType = StructureTypeGetTypeLst (t);
      NSubLstType = TypeListCast2SubList (fieldsType);
      Goon = pk_true;
      while (Goon)
	{
	  fieldName = NameSubListGetHead (NSubLst);
	  e = EM->mkConstantExp (opnom, fieldName);
	  TT = TypeSubListGetHead (NSubLstType);
	  rpartAdded1 = rpartAdded;	// defaut
	  if (EM->satisfiesList (rpart))
	    EM->getNaryArgs (&rpart, &velem);
	  else
	    {
	      velem = rpart;
	      // type(velem)= TT for a structure reduced to one field.
	      if (EM->getUnamedType (velem) != TT)
		rpartAdded1 = EM->mkNaryExp (oplesign, e, rpartAdded, NILBACKANNOTATION);
	    }
	  lpartAdded1 = EM->mkNaryExp (oplesign, e, lpartAdded, NILBACKANNOTATION);
	  genFieldsTupleStruct (lpart, lpartAdded1, TT, velem, rpartAdded1);
	  if (NameSubListTestLast (NSubLst))
	    Goon = pk_false;
	  else
	    {
	      NSubLst = NameSubListGetTail (NSubLst);
	      NSubLstType = TypeSubListGetTail (NSubLstType);
	    }
	}
    }
  else if (TypeTestArrayClass (t))
    {
      n = genOpenLoopForDimLst (ArrayMatrixTypeGetDimLst (t));
      if (getExprLastIndexes (n, &e))
	{
	  lpartAdded = EM->mkNaryExp (oplesign, e, lpartAdded, NILBACKANNOTATION);
	  rpartAdded = EM->mkNaryExp (oplesign, e, rpartAdded, NILBACKANNOTATION);
	}
      TextForm->printblockopening ();
      genFieldsTupleStruct (lpart, lpartAdded, ArrayMatrixTypeGetTerm (t), rpart, rpartAdded);
      TextForm->printBlockClosingBackIndent ();
      genCloseIteration (n);
    }
  else
    {
      b = extensionInhibit;
      extensionInhibit = pk_true;
      if (t == StringType)
	genAssignString (lpart, lpartAdded, rpart, rpartAdded);
      else
	{
	  gensig (lpart, pk_false);
	  genFieldsList (lpartAdded);
	  TextForm->printassign ();
	  genexpression (rpart, opnullary);
	  genFieldsList (rpartAdded);
	}
      TextForm->printSemiColonLine ();
      extensionInhibit = b;
    }
}

/** 
    It generates the assignment le := re  for re a list of expressions (Tuple)
    @param le (in) left expression of the assignment.
    @param re (in) right expression of the assignment; 
    @return "true" if re is a list of expression, "false" otherwise.
*/
_VIRTUAL_ boolean
cGenerator::genTupleStruct (ptexpcal le, ptexpcal re)
{
  TypeType t;
  ptexpcal velem;

  if (isScalarTypeAndConstant (re, &velem))
    re = velem;
  if (!isStructureExpr (le, &t))
    return pk_false;
  TextForm->printLBracket ();
  TextForm->printLine ();
  genFieldsTupleStruct (le, NILEXPCAL, t, re, NILEXPCAL);
  TextForm->printRBracket ();
  TextForm->printLine ();

  return pk_true;
}

/** 
    It generates the assignment le := re  for re a list of expressions (Tuple)
    @param le (in) left expression of the assignment.
    @param re (in) right expression of the assignment;
    @param n (in) number of array-blocks opened for the extension of the operators.
    @return "true" if re is a list of expression, "false" otherwise.
*/
_VIRTUAL_ boolean
cGenerator::genArrayTupleStruct (ptexpcal le, ptexpcal re, int n)
{
  TypeType t;
  ptexpcal le1, e;

  t = EM->getUnamedType (le);
  if (!TypeTestArrayClass (t))
    return pk_false;
  t = ArrayMatrixTypeGetTerm (t);
  if (TypeTestNamedClass (t))
    t = NamedTypeGetTerm (t);
  if (!TypeTestStructureClass (t))
    return pk_false;
  TextForm->printLBracket ();
  TextForm->printLine ();
  le1 = NILEXPCAL;
  if (getExprLastIndexes (n, &e))
    le1 = EM->mkNaryExp (oplesign, e, le1, NILBACKANNOTATION);
  genFieldsTupleStruct (le, le1, t, re, le1);
  TextForm->printSpace ();
  TextForm->printRBracket ();
  if (n == 0)
    TextForm->printLine ();
  return pk_true;
}

/** 
    It generates the assignment expg := expd  for expd a synchronous expression.
    
    @param expg (in) left expression of the assignment.
    @param expd (in) right expression of the assignment; synchronous expression except opelevec, opensvec, opconc operators.
*/
_VIRTUAL_ void
cGenerator::genAssign (ptexpcal expg, ptexpcal expd)
{
  noperators op;
  ptexpcal exp1, saveexpd, exp2;
  int m, n;
  boolean b;

  if (expg == getExpressionToGenerate (expd))
    endStatement (0, pk_true);
  else
    {
      saveexpd = expd;
      setOptimizedIndexes (expd);
      n = genOpenTravelIndexes (expd);
      op = EM->getOperator (expd);
      if (op == oprest)
	{
	  EM->getBinaryArgs (expd, &exp1, &expd);
	  expg = EM->mkBinaryExp (opelem, expg, exp1, NILBACKANNOTATION);
	}

      if ((contextOp == opsequ) || (EM->getNumberOfDimensions (expd) > EM->getNumberOfDimensions (expg)))
	n = n + genOpenIteration (expd);
      else
	{
	  m = n;
	  if (EM->getOperator (expg) == opelem)
	    {
	      EM->getBinaryArgs (expg, &exp1, &exp2);
	      n = n + genOpenIteration (exp1, exp2);
	    }
	  else
	    n = n + genOpenIteration (expg);
	}

      if (!genTupleStruct (expg, expd))
	{
	  if (genArrayTupleStruct (expg, expd, n - m))
	    b = pk_false;
	  else
	    {
	      b = pk_true;
	      if (!isGenAssignString (expg, expd))
		{
		  gensig (expg, pk_false);
		  TextForm->printassign ();
		  genexpression (expd, opnullary);
		}
	    }
	  endStatement (n, b);
	  genCloseIteration (n);
	}
      resetOptimizedIndexes (saveexpd);
    }
}

/**
   It generates the assignment expg := vselect : expr. ( example  x := [E1,E2] : expression ).
   
   @param expg (in) left expression of the assignment.
   @param vselect (in) a indices-defined coordinate points of expg array.
   @param expr (in) expression assigned to the vselect element of expg.
*/
_VIRTUAL_ void
cGenerator::genAssignVectElem (ptexpcal expg, ptexpcal vselect, ptexpcal expr)
{
  int n;

  n = genOpenIteration (expg, vselect);
  // A faire : expg[exp1] peut necessiter des indexes supplementaires
  // par exemple : x:= [0]:999 et x est une  matrice.
  genelemdim (expg, vselect, NILEXPCAL);
  TextForm->printassign ();
  genexpression (expr, opnullary);
  printptvlinebi (n);
  genCloseIteration (n);
}

/** 
    It generates the assignment expg := expd for expd an expression such that [E1,E2]:expression.
   
    @param expg (in) left expression of the assignment.
    @param expd (in) right expression of the assignment; 
    @param ajustDim (in) at pk_true, it specifies that the first dimension of expd must be ignored (used for initialization of delays).
*/
_VIRTUAL_ void
cGenerator::genAssignVectElem (ptexpcal expg, ptexpcal expd, boolean ajustDim)
{
  ptexpcal exp1, exp2;

  EM->getBinaryArgs (expd, &exp1, &exp2);
  if (ajustDim)
    expd = EM->getNaryArgs (&exp1);
  genAssignVectElem (expg, exp1, exp2);
}

/**
   It generates the assignment expg[i1,...,in] := expd where the indexes are given by embedded loops (parc).

   @param expg (in) left expression of the assignment.
   @param parc (in) set of (index, lowerBound, upperbound, step).
   @param expd (in) right expression of the assignment; 
*/
_VIRTUAL_ void
cGenerator::genAssignElemParc (ptexpcal expg, ptexpcal parc, ptexpcal expd)
{
  genelemdim (expg, NILEXPCAL, parc);
  TextForm->printassign ();
  genexpression (expd, opnullary);
  TextForm->printSemiColonLine ();
}




/**
   It generates the conditional assgigment : if (h) expg := expd.
   
   @param h (in) a clock that defines the condition if is not the current clock (given by currentClock member of the cGenerator class)
   @param expg (in) left expression of the assignment.
   @param expd (in) right expression of the assignment; must be a synchronous one.
*/
_VIRTUAL_ void
cGenerator::genConditionalAssign (ptinsth h, ptexpcal expg, ptexpcal expd)
{
  ptinsth hAt;
  boolean b, nblock, bsameLine;

  bsameLine = sameLine;
  b = (h != currentClock);
  //  if (b)  removed July 2016
    {
      if (!CKM->getBooleanClockAt (h, CKM->getBranching (h, currentClock), &hAt))
	hAt = h;
      genClockCondition (hAt);
    }
  
  if (!EM->satisfiesList (expd)) nblock = pk_false; else nblock = (EM->getSize (expd) > 1);
  if (nblock) TextForm->blockIndentOpening ();
  genAssignSynchronous (expg, expd);
  if (nblock) TextForm->printBlockClosingBackIndent ();
  if (b && !bsameLine)
    TextForm->backindent (1);
}

/** 
    It generates the conditional assgigment : if (clockExp) expg := expd.
   
    @param clockExp (in) a boolean expression defining the condition.
    @param expg (in) left expression of the assignment.
    @param expd (in) right expression of the assignment; must be a synchronous one.
*/
_VIRTUAL_ void
cGenerator::genConditionalAssign (ptexpcal clockExp, ptexpcal expg, ptexpcal expd)
{
  TextForm->printkwif ();
  TextForm->printLP ();
  genexpression (clockExp, opnullary);

  if (sameLine)
    {
      TextForm->printRP ();
      TextForm->printSpace ();
    }
  else
    TextForm->printRPlineindent ();

  genAssignSynchronous (expg, expd);
  if (!sameLine)
    TextForm->backindent (1);
}

/**
   It generates a cell expression ( expg := expd at h )
  
   @param h (in)  the modification clock of expg.
   @param expg (in) left expression.
   @param expd (in) right expression; when expg is a node of a delay tree, a "else" statement is generated 
   to take into account the specific management of dynamic expressions.
*/
_VIRTUAL_ void
cGenerator::genAssignCell (ptinsth h, ptexpcal expg, ptexpcal expd)
{
  ptmemnode mn;
  int nbic, n;

  genConditionalAssign (h, expg, expd);
  expg = EM->getCloneClassHead (expg);
  if (ptexpcal_satisfies_node_tree_delay (expg, &mn))
    {
      TextForm->printelseindent ();
      n = genOpenIteration (expg);
      gensig (expg, pk_false);
      TextForm->printassign ();
      get_tree_delay_indi (expg, &mn);	// getindielemdim(&expg, &mn);
      nbic = ptmemnode_get_nb_indexes (mn);
      gennnelemdelay (mn, NILEXPCAL, NILEXPCAL, nbic, pk_true);
      printptvlinebi (n + 1);	// why +1
      genCloseIteration (n);
    }
}

/** 
    It generates the assignment expg := hexp for hexp a sub-term of a default expression.
   
    @param expg (in) left expression of the assignment.
    @param hexp (in) right expression ; when or default expression.
*/
_VIRTUAL_ boolean
cGenerator::gencascademele (ptexpcal expg, ptexpcal hexp)
{
  ptexpcal hexp1, hexp2;
  ptexpcal expd;

  expd = EM->getWhenOrCellArg (hexp);
  if (EM->getOperator (expd) != opmele)
    return genExpWhenClockDefault (expg, hexp, CKM->getUndefClock ());
  EM->getDefaultArgs (expd, &hexp1, &hexp2);
  if (!gencascademele (expg, hexp1))
    return genExpWhenClockDefault (expg, hexp2, EM->getContributionClock (hexp1));
  if (!CKM->satisfiesNullClock (EM->getContributionClock (hexp2)))
    {
      TextForm->printkwelse ();
      genExpWhenClockDefault (expg, hexp2, EM->getContributionClock (hexp1));
    }
  return pk_true;
}

/** 
    It generates the assignment expg := expd for expd a default expression.
 
    @param expg (in) left expression of the assignment.
    @param expd (in) right expression ; default expression.
*/
_VIRTUAL_ void
cGenerator::genAssignDefault (ptexpcal expg, ptexpcal expd)
{
  ptexpcal hexp1, hexp2;
  TypeType t;

  sameLine = !isStructureExpr (expg, &t);
  EM->getDefaultArgs (expd, &hexp1, &hexp2);
  if (gencascademele (expg, hexp1))
    {
      if (!CKM->satisfiesNullClock (EM->getContributionClock (hexp2)))
	{
	  // For the last, no condition
	  TextForm->printkwelse ();
	  genAssignSynchronous (expg, EM->getWhenOrCellArg (hexp2));
	  TextForm->printLine ();
	}
      else
	TextForm->printLine ();
    }
  else // For the last, no condition
    genAssignSynchronous (expg, EM->getWhenOrCellArg (hexp2));
  sameLine = pk_false;
}


/**
   It generates the assignment expg := expd for expd ::= "default_expression" cell h.
   @param expg (in) left expression of the assignment.
   @param expd (in) right expression defined by (E1 default E2 ...default En) cell C
*/
_VIRTUAL_ void
cGenerator::genAssignDefaultCell (ptexpcal expg, ptexpcal expd)
{
  ptexpcal hexp1, hexp2;
  ptinsth h1, h2, vh;

  EM->getDefaultArgs (expd, &hexp1, &hexp2);
  gencascademele (expg, hexp1);
  TextForm->printkwelse ();
  genAssignSynchronous (expg, EM->getWhenOrCellArg (hexp2));
}


/**
   It generates the assignment of exp to its equivalents.
   Specific to the management of the nodes of delay trees (a signal can be a root of several trees of delays).
   @param exp (in) a signal.
*/
_VIRTUAL_ void
cGenerator::genAssignEquivalence (ptexpcal exp)
{
  ptmemnode mn;
  ptlmemnode lmn;
  boolean copieext;
  ptmemnode elem;
  int n;
  ptexpcal e;

  if (!EM->isNILEXPCAL (exp))
    {
      if (EM->satisfiesList (exp))
	while (EM->getNaryArgs (&exp, &e))
	  genAssignEquivalence (e);
      else
	{
	  exp = EM->getCloneClassHead (exp);
	  if (ptexpcal_satisfies_node_tree_delay (exp, &mn))
	    {
	      lmn = ptmemnode_get_equiv (mn);
	      copieext = (boolean) ((ptmemnode_get_use_in_external_call (mn) == usein) && copyForExternal (exp));
	      if ((lmn != NILLMEMNODE) || copieext)
		{
		  n = genOpenIteration (exp);
		  while (ptlmemnode_getNaryArgs (&lmn, &elem))
		    {
		      genelemdelay (elem, NILEXPCAL, NILEXPCAL);
		      TextForm->printassign ();
		      genelemdelay (mn, NILEXPCAL, NILEXPCAL);
		      // TextForm->printSemiColonLine();
		    }
		  if (copieext)
		    {
		      genelemsimple (exp, NILEXPCAL, NILEXPCAL);
		      TextForm->printassign ();
		      genelemdelay (mn, NILEXPCAL, NILEXPCAL);
		      // TextForm->printSemiColonLine();
		    }
		  printptvlinebi (n);
		  genCloseIteration (n);
		}
	    }
	}
    }
}

//----- ProcCall, NodeCall

/**
   It generates signal expressions of a list linputs as input parameters of an external call.
   @return bfirst when linputs is empty, "false" otherwise.
   @param linputs (in) a combed signal expression.
   @param bfirst (in) specified,  when it is pk_true (resp. pk_false) that a list-separator must not (resp. must) be generated first.
*/
_VIRTUAL_ boolean
cGenerator::genInputsExternalNode (ptexpcal linputs, boolean bfirst)
{
  boolean printident /*, bfirst */ ;
  ptexpcal exp;

  // TextForm->printLP(); bfirst = pk_true;
  while (linputs != NILEXPCAL)
    {
      if (bfirst)
	bfirst = pk_false;
      else
	TextForm->printComma ();
      exp = EM->getNaryArgs (&linputs);
      printident = pk_false;
      if (EM->getOperatorClass (exp) == variabl)
	if (copyForExternal (exp))
	  printident = pk_true;
      if (printident)
	printPrefixedSignalIdentifier (exp);
      else
	genexpression (exp, opnullary);
    };
  return bfirst;
}

/**
   It generates the signals expressions of the list loutputs as output parameters of an external call.
   @param bfirst (in) specified, when it is pk_true (resp. pk_false) that a list-separator must not (resp. must) be generated first.
   @param loutputs (in) combed signal expression.
*/
_VIRTUAL_ void
cGenerator::genOutputsExternalNode (ptexpcal loutputs, boolean bfirst)
{
  ptexpcal exp;

  while (loutputs != NILEXPCAL)
    {
      if (bfirst)
	bfirst = pk_false;
      else
	TextForm->printComma ();
      exp = EM->getNaryArgs (&loutputs);
      genOutputSignal (exp, EM->getNumberOfDimensions (exp) == 0);
    }
}


/**
   It generates the signal (expEff) for a read.
*/
_VIRTUAL_ void
cGenerator::genReadEnvSignal (ptexpcal expEff)
{
  boolean bref;
  if (!genindielemdim (&expEff, pk_true))
    {
      bref = EM->getNumberOfDimensions (expEff) == 0 && (EM->getUnamedType (expEff) != StringType);
      printPrefixedIdentRef (getIOCompleteIdentifier (expEff), bref, EM->IsField (expEff));
    }
}


/**
   It generates the code of a read environment.
   The form of the call is :
   \code
   if (!r_<processName>_<ident1)> (ident2)) return pk_false;
   \endcode
   where ident1(resp. ident2) denotes the identifier of the signal exp (resp. expEff).
   @param exp (in) a signal.
   @param expEff (in) a signal.
   @param processName (in) an identifier.
*/
_VIRTUAL_ void
cGenerator::genCodeReadEnv (ptexpcal exp, ptident processName, ptexpcal expEff)
{
  TextForm->printkwif ();
  TextForm->printLP ();
  TextForm->printkwnot ();
  TextForm->printReadFunctionLP (getIOCompleteIdentifier (exp), processName);

  genReadEnvSignal (expEff);

  TextForm->printRP ();
  TextForm->printRP ();
  TextForm->printSpace ();
  TextForm->printreturn (pk_false);
}

/**
   It generates the call to the read function of the signal expEff at a clock (h). 
   It manages also the equivalence induced by the delay trees.

   @param exp (in) signal.
   @param expEff (in) signal.
   @param processName (in) an identifier.
   @param h (in) clock of the read node.

   \see genCodeReadEnv()
*/
_VIRTUAL_ void
cGenerator::genReadEnv (ptexpcal exp, ptident processName, ptexpcal expEff, ptinsth h)
{
  genIncrIndexDelayTreeNode (expEff);
  genBlockOpeningOf (h);
  genCodeReadEnv (exp, processName, expEff);
  genAssignEquivalence (expEff);
}


/**
   It generates the reading of exp for separated compiling. It generates 
   the assignment ident1 = ident2 where ident1 denotes the identifier of exp, and ident2 is 
   defined by _ident1_ (name of the formal parameter of the entry point).

   (parameter h is unused, but kept).
*/
_VIRTUAL_ void
cGenerator::genReadEnvSeparate (ptexpcal exp, ptinsth h)
{
  int n;

  n = genOpenIteration (exp);
  gensig (exp, pk_false);
  TextForm->printassign ();

  genFormalParameterIdent (exp, pk_false, 0);
  genLindexes ();
  printptvlinebi (n);
  genCloseIteration (n);
  genAssignEquivalence (exp);
}

/**
   It generates the writing of an output signal (exp).
*/
_VIRTUAL_ void
cGenerator::genWriteEnvSignal (ptexpcal exp)
{
  ptexpcal exp1;

  if (!genindielemdim (&exp, pk_false))
    {
      exp1 = getExpressionToGenerate (exp);
      if (exp1 == exp)
	if (isNodeAtInitialClock (EM->getNode (exp), &exp1))
	  genexpression (exp1, opnullary);
	else
	  printPrefixedSignalIdentifier (exp);
      else
	genexpression (exp1, opnullary);
    }
}

/**
   It generates the call to the write function of the signal expEff at a clock (h).
   The form of the call is:  w_<processName_<ident1)> (ident2)), where 
   ident1(resp. ident2) denotes the identifier of the signal exp (resp. expEff).
*/
_VIRTUAL_ void
cGenerator::genWriteEnv (ptexpcal exp, ptident processName, ptexpcal expEff, ptinsth h)
{
  genBlockOpeningOf (h);
  genIncrIndexDelayTreeNode (exp);
  TextForm->printWriteFunctionLP (getIOCompleteIdentifier (exp), processName);
  genWriteEnvSignal (expEff);
  TextForm->printRPSemiColonLine ();
}

/**
   It generates the writing of a signal (exp) at a clock (h) for separated compiling. 
   It generates the assignment ident1 = ident2 where ident2 denotes the identifier of exp, and ident1 is 
   defined by _ident1_ (name of the formal parameter of the entry point).
*/
_VIRTUAL_ void
cGenerator::genWriteEnvSeparate (ptexpcal exp, ptinsth h)
{
  int n;
  ptexpcal exp1;

  genBlockOpeningOf (h);
  n = genOpenIteration (exp);
  genFormalParameterIdent (exp, EM->IsTypeExprPredefinedType (exp), 0);
  genLindexes ();
  TextForm->printassign ();
  exp1 = getExpressionToGenerate (exp);
  if (exp1 == exp) gensig (exp, pk_false); else genexpression (exp1, opnullary);
  genLindexes ();
  printptvlinebi (n);
  genCloseIteration (n);
}

/**
   It generates the name, or a reference to when refTo is "true, of the signal exp.
*/
_VIRTUAL_ void
cGenerator::genOutputSignal (ptexpcal exp, boolean refTo)
{
  if (copyForExternal (exp))
    printPrefixedSignalIdentifier (exp);
  else
    gensig (exp, refTo);
}



/**
   It generates the instructions for the predefined "affine_sample" node.
   For B: = Affine_sample{phase,period_d} (C) where B is a boolean clock, it manages the counter (N) assigned to
   the clock B. The generated code is \n
   if (N==0) { B=true; N=period_d-1; } else { B= false; N=N -1; }
*/
_VIRTUAL_ void
cGenerator::cgenSampleNode (ptexpcal params, ptexpcal ins, ptexpcal outs)
{
  ptexpcal boolClock, phase, period_d;
  ptident idCounter;
  long int d;

  EM->getNaryArgs (&params, &phase);
  EM->getNaryArgs (&params, &period_d);
  d = EM->getValue (period_d);

  EM->getNaryArgs (&outs, &boolClock);
  EM->getAttribute (boolClock, affineSampleAttribute, (attributeValue *) & idCounter);
  TextForm->printkwif ();
  TextForm->printLP ();
  printPrefixedIdent (idCounter);
  TextForm->printkwequal ();
  TextForm->printint (0);
  TextForm->printRP ();
  TextForm->printSpace ();
  // {
  // TextForm->printkwthen();
  TextForm->printLBracket ();
  printPrefixedIdent (idCounter);
  TextForm->printassign ();
  TextForm->printint (d - 1);
  TextForm->printSemiColon ();
  gensig (boolClock, pk_false);
  TextForm->printassign ();
  TextForm->printkwlogicons (pk_true);
  TextForm->printSemiColon ();
  TextForm->printRBracket ();
  TextForm->printSpace ();
  // } else {
  TextForm->printkwelse ();
  TextForm->printLBracket ();
  printPrefixedIdent (idCounter);
  TextForm->printassign ();
  printPrefixedIdent (idCounter);
  TextForm->printkwminus ();
  TextForm->printint (1);
  TextForm->printSemiColon ();
  gensig (boolClock, pk_false);
  TextForm->printassign ();
  TextForm->printkwlogicons (pk_false);
  TextForm->printSemiColon ();
  TextForm->printRBracket ();
  // }
  TextForm->printLine ();
}


_VIRTUAL_ void
cGenerator::genRendezVousNode(ptident nom, ptexpcal linputs, ptexpcal loutputs)
{
  // TextForm->printBeginComment ();
  genExternalCall (nom, linputs, loutputs, pk_true, NILPRAGMA, NILTREE);
  // TextForm->printEndComment ();
  // TextForm->printLine ();
}



/**
   @return "true" if the node node denotes an intrinsic function node, "false" otherwise.
   It generates the code (and return pk_true value) for the following cases:
   - the conversion pragma (PRAGMA_CONVERSION)
   - the exception operator (\\); it generates the code (genPragmaCode()) associated with it in the internal table (tableOpCode).
   - the indice functions (see genIntrinsicfuncIndices())
   - the catesian product (<<I1,... In>> operator) (see genCartesianProduct())
   - the assertion model; it generates the code (genPragmaCode()) associated with it in the internal table (tableOpCode).
   - the affine sample node (see cgenSampleNode())
   
   The  internal table (tableOpCode) is filled by $SignalLib_C/CodeGenerator.defs file.
*/
_VIRTUAL_ boolean
cGenerator::genNodeIntrinsicFunc (ptINode node)
{
  ptexpcal linputs, loutputs, par;
  ptident nom, id;

  if (!NM->is_intrinsicfunc (node)) return pk_false;
  NM->get_procCall_arg (node, &nom, &par, &linputs, &loutputs);

  if (existPredefinedPragma (NM->getPragmas (node), PRAGMA_CONVERSION))
    {
      genOutputSignal (EM->getNaryArgs (&loutputs), pk_false);
      TextForm->printassign ();
      TextForm->printcast (identType2TypeType (ident2repstr (nom)));
      TextForm->printLP ();
      genInputsExternalNode (linputs, pk_true);
      TextForm->printRPSemiColonLine ();
      return pk_true;
    }

  id = getExceptionId ();
  if (eq_idents (nom, id))
    {
      genPragmaCode (nom, getEntryCodeOp (tableOpCode, ident2repstr (id)), par, linputs, loutputs);
      return pk_true;
    }
  if (isIndicesId (nom))
    {
      genIntrinsicfuncIndices (linputs, loutputs);
      return pk_true;
    }

  if (isCartesianProductId (nom))
    {
      genCartesianProduct (loutputs, linputs);
      return pk_true;
    }

  if (isAssertionId (nom))
    {
      genPragmaCode (nom, getEntryCodeOp (tableOpCode, ident2repstr (getAssertionId ())), par, linputs, loutputs);
      return pk_true;
    }
  if (isSampleId (nom))
    {
      cgenSampleNode (par, linputs, loutputs);
      return pk_true;
    }

  //  if (!isintrinsicfuncdeccounter(nom)) return pk_false;

  if (isSendId(nom) || isReceiveId(nom))
    {
      genRendezVousNode(nom, linputs, loutputs);
      return pk_true;
    }
  return pk_false;
}

/**
  @return "true" if there is no generated code, "false" otherwise.
  It generates some code when BboxOrGbox is defined and the implementation language referenced in 
  vlanguage is C ( or another language than C++, it generates some code but an error is detected and signaled). 
  
  For C, it generates a reference to the data area given by the BboxOrGbox pragma (getClassAndObjectExternalCall()).

  The code generated here must be coherent with the code generated for separated compiling.
*/
_VIRTUAL_ boolean
cGenerator::genExternalSignalNodeCallAsProcedureCall (ptpragma BboxOrGbox, ntrees vlanguage)
{
  ptrepstr vl, externalObj, vdata;

  if (BboxOrGbox == NILPRAGMA)
    return pk_true;
  vl = getStrLanguageExternalDef (vlanguage);
  externalObj = getClassAndObjectExternalCall (BboxOrGbox, &vdata);
  if ((vl == nilstr) || (vl == string2repstr ("C")))
    {
      TextForm->printkwref ();
      genPrefix ();
      TextForm->printstr (vdata);
      return pk_false;
    }
  if (vl != string2repstr ("C++"))
    {
      pK_messenger->errorStrStr (errInterfaceToLanguage, string2repstr ("C"), vl);
      TextForm->printkwnotimpl ();
      TextForm->printkwref ();
      TextForm->printstr (vdata);
      setSomeProblem ();
      return pk_false;
    }
  return pk_true;
}

/**
  @return "true" if there is no generated code, "false" otherwise.
  It generates some code when BboxOrGbox is defined (getClassAndObjectExternalCall()) 
  and the implementation language referenced in vlanguage is C++: for C++, it generates the name of the object, 
  referenced in the BboxOrGbox pragma, followed by a dot character.

  The code generated here must be coherent with the code generated for separated compiling.
*/
_VIRTUAL_ boolean
cGenerator::genExternalSignalNodeCallAsObjectReference (ptpragma BboxOrGbox, ntrees vlanguage)
{
  ptrepstr vl, externalObj, vdata;

  if (BboxOrGbox == NILPRAGMA)
    return pk_false;
  vl = getStrLanguageExternalDef (vlanguage);
  externalObj = getClassAndObjectExternalCall (BboxOrGbox, &vdata);
  if (vl == string2repstr ("C++"))
    {
      TextForm->printstr (vdata);
      TextForm->printDot ();
      return pk_true;
    }
  return pk_false;
}

/** 
    It prints a reference to an external call called id.
    For C, it prints the identifier.( For Java, it will be specialized.)
*/
_VIRTUAL_ void
cGenerator::genIdentExternalCall (ptident id)
{
  TextForm->printident (id);
}

/**
   It generates a procedure (resp. function) call when asFunc is "false" (resp. "true")
   with input parameters (linputs) and output parameters (loutputs).

   The references to a Signal model takes into account Polychrony specific separated compiling informations 
   (BboxOrGbox: black boxes or Grey boxes mode, and the implementation language (vlanguage) of the referenced model).
   These informations are used in  genExternalSignalNodeCallAsObjectReference(), genExternalSignalNodeCallAsProcedureCall() methods.

*/
_VIRTUAL_ void
cGenerator::genExternalCall (ptident nom, ptexpcal linputs, ptexpcal loutputs, boolean asFunc, ptpragma BboxOrGbox, ntrees vlanguage)
{
  boolean bfirst;
  ptexpcal exp, saveloutputs;

  saveloutputs = loutputs;
  if (asFunc && (loutputs != NILEXPCAL))
    {
      exp = EM->getNaryArgs (&loutputs);
      if ((EM->getNumberOfDimensions (exp) != 0) || (loutputs != NILEXPCAL))
	loutputs = saveloutputs;
      else
	{
	  genOutputSignal (exp, pk_false);
	  TextForm->printassign ();
	}
    }
  saveloutputs = loutputs;
  genExternalSignalNodeCallAsObjectReference (BboxOrGbox, vlanguage);
  genIdentExternalCall (nom);
  TextForm->printLP ();

  bfirst = genExternalSignalNodeCallAsProcedureCall (BboxOrGbox, vlanguage);
  genOutputsExternalNode (loutputs, genInputsExternalNode (linputs, bfirst));
  TextForm->printRPSemiColonLine ();

  copyForExternalList (saveloutputs);
}

/**
   @return "true" if a node (node) denotes a communicating node.
   The node is generated taking account the context given by blackBox. When blackBox 
   - is pk_true, the operation is a reference to an external model assumed compiled in separated mode. It calls
   genReadEnvSeparate() for a reading and genWriteEnvSeparate() for a writing.

   - is pk_false, the operation is an acces to the environment. It calls genReadEnv() for a reading , genWriteEnv() for a writing.
*/
_VIRTUAL_ boolean
cGenerator::genIOCall (ptINode node, boolean blackBox)
{
  ptexpcal exp;
  ptINode NodeCall;
  ptinsth h;
  boolean b;

  h = NM->get_nodeClock (node);
  /* readings */
  if (NM->is_readNode (node))
    {
      b = NM->is_outputNodeCall (node, &NodeCall);
      if (b)
	b = NM->isNodeCall (NodeCall);
      if (!b)
	{
	  exp = NM->get_io_arg (node);
	  if (!blackBox)
	    genReadEnv (exp, progName, exp, h);
	  else
	    genReadEnvSeparate (exp, h);
	}
      // else genNodeCallNode(NodeCall); // Added March 2015 LBesnard. then removed April 2015 (must be retested)
      return pk_true;
    }

  /* writings */
  if (NM->is_writeNode (node))
    {
      exp = NM->get_io_arg (node);
      if (!blackBox)
	genWriteEnv (exp, progName, exp, h);
      else
	genWriteEnvSeparate (exp, h);
      return pk_true;
    }

  return pk_false;
}


/**
   @return "true" if it exists a pragma associated with a node (ProcCallNode) used by the generator , "false" otherwise.
   For C generator, it looks at the PragmaGencodeLanguage pragma.
   It calls genPragmaCodeInt() when the pragma exists.
   @param ProcCallNode (in) assumed to be a procedure node.

   \code
   Example: the definition of the intrinsic function (read) defined in the library IO_INTRINSICS.LIB gives the following specification:

   process read =(?string message; ! x )
        spec (| message ^= x |)
        pragmas
           C_CODE "pK_PRINT(&i1); pK_READ&t&o1(&o1) "
           CPP_CODE "cout << &i1 ; cin >> &o1 "
           Java_CODE "System.out.println( &i1 ) ; &o1 = KR.read&t&o1(); if (KR.isEof()) return false "
        end pragmas;

	
   For a reference such that read("Give me a value: ", x) and x an integer, in a Signal program, 
   the C code will be
   -----------------------------------------------
              pK_PRINT("Give me a value: ");
              pK_READint(&x);
   -----------------------------------------------
   pK_PRINT,  pK_READint are predefined C function known by Polychrony.
   &i1 stands for the first inputs, &t&o1 stands for the type of the first output,...
\endcode
*/
_VIRTUAL_ boolean
cGenerator::genProcCallNodeWithCodePragma (ptINode ProcCallNode)
{
  ptpragma valPragma;

  if (!getPragmaStrKeyword (NM->getPragmas (ProcCallNode), PragmaGencodeLanguage, &valPragma))
    return pk_false;
  genPragmaCodeInt (ProcCallNode, valPragma);
  return pk_true;
}


/**
   It generates (genPragmaCode()) the code associated with a pragma (valPragma) using interface of a node (ProcCallNode).
*/
_VIRTUAL_ void
cGenerator::genPragmaCodeInt (ptINode ProcCallNode, ptpragma valPragma)
{
  ptexpcal expg, expd, par;
  ptident nom;
  ntrees statement;

  NM->get_procCall_arg (ProcCallNode, &nom, &par, &expd, &expg);
  statement = getPragmaStatement (valPragma);
  genPragmaCode (nom, statement, par, expd, expg);
}



/**
   It generates (genParametrizedCode()) the statement of a pragma (statement) followed by 
   a end of C statement (;) if expg is not a sub term. It uses
   name, params as parameters, expg as outputs, expd as inputs as effectives values in the
   statement. 
   
   Example: see genProcCallNodeWithCodePragma().
*/
_VIRTUAL_ void
cGenerator::genPragmaCode (ptident name, ntrees statement, ptexpcal param, ptexpcal expd, ptexpcal expg)
{
  boolean b;

  if (expg == NILEXPCAL)
    b = pk_false;
  else
    b = (EM->getOperatorClass (expg) == call);	// in SubTerm
  genParametrizedCode (statement, name, param, expd, expg);
  if (!b)
    {
      TextForm->printSemiColon ();
      TextForm->printLine ();
    }

}


/**
   It generates the code for the procedure call node inst taking into account the context/clock (genIncrIndexDelayTreeNodes(), genBlockOpeningOf()).
   It calls:
   - genIOCall() for I/O operations
   - genProcCallNodeWithCodePragma() for model with associated code using pragma
   - genNodeIntrinsicFunc() for some intrinsic model
   - genExternalCall() otherwise.
   
   followed by genAssignEquivalence() method.
 
   The node is generated taking account the context given by blackBox for I/O operations only.
   When blackBox is "true", the referenced model is assumed to be a model compiled in separated mode.
*/
_VIRTUAL_ void
cGenerator::genProcCallNode (ptINode inst, boolean blackBox)
{
  ptexpcal expg, expd, par;
  ptident nom;
  ptinsth h;
  ptlpragmas lp;
  ptpragma valPragma;
  boolean asFunction;
  ntrees vlang;

  if (!genIOCall (inst, blackBox))
    {
      NM->get_procCall_arg (inst, &nom, &par, &expd, &expg);
      genIncrIndexDelayTreeNodes (expg);
      h = NM->get_nodeClock (inst);
      genBlockOpeningOf (h);
      if (!genProcCallNodeWithCodePragma (inst))
	if (!genNodeIntrinsicFunc (inst))
	  {
	    lp = NM->getPragmas (inst);
	    asFunction = pk_true;
	    valPragma = NILPRAGMA;
	    if (isAnnotatedGreyBox (inst, &valPragma))
	      asFunction = pk_false;
	    else if (isAnnotatedBlackBoxGraph (inst, &valPragma))
	      asFunction = pk_false;
	    if (!NM->getAttribute (inst, EXTERN_LANGUAGE_IMPL, (attributeValue *) & vlang))
	      SET_NILTREE (vlang);
	    genExternalCall (nom, expd, expg, asFunction, valPragma, vlang);
	  }
      genAssignEquivalence (expg);
    }
}


/**
   It generates the code for a node call node taking into account the context/clock ( genIncrIndexDelayTreeNodes(), genBlockOpeningOf()).
   It calls:
   - genPragmaCode() for model with associated code using pragma
   - genExternalCall() otherwise

   followed by genAssignEquivalence() method.
*/
_VIRTUAL_ void
cGenerator::genNodeCallNode (ptINode node)
{
  ptident nom;
  ptexpcal OrderedParams, OrderedInputs, OrderedOutputs;
  ptDCG aGraph;
  ptpragma valPragma;
  ntrees statement;
  ptlpragmas prags;
  ptrepstr vClass, vobj;
  boolean asFunc;
  ntrees vlang;

  genBlockOpeningOf (NM->get_nodeClock (node));
  NM->get_nodeCall_arg (node, &aGraph, &OrderedParams, &OrderedInputs, &OrderedOutputs);
  genIncrIndexDelayTreeNodes (OrderedOutputs);
  prags = aGraph->getPragmas ();
  nom = aGraph->getName ();
  if (getPragmaStrKeyword (prags, PragmaGencodeLanguage, &valPragma))
    {
      statement = getPragmaStatement (valPragma);
      genPragmaCode (nom, statement, OrderedParams, OrderedInputs, OrderedOutputs);
    }
  else
    {
      valPragma = NILPRAGMA;
      if (isAnnotatedBlackBoxGraph (node, &valPragma))
	{
	  vClass = getClassAndObjectExternalCall (valPragma, &vobj);
	  // nom = makeIdent_repstr_ident( vClass, idIterBlackBox);
	  nom = makeIdent_ident_string (repstr2ident (vClass), STRING_ITERATE);
	  asFunc = pk_false;
	}
      else
	asFunc = !existPredefinedPragma (prags, PRAGMA_SEPARATE) && !existPredefinedPragma (prags, PRAGMA_UNEXPANDED);

      if (!NM->getAttribute (node, EXTERN_LANGUAGE_IMPL, (attributeValue *) & vlang))
	SET_NILTREE (vlang);
      genExternalCall (nom, OrderedInputs, OrderedOutputs, asFunc, valPragma, vlang);
    }
  genAssignEquivalence (OrderedOutputs);
}

//----- Delays

/**
   It generates code for a delay node (node) taking into account the context/clock (genBlockOpeningOf()).
   It calls genIncrIndexDelayTreeNode(), genAssignEquivalence().
*/
_VIRTUAL_ void
cGenerator::genDelayNode (ptINode node)
{
  ptexpcal exp, expg, expd, vinit;
  ptmemnode mn;

  // Doesn't reference textform object.
  NM->get_delay_arg (node, &expg, &expd, &exp, &exp, &vinit);
  genBlockOpeningOf (NM->get_nodeClock (node));
  expg = EM->getCloneClassHead (expg);
  if (!ptexpcal_satisfies_node_tree_delay (expg, &mn))
    genAssign (expg, expd);
  else
    {
      genIncrIndexDelayTreeNode (expg, pk_true);
      genAssignEquivalence (expg);
    }
}

//----- Sub-Graph

/**
   It generates the declaration of the entry point assigned to a sub-graph cluster (aSubGraph). 
   The code is generated taking account the parameters declaration 
   and the context mode (SEPARATE).

   In SEPARATE mode, the cluster will be used for separated compiling, the inputs/outpus of 
   the sub graph are generated as parameters. Otherwise, the cluster is used only in the module, the entry 
   point is declared without formal parameters.
   
   The boolean declaration must be
   - "true" for only a declaration (generation of the header file) 
   - "false" for the definition of the entry point. 
*/
_VIRTUAL_ void
cGenerator::genEntryPointSubGraphHeader (ptINode aSubGraph, boolean declaration)
{
  // Doesn't reference textform object
  ptident name;
  ptexpcal linputs, loutputs;

  if (SEPARATE)
    NM->get_inOutGraphRef (aSubGraph, &linputs, &loutputs);
  else
    {
      linputs = NILEXPCAL;
      loutputs = NILEXPCAL;
    }
  NM->get_clusterName (aSubGraph, &name);
  genEntryPointDeclaration (name, linputs, loutputs, declaration, SEPARATE);
}

/**
   It generates code for a reference to the node aSubGraph when it is a Cluster one.
   Such a reference is a procedure call with or without parameters according to the context 
   given by the parameters blackBox and schedulerTrain.
   When blackBox is "false" or schedulerTrain  is "true", the procedure is called without parameters.
*/
_VIRTUAL_ void
cGenerator::genRefSubGraphNode (ptINode aSubGraph, boolean blackBox, boolean schedulerTrain)
{
  ptident nom;
  ptexpcal linputs, loutputs;
  ptlINode ln;

  if (NM->is_trainCluster (aSubGraph))
    {
      genBlockOpeningOf (NM->get_nodeClock (aSubGraph));
      NM->get_clusterName (aSubGraph, &nom);
      TextForm->printident (nom);
      if (!blackBox)
	NM->get_inOutGraphRef (aSubGraph, &linputs, &loutputs);
      else
	{
	  linputs = NILEXPCAL;
	  loutputs = NILEXPCAL;
	}
      if (schedulerTrain && !blackBox)
	{
	  linputs = NILEXPCAL;
	  loutputs = NILEXPCAL;
	}
      TextForm->printLP ();
      genOutputsExternalNode (loutputs, genInputsExternalNode (linputs, pk_true));
      TextForm->printRPSemiColonLine ();
    }
  else if (NM->getAttribute (aSubGraph, ARRAY_ORDER, (attributeValue *) & ln))
    {
      genIncrIndexDelayTreeNodeList (ln);
      genOrderedNodes (ln, NM->get_nodeClock (aSubGraph), blackBox, schedulerTrain);
    }
}

//----- Array: the array "processes" are partialy implemented in this version. 

/**
   It generates the loop for a index (indexNode) characterized by a name, a lower bound, an upper bound.
   It calls printIterateIndex().
*/
_VIRTUAL_ void
cGenerator::genArrayLoopBegin (ptINode indexNode)
{
  ptexpcal indexDef, uBound, lBound, vindex;

  vindex = NM->get_index (indexNode);
  indexDef = NM->get_index_def (indexNode);
  lBound = EM->getTernaryArgstAt (indexDef, 1);
  uBound = EM->getTernaryArgstAt (indexDef, 2);
  printIterateIndex (vindex, lBound, uBound, NILEXPCAL, pk_true);
}

/**
   It applies genIncrIndexDelayTreeNodes() for the outputs of each node of ln.
   It recalls this method for the nodes of a array node of ln.
*/
_VIRTUAL_ void
cGenerator::genIncrIndexDelayTreeNodeList (ptlINode ln)
{
  ptINode n;
  ptlINode ln1;

  ln = getNodesToBeGenerated (ln);
  while (ptlINode_get_elem (&ln, &n))
    {
      genIncrIndexDelayTreeNodes (NM->get_leftOperand (n));
      if (NM->is_array (n))
	if (NM->getAttribute (n, ARRAY_ORDER, (attributeValue *) & ln1))
	  genIncrIndexDelayTreeNodeList (ln1);
    }
}


/**
   It generates the code of an array node (node) taking into account the following context informations:
   - blackBox is "true" when the graph is without clusters, "false" otherwise;
   - schedulerTrain is "true" when the node is an element of the scheduller. Otherwise it is internal to a cluster. 
   \note the arrays "processes" are partialy implemented in this version.

   It manages a attribute member (ARRAY_CONTEXT).
   
   It calls the following methods:
   - genIncrIndexDelayTreeNodeList(),
   - genBlockOpeningOf(), 
   - genOrderedNodes() for the initialisations of the array,
   - pushIndex(),
   - genArrayLoopBegin(),
   - genOrderedNodes() for the definition of the body of the array,
   - popIndex()
   - genAssignEquivalence()
*/
_VIRTUAL_ void
cGenerator::genArrayNode (ptINode node, boolean blackBox, boolean schedulerTrain)
{
  ptINode indexNode;
  ptinsth h;
  ptlINode ln, initNodes;
  ptlinsth lh;
  ptexpcal indice;

  if (!NM->getAttribute (node, ARRAY_ORDER, (attributeValue *) & ln))
    return;			//  strange : array without node.
  genIncrIndexDelayTreeNodeList (ln);
  indexNode = NM->get_indexNode (node);
  indice = NM->get_index (indexNode);
  ARRAY_CONTEXT = pk_true;
  h = NM->get_nodeClock (node);
  genBlockOpeningOf (h);
  NM->get_array_initDefinitions (node, &initNodes, &lh);
  genOrderedNodes (initNodes, h, blackBox, schedulerTrain);
  pushIndex (indice);
  genArrayLoopBegin (indexNode);
  TextForm->printblockopening ();
  genOrderedNodes (ln, h, blackBox, schedulerTrain);
  /* bouclegeninstsansretards(inst); REVOIR */
  TextForm->printBlockClosingBackIndent ();
  popIndex (indice);
  if (getIndexStackSize () == 0)
    genAssignEquivalence (NM->get_leftOperand (node));
}



/**
   It generates code for a node (node) taking into account somes contexts.
   It realizes a switch on the type of the node : 
   - genEquNode() for an equation (excluding delays)
   - genDelayNode() for delay/window,
   - genProcCallNode() for an external procedure call,
   - genNodeCallNode() for a node call,
   - genRefSubGraphNode() for a reference to a sub-graph,
   - genArrayNode() for an array process.

  Contexts:
  - blackBox must be "true" when the graph is without clusters, "false" otherwise;
  - schedulerTrain must be "true" when the nodes lOrderedNodes are the nodes of the scheduller.
*/
_VIRTUAL_ void
cGenerator::genNode (ptINode inst, boolean blackBox, boolean schedulerTrain)
{
  if (!mustBeGenerated (inst))
    return;
  switch ((int) (NM->get_oper (inst)))
    {
    case opdefpart:
    case opdefsig:
      genEquNode (inst);
      break;
    case opmoti:
      genArrayNode (inst, blackBox, schedulerTrain);
      break;
    case opappe:
      genProcCallNode (inst, blackBox);
      break;
    case opnodecall:
      genNodeCallNode (inst);
      break;
    case opret:
    case opfen:
      genDelayNode (inst);
      break;

    case opsubgraph:
      genRefSubGraphNode (inst, blackBox, schedulerTrain);
      break;

    case opldefpart:		// Set of part definition node : no code.
    case opbund:		// struct./bundle --> extension (fields)
      break;

    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
}

/**
   Misc.
   @return the list of the unoptimized nodes of a list (lOrderedNodes).
*/
_VIRTUAL_ ptlINode cGenerator::getNodesToBeGenerated (ptlINode lOrderedNodes)
{
  ptlINode
    ln;
  ptINode
    aNode;

  ln = lOrderedNodes;
  while (ptlINode_get_elem (&lOrderedNodes, &aNode))
    if (!mustBeGenerated (aNode))
      ptlINode_remove_elem (&ln, aNode);
  return ln;
}



/**
   Code generation for a list of nodes(lOrderedNodes) taking into account somes contexts (sgClock, blackBox, schedulerTrain).
  
   @param lOrderedNodes (in) list of nodes to be generate assumed ordered.
   @param sgClock (in) denotes the branching clock of the nodes of lOrderedNodes
   @param blackBox (in) "true" when the graph is without clusters, "false" otherwise
   @param schedulerTrain (in) at "true" when the nodes are the nodes of the scheduller.
*/
_VIRTUAL_ void
cGenerator::genOrderedNodes (ptlINode lOrderedNodes, ptinsth sgClock, boolean blackBox, boolean schedulerTrain)
{
  ptINode aNode;

  lOrderedNodes = getNodesToBeGenerated (lOrderedNodes);
  if (lOrderedNodes == NILLINODE)
    return;

  initContextGenerator (sgClock);
  while (ptlINode_get_elem (&lOrderedNodes, &aNode))
    {
      genNode (aNode, blackBox, schedulerTrain);
      if (!schedulerTrain)
	set_booleanClocks_undefined (aNode);
    }
  CKM->resetMarkAll (alreadyVisisted);
  genBlockOpeningOf (sgClock);
}

/**
   @return "true" if the operand expression is a reference to a state variable (stateVar?) and its node (node) must be generated, "false" otherwise (node is not set).
*/
boolean
cGenerator::getNodePreStateVariable (ptexpcal exp, ptexpcal * stateVar, ptINode * node)
{
  ptINode n;

  if (EM->getOperator (exp) != opvvarin) return pk_false;
  *stateVar = EM->getUnaryArg (exp);
  n = EM->getNode (*stateVar);
  if (NM->is_undefined (n)) return pk_false;
  if (!mustBeGenerated (n)) return pk_false;
  *node = n;
  return pk_true;
}

/**
   A state variable  S value of exp parameter (S? named S_previous) must be updated when its value as changed during a step.
   It is done for each partial definition S::= Ei when hi.
   If hi is an undersampling such that hi=X?, S must be updated before X (this test is valid because in the code gneration, the clocks are not substituabled).

   @param exp a state variable
   @param node definiton of exp  (::= or a bundle/struct)
   @param lexp  list of state variables of the program
   @param lres  list of the ordered state variables
   @return  list of the ordered state variables

 */
ptlinstances
cGenerator::stateVariableOrdering (ptexpcal exp, ptINode node, ptlinstances lexpr, ptlinstances * lres)
{
  ptexpcal s, vcond;
  ptlINode ln;
  ptINode n;
  ptinsth h;
  noperators vop;
  ptlinstances ls;

  vop = NM->get_oper (node);
  if (vop == opldefpart)
    {
      ln = NM->get_setOfPartialDefinitions_arg (node);
      while (ptlINode_get_elem (&ln, &n))
	{
	  h = NM->get_nodeClock (n);
	  if (CKM->satisfiesSub (h))
	    {
	      vcond = CKM->getConditionSub (h);
	      vcond = getExpressionToGenerate (vcond);
	      if (EM->getOperator (vcond) == opvvarin)
		if (!ptlinstances_isin_exp (lexpr, vcond))	// i.e. not after.
		  {
		    *lres = ptlinstances_add_pre (*lres, exp, vcond);
		    lexpr = *lres;
		  }
	    }
	}
    }
  else if (vop == opbund)	// Struct/bundle
    {
      ls = NM->get_lfields_arg (node);
      while (ptlinstances_get_next (&ls, &s))
	lexpr = stateVariableOrdering (exp, EM->getNode (s), lexpr, lres);
    }
  return lexpr;
}


/**
   This method orders the state variables before the effective code generation.
   A state variable (S? named S_previous) must be updated when its value as changed during a step.
   It is done for each partial definition S::= Ei when hi.
   If hi is an undersampling such that hi=X?, S must be updated before X. This test is valid because in the code gneration, the clocks are not substituabled.
   
   It applies stateVariableOrderingFor() for state variable.
*/
ptlinstances
cGenerator::stateVariablesOrdering (ptlinstances lexpr)
{
  ptlinstances lres;
  ptexpcal vcond, exp, stateVar;
  ptINode node;

  lres = lexpr;
  while (ptlinstances_get_next (&lexpr, &exp))
    if (getNodePreStateVariable (exp, &stateVar, &node))
      lexpr = stateVariableOrdering (exp, node, lexpr, &lres);
  return lres;
}


/**
   It generates the updating of the state variables of the graph.
   For a state variable, it computes the list of the "modification clocks" of the state. These 
   clocks are by definition exclusive.

   The generation takes into account the AttributeNameGencode attribute associated with a state variable in stateVariablesAttribute() method.
   
*/
_VIRTUAL_ void
cGenerator::genUpdateStateVariables ()
{
  ptlinstances lexpr;
  ptexpcal exp, stateVar;
  ptINode node;
  ptinsth h;
  ptlinsth lh, lhsave;

  if (!EM->getExprList (opvvarin, &lexpr))
    return;
  // lexpr : contains the list of the expressions x? (reference to a state variable) of the program.
  lexpr = stateVariablesOrdering (lexpr);
  while (ptlinstances_get_next (&lexpr, &exp))
    if (getNodePreStateVariable (exp, &stateVar, &node))
      {
	lh = getListOfModificationClocks (node, NILLINSTH);
	lhsave = lh;
	while (ptlinsth_get_next (&lh, &h))
	  if (!CKM->satisfiesNullClock (h))
	    {
	      genBlockOpeningOf (h);
	      genAssign (exp, stateVar);
	    }
	ptlinsth_free (lhsave);
      }
  genBlockOpeningOf (tick);	// closing all blocks.
}

/**
   @return lh + the modification clocks associated with a node (node).
   \note not implemented for array, structure.
*/
_VIRTUAL_ ptlinsth
cGenerator::getListOfModificationClocks (ptINode node, ptlinsth lh)
{
  ptlINode ln;
  ptINode n;
  boolean b;
  ptlinstances ls;
  ptexpcal s;

  switch ((int) (NM->get_oper (node)))
    {
    case opdefpart:
    case opdefsig:
    case opappe:
    case opnodecall:
    case opret:
    case opfen:
      lh = ptlinsth_sorted_insert_elem (lh, NM->get_nodeClock (node), &b);
      break;

    case opldefpart:		// Set of part definition node.
      ln = NM->get_setOfPartialDefinitions_arg (node);
      while (ptlINode_get_elem (&ln, &n))
	lh = getListOfModificationClocks (n, lh);
      break;

    case opbund:		// struct./bundle --> extension (fields)
      ls = NM->get_lfields_arg (node);
      while (ptlinstances_get_next (&ls, &s))
	lh = getListOfModificationClocks (EM->getNode (s), lh);
      break;

    case opsubgraph:
      // not implemented.
      break;
    case opmoti:
      // not implemented.
      break;

    default:;

    }
  return lh;
}





/**
   It generates the assignment expg := expd for expd a concatenation (i.e || operator).
*/
_VIRTUAL_ void
cGenerator::genConcatenate (ptexpcal expg, ptexpcal expd)
{
  genConcatenateElement (expg, expd, Zero);
}

/**
   It generates the assignment expg := expd for a concatenate operator( || operator ).
   
   @param expg (in) left expression of the assignment.
   @param expd (in) right expression of the assignment to be subscripted using binf parameter.
   @param binf (inout) denotes the current index to use to generate the subscript on expg.
   
   @return the expression obtained by increasing binf by the size of the first dimension of the expression expd.
*/
_VIRTUAL_ ptexpcal
cGenerator::genConcatenateElement (ptexpcal expg, ptexpcal expd, ptexpcal binf)
{
  ptexpcal exp1, exp2, vDim, bi;
  int n;

  if (EM->getOperator (expd) == opconc)
    {
      EM->getBinaryArgs (expd, &exp1, &exp2);
      binf = genConcatenateElement (expg, exp1, binf);
      return genConcatenateElement (expg, exp2, binf);
    }
  if (EM->getIthDimension (expd, 1, &vDim))
    {
      n = genOpenIteration (expd);
      bi = topIndex ();
      if (!EM->isEquivalent (binf, Zero, pk_true))
	bi = EM->mkBinaryExp (opplus, binf, bi, NILBACKANNOTATION);
      bi = EM->mkNaryExp (oplesign, bi, NILEXPCAL, NILBACKANNOTATION);
      genAssignVectElem (expg, bi, expd);
      TextForm->backindent (n);
      genCloseIteration (n);
      return EM->mkBinaryExp (opplus, binf, vDim, NILBACKANNOTATION);
    }
  // Scalar.
  bi = EM->mkNaryExp (oplesign, binf, NILEXPCAL, NILBACKANNOTATION);
  genAssignVectElem (expg, bi, expd);
  return EM->mkBinaryExp (opplus, binf, One, NILBACKANNOTATION);
}
