/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   It generates the declaration of the list of the indexes of the index stack (indexStack).
*/
  _VIRTUAL_ void
cGenerator::genDeclarationOfIndexes ()
{
  ptlinstances l;
  ptexpcal vindex;

  l = indexStack;
  while (ptlinstances_get_next (&l, &vindex))
    genIndexDeclaration (vindex);
}

/**
   It generates the access to the n first indexes of the index stack (indexStack).
*/
_VIRTUAL_ void
cGenerator::genLindexesHead (int n)
{
  ptlinstances l;
  ptexpcal vindex;
  int i;

  if (n > 0)
    {
      l = indexStack;
      for (i = 0; i < n; i++)
	{
	  ptlinstances_get_next (&l, &vindex);
	  printIndexAccess (vindex);
	}
    }
}

/**
   @return 
   - "true" if the access (list of expressions) to the n last indexes of the index stack (indexStack) is defined in vexp,
   - "false" otherwise.
*/
_VIRTUAL_ boolean
cGenerator::getExprLastIndexes (int n, ptexpcal * vexp)
{
  ptlinstances l;
  ptexpcal vindex;
  int nbIterator, i;

  *vexp = NILEXPCAL;
  nbIterator = getIndexStackSize ();
  if (nbIterator <= 0)
    return pk_false;
  i = nbIterator - n;
  if (i < 0)
    return pk_false;

  l = indexStack;
  while (i > 0)
    {
      ptlinstances_get_next (&l, &vindex);
      i--;
    }
  if (ptlinstances_get_size (l) == 0)
    return pk_false;
  while (n != 0)
    {
      if (ptlinstances_get_ith_element (l, n, &vindex))
	*vexp = EM->mkNaryExp (oplesign, vindex, *vexp, NILBACKANNOTATION);
      n--;
    }
  return (*vexp != NILEXPCAL);
}

/**
   See genTravelIndexLst()
*/
_VIRTUAL_ void
cGenerator::genIndexes ()
{
  genLindexes ();
}

/** 
    @return "true" if exp is an optimized index, "false" otherwise.
    An index (object markbed by travelIndexAttribute) is optimized when it is attributed with OPTIMIZED_INDEX
    (see setOptimizedIndexes()).
*/
_VIRTUAL_ boolean
cGenerator::isOptimizedIndex (ptexpcal exp)
{
  if (!EM->isMarked (exp, travelIndexAttribute))
    return pk_false;
  return EM->isMarked (exp, OPTIMIZED_INDEX);
}


/**
   Stack of indexes management: it pushes vindex at the top of indexStack (attribute member of the class).
*/
_VIRTUAL_ void
cGenerator::pushIndex (ptexpcal vindex)
{
  indexStack = ptlinstances_push_end (indexStack, vindex);
}


/**
   Stack of indexes management: it pops vindex from the indexStack stack(attribute member of the class).
*/
_VIRTUAL_ void
cGenerator::popIndex (ptexpcal vindex)
{
  indexStack = ptlinstances_remove_exp (indexStack, vindex);
  if (indexStack == NILLINSTANCES)
    ARRAY_CONTEXT = pk_false;
}

/**
   @return the top of the indexStack stack.
*/
_VIRTUAL_ ptexpcal
cGenerator::topIndex ()
{
  ptexpcal x;
  ptlinstances l;

  l = indexStack;
  if (ptlinstances_get_next (&l, &x))
    return x;
  return NILEXPCAL;
}

/**
   @return the size of the the indexStack stack.
*/
_VIRTUAL_ int
cGenerator::getIndexStackSize ()
{
  return ptlinstances_get_size (indexStack);
}

/**
   It resets the array context (indexStack stack, ARRAY_CONTEXT°.
*/
_VIRTUAL_ void
cGenerator::resetTravelIndexes ()
{
  indexStack = NILLINSTANCES;
  ARRAY_CONTEXT = pk_false;
}


/**
   @return a new identifier of index.
*/
ptexpcal cGenerator::newIndex ()
{
  return EM->mkIndexIdent (NILIDENT, NILEXPCAL, NM->mk_node_undefined (), CKM->getUndefClock (), NILBACKANNOTATION);
}


/**
  Generation of a loop for each dimension of a signal.
  @return the number of opened loops.
  @param exp (in) a signal
*/
_VIRTUAL_ int
cGenerator::genOpenTravelIndexes (ptexpcal exp)
{
  ptexpcal vindex, ithDimExpr, assIndex;
  ptlinstances l;
  int n;

  if (!EM->getTravelIndexes (exp, &l))
    return 0;
  n = 1;
  while (ptlinstances_get_next (&l, &vindex))
    {
      if (EM->getIthDimension (vindex, n, &ithDimExpr))
	{
	  n++;
	  assIndex = genOpenLoopForDim (ithDimExpr);
	  EM->addAttribute (vindex, IteratorAttribute, (attributeValue) assIndex);
	}

      else;			// internal error
      EM->addMark (vindex, travelIndexAttribute);
    }
  return n - 1;
}

/**
  Generation of a loop for each dimension of a signal.
  It calls genOpenIteration(ptexpcal , ptexpcal).
  @return the number of opened loops.
  @param exp (in) a signal
*/
_VIRTUAL_ int
cGenerator::genOpenIteration (ptexpcal exp)
{
  return genOpenIteration (exp, NILEXPCAL);
}

/**
   For a signal (varray) of array type indexed by a list  of indexes (lindex), it opens loops for the dimensions
   not yet opened of the indexed signal.
   @return the number of opened loops.
*/
_VIRTUAL_ int
cGenerator::genOpenIteration (ptexpcal varray, ptexpcal lindex)
{
  TypeType TT;
  ptexpcal dimLst, x;
  ptlinstances l;
  boolean b;
  int n;

  if (!EM->getTypeAndDim (varray, &TT, &dimLst))
    return 0;
  n = 0;
  l = indexStack;
  b = pk_true;
  while (b && EM->getNaryArgs (&lindex, &x))
    {
      b = EM->getNaryArgs (&dimLst, &x);
      ptlinstances_get_next (&l, &x);
    }

  while (b && ptlinstances_get_next (&l, &x))
    b = EM->getNaryArgs (&dimLst, &x);
  if (!EM->isNILEXPCAL (dimLst))
    n = genOpenLoopForDimLst (dimLst);
  if (TypeTestNamedClass (TT))
    TT = NamedTypeGetTerm (TT);
  while (TypeTestArrayClass (TT))
    {
      dimLst = ArrayMatrixTypeGetDimLst (TT);
      n = n + genOpenLoopForDimLst (dimLst);
      TT = ArrayMatrixTypeGetTerm (TT);
      if (TypeTestNamedClass (TT))
	TT = NamedTypeGetTerm (TT);
    }
  return n;
}


/**
   @return the (i+1)-th internal indexes where i is the size of the index stack.
   \note 
   The internal indexes are computed during the declaration phasis (genInternalIndexes()).
*/
_VIRTUAL_ ptexpcal
cGenerator::getInternalIndex ()
{
  ptlinstances l1, l2;
  ptexpcal vindex;

  l1 = InternalIndexLst;
  l2 = indexStack;
  while (ptlinstances_get_next (&l2, &vindex))
    ptlinstances_get_next (&l1, &vindex);
  if (!ptlinstances_get_next (&l1, &vindex))
    // internal error
    {
      pK_messenger->error (errcompilerbug);	//  (send a mail to...)
      setSomeProblem ();
    }

  return vindex;
}

/**
   @return the index associated with the travel of a dimension (vdim).
   This index is an internal one (ie declared for the code generator).

   It generates the opening of a loop for a dimension (vdim) using the index.
   It calls getInternalIndex(), printForLoopHead(), printForLoopHead().
*/
_VIRTUAL_ ptexpcal
cGenerator::genOpenLoopForDim (ptexpcal vdim)
{
  ptexpcal vindex;

  vindex = getInternalIndex ();
  pushIndex (vindex);
  printForLoopHead (vindex, vdim, pk_false);
  return vindex;
}

/**
   Generation of a loop for each dimension of list of dimensions.
   It calls genOpenLoopForDim() for each dimension.
   @return the number of opened loops.
   @param  vDimLst (in) an assumed list of dimensions.
*/
_VIRTUAL_ int
cGenerator::genOpenLoopForDimLst (ptexpcal vDimLst)
{
  ptexpcal vdim;
  int n;

  n = 0;
  while (EM->getNaryArgs (&vDimLst, &vdim))
    {
      n++;
      genOpenLoopForDim (vdim);
    }
  return n;
}

_VIRTUAL_ void
cGenerator::genIndexDeclaration (ptexpcal vindex)
{
  TextForm->printkwIntSpace ();
  printIndex (vindex);
  TextForm->printchar (';');
}

/**
   It generates the opening of a loop with vindex as index, udound as upper bound.
   Moreover, the declaration of the index is integrated if withIndexDeclaration.

   It calls printIterateIndex().
*/
_VIRTUAL_ void
cGenerator::printForLoopHead (ptexpcal vindex, ptexpcal ubound, boolean withIndexDeclaration)
{
  // TextForm->printchar('{');

  // Index declaration (in a block)
  if (withIndexDeclaration)
    genIndexDeclaration (vindex);
  // Iteration on previous index;
  printIterateIndex (vindex, NILEXPCAL, ubound, NILEXPCAL, pk_false);
}

/* Generation of a loop.
   @return false if ident is optimized (for example [ i in 1 to 1 ] (i becomes 1)), true otherwise.
   @param ident (in) index of the loop.
   @param binf (in) lower bound.
   @param bsup (in) higher bound.
   @param pas (in) the step.
*/
_VIRTUAL_ boolean
cGenerator::genLoopBegin (ptexpcal ident, ptexpcal binf, ptexpcal bsup, ptexpcal pas)
{
  if (EM->getOperatorClass (ident) != variabl)
    return pk_false;		// cas [ i in e to e ] (i->e)
  printIterateIndex (ident, binf, bsup, pas, pk_true /* modified May 2008: false */ );
  return pk_true;
}


/**
   It generates the index of array expression (vindex).
   In C language, [] operator.
*/
_VIRTUAL_ void
cGenerator::printIndexAccess (ptexpcal vindex)
{
  TextForm->printLeftSquareBracket ();
  genexpression (vindex, opnullary);	// printIndex(vindex);
  TextForm->printRightSquareBracket ();
}


/*
  Generation of the beginning of an iteration on an index.
  Taking account prefix context. ("i" or  "data->i" in separated compiling).
  @param ident (in) index of the loop.
  @param binf (in) lower bound.
  @param bsup (in) higher bound.
  @param pas (in) the step.
*/
_VIRTUAL_ void
cGenerator::printIterateIndex (ptexpcal ident, ptexpcal lBound, ptexpcal uBound, ptexpcal vstep, boolean bEqual)
{
  boolean bincrease;

  TextForm->printForLP ();
  printIndex (ident);
  TextForm->printassign ();
  if (lBound == NILEXPCAL)
    TextForm->printint (0);
  else
    genexpression (lBound, opnullary);
  TextForm->printSemiColon ();
  printIndex (ident);
  if (vstep == NILEXPCAL)
    bincrease = pk_true;
  else
    bincrease = (boolean) (EM->getValue (vstep) > 0);
  if (bincrease)
    if (bEqual)
      TextForm->printkwLessthanEqual ();
    else
      TextForm->printkwLessthan ();
  else if (bEqual)
    TextForm->printkwGreaterthanEqual ();
  else
    TextForm->printkwGreaterthan ();
  genexpression (uBound, opnullary);
  TextForm->printSemiColon ();
  printIndex (ident);
  if (vstep == NILEXPCAL)
    TextForm->printkwincr ();
  else
    {
      TextForm->printassignincr ();
      genexpression (vstep, opnullary);
    }
  TextForm->printRPlineindent ();
}

/**
   It generates the end of a statement and signals n back indentations to the shaper.
*/
_VIRTUAL_ void
cGenerator::printptvlinebi (int n)
{
  TextForm->printSemiColonLineBackIndent (n);
}

/**
   It signals n back indentations to the shaper, where n is the current index stack size.
*/
_VIRTUAL_ void
cGenerator::backIndent ()
{
  TextForm->backindent (getIndexStackSize ());
}


/*
  It generates a list of contextual indexes: [0] ... [nbind]
*/
_VIRTUAL_ void
cGenerator::genIndexes (int nbind)
{
  int n;

  n = 0;
  while (n < nbind)
    {
      n++;
      TextForm->printLeftSquareBracket ();
      TextForm->printindex (n);
      TextForm->printRightSquareBracket ();
    }

}

/**
   It generates the list of the indexes of the index stack (indexStack).
*/
_VIRTUAL_ void
cGenerator::genTravelIndexLst (ptlinstances l)
{
  ptexpcal vindex;

  while (ptlinstances_get_next (&l, &vindex))
    printIndexAccess (vindex);
}


/**
   It generates the list of the indexes of the index stack (indexStack).
*/
_VIRTUAL_ void
cGenerator::genLindexes ()
{
  genTravelIndexLst (indexStack);
}

/**
   Assumed: n>0, i>=0.
   It generates n indexes of the indexes of the index stack (indexStack), from the i-th.
*/
_VIRTUAL_ void
cGenerator::genLindexes (int i, int n)
{
  ptexpcal vindex;
  ptlinstances l;

  l = indexStack;
  while (i > 0)
    {
      ptlinstances_get_next (&l, &vindex);
      i--;
    }
  if (ptlinstances_get_next (&l, &vindex))
    while (i != n)
      {
	printIndexAccess (vindex);
	ptlinstances_get_next (&l, &vindex);
	i++;
      }
}


/**
   It closes n loops.
*/
_VIRTUAL_ void
cGenerator::genCloseIteration (int n)
{
  ptlinstances x;

  if ((getIndexStackSize () > 0) && !ARRAY_CONTEXT)
    TextForm->printLine ();
  while (n > 0)
    {
      indexStack = ptlinstances_remove_last_elem (indexStack, &x);
      n--;
    }
}


/**
   It generates n indexes of the stack of indexes.
*/
_VIRTUAL_ void
cGenerator::genStackIndexes (int n)
{
  int nbIterator, i;

  if (n > 0)
    {
      nbIterator = getIndexStackSize ();
      if (nbIterator > n)
	i = nbIterator - n;
      else
	i = 0;
      genLindexes (i, n);
    }
}

/**
  Generation of a list of indexes.
  @param nbGeneratedIndex (in) gives the number of generated indexes (before the call)
  @param nbIndexToGenerate (in) gives the upper bound of the interval [nbGeneratedIndex+1, nbIndexToGenerate].
  @param ldim  (inout) list of dimensions; all the indexes are generated. Empty list at the end.
  @param lparc (inout) list of Signal travels; all the indexes are generated. Empty list at the end.
*/
_VIRTUAL_ void
cGenerator::genListOfIndexes (ptexpcal ldim, ptexpcal lparc, int nbGeneratedIndex, int nbIndexToGenerate)
{
  ptexpcal vindex;

  genRecListIndexes (ldim, 1);

  while (lparc != NILEXPCAL)
    {
      vindex = EM->getBinaryArgstAt (EM->getNaryArgs (&lparc), 1);
      printIndexAccess (vindex);
      // TextForm->printLeftSquareBracket(); genexpression(expindice, opnullary); TextForm->printRightSquareBracket();
    }

  if (!extensionInhibit)
    genStackIndexes (nbIndexToGenerate - nbGeneratedIndex);
}

/**
   @return "true" if an expression (exp) is attributed with indexes, "false" otherwise.
   
   It generates the expression when it is attributed with indexes: 
   - the name of the identifier (exp)
   - n indexes of array (n = number of dimensions of exp). These indexes are assumed available in the stack of indexes (indexStack)
   (genOpenTravelIndexes()).
*/
boolean cGenerator::genTravelIndex (ptexpcal exp)
{
  ptlinstances
    l;
  ptexpcal
    vindex,
    exp1;
  int
    i,
    n;

  exp = getExpressionToGenerate (exp);
  if (!EM->isMarked (exp, travelIndexAttribute))
    return pk_false;

  if (EM->isMarked (exp, OPTIMIZED_INDEX))
    {
      if (!EM->getAttribute (exp, IteratorAttribute, (attributeValue *) & exp1))
	return pk_false;
      genexpression (exp1, opnullary);
      return pk_true;
    }
  genPrefixedIdentifier (exp);

  n = EM->getNumberOfDimensions (exp);
  if (n > 0)
    {
      l = indexStack;
      for (i = 0; i < n; i++)
	{
	  ptlinstances_get_next (&l, &vindex);
	  printIndexAccess (vindex);
	}
    }
  return pk_true;
}

/**
   For an optimized index (exp) (see isOptimizedIndex())
   - it generates the n-th index of the stack of index (indexStack) assumed available.
   - otherwise, it generates the expression (genexpression()).
*/
_VIRTUAL_ void
cGenerator::genIndice (ptexpcal exp, int n)
{
  ptlinstances l;
  ptexpcal x;
  int i, m;

  if (!isOptimizedIndex (exp))
    genexpression (exp, opnullary);
  else
    {
      m = EM->getNumberOfDimensions (exp);
      if (n > m)
	n = m;
      l = indexStack;
      for (i = 1; i < n; i++)
	ptlinstances_get_next (&l, &x);

      ptlinstances_get_next (&l, &x);
      genPrefixedIdentifier (x);
    }
}

/**
   @return n + the number of generated indexes in this method.
   It generates all the indexes of a list (lindexes) starting at the n-th dimension.
   lindexes ::= [E1],...,[Ev]. 
   If an element is a list, it recalls recursively this method, otherwise it calls genIndice() method.
*/
_VIRTUAL_ int
cGenerator::genRecListIndexes (ptexpcal lindexes, int n)
{
  ptexpcal exp;

  while (EM->getNaryArgs (&lindexes, &exp))
    {
      if (EM->satisfiesList (exp))
	n = genRecListIndexes (exp, n);
      else
	{
	  TextForm->printLeftSquareBracket ();
	  genIndice (exp, n);
	  TextForm->printRightSquareBracket ();
	  n++;
	}
    }
  return n;
}

/**
   It generates the code for a reference to the predefined "indices" function.
   The number of outputs is equal to the number of inputs (a list of integers).
   \code
   Example : for a Signal specification, 
     (i, j):= indices(3,8) with [3,8] integer i,j. It stands for 
     a = [0, 2], b = [0, 7],
            i[a][b] = a
            j[a][b] = b
   \endcode
*/
boolean cGenerator::genIntrinsicfuncIndices (ptexpcal inputs, ptexpcal outputs)
{
  ptexpcal
    aux,
    x,
    vindex;
  int
    n;
  ptlinstances
    l;

  aux = outputs;
  EM->getNaryArgs (&aux, &x);
  n = genOpenIteration (x);
  l = indexStack;
  TextForm->printblockopening ();
  while (EM->getNaryArgs (&outputs, &x))
    {
      gensig (x, pk_false);
      TextForm->printassign ();
      ptlinstances_get_next (&l, &vindex);
      genPrefixedIdentifier (vindex);
      TextForm->printSemiColonLine ();
    }
  TextForm->printblockclosing ();
  TextForm->backindent (n);
  genCloseIteration (n);
  return pk_true;
}

/**
   It resets (unmarks) the optimized indexes associated with exp previously set by setOptimizedIndexes() method.
   (an internal OPTIMIZED_INDEX is used).
*/
void
cGenerator::resetOptimizedIndexes (ptexpcal exp)
{
  ptlinstances loptimizedIndices;
  ptexpcal vindex;

  loptimizedIndices = EM->getOptimizedTravelIndexes (exp);
  while (ptlinstances_get_next (&loptimizedIndices, &vindex))
    EM->removeMark (vindex, OPTIMIZED_INDEX);
}

/**
   It marks the optimized indexes associated with exp.
   (an internal OPTIMIZED_INDEX is used).
*/
void
cGenerator::setOptimizedIndexes (ptexpcal exp)
{
  ptlinstances loptimizedIndices;
  ptexpcal vindex;

  loptimizedIndices = EM->getOptimizedTravelIndexes (exp);
  while (ptlinstances_get_next (&loptimizedIndices, &vindex))
    EM->addMark (vindex, OPTIMIZED_INDEX);
}

/** 
    @param l (in) is a cartesian product expression  (<<I1,... In>> operator).

    It generates the opening of n loops where n is the 
    number of elements of a cartesian product (<<I1,... In>> operator).
    For each element, it generates (genOpenLoopForDim()) the opening of a loop that 
    takes into account the size of the dimension of the element.
    It associates the created index to the element. 
    The element is also marked by travelIndexAttribute attribute.
    \code
    Example:
    For <<0..4, 0..2>>, it generates the (pseudo) code
    for L = 0, 4;
    for U = 0, 2;
   \endcode
   @return the number of elements of l (i.e. number of opened loops)
*/
int
cGenerator::genLoopForList (ptexpcal l)
{
  ptexpcal vexp, ithDimExpr, assIndex;
  int n;
  // l is list of expression. (cartesian product) ; each element is a vector.
  n = 0;
  while (EM->getNaryArgs (&l, &vexp))
    if (EM->getIthDimension (vexp, 1, &ithDimExpr))
      {
	n++;
	assIndex = genOpenLoopForDim (ithDimExpr);
	EM->addAttribute (vexp, IteratorAttribute, (attributeValue) assIndex);
	EM->addMark (vexp, travelIndexAttribute);
      }
  return n;
}

/**  
     @param lexp (in) is a cartesian product expression (<<I1,... In>> operator).
     
     It assumes that the genLoopForList() method has been previously applied (the loops are opened).

     It creates and generates the definition of an internal variable (x) by \n
     <em> x  = II1 * sizeOf(I2) * ...* izeOf(IN) + II1 * sizeOf(I3) * ...* izeOf(IN)  + ... + IIN </em> \n
     where IIj are the associated indexes with I1, ..., In.
     
     The definition of x is produced by the mkSelectorDefinition() method.

     @return the  internal created variable (x).
     
     \code
     Example:  <<0..4, 0..2>>, with the following associated indexes
     for L = 0, 4;
     for U = 0, 2;
     
     this method generates the following assignment
     X = 3L + U
     \endcode
*/
_VIRTUAL_ ptexpcal
cGenerator::genDefIndex (ptexpcal lexp)
{
  ptexpcal vind, indexExpr;

  if (EM->getSize (lexp) == 1)
    EM->getNaryArgs (&lexp, &vind);
  else
    {
      vind = getInternalIndex ();
      indexExpr = mkSelectorDefinition (lexp, indexStack);
      genPrefixedIdentifier (vind);
      TextForm->printassign ();
      genexpression (indexExpr, opnullary);
      TextForm->printSemiColonLine ();
    }
  return vind;
}

/**
   @param lexp (in) is a cartesian product expression (<<I1,... In>> operator).
   @param OpenedBlockList (in) a list of indexes already opened (assumed).
   @return the expression <em>  II1 * sizeOf(I2) * ...* izeOf(IN) + II1 * sizeOf(I3) * ...* izeOf(IN)  + ... + IIN </em> \n
   where IIj are elements of OpenedBlockList associated indexes with I1, ..., In.
*/
_VIRTUAL_ ptexpcal
cGenerator::mkSelectorDefinition (ptexpcal lexp, ptlinstances OpenedBlockList)
{
  ptexpcal vexp, vl, vres, vnext, ithDimExpr, vindex, vnb;

  if (!EM->getNaryArgs (&lexp, &vexp))
    return NILEXPCAL;
  ptlinstances_get_next (&OpenedBlockList, &vindex);
  vl = mkSelectorDefinition (lexp, OpenedBlockList);
  if (vl == NILEXPCAL)
    return vindex;
  if (!EM->getNaryArgs (&lexp, &vnext))
    return vindex;

  EM->getIthDimension (vnext, 1, &vnb);
  while (EM->getNaryArgs (&lexp, &vnext))
    {
      EM->getIthDimension (vnext, 1, &ithDimExpr);
      vnb = EM->mkBinaryExp (opmult, vnb, ithDimExpr, NILBACKANNOTATION);
    }
  vres = EM->mkBinaryExp (opmult, vindex, vnb, NILBACKANNOTATION);
  return EM->mkBinaryExp (opplus, vres, vl, NILBACKANNOTATION);
}

/**
   It generates the selection of a element of an array, given by a list of indexes (lindexes).
*/
_VIRTUAL_ void
cGenerator::genAccessElementArray (ptexpcal lindexes)
{
  ptexpcal vindex;
  while (EM->getNaryArgs (&lindexes, &vindex))
    printIndexAccess (vindex);
}
