/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/** 
    Generation of code for each cluster. Each cluster is a Java thread.
    In separated mode, entry points are generated with parameters, otherwise without parameters ; in this case, 
    the i/o variables are updated by side effect
*/
  _VIRTUAL_ void
threadsJavaGenerator::genSubGraphs ()
{
  integer i;
  ptINode aSubGraph;
  ptinsth h;
  ptexpcal lins, louts;
  ptlINode ld;
  ptident name;

  if (nbSets == 1)
    return;

  for (i = 1; i < nbSets; i++)
    {
      aSubGraph = SubGraphTable[i]->aSubGraph;
      genEntryPointSubGraphHeader (aSubGraph, pk_false);
      NM->get_inOutGraphRef (aSubGraph, &lins, &louts);
      NM->get_clusterName (aSubGraph, &name);

      ((inteprintjavathread *) TextForm)->genBeginCluster (name);

      h = NM->get_nodeClock (aSubGraph);
      if (SEPARATE)
	genAssignIOSeparate (lins, h, pk_true);

      setUndefinedValueForBooleanList (h, NM->getInodelLocalFlows (aSubGraph));
      genOrderedNodes (SubGraphTable[i]->ln, h, pk_false, pk_false);
      if (existPredefinedPragma (NM->getPragmas (aSubGraph), PRAGMA_DELAYCLUSTER))
	{
	  ld = SubGraphTable[i]->ldelays;
	  genOrderedNodes (ld, h, pk_false, pk_false);
	  genUpdateStateVariables ();
	  if (SEPARATE)
	    genAssignIOSeparate (louts, h, pk_false);
	  printCallEntrypointStepInitialize ();
	}
      else if (SEPARATE)
	genAssignIOSeparate (louts, h, pk_false);

      ((inteprintjavathread *) TextForm)->genEndCluster ();
    }
}


/** 
    Code generation of the file foo.java, with foo the name of the current graph.
    It generates a part of the declaration of the class foo.

    It calls genBodyClassBegin(), genDclParametersAndIndexes(), 
    genDclrArea(), generateConstructor(), printJavaAreaConstructor(), genBodyClassEnd().
    \code
    The code looks like 
    class foo
    {
         private foo_io _io_;
	 private  KeyboardReader KR = new KeyboardReader();
	 public foo_Area XArea =new foo_Area();
	 
	public foo(){
	     _io_=new foo_io ();
        }

	public foo(foo_Area Xa){
         XArea = Xa;}

	 public void openIO(){
        _io_.foo_OpenIO();
	}

	public void closeIO(){
        _io_.foo_CloseIO();
	}
    \endcode
*/
_VIRTUAL_ void
threadsJavaGenerator::genBodyClass ()
{
  ((inteprintjavathread *) TextForm)->genBodyClassBegin (progName);

  genDclParametersAndIndexes ();

  if (nbSets != 1)
    {
      ((inteprintjavathread *) TextForm)->genDclrArea ();
      generateConstructor ();
      ((inteprintjavathread *) TextForm)->printJavaAreaConstructor (progName);
    }
  else				// never.
    {
      cGenerator::gendeclarationsBody ();
      generateConstructor ();
    }
  ((inteprintjavathread *) TextForm)->printJavaIOMethods ();
  if (externalObjects)
    ((inteprintjava *) TextForm)->printJavaSetExternalObject ();
}

/** 
    Code generation of the file foo.java
    - the entries: foo_initialize(), foo_STEP_initialize(), (genInitialisations())
    - the scheduler (genScheduler())
    - an entry point for each cluster (genSubGraphs()), one thread by cluster.
    - the shared object of the threads are genrated in genClassArea().
*/
_VIRTUAL_ void
threadsJavaGenerator::genBodyCode ()
{
  genInitialisations ();
  if (nbSets == 1)
    {				// never.
      genSubGraphs ();
      genScheduler ();
      TextForm->genClassEnd ();
    }
  else
    {
      genScheduler ();
      ((inteprintjavathread *) TextForm)->genRunFunction (progName);
      TextForm->genClassEnd ();

      genClassArea ();
      genSubGraphs ();
    }
  TextForm->closefile ();
}


/** 
    Generation of the Area class that contains the signals of the program shared by the threads.
    \code
    Such a code looks like (foo is the name of the graph)
    class foo_Area{
          public  KeyboardReader KR = new KeyboardReader();
          //  ==> parameters and indexes
          //  ==> input signals     
          public float a, b, c;
          //  ==> output signals
	  public float x2, x1;
	  public boolean error;
	  //  ==> local signals
	  public boolean OK;
	  public float bb, cc, dd, yy, next_yy;
	  public boolean biterate;
	  ...
	  }
     \endcode
*/
_VIRTUAL_ void
threadsJavaGenerator::genClassArea ()
{
  ((inteprintjavathread *) TextForm)->genClassAreaBegin ();
  cGenerator::gendeclarationsBody ();
  ((inteprintjavathread *) TextForm)->genClassAreaEnd ();
}

/** 
    Generation of the header of a procedure characterized by its name (name) its inputs (inputs)  and its outputs (outputs).
    This procedure is an external one when bExternal or a static one othewise. 
    It is the declaration when declaration, its definition otherwise.. 
*/
_VIRTUAL_ void
threadsJavaGenerator::genEntryPointDeclaration (ptident name, ptexpcal inputs, ptexpcal outputs, boolean declaration, boolean bExternal)
{
  if (bExternal)
    genEntryPoint (name, inputs, outputs, pk_false, pk_true, declaration);
  else
    ((inteprintjavathread *) TextForm)->genEntryPointDeclaration (name, declaration);
}



/** 
    This method generates code for a reference to the subgraph (aSubGraph) when it is a cluster one. 
    Such a reference is a procedure call with or without parameters according to the context 
    given by the parameters blackBox and schedulerTrain: when blackBox="false" or schedulerTrain="true", 
    the procedure is called without parameters. 
*/
_VIRTUAL_ void
threadsJavaGenerator::genRefSubGraphNode (ptINode aSubGraph, boolean blackBox, boolean schedulerTrain)
{
  ptident nom;
  ptexpcal linputs, loutputs;

  if (NM->is_trainCluster (aSubGraph))
    {
      genBlockOpeningOf (NM->get_nodeClock (aSubGraph));
      NM->get_clusterName (aSubGraph, &nom);
      ((inteprintjavathread *) TextForm)->genRefSubGraphNode (nom);
      if (!blackBox)
	NM->get_inOutGraphRef (aSubGraph, &linputs, &loutputs);
      else
	{
	  linputs = NILEXPCAL;
	  loutputs = NILEXPCAL;
	}
      if (schedulerTrain && !blackBox)
	{
	  linputs = NILEXPCAL;
	  loutputs = NILEXPCAL;
	}
      if (!schedulerTrain)
	genOutputsExternalNode (nom, loutputs, genInputsExternalNode (linputs, pk_true));
    }
}
