/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   Specialisation: return "true" if the node does not be generated, "false" otherwise.
   Examples: 
     - Sigali nodes,
     - for assertion calls when "no code" is asked for assertions.
*/
  _VIRTUAL_ boolean ColouringGenCode::filteredNode (ptINode node)
{
  ptexpcal v;
  
  // Sigali order is unmarked.
  if (existPredefinedPragma (NM->getPragmas (node), PRAGMA_SIGALI))
    return pk_true;
  // Default value x::= default value (x$1)
  if (NM->isDefaultValueNodeDelay (node, &v))
    return pk_true;
  // Assert()
  if (bCheck)
    return pk_false;
  return NM->is_intrinsicNode (node, idAssert);
}

/** 
    Specialisation: equivalent formula are not marked for code generation.
*/
_VIRTUAL_ void
ColouringGenCode::markConstraintsAndProvedEquiv (ptinsth vh)
{
  markClass (vh, pk_true);
}


/**
   @return "true" if exp = not(cond) && OtherExpr, "false" otherwise.
*/
_VIRTUAL_ boolean
ColouringGenCode::is_andNotCondition (ptexpcal exp, ptexpcal cond, ptexpcal * OtherExp)
{
  ptexpcal exp1, exp2;

  if (EM->getOperator (exp) != opet)
    return pk_false;
  EM->getBinaryArgs (exp, &exp1, &exp2);
  if (EM->getOperator (exp1) == opnon)
    {
      exp1 = EM->getUnaryArg (exp1);
      if (exp1 == cond)
	/*if (!satifies_memory_optimized(exp2)) */
	{
	  *OtherExp = exp2;
	  return pk_true;
	}
    }
  if (EM->getOperator (exp2) == opnon)
    {
      exp2 = EM->getUnaryArg (exp2);
      /*if (exp2 == cond) if (!satifies_memory_optimized(exp1)) */
      {
	*OtherExp = exp1;
	return pk_true;
      }
    }
  return pk_false;
}

/**
   @return "true" if h = K - hleft, "false" otherwise
   @param h (in) a clock.
   @param hleft (in) a clock.
   @param exp (out) when  h = k - hleft.
   \note : this text must be similar in the code generators.
*/
_VIRTUAL_ boolean
ColouringGenCode::isOptimizedClock (ptinsth h, ptinsth hleft, ptinsth * k)
{
  ptexpcal exp, condLeft, cond, expd;
  ptINode aNode;

  if (CKM->satisfiesUndefClock (hleft))
    return pk_false;
  condLeft = CKM->getConditionSub (hleft);
  cond = CKM->getConditionSub (h);
  if (!EM->IsSignal (cond))
    return pk_false;
  aNode = EM->getNode (cond);
  if (!NM->get_rightOperand (aNode, &expd))
    return pk_false;
  expd = EM->getBasicExpr (expd);
  // si expd = not(condLeft) && exp alors on engendre if (exp)
  if (!is_andNotCondition (expd, condLeft, &exp))
    return pk_false;
  if (!EM->IsSignal (exp))
    return pk_false;
  return EM->getExtractionLink (exp, k);
}


/**
   It marks a default expression (vexp), left argument of a cell if bcell.
   For the code generation, an expression such that \n
    X:= (E1 when C1) default (E2 when (not C1 and C2)) default ...\n
   will be generated
       if C1 X:=E1; else if C2 X:=E2...\n
   so, some clocks are not used.
     
*/
_VIRTUAL_ void
ColouringGenCode::markDefaultExpr (ptexpcal vexp, boolean bcell)
{
  ptexpcal vexpd, vexpg, vhorlexpcal, vexpLast;
  ptinsth h, hleft, k;

  markClock (EM->getContributionClock (vexp));
  vexpLast = EM->getDefaultArgstAt (vexp, 2);	// E when Clk
  markExpression (EM->getWhenOrCellArg (vexpLast));
  vhorlexpcal = EM->getDefaultArgstAt (vexp, 1);
  vexpg = vhorlexpcal;
  while (EM->getOperator (vexpg) == opmele)
    {
      vhorlexpcal = EM->getDefaultArgstAt (vexpg, 1);
      vexpd = EM->getDefaultArgstAt (vexpg, 2);
      h = EM->getContributionClock (vexpd);
      hleft = EM->getContributionClock (vhorlexpcal);
      if (CKM->getBooleanClockAt (hleft, CKM->getBranching (hleft, currentClock), &k)) hleft = k;

      if (isOptimizedClock (h, hleft, &k))
	{
	  markExpression (EM->getWhenOrCellArg (vexpd));
	  markClock (k);
	}
      else
	markExpression (vexpd);
      vexpg = EM->getWhenOrCellArg (vhorlexpcal);
    }
  markClockExpression (EM->getContributionClock (vhorlexpcal));
  markExpression (vhorlexpcal);
}

/** 
    Specialisation: the optimized indexes are NOT marked.
*/
_VIRTUAL_ void
ColouringGenCode::markReferencedSignalsIn (ptINode node)
{
  ptlinstances ls, lopt;
  ptexpcal vdef;
  boolean b;

  b = pk_false;
  if (NM->is_equation (node) || NM->is_partialDefinition (node))
    {
      if (NM->get_rightOperand (node, &vdef))
	if (!EM->IsCellExpr (vdef))
	  {
	    vdef = EM->getBasicExpr (vdef);
	    b = pk_true;
	  }
    }
  else if (NM->is_intrinsicNode (node, getCartesianProductId ()))
    {
      NM->get_rightOperand (node, &vdef);
      b = pk_true;
    }
  ls = NM->getInodelInputFlows (node);
  if (b)
    {
      lopt = EM->mkOptimizedIndexes (vdef);
      ls = ptlinstances_minus (ls, lopt);
    }
  marklsig (ls);
}


void
ColouringGenCode::unMarkTickSignal ()
{
  ptexpcal vTick;

  if (CKM->isClockRepresentedBySignal (CKM->getTickClock (), &vTick))
    NM->removeMark (EM->getNode (vTick), aMark);
}
