/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
   @return the type of a polynomial (p).
*/
extern typolynomial
getpolytype (ptpolynomial p)
{
  return p->nature;
}


/** @return the value of an assumed constant polynomial (p). */
extern int
getpolyvalue (ptpolynomial p)
{
  return p->U.value;
}

/** @return the name of an assumed variable polynomial (p). */
extern ptident
getpolyname (ptpolynomial p)
{
  return p->U.name;
}



/** @return the first operand of an assumed binary polynomial (p).*/
extern ptpolynomial
getop1 (ptpolynomial p)
{
  return p->U.binary.op1;
}


/** @return the second operand of an assumed binary polynomial (p).*/
extern ptpolynomial
getop2 (ptpolynomial p)
{
  return p->U.binary.op2;
}

/** @return the operand of an assumed unary polynomial (p).*/
extern ptpolynomial
getop (ptpolynomial p)
{
  return p->U.op;
}

/** @return the polynomial associated with a constant c (-1,0,1).*/
extern ptpolynomial
polyconstant (int c)
{
  return tabConstants[c + 2];
}

/** @return the polynomial associated with an identifier (id). */
extern ptpolynomial
polyvariable (ptident id)
{
  ptpolynomial poly;

  poly = new polynomial ();
  poly->nature = variable;
  poly->U.name = id;
  return poly;
}

/** @return a new binary polynomial with \e n as operator, \e p1 the fisrt child and \e p2 the second one.*/
extern ptpolynomial
polybinary (typolynomial n, ptpolynomial p1, ptpolynomial p2)
{
  ptpolynomial poly;

  poly = new polynomial ();
  poly->nature = n;
  poly->U.binary.op1 = p1;
  poly->U.binary.op2 = p2;
  return poly;
}



/** @return a new unary polynomial with \e n as operator, \e p1 its child.*/
extern ptpolynomial
polyunary (typolynomial n, ptpolynomial p1)
{
  ptpolynomial poly;

  poly = new polynomial ();
  poly->nature = n;
  poly->U.binary.op1 = p1;
  return poly;
}


/** @return "true"  if p=0, "false" otherwise. */
extern boolean
ispolynul (ptpolynomial p)
{
  if (p == NILPOLYNOMIAL)
    return pk_true;
  if (getpolytype (p) == constpoly)
    return (boolean) (getpolyvalue (p) == 0);
  return pk_false;
}



/** @return polyWhen p  */
extern ptpolynomial
polywhen1 (ptpolynomial p)
{
  return polyunary (polyWhen1, p);
}



/** @return not(p) */
extern ptpolynomial
polynot (ptpolynomial p)
{
  typolynomial t;

  t = getpolytype (p);
  if (t == constpoly)
    return tabConstants[-getpolyvalue (p) + 2];
  if (setpolyNotpolyMinus (p))
    return getop (p);
  return polyunary (polyNot, p);
}



/** @return -p  */
extern ptpolynomial
polyminus (ptpolynomial p)
{
  typolynomial t;

  t = getpolytype (p);
  if (t == constpoly)
    return tabConstants[2 - getpolyvalue (p)];
  if (setpolyNotpolyMinus (p))
    return getop (p);
  return polyunary (polyMinus, p);
}



/** @return P^2  */
extern ptpolynomial
polysquare (ptpolynomial p)
{
  typolynomial t;
  ptpolynomial pp;

  t = getpolytype (p);
  if (setpolyNotpolyMinus (p))
    {
      pp = getop (p);
      t = getpolytype (pp);
    }
  else
    pp = p;
  if (t == constpoly)
    return tabConstants[getpolyvalue (pp) * getpolyvalue (pp) + 2];
  if (t == polySquare)
    return pp;
  return polyunary (polySquare, pp);
}



/** @return p1 * p2  */
extern ptpolynomial
polymult (ptpolynomial p1, ptpolynomial p2)
{
  typolynomial t1, t2;
  boolean neg;
  ptpolynomial pp1, pp2, poly;

  t1 = getpolytype (p1);
  if (t1 == polyMinus)
    {
      pp1 = getop (p1);
      t1 = getpolytype (pp1);
      neg = pk_true;
    }
  else
    {
      pp1 = p1;
      neg = pk_false;
    }
  t2 = getpolytype (p2);
  if (t2 == polyMinus)
    {
      pp2 = getop (p2);
      t2 = getpolytype (pp2);
      neg = (boolean) (!neg);
    }
  else
    pp2 = p2;
  if (t1 == constpoly)
    if (t2 == constpoly)
      poly = tabConstants[getpolyvalue (pp1) * getpolyvalue (pp2) + 2];
    else
      switch ((int) (getpolyvalue (pp1)))
	{
	case 0:
	  poly = zero;
	  break;
	case 1:
	  poly = pp2;
	  break;
	case -1:
	  poly = pp2;
	  neg = (boolean) (!neg);
	  break;
	default:;
	}
  else if (t2 == constpoly)
    switch ((int) (getpolyvalue (pp2)))
      {
      case 0:
	poly = zero;
	break;
      case 1:
	poly = pp1;
	break;
      case -1:
	poly = pp1;
	neg = (boolean) (!neg);
	break;
      default:;
      }
  else if (t2 == polyWhen1)
    poly = polybinary (polyWhen, pp1, getop (pp2));
  else
    poly = polybinary (polyMult, pp1, pp2);

  if (neg)
    return polynot (poly);
  return poly;
}



/** @return P1 + P2  */
extern ptpolynomial
polysum (ptpolynomial p1, ptpolynomial p2)
{
  typolynomial t1, t2;

  t1 = getpolytype (p1);
  t2 = getpolytype (p2);
  if (t1 == constpoly)
    {
      if (t2 == constpoly)
	return tabConstants[getpolyvalue (p1) + getpolyvalue (p2) + 2];
      if (getpolyvalue (p1) == 0)
	return p2;
      if (t2 == polyMinus)
	return polybinary (polyDiff, p1, getop (p2));
      return polybinary (polySum, p1, p2);
    }
  if (t2 == constpoly)
    switch ((int) (getpolyvalue (p2)))
      {
      case 0:
	return p1;
      case 1:
	return polybinary (polySum, p1, un);
      case -1:
	return polybinary (polyDiff, p1, un);
      default:;
      }
  if (t2 == polyMinus)
    return polybinary (polyDiff, p1, getop (p2));
  return polybinary (polySum, p1, p2);
}



/** @return P1 - P2  */
extern ptpolynomial
polydifference (ptpolynomial p1, ptpolynomial p2)
{
  typolynomial t1, t2;

  t1 = getpolytype (p1);
  t2 = getpolytype (p2);
  if (t1 == constpoly)
    {
      if (t2 == constpoly)
	return tabConstants[getpolyvalue (p1) - getpolyvalue (p2) + 2];
      if (getpolyvalue (p1) == 0)
	return polyminus (p2);
      if (t2 == polyMinus)
	return polybinary (polySum, p1, getop (p2));
      return polybinary (polyDiff, p1, p2);
    }
  if (t2 == constpoly)
    switch ((int) (getpolyvalue (p2)))
      {
      case 0:
	return p1;
      case 1:
	return polybinary (polyDiff, p1, un);
      case -1:
	return polybinary (polySum, p1, un);
      default:;
      }
  if (t2 == polyMinus)
    return polybinary (polySum, p1, getop (p2));
  return polybinary (polyDiff, p1, p2);
}



/** @return P1 default P2  */
extern ptpolynomial
polydefault (ptpolynomial p1, ptpolynomial p2)
{
  if (ispolynul (p1))
    return p2;
  if (ispolynul (p2))
    return p1;
  return polybinary (polyDefault, p1, p2);
}



/** @return P1 polyWhen P2  */
extern ptpolynomial
polywhen (ptpolynomial p1, ptpolynomial p2)
{
  if (ispolynul (p1))
    return p2;
  if (ispolynul (p2))
    return p1;
  if (getpolytype (p2) == polyWhen1)
    return polybinary (polyWhen, p1, getop (p2));
  return polybinary (polyWhen, p1, p2);
}



/** @return P1 and P2  */
extern ptpolynomial
polyand (ptpolynomial p1, ptpolynomial p2)
{
  return polybinary (polyAnd, p1, p2);
}



/** @return P1 = P2  */
extern ptpolynomial
polyequa (ptpolynomial p1, ptpolynomial p2)
{
  return polybinary (equation, p1, p2);
}

/** @return P1 or P2  */
extern ptpolynomial
polyor (ptpolynomial p1, ptpolynomial p2)
{
  return polybinary (polyOr, p1, p2);
}



/** Initializing: it creates the representation of the z3z constants (-1,0,1). */
extern void
initconstants (void)
{
  zero = makeconstant (0);
  un = makeconstant (1);
  polyMinusun = makeconstant (-1);
  tabConstants[0] = un;
  tabConstants[1] = polyMinusun;
  tabConstants[2] = zero;
  tabConstants[3] = un;
  tabConstants[4] = polyMinusun;
}
