/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PROTECTED_
/**
   @return "true" if a clock (vh) is selected for the current translation, "false" otherwise.
 */
  boolean DCG2Syndex::isSelectedClock (ptinsth vh)
{
  return !ptlinsth_sorted_isin_elem (unClocks, vh);
}

/** Array managment.
   @return true if the clock is an internal clock of current array block.   
*/
boolean DCG2Syndex::isClockOfCurrentBlock (ptinsth vClk)
{
  ptINode
    cn;

  if (!isSelectedClock (vClk))
    return pk_false;
  CKM->getInComposedNode (vClk, &cn);
  return (cn == ArrayBlock);
}


/** 
    Translating a subgraph (vgfd): a set of nodes assigned to a clock. 
    Two steps : 
    - the declaration (declareReferences()) of Syndex objects (references); 
    - the connections (mkConnectionsNodes()).
    - the distribution constraints  (genRunOnPartionning()).
    The attribute member lwrites is reset before the first step.
    During the translation of the nodes of the grapg(vgfd) the set of the produced syndex operations is kept in the member SyndexOperations 
    when the specification contains distribution constraints.
*/
_VIRTUAL_ void
DCG2Syndex::translateSubGraph (ptgfd vgfd)
{
  ptlINode ln;

  TRACE (cout << endl << "A new subclock";
    );
  ln = ptgfd_get_nodes (vgfd);
  lwrites = NILLINODE;
  declareReferences (ln);	// References (first  Step)
  mkConnectionsNodes (ln);	// connections (second Step)
  if (isRunOnPartionning)
    {
      genRunOnPartionning (ln);	// distribution constraints (third Step)
      SyndexOperations.clear ();
    }
}

/**
  Initializing \ref translateSubGraph algorithm for a clock (aClock) : 
  - a Syndex-algorithm is built ; its becomes the current Syndex-algorithm in the Tree Builder Interface (TBI).
  - and global informations about interface are set (inputs, outputs, parameters).
*/

_VIRTUAL_ void
DCG2Syndex::translateGraphClockInit (ptinsth aClock)
{
  ptinterface interface;
  ptrepstr name;

  addClockTreePost = pk_false;	//for each clock, the default value is to add "pre"

  CurrentSubGraph++;
  if (aClock == TheTick)
    name = ident2repstr (graphName);
  else
    name = getStrClockName (aClock);
  TBI->mkSubGraphSkeletonAlgorithm (name, NILTREE);
  // Interface infos
  interface = CKM->getInterfaceSubHierarchy (aClock);
  inputs = ptinterface_get_inputSignals (interface);
  outputs = ptinterface_get_outputSignals (interface);
  parameters = ptinterface_get_parameters (interface);
}

/**
   It finalyzes the translation of the sub-graph associated with a clock (the current one).
   The tree is pre/post added in the Syndex Tree program, according to the value of addPostClock.
   Moreover, it cleans some marks: equivAttribute, vConnectionMark.
*/
_VIRTUAL_ void
DCG2Syndex::translateGraphClockFinal (boolean addPostClock)
{
  TBI->addCurrentClock (addPostClock);
  EM->resetMarkAll (equivAttribute);
  EM->resetMarkAll (vConnectionMark);
}

/**
   @return "true" if a sub-graph (vgfd) must be visited, "false" otherwise (not nodes for the Syndex program).
 */
_VIRTUAL_ boolean
DCG2Syndex::gfd_isSelected (ptgfd vgfd)
{
  ptlINode ln;
  ptINode n;

  ln = ptgfd_get_nodes (vgfd);
  if (ln == NILLINODE)
    return pk_false;
  while (ptlINode_get_elem (&ln, &n))
    if (!NM->is_readNode (n))
      if (!NM->is_writeNode (n))
	return pk_true;
  return pk_false;
}

/** Translating for a Clock (Definition and subgraph).
   @param aClock (in) the current clock of a hierarchy.
   @return true is the sub-graph assigned to aClock is not empty, false otherwise.
*/
_VIRTUAL_ boolean
DCG2Syndex::translateGraphClock (ptinsth aClock)
{
  ptgfd vgfd;
  ptexpcal vcond;
  int code;

  vgfd = CKM->getClockSubGraph (aClock);	// Subgraph assigned to the clock aClock
  if (gfd_isSelected (vgfd))
    {
      translateGraphClockInit (aClock);
      PushInterfaceDeclarations ();
      vcond = CKM->getConditionSub (aClock);
      if (CKM->getOperator (aClock) == sub)
	code = 1;
      else
	code = 0;
      TBI->addCondition (getRepstrIdName (vcond), code);
      translateSubGraph (vgfd);	// references; connexions; partitionning constraints
      PopInterfaceDeclarations ();
      translateGraphClockFinal (addClockTreePost);
      return pk_true;
    }
  return pk_false;
}

/** 
    Translating for a Clock (Definition and subgraph).
    @param aClock (in) the current clock of a hierarchy.
*/
_VIRTUAL_ boolean
DCG2Syndex::translateMainGraphClock (ptinsth aClock)
{
  ptgfd vgfd;

  vgfd = CKM->getClockSubGraph (aClock);	// Subgraph assigned to the clock aClock
  if (ptgfd_satisfies_empty (vgfd))
    return pk_false;

  if (asMain)
    {
      TBI->mkSubGraphSkeletonAlgorithm (getStrClockName (aClock), mkProgramParameters ());
      CurrentSubGraph++;
      // Interface infos
      inputs = NILLINSTANCES;
      outputs = NILLINSTANCES;
      parameters = NILLINSTANCES;
      TBI->addTrueCondition ();
      translateSubGraph (vgfd);	// references; connexions; partitionning constraints
      EM->resetMarkAll (equivAttribute);
      EM->resetMarkAll (vConnectionMark);
    }
  else
    {
      translateGraphClockInit (aClock);
      PushInterfaceDeclarations ();
      TBI->addTrueCondition ();	// pk_true
      translateSubGraph (vgfd);	// references; connexions; partitionning constraints
      // PopInterfaceDeclarations();  must be kept in comment!!!
      translateGraphClockFinal (addClockTreePost);
    }
  TBI->setCurrentClockAlgo (TBI->getCurrentAlgo ());	//set main algorithm
  return pk_true;
}


/**
   For a clock (aClock), it defined the name of the algorithm Syndex. This name is then used to build an attribute
   that is associated with each output of the sub-graph  (setReferenceTo(ptexpcal, ptSyndexAttribute)).
   Such an attribute (mkSyndexAttribute()) is composed of: 
   - the name of the algorithm, 
   - the name of the output, 
   - number of the sub-graph.
*/
_VIRTUAL_ void
DCG2Syndex::setDefinedInSubGraph (ptinsth aClock)
{
  ptexpcal elem;
  ptrepstr id;
  ptlinstances l;

  id = TBI->makeInstSubGraphName (getStrClockName (aClock));
  l = ptinterface_get_outputSignals (CKM->getInterfaceSubHierarchy (aClock));
  while (ptlinstances_get_next (&l, &elem))
    setReferenceTo (elem, mkSyndexAttribute (id, getRepstrIdName (elem)));
}
