/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/



/**
   It assigns an attribute to an expression.
   This attribute is a triplet (the name of the Syndex operator associated with a signal operator, ".o", 
   the index of the current subgraph).
   \see getLanguageOperator(ptexpcal), mkSyndexAttribute() .
*/
_VIRTUAL_ void
DCG2Syndex::setReferenceToForOperatorOf (ptexpcal e)
{
  setReferenceTo (e, mkSyndexAttribute (getLanguageOperator (e), MyStdOutput));
}

/** 
    Giving the Nth usage of a Signal operator (op).
    @param op (in) the operator of signal.
    @param  Nth (out) the Nth usage of signal operator if the operator is known (declare in the global table OperatorNamesTable).
    @param name (out) the name of the op operator if it exists.
    @return "true" if the operator op is kwnown, "false" otherwise.
*/
_VIRTUAL_ boolean
DCG2Syndex::getNewOperatorReference (noperators op, ptrepstr * name, int *Nth)
{
  (*name) = OperatorNamesTable[op];
  if (*name == 0)
    return pk_false;
  ReferenceCountersTable[*name]++;
  *Nth = ReferenceCountersTable[*name];
  return pk_true;
}


/** 
    @return the number associated with a new usage of an identifier (name).
    It manages the \c ReferenceCountersTable attribute member of the class.
*/
_VIRTUAL_ int
DCG2Syndex::getNewNameReference (ptrepstr name)
{
  if (ReferenceCountersTable.find (name) == ReferenceCountersTable.end ())
    ReferenceCountersTable[name] = 0;
  ReferenceCountersTable[name]++;
  return ReferenceCountersTable[name];
}

/**
   @return "true" if a Signal operator is implemented in Syndex for vectors.
   The table is built in TablesInitialize() method.
*/
_VIRTUAL_ boolean
DCG2Syndex::isDefinedForVector (noperators op)
{
  if (implementedForVector.find (op) == implementedForVector.end ())
    return pk_false;
  return implementedForVector[op];
}

/**
   @return "true" if a type (t) is a vector type with size vectorSize, "false" otherwise.
*/
_VIRTUAL_ boolean
DCG2Syndex::isVectorType (TypeType t, ptexpcal * vectorSize)
{
  ptexpcal x, vdim;

  if (!TypeTestArrayClass (t))
    return pk_false;
  vdim = ArrayMatrixTypeGetDimLst (t);
  EM->getNaryArgs (&vdim, vectorSize);
  return !EM->getNaryArgs (&vdim, &x);
}


/**
   @return the Syndex type associated with a Signal type (t). 
   The returned type is different from t 
   - for enumeration; it is replaced by an int type,
   - for an array type; the type of the element is returned.
   
   @param t (in) a signal type of an expression with op as operator.
   @param op (in) a signal operator used only to fix the dim expression. 
   @param dim (out) defined by the size of the vector when it is a vectorial type.
   or by 1 when the type t not a vector one, and an operator (op) is defined for vector.
   @param bLibrary (out) set to pk_true for predefined type, "false" otherwise. For predefined type, it exists a predefined Syndex library
   associated with.   
*/
_VIRTUAL_ TypeType
DCG2Syndex::getInfosType (noperators op, TypeType t, ntrees * dim, boolean * bLibrary)
{
  ptexpcal vectorSize;

  if (isVectorType (t, &vectorSize))
    {
      *dim = translateConstantExpression (vectorSize);
      t = ArrayMatrixTypeGetTerm (t);
    }
  else if (isDefinedForVector (op))
    *dim = TBI->makeConstantInt (1);
  else
    SET_NILTREE (*dim);
  if (TypeTestEnumerationClass (t))
    t = get_IntegerType ();
  *bLibrary = TypeTestPredef (t);
  return t;
}

/**
   @return the name of a type. It looks at the table DCG2Syndex::TypeTable initialized in initTypeTable() method.
   For type not defined in the table (user type), it returns the associated name.
*/
_VIRTUAL_ ptrepstr
DCG2Syndex::getTypeName (TypeType t)
{
  ptrepstr vtype;

  if (t == DRealType)
    restriction (1);
  vtype = TypeTable[t];
  if (vtype != 0)
    return vtype;
  // Definir une methode pour l'obtention du nom (cf cGeneratorType)
  if (TypeTestExternClass (t))
    return ExternTypeGetName (t);
  if (TypeTestMetaClass (t))
    return MetaTypeGetName (t);
  if (TypeTestNamedClass (t))
    return NamedTypeGetName (t);
  return string2repstr ("UNKNOWN_type");
}

/** Giving the operator name in syndex.
   @param op (in) the operator name of signal.
   @param t (in) the operator type of signal.
   @param i (out) the Nth operator of signal.
   @param vtype (out) the operator type of syndex (base type when vector)
   @param dim (out) the vector dimension when t is a vector type.
   @param bLibrary (out) at "true" when it exists a "Signal_t" predefined library for the type t, false otherwise.
   @return syndex name.
*/
_VIRTUAL_ ptrepstr
DCG2Syndex::getInfosForReferenceDeclaration (noperators op, TypeType t, int *i, ptrepstr * vtype, ntrees * dim, boolean * bLibrary)
{
  ptrepstr operInSyndex;

  if (!getNewOperatorReference (op, &operInSyndex, i))
    operInSyndex = getNewNameReference (string2repstr ("UNKNOWN_OPERATOR"));
  t = getInfosType (op, t, dim, bLibrary);
  *vtype = getTypeName (t);
  return operInSyndex;
}


/** 
    Giving the operator name in syndex.
    @param op (in) the operator name of signal.
    @param ty (in) the operator type of signal.
    @return syndex name+Nth. eg:plus234
*/
_VIRTUAL_ ptrepstr
DCG2Syndex::getLanguageOperator (noperators op, TypeType ty)
{
  return getLanguageOperator (op, ty, NILTREE);
}

/**
   @return the name of the operator associated with the Signal operator (op) for a particular type (ty).
   For a Signal operator (op) of an expression with ty as type, it adds the declaration of the Syndex operator in
   the syndex algorithm. It takes into account the predefined library (signal_TT, for TT a predefined type), and the possible initial values.
   For example, it produces the following declarations:
   - Signal_float/cst<{2.0}> cst_13;
   - Signal_float/Log_ne Log_ne_3;
*/
_VIRTUAL_ ptrepstr
DCG2Syndex::getLanguageOperator (noperators op, TypeType ty, ntrees initialvalue)
{
  int theNthOperation;
  ptrepstr nameOperSyn, typeSyn;
  ntrees t, dim;
  boolean bLibrary;

  nameOperSyn = getInfosForReferenceDeclaration (op, ty, &theNthOperation, &typeSyn, &dim, &bLibrary);
  t = TBI->addReference (bLibrary, typeSyn, nameOperSyn, theNthOperation, TBI->mkArgsValues (initialvalue, NILTREE, NILTREE, dim));
  return getReferenceName (t);
}

/**
   @return the type (for Syndex) of an expression (vexp)
   (enum type is implemented as int).
*/
_VIRTUAL_ TypeType
DCG2Syndex::getTypeOperator (ptexpcal vexp)
{
  noperators op;
  TypeType TT;

  op = EM->getOperator (vexp);
  if (ISRELATIONNALOP (op) || (op == opelem))
    vexp = EM->getBinaryArgstAt (vexp, 1);
  TT = getType (vexp);
  if (TypeTestEnumerationClass (TT))
    return get_IntegerType ();	//if it's enum, use as int.
  return TT;
}

/** Giving the operator name in syndex.
   @param vexp (in) the expression of signal.
   @return syndex name+Nth. eg:plus234
*/
_VIRTUAL_ ptrepstr
DCG2Syndex::getLanguageOperator (ptexpcal vexp)
{
  return getLanguageOperator (EM->getOperator (vexp), getTypeOperator (vexp));
}


/**
   Cell translation:
   It solves the references for each operand of the expression (e) 
   Then it declares and assigns (attribute) a new Cell Syndex operator to the expression (e). 

*/
_VIRTUAL_ void
DCG2Syndex::referenceForCell (ptexpcal e, ptexpcal init, TypeType type)
{
  ptexpcal e1;
  e1 = EM->getWhenOrCellArg (e);
  referenceForExpression (EM->getWhenOrCellArg (e1));
  referenceForExpression (CKM->getBooleanSignal (EM->getContributionClock (e1)));
  setReferenceTo (e, mkSyndexAttribute (getLanguageOperator (EM->getOperator (e),
							     getTypeOperator (e), translateConstantExpression (init)), MyStdOutput));
}

/** When translation:
    For Syndex, the translation of a subgraph produces an algorithm 
    that takes into account the condition (condition: c=1).
    So, for E When C expression, the reference is solved for E.
 */
_VIRTUAL_ void
DCG2Syndex::referenceForWhen (ptexpcal e)
{
  referenceForExpression (EM->getWhenOrCellArg (e));
}


/** Default translation:
    It solves the references for each operand of the expression (e).
    Then it declares and assigns (attribute) a new default Syndex operator to the expression (e).
*/
_VIRTUAL_ void
DCG2Syndex::referenceForDefault (ptexpcal e)
{
  referenceForExpression (EM->getBinaryArgstAt (e, 1));
  referenceForExpression (EM->getBinaryArgstAt (e, 2));
  setReferenceToForOperatorOf (e);
}

/**
   Constant value expression:
   It declares and assigns (attribute) a new constant Syndex operator to the expression (vexp). 
   The value associated with the constant Syndex operator is produced by the translation of  the constant 
   expression (translateConstantExpression()).

   For enum type, it applied referenceForEnumerate().
*/
void
DCG2Syndex::referenceForConstantExpression (ptexpcal vexp)
{
  ntrees e;
  TypeType ty;
  ptrepstr t;

  if (EM->getOperator (vexp) == opcstenum)
    referenceForEnumerate (vexp);
  else
    {
      ty = getType (vexp);
      e = translateConstantExpression (vexp);
      t = getLanguageOperator (EM->getOperator (vexp), ty, e);
      setReferenceTo (vexp, mkSyndexAttribute (t, MyStdOutput));
    }
}

/**
   Assumed : vexp is a reference to an enum value (i.e. Type#enumVal).
   It is replaced in Syndex by a int constant, the ordinal of the value in the type.
*/
void
DCG2Syndex::referenceForEnumerate (ptexpcal vexp)
{
  int val;
  ntrees e;
  TypeType ty;
  ptexpcal exp1, exp2;
  ptrepstr enumName, t;

  EM->getBinaryArgs (vexp, &exp1, &exp2);
  ty = get_IntegerType ();
  enumName = EM->getConstant (exp2);
  EnumTypeOrdinal (getType (vexp), enumName, &val);
  e = TBI->makeConstantInt (val);
  t = getLanguageOperator (opcstent, ty, e);
  setReferenceTo (vexp, mkSyndexAttribute (t, MyStdOutput));
}

/** Unary expression translation:
    It solves the references for each operand of the expression (vexp).
    Then it declares and assigns (attribute) a new Syndex operator to the expression.
    \note the references to state variable (x?) are filtered. (why?)
*/
void
DCG2Syndex::referenceForUnaryExpression (ptexpcal vexp)
{
  referenceForExpression (EM->getUnaryArg (vexp));
  //when it is X?, do nothing for reference.
  if (EM->getOperator (vexp) != opvvarin)
    setReferenceToForOperatorOf (vexp);
}

/** Binary expression translation:
    It solves the references for each operand of the expression (vexp).
    Then it declares and assigns (attribute) a new Syndex operator to the expression.

    For enum type, it applied referenceForEnumerate().
*/
void
DCG2Syndex::referenceForBinaryExpression (ptexpcal vexp)
{
  ptexpcal arg1, arg2;
  noperators op;
  ptinsth vh;

  op = EM->getOperator (vexp);
  if (op == opvenum)
    referenceForEnumerate (vexp);
  else
    {
      arg1 = EM->getBinaryArgstAt (vexp, 1);
      if (op == opleft_tt)
	{
	  EM->getSynchronousExprHorl (arg1, &vh);
	  arg2 = CKM->getBooleanSignal (vh);
	}
      else
	arg2 = getSecondArgBinaryExpression (vexp, op);

      referenceForExpression (arg1);
      referenceForExpression (arg2);
      setReferenceToForOperatorOf (vexp);
    }
}

/** Ternary expression translation:
    It solves the references for each operand of the expression (vexp).
    Then it declares and assigns (attribute) a new Syndex operator to the expression.
*/
void
DCG2Syndex::referenceForTernaryExpression (ptexpcal vexp)
{
  referenceForExpression (EM->getTernaryArgstAt (vexp, 1));
  referenceForExpression (EM->getTernaryArgstAt (vexp, 2));
  referenceForExpression (EM->getTernaryArgstAt (vexp, 3));
  setReferenceToForOperatorOf (vexp);
}

/**
   @return a new instance (id_N) of a syndex operator typeSyn/id (or Signal_typeSyn/id).
   This instance is delared in the Syndex program.
*/
ptrepstr DCG2Syndex::mkDeclarationOperator (ptident id, ptrepstr typeSyn)
{
  ptrepstr
    name;
  ntrees
    t;
  int
    nth;

  name = ident2repstr (id);
  nth = getNewNameReference (name);
  t = TBI->addReference (pk_false, typeSyn, name, nth);
  return getReferenceName (t);
}

/** function call expression translation:
    It solves the references for each operand of the expression (vexp).
    Then it declares and assigns (attribute) a new Syndex operator to the expression.

    Here, we assume that the "user" will provide the definition of the associated Syndex algorithm of the function.
   \see mkDeclarationOperator
*/
void
DCG2Syndex::referenceForFunctionCallExpression (ptexpcal vexp)
{
  ptident id;
  ptexpcal vexpcal1, vexpcal2;
  ptrepstr x;
  boolean b;

  EM->getFunctionCall (vexp, &id, &vexpcal2, &vexpcal1);	// vexpcal2: parameters, vexpcal1: input signals.
  referenceForExpression (vexpcal1);

  b = EM->isMarked (vexp, INTRINSIC);
  id = explicitConversion (EM->getPragmas (vexp), vexpcal1, id);
  x = mkDeclarationOperator (id, TBI->getLibraryName (b));
  setReferenceTo (vexp, mkSyndexAttribute (x, MyStdOutput));
}

/**
   It declares an external Signal constant (vexp) in the Syndex program.
   The created Syndex operator is assigned (attribute) to the expression.
   
   Here, we assume that the "user" will provide the definition of the associated Syndex algorithm.
   Example: for a undefined parameter n, it declares
   - Signal_int/n n_1;
   \see mkDeclarationOperator
*/
void
DCG2Syndex::declareExternalConstant (ptexpcal vexp)
{
  ntrees dim;
  ptrepstr t, vlibname;
  ptident name;

  t = getTypeLibrary (vexp, &dim, &vlibname);
  name = EM->getName (vexp);
  t = mkDeclarationOperator (name, vlibname);
  setReferenceTo (vexp, mkSyndexAttribute (t, MyStdOutput));
}

/**
   For an identifier, constant or parameter, it applied declareExternalConstant() for external one 
   or referenceForEqu() otherwise (i.e. as a signal definition).
   
   For other identifiers (Signals), nothing is done: their references will be replaced by connections 
   from the node where they are produced.
*/
void
DCG2Syndex::referenceForVariable (ptexpcal vexp)
{
  ptINode n;
  if (EM->IsConstant (vexp) || EM->IsParameter (vexp))
    {
      n = EM->getNode (vexp);
      if (NM->is_undefined (n))
	declareExternalConstant (vexp);
      else
	referenceForEqu (n);
    }
}

/** List expression translation:
    It solves the references for each operand of the list (vexp).
*/
void
DCG2Syndex::referenceForListExpression (ptexpcal vexp)
{
  ptexpcal elem;
  while (EM->getNaryArgs (&vexp, &elem))
    referenceForExpression (elem);
}

/** Expression translation:
    
    It solves the references for each operand of the expression (vexp).
    It declares and assigns (attribute) a new Syndex operator to the expression. Recall that the connection between theses
    operators will be done in a second step.

    Here, we assume that there are Syndex libraries that defines the algorithms of the Signal operators.
    \note enum type available in Signal is translated as int in Syndex.

    It realises a switch according to the operator of an expression (vexp).
    - referenceForVariable()                identifier
    - referenceForConstantExpression()      constant
    - referenceForUnaryExpression()         unary expression
    - referenceForListExpression()          list expression
    - referenceForBinaryExpression()        binary expression 
    - referenceForTernaryExpression()       ternary expression
    - referenceForFunctionCallExpression()  function call 
    - referenceForWhen()       when operator
    - referenceForDefault()   default operator
*/
void
DCG2Syndex::referenceForExpression (ptexpcal vexp)
{
  ptexpcal nexp;

  //To do: is this right when an expression include an array?eg: S = A+ B[3],or others, see agian.
  if (EM->isNILEXPCAL (vexp))
    return;
  if (EM->getIndirection (vexp, &nexp))
    vexp = nexp;
  if (isAlreadyReferenced (vexp))
    return;

  TRACE (cout << EM->getOperatorClass (vexp) << " the operator class in reference\n");
  TRACE (cout << EM->getOperator (vexp) << " the operator in reference \n");
  TRACE (cout << ident2string (EM->getName (vexp)) << " the name of expression in reference\n");

  switch ((int) (EM->getOperatorClass (vexp)))
    {
      // reference to an identifier (nothing to do for reference step)
    case variabl:
      TRACE (cout << ident2string (EM->getName (vexp)) << " enter variable\n");
      referenceForVariable (vexp);
      break;

      // reference to a constant expression.
    case constant:
      referenceForConstantExpression (vexp);
      break;

      // Unary expression : not , -
    case oper1:
      referenceForUnaryExpression (vexp);
      break;

      // Binary Expression : +,*,...
    case oper2:
      if (EM->satisfiesList (vexp))
	referenceForListExpression (vexp);
      else
	referenceForBinaryExpression (vexp);
      break;

      // Delay/Window delay  ( handed in NodeRef_i.c )
    case oper3:
      referenceForTernaryExpression (vexp);
      break;

      // function call id( i1,...., in)
    case call:
      referenceForFunctionCallExpression (vexp);
      break;

      // E1 when clk ;  E1 cell clk has been filtered before.
    case opereh:
      referenceForWhen (vexp);
      break;

    case oper2tp:		// default operator
      referenceForDefault (vexp);
      break;

      // no occurrence (for syndex translating)
    case oper1tp:
    case oper3tp:
      break;

      // not yet implemented.
    case opertyp:
      break;

      // Internal error
    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
}



/**
   It adds a declaration of a Syndex operator associated with a initial value (vexpInit) characterized by its type (ty).
   The declared operator is associated with the expression as an attribute.
   \see mkSyndexAttribute(),setReferenceTo() 
_VIRTUAL_ void
DCG2Syndex::referenceForInitialValues(ptexpcal vexpInit, TypeType ty)
{
  noperators op;
  ntrees init;
  ptrepstr t; 

  if (! EM->isNILEXPCAL(vexpInit))
    {
      if (!EM->satisfiesList(vexpInit)) op = opcstent; 
      else if (EM->getSize(vexpInit) == 1) op = opcstent; else op = oplelevec;

      init = translateConstantExpression(vexpInit);
      t = getLanguageOperator(op, ty, init);
      setReferenceTo(vexpInit, mkSyndexAttribute(t, MyStdOutput ));
    }
}
*/
