/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   Declaration of the references for an equation (vnode), first step of the translation.
   It calls referenceForCell() for a cell. It applies referenceForExpression() on the definition expression.
   An attribute, see mkSyndexAttribute() method, is associated with each visited expression.

   Example:
   \verbatim
   references: 
       Signal_float/Arit_div<1> Arit_div_9;
       Signal_float/Arit_minus<1> Arit_minus_5;
   \endverbatim
    are the references declared for the float expression -(x/y).
*/
  _VIRTUAL_ void
DCG2Syndex::referenceForEqu (ptINode vnode)
{
  ptexpcal leftPart, rightPart, basicExpr;
  ptSyndexAttribute def;

  if (NM->get_rightOperand (vnode, &rightPart))
    {
      leftPart = NM->get_leftOperand (vnode);
      if (EM->IsCellExpr (rightPart))	// Cell operator
	{
	  referenceForCell (rightPart, EM->getInitialValue (leftPart), getType (leftPart));
	  if (getSyndexAttributeOf (rightPart, &def))
	    setReferenceTo (leftPart, def);
	}
      else
	{
	  basicExpr = EM->getBasicExpr (rightPart);
	  referenceForExpression (basicExpr);	//for one exp,eg (not A) and B
	  if (getSameLevelSyndexAttributeOf (basicExpr, &def))
	    setReferenceTo (leftPart, def);
	  else
	    setIndirectionTo (leftPart, basicExpr);
	}
    }
}


/**
   Declaration of the references for an delay (aNode), first step of the translation.
   An attribute, see mkSyndexAttribute() method, is associated with each visited expression.
  Example:
   \verbatim
   references: 
      Signal_bool/delay<{1};1> delay_1;
   \endverbatim
   are the references declared for a boolean delay <tt> x $ 1 init pk_true </tt>.
*/
_VIRTUAL_ void
DCG2Syndex::referenceForDelay (ptINode aNode)
{
  ptexpcal vexpsigres, vexpsigret, vexpret, vexpfen, vexpInit;
  boolean simple, b;
  int i;
  compilerInteger v;
  ptrepstr name, typeName;
  TypeType ty;
  ntrees initValTree, dim, t, w, d;
  noperators op;

  NM->get_delay_arg (aNode, &vexpsigres, &vexpsigret, &vexpret, &vexpfen, &vexpInit);
  ty = getType (vexpsigret);
  if (TypeTestEnumerationClass (ty))
    ty = get_IntegerType ();	//if it's enum, use as int.
  initValTree = translateConstantExpression (vexpInit);
  if (vexpfen != NILEXPCAL)
    if (vexpret != NILEXPCAL)
      // ZX:= X $m window n
      {
	w = translateConstantExpression (vexpfen);
	d = translateConstantExpression (vexpret);
	op = opretfen;
      }
    else
      // ZX:= X window n.
      {
	w = translateConstantExpression (vexpfen);
	op = opfen;
	SET_NILTREE (d);
      }
  else				// ZX:= X $m
    {
      if (EM->getIntegerValue (vexpret, &v))
	simple = (v == 1);
      else
	simple = pk_false;
      if (simple)
	d = TBI->makeConstantInt (1);
      else
	d = translateConstantExpression (vexpret);
      op = opretN;
      SET_NILTREE (dim);
      SET_NILTREE (w);
    }
  name = getInfosForReferenceDeclaration (op, ty, &i, &typeName, &dim, &b);
  if (op == opretN)
    SET_NILTREE (dim);
  t = TBI->addReference (TypeTestPredef (ty), typeName, name, i, TBI->mkArgsValues (initValTree, d, w, dim));
  setReferenceTo (vexpsigres, mkSyndexAttribute (getReferenceName (t), MyStdOutput));
}


/**
   Specific processing for the outputs (attribute member: \c lwrites) of the current graph defined in the current sub-graph
   and the outputs of the subgraph (attribute member: \c ouputs):
   - For an node of lwrites, it declares a reference to an actuator, 
   and adds the connection from the node that produced the signal to the actuator.

   \verbatim
   references:
      Signal_bool/output<1> output_3;
   dependences:
      strong_precedence_data default_12.o -> output_3.i;
   \endverbatim

   - For an output of the subgraph, it adds the connection from the node that produced the signal to the declared output in the interface.
     \verbatim
       ...
      ! float  x21;

      conditions: C_x_511 = 1;
      references: 
          Signal_float/default<1> default_10;
      dependences:
          strong_precedence_data default_10.o -> x21;
   \endverbatim

   \remarks 
   - The list \c lwrites is updated in referenceSpecialExternalCall().
   - here, some connections are managed.
*/
_VIRTUAL_ void
DCG2Syndex::referenceForOutputs ()
{
  ptlINode l;
  ptexpcal p, expg, expd, par;
  ptident nom;
  ptrepstr actuatorName;
  ptINode aNode;
  ptlinstances list;

  l = lwrites;
  while (ptlINode_get_elem (&l, &aNode))
    {
      NM->get_procCall_arg (aNode, &nom, &par, &expd, &expg);
      EM->getNaryArgs (&expd, &p);
      actuatorName = addIODeclaration (OutSignal, p);
      p = EM->getBasicExpr (p);	// we can have constant value (true when C for example)
      referenceForExpression (p);
      connectionFromTo (p, actuatorName, -1);
      setNodeReference (aNode, actuatorName);
    }

  list = outputs;
  while (ptlinstances_get_next (&list, &p))
    connectionFromTo (p, getRepstrIdName (p), -2);	// (-2) means without suffix.
}


/**
   Declaration of the references for some particular external calls (aNode), first step of the translation. 
   The following intrinsics are treated: \c writeS, \c readS, \c Status (Signal state variable are replaced by such an external).
   For the \c writeS external call, the node is added to the lwrite list (attribute member), 
   and will be treated later (see declareReferences()).
   @return "true" if the external call is treated, "false" otherwise.
   An attribute, see mkSyndexAttribute() method, is associated with each visited expression.
*/
_VIRTUAL_ boolean
DCG2Syndex::referenceSpecialExternalCall (ptINode aNode)
{
  ptexpcal p, expg, expd, par, arg1, arg2, arg3, x;
  ptident nom;
  ptrepstr sensorName;

  if (!NM->is_intrinsicfunc (aNode))
    return pk_false;

  NM->get_procCall_arg (aNode, &nom, &par, &expd, &expg);
  if (eq_idents (nom, string2ident ("writeS")))
    {
      ptlINode_insert_elem (&lwrites, aNode);
      return pk_true;
    }
  if (eq_idents (nom, string2ident ("readS")))
    {
      EM->getNaryArgs (&expg, &p);
      sensorName = addIODeclaration (InSignal, p);
      setReferenceTo (p, mkSyndexAttribute (sensorName, MyStdOutput));
      setNodeReference (aNode, sensorName);
      return pk_true;
    }

  if (isStatusId (nom))		// STATUS
    {				//state{0}...
      EM->getNaryArgs (&expg, &p);
      EM->getNaryArgs (&par, &arg1);
      EM->getNaryArgs (&expd, &x);
      arg2 = EM->getWhenOrCellArg (x);
      arg3 = CKM->getBooleanSignal (EM->getContributionClock (x));
      referenceForExpression (arg2);
      // referenceForExpression(arg1); init
      referenceForExpression (arg3);
      sensorName = getLanguageOperator (opvvarin, getTypeOperator (p), translateConstantExpression (arg1));
      setReferenceTo (p, mkSyndexAttribute (sensorName, MyStdOutput));
      setNodeReference (aNode, sensorName);
      return pk_true;
    }
  return pk_false;
}

/**
   It applies referenceForExpression() for each element of inputs.
   It delares a new Syndex operator built with vLib and name, as UserLibrary/name name_60, or IntrinsicLibrary/name name_60.
   It links (attribute) each signal output, element of outputs, to an output of the model (Syndex node).
*/
_VIRTUAL_ void
DCG2Syndex::setCallreference (ptINode n, ptident name, ptexpcal lpars, ptexpcal inputs, ptexpcal outputs, ptrepstr vLib)
{
  ptrepstr x;
  ptexpcal p;
  int i;

  //lpars not used.

  referenceForExpression (inputs);
  name = explicitConversion (NM->getPragmas (n), inputs, name);	// name is changed for an explicit conversion.
  x = mkDeclarationOperator (name, vLib);
  if (EM->getSize (outputs) == 1)
    {
      EM->getNaryArgs (&outputs, &p);
      setReferenceTo (p, mkSyndexAttribute (x, MyStdOutput));
    }
  else
    {
      i = 1;
      while (EM->getNaryArgs (&outputs, &p))
	{
	  setReferenceTo (p, mkSyndexAttribute (x, TBI->mkOName (i)));
	  i++;
	}
    }

  setNodeReference (n, x);	// assigns syndexOperator for the node.
}

/**
   Declaration of the references for an external call (aNode), first step of the translation.
   An attribute, see mkSyndexAttribute() method, is associated with each visited expression.
   It calls  referenceSpecialExternalCall() for some particular cases.
   \see setCallreference()
*/
_VIRTUAL_ void
DCG2Syndex::referenceForExternalCall (ptINode aNode)
{
  ptexpcal expg, expd, par;
  ptident nom;

  if (!referenceSpecialExternalCall (aNode))
    if (!NM->is_readNode (aNode))
      {
	NM->get_procCall_arg (aNode, &nom, &par, &expd, &expg);
	setCallreference (aNode, nom, par, expd, expg, TBI->getLibraryName (NM->is_intrinsicfunc (aNode)));
      }
}


/**
   Internal to referenceForNodeCall().
   Declaration of the references for a reference to a model (aGraph), first step of the translation, 
   with some effective parameters (par),  with some effective inputs (inputs),  with some effective outputs (outputs).
   An attribute, see mkSyndexAttribute() method, is associated with each visited expression.

   For a graph with foo as name, it declares \n
   foo foo1 \n
   in the main Syndex program.
   
*/
_VIRTUAL_ ptrepstr
DCG2Syndex::setNodeCallreference (ptDCG aGraph, ptexpcal par, ptexpcal inputs, ptexpcal outputs)
{
  ptrepstr x, fname;
  ptexpcal p, vs;
  ptident name;
  ptlinstances l;

  referenceForExpression (inputs);
  name = aGraph->getName ();

  x = mkDeclarationOperator (name, nilstr);
  l = aGraph->getOutputs ();
  while (EM->getNaryArgs (&outputs, &p))
    {
      ptlinstances_get_next (&l, &vs);
      fname = getRepstrIdName (vs);
      setReferenceTo (p, mkSyndexAttribute (x, fname));
    }
  return x;
}


/**
   Declaration of the references for a node call (aNode), first step of the translation.
   An attribute, see mkSyndexAttribute() method, is associated with each visited expression.
   \see setCallreference()
*/
_VIRTUAL_ void
DCG2Syndex::referenceForNodeCall (ptINode aNode)
{
  ptexpcal expg, expd, par;
  ptDCG aGraph;
  ptrepstr vlib, x;

  NM->get_nodeCall_arg (aNode, &aGraph, &par, &expd, &expg);
  if (aGraph->isGraphOfExternalProcess ())
    {
      vlib = TBI->getLibraryName (NM->is_intrinsicfunc (aNode));
      setCallreference (aNode, aGraph->getName (), par, expd, expg, vlib);
    }
  else
    {
      vlib = nilstr;
      x = setNodeCallreference (aGraph, par, expd, expg);
      setNodeReference (aNode, x);	// assigns syndexOperator for the node.
    }

}

/**
   Declaration of the references for a reference to a subgraph node (aNode), first step of the translation.
   An attribute, see mkSyndexAttribute() method, is associated with each visited expression.
   \see setCallreference()
*/
_VIRTUAL_ void
DCG2Syndex::referenceForRefSubGraph (ptINode aNode)
{
  ptexpcal lpars, lins, louts;
  ptident name;

  if (!NM->get_clusterName (aNode, &name))
    name = makeIdent_ident_int (AGRAPH->getName (), NM->get_num (aNode));
  getInterface (aNode, &lpars, &lins, &louts);
  setCallreference (aNode, name, lpars, lins, louts, nilstr);
}

/**
   Declaration of the references for an array node (aNode), first step of the translation.
   An attribute, see mkSyndexAttribute() method, is associated with each visited expression.
   \note Not yet implemented.
*/
_VIRTUAL_ void
DCG2Syndex::referenceForArray (ptINode aNode)
{
  returnCode = pk_false;
  pK_messenger->printString ("***** WARNING Syndex translation: Array of processes not yet implemented. ******");
}

/**
   Declaration of the references for a node (aNode), first step of the translation.
   An attribute, see mkSyndexAttribute() method, is associated with each visited expression.

   It switches according to the operator of the node:
   - referenceForEqu()  equation
   - referenceForDelay() delay 
   - referenceForRefSubGraph() subgraph
   - referenceForNodeCall() node call
   - referenceForExternalCall() external call
   - referenceForArray() (array, not implemented)   
*/
void
DCG2Syndex::referenceForNode (ptINode aNode)
{
  ptINode oCurrentNode;

  if (!isSelectedNode (aNode))
    return;

  switch ((int) NM->get_oper (aNode))
    {
    case opdefsig:
    case opdefpart:
      referenceForEqu (aNode);	/* X:= E */
      break;

    case opret:
      referenceForDelay (aNode);	/* x:= E1 $ n window m  */
      break;

    case opnodecall:
      referenceForNodeCall (aNode);
      break;

    case opappe:
      if (!NM->is_ioNode (aNode))
	referenceForExternalCall (aNode);
      break;

    case opsubgraph:
      referenceForRefSubGraph (aNode);
      break;

    case opmoti:		/* array structure */
      referenceForArray (aNode);
      break;

    case opbund:		// No implemented
    case opldefpart:		// nothing to do
    case opnil:
      break;

    default:
      pK_messenger->Caseerror (currentLine, currentFile);
      break;
    }
}

/**
   It declares, first step of the translaion, the Syndex operators for a list of nodes (lnodes).
   It calls referenceForNode() for each nodeof the list.
   Then, it calls referenceForOutputs() 
   for a specific processing of the outputs (lwrite attribute member) of the current graph defined in the current sub-graph
   and the outputs of the subgraph (ouputs attribute member).
   \see referenceSpecialExternalCall, referenceForNode.
*/
void
DCG2Syndex::declareReferences (ptlINode lnodes)
{
  ptINode aNode;

  while (ptlINode_get_elem (&lnodes, &aNode))
    referenceForNode (aNode);
  referenceForOutputs ();
}
