/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PRIVATE_
/**
   @return "true" if the expression "exp" is an identifier or a constant expression, "false" otherwise.
*/
  boolean DCG2KernelTranslator::isIdentOrConstantExp (ptexpcal exp)
{
  noperators
    op;

  if (exp == NILEXPCAL)
    return pk_true;		// blindage
  switch (EM->getOperatorClass (exp))
    {
    case variabl:
    case constant:
      return pk_true;
    default:
      op = EM->getOperator (exp);
      if (op == opvenum)
	return pk_true;
      if (op == opelem)
	return pk_true;
      if (op == opvvarin)
	return pk_true;
    }
  return pk_false;
}


/**
   It propagates the partitionning pragmas of the current visited node (currentNode member of the class) to n.
*/
void
DCG2KernelTranslator::propagatePragmasOn (ptINode n)
{
  ptlpragmas LPrag;
  ptpragma aPragma;

  LPrag = NM->getPragmas (currentNode);
  if (getPragmaKeyword (LPrag, PRAGMA_RUN_ON, &aPragma))
    NM->addPragma (n, aPragma);
}


/**
   Uses "idContext" attribute that must be set before calling it.
   Adds to the graph. a new node "x:= ne" where "ne" is a synchronous expression.
   When "ne" is a constant expression, its clock is the context clock "vh".
   @return the created signal.
*/
ptexpcal DCG2KernelTranslator::addSignal (ptexpcal ne, ptinsth vh)
{
  ptINode
    n;
  ptinsth
    vhcal,
    h;

  if (EM->getExprClock (ne, &h))
    vhcal = h;
  else
    {
      h = vh;
      vhcal = hinit;
    }
  ne = NM->insertEquation (idContext, EM->getType (ne), ne, h, vhcal, &n);
  propagatePragmasOn (n);
  return ne;
}

/**
   Reduces to Basic Signal a sampling ("when") expression "exp" with the context clock "vh".
   @return the reduced expression.
*/
ptexpcal DCG2KernelTranslator::kernelRewritingWhen (ptexpcal exp, ptinsth vh, integer holoph)
{
  ptexpcal
    e1,
    ne1,
    ne;
  ptinsth
    vh1;
  boolean
    b;

  e1 = EM->getWhenOrCellArg (exp);
  if (isIdentOrConstantExp (e1))
    return exp;
  vh = EM->getContributionClock (exp);
  ne1 = kernelRewriting (e1, vh, holoph + 1);
  if (ne1 == e1)
    {
      if (holoph > 0)
	b = pk_true;
      else
	b = (EM->getSynchronousExprHorl (e1, &vh1) && (vh != CKM->getRepresentativeFormula (vh1)));
      if (b)
	{
	  ne1 = addSignal (ne1, vh);
	  ne = EM->mkWhenOrCellExp (opextr, ne1, vh);
	}
      else
	ne = exp;
    }
  else
    {
      b = pk_false;
      if (!isIdentOrConstantExp (ne1))
	if (holoph > 0)
	  b = pk_true;
	else
	  b = (EM->getSynchronousExprHorl (ne1, &vh1) && (vh != CKM->getRepresentativeFormula (vh1)));
      if (b)
	ne1 = addSignal (ne1, vh1);
      ne = EM->mkWhenOrCellExp (opextr, ne1, vh);
    }
  return ne;
}

/**
   Reduces to Basic Signal a (default) expression "exp", with the context clock "vh".
   @return the reduced expression.
*/
ptexpcal DCG2KernelTranslator::kernelRewritingDefault (ptexpcal exp, ptinsth vh, integer holoph)
{
  ptexpcal
    e1,
    e2,
    ne1,
    ne2;

  e1 = EM->getBinaryArgstAt (exp, 1);
  vh = EM->getContributionClock (e1);
  ne1 = kernelRewriting (e1, vh, holoph + 1);
  if (EM->getOperator (EM->getWhenOrCellArg (ne1)) == opmele)
    {
      ne1 = addSignal (ne1, vh);
      ne1 = EM->mkWhenOrCellExp (opextr, ne1, vh);
    }
  e2 = EM->getBinaryArgstAt (exp, 2);
  vh = EM->getContributionClock (e2);
  ne2 = kernelRewritingWhen (e2, vh, holoph + 1);
  if ((ne1 == e1) && (ne2 == e2))
    return exp;
  ne1 = EM->mkBinaryExpTp (opmele, ne1, ne2, EM->getBackAnnotation (exp));
  EM->setContributionClock (ne1, CKM->getRepresentativeFormula (EM->getContributionClock (exp)));
  return ne1;
}

/**
   Reduces to Basic Signal a (external function call) expression "exp", with the context clock "vh".
   @return the operand expression (exp) because it is already in kernel Signal.
*/
ptexpcal DCG2KernelTranslator::kernelRewritingFunctionCall (ptexpcal exp, ptinsth vh, integer holoph)
{
  // We accept program with external function calls.
  return exp;
}

/**
   Reduces to Basic Signal an expression "exp" that is neither a constant expression nor a signal identifier, with the context clock "vh".
   It is referred to only in expr2kernel() method.
   @return the reduced expression.
*/
ptexpcal DCG2KernelTranslator::kernelRewriting (ptexpcal exp, ptinsth vh, integer holoph)
{
  noperators
    op;
  ptexpcal
    e1,
    e2,
    ne1,
    ne2;

  op = EM->getOperator (exp);
  switch (EM->getOperatorClass (exp))
    {
      // case variabl: // case constant: return pk_false;

    case oper1:
    case oper1tp:
      e1 = EM->getUnaryArg (exp);
      if (isIdentOrConstantExp (e1))
	return exp;

      ne1 = kernelRewriting (e1, vh, holoph + 1);
      ne1 = addSignal (ne1, vh);
      // if (ne1 == e1) return exp;
      return EM->mkUnaryExp (op, ne1, EM->getBackAnnotation (exp));

    case oper2:
      e1 = EM->getBinaryArgstAt (exp, 1);
      e2 = EM->getBinaryArgstAt (exp, 2);
      if (!isIdentOrConstantExp (e1))
	{
	  ne1 = kernelRewriting (e1, vh, holoph + 1);
	  ne1 = addSignal (ne1, vh);
	}
      else
	ne1 = e1;
      if (!isIdentOrConstantExp (e2))
	{
	  ne2 = kernelRewriting (e2, vh, holoph + 1);
	  ne2 = addSignal (ne2, vh);
	}
      else
	ne2 = e2;
      if ((ne1 == e1) && (ne2 == e2))
	return exp;
      return EM->mkBinaryExp (op, ne1, ne2, EM->getBackAnnotation (exp));

    case opereh:
      return kernelRewritingWhen (exp, vh, holoph);

    case oper2tp:		// Default
      return kernelRewritingDefault (exp, vh, holoph);

    case opertyp:
    case oper3:		// Pas d'occurrence
    case oper3tp:
      return exp;
    case call:
      return kernelRewritingFunctionCall (exp, vh, holoph);

    default:
      pK_messenger->Caseerror (currentLine, currentFile);	// abort
      return NILEXPCAL;
    }
}

/**
   Reduces to Basic Signal expression BASIC_EXP, the expression "exp" with the context clock "vh"
   @return "true" if "exp" and  BASIC_EXP are distinct expressions; return "false" otherwise.
   @return BASIC_EXP in "n_exp" if "exp" and  BASIC_EXP are distinct expressions;
 */
boolean DCG2KernelTranslator::expr2kernel (ptexpcal exp, ptinsth vh, ptexpcal * n_exp)
{
  if (isIdentOrConstantExp (exp))
    return pk_false;
  *n_exp = kernelRewriting (exp, vh, 0);
  return (exp != (*n_exp));
}

/**
   Reduces to Basic Signal an equation "node".
   The reduction Basic(x:=E) is x:=Basic(E,^x).
 */
void
DCG2KernelTranslator::equ2kernel (ptINode node)
{
  ptexpcal expg, expd, nexpd;
  ptinsth vhcal, k;

  expg = NM->get_leftOperand (node);
  if (!EM->IsSignal (expg))
    setFunctionalityNotApplied (node);	// restriction
  else
    {
      idContext = EM->getOriginalName (expg);	// warning Global Symbol
      NM->get_rightOperand (node, &expd);
      k = CKM->getRepresentativeFormula (EM->getContributionClock (expd));
      if (EM->IsCellExpr (expd))
	{
	  // expd  = (E when XXX) cell H, expd := (E1 default E2) cell H
	  expd = EM->getWhenOrCellArg (expd);
	  vhcal = CKM->getRepresentativeFormula (EM->getContributionClock (expd));
	  expd = EM->getBasicExpr (expd);
	  if (!isIdentOrConstantExp (expd))
	    {
	      nexpd = addSignal (expd, vhcal);
	      NM->modifiedDefinition (expg, EM->getWhenOrCellArg (nexpd), vhcal, k);
	    }
	}
      else if (!CKM->satisfiesInitClock (k))
	if (expr2kernel (expd, k, &nexpd))
	  NM->modifiedDefinition (expg, EM->getWhenOrCellArg (nexpd), k, k);
    }
}

/**
   Reduces to Basic Signal a procedure call or a node call (node).
   The reduction Basic(P(E1,....En)) creates a new signal Xi (see mkIdentifiersForExpression()) defined by Xi:= Basic(Ei) for each input expression Ei such Basic(Ei) is not an identifier.
 */
void
DCG2KernelTranslator::call2kernel (ptINode node)
{
  ptDCG g1;
  ptexpcal Params, Inputs, Outputs, nInputs;
  ptinsth clkNode;
  ptident name;

  if (NM->isNodeCall (node))
    NM->get_nodeCall_arg (node, &g1, &Params, &Inputs, &Outputs);
  else
    NM->get_procCall_arg (node, &name, &Params, &Inputs, &Outputs);
  clkNode = NM->get_nodeClock (node);

  // if (Params != nParams) NM->update(node, Params, nParams);
  nInputs = mkIdentifiersForExpression (Inputs, clkNode);
  if (Inputs != nInputs)
    NM->update (node, Inputs, nInputs);
}

/**
   Reduces to Basic Signal a delay (node).
   
   \note not implemented: When the delay is greater than 1, it is not translated.
*/
void
DCG2KernelTranslator::delay2kernel (ptINode node)
{
  ptexpcal resret, sigret, vret, vfen, vinit;
  compilerInteger bi, bs;
  boolean b;

  NM->get_delay_arg (node, &resret, &sigret, &vret, &vfen, &vinit);
  if (EM->isNILEXPCAL (vret))
    return;
  if (EM->getBounds (vret, &bi, &bs))
    {
      if (bi == 1)
	return;
    }
  setFunctionalityNotApplied (node);
}

/**
   If the expression "e" is not a list of expressions,
   - reduces "e" to Basic Signal with the context clock "contextClock" and
   - if Basic("e") is not an identifier the method creates an equation X:= Basic("e") and return X, where X is a new identifier containing the string "_Added_"
   If the expression "e" is a list E1, ... En,
   - recursively apply :mkIdentifiersForExpression to each sub-expression Ei 
   - returns the list X1, ..., Xn of the corresponding results
*/
ptexpcal DCG2KernelTranslator::mkIdentifiersForExpression (ptexpcal e, ptinsth contextClock)
{
  ptexpcal
    e1,
    elem,
    ne1,
    nelem;
  ptINode
    n;
  ptinsth
    h,
    hcal;

  if (EM->isNILEXPCAL (e))
    return e;
  // a list
  if (EM->satisfiesList (e))
    {
      e1 = e;
      EM->getNaryArgs (&e1, &elem);
      ne1 = mkIdentifiersForExpression (e1, contextClock);
      nelem = mkIdentifiersForExpression (elem, contextClock);
      if ((e1 == ne1) && (nelem == elem))
	return e;
      return EM->mkNaryExp (EM->getOperator (e), nelem, ne1, NILBACKANNOTATION);
    }
  // not a list
  if (EM->getOperatorClass (e) == variabl)
    return e;
  if (!EM->getSynchronousExprHorl (e, &h))
    {
      hcal = CKM->getInitClock ();
      h = contextClock;
    }
  else
    hcal = h;
  e1 = EM->getBasicExpr (e);
  return NM->insertEquation (string2ident ("_Added_"), EM->getType (e1), e1, h, hcal, &n);
}

/**
   @return the current node before the call to this method.
   Sets "n" as the current visiting node.
*/
ptINode
DCG2KernelTranslator::setCurrentNode (ptINode n)
{
  ptINode x;
  x = currentNode;
  currentNode = n;
  return x;
}

/**
   Reduces to Basic Signal a node (n).
   Restriction: not applied for arrays, structures/bundle, delays.
*/
void
DCG2KernelTranslator::node2Kernel (ptINode n)
{
  ptlINode ln;
  ptlinsth lh;
  ptINode n1, old_n;

  old_n = setCurrentNode (n);
  switch (NM->get_oper (n))
    {
    case opdefsig:		/* X := E */
      equ2kernel (n);
      break;

    case opsubgraph:
      NM->get_subGraphDefinitions (n, &ln, &lh);
      while (ptlINode_get_elem (&ln, &n1))
	node2Kernel (n1);
      break;

    case opappe:
    case opnodecall:
      call2kernel (n);
      break;

    case opret:		/* Delay : to be completed. */
      delay2kernel (n);
      break;

    case opldefpart:
    case opnil:
      break;

    case opdefpart:		/* X ::= E */
      // It is not defined in the kernel. 
      setFunctionalityNotApplied (n);
      break;

    default:			/* To be completed for: case opmoti: case opbund: */
      // Assumed: the method splitDataStructure() has been applied before this transformation.
      setFunctionalityNotApplied (n);
      break;

    }
  setCurrentNode (old_n);
}



void
DCG2KernelTranslator::setFunctionalityNotApplied (ptINode node)
{
  ptlinstances lsig;
  ptexpcal vsig;
  ptrepstr vstr;

  bachieved = pk_false;
  lsig = NM->getInodelOutputFlows (node);
  if (ptlinstances_get_next (&lsig, &vsig))
    vstr = ident2repstr (EM->getName (vsig));
  else
    vstr = nilstr;
  pK_messenger->warningstr (war2kernel, vstr, pk_true);
}
