/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** It adds the clock equations induced by the process CLK_AFFINE.
    The affine relation is in canonical form. 
    Extract from the Irina Smarandache Ph.D (Proposition 16 page 141):
    " If X and Y are in relation (n, phi, d), then it exists U such that X and U in relation (1, max(0, -phi), n), and,
      U and Y in relation (1, max(0, phi), d)."
    So, it creates of a new variable clock (U).
    @return the null clock.
*/
    void
 intemkIGraph::makeAffineClock(ptexpcal n, ptexpcal phi, ptexpcal d, ptinsth * vh1, ptinsth * vh2)
{
  ptinsth vhu;
  ptAffineRelation affr1, affr2;

  affrmkdivide(n, phi, d, &affr1, &affr2);
  vhu = CKM->new_variFormulaRootClock();
  *vh1 = CKM->makeSubAffine(vhu, affr1);
  *vh2 = CKM->makeSubAffine(vhu, affr2);
}

/**
  It adds the clock equations induced by the process SAMPLE.
  The affine relation is in canonical form.
  @param lparaff (in) the parameters (phi, d) of the process SAMPLE.
  @param vh (in) the sampled clock.
  @return the undersampling of the clock (vh) with phi phase and the d period d.
*/
ptinsth intemkIGraph::makeSampleClock(ptexpcal lparaff, ptinsth vh)
{
  compilerInteger period_n, phase, period_d;
  ptAffineRelation aux;

  period_n = 1;
  EM->getIntegerValue(EM->getNaryArgs(&lparaff), &phase);
  EM->getIntegerValue(EM->getNaryArgs(&lparaff), &period_d);
  aux = mkAffineRelation(period_n, phase, period_d);
  return CKM->makeSubAffine(vh, aux);
}

/**  It adds the clock equations induced by the process UNSAMPLE.
     The affine relation is in canonical form.
     @param lparaff (in) the parameters (n, phi) of the process UNSAMPLE.
     It creates the affine clocks induced by the affine relation (n, phi, 1) -- see makeAffineClock() --
     @param vh1 (out) the undersampling by (1, max(0, -phi), n).
     @return the undersampling by (1, max(0, phi), 1).
     
*/
ptinsth intemkIGraph::makeUnsampleClock(ptexpcal lparaff, ptinsth * vh1)
{
  ptinsth vh;
  ptexpcal n, phi, d;

  n = EM->getNaryArgs(&lparaff);
  phi = EM->getNaryArgs(&lparaff);
  d = EM->mkIntegerConstant(1);
  makeAffineClock(n, phi, d, vh1, &vh);
  return vh;
}

/** Adding the definition y := x when hy for a reference to the process 
    y := sample{phi, n}(x) reference. 
*/
void intemkIGraph::mkSampleNode(ptexpcal vy, ptexpcal vx)
{
  ptlinstances instent;
  ptinsth vhvar;

  instent = ptlinstances_insert(NILLINSTANCES, vx);
  vhvar = setClockVariable(vy);
  createEquation(vy, vx, vhvar, vhvar, instent, NILBACKANNOTATION);
}

/** Adding the definition vy := (vx when  ^vx ^* ^vy) default (vz when ^vy ^- ^vx).
    for a reference to the process
    vy := unsample {n,phi}(vx, vz)
*/
void intemkIGraph::mkUnsampleNode(ptexpcal vy, ptexpcal vx, ptexpcal vz)
{
  ptinsth vhx, vhy;
  vhx = setClockVariable(vx);
  vhy = setClockVariable(vy);
  addSubEquation(vy, vx, CKM->makeInf(vhx, vhy), ptlinstances_insert(NILLINSTANCES, vx), NILBACKANNOTATION);
  createEquation(vy, vz, CKM->makeMinus(vhy, vhx), vhy, ptlinstances_insert(NILLINSTANCES, vz), NILBACKANNOTATION);
}
