/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor. The object is linked with a graph (g).
   The provided methods work on this graph.
   It provides methods to split the structure/bundle type of the linked graph.
*/
  splitting::splitting (ptDCG g)
{
  AGRAPH = g;
  EM = AGRAPH->getExpressionManager ();
  CKM = AGRAPH->getClockManager ();
  NM = AGRAPH->getNodeManager ();
  linputs = AGRAPH->getInputs ();
  loutputs = AGRAPH->getOutputs ();
  bresult = pk_false;
}

/**
   Destructor.
*/
splitting::~splitting ()
{
}

/** @defgroup exprSplitGroup: splitting:Expressions  */

/** @ingroup exprSplitGroup
    Getting the i-th field of a synchronous expression.
    @param exp (in) : an (assumed) synchronous expression.
    @param i (in) : the wanted field number.
    @return the i-th field of a synchronous expression.
 */
ptexpcal splitting::mkFieldOfSynchronousExpression (ptexpcal exp, integer i)
{
  noperators
    op;
  ptexpcal
    e1,
    e2,
    e3;

  if (exp == NILEXPCAL)
    return exp;
  op = EM->getOperator (exp);
  switch (EM->getOperatorClass (exp))
    {
    case constant:
      return exp;
    case opertyp:
      return exp;

    case variabl:
      return getIthFieldForVariable (exp, i);

    case oper1:
      if (op == opvvarin)
	return getIthFieldForVariable (exp, i);	// x?
      e1 = mkFieldOfSynchronousExpression (EM->getUnaryArg (exp), i);
      return EM->mkUnaryExp (op, e1, EM->getBackAnnotation (e1));

    case oper2:
      if (EM->satisfiesList (exp))
	return getIthElementOfList (exp, i);

      if (op == opcham)
	return getIthFieldForVariable (exp, i);
      if (op == opelem)
	{
	  e1 = mkFieldOfSynchronousExpression (EM->getBinaryArgstAt (exp, 1), i);
	  e2 = EM->getBinaryArgstAt (exp, 2);
	}
      else
	{
	  e1 = mkFieldOfSynchronousExpression (EM->getBinaryArgstAt (exp, 1), i);
	  e2 = mkFieldOfSynchronousExpression (EM->getBinaryArgstAt (exp, 2), i);
	}
      return EM->mkBinaryExp (op, e1, e2, NILBACKANNOTATION);

    case oper3:
      e1 = mkFieldOfSynchronousExpression (EM->getTernaryArgstAt (exp, 1), i);
      e2 = mkFieldOfSynchronousExpression (EM->getTernaryArgstAt (exp, 2), i);
      e3 = mkFieldOfSynchronousExpression (EM->getTernaryArgstAt (exp, 3), i);
      return EM->mkTernaryExp (op, e1, e2, e3, NILBACKANNOTATION);

    case call:
      // mkProcCall from expression, then take the i-th output;
      return exp;		// to be completed.
    default:
      return NILEXPCAL;
    }
}


/**
   @ingroup exprSplitGroup
   @return the list of the fields associated with an expression (exp) with structrue/bundle as type or an array type built on structured type,
   an empty list for a non structure/bundle type.
   Assumed: the expression (exp) must be an identifier (example: for a signal x with struct t1 y; t2 z; end, it produces {x.y, x.z})
   For more details, see 
     - mkFieldsDescription(TypeType, ptexpcal , ptexpcal , boolean ).
     - isStructureBundleType(TypeType , TypeType , ptexpcal , boolean)
*/
ptlinstances splitting::mkFieldsDescription (ptexpcal exp)
{
  TypeType
    t;
  ptlinstances
    l;
  ptexpcal
    vdim;
  boolean
    bstruct;

  if (EM->getFieldsOf (exp, &l))
    return l;
  t = EM->getUnamedType (exp);
  if (!isStructureBundleType (t, &t, &vdim, &bstruct))
    return NILLINSTANCES;
  l = mkFieldsDescription (t, exp, vdim, bstruct);
  EM->setFieldsOf (exp, l);
  return l;
}


/** @defgroup nodeSplitGroup: splitting:nodes  */

/** @ingroup nodeSplitGroup
    Splitting structured type definition for a delay (n).
*/
boolean splitting::splitDelay (ptINode n)
{
  return splitCommon (n);
}

/** @ingroup nodeSplitGroup 
    Splitting structured type definition for an equation (n)
 */
boolean splitting::splitEqu (ptINode n)
{
  return splitCommon (n);
}

/**@ingroup nodeSplitGroup
   Splitting structured type definition for a partial definition (n).
*/
boolean splitting::splitPartialDefinition (ptINode n)
{
  ptlINode
    lnodes;
  ptINode
    node;
  boolean
    b;

  lnodes = NM->get_setOfPartialDefinitions_arg (n);
  while (ptlINode_get_elem (&lnodes, &node))
    if (splitCommon (node))
      b = pk_true;
  return b;
}

/**
   @ingroup nodeSplitGroup
   Splitting structured type definition for all the delays of the graph.
*/
void
splitting::splitStructuresDelays ()
{
  ptlINode lnodes;
  ptINode node;

  if (NM->getDelays (&lnodes))
    while (ptlINode_get_elem (&lnodes, &node))
      if (splitDelay (node))
	bresult = pk_true;
}

/**
   @ingroup nodeSplitGroup
   Splitting structured type definition for all the equations of the graph.
*/
void
splitting::splitStructuresEquations ()
{
  ptlINode lnodes;
  ptINode node;

  if (NM->getEquations (&lnodes))
    while (ptlINode_get_elem (&lnodes, &node))
      if (splitEqu (node))
	bresult = pk_true;
}

/**
   @ingroup nodeSplitGroup
   Splitting structured type definition for all the partial definitions of the graph.
*/
void
splitting::splitStructuresPartialDefinitions ()
{
  ptlINode lnodes;
  ptINode node;

  if (NM->getGroupsOfPartialEquations (&lnodes))
    while (ptlINode_get_elem (&lnodes, &node))
      if (splitPartialDefinition (node))
	bresult = pk_true;

}

/**
   @ingroup nodeSplitGroup
   Splitting structured type definition for all the external calls of the graph.
   \warning Not applied for node calls.
*/
void
splitting::splitStructuresCalls ()
{
  ptlINode lnodes;
  ptINode node;

  if (NM->getInputs (&lnodes))
    while (ptlINode_get_elem (&lnodes, &node))
      if (splitInput (node))
	bresult = pk_true;

  if (NM->getOutputs (&lnodes))
    while (ptlINode_get_elem (&lnodes, &node))
      if (splitOuput (node))
	bresult = pk_true;

  if (NM->getProcCalls (&lnodes))
    while (ptlINode_get_elem (&lnodes, &node))
      if (splitProcCall (node))
	bresult = pk_true;

  // case opnodecall: 
}


/**
   @ingroup nodeSplitGroup
   Splitting structured type definition for all the nodes of the graph and the definitions of constants.
   Then
      - the field references are replaced by standart signals (see replaceFieldBySignal()), 
      - the structured nodes are removed (see removeGlobalStructureNode()).
      - the interface of the graph is updated (can be modified).
   \warning Not applied for node calls. For arrays and sub-graph it is without effect : the effect is in the inner nodes.
*/
void
splitting::splitStructures ()
{
  splitStructuresEquations ();
  splitStructuresDelays ();
  splitStructuresPartialDefinitions ();
  splitStructuresCalls ();

  splitStructuresConstants ();
  // case opmoti:
  // case opsubgraph:    case opbund: 
  // if (bresut) 
  {
    EM->replaceFieldBySignal ();
    removeGlobalStructureNode ();
    AGRAPH->setInputs (linputs);
    AGRAPH->setOutputs (loutputs);
  }
}
