/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PRIVATE_
/**
   @return "true" if a type (t) is a structure, a bundle type or an array type built on structure/bundle type, "false" otherwise.
*/
  boolean splitting::isStructureBundleType (TypeType t)
{
  TypeType t1;

  if (TypeTestStructureClass (t))
    return pk_true;
  if (TypeTestBundleClass (t))
    return pk_true;
  if (TypeTestArrayClass (t))
    {
      t1 = ArrayMatrixTypeGetTerm (t);
      if (TypeTestNamedClass (t1))
	t1 = NamedTypeGetTerm (t1);
      return isStructureBundleType (t1);
    }
  return pk_false;
}


/** 
    @return "true" if a type (t) is a structure or bundle type or a array type built on a structure/bundle type, "false" otherwise.
    @param t (in) a type.
    @param TT (out) different from t if it is an array type, in this case it is the structured type.
    @param vdim (out) dimensions of the operand type (t)
    @param bstruct (out) "true" when TT it is a structured type, "false" for a bundle.
*/
boolean
splitting::isStructureBundleType (TypeType t, TypeType * TT, ptexpcal * vdim, boolean * bstruct)
{
  TypeType t1;

  if (TypeTestStructureClass (t))
    {
      *vdim = NILEXPCAL;
      *bstruct = pk_true;
      *TT = t;
      return pk_true;
    }
  if (TypeTestBundleClass (t))
    {
      t = BundleTypeGetTerm (t);
      *TT = t;
      *vdim = NILEXPCAL;
      *bstruct = pk_false;
      return pk_true;
    }

  if (TypeTestArrayClass (t))
    {
      t1 = ArrayMatrixTypeGetTerm (t);
      if (TypeTestNamedClass (t1))
	t1 = NamedTypeGetTerm (t1);
      if (isStructureBundleType (t1, TT, vdim, bstruct))
	{
	  *vdim = ArrayMatrixTypeGetDimLst (t);
	  return pk_true;
	}
    }
  return pk_false;
}

/**
   @return the created node (equation) for the definition of a signal (vField) by the i-th field of an expression (definition).
   The i-th expression defines the signal at a given clock (hc).
*/
ptINode
splitting::mkNodeIthFieldWhen (ptexpcal vField, ptexpcal definition, integer i, ptinsth hc)
{
  ptexpcal e;
  ptlinstances instent;

  e = mkFieldOfSynchronousExpression (definition, i);
  instent = EM->getSignalsReferencedIn (e);
  return NM->mk_node_equation (vField, e, hc, instent, NILBACKANNOTATION /* a revoir */ );
}

/**
   @return the created node (delay node) for the definition of a signal (vField) by the i-th field of an expression (definition).
*/
ptINode
splitting::mkNodeIthFieldDelay (ptexpcal vField, ptexpcal definition, integer i)
{
  ptexpcal e1, vinit, vdelay, vwin, delayedSignal;
  ptexpcal delayedSignalField, vinitField;

  e1 = EM->getTernaryArgstAt (definition, 1);
  delayedSignal = EM->getBinaryArgstAt (e1, 1);
  vinit = EM->getBinaryArgstAt (e1, 2);
  vdelay = EM->getTernaryArgstAt (definition, 2);
  vwin = EM->getTernaryArgstAt (definition, 3);

  delayedSignalField = mkFieldOfSynchronousExpression (delayedSignal, i);
  vinitField = mkFieldOfSynchronousExpression (vinit, i);
  EM->setInitialValue (vField, vinitField);
  return NM->mk_node_delay (vField, delayedSignalField, vdelay, vwin, vinitField, NILBACKANNOTATION);
}

/**
   @return the created node for the definition of a signal (vField) by the i-th field of an expression (definition).
   It creates a delay node(mkNodeIthFieldDelay()) or an equation node (mkNodeIthFieldWhen()).
*/
ptINode
splitting::splitExpression (ptexpcal vField, ptexpcal definition, integer i)
{
  ptinsth hc;
  ptexpcal e1, e2;

  if (EM->getOperator (definition) == opret)
    return mkNodeIthFieldDelay (vField, definition, i);
  if (EM->IsCellExpr (definition))
    {
      // (e when hc) cell k, k is the clock of  vField.
      definition = EM->getWhenOrCellArg (definition);
      hc = EM->getContributionClock (definition);
      definition = EM->getBasicExpr (definition);
      return mkNodeIthFieldWhen (vField, definition, i, hc);
    }
  hc = EM->getContributionClock (definition);

  if (EM->getOperator (definition) == opmele)
    {
      e1 = EM->getBinaryArgstAt (definition, 1);
      splitExpression (vField, e1, i);
      e2 = EM->getBinaryArgstAt (definition, 2);
      return splitExpression (vField, e2, i);
    }

  // SynchronousExpression when hc
  definition = EM->getBasicExpr (definition);
  return mkNodeIthFieldWhen (vField, definition, i, hc);
}

/**
   @return the i-th element of a list (exp), undefined element it does not exist.
*/
ptexpcal
splitting::getIthElementOfList (ptexpcal exp, integer i)
{
  ptexpcal arg;
  integer n;

  n = 1;
  while (EM->getNaryArgs (&exp, &arg))
    if (n == i)
      return arg;
    else
      n++;
  return NILEXPCAL;
}

/**
   @return the i-th field of a signal (exp) or exp if it does not exist.
*/
ptexpcal
splitting::getIthFieldForVariable (ptexpcal exp, integer i)
{
  ptlinstances l;
  ptexpcal x;

  l = mkFieldsDescription (exp);
  if (ptlinstances_get_ith_element (l, i, &x))
    return x;
  return exp;			// error ?
}

/**
   @return a new signal with vstr as prefix for its name, with TT (an array type) as type, and h as clock.
   A link is kept for removing the field access.
*/
ptexpcal
splitting::mkSignalForArrayOfField (ptexpcal exp, ptrepstr vstr, TypeType TT, ptinsth h)
{
  ptexpcal vexp;

  vexp = EM->mkIdentExp (repstr2ident (vstr), tsignal, TT, NM->mk_node_undefined (), h, NILBACKANNOTATION);
  EM->setClock (vexp, h);
  return vexp;
}

/**
   @return a new signal (tfield) with vstr as prefix for its name, with TT as type, and h as clock.
   This signal is a renaning of the field access (exp.str).
   A link is kept for removing the field access.
   For example, \e foo5 can be the signal that represents the field access \e x.foo.
   \warning exp is assumed to be an identifier.
*/
ptexpcal
splitting::mkSignalForField (ptexpcal exp, ptrepstr vstr, TypeType TT, ptinsth h)
{
  ptexpcal vexp, vexpacc;

  vexpacc = EM->mkBinaryExp (opcham, exp, EM->mkConstantExp (opnom, vstr), NILBACKANNOTATION);
  vexp = EM->mkIdentField (repstr2ident (vstr), TT, h);
  EM->setClock (vexp, h);
  EM->setSignalAssignedTo (vexpacc, vexp);
  return vexp;
}


/**
   @return a list of signals that represent the fields of a signal (exp).
   @param t (in) a structured type (structure or bundle)
   @param exp (in) a signal with t as base type (completed by the dimensions vdim)
   @param vdim (in) the dimension associated with  the base type (t)  -- for an array type --.
   @param bstructure (in) when it is "true", t is a structure type, otherwise a bundle one.
   
    For an X of an array of structured type such that
     [N] structure
            T1 x1;
	    ...
	    Tn xn; 
          end structure
     It is splitted into several signals: Y1 declared [N] T1 X_x1, ...; Yn declared [N] Tn X_xn.
*/
ptlinstances
splitting::mkFieldsDescription (TypeType t, ptexpcal exp, ptexpcal vdim, boolean bstructure)
{
  ptinsth hundef, h1;
  TypeType TT;
  NameList NL;
  TypeList fieldsType;
  TypeSubList NSubLstType;
  boolean Goon, bstate, bconst;
  NameSubList NSubLst;
  ptexpcal vexp, s;
  ptlinstances description_type;
  ptrepstr vstr, vstr1, idstr;

  bconst = EM->IsConstant (exp);

  bstate = EM->isStateVar (exp) || (EM->getOperator (exp) == opvvarin);
  idstr = ident2repstr (EM->getName (exp));
  description_type = NILLINSTANCES;
  NL = StructureTypeGetNameLst (t);	// field Name
  NSubLst = NameListCast2SubList (NL);
  fieldsType = StructureTypeGetTypeLst (t);
  NSubLstType = TypeListCast2SubList (fieldsType);
  Goon = pk_true;
  hundef = CKM->getUndefClock ();
  while (Goon)
    {
      if (bconst)
	h1 = hundef;
      else if (bstate)
	h1 = CKM->getFictitiousRoot ();
      else if (bstructure)
	{
	  h1 = EM->getClock (exp);
	  if (CKM->satisfiesUndefClock (h1))
	    {
	      h1 = CKM->new_variFormulaRootClock ();
	      EM->setClock (exp, h1);
	    }
	}
      else
	h1 = CKM->new_variFormulaRootClock ();

      TT = TypeSubListGetHead (NSubLstType);
      vstr = NameSubListGetHead (NSubLst);
      if (vdim == NILEXPCAL)
	vexp = mkSignalForField (exp, vstr, TT, h1);
      else
	{
	  TT = EM->mkArrayType (vdim, TT);
	  vexp = mkSignalForArrayOfField (exp, vstr, TT, h1);
	}
      if (bstate)
	EM->setStateTag (vexp, pk_true);
      if (bconst)
	AGRAPH->addConstant (vexp);

      if (!EM->getSignalAssignedTo (exp, &s))
	s = exp;
      ptident id = EM->getName (s /*exp */ );
      EM->setName (vexp, createSuffIdentField (id, vstr));

      description_type = ptlinstances_push_end (description_type, vexp);
      if (NameSubListTestLast (NSubLst))
	Goon = pk_false;
      else
	{
	  NSubLst = NameSubListGetTail (NSubLst);
	  NSubLstType = TypeSubListGetTail (NSubLstType);
	}
    }
  return description_type;
}


/**
   It initialises each node of a list signals (ls) for splitting operation.
   (referenced in initSplitNode()).
*/
void
splitting::initSplitNodes (ptlinstances ls, ptINode n)
{
  ptexpcal s;
  ptINode n1;

  while (ptlinstances_get_next (&ls, &s))
    {
      n1 = EM->getNode (s);
      if (n1 == n)
	{
	  EM->setSignalAssignedTo (s, NILEXPCAL);
	  EM->setNode (s, NM->mk_node_undefined ());
	}
    }
}

/**
   @return the list of the signals associated with the different fields of a signal.
   For a node (n) defining  a signal (s), it initializes the splitting operation:
     - n becomes an undefined node
     - it creates the signals associated with the different fields (mkFieldsDescription() method)
     - it initializes each created signal
     - n becomes a "composed node" that groups the nodes of the different fields (mk_node_lfields()).
*/
ptlinstances
splitting::initSplitNode (ptINode n, ptexpcal s)
{
  ptlinstances l;

  NM->becomes_undefined (n);
  l = mkFieldsDescription (s);
  initSplitNodes (l, n);
  NM->mk_node_lfields (s, n, l, NM->getBackAnnotation (n));
  return l;
}


/**
   It finalizes the splitting operation for a node (n) replaced by a set of nodes (ln):
     - the pragmas associated with n are propagated to ln.
*/
void
splitting::finalyzeSplitNode (ptINode n, ptlINode ln)
{
  ptlINode laux;
  ptINode node;
  ptlpragmas lp;

  // Setting pragmas.
  lp = NM->getPragmas (n);
  laux = ln;
  while (ptlINode_get_elem (&laux, &node))
    NM->addPragmas (node, lp);
}

/** 
    Splitting structured type definition for Delays, equation, and partial definition.
*/
boolean
splitting::splitCommon (ptINode n)
{
  ptexpcal definedSignal, definition;
  ptinsth h;
  TypeType t;
  boolean partialDef, bdefaultValue;
  ptlINode ln;
  ptlinstances l;

  definedSignal = NM->get_leftOperand (n);
  t = EM->getUnamedType (definedSignal);
  if (!isStructureBundleType (t))
    return pk_false;
  // a voir pour les successeurs
  //  lh = NM->get_clockSuccessors(n);
  //  lsucc = get_successors(n);
  partialDef = (NM->get_oper (n) == opdefpart);
  bdefaultValue = (partialDef && NM->is_partialDefinitionDefaultValue (n));
  NM->get_rightOperand (n, &definition);
  l = initSplitNode (n, definedSignal);
  h = EM->getClock (definedSignal);
  ln = mkDefinitionFields (l, definition, h, partialDef, bdefaultValue);
  finalyzeSplitNode (n, ln);
  return pk_true;
}

/**
   It defines the i-th element of l by the acces to the i-th element of a definition (definition). All the created nodes, have h as clock.
   Some additionnal information are used:
      - partialDef when it is "true", it specify that it is a partial definition.
      - bpartdefault when it is "true", it specify that it is the default value of a partial definition.
  @return the create nodes.
  \see splitExpression
*/
ptlINode
splitting::mkDefinitionFields (ptlinstances l, ptexpcal definition, ptinsth h, boolean partialDef, boolean bpartdefault)
{
  ptexpcal vf;
  ptINode node;
  integer currentFieldNumber;
  ptlINode ln;

  // for each field, ....
  currentFieldNumber = 1;
  ln = NILLINODE;
  while (ptlinstances_get_next (&l, &vf))
    {
      if (partialDef)
	EM->addMark (vf, PART_DEFINITION);
      if (bpartdefault)
	EM->addMark (vf, PART_DEFAULT_DEFINITION);

      node = splitExpression (vf, definition, currentFieldNumber);
      NM->insertNode (node, h);
      ln = ptlINode_append_elem (ln, node);

      if (partialDef)
	EM->removeMark (vf, PART_DEFINITION);
      if (bpartdefault)
	EM->removeMark (vf, PART_DEFAULT_DEFINITION);

      currentFieldNumber++;
    }
  return ln;
}


/**
   Splitting structured type definition for a "write" node (associated with an output signal).
*/
boolean
splitting::splitOuput (ptINode node)
{
  ptexpcal e, vf;
  TypeType t;
  ptinsth h;
  ptlinstances l;
  ptINode n;
  ptlINode ln;

  e = NM->get_io_arg (node);
  t = EM->getUnamedType (e);
  if (!isStructureBundleType (t))
    return pk_false;
  loutputs = ptlinstances_remove_exp (loutputs, e);
  NM->removeNode (node);
  l = mkFieldsDescription (e);
  h = EM->getClock (e);
  ln = NILLINODE;
  while (ptlinstances_get_next (&l, &vf))
    {
      n = NM->mk_node_output (vf, NM->getBackAnnotation(node));
      NM->insertNode (n, h);
      ln = ptlINode_append_elem (ln, n);
    }
  return pk_true;
}


/**
   Splitting structured type definition for a "read" node (associated with an input signal).
*/
boolean
splitting::splitInput (ptINode node)
{
  ptexpcal e, vf;
  TypeType t;
  ptinsth h;
  ptlinstances l;
  ptINode n;
  ptlINode ln;

  e = NM->get_io_arg (node);
  t = EM->getUnamedType (e);
  if (!isStructureBundleType (t))
    return pk_false;
  linputs = ptlinstances_remove_exp (linputs, e);
  l = initSplitNode (node, e);
  linputs = ptlinstances_sortedfusion (l, linputs);
  h = EM->getClock (e);
  ln = NILLINODE;
  while (ptlinstances_get_next (&l, &vf))
    {
      n = NM->mk_node_input (vf, NILINODE, NM->getBackAnnotation(node));
      NM->insertNode (n, h);
      ln = ptlINode_append_elem (ln, n);
    }

  finalyzeSplitNode (node, ln);
  return pk_true;
}

/**
   @return "true" if the type of a signal (e) is structure/bundle one, "false" otherwise.
   When it returns "true", l is the list of the signals that represent the fields of the operand signal.
*/
boolean
splitting::getListOfFields (ptexpcal e, ptlinstances * l)
{
  TypeType t;

  if (EM->getOperatorClass (e) != variabl)
    return pk_false;
  t = EM->getUnamedType (e);
  if (!isStructureBundleType (t))
    return pk_false;
  *l = mkFieldsDescription (e);
  return pk_true;
}

/**
   @return  the expanded form of e (a signal or a list of signal) when structure type, bundle type or array type built on structure/bundle type are
   splitted (in this case, a signal is replaced by N signals, N is the number of attribute of the structured type.)
*/
ptexpcal
splitting::expandStructureToFields (ptexpcal e)
{
  ptexpcal e1, elem, ne1;
  ptlinstances l;
  noperators op;

  if (EM->isNILEXPCAL (e))
    return e;
  if (EM->satisfiesList (e))
    {
      e1 = e;
      op = EM->getOperator (e);
      EM->getNaryArgs (&e1, &elem);
      ne1 = expandStructureToFields (e1);	// list
      if (!getListOfFields (elem, &l))
	{
	  if (e1 == ne1)
	    return e;
	  return EM->mkNaryExp (op, elem, ne1, NILBACKANNOTATION);
	}
      l = ptlinstances_reverse (l);
      while (ptlinstances_get_next (&l, &elem))
	ne1 = EM->mkNaryExp (op, elem, ne1, NILBACKANNOTATION);
      return ne1;
    }
  if (!getListOfFields (e, &l))
    return e;
  return EM->convertLinstances2ListExpcal (l);
}

/**
   Splitting structured type definition for an external procedure call node.
*/
boolean
splitting::splitProcCall (ptINode node)
{
  ptexpcal params, OrderedInputs, OrderedOutputs;
  ptexpcal nparams, nOrderedInputs, nOrderedOutputs;
  ptident name;

  NM->get_procCall_arg (node, &name, &params, &OrderedInputs, &OrderedOutputs);
  nparams = expandStructureToFields (params);
  nOrderedInputs = expandStructureToFields (OrderedInputs);
  nOrderedOutputs = expandStructureToFields (OrderedOutputs);
  if (params != nparams)
    NM->update (node, params, nparams);
  if (OrderedInputs != nOrderedInputs)
    NM->update (node, OrderedInputs, nOrderedInputs);
  if (OrderedOutputs != nOrderedOutputs)
    NM->update (node, OrderedOutputs, nOrderedOutputs);
  return pk_true;
}

/**
   For a "composed node" (node) that groups the nodes of the different fields
   It removes the node (node) from the graph after propagating its dependences to its components.
*/
void
splitting::removeGlobalStructureNode (ptINode node)
{
  ptlinstances ls;
  ptexpcal aField;
  ptINode n;

  ls = NM->get_lfields_arg (node);	// set of fields
  while (ptlinstances_get_next (&ls, &aField))
    {
      n = EM->getNode (aField);
      if (n != node)
	{
	  NM->removeDependence (n, node);
	  NM->dependencesMove (node, n);
	  NM->removeNode (node);
	}
    }
}

/**
   It applies removeGlobalStructureNode(ptINode) for each "composed node" of the 
   graph that groups the nodes of the different fields.
   Then, it removes the structured type constant.
*/
void
splitting::removeGlobalStructureNode ()
{
  ptlINode lNodes;
  ptINode node;
  ptlinstances ls;
  ptexpcal exp;

  if (NM->getBundles (&lNodes))
    while (ptlINode_get_elem (&lNodes, &node))
      removeGlobalStructureNode (node);

  ls = AGRAPH->getConstants ();
  while (ptlinstances_get_next (&ls, &exp))
    {
      node = EM->getNode (exp);
      if (NM->is_lfields (node))
	AGRAPH->removeConstant (exp);
    }
}


/**
   Splitting structured type definition for the constants of the graph.
   (similar to splitEqu() method).
*/
void
splitting::splitStructuresConstants ()
{
  ptlinstances liste;
  ptexpcal exp;
  ptINode n;

  liste = AGRAPH->getConstants ();
  while (ptlinstances_get_next (&liste, &exp))
    {
      n = EM->getNode (exp);
      splitEqu (n);
    }
}
