/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_

/** Clock.
   @return true if h is a predefined clock (i.e undef, null, init, FictitiousRoot).
*/
  boolean ClockFormula::satisfiesPredefinedClock (ptinsth h)
{
  return ptinsth_satisfies_predef (h);
}

/** Clock.
   @return the representative formula assigned to the clock ?aFormula.
*/
insthClock ClockFormula::get_insthClock (insthFormula aFormula)
{
  if (aFormula == NILINSTH)
    return NILINSTH;
  return ptinsth_get_clhead (aFormula);
}

/** Clock.
   @return the representative formula assigned to the clock ?vh. */
ptinsth ClockFormula::getRepresentativeFormula (ptinsth vh)
{
  return ptinsth_get_clhead (vh);
}

/** Clock.
   @return the successor of the formula vh in its class with filtering of the equivalent 
   formula when constraints is "true". 
*/
ptinsth ClockFormula::getNextFormula (ptinsth vh, boolean constraints)
{
  return ptinsth_get_clsucc (vh, constraints);
}

/** Clock.
   @return true if it exists an assertion in the successors of the formula vh in its class.
   @param vh (in) : a clock.
   @param vhassert (out) : the assertion if it exists.
*/
boolean ClockFormula::getNextAssertion (ptinsth vh, ptinsth * vhassert)
{
  return ptinsth_get_nextassertion (vh, vhassert);
}

/** Clock.
    @return true if the clock vh is reduced to one formula.
 */
boolean ClockFormula::satisfiesSingleton (ptinsth vh)
{
  return ptinsth_satisfies_clsingleton (vh);
}

/** Clock.
    @return true if the clock vh is a variable..
 */
boolean ClockFormula::satisfiesVari (ptinsth vh)
{
  return ptinsth_satisfies_vari (vh);
}

/** Clock.
   We have getClock(x) = vh, for each x of the returned list.
   @return the set of signals assigned to the clock.
 */
ptlinstances ClockFormula::getSynchronousSignals (ptinsth vh)
{
  return ptinsth_get_lsig (vh);
}

/** Clock.
   We have getClock(x) = vh, for each x of the returned list, but x has been removed from the graph (tracability).
   @return the set of signals assigned to the clock.
 */
ptlinstances ClockFormula::getSynchronousSignalsKept (ptinsth vh)
{
  return ptinsth_get_lsigkept (vh);
}

/** It assigns the identifier id as the name of the clock */
void
ClockFormula::setName (ptinsth vh, ptident id)
{
  ptinsth_set_name (vh, id);
}

/** Clock.
   Assumes : getClock(x) = vh, for each x of lsig.
   It assignes to the clock vh the set of synchronous signals lsig.
*/
void
ClockFormula::setSynchronousSignals (ptinsth vh, ptlinstances lsig)
{
  ptinsth_set_lsig (vh, lsig);
}

/** Clock.
   Assumes : getClock(x) = vh, for each x of lsig, and x is removed from the graph. It is kept for tracability.
   It assignes to the clock vh the set of synchronous signals lsig.
*/
void
ClockFormula::setSynchronousSignalsKept (ptinsth vh, ptlinstances lsig)
{
  ptinsth_set_lsigkept (vh, lsig);
}

/**
   @return
   - "true" if a formula v is an affine clock, affr is its parameters
   - "false" otherwise, affr is not modified.
*/
boolean ClockFormula::getAffineRelation (ptinsth v, ptAffineRelation * affr)
{
  return ptinsth_get_affinerelation (v, affr);
}

/**
   @return
   - "true" if a formula v is an affine clock.
   - "false" otherwise.
*/
boolean ClockFormula::isAnAffineClock (ptinsth v)
{
  ptAffineRelation affr;
  return ptinsth_get_affinerelation (v, &affr);
}

/**
   @return
   - "true" if a formula v is an affine clock, affr is its parameters
   - "false" otherwise, affr is not modified.
*/
void
ClockFormula::setAffineRelation (ptinsth v, ptAffineRelation affr)
{
  ptinsth_set_affinerelation (v, affr);
}

/**
   It declares a formula (vh) as a equivalent one if b.
*/
void
ClockFormula::setIsAnEquivalentForm (ptinsth vh, boolean b)
{
  ptinsth_set_isAnEquivalentForm (vh, b);
}

/**
   It removes a formula (vform) if it is a formula of a clock (vhcl) and if it is not the representative formula of this class.
   When it is removed, the formula (vform) becomes a new classes of formula (reduced to vform).
   @return "true" if the operation succeeds, "false" otherwise (it is already a representative formula).
*/
boolean ClockFormula::removeFormulaFromClass (ptinsth vclk, ptinsth vform)
{
  return ptinsth_remove_form (vclk, vform);
}

/**
   It removes a formula (vform) from its class and updates the references.
   @param vform (in): an undersampling formula (assumed).
   @param vcond (in): the condition associated with vform.
   @param becomesvari (out): set to "true" value if the new class is reduced to a free variable, "false" otherwise.
   @return the representative formula of the new class.
*/
ptinsth ClockFormula::removeSubFormulaFromClass (ptinsth vform, ptcondition vcond, boolean * becomesvari)
{
  return ptinsth_remove_sub (vform, vcond, becomesvari);
}

/**
   Assumed: vh is an undersampling formula.
   It replaces the clock of the condition by another one (argh).
*/
void
ClockFormula::replaceArgSub (ptinsth vh, ptinsth argh)
{
  ptinsth_replace_argsub (vh, argh);
}

/**
   It assigns a condition (vcond) to an undersampling formula (vh). 
   Assumed : vh is a undersampling.
*/
void
ClockFormula::setArgSub (ptinsth vh, ptcondition vcond)
{
  ptinsth_set_argsub (vh, vcond);
}

/**
   It removes a formula (vform) from its class and updates the references.
   @param vform (in): a not-formula (assumed).
   @param becomesvari (out): set to "true" value if the new class is reduced to a free variable, "false" otherwise.
   @return the representative formula of the new class.
*/
ptinsth ClockFormula::removeNotFormulaFromClass (ptinsth vform, boolean * becomesvari)
{
  return ptinsth_remove_not (vform, becomesvari);
}

/**
   It declares a formula (vh) as a normalized one.
*/
void
ClockFormula::setNormalized (ptinsth vh)
{
  // definition of normalized?
  ptinsth_belongs_normalized (vh);
}

/**
   @return "true" is a clock (vh) is a normalized one, "false" otherwise.
*/
boolean ClockFormula::iSNormalized (ptinsth vh)
{
  // definition of normalized?
  return ptinsth_satisfies_normalized (vh);
}

/**
   It removes a formula (vform) of its class if it is not the representative formula.
   @return the representative formula of the class of the formula.
*/
ptinsth ClockFormula::removeFormulaFromClass (ptinsth vform)
{
  return ptinsth_remove_form_belongs_class (vform);
}

/** Assumed : vh is a binary formula.
    @return the other operand than a given operand (vh1).
*/
ptinsth ClockFormula::otherArgOf (ptinsth vh, ptinsth vh1)
{
  return ptinsth_get_otherarg (vh, vh1);
}

/**
   The formula (vh) is an assumption if b.
*/
void
ClockFormula::setAssumption (ptinsth vh, boolean b)
{
  ptinsth_set_AnAssumption (vh, b);
}

/**
  @return the class in which the representative formula is a given formula (vform).
  @param vnewhead (out): is "true" if the representative formula has been changed, "false" otherwise.
*/
ptinsth ClockFormula::resetclass (ptinsth vform, boolean * vnewhead)
{
  return ptinsth_resetclass (vform, vnewhead);
}

/** 
    It assigns the operands (arg1, arg2) to an assumed binary formula (vh).
 */
void
ClockFormula::setBinaryArgs (ptinsth vh, ptinsth arg1, ptinsth arg2)
{
  ptinsth_set_2args (vh, arg1, arg2);
}

/** 
    It assigns the operand (arg) to an assumed unary formula (vh).
 */
void
ClockFormula::setUnaryArg (ptinsth vh, ptinsth arg)
{
  ptinsth_set_1arg (vh, arg);
}

/** -- misc operation--
    If the operand (vh) is a complementary formula (vh = H * U, U = NOT(K))
    the reference from K to vh is removed, and the reference K to U is set.
    @return "true" if the operation is applied, "false" otherwise.
    \See ptinsth_switch_referenceshnot
*/
boolean ClockFormula::switchReferencesNotFormula (ptinsth vh)
{
  ptinsth
    vh1,
    vh2;
  if (getOperator (vh) != infer)
    return pk_false;
  getBinaryArgs (vh, &vh1, &vh2);
  if (ptinsth_switch_referenceshnot (vh, vh1))
    return pk_true;
  return ptinsth_switch_referenceshnot (vh, vh2);
}

/**
   It merges  two clasees (vh1 and vh2) in vh1.
*/
void
ClockFormula::fusionclass (ptinsth vh1, ptinsth vh2)
{
  ptinsth_fusionclass (vh1, vh2);
}

/** Set a reference form a clock (vFrom) to an another one (vTo).
 */
void
ClockFormula::setReferenceToFrom (ptinsth vTo, ptinsth vFrom)
{
  ptinsth_insert_dep (vTo, vFrom);
}

/**  It removes the reference from vFrom to vTo. */
void
ClockFormula::removeReferenceFromTo (ptinsth vFrom, ptinsth vTo)
{
  ptinsth_remove_ref (vFrom, vTo);
}

/**
  It removes the references to an assumed binary formula (vform) for its operands.
*/
void
ClockFormula::removeReferencesBinaryFormula (ptinsth vform)
{
  ptinsth_remove_binaryformula_link (vform);
}

/**
   It removes the references to an assumed undersampling formula (vform) with the condition vcond.
*/
void
ClockFormula::removeSubLink (ptinsth vform, ptcondition vcond)
{
  ptinsth_remove_sub_link (vform, vcond);
}

/**
   It removes a formula (vform) from its class and updates the references.
   @param   vform (in): a binary formula (assumed).
   @param becomesvari (out): set to "true" value if the new class is reduced to a free variable, "false" otherwise.
   @return the representative formula of the new class.
*/
ptinsth ClockFormula::removeBinaryFormula (ptinsth vform, boolean * becomesvari)
{
  return ptinsth_remove_binaryformula (vform, becomesvari);
}

/** It specifies that the clocks (vh1,vh2) are complementary in their father clock. */
void
ClockFormula::setComplementary (ptinsth vh1, ptinsth vh2)
{
  ptinsth_set_compl (vh1, vh2);
  ptinsth_set_compl (vh2, vh1);
}

/**
   It sets the operator (op) of a formula (clk).
*/
void
ClockFormula::setOperator (ptinsth clk, pkClockType op)
{
  ptinsth_set_operator (clk, op);
}

/** Clock.
   @return true if aClock is the predefined undefined clock */
boolean ClockFormula::satisfiesUndefClock (ptinsth aClock)
{
  return ptinsth_is_hundef (aClock);
}

/** Clock.
   @return the predefined undefined clock.
*/
ptinsth ClockFormula::getUndefClock ()
{
  return ptinsth_get_hundef ();
}

/** Clock.
   @return true if aClock is the predefined null clock.
*/
boolean ClockFormula::satisfiesNullClock (ptinsth aClock)
{
  return ptinsth_is_hnull (aClock);
}

/** Clock.
   @return the predefined null clock.
*/
ptinsth ClockFormula::getNullClock ()
{
  return ptinsth_get_hnull ();
}

/** Clock.
   @return true if aClock is the predefined init clock.
*/
boolean ClockFormula::satisfiesInitClock (ptinsth aClock)
{
  return ptinsth_is_hinit (aClock);
}

/** Clock.
   @return the predefined init clock.
*/
ptinsth ClockFormula::getInitClock ()
{
  return ptinsth_get_hinit ();
}

/**
   The formula (vh) looses the status of representative element.
   If such a formula is a free variable and its class is not reduced to one element, it is removed from this class.
*/
void
ClockFormula::belongsNotclhead (ptinsth vh)
{
  ptinsth_belongs_notclhead (vh);
}

/** The operand formula (vh) becomes the representative formula of the class when ishead is "true"
    otherwise the formula (vh) is not the representative one.
*/
void
ClockFormula::setIsRepresentativeFormula (ptinsth vh, boolean ishead)
{
  ptinsth_set_clhead (vh, ishead);
}

void
ClockFormula::moveReferences (ptinsth vh, ptinsth vform)
{
  ptinsth_move_references (vh, vform);
}

void
ClockFormula::removeSignalFromClass (ptinsth vh, ptexpcal sig)
{
  ptinsth vh1;

  ptinsth_remove_sig (vh, sig);
  vh1 = getNextFormula (vh, pk_true);
  while (vh1 != vh)
    {
      ptinsth_remove_sig (vh1, sig);
      vh1 = getNextFormula (vh1, pk_true);
    }
}

/**
   @return "true" is the formula(h1,h2) are in the same class (i.e. same clock), "false" otherwise.
*/
boolean ClockFormula::eqClock (ptinsth h1, ptinsth h2)
{
  return (getRepresentativeFormula (h1) == getRepresentativeFormula (h2));
}

/**
   @return "true" if a clock (vh) is unified with a signal (aSig), "false" otherwise (aSig is not modified).
   
*/
boolean ClockFormula::isClockUnifiedWithSignal (ptinsth vh, ptexpcal * aSig)
{
  return getAttribute (vh, CLKSIGUNIFICATION, (attributeValue *) aSig);
}

/**
   It sets that a clock (vh) is unified with a signal (aSig).
   if aSig is undefined, the attribute is removed.
*/
void
ClockFormula::setClockUnifiedWithSignal (ptinsth vh, ptexpcal aSig)
{
  if (aSig == NILEXPCAL)
    removeAttribute (vh, CLKSIGUNIFICATION);
  else
    addAttribute (vh, CLKSIGUNIFICATION, (attributeValue) aSig);
}

/**
   It sets that a clock (vh) is represented with a signal (aSig).
   if aSig is undefined, the attribute is removed.
*/
void
ClockFormula::setClockRepresentedBySignal (ptinsth vh, ptexpcal aSig)
{
  if (aSig == NILEXPCAL)
    removeAttribute (vh, CLK_REPRESENTED_BY);
  else
    addAttribute (vh, CLK_REPRESENTED_BY, (attributeValue) aSig);
}

/**
   @return "true" if a clock (vh) is represented by a signal (aSig), "false" otherwise (aSig is not modified).
*/
boolean ClockFormula::isClockRepresentedBySignal (ptinsth vh, ptexpcal * aSig)
{
  return getAttribute (vh, CLK_REPRESENTED_BY, (attributeValue *) aSig);
}

/**
   It sets that a clock (vh) may be replaced by its definition.
*/
void
ClockFormula::setSubstitutable (ptinsth vh)
{
  addMark (vh, SUBSTITUABLE_CLOCK);
}

/**
   @return "true" if a clock (vh) may be replaced by its defnition, "false" otherwise.
*/
boolean ClockFormula::isSubstitutable (ptinsth vh)
{
  return satisfiesMarked (vh, SUBSTITUABLE_CLOCK);
}

void
ClockFormula::addRDVClocks (ptinsth vh1, ptinsth vh2)
{
  addAttribute (vh1, POLYENDO_RDV, (attributeValue) vh2);
  addAttribute (vh2, POLYENDO_RDV, (attributeValue) vh1);
}

boolean
ClockFormula::isWaitingClock (ptinsth vh1, ptinsth *vh2)
{
  // Added for POLYENDO  LBesnard
  return getAttribute (vh1, POLYENDO_RDV, (attributeValue *) vh2); // il faudra une liste...
}


