/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_ ClockHierarchy::ClockHierarchy ()
{
  INIT_CLOCKS ();
}

/** Clock hierarchy.
    It initializes the hierarchy. clkroot is the predefined "fictitious" root of a system.
*/
void
ClockHierarchy::initHierarchy (ptinsth clkroot)
{
  hundef = ptinsth_get_hundef ();
  htick = hundef;
  froot = clkroot;
  setFather (froot, froot);
}

/**
   Clock Hierarchy.
   @return the "tick" of the program.
   For a endochronous program "tick" is the synthetized master clock,
   otherwise it is the fictitious root (predefined) of the system.
*/
ptinsth ClockHierarchy::getTickClock ()
{
  return ptinsth_get_clhead (htick);
}

/** Clock Hierarchy.
    Set the tick clock of the program.
*/
void
ClockHierarchy::setTickClock (ptinsth h)
{
  htick = h;
}

/** Clock Hierarchy.
    @return true if h is the tick clock of the program, false otherwise.
*/
boolean ClockHierarchy::satisfiesTickClock (ptinsth h)
{
  h = ptinsth_get_clhead (h);
  if (htick == hundef)
    return satisfiesFictitiousRoot (h);
  return (h == htick);
}

/** Clock hierarchy.
    @return true if h is the fictitious root of a system, false otherwise.
*/
boolean ClockHierarchy::satisfiesFictitiousRoot (ptinsth h)
{
  return (h == froot);
}

/** Clock hierarchy.
    @return the predefined "fictitious" root of a system.
*/
ptinsth ClockHierarchy::getFictitiousRoot ()
{
  return froot;
}

/** 
    @return "true" if a clock (pt) is a fictitious root one
    (ie its children are the real roots of the hierarchy), "false" otherwise.
*/
boolean ClockHierarchy::isFictitiousRootFormula (ptinsth pt)
{
  return ptinsth_isFictitiousRootFormula (pt);
}

/** Clock Hierarchy.
    @return the root of a clock (vh) and its depth (depth)
 */
ptinsth ClockHierarchy::getRoot (ptinsth vh, int *depth)
{
  return ptinsth_get_root (vh, depth);
}

/** Clock Hierarchy.
    It assigns the root clock (vhr) to a clock (vh).
 */
void
ClockHierarchy::setRoot (ptinsth vh, ptinsth vhr)
{
  ptinsth_set_rootClock (vh, vhr);
}

/** Clock Hierarchy.
    @return the set of children of a clock (vh).
 */
ptlinsth ClockHierarchy::getChildren (ptinsth vh)
{
  return ptinsth_get_lchild (vh);
}

/** Clock Hierarchy.
    @return the set of children of a clock (vh).
 */
void
ClockHierarchy::setChildren (ptinsth vh, ptlinsth lh)
{
  ptinsth_set_lchild (vh, lh);
}

/** Clock Hierarchy.
    @return true if the clock vh is a root of the hiecharchy, false otherwise.
*/
boolean ClockHierarchy::satisfiesRoot (ptinsth vh)
{
  return ptinsth_satisfies_root (vh);
}

/** Clock Hierarchy.
    It adds a clock (child) as a child of another clock (vh) in the hierarchy.
 */
void
ClockHierarchy::addChild (ptinsth vh, ptinsth child)
{
  ptinsth_insert_child (vh, child);
}

/** Clock Hierarchy.
    It removes a clock (child) of the children of another clock (vh) in the hierarchy.
 */
void
ClockHierarchy::removeChild (ptinsth vh, ptinsth child)
{
  ptinsth_remove_child (vh, child);
}

/** Clock  Hierachy.
    @return the depth of a clock in the hierarchy.
 */
int
ClockHierarchy::getDepth (ptinsth vh)
{
  return ptinsth_get_depth (vh);
}

/** Clock  Hierachy.
    Set the depth of a clock in the hierarchy.
 */
void
ClockHierarchy::setDepth (ptinsth vh, int d)
{
  ptinsth_set_depth (vh, d);
}

/** Clock  Hierachy.
    @return true if the depth of a clock is null, false otherwise.
 */
boolean ClockHierarchy::satisfiesNullDepth (ptinsth vh)
{
  return ptinsth_satisfiesNullClockDepth (vh);
}

/** Clock  Hierachy.
    @return the clock father of a clock (vh). 
    For a root, vh = getFather(vh).
*/
ptinsth ClockHierarchy::getFather (ptinsth vh)
{
  return ptinsth_get_father (vh);
}

/** Clock  Hierachy.
    It assigns the father clock (vhfather) to a clock (vh).
*/
void
ClockHierarchy::setFather (ptinsth vh, ptinsth vhfather)
{
  return ptinsth_set_father (vh, vhfather);
}

/** Hierachy.
    @return the complementary clock of a clock (vh) in the hierarchy if is available, or undefined otherwise.
    This information is available for placed clock and for extractions.
 */
boolean ClockHierarchy::getComplInFather (ptinsth vh, ptinsth * r)
{
  if (ptinsth_get_complInFather (vh, r))
    return pk_true;
  return pk_false;
}

/**
   @return the root and the depth (vdepth) of the clock vh. 
   When the root if the fictitious root, the depth is re-computed.
*/
ptinsth ClockHierarchy::getRootAndDepth (ptinsth vh, int *vdepth)
{
  ptinsth
    vhroot,
    vfather;

  vhroot = ptinsth_get_root (vh, vdepth);
  if (isFictitiousRootFormula (vhroot))
    {
      *vdepth = 0;
      vfather = ptinsth_get_father (vh);
      while ((vfather != vhroot))
	{
	  *vdepth = (*vdepth) + 1;
	  vfather = ptinsth_get_father (vfather);
	}
    }
  return vhroot;
}

/**
   It updates the depth of the subtree with vh as root. The depth of vh is vdepth.
   It sets the root of the clocks of the subtree to vhroot.
*/
void
ClockHierarchy::updateDepthSubTree (ptinsth vh, ptinsth vhroot, int vdepth)
{
  ptinsth_updateDepthSubTree (vh, vhroot, vdepth);
}

/** @return the list of the clocks on the branch from a clock (vh) until another one (vhemb).
    Assumed : vhemb and vh are in the same branch and vhemb is an ascendant of vh. 
*/
ptlinsth ClockHierarchy::getClocksUntil (ptinsth vh, ptinsth vhemb)
{
  return ptinsth_getClocksUntil (vh, vhemb);
}

/** 
    It places a clock (vh) in the clock hierarchy.
    The root of vh is vhrac, its father clock (vhfather) and its depth (vdepth).
*/
void
ClockHierarchy::placeInClockHierarchy (ptinsth vh, ptinsth vhrac, ptinsth vhfather, int vdepth)
{
  ptinsth_set_root (vh, vhrac, vdepth);
  ptinsth_set_father (vh, vhfather);
}

int
ClockHierarchy::getClockCounter ()
{
  return ptinsth_getClockCounter ();
}

/**
   @return 
    - pk_true  if vharg1 and vharg2 have the same root (but not the fictitious root)  and are on different branches. Then :
           - vhemb (out) : is the class of the smaller placed upper bound (branching);
	   - vharg1 (out): the nearest of vhemb;
	   - vharg2 (out): the other;
           - vhson1 (out): child of vhemb on the branch of vharg1
           - vhson2 (out): child of vhemb on the branch of vharg2
    - false if vharg1 and vharg2 have not the same root or the root is  the fictitious clock. Then,
           - vhemb (out) : is the undefined clock.
*/
boolean ClockHierarchy::getCompleteBranching (ptinsth * vharg1, ptinsth * vharg2, ptinsth * vhson1, ptinsth * vhson2, ptinsth * vhemb)
{
  boolean
    bres;
  ptinsth
    vh1,
    vh2;
  int
    vdepth1,
    vdepth2;

  vh1 = getRootAndDepth (*vharg1, &vdepth1);
  vh2 = getRootAndDepth (*vharg2, &vdepth2);

  if (vh1 != vh2)
    {
      *vhson1 = vh1;
      *vhson2 = vh2;
      *vhemb = hundef;
      return pk_false;
    }
  if (vdepth1 == vdepth2)
    {
      vh1 = *vharg1;
      vh2 = *vharg2;
      vdepth2 = 0;
    }
  else
    {
      if (vdepth1 > vdepth2)
	{
	  vh1 = *vharg2;
	  vh2 = *vharg1;
	  vdepth2 = vdepth1 - vdepth2;
	  *vharg1 = vh1;
	  *vharg2 = vh2;
	}
      else
	{
	  vh1 = *vharg1;
	  vh2 = *vharg2;
	  vdepth2 = vdepth2 - vdepth1;
	}
      vh2 = ptinsth_get_nieme_aieul (vh2, vdepth2);
    }

  if (vh1 == vh2)
    {
      *vhson1 = ptinsth_get_nieme_aieul (*vharg2, vdepth2 - 1);
      bres = pk_false;
    }
  else
    {
      vdepth1 = 0;
      while (vh1 != vh2)
	{
	  *vhson1 = vh1;
	  *vhson2 = vh2;
	  vh1 = getFather (vh1);
	  vdepth1 = vdepth1 + 1;
	  vh2 = getFather (vh2);
	  vdepth2 = vdepth2 + 1;
	}
      bres = pk_true;
    }
  *vhemb = vh1;
  if (isFictitiousRootFormula (vh1))
    {
      *vhemb = hundef;
      bres = pk_false;
    }
  return bres;
}

/** @return
     - the branching of two clocks (vharg1 and vharg2), if these clocks are in the same clock tree, 
     - the undefined clock otherwise. 
*/
ptinsth ClockHierarchy::getBranching (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth vhson1, vhson2, vhemb;

  vharg1 = ptinsth_get_clhead (vharg1);
  vharg2 = ptinsth_get_clhead (vharg2);
  if (vharg1 == hundef) return vharg2;
  if (vharg2 == hundef) return vharg1;
  getCompleteBranching (&vharg1, &vharg2, &vhson1, &vhson2, &vhemb);
  return vhemb;
}

/** Formula.
   @return "true" it the formula ?vh is the canonical form of its class.
   (ie. It is the representative formula and placed in the hierarchy) */
boolean ClockHierarchy::satisfiesIsCanonicalRepr (ptinsth vh)
{
  return ptinsth_satisfies_CanonicalRepr (vh);
}

void
ClockHierarchy::cloneClocks (ptinsth vh1, ptinsth vh2)
{
  if (getDepth (vh2) != 0)
    CH2BDD_clone (vh1, vh2, getFather (vh2));
}
