/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   @return the number of created formulas.
*/
  int
ClockSystem::getNumberOfCreatedFormula ()
{
  return getClockCounter ();
}

/**
   Getting number of free variables.
*/
int
ClockSystem::getNbFreeVariables ()
{
  ptinsth vh;
  int nb;
  ptlinsth list;

  nb = 0;
  list = getlhroots ();
  while (ptlinsth_get_next (&list, &vh))
    {
      if (vh == getRepresentativeFormula (vh))
	if (!ptinsth_satisfies_predef (vh))
	  if (!satisfiesSub (vh))
	    nb = nb + 1;
    };

  return nb;
}

/** @return the number of trees of the hierachy */
int
ClockSystem::getNbRoots ()
{
  /* ---added Nov 2017 */
  ptlinsth lchild = getChildren(getFictitiousRoot ());
  int n = ptlinsth_get_size(lchild);
  if (n != 0) return n;
  /* ---end added */
  return ptlinsth_get_size (getlhroots ());
}

/** Getting set of formula.
   @param lop (in) formula-operator code.
   @return the list of formula according to lop.
*/
ptlinsth
ClockSystem::getFormulaList (pkClockType lop)
{
  /* Il faudrait gerer les minus */
  if ((lop == sub) || (lop == nosub))
    return lssub;
  if (lop == infer)
    return lsinfer;
  if (lop == super)
    return lssuper;
  if (lop == vari)
    {				/* Suppression des variables inutiles et non supprimees lors des reecritures */
      clean_variableClockList ();
      return lvariables;
    }
  if (lop == hnot)
    return lsnot;
  return NILLINSTH;
}

/** Getting constraints.
   @param lhorl1 (out) list of placed formula (f1,...fn).
   @param lhorl2 (out) list of placed formula (l1,...ln).
   @return true if it exists some constraints on placed formula(fi is equivalent to li, but not proved).
*/
boolean
ClockSystem::getConstraintsOnPlacedClocks (ptlinsth * lhorl1, ptlinsth * lhorl2)
{
  ptlinsth lh1, lh2, lh1save, lh2save;
  ptinsth vh1, vh2;

  lh1 = lterm1equpl;
  lh2 = lterm2equpl;
  lh1save = lh1;
  lh2save = lh2;
  lterm1equpl = NILLINSTH;
  lterm2equpl = NILLINSTH;
  while (ptlinsth_get_next (&lh1, &vh1))
    {
      ptlinsth_get_next (&lh2, &vh2);
      addConstraintPlacedClocks (vh1, vh2);
    }
  *lhorl1 = lterm1equpl;
  *lhorl2 = lterm2equpl;
  ptlinsth_free (lh1save);
  ptlinsth_free (lh2save);  
  return (lterm1equpl != NILLINSTH);
}

/**
   @return the list of placed clocks equivalent to vh.
 */
boolean ClockSystem::getClocksPlacedEquivTo(ptinsth vh, ptlinsth *lh)
{
  ptlinsth lh1, lh2;
  ptinsth vh1, vh2;
  boolean bres, badded;

  bres = pk_false;
  lh1 = lterm1equpl;
  lh2 = lterm2equpl;
  while (ptlinsth_get_next (&lh1, &vh1))
    {
      ptlinsth_get_next (&lh2, &vh2);
      if (vh1 == vh)
	{
	  *lh = ptlinsth_sorted_insert_elem(*lh, vh2, &badded);
	  bres = badded;
	}
      else 
	if (vh2 == vh) 
	  {
	    *lh = ptlinsth_sorted_insert_elem(*lh, vh1, &badded);
	    bres = badded;
	  }
    }
  return bres;
}


/**
   @return the list of placed clocks equivalent to vh.
 */
ptlinsth ClockSystem::getClocksPlacedEquivTo(ptinsth vh)
{
  ptlinsth lhaux, lh;
  ptinsth velem;
  boolean bcont;

  lh = NILLINSTH;
  bcont = getClocksPlacedEquivTo(vh, &lh);
  // and transitively
  while (bcont)
    {
      bcont = pk_false;
      lhaux = lh;
      while (ptlinsth_get_next (&lhaux, &velem))
	{
	  if (getClocksPlacedEquivTo(velem, &lh))
	    bcont = pk_true;
	}
    }
  lh = ptlinsth_remove_elem (lh, vh, &bcont);
  return lh;
}


/** Getting assertions.
   @param lhorl1 (out) list of placed formula (f_1,...f_n).
   @param lhorl2 (out) list of placed formula (l_1,...l_n).
   @return true if it exists some assertions on placed formula (f_i assumed equivalent to l_i).
*/
boolean
ClockSystem::getAssertionsOnPlacedClocks (ptlinsth * lhorl1, ptlinsth * lhorl2)
{
  ptlinsth lh1, lh2;
  ptinsth vh1, vh2;

  lh1 = lterm1assertpl;
  lh2 = lterm2assertpl;
  lterm1assertpl = NILLINSTH;
  lterm2assertpl = NILLINSTH;
  while (ptlinsth_get_next (&lh1, &vh1))
    {
      ptlinsth_get_next (&lh2, &vh2);
      vh1 = getRepresentativeFormula (vh1);
      vh2 = getRepresentativeFormula (vh2);
      hcladdassertion (vh1, vh2);
    }
  *lhorl1 = lterm1assertpl;
  *lhorl2 = lterm2assertpl;
  return (lterm1assertpl != NILLINSTH);
}

/**
   @return the list of the placed clocks constrainted to be the null clock.
   The references to this clock must be visited.
*/
ptlinsth
ClockSystem::getPlacedNullClocks ()
{
  ptlinsth lh, lh1;
  ptinsth vh1;

  lh = NILLINSTH;
  lh1 = lsnullpl;
  while (ptlinsth_get_next (&lh1, &vh1))
    if (!satisfiesIsAnAssumption (vh1))
      lh = ptlinsth_insert_elem (lh, vh1);
  return lh;
}

/**
   @return the list of the placed clocks assumed equivalent to the null clock.
   The references to this clock must be visited.
*/
ptlinsth
ClockSystem::getPlacedAssertionNullClocks ()
{
  ptlinsth lh, lh1;
  ptinsth vh1;

  lh = NILLINSTH;
  lh1 = lsnullpl;
  while (ptlinsth_get_next (&lh1, &vh1))
    if (satisfiesIsAnAssumption (vh1))
      lh = ptlinsth_insert_elem (lh, vh1);
  return lh;
}

/**
   @return true if the system is constrainted (according to the solving method), false otherwise.
*/
boolean
ClockSystem::isConstrainedSystem ()
{
  ptlinsth lcour;
  ptinsth vh;

  if (existsynchropl ())
    return pk_true;
  lcour = lsplaced;
  while (ptlinsth_get_next (&lcour, &vh))
    if (!satisfiesSingleton (vh))
      return pk_true;

  lcour = lsroots;
  while (ptlinsth_get_next (&lcour, &vh))
    {
      vh = getRepresentativeFormula (vh);
      if (!ptinsth_is_hundef (vh))
	if (!satisfiesSingleton (vh))
	  return pk_true;
    }
  return pk_false;
}

/**
   @return "true" if the current system is composed of some assertions, "false" otherwise.
*/
boolean
ClockSystem::isSystemWithAssertions ()
{
  ptlinsth lh1, lh2;
  ptinsth vh;

  lh1 = lsnullpl;
  while (ptlinsth_get_next (&lh1, &vh))
    if (satisfiesIsAnAssumption (vh))
      return pk_true;

  if (getAssertionsOnPlacedClocks (&lh1, &lh2))
    return pk_true;
  lh1 = getlhroots ();
  while (ptlinsth_get_next (&lh1, &vh))
    {
      vh = getRepresentativeFormula (vh);
      if (getNextAssertion (vh, &vh))
	return pk_true;
    }
  lh1 = getlhplaced ();
  while (ptlinsth_get_next (&lh1, &vh))
    {
      vh = getRepresentativeFormula (vh);
      if (getNextAssertion (vh, &vh))
	return pk_true;
    }
  return pk_false;
}

/** @return "true" if it exists some formula (lConstraintsNullClock) not proved to the null clock, "false" otherwise.
 */
boolean
ClockSystem::getNullClockConstraints (ptlinsth * lConstraintsNullClock)
{
  ptinsth vh1;
  boolean b;
  ptlinsth listSub, listInf, listSup, lh;

  *lConstraintsNullClock = NILLINSTH;

  listSub = getFormulaList (sub);
  listInf = getFormulaList (infer);
  listSup = getFormulaList (super);

  vh1 = getNextFormula (the_nullClock_CKS_copy, pk_true);
  while (vh1 != the_nullClock_CKS_copy)
    {
      if (!satisfiesIsAnAssumption (vh1))
	{
	  b = pk_false;
	  switch ((int) (getOperator (vh1)))
	    {
	    case nosub:
	    case sub:
	      b = ptlinsth_isin_elem (listSub, vh1);
	      break;
	    case infer:
	      b = ptlinsth_isin_elem (listInf, vh1);
	      break;
	    case super:
	      b = ptlinsth_isin_elem (listSup, vh1);
	      break;
	    default:;
	    }
	  if (b) *lConstraintsNullClock = ptlinsth_insert_elem (*lConstraintsNullClock, vh1);
	}
      
      vh1 = getNextFormula (vh1, pk_true);
    }
  
  lh = getPlacedNullClocks ();	// Placed null clocks (constraint: not proved equal to).
  while (ptlinsth_get_next (&lh, &vh1))
    if (!satisfiesIsAnAssumption (vh1))
      *lConstraintsNullClock = ptlinsth_insert_elem (*lConstraintsNullClock, vh1);

  return ((*lConstraintsNullClock) != NILLINSTH);
}

/** @return "true" if it exists some assertions (lasserts) on the null clock. */
boolean
ClockSystem::getNullClockAssertions (ptlinsth * lasserts)
{
  ptinsth vh1;
  boolean b;
  ptlinsth listSub, listInf, listSup, lh;

  *lasserts = NILLINSTH;

  listSub = getFormulaList (sub);
  listInf = getFormulaList (infer);
  listSup = getFormulaList (super);

  vh1 = the_nullClock_CKS_copy;
  while (getNextAssertion (vh1, &vh1))
    if (satisfiesIsAnAssumption (vh1))
      {
	b = pk_false;
	switch ((int) (getOperator (vh1)))
	  {
	  case nosub:
	  case sub:
	    b = ptlinsth_isin_elem (listSub, vh1);
	    break;
	  case infer:
	    b = ptlinsth_isin_elem (listInf, vh1);
	    break;
	  case super:
	    b = ptlinsth_isin_elem (listSup, vh1);
	    break;
	  default:;
	  }
	if (b) *lasserts = ptlinsth_insert_elem (*lasserts, vh1);
      }

  lh = lsnullpl; // Placed null clocks: selecting the assertions.
  while (ptlinsth_get_next (&lh, &vh1))
    if (satisfiesIsAnAssumption (vh1)) *lasserts = ptlinsth_insert_elem (*lasserts, vh1);

  return ((*lasserts) != NILLINSTH);
}

/**
   @return the list of the clocks of the system.
   The user may freed the returned list.
*/
ptlinsth
ClockSystem::getClocks ()
{
  ptlinsth lh, lh1;
  ptinsth vh;
  boolean bdummy;

  lh1 = getlhroots ();
  lh = NILLINSTH;
  while (ptlinsth_get_next (&lh1, &vh))
    {
      vh = getRepresentativeFormula (vh);
      lh = ptlinsth_sorted_insert_elem (lh, vh, &bdummy);
    }
  lh1 = getlhplaced ();
  while (ptlinsth_get_next (&lh1, &vh))
    {
      vh = getRepresentativeFormula (vh);
      lh = ptlinsth_sorted_insert_elem (lh, vh, &bdummy);
    }
  return lh;
}

/**
   @return the list of clocks built on IO clocks.
   @param lOthers (out) the list of the clocks npot built on IO clocks.
   @param ls (out) the list of signals (ls) that define the projected system.
   A outside algorithm (out of the clock system) must be used to produce these informations.
   \see setProjectedSystem 
*/
ptlinsth
ClockSystem::getProjectedSystem (ptlinsth * lOthers, ptlinstances * ls)
{
  *lOthers = ExportedClocksOthers;
  *ls = ExportedSignalsInSpec;
  return ExportedClocksIOClocks;
}

/**
   @return the list of the root clocks of the system.
*/
ptlinsth
ClockSystem::getlhroots ()
{
  return lsroots;
}

/**
   @return the list of the placed clocks of the system.
*/
ptlinsth
ClockSystem::getlhplaced ()
{
  return lsplaced;
}

/**
    @return "true" if a formula (vh) is a recursive expression by transitive syntatic verification, "false" otherwise.
*/
boolean
ClockSystem::isRecursiveFormula (ptinsth vh)
{
  int p;
  ptinsth vh1;
  boolean b;

  vh1 = getRoot (vh, &p);
  if (!satisfiesFictitiousRoot (vh1))
    vh = vh1;
  b = ishodefrec (vh, vh);
  resetMarkAll (mark3);
  return b;
}

/**
   @return true if the clock system is empty.
*/
boolean
ClockSystem::isAnEmptySystem ()
{
  return ((getlhroots () == NILLINSTH) && (getlhplaced () == NILLINSTH));
}

/**
   @return "true" if a formula (vh) depends on the value of a signal (s), "false" otherwise.
*/
boolean
ClockSystem::__isaFormulaDependOnValueOf (ptinsth vh, ptexpcal s)
{
  ptinsth vh1, vh2;
  boolean b;
  ptlinstances ls;

  if (satisfiesMarked (vh, mark3))
    return pk_false;
  addMark (vh, mark3);

  switch ((int) (getOperator (vh)))
    {
    case super:
      getSupArgs (vh, &vh1, &vh2);
      if (__isaFormulaDependOnValueOf (vh1, s))
	b = pk_true;
      else
	b = __isaFormulaDependOnValueOf (vh2, s);

      break;
    case infer:
      getInfArgs (vh, &vh1, &vh2);
      if (__isaFormulaDependOnValueOf (vh1, s))
	b = pk_true;
      else
	b = __isaFormulaDependOnValueOf (vh2, s);
      break;
    case sub:
    case nosub:
      ls = getSignalsOfCondition (vh);
      if (ptlinstances_isin_exp (ls, s))
	b = pk_true;		// else xxxxxx(); // Il faut aller voir les noeuds de ces signaux.

      break;

    case hcomp:
    case hnot:
      if (getUnaryArg (vh, &vh1))
	b = __isaFormulaDependOnValueOf (vh1, s);
      break;
    default:
      b = pk_false;
    }

  removeMark (vh, mark3);
  return b;
}



