/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
  @return a new affine relation (period_n, phasis, period_d)
*/
extern ptAffineRelation
mkAffineRelation (long int period_n, long int phasis, long int period_d)
{
  ptAffineRelation aux;

  aux = new AffineRelation ();
  aux->period_n = (AffineParam) period_n;
  aux->phasis = (AffineParam) phasis;
  aux->period_d = (AffineParam) period_d;
  return aux;
}

/**
   @return a copy of an affine relation (pt).
*/
extern ptAffineRelation
ptAffineRelation_duplicate (ptAffineRelation pt)
{
  AffineParam period_n, phasis, period_d;

  ptAffineRelation_get_elems (pt, &period_n, &phasis, &period_d);
  return mkAffineRelation (period_n, phasis, period_d);
}

/**
   Copies an affine relation (pt1) into another one (pt2).
*/
extern void
ptAffineRelation_copy (ptAffineRelation pt1, ptAffineRelation pt2)
{
  pt2->period_n = pt1->period_n;
  pt2->phasis = pt1->phasis;
  pt2->period_d = pt1->period_d;
}

/**
   Getting the arguments (period_n, phasis, period_d) of an affine relation (pt).
*/
extern void
ptAffineRelation_get_elems (ptAffineRelation pt, AffineParam * period_n, AffineParam * phasis, AffineParam * period_d)
{
  *period_n = pt->period_n;
  *phasis = pt->phasis;
  *period_d = pt->period_d;
}

/**
   Setting the arguments (period_n, phasis, period_d) of an affine relation (pt).
*/
extern void
ptAffineRelation_set_elems (ptAffineRelation pt, AffineParam period_n, AffineParam phasis, AffineParam period_d)
{
  pt->period_n = period_n;
  pt->phasis = phasis;
  pt->period_d = period_d;
}

/**
  @return 
  - "true" if the parameters of affine relations (pt1, pt2) are equal: for pt1 = (n1, phi1, d1)
  and pt2 = (n2, phi2, d2), n1 = n2, phi1 = phi2, d1 = d2.
  - "false" otherwise.
*/
extern boolean
ptAffineRelation_compare (ptAffineRelation pt1, ptAffineRelation pt2)
{
  AffineParam period_n1, phasis1, period_d1, period_n2, phasis2, period_d2;

  ptAffineRelation_get_elems (pt1, &period_n1, &phasis1, &period_d1);
  ptAffineRelation_get_elems (pt2, &period_n2, &phasis2, &period_d2);
  return ((period_n1 == period_n2) && (phasis1 == phasis2) && (period_d1 == period_d2));
}

/**
   @return "true" if an affine relation pt=(period1, phasis, period2) satisfies period1= 1, phasis= 0 and period2= 1, "false" otherwise.
*/
extern boolean
ptAffineRelation_isOneZeroOne (ptAffineRelation pt)
{
  AffineParam period_n, phasis, period_d;

  ptAffineRelation_get_elems (pt, &period_n, &phasis, &period_d);
  return ((period_n == 1) && (phasis == 0) && (period_d == 1));
}

/**
   It builts the canonical form of an affine relation (pt).
   (see AffineParam_gcd_quotient(), AffineParam_div_and_mod()).
   For pt=(n, phi, d):
   - AffineParam_gcd_quotient() produces v_gcd = gcd(n,d), n1= n/v_gcd and d1= d/v_gcd.
   - AffineParam_div_and_mod() produces q, r such that phi = q* v_gcd + r, applied on phi, gcd
   - then the canocical form is: (n1, q, d1) if r=0, (2*n1, 2*q+1, 2*d1) otherwise
*/
extern void
ptAffineRelation_canonical_form (ptAffineRelation pt)
{
  AffineParam gcd;
  AffineParam period_n, phasis, period_d, period_ng, period_dg, phasis_qg, phasis_rg;

  ptAffineRelation_get_elems (pt, &period_n, &phasis, &period_d);
  gcd = AffineParam_gcd_quotient (period_n, period_d, &period_ng, &period_dg);
  AffineParam_div_and_mod (phasis, gcd, &phasis_qg, &phasis_rg);

  if (phasis_rg == 0)
    ptAffineRelation_set_elems (pt, period_ng, phasis_qg, period_dg);
  else
    ptAffineRelation_set_elems (pt, 2 * period_ng, 2 * phasis_qg + 1, 2 * period_dg);
  /*as gcd > 0 always, phasis_qg is equal to the integer part of (phasis/gcd) */
  /*a procedure to compute the integer part of (a/b) for a, b general integers
     is nevertheless provided */
}

/**
   @return "true" if pt1=(n1, phi1, d1), and pt2=(n2, phi2, d2) are equivalent, "false" otherwise.
   - Particular cases:
    - n1=n2=1, phi1=phi2,d1=d2
    - d1=d2=1, phi1=phi2,n1=n2
   - Otherwise the canonical form of the affine relations (ptAffineRelation_canonical_form()) are built and compared (ptAffineRelation_compare()).
*/
extern boolean
ptAffineRelation_isEquivalent (ptAffineRelation pt1, ptAffineRelation pt2)
{
  AffineParam period_n1, phasis1, period_d1, period_n2, phasis2, period_d2;
  AffineRelation affr1, affr2;

  ptAffineRelation_get_elems (pt1, &period_n1, &phasis1, &period_d1);
  ptAffineRelation_get_elems (pt2, &period_n2, &phasis2, &period_d2);

  if ((period_n1 == 1) && (period_n2 == 1))
    return ((phasis1 == phasis2) && (period_d1 == period_d2));
  if ((period_d1 == 1) && (period_d2 == 1))
    return ((phasis1 == phasis2) && (period_n1 == period_n2));
  
  /* General case: using canconical form. */
  ptAffineRelation_set_elems (&affr1, period_n1, phasis1, period_d1);
  ptAffineRelation_set_elems (&affr2, period_n2, phasis2, period_d2);
  ptAffineRelation_canonical_form (&affr1);
  ptAffineRelation_canonical_form (&affr2);
  return ptAffineRelation_compare (&affr1, &affr2);
}

/**
   The affine relation ptin = (n, phi, d) is divided in ptout1 = (1, max(0, -phi), n) and ptout2 = (1, max(0, phi), d).
   \note Side effect on ptin (=ptout1) and a new object is created for ptout2.
*/
extern void
ptAffineRelation_divide (ptAffineRelation ptin, ptAffineRelation * ptout1, ptAffineRelation * ptout2)
{
  AffineParam phasis, phasisMinus, phasisPlus, period_n;

  *ptout1 = ptin;
  *ptout2 = ptAffineRelation_duplicate (ptin);
  /*computing: max(0, -phi) and max(0, phi) */
  phasis = ptAffineRelation_get_phasis (ptin);
  phasisMinus = AffineParam_max (0, -phasis);
  phasisPlus = AffineParam_max (0, phasis);

  period_n = ptAffineRelation_get_periodN (*ptout1);
  ptAffineRelation_set_periodN (*ptout1, 1);
  ptAffineRelation_set_periodD (*ptout1, period_n);
  ptAffineRelation_set_phasis (*ptout1, phasisMinus);

  ptAffineRelation_set_periodN (*ptout2, 1);
  ptAffineRelation_set_phasis (*ptout2, phasisPlus);
}

/**
   It produces (ptout) the (simple) composition of the relations ptin1 = (1, phi1, d1) and ptin2 = (1, phi2, d2).
   It is defined by ptout = (1, phi1 + d1*phi2, d1*d2).
*/
extern void
ptAffineRelation_simple_composition (ptAffineRelation ptin1, ptAffineRelation ptin2, ptAffineRelation ptout)
{
  AffineParam phasis1, phasis2, period_d1, period_d2;

  phasis1 = ptAffineRelation_get_phasis (ptin1);
  phasis2 = ptAffineRelation_get_phasis (ptin2);
  period_d1 = ptAffineRelation_get_periodD (ptin1);
  period_d2 = ptAffineRelation_get_periodD (ptin2);
  ptAffineRelation_set_elems (ptout, 1, phasis1 + phasis2 * period_d1, period_d1 * period_d2);
}

/**
   It produces (ptout) the composition of the relations ptin1 = (1, phi1, d1) and ptin2 = (1, phi2, d2).
   Definition: for ptin1 = (n1, phi1, d1) and ptin2 = (n2, phi2, d2), ptout = (n1*n2, phi1*n2 + d1*phi2, d1*d2).

   \note We can composed (only?) the affine relations 
   - (n, phi, d) with (1, phi', d'), result = (n, phi + d*phi', d*d') with phi integer, (phi' >= 0)
   - (1, phi, d) with (1, phi', d') is (1, phi + d phi', d d') phi, phi' >= 0.
   - (n, phi, 1) with (1, phi', d') is (n, phi + phi', d').
*/
extern void
ptAffineRelation_general_composition (ptAffineRelation ptin1, ptAffineRelation ptin2, ptAffineRelation ptout)
{
  AffineParam period_n1, period_n2, phasis1, phasis2, period_d1, period_d2;

  period_n1 = ptAffineRelation_get_periodN (ptin1);
  period_n2 = ptAffineRelation_get_periodN (ptin2);
  phasis1 = ptAffineRelation_get_phasis (ptin1);
  phasis2 = ptAffineRelation_get_phasis (ptin2);
  period_d1 = ptAffineRelation_get_periodD (ptin1);
  period_d2 = ptAffineRelation_get_periodD (ptin2);
  if ((period_n1 != 1) && (period_n2 != 1))
    /* cout << "severe error\ntry to compute (momentarily) not allowed affine relation composition" << endl; */
    ;
  else
    ptAffineRelation_set_elems (ptout, period_n1 * period_n2, phasis1 * period_n2 + phasis2 * period_d1, period_d1 * period_d2);
}

/**
  @return "true" if the complementary of ptin exists (ptout), "false" otherwise.

  For ptin = (1, phi, d), it exists if d = 2, phi=0 or phi=1, then ptout = (1, 1-phi, d).
*/
extern boolean
ptAffineRelation_complementary (ptAffineRelation ptin, ptAffineRelation ptout)
{
  AffineParam period_n, phasis, period_d;

  ptAffineRelation_get_elems (ptin, &period_n, &phasis, &period_d);
  if (period_d != 2)
    return pk_false;
  if (phasis == 0)
    {
      ptAffineRelation_set_elems (ptout, period_n, 1, 2);
      return pk_true;
    }
  if (phasis == 1)
    {
      ptAffineRelation_set_elems (ptout, period_n, 0, 2);
      return pk_true;
    }
  return pk_false;
}

/**
   It defines the inverse (ptout) of an affine relation (ptin). With ptin = (n, phi, d), ptout = (d, -phi, n).
*/
extern void
ptAffineRelation_inverse (ptAffineRelation ptin, ptAffineRelation ptout)
{
  ptout->period_n = ptin->period_d;
  ptout->phasis = -ptin->phasis;
  ptout->period_d = ptin->period_n;
}

/**
   @return "true" if  inverse(ptin1) composed with ptin2 is an affine relation whose 
   the UNION of the affine operand clocks is in affine relation with each operand, "false" otherwise.
   \see ptAffineRelation_sup_is_affine
*/
extern boolean
ptAffineRelation_comp_sup_is_affine (ptAffineRelation ptin1, ptAffineRelation ptin2, ptAffineRelation * ptout)
{

  AffineRelation affr12;
  AffineParam phasis12, phasisout;
  boolean result, inverse;

  // ptout is the affine relation (1, phi, d) entre l'embranchement commun des horloges operandes et l'horloge union.
  ptAffineRelation_inverse (ptin1, &affr12);
  ptAffineRelation_general_composition (&affr12, ptin2, &affr12);
  phasis12 = ptAffineRelation_get_phasis (&affr12);
  /* if (phasis12 <= 0) {ptAffineRelation_inverse(&affr12, &affraux); ptAffineRelation_copy(&affraux, &affr12); inverse = pk_true; } */
  result = ptAffineRelation_sup_is_affine (&affr12, ptout);
  if (result)
    {
      inverse = (phasis12 <= 0);
      if (inverse)
	phasisout = ptAffineRelation_get_phasis (ptin2);
      else
	phasisout = ptAffineRelation_get_phasis (ptin1);
      ptAffineRelation_set_phasis (*ptout, phasisout);
    }
  return result;
}

/**
   @return "true" if  inverse(ptin1) composed with ptin2 is an affine relation whose 
   the INTERSECTION of the affine operand clocks is in affine relation with each operand, "false" otherwise.
   Currently, it returns always "false".
   
   \see  ptAffineRelation_inf_is_affine
*/
extern boolean
ptAffineRelation_comp_inf_is_affine (ptAffineRelation ptin1, ptAffineRelation ptin2, ptAffineRelation * ptout)
{
  AffineRelation affr12;
  ptAffineRelation_inverse (ptin1, &affr12);
  ptAffineRelation_general_composition (&affr12, ptin2, &affr12);
  return ptAffineRelation_inf_is_affine (&affr12, ptout);
}
