/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
   @return the class head of the root of the clock-tree of a clock (vh)
   and the depth (vdepth) of this clock in the tree.
*/
extern ptinsth
ptinsth_get_root (ptinsth vh, int *vdepth)
{
  *vdepth = ptinsth_get_depth (vh);
  if (*vdepth > 0)
    return recroot (vh, *vdepth, vdepth);
  return vh;
}

/** @return the father clock of a clock (vh).
 */
extern ptinsth
ptinsth_get_father (ptinsth vh)
{
  ptinsth hf;

  vh = ptinsth_get_clhead (vh);
  hf = vh->father;
  if (hf == NILINSTH)
    hf = vh;
  else
    hf = ptinsth_get_clhead (hf);
  vh->father = hf;
  return hf;
}

/** @return the n-th ascendant of a clock (vh). */
extern ptinsth
ptinsth_get_nieme_aieul (ptinsth vh, int n)
{
  while (n > 0)
    {
      vh = ptinsth_get_father (vh);
      n--;
    }
  return vh;
}

/** It sets to vdepth value the depth of a clock (vh).
 */
extern void
ptinsth_set_depth (ptinsth vh, int vdepth)
{
  vh->vdepth = vdepth;
}

/** @return the depth of a clock (vh)
 */
extern int
ptinsth_get_depth (ptinsth vh)
{
  vh = ptinsth_get_clhead (vh);
  return vh->vdepth;
}

/** @return "true" if the depth of vh is equal to zero, "false" otherwise.
 */
extern boolean
ptinsth_satisfiesNullClockDepth (ptinsth vh)
{
  vh = ptinsth_get_clhead (vh);
  return (vh->vdepth == 0);
}

/** It sets to vhfather the father clock of a clock (vh).
 */
extern void
ptinsth_set_father (ptinsth vh, ptinsth vhfather)
{
  vh = ptinsth_get_clhead (vh);
  vh->father = ptinsth_get_clhead (vhfather);
}

/** It sets the root clock (vhrac) of a clock (vh) (ie vhrac is the root of the tree-clock where vh is a node).
 */
extern void
ptinsth_set_rootClock (ptinsth vh, ptinsth vhrac)
{
  vh->root = vhrac;
}

/** It sets the root clock (vhrac) and the depth (vdepth) of a clock (vh).
 */
extern void
ptinsth_set_root (ptinsth vh, ptinsth vhrac, int vdepth)
{
  vh = ptinsth_get_clhead (vh);
  ptinsth_set_rootClock (vh, vhrac);
  ptinsth_set_depth (vh, vdepth);
}

/** @return the list of the clocks children of a clock (vh).
 */
extern ptlinsth
ptinsth_get_lchild (ptinsth vh)
{
  vh = ptinsth_get_clhead (vh);
  return vh->children;
}

/** it sets to a clock (vh) its children (lh).
 */
extern void
ptinsth_set_lchild (ptinsth vh, ptlinsth lh)
{
  vh->children = lh;
}

/** It adds a clock (vhapp) in the list of the children of a clock (vh).
 */
extern void
ptinsth_insert_child (ptinsth vh, ptinsth vhapp)
{
  boolean dummy;

  vh->children = ptlinsth_sorted_insert_elem (vh->children, vhapp, &dummy);
}

/** It removes a clock (vh) from the list of the children of a clock (vhfather).
 */
extern void
ptinsth_remove_child (ptinsth vhfather, ptinsth vh)
{
  boolean b;
  vhfather->children = ptlinsth_remove_elem (vhfather->children, vh, &b);
}

/** @return "true" if a clock (vh) is a root clock.
 */
extern boolean
ptinsth_satisfies_root (ptinsth vh)
{
  ptinsth vh1;

  vh = ptinsth_get_clhead (vh);
  if (ptinsth_get_depth (vh) <= 0)
    return pk_true;
  vh1 = ptinsth_get_father (vh);
  if (vh1 == NILINSTH)
    return pk_true;
  if (vh1 == vh)
    return pk_true;
  return (ptinsth_isFictitiousRootFormula (vh1));
}

/** @return the list of the clocks on the branch from a clock (vh) until another one (vhemb).
    Assumed : vhemb and vh are in the same branch and vhemb is an ascendant of vh. 
*/
extern ptlinsth
ptinsth_getClocksUntil (ptinsth vh, ptinsth vhemb)
{
  ptlinsth lh;
  boolean binsert;

  if (ptinsth_satisfies_predef (vh))
    return NILLINSTH;
  if (ptinsth_isFictitiousRootFormula (vh))
    return NILLINSTH;
  lh = NILLINSTH;
  while (vh != vhemb)
    {
      lh = ptlinsth_sorted_insert_elem (lh, vh, &binsert);
      vh = ptinsth_get_father (vh);
    }
  return lh;
}

/** Hierarchy.
    @return "true" it a formula (vh) is a canonical formula (ie placed in a tree) , "false" otherwise.
*/
extern boolean
ptinsth_satisfies_CanonicalRepr (ptinsth vh)
{
  if (ptinsth_satisfies_clhead (vh))
    return !ptinsth_satisfiesNullClockDepth (vh);
  return pk_false;
}

/**
   Clock Hierarchy.
   @return "true" if the complementary clock (vhc) of a clock (clk) relatively to its father clock is known, "false" otherwise.
*/
extern boolean
ptinsth_get_complInFather (ptinsth clk, ptinsth * vhc)
{
  if (clk->hocompl != NILINSTH)
    {
      *vhc = clk->hocompl;
      return pk_true;
    }
  return ptinsth_get_complExtraction (clk, vhc);
}

/**
   It updates the depth of the subtree with vh as root. The depth of vh is vdepth.
   It sets the root of the clocks of the subtree to vhroot.
*/
extern void
ptinsth_updateDepthSubTree (ptinsth vh, ptinsth vhroot, int vdepth)
{
  ptlinsth lh;
  ptinsth vh1;

  lh = ptinsth_get_lchild (vh);
  vdepth++;
  while (ptlinsth_get_next (&lh, &vh1))
    {
      ptinsth_set_rootClock (vh1, vhroot);
      ptinsth_set_depth (vh1, vdepth);
      ptinsth_updateDepthSubTree (vh1, vhroot, vdepth);
    }
}
