/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   @return 
   - pk_true if the clock is not a placed one and if a formula (vh), not representative formula of its class, 
   is a syntactic formula equivalent to the representative element.
   - false otherwise (in particular when the class is a singleton, or the clock is place placed).
*/
  boolean ClockManager::isOtherSyntacticFormOfRepresentativeFormula (ptinsth vh)
{
  ptinsth vhcl;

  vhcl = getRepresentativeFormula (vh);
  if (vhcl == vh)
    return pk_false;
  if (!satisfiesNullDepth (vhcl))
    return pk_false;
  if (!isIncludedIn (vh, vhcl))
    return pk_false;
  if (!isIncludedIn (vhcl, vh))
    return pk_false;
  return pk_true;
}

/** Building elementary tree.
    @param vh1 (in) : an positive extraction formula ([C]),
    @param vh2 (in) : an negative extraction formula on the same condition than vh1 ([not C]),
    @param vh (in) the clock of C (assumed);
    @return "true" the elementary tree (vh1 and vh2 added as children of vh) is build,
    false when the clocks (in which vh1, or vh2 are elements) are already placed, or must be reduce to null clock.
*/
boolean
ClockManager::buildElementaryTreeFromSub (ptinsth vh, ptinsth vh1, ptinsth vh2)
{
  ptinsth vh1head, vh2head;
  boolean becomesvari;

  if (satisfiesNullClock (vh))
    {
      setClockToNullClock (vh1);
      setClockToNullClock (vh2);
      return pk_true;
    }
  vh1head = getRepresentativeFormula (vh1);
  vh2head = getRepresentativeFormula (vh2);
  if (vh1head == vh2head)
    {
      setClockToNullClock (vh);
      setClockToNullClock (vh1);
      setClockToNullClock (vh2);
      return pk_true;
    }

  if (clockEqAffine (vh1head, vh1))
    {
      removeFormulaFromClass (vh1head, vh1);
      removeSub (vh1, pk_true, &becomesvari);
      makeSynchro (vh1head, vh1);
      removeFormulaFromClass (vh2head, vh2);
      removelhroots (vh2);
      makeSynchro (vh2head, vh2);
      return pk_true;
    }

  if (satisfies_condintree (vh, vh1, vh2)) return pk_false;
  addClockAsRoot (vh);
  vh1 = addPlaced (vh1, vh);
  vh2 = addPlaced (vh2, vh);
  hcladdmodified (vh1);
  hcladdmodified (vh2);
  return pk_true;
}

/**
   @return 
   - pk_true if the formula (vharg1 ^* vharg2) is placed in a tree, \b otherterm contains the placed formula.
   - false if vharg1 and vharg2 are not placed in the same tree, \b otherterm is not available.
*/
boolean
ClockManager::treeReduceInf (ptinsth vharg1, ptinsth vharg2, ptinsth * otherterm)
{
  ptinsth vhson2, vhson1, vhemb, vh;

  if (!treeOperationAuthorized)
    return pk_false;

  /** Added Jan 2016 LBesnard */

  ptinsth  vhs2 = vharg2;
  if (infReduceNullClock (3, vharg1, &vhs2)) 
    {
      *otherterm = the_nullClock_CKM_copy;
       return pk_true;
    }
  /** END Added Jan 2016 LBesnard */

  if (getRealBranching (&vharg1, &vharg2, &vhson1, &vhson2, &vhemb))
    {
      *otherterm = inftreenormalize (vharg1, vharg2, vhemb);
      return pk_true;
    }
  if (satisfiesUndefClock (vhemb))
    {
      if (!isFormOntick (vharg1, vharg2, &vh))
	return pk_false;
      *otherterm = vh;
      return pk_true;
    }
  *otherterm = vharg2;
  return pk_true;
}

/**
   @return 
   - pk_true if the formula (vharg1 ^+ vharg2) is placed in a tree, \b otherterm contains the placed formula.
   - false if vharg1 and vharg2 are not placed in the same tree, \b otherterm is not available.
*/
boolean
ClockManager::treeReduceSup (ptinsth vharg1, ptinsth vharg2, ptinsth * otherterm)
{
  ptinsth vh, vhson1, vhson2, vhemb;

  if (!treeOperationAuthorized)
    return pk_false;

  if (getRealBranching (&vharg1, &vharg2, &vhson1, &vhson2, &vhemb))
    {
      *otherterm = suptreenormalize (vharg1, vharg2, vhemb);
      return pk_true;
    }
  if (satisfiesUndefClock (vhemb))
    {
      if (!isFormOntick (vharg1, vharg2, &vh))
	return pk_false;
      *otherterm = getTickClock ();
      return pk_true;
    }
  *otherterm = vhemb;
  return pk_true;
}

/**
   @return 
   - pk_true if the formula vfinf = vharginf ^- vharghnot is placed in a tree, \b otherterm contains the placed formula.
   - false if vharginf and vharghnot are not placed in the same tree, \b otherterm is not available.
*/
boolean
ClockManager::treeReduceInfhnot (ptinsth vfinf, ptinsth vharghnot, ptinsth vharginf, ptinsth * newterm)
{
  ptinsth vhson1, vhson2, vhemb, svharginf, svharghnot;

  if (treeOperationAuthorized)
    {
      svharghnot = vharghnot;
      svharginf = vharginf;
      if (getCompleteBranching (&vharghnot, &vharginf, &vhson1, &vhson2, &vhemb))
	{
	  *newterm = compltreenormalize (vfinf, svharghnot, svharginf, vhemb);
	  return (boolean) (!satisfiesUndefClock (*newterm));
	}
      if (satisfiesUndefClock (vhemb))
	return isFormComplOnTick (vharginf, vharghnot, newterm);
      *newterm = compltreenormalize (vfinf, svharghnot, svharginf, vhemb);
      return (boolean) (!satisfiesUndefClock (*newterm));
    }
  return pk_false;
}

/** 
    It finalizes the hierarchy by setting the access to the children of each clock (during the clock calculus, 
    trees are build using the "father" fonction).
    @return the clock tick (master clock) for endochonous system, "fictitious" root otherwise.
*/
ptinsth
ClockManager::finalizeClockHierarchy ()
{
  ptlinsth lcour;
  ptinsth vh, tick, vhrac, vhroot;
  int nb, vdepth;

  tick = getTickClock ();
  if (satisfiesUndefClock (tick) || isFictitiousRootFormula (tick))
    {
      setClocksAsRoots ();
      vhroot = getFictitiousRoot ();
      nb = 0;
      lcour = getlhroots ();
      while (ptlinsth_get_next (&lcour, &vh))
	if (!satisfiesPredefinedClock (vh))
	  if (vh == getRepresentativeFormula (vh))
	    {
	      if (satisfiesRoot (vh))
		{
		  placeInClockHierarchy (vh, vhroot, vhroot, 1);
		  insertlhplaced (vh);	// NEW NEW October 2013.
		  addChild (vhroot, vh);
		  {
		    nb = nb + 1;
		    tick = vh;
		  }
		}
	    }
	  else
	    removelhroots (vh);

      lcour = getlhplaced ();
      while (ptlinsth_get_next (&lcour, &vh))
	{
	  vhrac = getRoot (vh, &vdepth);	/* force le recalcul de la profondeur */
	  addChild (getFather (vh), vh);
	}

      if (nb != 1)
	tick = vhroot;
      setTickClock (tick);
    }
  return tick;
}

/** Clock  Hierachy.
    The tree assigned to a clock (vh) is placed under (as a child) the tree assigned to another clock (f).
    @return the placed clock.
*/
ptinsth
ClockManager::addPlaced (ptinsth vh, ptinsth f)
{
  boolean vnewhead;

  if (satisfiesVari (vh))
    if (!isFictitiousRootFormula (vh))
      vh = getNextFormula (vh, pk_true);

  vh = resetclass (vh, &vnewhead);
  if (satisfiesRoot (vh))
    placeInClockHierarchy (vh, vh, vh, 0);
  else if (vnewhead)
    placeInClockHierarchy (vh, vh, vh, 0);
  placeInClockHierarchy (vh, f, f, 1);
  insertlhplaced (vh);
  setNormalized (vh);
  return vh;
}

/** Clock  Hierachy.
    It calls ClockManager::addPlaced() and assocites to vh the same bdd than vo.
    vh is a clone of vo (other clock System, but they have the same associated bdd).
    @return the placed clock.
*/
ptinsth
ClockManager::addPlacedClone (ptinsth vh, ptinsth f, ptinsth vo)
{
  ptinsth vres;

  vres = addPlaced (vh, f);
  if (!CH2BDD_clone (vo, vh, f));	/* printf("Internal Error: CH2BDD_clone returns false value (clock name = %s)!\n", ident2string(getName(vo))) */
  return vres;
}

/** 
   It resets a clock (vh) as a root in the hierarchy and recursively 
   the clocks with a formula that references it.
   @return true if the clock is initialy placed (and it re-becomes a root), false otherwise.
*/
boolean
ClockManager::resetClockAsRoot (ptinsth vh)
{
  ptlinsth luse;
  ptinsth use, vhemb;

  if (vh == getRepresentativeFormula (vh))
    if (!satisfiesNullDepth (vh))
      {
	removelhplaced (vh, pk_true);
	insertlhroots (vh);
	removeChild (getFather (vh), vh);

	placeInClockHierarchy (vh, vh, vh, 0);
	updateDepthSubTree (vh, vh, 0);
	luse = getClockReferencesTo (vh);
	while (ptlinsth_get_next (&luse, &use))
	  {
	    if (getRepresentativeFormula (use) == use)
	      {
		vhemb = getBranching (getRepresentativeFormula (vh), getRepresentativeFormula (use));
		if (vhemb != vh)
		  if (resetClockAsRoot (use))
		    addFormulaToListRoot (use);
	      }
	  }
	return pk_true;
      }
  return pk_false;
}

/**
   It resets the clocks of a list (lh) as roots of the hierarchy.
   \see resetClockAsRoot
*/
void
ClockManager::resetClocksAsRoot (ptlinsth lh)
{
  ptinsth vh;

  while (ptlinsth_get_next (&lh, &vh))
    if (resetClockAsRoot (vh))
      addFormulaToListRoot (vh);
}

/** clocktreefront is initialized from systemSolving
    It initializes the BDD system solving: ClockHierarchy2BDD, BDDHierarchy, BDD2ClockHierarchy
    The null clock is assumed to be defined
   \see treeEndSolving
*/
void
ClockManager::clocktreefront_init ()
{
  setTreeOperationMode (pk_true);
  ClockHierarchy2BDD_init (this);
  initSystemForModification ();
}

/**
   It stops the tree placing mechanism.
   \see clocktreefront_init
*/
void
ClockManager::treeEndSolving (void)
{
  // print_all_bdd();
  setTreeOperationMode (pk_false);
}

/**
   It prepares the system for new modifations.
   It must be called when the hierarchy is build (\ref finalizeClockHierarchy  has been applied).
   Then the system can be completed and the \ref clockCalculus "clock calculus" re-applied.
   @return the list of the clocks, children, before the call, of the fictitious root.
 */
ptlinsth
ClockManager::initSystemForModification ()
{
  ptlinsth lh;
  setTreeOperationMode (pk_true);
  lh = getChildren (getFictitiousRoot ());
  resetClocksAsRoot (lh);
  if (!satisfiesUndefClock (getTickClock ()))
    setTickClock (getUndefClock ());
  return lh;
}

void
ClockManager::removeBDDClock (ptinsth vh)
{
  CH2BDD_remove (vh);
}
