/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   @return true if one of the operand vh1 or vh2 is the fictitious root.
   In this case, vh (out) is the other operand.
 */
  boolean ClockManager::isFormOntick (ptinsth vh1, ptinsth vh2, ptinsth * vh)
{
  if (isFictitiousRootFormula (vh1))
    {
      *vh = vh2;
      return pk_true;
    }
  if (isFictitiousRootFormula (vh2))
    {
      *vh = vh1;
      return pk_true;
    }
  return pk_false;
}

/**
   @return 
   - pk_true if vh2 is the fictitious root and vh1 is not the fictitious root, then vh (out) is the null clock;
   - false otherwise.
*/
boolean
ClockManager::isFormComplOnTick (ptinsth vh1, ptinsth vh2, ptinsth * vh)
{
  if (isFictitiousRootFormula (vh1))
    return pk_false;
  if (!isFictitiousRootFormula (vh2))
    return pk_false;
  *vh = the_nullClock_CKM_copy;
  return pk_true;
}

/**
   @return 
    - pk_true  if vharg1 and vharg2 are in the same tree (the fictitious root is filtered)  and are on different branches. Then :
           - vhemb (out) : is the class of the smaller placed upper bound (branching);
	   - vharg1 (out): the nearest of vhemb;
	   - vharg2 (out): the other;
           - vhson1 (out): child of vhemb on the branch of vharg1
           - vhson2 (out): child of vhemb on the branch of vharg2
    - false 
        - if vharg1 and vharg2 have not the same root or the root is  the fictitious clock. In this case, vhemb (out) is the undefined clock.
	- if vharg1 and vharg2 are in the same tree using the assertions. In this case,  vhemb (out) is the greater clock.            
*/
boolean
ClockManager::getRealBranching (ptinsth * vharg1, ptinsth * vharg2, ptinsth * vhson1, ptinsth * vhson2, ptinsth * vhemb)
{
  boolean b;

  b = getCompleteBranching (vharg1, vharg2, vhson1, vhson2, vhemb);
  if (getBranchingAssertions (vharg1, vharg2, vhemb))
    b = pk_false;
  return b;
}

/** 
    Similar to \ref getCompleteBranching but here the operands (vharg1, vharg2) are not modified.
    vhson1 (resp. vhson2) is the child of vhemb in the branch of vharg1 (vharg2).
*/
boolean
ClockManager::hoclembr (ptinsth vharg1, ptinsth vharg2, ptinsth * vhson1, ptinsth * vhson2, ptinsth * vhemb)
{
  ptinsth saveh;

  saveh = vharg1;
  if (!getCompleteBranching (&vharg1, &vharg2, vhson1, vhson2, vhemb))
    return pk_false;
  if (saveh != vharg1)
    {
      saveh = *vhson1;
      *vhson1 = *vhson2;
      *vhson2 = saveh;
    }
  return pk_true;
}

/**
   @return 
   - pk_true if x and hup are in the same branch and hup is faster than x using the assertions.
   - false otherwise
*/
boolean
ClockManager::isInSameBranchInUsingAssertions (ptinsth x, ptinsth hup)
{
  ptlinsth lh1, lh2;
  ptinsth vh1, vh2, k;
  int d;

  lh1 = lterm1assertpl;
  lh2 = lterm2assertpl;
  x = getRepresentativeFormula (x);
  while (ptlinsth_get_next (&lh1, &vh1))
    {
      ptlinsth_get_next (&lh2, &vh2);
      vh1 = getRepresentativeFormula (vh1);
      vh2 = getRepresentativeFormula (vh2);
      if (vh1 == x)
	{
	  k = getRoot (vh2, &d);
	  if (k == hup)
	    return pk_true;
	}
      if (vh2 == x)
	{
	  k = getRoot (vh1, &d);
	  if (k == hup)
	    return pk_true;
	}
    }
  return pk_false;
}

/**
   @return 
   - pk_true if vh and vhup are in the same branch and vhup is faster than vh using the assertions. 
   The test is applied for the ancestors of vh from vh to its root (using the assertions).
   - false otherwise
*/
boolean
ClockManager::isUpUsingAssertions (ptinsth vhup, ptinsth vh)
{
  ptinsth x, vhr;
  int vd;

  vhr = getRootAndDepth (vh, &vd);
  x = vh;
  while (x != vhr)
    {
      if (isInSameBranchInUsingAssertions (x, vhup))
	return pk_true;
      x = getFather (x);
    }
  return pk_false;
}

/**
   @return 
    - pk_true if the arguments vharg1 and  vharg2 are in the same branch using the assertions. In this case, vhem is the greater clock.
    -false if there are not assertions, or the arguments are not in the same branch.
*/
boolean
ClockManager::getBranchingAssertions (ptinsth * vharg1, ptinsth * vharg2, ptinsth * vhemb)
{
  ptinsth x;
  if (lterm1assertpl == NILLINSTH)
    return pk_false;
  if (isUpUsingAssertions (*vharg1, *vharg2))
    {
      *vhemb = *vharg1;
      return pk_true;
    }
  if (isUpUsingAssertions (*vharg2, *vharg1))
    {
      *vhemb = *vharg2;
      x = *vharg1;
      *vharg1 = *vharg2;
      *vharg2 = x;
      return pk_true;
    }
  return pk_false;
}

/**
   if vh1 is in a clock already placed in a tree then vh1 is isolated in a clock (the constraint is kept in the system using \ref addConstraintPlacedClocks() )
   else there is no action.
   @return the representative formula of the class of vh1.
*/
ptinsth
ClockManager::condintree (ptinsth vh1)
{
  ptinsth vh1head;
  vh1head = getRepresentativeFormula (vh1);
  if (!satisfiesNullDepth (vh1head))
    // if (assumedWithoutConstraints) makeEquivForm(vh1head, vh1); else  
    {
      removeFormulaFromClass (vh1head, vh1);
      addConstraintPlacedClocks (vh1head, vh1);
      vh1head = getRepresentativeFormula (vh1);
    }
  return vh1head;
}

/** @return
    - false if the extractions vh1 ([C]), vh2 ([not C]) with vhcond (^C) as mother clock can be used to
    build an elementary tree of clocks.
    - true otherwise (it exists some placed formula in one of the classes vh1 or vh2).    
*/
boolean
ClockManager::satisfies_condintree (ptinsth vhcond, ptinsth vh1, ptinsth vh2)
{
  ptinsth vh1head, vh2head, vhrac1;
  int depth1;
  boolean b;

  vh1head = condintree (vh1);
  vh2head = condintree (vh2);

  if (assumedWithoutConstraints)
    return pk_false;

  vhcond = getRepresentativeFormula (vhcond);
  if (vhcond == vh1head)
    {
      if (removeFormulaFromClass (vh1head, vh1))
	b = pk_true;
      else
	b = removeFormulaFromClass (vh1head, vhcond);
      if (!b)
	return pk_true;
      vh1head = getRepresentativeFormula (vh1);
      vhcond = getRepresentativeFormula (vhcond);
      addConstraintPlacedClocks (vh1head, vhcond);
      return pk_false;
    }

  if (vhcond == vh2head)
    {
      if (removeFormulaFromClass (vh2head, vh2))
	b = pk_true;
      else
	b = removeFormulaFromClass (vh2head, vhcond);
      if (!b)
	return pk_true;
      vh2head = getRepresentativeFormula (vh2);
      vhcond = getRepresentativeFormula (vhcond);
      addConstraintPlacedClocks (vh2head, vhcond);
      return pk_false;
    }

  if (satisfiesNullClock (vh1head))
    {
      removeFormulaFromClass (vh1head, vh1);
      addConstraintPlacedClocks (vh1head, vh1);
      vh1head = getRepresentativeFormula (vh1);
      return pk_false;
    }

  if (satisfiesNullClock (vh2head))
    {
      removeFormulaFromClass (vh2head, vh2);
      addConstraintPlacedClocks (vh2head, vh2);
      vh2head = getRepresentativeFormula (vh2);
      return pk_false;
    }

  vhrac1 = getRoot (vhcond, &depth1);
  if (depth1 > 0)
    {
      if (vhrac1 == vh1head)
	{
	  removeFormulaFromClass (vh1head, vh1);
	  addConstraintPlacedClocks (vhcond /*vh1head */ , vh1);
	  return pk_true;
	}
      if (vhrac1 == vh2head)
	{
	  removeFormulaFromClass (vh2head, vh2);
	  addConstraintPlacedClocks (vhcond /*vh2head */ , vh2);
	  return pk_true;
	}
    }
  return pk_false;
}

/** 
    It authorises (when b is "true")/suspends (when b is "false") the operations in the trees of clocks.
*/
void
ClockManager::setTreeOperationMode (boolean b)
{
  treeOperationAuthorized = b;
}
