/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   It solves the clock system using a triangularisation method:
      - (1) the sub-formula are placed: [c] and [not c] are children of ^c (elementary trees are build) (\ref buildExtractionTree ())
      - (2) when two clocks are placed in the same tree, formula build on them can be 
        placed using a hierarchical canonical form using BDD technique; (\ref NormalFormTreePlacing())
      - (3) when (2) cannot be applied, classical boolean reductions are applied to try to start again (2).
        This reductions are parametrized by \b vholophmax, called depth of solving  (\ref BooleanReducing()).

      Each step can induce the reduction of number root clock (merging tree by unification). 

      This technique:
         - tries to proved the constraints explained in the system (two formula in the same clock must be proved equivalent),
         - tries to synthetize, if it exists, the "master clock" of the system,
	 - structures the control of the application for code generation (trees give imbricated blocks of statements).

   Before calling this, you must call \ref systemSolving_init method.
   After calling this, you must call \ref systemSolving_postlude method.
   A typical example is : 

    \code
    Example: CKM is the clock manager and EM is the expression manager of the same graph.
      CKM->systemSolving_init();
      do
       {
           CKM->clockCalculus(ilevelmax);
           EM -> rewritingExprEH();      rewriting when/cell expression
       }
      while(CKM->updateHierarchy());
     CKM->systemSolving_postlude();
     \endcode

\see { buildExtractionTree(), NormalFormTreePlacing(), BooleanReducing(), 
       systemSolving_init(), systemSolving_postlude(), updateHierarchy()
       }
*/
void
ClockManager::clockCalculus (int vholophmax)
{
  boolean constraintsOnExtracts;

  if (vholophmax > 0)
    {
      initSystemForModification ();
      constraintsOnExtracts = buildExtractionTree ();
      NormalFormTreePlacing ();
      BooleanReducing (vholophmax);
      if (!constraintsOnExtracts)
	{
	  solvingModifyRepr ();
	  verifyConstraintsOnRoots ();
	  // The following line has been removed by Loïc Besnard, to be fixed ("bad solving results").
	  // booleanNormalFormBuilting(); 
	}
      if (rewritingUsingPlacedConstrainst ())
	{
	  NormalFormTreePlacing ();
	  BooleanReducing (vholophmax);
	}
    }
}

/**
   systemSolving_init is initialized from DCG_basic and secondarily from exclusiveClocks
   It initializes clocktreefront and recursively the BDD package and its interfaces
   The null clock is assumed to be defined
   Must be called before \ref clockCalculus.
*/
void
ClockManager::systemSolving_init ()
{
  clocktreefront_init ();
  setNotFormulaReferences ();
}

/**
   Finalizing the system after solving.
   Must be call after \ref clockCalculus.
*/
void
ClockManager::systemSolving_postlude ()
{
  treeEndSolving ();
}

/** 
    Updating the hierachy after clock calculus.
    @return true if some placed clocks become roots.
    Currently, only placed clock reduced to null clock are considered.
    Such  null clock can appear after rewritting of expressions.

    \see resetClockAsRoot
*/
boolean ClockManager::updateHierarchy ()
{
  return pk_false;

  // Modified By Loïc Besnard, 
  // it induces bad solving results (constraints with null clocks are removed).
  /* ptinsth vh;
     ptlinsth lh;

     lh = getPlacedNullClocks();
     if (lh == NILLINSTH) return pk_false;
     while (ptlinsth_get_next(&lh, &vh))
     {
     if (resetClockAsRoot(vh)) addFormulaToListRoot(vh);
     hclunionwithhnull(vh);
     }
     return pk_true; 
   */
}

/**
   Applying boolean reductions on the formula of the clock system using a value (holoph) as the max depth for rewritings.
*/
void
ClockManager::booleanReducing (int holoph)
{
  BooleanReducing (holoph);
}


/** It tries to remove the not formula (h*not(k)) using the placed formula but 
    not placed in the same trees. Useful to test if a program is polyendochronous.
*/
void
ClockManager::ReduceMinusFormulae(int vholophmax)
{
  // Added for PolyEndo.
  initSystemForModification ();
  reduceMinusFormulae(vholophmax);
  finalizeClockHierarchy ();
}
