/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
  Getting the node of a signal. 
  @param  vsig (in) is a signal identifier or an access to an element of an array.
  @returns the node of the signal or the node associated with the element of an array (previously sets by setNodeAccessToArrayElement).
*/
  ptINode ExprManager::getNode (ptexpcal vsig)
{
  ptINode n;

  if (getAttribute (vsig, AccessToArrayElement, (attributeValue *) & n))
    return n;			// access to an element of array 
  return ptexpcal_get_node (vsig);
}

/**
  Getting the clock of a signal. 
  @param  vsig (in) is a "variable" expression (signal identifier). 
  @returns the clock of the signal (vsig).
*/
ptinsth
ExprManager::getClock (ptexpcal vsig)
{
  ptinsth h;
  h = CKM->get_insthClock (ptexpcal_get_clock (vsig));
  ptexpcal_set_clock (vsig, h);
  return h;
}


/** 
    Testing NIL expression. 
    @returns "true" if exp is the NIL expression, "false" otherwise.
*/
boolean ExprManager::isNILEXPCAL (ptexpcal exp)
{
  return ptexpcal_isNil (exp);
}

/** 
    Testing global signal status. 
    @param vexp (in) any signal expression. 
    @returns "true" if vexp is a signal defined by a set of part definitions nodes (set of ::=), 
    "false" otherwise. 
*/
boolean
ExprManager::isSharedSignal (ptexpcal vexp)
{
  vexp = getDeclaredSignal (vexp);
  if (IsSignal (vexp))
    return ptexpcal_get_sharedTaq (vexp);
  if (getSignalAssignedTo (vexp, &vexp))
    return ptexpcal_get_sharedTaq (vexp);
  return pk_false;
}

/**
   @return "true" if a signal (vsig) is marked as a shared signal by several graphs, "false" otherwise.
*/
boolean ExprManager::isSharedSignalBySeveralGraphs (ptexpcal vsig)
{
  return isMarked (vsig, SHARED_BY_SEVERAL_GRAPHS);
}


/** 
    Testing state variable status. 
    @param vexp (in) any signal expression. 
    @returns "true" if vexp is a state variable, "false" otherwise. 
*/
boolean
ExprManager::isStateVar (ptexpcal vexp)
{
  vexp = getDeclaredSignal (vexp);
  if (IsSignal (vexp))
    return ptexpcal_get_stateTaq (vexp);
  if (getSignalAssignedTo (vexp, &vexp))
    return ptexpcal_get_stateTaq (vexp);
  return pk_false;
}

/** 
    Testing output signal. 
   @param exp (in) any signal expression. 
   @returns  "true" if the expression exp is a output signal, "false" otherwise. 
*/
boolean
ExprManager::IsOutputFlow (ptexpcal exp)
{
  return ptexpcal_IsOutputFlow (exp);
}

/** 
    Testing Input signal. 
    @param exp (in) any signal expression. 
   @returns  "true" if the expression exp is a input signal, "false" otherwise. 
*/
boolean
ExprManager::IsInputFlow (ptexpcal exp)
{
  ptexpcal_getSignalAssignedTo (exp, &exp);
  return ptexpcal_IsInputFlow (exp);
}


/** 
    Getting identifier class. 
    @param  exp (in) is a "variable" expression (signal identifier). 
    @returns the class type of a signal (index, signal, parameter,. . . )
*/
typinst ExprManager::getClassIdentifier (ptexpcal exp)
{
  return (typinst) ptexpcal_get_typeinst (exp);
}

/** 
    @returns "true" if an expression (exp) is a parameter identifier, "false" otherwise. 
*/
boolean ExprManager::IsParameter (ptexpcal exp)
{
  return (boolean) ptexpcal_IsParameter (exp);
}

/** 
    @returns  "true" if an expression (exp) is a constant identifier, "false" otherwise.
*/
boolean ExprManager::IsConstant (ptexpcal exp)
{
  return (boolean) ptexpcal_IsConstant (exp);
}

/** 
    @returns  "true" if an expression (exp) is a signal identifier, "false" otherwise.  
*/
boolean ExprManager::IsSignal (ptexpcal exp)
{
  return (boolean) ptexpcal_IsSignal (exp);
}

/** 
    @returns  "true" if an expression (exp) is a field of a structure/bundle identifier, "false" otherwise.  
*/
boolean ExprManager::IsField (ptexpcal exp)
{
  return (boolean) ptexpcal_IsFieldInstance (exp);
}

/** 
    @returns  "true" if an expression (exp) is a label identifier, "false" otherwise. 
*/
boolean ExprManager::IsLabel (ptexpcal exp)
{
  return (boolean) ptexpcal_IsLabel (exp);
}

/** 
    @returns "true" if an expression (exp) is an index identifier, "false" otherwise. 
 */
boolean ExprManager::IsIndex (ptexpcal exp)
{
  return (boolean) ptexpcal_IsIndex (exp);
}


/** 
   @returns "true" if an expression (exp) is an external constant, "false" otherwise.
*/
boolean ExprManager::IsExternalConstant (ptexpcal exp)
{
  if (!ptexpcal_IsConstant (exp))
    return pk_false;
  return NM->is_undefined (ptexpcal_get_node (exp));
}

/** 
   @returns "true" if a signal (exp) is a signal added by the tool, "false" otherwise.
*/
boolean ExprManager::IsAuxiliarySignal (ptexpcal exp)
{
  if (isMarked (exp, AUXILIARY_IDENT))
    return pk_true;
  return (ptexpcal_get_treeIdentifier (exp) == getIdentNil ());
}

/**
   @param sig (in) assumed to be a "variable" expression (signal identifier). 
   @returns the name of the original signal on which the signal (sig) is built. 
*/
ptident ExprManager::getOriginalName (ptexpcal sig)
{
  return (ptident) ptexpcal_get_treeIdentifier (sig);
}

/** 
   @param sig (in) assumed to be a "variable" expression (signal identifier). 
   @returns  the name of the signal (sig). 
 */
ptident ExprManager::getName (ptexpcal sig)
{
  return (ptident) ptexpcal_get_identifier (sig);
}


/**
   @param exp (in) assumed to be a "variable" expression (signal identifier). 
   @returns the tree of the declaration of the signal. This information (abstract tree) is produced by the parser.
*/
ntrees ExprManager::getTreeOfSignalDeclaration (ptexpcal exp)
{
  return (ntrees) ptexpcal_get_treeOfSignalDeclaration (exp);
}

/**
   @param  exp (in) is a "variable" expression (signal identifier). 
   @returns  the initialization expression of the signal (exp). 
*/
ptexpcal ExprManager::getInitialValue (ptexpcal exp)
{
  return (ptexpcal) ptexpcal_get_initExp (exp);
}


/** 
    @returns the declared signal of the expression (exp), defined by
    - for a identifier x: x.
    - for a field access (x.y): getDeclaredSignal(x).
    - for an array access (x[y]): getDeclaredSignal(x).
*/
ptexpcal ExprManager::getDeclaredSignal (ptexpcal exp)
{
  return ptexpcal_getDeclaredSignal (exp);
}

/**
   Fields of structure/bundle management: for a signal of structured type, the compiler assigns a fictitious signal 
   to each field to have the same access (getting clock, getting node,...) than a standart signal. 
   @returns 
   - "true" if an signal instance s is assigned to exp, s is defined by 
      - exp if exp is a standart signal, 
      - getSignalAssignedTo(x,s), when exp is an array access (x[])
      - the value of the attribute FIELD_SIGNAL for a field access (\see setSignalAssignedTo()).

   - "false" otherwise, s is not modified.
*/
boolean ExprManager::getSignalAssignedTo (ptexpcal exp, ptexpcal * s)
{
  return ptexpcal_getSignalAssignedTo (exp, s);
}

/** 
    @returns 
    - "true" if a list of signals (ls) is assigned to e. This information is available for a standart signal, or a structure (\see setFieldsOf()).
    - "false" otherwise.
*/
boolean ExprManager::getFieldsOf (ptexpcal e, ptlinstances * ls)
{
  return ptexpcal_getAttribute (e, FIELDSLIST, (attributeValue *) ls);
}


/**
   @returns the SIGNAL operator of the expression (exp). 
   See opersignal.h file to have the set of signal operators (incl directory). 
   \see getOperatorClass()
*/
noperators ExprManager::getOperator (ptexpcal exp)
{
  return (noperators) ptexpcal_get_operator (exp);
}

/** 
    @returns the operator class of the expression (exp).
    \see{cloper, getOperator}. 
*/
cloper ExprManager::getOperatorClass (ptexpcal exp)
{
  return (cloper) ptexpcal_get_operatorClass (exp);
}

/**
   @param  exp (in) is a "constant" expression (assumed).
   @returns the value of the entry point in the string table of the constant expression (exp).  
   \see{getOperatorClass, getOperator}
*/
ptrepstr ExprManager::getConstant (ptexpcal exp)
{
  return (ptrepstr) ptexpcal_get_constant (exp);
}

/**
   @param  exp (in) is a "oper1" (unary expression) expression (assumed)
   @returns  the operand of the expression (exp). 
   \see{getOperatorClass, getOperator}
*/
ptexpcal ExprManager::getUnaryArg (ptexpcal exp)
{
  return (ptexpcal) ptexpcal_get_unaryArg (exp);
}

/** 
    Getting operands of a binary expression. 
   @param exp (in) a binary expression (class operator is equal to "oper2"). 
   @param op1 (out) first child of exp
   @param op2 (out) second operand of exp. 
   \see{getOperatorClass, getOperator}
*/
void
ExprManager::getBinaryArgs (ptexpcal exp, ptexpcal * op1, ptexpcal * op2)
{
  ptexpcal_get_binaryArgs (exp, op1, op2);
}

/** 
    Getting an operand of a binary expression. 
    @param exp (in) a binary expression (class operator is  equal to "oper2"). 
    @param pos  (in) int in [1,2]. 
    @returns the pos-th operand of a binary expression exp. 
    \see{getOperatorClass, getOperator}  */
ptexpcal ExprManager::getBinaryArgstAt (ptexpcal exp, int pos)
{
  return (ptexpcal) ptexpcal_get_binaryArgstAt (exp, pos);
}

/** 
    Getting operands of a ternary expression. 
    @param exp (in) a ternary expression (class operator is  equal to "oper3"). 
    @param op1 (out) first child of exp. 
    @param op2 (out) second child of exp. 
    @param op3 (out) third child of exp. 
    \see{getOperatorClass, getOperator}
*/
void
ExprManager::getTernaryArgs (ptexpcal exp, ptexpcal * op1, ptexpcal * op2, ptexpcal * op3)
{
  ptexpcal_get_ternaryArgs (exp, op1, op2, op3);
}

/** 
    Getting an operand of a ternary expression. 
    @param exp (in) a ternary expression.  (class operator is  equal to "oper3"). 
    @param pos  (in) int in [1,3]. 
    @returns the pos-th operand of a ternary expression. 
    \see{getOperatorClass, getOperator}
*/
ptexpcal ExprManager::getTernaryArgstAt (ptexpcal exp, int pos)
{
  return (ptexpcal) ptexpcal_get_ternaryArgstAt (exp, pos);
}

/**
   Getting operands of a temporal binary expression (default operator). 
    @param exp (in) is a "oper2tp" expression. 
    @param op1 (out) first operand of exp. 
    @param op2 (out) second operand of exp. 
    \see{getOperatorClass, getOperator}
*/
void
ExprManager::getDefaultArgs (ptexpcal exp, ptexpcal * op1, ptexpcal * op2)
{
  ptexpcal_get_defaultArgs (exp, op1, op2);
}

/** 
    Getting an operand of a temporal binary expression (default operator). 
    @param exp (in) is a "oper2tp" expression. 
    @param pos  (in) int in [1,2]. 
    @returns the pos-th operand of exp. 
    \see{getOperatorClass, getOperator}
*/
ptexpcal ExprManager::getDefaultArgstAt (ptexpcal exp, int pos)
{
  return (ptexpcal) ptexpcal_get_defaultArgstAt (exp, pos);
}


/**
   Getting left operand of a when/cell expression. 
   @param exp (in) is a "opereh" expression. 
   @returns the left operand of a when expression. 
   \see{getOperatorClass, getOperator}
*/
ptexpcal ExprManager::getWhenOrCellArg (ptexpcal exp)
{
  return (ptexpcal) ptexpcal_get_expWhen (exp);
}


/**
   Getting left operand of a when/cell expression. 
   @param exp (in) is a "opereh" expression. 
   @returns the left operand of a when expression. 
   \see{getOperatorClass, getOperator}
*/
boolean ExprManager::eq_trueExpcal (ptexpcal vexpcal)
{
  return ptexpcal_equalTrue (vexpcal);
}


/**
   Getting left operand of a when/cell expression. 
   @param exp (in) is a "opereh" expression. 
   @returns the left operand of a when expression. 
   \see{getOperatorClass, getOperator}
*/
boolean ExprManager::eq_falseExpcal (ptexpcal vexpcal)
{
  return ptexpcal_equalFalse (vexpcal);
}


/**
   Getting right operand of a when/cell expression. 
   @param vexp (in) is a "when/cell" expression (OperatorClass(exp) = opereh).
   @returns the right operand of a when expression.  
   The test on class operator must be done before this call. 
   \see{getOperatorClass, getOperator}
*/
ptinsth
ExprManager::getContributionClock (ptexpcal vexp)
{
  ptinsth h;

  h = ptexpcal_getContributionClock (vexp);
  if (h == NILINSTH)
    return hundef;
  h = CKM->get_insthClock (h);
  ptexpcal_set_ContributionClock (vexp, h);
  return h;
}


/**
   @returns "true" if an expression (exp) is a list of expressions, "false" otherwise.
   \see{getOperatorClass, getOperator}
 */
boolean ExprManager::satisfiesList (ptexpcal exp)
{
  noperators
    op;

  if (isNILEXPCAL (exp))
    return pk_false;
  op = getOperator (exp);
  return (LL_is_listOp (op));
}

/** 
    Getting operands of a list of expressions. 
    @param list (in/out) a list expression represented as a combed binary expression 
    (class operator is equal to "oper2").  
    @param list (in/out) tail of the input list. 
    @param arg (out) the head element of list (in), if input list is not empty. 
    @returns "true" if input list is not empty, "false" otherwise.
    \see{getNaryArgs, getSize}
*/
boolean ExprManager::getNaryArgs (ptexpcal * list, ptexpcal * arg)
{
  return ptexpcal_get_Nary (list, arg);
}


/** 
    Getting operands of a list of expressions. 
    @param list (in/out) a list expression represented as a combed binary expression 
    (class operator of list is equal to "oper2"). 
    @param list (in/out) tail of the input list. 
    @returns the head element of the input list, NILEXPCAL for an empty list. 
    \see{getNaryArgs, getSize}
*/
ptexpcal
ExprManager::getNaryArgs (ptexpcal * list)
{
  return ptexpcal_getNaryArgs (list);
}

/**
   Getting size of a list. 
   @param list (in) a list expression represented as a combed binary expression. 
   @returns the size of the list operand
*/
int
ExprManager::getSize (ptexpcal list)
{
  return ptexpcal_get_nbelem (list);
}

/**
   Getting operands of an external function expression (e), i.e. its class operator is equal to "appel". 
   @param e (in) an expression
   @param id (out) : the name of the function. 
   @param lp (out) : the list of parameters. 
   @param inputs (out) : the list of input expressions. 
   @returns 
   - "true" if e is a function call, the output parameters are set,
   - "false" otherwise, the output parameters are not modified.
*/
boolean ExprManager::getFunctionCall (ptexpcal e, ptident * id, ptexpcal * lp, ptexpcal * inputs)
{
  return ptexpcal_get_funcCall (e, id, lp, inputs);
}


/** Testing expression class to when/cell expression. 
  @param vexp (in) an any expression
  @returns "true" if vexp denotes an "opereh" expression (E when h, E cell h), "false" otherwise. 
*/
boolean
ExprManager::IsWhenOrCellExpr (ptexpcal vexp)
{
  return ptexpcal_satisfies_WhenCell (vexp);
}


/** Testing expression to cell expression. 
   @param exp (in) any signal expression. 
   @returns  "true" if the expression is a cell one (E cell h), "false" otherwise. 
*/
boolean
ExprManager::IsCellExpr (ptexpcal exp)
{
  return ptexpcal_satisfies_cell (exp);
}


/** Getting operand of a type expression. 
    @param exp (in) an expression assumed to be a reference to a type (i.e. its class operator must be equal to "opertyp").
    @returns the referenced type T of the expression. 
    \see{getOperatorClass}
*/
TypeType ExprManager::getExpType (ptexpcal exp)
{
  return ptexpcal_get_exprType (exp);
}

/** 
    Getting the type of an expression. 
    @param exp (in) any signal expression. 
    @returns  the type of the expression exp. 
*/
TypeType ExprManager::getType (ptexpcal exp)
{
  return (TypeType) ptexpcal_get_type (exp);
}

/**
   Getting the array type information of an expression. 
   @param exp (in) any signal expression. 
   @param baseType (out) the elementary type of the expression exp if it is a array 
   or the type of the expression.
   @param dimLst (out) the list of Dimensions if it is a array, NILEXPCAL otherwise.
   @returns "true" if the type of the operand expression is an array, "false" otherwise.
*/
boolean ExprManager::getTypeAndDim (ptexpcal exp, TypeType * baseType, ptexpcal * dimLst)
{
  TypeType
    t;

  t = getUnamedType (exp);
  if (TypeTestArrayClass (t))
    {
      *baseType = ArrayMatrixTypeGetTerm (t);
      *dimLst = ArrayMatrixTypeGetDimLst (t);
      return pk_true;
    }
  *dimLst = NILEXPCAL;
  *baseType = t;
  return pk_false;
}


/** Getting the unamed type of an expression (misc). 
  @param exp (in) any signal expression. 
  @returns the unamed type of the expression exp. 
*/
TypeType ExprManager::getUnamedType (ptexpcal exp)
{
  return (TypeType) ptexpcal_getUnamedType (exp);
}

/** Testing type of an expression to event type. 
   @returns "true" if the type of the expression  exp is event, "false" otherwise. 
*/
boolean ExprManager::IsEvent (ptexpcal exp)
{
  return (boolean) ptexpcal_IsEvent (exp);
}

/** Testing type of an expression to boolean type. 
   @returns "true" if the type of the expression exp is boolean, "false" otherwise. 
*/
boolean ExprManager::IsBoolean (ptexpcal exp)
{
  return ptexpcal_IsBoolean (exp);
}

/** Testing type of an expression to free type (misc). 
   @returns "true" if the expression is not typed, "false" otherwise.
*/
boolean ExprManager::IsFreeType (ptexpcal exp)
{
  return (boolean) ptexpcal_IsFreeType (exp);
}

/** Testing type of an expression to int type. 
   @returns "true" if the type of exp is int (predefined int, short,...), "false" otherwise.
*/
boolean ExprManager::IsInteger (ptexpcal exp)
{
  return (boolean) ptexpcal_hasIntegerType (exp);
}

/**
  Testing type of an expression to predefined type. 
  @returns "true" if the type of the operand expression is a predefined one, "false" otherwise. 
*/
boolean
ExprManager::IsTypeExprPredefinedType (ptexpcal exp)
{
  return ptexpcal_satisfies_predefinedType (exp);
}

/** Getting number of a dimensions of a type of an expression. 
   @param exp (in) any signal expression. 
   @returns the number of dimensions of the type of the expression exp. 
*/
int
ExprManager::getNumberOfDimensions (ptexpcal exp)
{
  return ptexpcal_get_nbDim (exp);
}

/**
   @returns 
   - "true" if the i-th dimension (ithDim) of a signal (vsig) exists, ithDimExpr is then its expression
   - "false" otherwise, ithDimExpr is not modified.
*/
boolean ExprManager::getIthDimension (ptexpcal vsig, long int ithDim, ptexpcal * ithDimExpr)
{
  return ptexpcal_get_aDim (vsig, (int) ithDim, ithDimExpr);
}

/** 
    Getting back information to clock system. 
    @param vexp (in) any signal expression. 
    @param vhsub (out) the extraction (clock) built on vexp. 
    @returns "true" if the extraction vhsub exists in the clock system, "false" otherwise. 
*/
boolean
ExprManager::getExtractionLink (ptexpcal vexp, ptinsth * vhsub)
{
  return ptexpcal_get_extraction (vexp, vhsub);
}

/** Misc. It should be internal of the manager.\n
    @param vexp (in) any signal expression. 
    @param nexp (out) the current equivalent expression of vexp (rewritting result)
    @returns "true" is such expression exists, "false" otherwise. 
*/
boolean
ExprManager::getIndirection (ptexpcal vexp, ptexpcal * nexp)
{
  return ptexpcal_get_indir (vexp, nexp);
}

/** 
    Getting the set of nodes connected to an expression. 
    For example, for the node n defined by x:=E, x and E have n in such a set. 
    @param vexp (in) any signal expression. 
    @returns the list of the nodes connected to the expression. 
*/
ptlINode
ExprManager::getNodesFeedBy (ptexpcal vexp)
{
  return ptexpcal_get_NodesFeedBy (vexp);
}

/**
   Getting all the identifiers referenced in an expression.
   @param vexp (in) any signal expression.
   @returns the list of identifiers (signal, constants, labels . . . ) referenced in the operand expression (vexp). 
*/
ptlinstances
ExprManager::getIdsReferencedIn (ptexpcal vexp)
{
  return ptexpcal_getIdsReferencedIn (vexp);
}

/**
   Getting all the constants referenced in an expression.
   @param vexp (in) any signal expression.
   @returns the list of identifiers (signal, constants, labels . . . ) referenced in the operand expression (vexp). 
*/
ptlinstances
ExprManager::getConstantsReferencedIn (ptexpcal vexp)
{
  return ptexpcal_getConstantsReferencedIn (vexp);
}


/**
   Getting signals and fields identifiers referenced in an expression. 
   @param vexp (in) any signal expression. 
   @returns the list of identifiers (signal) referenced in the operand expression.
*/
ptlinstances
ExprManager::getFieldsAndSignalsReferencedIn (ptexpcal vexp)
{
  return ptexpcal_getFieldsAndSignalsReferencedIn (vexp);
}


/** Getting signals referenced in an expression. 
   @param vexp (in) any signal expression. 
   @returns the list of identifiers (signal) referenced in the operand expression. 
*/
ptlinstances
ExprManager::getSignalsReferencedIn (ptexpcal vexp)
{
  return ptexpcal_getSignalsReferencedIn (vexp);
}


/** Getting indexes referenced in an expression. 
   @param vexp (in) any signal expression. 
   @returns the list of identifiers (indexes) referenced in the operand expression. 
*/
ptlinstances
ExprManager::getIndexesReferencedIn (ptexpcal vexp)
{
  return ptexpcal_getIndexesReferencedIn (vexp);
}

/** 
    Getting signals annotated as possibly used as indexes referenced in an expression. 
    @param vexp (in) any signal expression. 
    @returns the list of identifiers (indexes) referenced in the operand expression. 
*/
ptlinstances
ExprManager::getPossiblyIndexesReferencedIn (ptexpcal vexp)
{
  return ptexpcal_getPossiblyIndexesReferencedIn (vexp);
}


/**
   @returns the list of the signals referenced in an expression (e) taking into account the associations:
   for a field reference (x.f) if a signal (y) is associated with it, y is considered not x.f.
   (\see setSignalAssignedTo())
*/
ptlinstances
ExprManager::getSignalsDefinedIn (ptexpcal e)
{
  ptlinstances l;
  noperators op;
  ptexpcal x;

  if (e == NILEXPCAL)
    return NILLINSTANCES;
  if (IsSignal (e))
    return ptlinstances_push_front (NILLINSTANCES, e);
  op = ptexpcal_get_operator (e);
  if (op == opcham)
    return getSignalsDefinedIn (ptexpcal_get_binaryArgstAt (e, 1));
  if (getSignalAssignedTo (e, &e))
    return getSignalsDefinedIn (e);
  if (op == opelem)
    return getSignalsDefinedIn (ptexpcal_get_binaryArgstAt (e, 1));
  if (satisfiesList (e))
    {
      l = NILLINSTANCES;
      while (ptexpcal_get_Nary (&e, &x))
	l = ptlinstances_sortedfusion (getSignalsDefinedIn (x), l);
      return l;
    }
  pK_messenger->Caseerror (currentLine, currentFile);
  return NILLINSTANCES;
}

/** 
    Getting inner expression filtering when/cell operators. 
    @param vexp (in) an any expression
    @returns the left operand of the operand expression until such operand is not a cell/when expression.
    \note no effect for a default expression: vexp is returned.
*/
ptexpcal
ExprManager::getBasicExpr (ptexpcal vexp)
{
  return ptexpcal_get_basicExpr (vexp);
}

/** 
    Testing an expression: is it an expression on constants ?.
    @param exp (in) an any expression
    @returns "true" if an expression (exp) is an expression on constants, "false" otherwise
*/
boolean
ExprManager::IsConstantExpr (ptexpcal exp)
{
  return ptexpcal_satisfies_ConstantExpr (exp);
}

/** 
    @returns "true" if all elements of an expression (exp) are defined by the same expression (vce), "false" otherwise.
    @param vce (out) the expression that defines exp expression.
    @param exp (in) exp expression sur signal.
    \note \DocNotComplete (spacial processing).
*/
boolean
ExprManager::IsSameExpressionForEachElement (ptexpcal exp, ptexpcal * vce)
{
  return ptexpcal_satisfies_DefinedBySameExpression (exp, vce);
}

/**
   Getting the root (representative signal) of aggregations from a signal (optimization). 
   @param exp (in) any signal expression. 
   @returns the representative signal of the signal ?exp.  
   It uses the attribute AGGREGATTRIBUTE (predefined attribute, declared in CompilerAttributes.h file).

   Example:
   For the following SIGNAL example (| x := y when c | z := x cell d |),
   the signals x, y and z can be represented by the same variable. It is used for optimisations.
*/
ptexpcal
ExprManager::getAggregation (ptexpcal exp)
{
  ptexpcal v;

  if (!getAttribute (exp, AGGREGATTRIBUTE, (attributeValue *) & v))
    return exp;
  return getAggregation (v);
}


/**
   Testing aggregation informations for a signal (optimizations).  
   @param exp (in) any signal expression. 
   @returns "true" if the signal ?exp is not a representative signal for the code generation, "false" otherwise.
   \note 
   It uses the attribute AGGREGATTRIBUTE (predefined attribute, declared in CompilerAttributes.h file).
*/
boolean
ExprManager::isAggregatedSignal (ptexpcal exp)
{
  ptexpcal v;
  return getAttribute (exp, AGGREGATTRIBUTE, (attributeValue *) & v);
}

/**
  Getting int value of an expression. 
  @param vexp (in) any signal expression. 
  @param val (out) value assigned if it exists to the exp expression. 
  @returns "true" if val (out) is defined, "false" otherwise.
*/
boolean
ExprManager::getIntegerValue (ptexpcal vexp, compilerInteger * val)
{
  if (ptexpcal_isNil (vexp))
    return pk_false;
  if (!ptexpcal_hasIntegerType (vexp))
    return pk_false;
  if (!ptexpcal_IsAssigned (vexp))
    return pk_false;
  *val = ptexpcal_get_integerValue (vexp);
  return pk_true;
}

/**  Testing "bounded" property. 
     @returns "true" if the exp expression is bounded, "false" otherwise.
 */
boolean ExprManager::IsBounded (ptexpcal exp)
{
  return (boolean) ptexpcal_IsBounded (exp);
}

/**
   @return "true" if the expression lb is bounded and the values are available (bi: lower bound, bs: upper bounbd), "false" otherwise.
   bi = bs for a valued signal.
*/
boolean
ExprManager::getBounds (ptexpcal exp, compilerInteger * bi, compilerInteger * bs)
{
  if (exp == NILEXPCAL)
    return pk_false;
  if (ptexpcal_IsAssigned (exp))
    {
      *bi = ptexpcal_get_integerValue (exp);
      *bs = *bi;
      return pk_true;
    }
  if (ptexpcal_IsBounded (exp))
    {
      ptexpcal_get_integerBounds (exp, bi, bs);
      return pk_true;
    }
  return pk_false;
}

/**  Testing "defined" property. 
    @returns "true" if the exp expression is defined but not valued, "false" otherwise.
*/
boolean ExprManager::IsDefinedSignal (ptexpcal exp)
{
  return (boolean) ptexpcal_IsDefinedSignal (exp);
}

/**  
     Testing  if a signal is free (not defined, not assigned, ...). 
     @returns "true" if the exp expression is defined but not valued, "false" otherwise.
*/
boolean ExprManager::IsFreeSignal (ptexpcal exp)
{
  return (boolean) ptexpcal_IsFreeSignal (exp);
}

/**  Testing value of an expression. 
     @returns "true" if the value of exp expression is known: int greater/equal to 0, "false" otherwise.
*/
boolean ExprManager::hasIntegerValue (ptexpcal exp)
{
  return ptexpcal_hasIntegerValue (exp);
}

/** Testing "assigned" property.
   @returns  "true" if the expression exp is an valued int expression, "false" otherwise.
*/
boolean ExprManager::IsAssigned (ptexpcal exp)
{
  return (boolean) ptexpcal_IsAssigned (exp);
}

/** Getting int value of an expression. 
    Assumed : exp is supposed to be an "assigned" expression.
    @returns  the value assigned to the expression (exp).  
   \see{IsAssigned}. 
*/
long int
ExprManager::getValue (ptexpcal exp)
{
  return ptexpcal_get_integerValue (exp);
}

/** Getting the list of referenced signals in an expression without state variable reference.
   @param exp  (in) an Synchronous expression. 
   @returns a list of signals such that x[], x[?], x, x. y , without state variables. 
*/
ptlinstances
ExprManager::getPredecessors (ptexpcal exp)
{
  return ptexpcal_getPredecessors (exp);	// x[], x[?], x, x. y , pas de variables d'etat
}


/** Testing the existence of a opIndex-nth child for the expression e. 
   @param e (in) an signal expression. 
   @param opIndex  (in) an int. 
   @returns "true" if the opIndex-nth child exists, "false" otherwise. 
*/
boolean ExprManager::existChild (ptexpcal e, int opIndex)
{
  cloper
    cl;
  if (isNILEXPCAL (e))
    return pk_false;
  if (opIndex < 0)
    return pk_false;
  cl = getOperatorClass (e);
  if (cl == variabl)
    return (opIndex == 0);
  if (cl == constant)
    return (opIndex == 0);
  if (opIndex == 0)
    return pk_false;
  if (cl == oper1)
    return (opIndex == 1);
  if (cl == oper1tp)
    return (opIndex == 1);
  if (cl == oper2)
    return (opIndex <= 2);
  if (cl == oper2tp)
    return (opIndex <= 2);
  if (cl == oper3)
    return (opIndex <= 3);
  if (cl == oper3tp)
    return (opIndex <= 3);
  if (cl == opereh)
    return (opIndex <= 2);
  if (cl == opertyp)
    return (opIndex == 1);
  return pk_false;
}




/** Getting clock of some expression. 
   @param pt (in) a expression of { variabl, oper1, oper2,opereh} class. 
   @param vh (out) the clock of the expression (if found)
   @returns "true" if the clock (vh) of the expression is set, "false" otherwise. 
*/
boolean
ExprManager::getSynchronousExprHorl (ptexpcal pt, ptinsth * vh)
{
  noperators op;
  ptexpcal pt1;
  cloper t;
  typinst tinst;
  TypeType TT;

  t = ptexpcal_get_operatorClass (pt);
  switch (t)
    {
    case variabl:
      tinst = ptexpcal_get_typeinst (pt);
      if ((tinst == tsignal) || (tinst == tfield) || (tinst == tlabel) || (tinst == tindex))
	{
	  *vh = ptexpcal_get_clock (pt);
	  return pk_true;
	}
      return pk_false;

    case oper1:
      if (ptexpcal_get_operator (pt) == opvvarin)
	return pk_false;	// attention 
      return getSynchronousExprHorl (ptexpcal_get_unaryArg (pt), vh);

    case oper2:
      op = ptexpcal_get_operator (pt);
      if (op == opleft_tt)
	return pk_false;	// attention
      if (op == opcham)
	{
	  // bundle or structure type
	  if (ptexpcal_getSignalAssignedTo (pt, &pt1))
	    {
	      *vh = ptexpcal_get_clock (pt1);
	      return pk_true;
	    }
	  pt1 = ptexpcal_getDeclaredSignal (pt);
	  TT = getType (pt1);
	  if (TypeTestStructureClass (TT))
	    {
	      *vh = ptexpcal_get_clock (pt1);
	      return pk_true;
	    }
	  return pk_false;	// bundle.
	}

      if (getSynchronousExprHorl (ptexpcal_get_binaryArgstAt (pt, 1), vh))
	return pk_true;
      pt1 = ptexpcal_get_binaryArgstAt (pt, 2);
      if (pt1 != NILEXPCAL)
	return getSynchronousExprHorl (pt1, vh);
      return pk_false;

    case opereh:
      if (getSynchronousExprHorl (ptexpcal_get_expWhen (pt), vh))
	return pk_false;
      // Expression constante.
      *vh = ptexpcal_getContributionClock (pt);
      return pk_true;

      /* case constant: case oper2tp: case oper3: case oper1tp: case oper3tp: case opertyp: */
    default:
      return pk_false;
    }
}

/** Getting clock of an expression. 
   @param exp (in) a signal expression. 
   @param h  (out) clock of the expression. 
   @returns "true" if the clock (h) has been found, "false" otherwise (constant expression for example). 
*/

boolean
ExprManager::getExprClock (ptexpcal exp, ptinsth * h)
{
  cloper cl;
  if (isNILEXPCAL (exp)) return pk_false;
  cl = getOperatorClass (exp);
  switch (cl)
    {
    case oper1tp:		// Pas d'occurence
    case oper3tp:		// Pas d'occurence
      return pk_false;

    case oper2tp:		// default
    case opereh:		// E when h or E cell h
      (*h) = CKM->get_insthClock (getContributionClock (exp));
      return pk_true;

    case call:
      (*h) = CKM->get_insthClock (getContributionClock (exp));
      break;


    default:
      // variabl, constant, oper1, oper2, oper3
      if (!getSynchronousExprHorl (exp, h)) return pk_false;
      *h = CKM->get_insthClock (*h);
      return pk_true;
    }
  return pk_false;
}

/**
   Indexed signals (array process).
   @returns  the exception expression assigned to the signal (exp).

   \note \DocNotComplete.
   It is managed for "with part" of array processes.  
*/
ptexpcal ExprManager::getException (ptexpcal exp)
{
  return (ptexpcal) ptexpcal_get_exception (exp);
}

/**
   Indexed signals (array process). 
   @returns the signal (exp) without array indexes. 
*/
ptexpcal
ExprManager::getUnindexedSignal (ptexpcal vexp)
{
  return ptexpcal_getUnindexedSignal (vexp);
}

/**
   @returns the abstract tree assigned, previously set by a call to the method setBackAnnotation(), with an expression (n).
*/
BackAnnotationType ExprManager::getBackAnnotation (ptexpcal n)
{
  return ptexpcal_get_backAnnotation (n);
}


/**
   @returns "true" if for the expression (vexp) is  x, x[], x?, x.y ans x a state variable, "false" otherwise.
*/
boolean
ExprManager::isReferenceToStateVar (ptexpcal vexp)
{
  return ptexpcal_isReferenceToStateVar (vexp);
}


/**
   Getting travel indexes assigned to an expression.
   @param exp (in) a signal expression.
   @param l (out) the list of the travel indexes assigned to the operand expression.
   @returns "true" if (*l) is not empty, "false" otherwise.
   \note \DocNotComplete (what is a "travel index"?)
*/
boolean
ExprManager::getTravelIndexes (ptexpcal exp, ptlinstances * l)
{
  *l = ptexpcal_get_travelIndexes (exp);
  return ((*l) != NILLINSTANCES);
}

/**
   Array processes management:
   @returns "true" if it exists a clone (vClone) of v, "false" otherwise (vClone is not modified).
   A clone is a copy of an identifier used in array of processes (x[I] is a clone of x, x[I][J] is a clone of x[I],...)
   Note that the clones are managed by a circular list with a header element.
   See isCloneClassHead(), setClone().
*/
boolean
ExprManager::getClone (ptexpcal v, ptexpcal * vClone)
{
  return getAttribute (v, CLONE_ATTRIBUTE, (attributeValue *) vClone);
}

/** Array processes management:
    @returns "true" if a signal (v) is the "clone class head" element, "false" otherwise.
    \see {setClone(), setCloneClassHead()}
*/
boolean
ExprManager::isCloneClassHead (ptexpcal v)
{
  return isMarked (v, CLONE_CLASS_HEAD);
}

/**
   @returns the clone class head for a signal (v). It is v when there is no associated clone.
*/
ptexpcal
ExprManager::getCloneClassHead (ptexpcal v)
{
  ptexpcal x;

  if (isCloneClassHead (v))
    return v;
  if (!getClone (v, &x))
    return v;
  return getCloneClassHead (x);
}


/**
   @returns "true" the utility clock (hu) of a signal (exp) is defined, "false" otherwise.
*/
boolean
ExprManager::getUtilityClock (ptexpcal exp, ptinsth * hu)
{
  if (!getAttribute (exp, UTILITYCLOCK, (attributeValue *) hu)) return pk_false;
  (*hu) = CKM->getRepresentativeFormula (*hu);
  return pk_true;
}

/**
   @returns "true" the utility clock (hu) of a signal (exp) in a subgraph (sg) is defined, "false" otherwise.
*/
boolean
ExprManager::getUtilityClockInSubGraph (ptexpcal exp, ptINode sg, ptinsth * hu)
{
  return NM->getUtilityClockInSubGraph (sg, exp, hu);
}


/**
   @returns pk_true the assignment clock (hu) of a signal (exp) is defined, false otherwise.
*/
boolean
ExprManager::getAssignmentClock (ptexpcal exp, ptinsth * vassclk)
{
  if (!getAttribute (exp, ASSIGNMENTCLOCK, (attributeValue *) vassclk)) return pk_false;
  (*vassclk) = CKM->getRepresentativeFormula (*vassclk);
  return !CKM->eqClock (*vassclk, hundef);
}


/**
   @returns pk_true the assignment clock (hu) of a signal (exp) is defined, false otherwise.
*/
boolean
ExprManager::getDefaultConstantValueInfos (ptexpcal exp, ptinsth * vh, ptexpcal * vExpr)
{
  if (!getAttribute (exp, ASSIGNMENTCLOCKPART, (attributeValue *) vh))
    return pk_false;
  if (!getAttribute (exp, DEFAULCONSTANTVALUE, (attributeValue *) vExpr))
    return pk_false;
  return pk_true;
}

/** 
   @param vexp (in) any expression.
   @returns the list of contribution clock referenced in the expression.
*/
ptlinsth
ExprManager::getClocksReferencedIn (ptexpcal vexp)
{
  ptexpcal vexpcal, exp1, exp2;
  ptinsth vh;
  ptlinsth lhcontr;
  boolean b;

  if (ptexpcal_isNil (vexp))
    return NILLINSTH;

  if (satisfiesList (vexp))
    {
      lhcontr = NILLINSTH;
      while (vexp != NILEXPCAL)
	{
	  exp1 = ptexpcal_getNaryArgs (&vexp);
	  lhcontr = ptlinsth_sorted_fusion (lhcontr, getClocksReferencedIn (exp1));
	}
      return lhcontr;
    }
  // Cell, default, when
  if (ptexpcal_satisfies_cell (vexp))
    vexp = ptexpcal_get_expWhen (vexp);
  vexpcal = ptexpcal_get_expWhen (vexp);
  lhcontr = NILLINSTH;
  while (ptexpcal_get_operator (vexpcal) == opmele)
    {
      lhcontr = ptlinsth_sorted_insert_elem (lhcontr, getContributionClock (vexpcal), &b);
      exp2 = ptexpcal_get_defaultArgstAt (vexpcal, 2);
      if (ptexpcal_satisfies_WhenCell (exp2))
	lhcontr = ptlinsth_sorted_insert_elem (lhcontr, getContributionClock (exp2), &b);
      vexp = ptexpcal_get_defaultArgstAt (vexpcal, 1);
      vexpcal = ptexpcal_get_expWhen (vexp);
    }

  if (ptexpcal_get_operator (vexpcal) == opleft_tt)
    {
      ptexpcal_get_binaryArgs (vexpcal, &exp1, &exp2);
      if (getSynchronousExprHorl (exp1, &vh))
	lhcontr = ptlinsth_sorted_insert_elem (lhcontr, vh, &b);
    }
  if (ptexpcal_satisfies_WhenCell (vexp))
    lhcontr = ptlinsth_sorted_insert_elem (lhcontr, getContributionClock (vexp), &b);
  else if (ptexpcal_satisfies_defaultExpression (vexp))
    lhcontr = ptlinsth_sorted_insert_elem (lhcontr, getContributionClock (vexp), &b);

  return lhcontr;
}


/**
   @return "true" if exp is a default, "false" otherwise.
*/
boolean
ExprManager::isDefaultExpression (ptexpcal exp)
{
  return ptexpcal_satisfies_defaultExpression (exp);
}

/**
   @returns 
     "true" if an expression (exp) is a logical expression (i.e. identifier | not(E) | E1 or E2 | E1 and E2), 
     "false" otherwise.
*/
_VIRTUAL_ boolean ExprManager::isLogicalBooleanExpression (ptexpcal exp)
{
  noperators
    op;

  if (exp == NILEXPCAL)
    return pk_true;
  op = getOperator (exp);
  switch (getOperatorClass (exp))
    {
    case variabl:
    case constant:
      return pk_true;

    case oper1:
    case oper1tp:
      return isLogicalBooleanExpression (ptexpcal_get_unaryArg (exp));

    case oper2:
      if ((op == opnon) || (op == opet) || (op == opou) || (op == opleft_tt))
	{
	  if (!isLogicalBooleanExpression (ptexpcal_get_binaryArgstAt (exp, 1)))
	    return pk_false;
	  return isLogicalBooleanExpression (ptexpcal_get_binaryArgstAt (exp, 2));
	}
      return pk_false;

    case oper2tp:
      if (!isLogicalBooleanExpression (ptexpcal_get_binaryArgstAt (exp, 1)))
	return pk_false;
      return isLogicalBooleanExpression (ptexpcal_get_binaryArgstAt (exp, 2));

    case oper3:
    case oper3tp:
      if (!isLogicalBooleanExpression (ptexpcal_get_ternaryArgstAt (exp, 1)))
	return pk_false;
      if (!isLogicalBooleanExpression (ptexpcal_get_ternaryArgstAt (exp, 2)))
	return pk_false;
      return isLogicalBooleanExpression (ptexpcal_get_ternaryArgstAt (exp, 3));

    case opereh:
      return isLogicalBooleanExpression (ptexpcal_get_expWhen (exp));

    case opertyp:
      return pk_false;

    default:
      pK_messenger->Caseerror (currentLine, currentFile);
      return pk_false;
    }
}

/**
   Specific to array processes.
   @returns "true" if the expression e does not depend on "current" indexes of array processes, "false" otherwise.   
   \note \DocNotComplete.
*/
boolean
ExprManager::isIndexDependent (ptexpcal e)
{
  ptexpcal e1;

  if (!satisfiesList (e))
    return analyzeIndexDependence (e);
  while (ptexpcal_get_Nary (&e, &e1))
    if (isIndexDependent (e1))
      return pk_true;
  return pk_false;
}


/**
   Syntactic expression equivalence.
   @param exp1, exp2 (in) signal expressions.
   @param lookvalue (in) a boolean used to 
   @returns 
   - "true" if the operand expressions are syntacticly identical (lookvalue is "false") or have the same value (lookvalue is "true")
   - "false" otherwise.
*/
boolean ExprManager::isEquivalent (ptexpcal exp1, ptexpcal exp2, boolean lookvalue)
{
  return (boolean) ptexpcal_isEquivalent (exp1, exp2, lookvalue);
}


/**
   Getting references to a signal. 
   @returns the list of the references to the expression exp.
*/
ptlexpcal ExprManager::getLuse (ptexpcal exp)
{
  return (ptlexpcal) ptexpcal_get_luse (exp);
}

/**
   Getting the number assigned to an expression. 
   @returns the number (unicity is guaranteed) assigned to exp.
  */
int
ExprManager::getNum (ptexpcal exp)
{
  return (int) ptexpcal_get_num (exp);
}


/** 
    Optimized travel indexes.
    @param exp (in) a signal expression.
    @returns the list of optimized travel indexes assigned to an expression.
    \note \DocNotComplete (what is a "travel index"?, what is an optimized "travel index"?)
*/
ptlinstances
ExprManager::getOptimizedTravelIndexes (ptexpcal exp)
{
  ptlinstances l;

  if (getAttribute (exp, OPTIMIZED_INDEX_LIST, (attributeValue *) & l))
    return l;
  return NILLINSTANCES;
}


/** 
    It tests the initial value compatibility: 
    two expressions exp1, exp2 are compatibles if they are constant expressions with the same value.
    A constant expression is a constant, a parameter identifier, a constant identifier, or an op(e1,..,en) and each ei are constant expressions.
    
    @returns "true" if such property is satisfied, "false" otherwise.
    \note incomplete test.
*/
boolean
ExprManager::IsCompatibleWith (ptexpcal exp1, ptexpcal exp2)
{
  ptexpcal vexp1, vexp2;

  if (exp1 == exp2)
    return pk_true;
  if ((exp1 == NILEXPCAL) || (exp2 == NILEXPCAL))
    return pk_true;
  if (!ptexpcal_get_ConstantInitExpr (exp1, &vexp1))
    return pk_false;
  if (!ptexpcal_get_ConstantInitExpr (exp2, &vexp2))
    return pk_false;
  return ptexpcal_isEquivalent (vexp1, vexp2, pk_true);
}

/**
   @param vexp  (in) any expression.
   @param nexp (out) equivalent expression of vexp obtained using valuations.
   @param brestrict (in) with the value "true" only the int synchronous expressions are tested.
   When it is pk_false, the expression must be the definition of a signal (ie X:= vexp), to have no problem with the clocks.
   @returns "true" if nexp is set, "false" otherwise.
*/
boolean
ExprManager::getEquivConstant (ptexpcal vexp, ptexpcal * nexp, boolean brestrict)
{
  long int val;
  cloper classOp;

  classOp = ptexpcal_get_operatorClass (vexp);
  if (classOp == constant)
    return pk_false;		// already in the form.
  if (!ptexpcal_IsAssigned (vexp))
    return pk_false;

  if (brestrict)
    {
      if (ptexpcal_get_type (vexp) != IntegerType)
	return pk_false;	// Restriction  
      if (classOp == oper1tp)
	return pk_false;
      if (classOp == oper2tp)
	return pk_false;
      if (classOp == oper3tp)
	return pk_false;
      if (classOp == opereh)
	return pk_false;
      if (classOp == opertyp)
	return pk_false;
      val = ptexpcal_get_integerValue (vexp);
      if (val >= 0)
	{
	  *nexp = mkIntegerConstant (val);
	  return pk_true;
	}
      if (classOp != oper2)
	return pk_false;	// why ?
      *nexp = mkIntegerConstant (val);
      return pk_true;
    }
  return getExprOfConstant (vexp, nexp);
}

/**
   @return "true" if the operation succeeds, "false" otherwise (vexp is not a constant expression).
   For an evaluated expression (vexp), nexp is the expression of the associated constant.
*/
boolean
ExprManager::getExprOfConstant (ptexpcal vexp, ptexpcal * nexp)
{
  long int val;
  TypeType T;

  T = getUnamedType (vexp);
  if (T == IntegerType)
    {
      val = ptexpcal_get_integerValue (vexp);
      *nexp = mkIntegerConstant (val);
      return pk_true;
    }

  if (T == BooleanType)
    {
      val = ptexpcal_get_integerValue (vexp);
      if (val == 0)
	*nexp = ptexpcal_getFalse ();
      else
	*nexp = ptexpcal_getTrue ();
      return pk_true;
    }

  if (T == PureType)
    {
      *nexp = ptexpcal_getTrue ();
      return pk_true;
    }
  return pk_false;
}

/**
   @return "true" if vsig is defined by (expr when ^expr), with expr ^= vsig, expr is a synchronous expression, 
   "false" otherwise.
*/
boolean ExprManager::getSyncExprDefSig (ptexpcal vsig, ptexpcal * expr)
{
  ptINode
    node;
  ptexpcal
    vexpdedef,
    vexp;
  ptinsth
    vh,
    vhs;
  noperators
    op;
  boolean
    b;

  if (!IsSignal (vsig))
    return pk_false;
  node = getNode (vsig);
  if (!NM->is_equation (node))
    return pk_false;
  if (!NM->get_rightOperand (node, &vexp))
    return pk_false;
  if (IsCellExpr (vexp))
    return pk_false;
  vexpdedef = getBasicExpr (vexp);
  *expr = vexpdedef;
  op = getOperator (vexpdedef);
  if ((op == opleft_tt) || (op == opmele))
    return pk_false;
  b = pk_false;
  vh = CKM->getRepresentativeFormula (getContributionClock (vexp));
  if (!CKM->satisfiesInitClock (vh))
    if (getSynchronousExprHorl (vexpdedef, &vhs))
      {
	vhs = CKM->getRepresentativeFormula (vhs);
	b = (vh == vhs);
      }
  if (!b)
    return pk_false;
  return pk_true;
}

/**
   It returns the expression to be exported (boolean clocks) for the clock "when vsig" where vsig is a signal (X).
   For X := (E when h), ^E = h, E not a default, not a left_tt, if
     - E is an expression built on I/O signals, E is returned, bpos = pk_true
     - not E  and E is an expression built on I/O signals, E is returned, bpos = false
     - not Y, and Y is a signal, it returns the result of the method applied to Y, bpos = 
     - otherwise, it returns X

   The second parameter bpos is set to "true" when the result is the expression, "false" when it is the not of the result.
*/

ptexpcal ExprManager::getExprExportBooleanAbtraction (ptexpcal vsig, ptlinstances laddedSignals, boolean * bpos)
{
  ptlinstances
    lsave,
    ls,
    l;
  ptINode
    node;
  ptexpcal
    s,
    vexpdedef,
    vexp;
  ptinsth
    vh,
    vhs;
  noperators
    op;
  boolean
    b;

  *bpos = pk_true;
  if (IsSignal (vsig))
    {
      node = getNode (vsig);
      if (!NM->is_equation (node))
	return vsig;
      if (!NM->get_rightOperand (node, &vexp))
	return vsig;
      if (IsCellExpr (vexp))
	return vsig;
      vexpdedef = getBasicExpr (vexp);
      op = getOperator (vexpdedef);
      if ((op == opleft_tt) || (op == opmele))
	return vsig;
      b = pk_false;
      vh = CKM->get_insthClock (getContributionClock (vexp));
      if (!CKM->satisfiesInitClock (vh))
	if (getSynchronousExprHorl (vexpdedef, &vhs))
	  {
	    vhs = CKM->get_insthClock (vhs);
	    b = (vh == vhs);
	  }
      if (!b)
	return vsig;
      if (op == opnon)
	{
	  *bpos = pk_false;
	  vexpdedef = getUnaryArg (vexpdedef);
	  if (IsSignal (vexpdedef))
	    {
	      vexpdedef = getExprExportBooleanAbtraction (vexpdedef, laddedSignals, &b);
	      if (!b)
		*bpos = pk_true;
	      return vexpdedef;
	    }
	}
    }
  else
    vexpdedef = vsig;
  ls = getSignalsReferencedIn (vexpdedef);
  // If all signals referenced in ls are I/O, the definition of vsig is exported.
  b = pk_true;
  while (b && ptlinstances_get_next (&ls, &s))
    b = IsInputFlow (s) || IsOutputFlow (s) || ptlinstances_isin_exp (laddedSignals, s);
  if (!b)
    return vsig;
  return vexpdedef;
}


/**
   @return "true" if a signal is unified with a clock (clk), "false" otherwise.
*/
boolean ExprManager::isSignalUnifiedWithClock (ptexpcal aSig, ptinsth * clk)
{
  return getAttribute (aSig, CLKSIGUNIFICATION, (attributeValue *) clk);
}


/**
   @return the identifier associated with a boolean expression (exp) according the "naming in boolean expressions" rules.
*/
ptident ExprManager::getNameBooleanExpression (ptexpcal exp)
{
  ptident
    id;
  noperators
    op;
  ptexpcal
    arg1;

  op = getOperator (exp);
  if (!getNameAssociatedWithOperator (op, &id))
    id = getPredicateId ();
  else if (op == opleft_tt)
    {
      arg1 = getBinaryArgstAt (exp, 1);
      if (IsSignal (arg1))
	id = makeIdent_ident_ident (id, getOriginalName (arg1));
    }

  aNameManager->insertInIdentNameTable (id);	// to always add an suffix for the created identifier.
  return id;
}


/**
   @return "true" if vExpr is the expression "vsig $1 ", "false" otherwise.
*/
boolean ExprManager::isDelayOneOf (ptexpcal vExpr, ptexpcal vsig, ptexpcal * vinit)
{
  ptINode delayNode;
  ptexpcal resret, sigret, vret, vfen;
  compilerInteger valDelay;

  if (!IsSignal (vExpr))
    return pk_false;
  delayNode = getNode (vExpr);
  if (!NM->is_delay (delayNode))
    return pk_false;
  NM->get_delay_arg (delayNode, &resret, &sigret, &vret, &vfen, vinit);
  if (sigret != vsig)
    return pk_false;
  if (!getIntegerValue (vret, &valDelay))
    return pk_false;
  return (valDelay == 1);
}


/**
   @return the list composed of the elements of ls, declared as input/output.
*/
ptlinstances ExprManager::getIOflows (ptlinstances ls)
{
  ptlinstances
    lsio;
  ptexpcal
    aSig;

  lsio = NILLINSTANCES;
  while (ptlinstances_get_next (&ls, &aSig))
    if (IsInputFlow (aSig) || IsOutputFlow (aSig))
      lsio = ptlinstances_insert (lsio, aSig);
  return lsio;
}



/**
   @return "true" if v1 and v2 are not equal integer constant value, or not equal enumetared value, "false" otherwise.
 */
boolean
ExprManager::TestExclusiveConstants (ptexpcal v1, ptexpcal v2)
{
  cloper clexp1, clexp2;
  noperators op1, op2;

  if (v1 == v2)
    return pk_false;
  clexp1 = getOperatorClass (v1);
  clexp2 = getOperatorClass (v2);

  if (clexp1 != clexp2)
    return pk_false;
  if (clexp1 == constant)
    return (getConstant (v1) != getConstant (v2));
  op1 = getOperator (v1);
  op2 = getOperator (v2);
  if (op1 != op2)
    return pk_false;
  return (op1 == opvenum);
}

boolean
ExprManager::TestExclusiveConditions (ptexpcal cond1, boolean bpos1, ptexpcal cond2, boolean bpos2)
{
  noperators op1, op2;
  ptexpcal v11, v12, v21, v22;

  if (cond1 == NILEXPCAL)
    return pk_false;
  if (cond2 == NILEXPCAL)
    return pk_false;
  if (cond1 == cond2)
    return (bpos1 != bpos2);	// [c], [not c]

  if (bpos1 != bpos2)
    return pk_false;		// restriction.

  // test on enum value.
  op1 = getOperator (cond1);
  op2 = getOperator (cond2);

  if (op1 != op2)
    return pk_false;		// restriction.
  if ((op1 != opegal) && (op1 != opegeg))
    return pk_false;		// restriction.
  // opegal opegeg
  getBinaryArgs (cond1, &v11, &v12);
  getBinaryArgs (cond2, &v21, &v22);
  if (v11 == v21)
    return TestExclusiveConstants (v12, v22);
  if (v11 == v22)
    return TestExclusiveConstants (v12, v21);
  return pk_false;
}

/**
   @return the declared type of a signal(vsig), or undefined type if vsig is not an identifier.
 */
TypeType
ExprManager::getDeclaredType (ptexpcal vsig)
{
  return ptexpcal_get_declared_type (vsig);
}


/**
   Code Generation.
   @return the expression to generate associated with exp.
   The returned expression can be different from exp when some compiler attributes are associated 
   with the expression (OPTIMIZED_ATTRIBUTE, OPTIMIZED_MEMORY_ATTRIBUTE). It also looks at 
   the clone (getCloneClassHead()) and the aggregation getAggregation()).
*/
ptexpcal ExprManager::getExpressionToGenerate (ptexpcal exp)
{
  ptexpcal exp1;

  if (exp == NILEXPCAL) return exp;
  exp = getCloneClassHead (exp);
  // while (getAttribute(exp, OPTIMIZED_ATTRIBUTE, (attributeValue *) &exp1)) exp = exp1;
  if (getAttribute (exp, OPTIMIZED_ATTRIBUTE, (attributeValue *) & exp1))
    return getExpressionToGenerate (exp1);
  if (getAttribute (exp, OPTIMIZED_MEMORY_ATTRIBUTE, (attributeValue *) & exp1))
    return getExpressionToGenerate (exp1);
  exp1 = getAggregation (exp);
  if (exp1 != exp)
    return getExpressionToGenerate (exp1);
  return exp;
}
