/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


_PUBLIC_
/**
   Fields of structure management: for a signal of structured type, the compiler assigns a fictitious signal 
   to each field to have the same access (getting clock, getting node,...) than standart signal. 
   This method assigns a signal (s) to a reference of a field of a structure (e):
   for x.y, a signal s is created and associated with it. 
   \note
   - The link (e,s) is managed using an attribute called FIELD_SIGNAL.
   - When s is NILEXPCAL, the association of e with its "fictitious" signal is deleted.
*/
  void
ExprManager::setSignalAssignedTo (ptexpcal e, ptexpcal s)
{
  if (s == NILEXPCAL)
    removeAttribute (e, FIELD_SIGNAL);
  else
    addAttribute (e, FIELD_SIGNAL, (attributeValue) s);
}


/**
   Fields of structure management: it assigns a list of signals (ls) to a signal (e) for e a structured signal.
   A signal (u) of ls is the signal associated with a field of e (x.f) \see setSignalAssignedTo().
   \note the list (ls) is associated with e as an attribute called FIELDSLIST.
*/
void
ExprManager::setFieldsOf (ptexpcal e, ptlinstances ls)
{
  addAttribute (e, FIELDSLIST, (attributeValue) ls);
}


/**
   It associates the definition (node) to a signal (vsig) or an access to an element (x[]) of an array.
   For x[], the node is associated with x and an attribute is set for linking x[] to node (setNodeAccessToArrayElement()). 
   So, the node of x can evolved during the compiling (in array blocks for example).
*/
void
ExprManager::setNode (ptexpcal vsig, ptINode node)
{
  ptexpcal s;
  ptINode curnode;

  s = getDeclaredSignal (vsig);
  curnode = getNode (s);
  ptexpcal_set_node (s, node);
  if (s != vsig)
    setNodeAccessToArrayElement (vsig, node);
  else
    NM->dependencesMove (curnode, node);	// Added Oct 2014 LB.
}


/**
   It sets to an expression (exp) its type (vtyp).
*/
void
ExprManager::setType (ptexpcal exp, TypeType vtyp)
{
  ptexpcal_set_type (exp, (TypeType) vtyp);
}



/**
   @param exp (in) is a "opereh" expression (when/cell expression).
   @param h  (in) a clock.
   It assigns the clock h as second operand of the expression.
*/
void
ExprManager::setContributionClock (ptexpcal exp, ptinsth h)
{
  ptexpcal_set_ContributionClock (exp, h);
}


/**
   @param vsig (in) is a "variabl" expression (signal identifier).
   @param id (in) an identifier.
   It assigns the name (id) of a signal (vsig).
*/
void
ExprManager::setName (ptexpcal vsig, ptident id)
{
  ptexpcal vClone;
  boolean stop;

  vsig = getCloneClassHead (vsig);
  ptexpcal_set_identifier (vsig, id);
  do
    if (!getClone (vsig, &vClone))
      stop = pk_true;
    else if (isCloneClassHead (vClone))
      stop = pk_true;
    else
      {
	ptexpcal_set_identifier (vClone, id);
	stop = pk_false;
	vsig = vClone;
      }
  while (!stop);
}

/** 
    @param  exp (in) is a "variabl" expression (signal identifier).
    @param vdecl (in) a tree address (declaration of the signal in the abstract tree produced by the parser).
    It links a signal (exp) to its tree-declaration (vdecl).
*/
void
ExprManager::setTreeOfSignalDeclaration (ptexpcal exp, ntrees vdecl)
{
  ptexpcal_set_treeOfSignalDeclaration (exp, vdecl);
}


/** 
   It sets  the initial expression of a signal.
   @param  vsig (in) is a signal identifier (assumed).
   @param valini (in) initialization expression.
*/
void
ExprManager::setInitialValue (ptexpcal vsig, ptexpcal valini)
{
  if (!isNILEXPCAL (valini))
    {
      TypePropagate (ptexpcal_get_type (vsig), valini);
      ptexpcal_set_initExp (vsig, valini);
    }
}


/** 
   @param vexp (in) any signal expression.
   @param vhsub (in) the extraction (clock) built on vexp.
   It sets  the link clock-expression vhsub = [vexp].
 */
void
ExprManager::setExtractionLink (ptexpcal vexp, ptinsth vhsub)
{
  if (!isNILEXPCAL (vexp))
    ptexpcal_set_extraction (vexp, vhsub);
}


/**
   It defines the state variable attribute of an identifier by b.
   When b is pk_true, it specified that exp is a stqte variable. By default a signal IS NOT a state variable.
*/
void
ExprManager::setStateTag (ptexpcal exp, boolean b)
{
  if (IsSignal (exp))
    ptexpcal_set_stateTaq (exp, b);
}

/**
   It declared exp as a shared signal if b.
   By default a signal IS NOT a shared variable.
*/
void
ExprManager::setSharedSignal (ptexpcal exp, boolean b)
{
  if (IsSignal (exp))
    ptexpcal_set_sharedTaq (exp, b);
}

/**
   It defines a int signal(exp) as a constant signal with val as value.
*/
void
ExprManager::set_integerValueHolder (ptexpcal exp, long int val)
{
  ptexpcal_set_integerValueHolder (exp, val);
}

/**
   It assigns the bounds [lb, up] to an expression (exp).
*/
void
ExprManager::setBounds (ptexpcal exp, compilerInteger lb, compilerInteger up)
{
  ptexpcal_set_integerBoundsHolder (exp, lb, up);
}



/**  Indexed signals (array process)
     \DocNotComplete
*/
void
ExprManager::setException (ptexpcal exp, ptexpcal vexception)
{
  ptexpcal_set_exception (exp, vexception);
}

/** 
    @param exp (in) an identifier.
    It specifies that exp is a signal
    \note an identifier can be a signal, a parameter, an index ...
*/
void
ExprManager::belongsSignal (ptexpcal exp)
{
  ptexpcal_belongs_signal (exp);
}

/** @param exp (in) an identifier.
    It specifies that exp is an input signal. 
*/
void
ExprManager::belongsInFlow (ptexpcal exp)
{
  ptexpcal_belongs_InFlow (exp);
}

/** @param exp (in)  an identifier.
    It specifies that exp is an output signal.
*/
void
ExprManager::belongsOutFlow (ptexpcal exp)
{
  ptexpcal_belongs_OutFlow (exp);
}


/** @param exp (in)  an identifier.
    It specifies that the signal (exp) is a local signal.
*/
void
ExprManager::belongsLocalFlow (ptexpcal exp)
{
  ptexpcal_belongs_LocalFlow (exp);
}

/**
   A signal (pt) belongs a free (not defined, not assigned,...) signal.
*/
void
ExprManager::belongsFreeSignal (ptexpcal pt)
{
  ptexpcal_belongs_freeSignal (pt);
}


/**
   A signal (pt) belongs defined (but not valued).
*/
void
ExprManager::belongsDefined (ptexpcal exp)
{
  ptexpcal_belongsDefined (exp);
}

/** 
    It links an expression (n) with the original tree (ba) it comes from.
*/
void
ExprManager::setBackAnnotation (ptexpcal n, BackAnnotationType ba)
{
  ptexpcal_set_backAnnotation (n, ba);
}

/** 
    Setting travel indexes to an expression.
    @param exp (in) a signal expression.
    @param l (in) a list of indexes.
    \note \DocNotComplete (what is a "travel index"?)
*/
void
ExprManager::setTravelIndexes (ptexpcal exp, ptlinstances l)
{
  ptexpcal_set_travelIndexes (exp, l);
}

/**
   Setting the list of optimized travel indexes to an expression.
   @param exp (in) a signal expression.
   @param l (in) the list of indexes.
   \note \DocNotComplete (what is a "travel index"?, what is an optimized "travel index"?)
*/
void
ExprManager::setOptimizedTravelIndexes (ptexpcal exp, ptlinstances l)
{
  if (l != NILLINSTANCES)
    addAttribute (exp, OPTIMIZED_INDEX_LIST, (attributeValue) l);
}


/**
   It links a node (node) to an access to an element (s) of an array.
*/
void
ExprManager::setNodeAccessToArrayElement (ptexpcal s, ptINode node)
{
  addAttribute (s, AccessToArrayElement, (attributeValue) node);
}

/** Array management:
    It links (bidirectionnal) a clone (vClone) to a signal (v). (x[I] is a clone of x, x[I,J] a clone of x[I],...).
    An object (v) not yet cloned is the "clone class head" element.
    Note that it is a circular list with a header element.
    \see setCloneClassHead(), isCloneClassHead(), setClone().
*/
void
ExprManager::setClone (ptexpcal v, ptexpcal vClone)
{
  ptexpcal x;

  if (!getClone (v, &x))
    {
      setCloneClassHead (v);
      x = v;
    }
  addAttribute (vClone, CLONE_ATTRIBUTE, (attributeValue) x);
  addAttribute (v, CLONE_ATTRIBUTE, (attributeValue) vClone);
}


/**
   v is declared as the "clone class head" element of its class: 
    - x[I] is a clone of x, 
    - x[I,J] a clone of x[I],...),
    - an object (v) not yet cloned is the "clone class head" element (setCloneClassHead()).
*/
void
ExprManager::setCloneClassHead (ptexpcal v)
{
  addMark (v, CLONE_CLASS_HEAD);
}


/** 
    It assigns a clock (vclk) to a signal (vsig). No effect if vsig is not a signal.
*/
void
ExprManager::setClock (ptexpcal vsig, ptinsth vclk)
{
  ptexpcal_set_clock (vsig, vclk);
  if (!isStateVar (vsig) && !IsIndex (vsig) && !IsParameter (vsig) && !IsConstant (vsig))
    {
      CKM->setSynchronousSignals (vclk, ptlinstances_insert (CKM->getSynchronousSignals (vclk), vsig));
    }
}

/**
   If the name (id) of vsig is without suffix, and if it exists an another signal (vs) with the same prefix identifier than vsig, 
   then vs is recalled id.
   
   \note the removeIdent methode must be applied before calling this method. It must be used when a signal (vsig) is removed from the graph.

   \see{removeIdent}
*/
void
ExprManager::changeNameSignalSamePrefix (ptexpcal vsig)
{
  ptlinstances lident;
  ptident idPrefix, id;
  boolean found;
  ptexpcal vClone, vs;

  idPrefix = getOriginalName (vsig);
  if (eq_idents (idPrefix, getIdentNil ()))
    return;
  id = getName (vsig);
  if (!eq_idents (id, idPrefix))
    return;
  if (!getExprList (opident, &lident))
    return;
  found = pk_false;
  while (!found && ptlinstances_get_next (&lident, &vs))
    found = (eq_idents (getOriginalName (vs), idPrefix));

  if (found)
    if (aNameManager->insertInIdentNameTable (idPrefix))
      setName (vs, idPrefix);	// It updates also the clones.
}

/**
   @param vexp (in) any signal expression.
   @param nexp (in) any signal expression.
   It sets an indirection from vexp to nexp.
   \note MISC.
*/
void
ExprManager::setIndirection (ptexpcal vexp, ptexpcal nexp)
{
  ptexpcal_set_indir (vexp, nexp);
}


/** 
    It sets the identifier assigned to a signal (aSig).
    If a signal (aSig) is not a declared signal (i.e. it has been added by the compiler),
    and if codeOp is the delay operator (opret), the name of aSig is defined by 
    \e pre_X where \e X denotes the identifier associated with vfrom. The prefix (\e pre) is given by getDelayId(). 

    @return "true" if the operation succeeds, "false" otherwise.
    \note
    - MISC operation.
    - Restriction: implemented only for delay.
*/
boolean ExprManager::changeUndefinedName (ptexpcal aSig, ptexpcal vfrom, noperators codeOp)
{
  ptident
    name;

  /* Currently : only for delays */
  if (codeOp != opret)
    return pk_false;
  if (getOriginalName (aSig) != getIdentNil ())
    return pk_false;
  if (!IsSignal (vfrom))
    return pk_false;
  name = newIdent_concIdentIdent (getDelayId (), getOriginalName (vfrom));
  mkSignalName (aSig, name, pk_true);
  return pk_true;
}


/**
   The name of a signal (vsig) is defined by the name of the original identifier on which is is built.
   This name is propagated to its clones.
*/
void
ExprManager::setOriginalName (ptexpcal vsig)
{
  setName (vsig, ptexpcal_get_treeIdentifier (getCloneClassHead (vsig)));
}

/** 
    The operand identifier (vsig) is a Signal added by the Signal ToolBox.
*/
void
ExprManager::setAuxiliarySignal (ptexpcal vsig)
{
  addMark (vsig, AUXILIARY_IDENT);
}

/**
   It sets that a signal (aSig) is unified with a clock (clk).
   If clk is the predefined undefined clock, the association is removed.
*/
void
ExprManager::setSignalUnifiedWithClock (ptexpcal aSig, ptinsth clk)
{
  if (CKM->eqClock (clk, hundef))
    removeAttribute (aSig, CLKSIGUNIFICATION);
  else
    addAttribute (aSig, CLKSIGUNIFICATION, (attributeValue) clk);
}

/**
   It assigns to a signal (vsig)  the clock (vh) at which the signal 
   is not defined by  the (assumed) operand constant value (vExpr).
*/
void
ExprManager::setDefaultConstantValueInfos (ptexpcal vsig, ptinsth vh, ptexpcal vExpr)
{
  addAttribute (vsig, ASSIGNMENTCLOCKPART, vh);
  addAttribute (vsig, DEFAULCONSTANTVALUE, vExpr);
  NM->setDefaultConstantValueInfos (getNode (vsig), vh, vExpr);
}


/**
   It defines the assignment clock of a signal (exp) by assClock.
*/
void
ExprManager::setAssignmentClock (ptexpcal exp, ptinsth assClock)
{
  addAttribute (exp, ASSIGNMENTCLOCK, (attributeValue) assClock);
  NM->setAssignmentClock (getNode (exp), assClock);
}


/**
   It defines the utility clock of a signal (exp) by hu.
*/
void
ExprManager::setUtilityClock (ptexpcal exp, ptinsth hu)
{
  addAttribute (exp, UTILITYCLOCK, (attributeValue) hu);
  NM->setUtilityClock (getNode (exp), hu);
}


/**
   It defines the utility clock of a signal (exp) in a subgraph (sg) by hu.
*/
void
ExprManager::setUtilityClockInSubGraph (ptexpcal exp, ptINode sg, ptinsth hu)
{
  NM->setUtilityClockInSubGraph (sg, exp, hu);
}


/**
   A signal (vsig) is marked as a shared signal by several graphs if b, otherwise it is unmarked.
*/
void
ExprManager::setSharedSignalBySeveralGraphs (ptexpcal vsig, boolean b)
{
  ptinsth h;
  ptexpcal vaux;

  if (b)
    {
      addMark (vsig, SHARED_BY_SEVERAL_GRAPHS);
      if (isSignalUnifiedWithClock (vsig, &h))
	{
	  CKM->setClockUnifiedWithSignal (h, NILEXPCAL);
	  setSignalUnifiedWithClock (vsig, hundef);
	}
      /* else
         {
         h = getClock(vsig);
         if (CKM-> isClockRepresentedBySignal(h, &vaux) && (vaux == vsig))
         CKM->setClockRepresentedBySignal(h, NILEXPCAL);
         } */
    }
  else
    removeMark (vsig, SHARED_BY_SEVERAL_GRAPHS);
}
