/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   It builts and assigns the name of a signal.
   @param vexpcal (in) must be an identifier expression.
   @param r (in) tree-identifier prefix for the name construction.
   @param appendSuffix (in) Tag for ident manipulation.
   @return the tree-identifier obtained :
   - by concatenating r and the number of vexpcal when appendSuffix
   - by substituing the suffix of r by the number of vexpcal when not appendSuffix
*/
ptident ExprManager::mkSignalName (ptexpcal vexpcal, ptident r, boolean appendSuffix)
{  
  ptident id;
  int numero;

  nbIdent++;
  id = r;
  if (id != NILIDENT)
    {
      if (IsAuxiliarySignal (vexpcal))
	ptexpcal_set_treeIdentifier (vexpcal, id);
      id = getSignalIdentifier (id);
      if (!aNameManager->insertInIdentNameTable (id))
	{
	  numero = ptexpcal_get_num (vexpcal);
	  if (appendSuffix)
	    id = newIdent_intSuffixed2 (id, numero);
	  else
	    id = identChangeSuffix (id, numero);
	  id = aNameManager->mkIdentName (id);
	}
    }
  ptexpcal_set_identifier (vexpcal, id);
  return id;
}

/** @return the expression of a type reference (effective parameter for a call) */
ptexpcal
ExprManager::mkExpType (TypeType t)
{
  ptexpcal exp;

  if (bunifexp)
    if (isinTypeExprSet (t, &exp))
      {
	setType (exp, t);
	return exp;
      }
  exp = ptexpcal_RIEN_create_type (t);
  setType (exp, t);
  addintabexp (optype, exp);
  return exp;
}

/** @return the expression of a call reference to an external model 
    characterized by 
    - its name (id),
    - some parameters (lparams),
    - some inputs (ins), 
    - its clock (vh) 
    - and its result type (typeRes).
*/
ptexpcal
ExprManager::mkCallExp (ptident id, ptexpcal lparams, ptexpcal ins, TypeType typeRes, ptinsth vh)
{
  ptexpcal exp;

  if (bunifexp)
    if (isinCallExprSet (id, lparams, ins, &exp))
      return exp;
  exp = ptexpcal_RIEN_create_funcCall (id, lparams, ins);
  ptexpcal_set_clock (exp, vh);
  setType (exp, typeRes);
  addintabexp (opappe, exp);
  SpatialFuncCall (exp, ins);
  return exp;
}


/**
   @return the expression which denotes a structure/bundle field identifier.
   \see{mkIdentExp} for the description of the parameter.
*/
ptexpcal
ExprManager::mkIdentField (ptident r, TypeType TT, ptinsth clk)
{
  return newIdent (r, NILTREE, tfield, TT, NM->mk_node_undefined (), clk);
}

/**
   It creates a new expression which denotes an identifier (instance of a flow).
   @param r (in) tree-identifier used as prefix to build the name of the instance.
   @param ti (in) specialisation of the identifier as a signal, constant, ...
   @param TT  (in) type of the signal.
   @param clk clock of the signal (it can be "hundef" value)
   @param node (in) node of the signal (can be "opnil" one)
   @param ar (in) abstract tree of the declaration of the flow(parser information).
   @return the created identifier.
*/
ptexpcal
ExprManager::mkIdentExp (ptident r, typinst ti, TypeType TT, ptINode node, ptinsth clk, BackAnnotationType ar)
{
  if (r == NILIDENT) r = getIdentNil ();
  return newIdent (r, BACK_ANNOT_GET_ABSTRACT_TREE(ar), ti, TT, node, clk);
}

/** 
    Similar to mkIdentExp(ptident, typinst, TypeType, ptINode, ptinsth, ntrees).
    Moreover, the signal is initialized by vinit expression.
*/
ptexpcal
ExprManager::mkIdentExp (ptident r, typinst ti, TypeType TT, ptexpcal vinit, ptINode node, ptinsth clk, BackAnnotationType ar)
{
  ptexpcal vexpcal;

  vexpcal = mkIdentExp (r, ti, TT, node, clk, ar);
  ptexpcal_set_initExp (vexpcal, vinit);
  return vexpcal;
}

/**  It creates a new expression which denotes an Index identifier.
     @param id (in) tree-identifier used as prefix to build the name of the instance.
     @param uBound (in) the max value of the index (NILEXPCAL if unkown).
     @param node (in) node of the signal (can be "opnil" one)
     @param clock (in) clock of the signal (can be "hundef" value)
     @param vt (in) abstract tree of the declaration of the flow(parser information)
     @return the created identifier.
*/
ptexpcal
ExprManager::mkIndexIdent (ptident id, ptexpcal uBound, ptINode node, ptinsth clock, BackAnnotationType vt)
{
  ptexpcal vexpcal;

  if (id == NILIDENT) id = getIndexId ();
  vexpcal = mkIdentExp (id, tindex, get_IntegerType (), node, clock, NILBACKANNOTATION);
  if (uBound != NILEXPCAL) setIndexBounds (vexpcal, uBound, vt);
  return vexpcal;
}

/**  It creates a new expression which denotes an Label identifier.
     @param id (in) tree-identifier used as prefix to build the name of the instance.
     @param node (in) node of the signal (can be "opnil" one)
     @param clock (in) clock of the signal (can be "hundef" value)
     @param vt (in) abstract tree of the declaration of the flow(parser information)
     @return the created identifier.
*/
ptexpcal
ExprManager::mkLabelIdent (ptident id, ptINode node, ptinsth clock, BackAnnotationType vt, boolean badded)
{
  if (badded) aNameManager->insertInIdentNameTable (id);
  return mkIdentExp (id, tlabel, get_LabelType (), node, clock, vt);
}


/**
   @return a new signal, copy of a model (vexpcal), characterized by a node (vinstr), and the same name than the model.
   Moreover, the signals (vexpcal and the copy) are chained (see setClone()) to keep the link between clones. It is used for the compiling of arrays
   to keep the link x, x[], x[][],...
*/
ptexpcal
ExprManager::mkIdentClone (ptexpcal vexpcal, ptINode vinstr)
{
  ptexpcal vcopy;
  ptlinstances ls;

  vcopy = ptexpcal_cloneIdent (vexpcal, vinstr);
  NM->set_leftOperand (vinstr, vcopy);
  setClone (vexpcal, vcopy);
  ptexpcal_set_identifier (vcopy, ptexpcal_get_identifier (vexpcal));
  return vcopy;
}

/**
   @return a new signal, copy of a model (vexpcal), characterized by a node (vinstr) and a clock (vclk). 
   A new name is assigned to the created signal if bkeepname, otherwise the same name is used.
*/
ptexpcal
ExprManager::mkIdentCopy (ptexpcal vexpcal, ptINode vinstr, ptinsth vclk, boolean bkeepname)
{
  ptexpcal vcopy;
  ptident id;
  ptlinstances ls;

  vcopy = atcpinstidexpcal (vexpcal, vinstr, vclk);
  if ( ! ptexpcal_IsLabel (vexpcal))  NM->set_leftOperand (vinstr, vcopy);
  id = ptexpcal_get_identifier (vexpcal);
  if (bkeepname)
    ptexpcal_set_identifier (vcopy, id);
  else
    {
      boolean b = (ptexpcal_get_treeIdentifier(vexpcal) == id);
      mkSignalName (vcopy, id, b /* pk_false*/);
    }
  addintabexp (opident, vcopy);
  return vcopy;
}

/**
   @return a new signal, copy of a model (vexpcal), characterized by a node (vinstr) and a clock (vclk). 
   A new name is assigned to the created signal.
*/
ptexpcal
ExprManager::mkIdentCopy (ptexpcal vexpcal, ptINode vinstr, ptinsth vclk)
{
  return mkIdentCopy (vexpcal, vinstr, vclk, pk_false);
}


/**
   Builting a constant nil expression.
   @param cst  (in) value of the constant (string memory value)
   @return the representation of the constant, (constant operator Class).
*/
ptexpcal
ExprManager::mk_nilPtexpcal (ptrepstr cst)
{
  ptexpcal vexp;

  if (mkesigconst (opnil, cst, &vexp));
  return vexp;
}

/**
   Builting a constant name expression.
   @param cst  (in) value of the constant (string memory value)
   @return the representation of the constant, (constant operator Class).
*/
ptexpcal
ExprManager::mk_namePtexpcal (ptrepstr cst)
{
  ptexpcal vexp;

  if (mkesigconst (opnom, cst, &vexp));
  return vexp;
}

/**
   Builting a constant enum value string.
   @param cst  (in) value of the constant (string memory value)
   @return the representation of the constant, (constant operator Class).
*/
ptexpcal
ExprManager::mk_enumPtexpcal (ptrepstr cst)
{
  ptexpcal vexp;

  if (mkesigconst (opcstenum, cst, &vexp));
  return vexp;
}

/**
   Builting a constant char value string.
   @param cst  (in) value of the constant (string memory value)
   @return the representation of the constant, (constant operator Class).
*/
ptexpcal
ExprManager::mk_charPtexpcal (ptrepstr cst)
{
  ptexpcal vexp;

  if (!mkesigconst (opcstcar, cst, &vexp))
    return vexp;
  setType (vexp, get_CharType ());
  return vexp;
}

/**
   Builting a constant string value string.
   @param cst  (in) value of the constant (string memory value)
   @return the representation of the constant, (constant operator Class).
*/
ptexpcal
ExprManager::mk_stringPtexpcal (ptrepstr cst)
{
  ptexpcal vexp;

  if (!mkesigconst (opcstcha, cst, &vexp))
    return vexp;
  setType (vexp, get_StringType ());
  return vexp;
}

/**
   Builting a Constant integer value string.
   @param op (in) signal constant operator
   @param cst  (in) value of the constant (string memory value)
   @return the representation of the constant, (constant operator Class).
*/
ptexpcal
ExprManager::mk_integerPtexpcal (ptrepstr cst)
{
  return ptexpcal_provideStringIntegerConstant (cst);
}

/**
   Builting a Constant real value string
   @param op (in) signal constant operator
   @param cst  (in) value of the constant (string memory value)
   @return the representation of the constant, (constant operator Class).
*/
ptexpcal
ExprManager::mk_realPtexpcal (ptrepstr cst)
{
  ptexpcal vexp;

  if (!mkesigconst (opcstreel, cst, &vexp))
    return vexp;
  setType (vexp, get_RealType ());
  return vexp;
}

/**
   Builting a Constant double precision real value string.
   @param op (in) signal constant operator
   @param cst  (in) value of the constant (string memory value)
   @return the representation of the constant, (constant operator Class).
*/
ptexpcal
ExprManager::mk_dprealPtexpcal (ptrepstr cst)
{
  ptexpcal vexp;

  if (!mkesigconst (opcstdp, cst, &vexp))
    return vexp;
  setType (vexp, get_DRealType ());
  return vexp;
}

/**
   Builting a Constant expression.
   @param op (in) signal constant operator
   @param cst  (in) value of the constant (string memory value)
   @return the representation of the constant, (constant operator Class).
*/
ptexpcal
ExprManager::mkConstantExp (noperators op, ptrepstr cst)
{
  if (op == opnil)
    return mk_nilPtexpcal (cst); 
  if (op == opnom)
    return mk_namePtexpcal (cst);
  if (op == optrue)
    return ptexpcal_getTrue ();
  if (op == opfalse)
    return ptexpcal_getFalse ();
  if (op == opcstenum)
    return mk_enumPtexpcal (cst);
  if (op == opcstcar)
    return mk_charPtexpcal (cst);
  if (op == opcstcha)
    return mk_stringPtexpcal (cst);
  if (op == opcstent)
    return ptexpcal_provideStringIntegerConstant (cst);
  if (op == opcstreel)
    return mk_realPtexpcal (cst);
  if (op == opcstdp)
    return mk_dprealPtexpcal (cst);
  printf ("ERROR in ExprManager::mkConstantExp code op %i is not a correct operator\n", (int) op);
  fflush (stdout);
  return NILEXPCAL;
}

/**
   Builting an int Constant expression.
   @param val (in) value of the constant.
   @return the representation of the constant (constant operator Class)
*/
ptexpcal
ExprManager::mkIntegerConstant (compilerInteger val)
{
  if (val >= 0) return ptexpcal_provideIntegerConstant (val);
  return mkUnaryExp (opnega, ptexpcal_provideIntegerConstant (-val), NILBACKANNOTATION);
}

/**
   Builting a Synchronous Unary expression.
   @param op (in) Signal Unary operator (assumed).
   @param e1 (in) child of the Unary expression.
   @param vt (in) abstract tree of the declaration of the flow(parser information)
   @return Unary expression op(e1), (oper1 operator Class)
*/
ptexpcal
ExprManager::mkUnaryExp (noperators op, ptexpcal e1, BackAnnotationType /* ntrees*/  vt)
{
  ptexpcal vexp;

  if (bunifexp) if (isinUnaryExprSet (op, e1, &vexp)) return vexp;
  vexp = mkExp1 (op, e1, vt);
  setBackAnnotation (vexp, vt);
  SpatialUnary (vexp, op, e1);
  return vexp;
}



/**
   Builting a Synchronous binary expression.
   @param op (in) Signal binary operator (assumed).
   @param e1,e2 (in) children of the Binary expression.
   @param vt (in) abstract tree of the declaration of the flow(parser information)
   @return Binary expression op(e1, e2), (oper2 operator Class)
*/
ptexpcal
ExprManager::mkBinaryExp (noperators op, ptexpcal e1, ptexpcal e2, BackAnnotationType /* ntrees*/ vt)
{
  ptexpcal vexp;

  if (op == opegal) if (!verifyScalarType (e1)) op = opegeg;
  if (op == opinfe) if (!verifyScalarType (e1)) op = opiieg;
  if (LL_is_listOp (op)) return mkNaryExp (op, e1, e2, vt);
  if (bunifexp) if (isinBinExprSet (op, e1, e2, &vexp)) return vexp;
  vexp = mkExp2 (op, e1, e2, vt);
  setBackAnnotation (vexp, vt);
  SpatialBinary (vexp, op, e1, e2);
  return vexp;
}

/**
   Builting a Synchronous Ternary expression. 
   @param op (in) Signal ternary operator (assumed).
   @param e1,e2,e3 (in) children of the ternary expression.
   @param vt (in) abstract tree of the declaration of the flow(parser information)
   @return Ternary expression op(e1, e2, e3) (oper3 operator Class).
*/
ptexpcal
ExprManager::mkTernaryExp (noperators op, ptexpcal e1, ptexpcal e2, ptexpcal e3, BackAnnotationType vt)
{
  ptexpcal vexp;
  if (bunifexp) if (isinTerExprSet (op, e1, e2, e3, &vexp)) return vexp;
  vexp = mkExp3 (op, e1, e2, e3, vt);
  setBackAnnotation (vexp, vt);
  SpatialTernary (vexp, op, e1, e2, e3);
  return vexp;
}


/**
   Builting a Synchronous list expression. Implemented as a combed binary expression.
   @param op (in) Signal list operator (assumed).
   @param e1 (in) first child of the list.
   @param e2 (in) tail of the list.
   @param vt (in) abstract tree of the declaration of the flow(parser information)
   @return list expression op(e1, e2), (oper2 operator Class).
*/
ptexpcal
ExprManager::mkNaryExp (noperators op, ptexpcal e1, ptexpcal e2, BackAnnotationType vt)
{
  ptexpcal vexp;
  if (bunifexp) if (isinListSet (op, e1, e2, &vexp)) return vexp;
  vexp = mkExpN (op, e1, e2, vt);
  setBackAnnotation (vexp, vt);
  SpatialNary (vexp, op, e1, e2);
  return vexp;
}

/**
   Builting a list of expressions reduced to one element (e1).
   @param op (in) Signal list operator (assumed).
   @param e1 (in) first child of the list.
   @param vt (in) abstract tree of the declaration of the flow(parser information)
   @return the list expression.
*/
ptexpcal
ExprManager::mkNaryExp (noperators op, ptexpcal e1, BackAnnotationType vt)
{
  return mkNaryExp (op, e1, NILEXPCAL, vt);
}

/**
   Concatenating two lists of expressions: the elements of a list (l1) are added before the elements of a second list (l2).
   When l2 is empty, l1 is duplicated !!!.
   @param l1 (in) a list of expressions (assumed)
   @param l2 (in) a list of expressions (assumed)
   @param vt (in) abstract tree of the declaration of the flow(parser information).
*/
ptexpcal
ExprManager::concatenate (ptexpcal l1, ptexpcal l2, BackAnnotationType vt)
{
  if (l1 == NILEXPCAL) return l2;
  return fusionlist (getOperator (l1), l1, l2, vt);
}

/**
   Builting a temporal binary expression (default). 
   @param op (in) Signal temporal binary operator (in fact default operator).
   @param e1 (in) first operand of the resulting expression
   @param e2 (in) second operand of the resulting expression
   @param vt (in) abstract tree of the declaration of the flow(parser information)
   @return expression op(e1, e2) (oper2tp operator Class), but the contribution clock is not set (see mkClockedDefault()).
   \note {The expressions e1 and e2  MUST be "when" expressions.}
*/
ptexpcal
ExprManager::mkBinaryExpTp (noperators op, ptexpcal e1, ptexpcal e2, BackAnnotationType vt)
{
  ptexpcal vexp;

  if (bunifexp) if (isinBinTpExprSet (op, e1, e2, &vexp)) return vexp;
  vexp = mkExpTp2 (op, e1, e2, vt);
  setBackAnnotation (vexp, vt);
  SpatialBinaryTp (vexp, op, e1, e2);
  return vexp;
}


/** @return the expression denoting vexp1 \b default vexptoadd and the contribution clock is assigned to the expression.
    It is assumed that
    - vexp1 is a normalized default with a contribution clock
    - vexpToadd is a normalized when expression.
*/
ptexpcal
ExprManager::mkClockedDefault (ptexpcal vexp1, ptexpcal vexpToadd)
{
  ptinsth hcontr, vh;
  ptexpcal vexp;

  hcontr = getContributionClock (vexp1);
  vh = getContributionClock (vexpToadd);
  if (CKM->satisfiesUndefClock (hcontr))
    hcontr = vh;
  else
    hcontr = CKM->makeSup (hcontr, vh);
  vexp = mkBinaryExpTp (opmele, vexp1, vexpToadd, NILBACKANNOTATION);
  setContributionClock (vexp, hcontr);
  return vexp;
}


/**
   Builting a temporal Unary expression (event, unary when).
   @param op (in) Signal temporal unary operator (assumed).
   @param vexp1 (in) child of the expression.
   @param vt (in) abstract tree of the declaration of the expression (parser information).
   @return expression op(e1), (oper1tp operator Class)

   \note Currently, unused by the compiler.
*/
ptexpcal
ExprManager::mkUnaryExpTp (noperators op, ptexpcal vexp1, BackAnnotationType /* ntrees*/ vt)
{
  ptexpcal vexp;

  if (bunifexp)
    if (isinUnaryExprSet (op, vexp1, &vexp))
      return vexp;
  vexp = mkesigexp1 (op, oper1tp, vexp1);
  setBackAnnotation (vexp, vt);
  SpatialUnaryTp (vexp, op, vexp1);
  return vexp;
}

/**
   Builting a temporal binary expression (when, cell).
   @param op (in) a signal operator; assumed to be the "when" operator or the "cell" operator.
   @param e  (in) left expression of the cell/when expression
   @param ck (in) contribution clock assigned to e
   @return op(e1, ck) (opereh operator Class)
*/
ptexpcal
ExprManager::mkWhenOrCellExp (noperators op, ptexpcal e, ptinsth ck)
{
  ptexpcal vexp;

  if (bunifexp) if (isinWhenCellSet (op, e, ck, &vexp)) return vexp;
  vexp = mkExtractOrMemory (op, e, ck);
  SpatialExtractOrMemory (vexp, op, e, ck);
  return vexp;
}


/** 
    @param vsig  (in) a signal identifier expression (assumed)
    @param nbElements (in) is the number of elements expected for the initializing.
    @param vind (out) new index identifier used to defined the initialisation expression. NIL expression can be returned.
    @param vinit (out) initial value defined fot the signal vsig. This expression is the default value 
    of its type (see Signal reference manual).
    @return "true" if the initial value is defined, false otherwise.
*/
boolean
ExprManager::mkDefaultInit (ptexpcal vsig, long int nbElements, ptexpcal * vind, ptexpcal * vinit)
{
  TypeType TT, TTbase;
  ptexpcal exp1, vexpIndex, velevect, aParc, bsup, vindice, lparc, born;

  /* Restriction: predef, array, enum */
  *vind = NILEXPCAL;
  *vinit = NILEXPCAL;
  TT = ptexpcal_getUnamedType (vsig);
  if (TypeTestPredef (TT))
    {
      exp1 = getDefaultInit (TT);
      if (nbElements == 1)
	{
	  *vinit = exp1;
	  return pk_true;
	}
    }
  else
    {
      if (TypeTestArrayClass (TT))
	{
	  TTbase = ArrayMatrixTypeGetTerm (TT);	// if ( TTbase == CharType) { *vinit = getDefaultInit(StringType); return pk_true; }
	}
      else
	{
	  if (!TypeTestEnumerationClass (TT))
	    return pk_false;
	  *vinit = getDefaultInit (TT);
	  return pk_true;
	}
      exp1 = getDefaultInit (TTbase);
      // Il faudrait "construire" les parcours pour les dim de TT
    }
  if (exp1 == NILEXPCAL) return pk_false;
  
  if (nbElements == 1)
    {
      vexpIndex = mkNaryExp (oplesign, mkIntegerConstant (0), NILEXPCAL, NILBACKANNOTATION);
      velevect = mkBinaryExp (opelevec, vexpIndex, exp1, NILBACKANNOTATION);
    }
  else
    {
      bsup = mkIntegerConstant (nbElements - 1);
      if (!getIndexForInit (&vindice, 1))
	{
	  vindice = mkIndexIdent (getIndexId (), bsup, NM->mk_node_undefined (), hundef, NILBACKANNOTATION);
	  addIndexForInit (vindice);
	}
      born = mkTernaryExp (opborn, mkIntegerConstant (0), bsup, mkIntegerConstant (1), NILBACKANNOTATION);
      aParc = mkBinaryExp (opparc, vindice, born, NILBACKANNOTATION);
      lparc = mkNaryExp (oplparc, aParc, NILEXPCAL, NILBACKANNOTATION);
      *vind = vindice;
      velevect = mkBinaryExp (opensvec, lparc, exp1, NILBACKANNOTATION);
    }
  *vinit = mkNaryExp (oplelevec, velevect, NILEXPCAL, NILBACKANNOTATION);
  return pk_true;
}


/** 
    It removes an expression (e) of the expression manager.
    @return true if e was in the manager, false otherwise.
*/
boolean
ExprManager::remove (ptexpcal e)
{
  return rmintabexp (ptexpcal_get_operator (e), e);
}


/**
   It applies the distributivity of the when on the default.
   For e default expression ( (e1 when h1) default ... default (en when hn)) when h 
   it returns (e1 when h1 when h) default ... default (en when hn when h)
*/
ptexpcal ExprManager::distributivityDefaultClock (ptexpcal e, ptinsth h)
{
  ptexpcal arg1, arg2, narg1, narg2;
  ptinsth hc;

  if (getOperator (e) == opmele)
    {
      arg1 = getDefaultArgstAt (e, 1);
      narg1 = distributivityDefaultClock (arg1, h);
      arg2 = getDefaultArgstAt (e, 2);
      narg2 = distributivityDefaultClock (arg2, h);
      if ((arg1 == narg1) && (arg2 == narg2))
	{
	  return e;
	}
      return mkBinaryExpTp (opmele, narg1, narg2, getBackAnnotation (e));
    }
  arg1 = getWhenOrCellArg (e);
  hc = getContributionClock (e);
  return mkWhenOrCellExp (opextr, arg1, CKM->makeInf (h, hc));
}


ptinsth ExprManager::getSetAssignmentClock(ptexpcal vexp)
{
  ptinsth vassclk;
  if (getAttribute (vexp, ASSIGNMENTCLOCK, (attributeValue *) &vassclk))
      vassclk = CKM->getRepresentativeFormula (vassclk);  
  else
    if (IsEvent(vexp))
      vassclk = getClock(vexp);
    else
      vassclk = NM->ComputeAssignmentClock(getNode(vexp));
  setAssignmentClock(vexp, vassclk);
  return vassclk;
}

