/*
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url
   http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
*/

ptexpcal ExprManager::ptexpcal_FALSE = NILEXPCAL;	///< internal representation of "true" expression
ptexpcal ExprManager::ptexpcal_TRUE = NILEXPCAL;	///< internal representation of "true" expression
ptexpcal ExprManager::ptexpcal_0 = NILEXPCAL;	///< internal representation of "0" expression
ptexpcal ExprManager::ptexpcal_1 = NILEXPCAL;	///< internal representation of "1" expression

_PUBLIC_
/**
 */
  void
ExprManager::DEBUG_TRUE (char *mess)
{
  DEBUG_ptexpcal_TRUE (mess);
}

/** Expression manager constructor.
    The returned object must be initialized \see{initialize}.
    The following expressions are created:
    - "true": to get it, use getTrue()
    - "false" to get it, use getFalse()
    By default,
    - the unicity of the expressions is guaranteed: see setUnifyExpressionMode() to modify this parameter.
    - the type ckecking is applied for each created expression: see setModeTypeChecking() to modify this parameter.
*/
ExprManager::ExprManager ()
{
  int i;

  nbIdent = 0;
  modeNoEvents = pk_false;	// Event type is allowed.
  bunifexp = pk_true;
  lequ1exp = NILLEXPCAL;
  lequ2exp = NILLEXPCAL;
  lexpmodified = NILLEXPCAL;
  lexptotest = NILLEXPCAL;
  modeEval = pk_true;
  modeArithmRules = pk_false;	// arithmetic reductions are not applied.
  modeTypeCheckingOutOfBounds = pk_true;	// will be set to false for recovery operator (e1 \\ e2)
  lastTypeCheckingelem = -1;
  ptexpcal_init (&ptexpcal_FALSE, &ptexpcal_TRUE, &ptexpcal_0, &ptexpcal_1);
  for (i = 0; i < SIG_NB_OPERATORS; i++)
    {
      table[i] = NILLEXPCAL;
      tableEnd[i] = NILLEXPCAL;
    }
  addintabexp (optrue, ptexpcal_getTrue ());
  addintabexp (opfalse, ptexpcal_getFalse ());
  lModifiedNodes = NILLINODE;
  lIndexForInit = NILLINSTANCES;
}

/** Destructor  */
ExprManager::~ExprManager ()
{
  int i;
  ptlinstances lexp;
  ptexpcal vexp;

  ptlINode_free (lModifiedNodes);
  ptlinstances_free (lIndexForInit);
  for (i = 0; i < SIG_NB_OPERATORS; i++)
    {
      lexp = table[i];
      while (ptlinstances_get_next (&lexp, &vexp))
	{
	  ptexpcal_free (vexp);
	}
    }
  for (i = 0; i < SIG_NB_OPERATORS; i++)
    {
      ptlinstances_free (table[i]);
    }

}

/**
    It links the expression manager to
    - a node manager (nm),
    - a clock  manager (ckm),
    - a name manager (vNameManager)
    - and a traceability manager (tm)
*/
void
ExprManager::initialize (ptNodeManager nm, ptClockManager ckm, ptNameManager vNameManager, ptTraceabilityManager tm)
{
  aNameManager = vNameManager;
  NM = nm;
  CKM = ckm;
  hundef = CKM->getUndefClock ();
  TM = tm;
}

/**
   In an expression manager, unicity of syntactic expressions can be asked (b with the value "true") or not (b with the value "false").
   By default the property is guaranted.
*/
void
ExprManager::setUnifyExpressionMode (boolean b)
{
  bunifexp = b;
}


/**
   In an expression manager, the type checking can be asked (b with the value "true") or not (b "false").
   By default, it is applied.
   @return the value of the mode before the calling to the method.
*/
boolean ExprManager::setModeTypeChecking (boolean b)
{
  boolean
    x;
  x = modeEval;
  modeEval = b;
  return x;
}

/**
   In an expression manager, the type checking for the access to the elements of an array can
   be asked (b with the value "true") or not (b "false"). By default, it is applied.

   Moreover, the result of the last control is returned (vLastOutOfBoundControl),
   using the following encoding:
   - 0 the bounds are not controlable (default),
   - 1 the indexes are all in the bounds,
   - 2 at lest one of the indexes is out of the bounds.

   @return the value of the mode before the calling to the method.
*/
boolean ExprManager::setTypeCheckingOutOfBounds (boolean b, int *vLastOutOfBoundControl)
{
  boolean
    x;
  x = modeTypeCheckingOutOfBounds;
  modeTypeCheckingOutOfBounds = b;
  *vLastOutOfBoundControl = lastTypeCheckingelem;
  return x;
}

/**
   In an expression manager, the application of some arithmetic rules for reductions can be asked (b with the value "true")
   or not (b "false").
   By default, they are not applied.
   @return the value of the mode before the calling to the method.
*/
boolean ExprManager::setArithmeticRuleMode (boolean b)
{
  boolean x;
  x = modeArithmRules;
  modeArithmRules = b;
  return x;
}

/** Access to a typed-list of expressions of an expression manager.
    @param vop (in) a Signal operator.
    @param list (out) list of expressions with the vop operator.
    @return true if the output list is not empty.
*/
boolean ExprManager::getExprList (noperators vop, ptlexpcal * list)
{
  if (vop <= 0)
    {
      return pk_false;
    }
  if (vop > SIG_NB_OPERATORS)
    {
      return pk_false;
    }
  *list = table[vop];
  return ((*list) != NILLEXPCAL);
}


/** Predefined expression in an expression manager.
    @return the representation of the "false" value.
*/
ptexpcal ExprManager::getFalse ()
{
  return ptexpcal_getFalse ();
}


/** Predefined expression in an expression manager.
    @return the representation of the "true" value.
*/
ptexpcal ExprManager::getTrue ()
{
  return ptexpcal_getTrue ();
}

/** Predefined expression in an expression manager.
    @return the representation of the "false" value.
*/
ptexpcal ExprManager::get0 ()
{
  return ptexpcal_0;
}


/** Predefined expression in an expression manager.
    @return the representation of the "true" value.
*/
ptexpcal ExprManager::get1 ()
{
  return ptexpcal_1;
}

void
ExprManager::get_ptexpcalConstants (ptexpcal * expcalFalse, ptexpcal * expcalTrue, ptexpcal * expcal0, ptexpcal * expcal1)
{
  (*expcalFalse) = ptexpcal_getFalse ();
  (*expcalTrue) = ptexpcal_getTrue ();
  (*expcal0) = ptexpcal_0;
  (*expcal1) = ptexpcal_1;
}



/** Garbaging.
    It adds an expression (e) in the set of unused expressions.
*/
void
ExprManager::free (ptexpcal e)
{
  /* To be tested.
     ptexpcal x, y;
     if (e == NILEXPCAL) return;
     if (satisfiesList(e))
     { y = e; while (get_Nary(&e,&x)) { ptexpcal_free(y); y = e; } ptexpcal_free(y); }
     // else ptexpcal_free(e);
   */
}


/** Solving conflict names :
   - for enumeration type value : a name can be in two enumeration type declaration. It is
   allowed in ADA but not for C, C++.
   This method appends a suffix to a name when a conflict is detected.
   @return "true" if a name has been modified, "false" otherwise.
*/
boolean
ExprManager::enumNameConflicts ()
{
  ptlinstances ltypes;
  TypeType TT;
  ptexpcal elem;
  boolean b;
  boolean Goon;
  NameSubList NSubLst;
  NameList NL;
  ptrepstr velem, newvelem;
  ptident id, id1;

  b = pk_false;
  if (getUserTypes (&ltypes))
    {
      while (ptlinstances_get_next (&ltypes, &elem))
	{
	  TT = getExpType (elem);
	  if (TypeTestEnumerationClass (TT))
	    {
	      NL = EnumerationTypeGetNameLst (TT);
	      NSubLst = NameListCast2SubList (NL);
	      Goon = pk_true;
	      while (Goon)
		{
		  velem = NameSubListGetHead (NSubLst);
		  id = repstr2ident (velem);
		  id1 = aNameManager->mkIdentName (id);
		  if (!eq_idents (id, id1))
		    {
		      newvelem = ident2repstr (id1);
		      NameListChangeName (NL, velem, newvelem);
		      refTypeSubstitute (TT, velem, newvelem);
		      b = pk_true;
		    }
		  if (NameSubListTestLast (NSubLst))
		    {
		      Goon = pk_false;
		    }
		  else
		    {
		      NSubLst = NameSubListGetTail (NSubLst);
		    }
		}
	    }
	}
    }
  return b;
}


void
ExprManager::printTracebilityInfos (ptident graphname)
{
  ofstream ofile;
  ptrepstr vfilename;
  int i;
  ptlinstances lexp;
  ptexpcal vexp;
  ptident vid;

  vid = makeIdent_ident_string (graphname, "EM");
  if (sys_suffixedFilePathInUserOutputDir (ident2repstr (vid), "txt", &vfilename))
    {
      ofile.open (repstr2string (vfilename));
      for (i = 0; i < SIG_NB_OPERATORS; i++)
	{
	  lexp = table[i];
	  while (ptlinstances_get_next (&lexp, &vexp))
	    {
	      ofile << " Num Expression: " << getNum (vexp);
	      if ( IsSignal(vexp)) ofile << " IOdentifier = " << ident2string(getName(vexp));
	      BackAnnotationType ba = getBackAnnotation (vexp);
	      if (ba == NILBACKANNOTATION)
		{
		  ofile << " No BACKANNOTATION, oper = " << getOperator (vexp);
		}
	      else
		{
		  ntrees vAbsTree = ba->getAbsTree ();
		  ntrees vCallTree = ba->getCallTree ();
		  ofile << "( " << vAbsTree << ", " << vCallTree << " )" << " oper = " << getOperator (vexp);
		}
	      ofile << endl;
	    }
	}
      ofile.close ();
    }
}
