/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
   http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_
/**
   It substitutes (see expReplace()) an expression (e1) by an another one (e2), given a context clock (h).
   The clock is used only when e2 is a constant and h is not the undefined clock, in this case 'e1' is replaced by 'e2 when h'.
*/
    void
 ExprManager::substitute(ptexpcal e1, ptexpcal e2, ptinsth h)
{
  ptlinstances lref_e2;

  lref_e2 = ptexpcal_getSignalsReferencedIn(e2);
  // Rewriting if non-recursive term.
  if (!ptlinstances_isin_exp(lref_e2, e1))
    expReplace(e1, e2, h);
  ptlinstances_free(lref_e2);
  updateInterfaceModifiedNodes();
}

/**
   Updating the nodes that reference the expression e1.
   The reference to e1 is replaced by e2 and the node is added to the list of the modified nodes (lModifiedNodes).
*/
void ExprManager::updateNodes(ptexpcal e1, ptexpcal e2)
{
  ptlINode ln;
  ptINode node;
  boolean b;

  ln = ptexpcal_get_NodesFeedBy(e1);
  b = (e1 != e2);
  while (ptlINode_get_elem(&ln, &node))
    {
      if (b)
	NM->update(node, e1, e2);
      NM->set_interface(node, NILINTERFACE);
      ptlINode_insert_elem(&lModifiedNodes, node);
    }
}

/**
   It replaces an expression (e1) by an another one (e2), given a context clock (h). 
   If the two expressions are signal identifiers, there is a possible change of the name.

   The clock is used only when e2 is a constant and h is not the undefined clock, 
   in this case 'e1' is replaced by 'e2 when h'.

*/
void ExprManager::expReplace(ptexpcal e1, ptexpcal e2, ptinsth h)
{
  ptlinstances luse;
  ptexpcal use, e2prim;

  changeNameSignal(e1, e2);
  // Visiting the uses.
  luse = ptexpcal_get_luse(e1);
  while (ptlinstances_get_next(&luse, &use))
    {
      if (use != e2)
	{
	  e2prim = e2;
	  // in a list, a constant signal is replaced by ( value<constant> when clock<constant signal> )
	  if (satisfiesList(use) && !CKM->satisfiesUndefClock(h))
	    if (ptexpcal_getSignalsReferencedIn(e2) == NILLINSTANCES)
	      e2prim = mkWhenOrCellExp(opextr, e2, h);
	  if (rewrite(e1, e2prim, use))
	    rewriting();
	}
    }
  propagatePragmas(e1, e2);
  updateNodes(e1, e2);
  updateClockManager(e1, e2);
  updateClockRepresentedBySignal(e1, e2);
}

/** 
    It updates the clocks of all the expressions (when, cell) that reference clocks.
    It calls rewritingWhenExpr(), rewritingCellExpr().
*/
void ExprManager::rewritingExprEH()
{
  rewritingWhenExpr();
  rewritingCellExpr();
  updateInterfaceModifiedNodes();
}

/**
   It substitutes all the evaluated expressions of the manager by their values excepted the output signals
   and some specific cases.
   @return "true" if a substitution is applied, "false" otherwise.
*/
boolean ExprManager::substituteEvaluatedExpression()
{
  ptexpcal vexp, nexp;
  ptlinstances liste;
  int i;
  ptinsth h;
  boolean b;

  b = pk_false;
  for (i = 0; i < SIG_NB_OPERATORS; i++)
    if ((i != opmemo) && (i != opextr) && (i != opelem))	// Warning.
      {
	liste = table[i];
	while (ptlinstances_get_next(&liste, &vexp))
	  if (!ptexpcal_IsOutputFlow(vexp))
	    if (!ptexpcal_IsInputFlow(vexp))
	      if (getEquivConstant(vexp, &nexp, pk_true))
		{
		  if (ptexpcal_IsSignal(vexp))
		    h = ptexpcal_get_clock(vexp);
		  else
		    h = hundef;
		  substitute(vexp, nexp, h);
		  b = pk_true;
		  if ((i != opident) && (i != opcham))
		    setIndirection(vexp, nexp);
		}
      }
  return b;
}

/**
   It replaces TT#velem expression by TT#newelem in the expression manager.
   @return "true" is such an expression exists in the manager, "false" otherwise.
*/
boolean ExprManager::refTypeSubstitute(TypeType TT, ptrepstr velem, ptrepstr newvelem)
{
  ptexpcal vexp, exp1, exp2;
  ptlinstances liste;
  boolean b;

  b = pk_false;
  liste = table[opvenum];
  while (ptlinstances_get_next(&liste, &vexp))
    {
      getBinaryArgs(vexp, &exp1, &exp2);
      if ((getExpType(exp1) == TT) && (ptexpcal_get_constant(exp2) == velem))
	{
	  ptexpcal_set_binaryArgstAt(vexp, 2, mk_enumPtexpcal(newvelem));
	  b = pk_true;
	}
    }
  return b;
}

/**
   It replaces all references to field structures by signals.
   An acces x.t is replaced by the internal signal assigned to it in the manager.
   \see getSignalAssignedTo(), setSignalAssignedTo()
*/
void ExprManager::replaceFieldBySignal()
{
  ptexpcal vexp, e, u;
  ptlinstances liste;

  liste = table[opcham];
  while (ptlinstances_get_next(&liste, &vexp))
    if (getSignalAssignedTo(vexp, &e))
      {
	u = getDeclaredSignal(vexp);
	ptexpcal_belongs_signal(e);
	setClock(e, getClock(e));
	substitute(vexp, e, hundef);
	setIndirection(vexp, e);
	CKM->removeFromClock(getClock(u), u, pk_false);
      }

  liste = table[opident];
  while (ptlinstances_get_next(&liste, &vexp))
    if (ptexpcal_has_typeinst(vexp, tfield))
      {
	ptexpcal_belongs_signal(vexp);
	setClock(vexp, getClock(vexp));
      }
}

/** 
    It substitutes an expression (vexp), where vexp is a state variable, by an another one (nexp).
    It can be used when a state variable may be replaced by a signal.
*/
void ExprManager::substituteStateVariablePreviousBy(ptexpcal vStateVariable, ptexpcal vSignal)
{
  ptexpcal exp;

  if (isinUnaryExprSet(opvvarin, vStateVariable, &exp))
    // exp is the reference to vStateVariable state variable (ie vStateVariable?)
    substitute(exp, vSignal, hundef);
}

/**
   @return "true" if the operand (vexp) is a left_tt expression, "false" otherwise.
   A vexp=left_tt(e1, e2) expression is rewritten in nvexp = (e1 when ^e1) default pk_false, and the
   contribution clock of nvexp is set to vContribClock.
*/
boolean ExprManager::lefttts2defaults(ptexpcal vexp, ptinsth vContribClock, ptexpcal * nvexp)
{
  ptexpcal arg1, e1;
  ptinsth vclk;

  *nvexp = NILEXPCAL;
  if (IsCellExpr(vexp))
    return pk_false;
  vexp = ptexpcal_get_basicExpr(vexp);
  if (ptexpcal_get_operator(vexp) != opleft_tt)
    return pk_false;
  arg1 = getBinaryArgstAt(vexp, 1);
  getSynchronousExprHorl(arg1, &vclk);
  e1 = mkWhenOrCellExp(opextr, arg1, vclk);
  *nvexp = mkBinaryExpTp(opmele, e1, ptexpcal_getFalse(), getBackAnnotation(vexp));
  setContributionClock(*nvexp, vContribClock);
  return pk_true;
}

/**
   It substitutes some intrinsic expressions (dim(T), size(T,i)) by their values.
*/
void ExprManager::intrinsicEvaluation()
{
  ptlinstances liste;
  ptexpcal vexp, lp, le, e1, e2, e3;
  ptident id;
  compilerInteger n;

  liste = table[opappe];
  while (ptlinstances_get_next(&liste, &vexp))
    if (isMarked(vexp, INTRINSIC))
      {
	getFunctionCall(vexp, &id, &lp, &le);
	if (isDimId(id))
	  {
	    // dim(T)
	    getBinaryArgs(le, &e1, &le);
	    n = ptexpcal_get_nbDim(e1);
	    substitute(vexp, mkIntegerConstant(n), hundef);
	  }
	else if (isSizeId(id))
	  // size(T,i)
	  {
	    getBinaryArgs(le, &e1, &le);
	    getBinaryArgs(le, &e2, &le);
	    if (getIntegerValue(e2, &n))
	      if (getIthDimension(e1, (int)n, &e3))
		substitute(vexp, e3, hundef);
	  }
      }
}

/**
   It replaces an expression (vexp) by an another one (nexp) in the manager.
   It applied substitute(), rewriting() methods and set an indirection from vexp to nexp.
*/
void ExprManager::replace(ptexpcal vexp, ptexpcal nexp)
{
  noperators op;
  op = ptexpcal_get_operator(vexp);
  substitute(vexp, nexp, hundef);
  rewriting();
  if ((op != opident) && (op != opcham))
    setIndirection(vexp, nexp);
  cleanDependences(vexp);
}

/**
   @param vexp (in) any expression.
   @param nexp (out) a equivalent expression of vexp using the following rules:
   -  E >= E, E<=E, E=E ---> 'true' expression
   -  E > E, E < E, E <> E ---> 'false' expression
   -  E = F and F << E ---> F=E   (<< Order on expressions)
   -  E < F ---> (E <= F) and not(E=F) 
   -  E > F ---> (F <= E) and not(E=F)
   -  E >= F --->  F <= E
   -  E <= F and F << E --->  (E=F) or not (F <= E)
   -  E != F --> not (E=F)
   -  Warning : Signal substitution is NOT applied.

   @return "true" if no rule have been applied (nexp = vexp), "false" otherwise.
*/
boolean ExprManager::normFormRelationalExpr(ptexpcal vexp, ptexpcal * nexp)
{
  BackAnnotationType arbreerreur;
  noperators ope;
  ptexpcal e1, ne1, e2, ne2;

  SET_NILBACKANNOTATION(arbreerreur);
  ope = ptexpcal_get_operator(vexp);
  e1 = ptexpcal_get_binaryArgstAt(vexp, 1);
  e2 = ptexpcal_get_binaryArgstAt(vexp, 2);
  switch (ope)
    {
    case opegal:
      if (isExprEqualLookingValue(e1, e2))
	*nexp = ptexpcal_getTrue();
      else if (e1beforee2(e1, e2))
	*nexp = vexp;
      else
	*nexp = mkBinaryExp(opegal, e2, e1, arbreerreur);
      break;

    case opdiff:
      if (isExprEqualLookingValue(e1, e2))
	*nexp = ptexpcal_getFalse();	// E != E :false
      else
	// E != F, E << F : not (E=F)
	// E != F, F << E : not (F=E)
      if (e1beforee2(e1, e2))
	*nexp = mkUnaryExp(opnon, mkBinaryExp(opegal, e1, e2, arbreerreur), arbreerreur);
      else
	*nexp = mkUnaryExp(opnon, mkBinaryExp(opegal, e2, e1, arbreerreur), arbreerreur);
      break;

    case opiegal:
      if (isExprEqualLookingValue(e1, e2))
	*nexp = ptexpcal_getTrue();	// E <= E : pk_true
      else if (e1beforee2(e1, e2))
	*nexp = vexp;		// normalized
      else
	// E <= F, F << E : (F=E) or not (F <= E)
	*nexp =
	    mkBinaryExp(opou, mkBinaryExp(opegal, e2, e1, arbreerreur), mkUnaryExp(opnon, mkBinaryExp(opiegal, e2, e1, arbreerreur), arbreerreur),
			arbreerreur);
      break;

    case opinfe:
      if (isExprEqualLookingValue(e1, e2))
	*nexp = ptexpcal_getFalse();	// E < E : false
      else
	{			// E < F, E<<F : (E <= F) and not (E = F)
	  // E < F, F<<E : (E <= F) and not (F = E)
	  ne1 = mkBinaryExp(opiegal, e1, e2, arbreerreur);
	  if (e1beforee2(e1, e2))
	    ne2 = mkBinaryExp(opegal, e1, e2, arbreerreur);
	  else
	    ne2 = mkBinaryExp(opegal, e2, e1, arbreerreur);
	  *nexp = mkBinaryExp(opet, ne1, mkUnaryExp(opnon, ne2, arbreerreur), arbreerreur);
	}
      break;

    case opsegal:
      if (isExprEqualLookingValue(e1, e2))
	*nexp = ptexpcal_getTrue();	// E >= E : pk_true
      else			// E >= F : F <= E  
	*nexp = mkBinaryExp(opiegal, e2, e1, arbreerreur);
      break;

    case opsupe:
      if (isExprEqualLookingValue(e1, e2))
	*nexp = ptexpcal_getFalse();	// E > E : false
      else
	{
	  // E > F, E << F : (F <= E) and not (E=F)
	  // E > F, F << E : (F <= E) and not (F=E)
	  ne1 = mkBinaryExp(opiegal, e2, e1, arbreerreur);
	  if (e1beforee2(e1, e2))
	    ne2 = mkBinaryExp(opegal, e1, e2, arbreerreur);
	  else
	    ne2 = mkBinaryExp(opegal, e2, e1, arbreerreur);
	  *nexp = mkBinaryExp(opet, ne1, mkUnaryExp(opnon, ne2, arbreerreur), arbreerreur);
	}
      break;
    default:
      *nexp = vexp;
    }
  return (boolean) ((*nexp) == vexp);
}

/** 
    Event signals becomes booleans (always "true").
    After this method, the pure type is not referenced in the expression manager, only for the "true" expression.
*/
_VIRTUAL_ void ExprManager::events2booleans()
{
  ptlinstances lexp;
  ptexpcal aSig;
  TypeType boolType;
  int i;

  boolType = get_BooleanType();
  for (i = 0; i < SIG_NB_OPERATORS; i++)
    {
      lexp = table[i];
      while (ptlinstances_get_next(&lexp, &aSig))
	if (!ptexpcal_equalTrue(aSig))
	  if (IsEvent(aSig))
	    {
	      ptexpcal_set_type(aSig, boolType);
	      belongsDefined(aSig);
	    }
    }
  modeNoEvents = pk_true;
}

/**
   It adds in an expression (exp) the explicit conversions to a type (vtype)  of the sub-expressions of exp.
   Assmed : vtype is a predefined type.
*/
ptexpcal ExprManager::explicitConversions(TypeType vtype, ptexpcal exp)
{
  noperators op;
  cloper ope;
  ptexpcal vexp1, vexp2, vexp3, nexp1, nexp2, nexp3;
  ptident id;
  TypeType expType;

  if (exp == NILEXPCAL)
    return exp;

  ope = ptexpcal_get_operatorClass(exp);
  op = ptexpcal_get_operator(exp);
  switch ((int)ope)
    {
    case oper2tp:		// default, but always with the form (e1 when h1) default (e2 when h2) default ...
      ptexpcal_get_defaultArgs(exp, &vexp1, &vexp2);
      nexp1 = explicitConversions(vtype, vexp1);
      if (nexp1 != vexp1)
	ptexpcal_set_defaultArgstAt(exp, 1, nexp1);
      nexp2 = explicitConversions(vtype, vexp2);
      if (nexp2 != vexp2)
	ptexpcal_set_defaultArgstAt(exp, 2, nexp2);
      return exp;
      break;

    case opereh:
      vexp1 = ptexpcal_get_expWhen(exp);
      nexp1 = explicitConversions(vtype, vexp1);
      if (nexp1 != vexp1)
	ptexpcal_set_expWhen(exp, nexp1);
      return exp;
      break;

    case constant:
      if ((op == optrue) || (op == opfalse))
	return exp;

    case variabl:
    case call:
      expType = getType(exp);
      if ((vtype != expType) && (expType != PureType))
	{			// adding an explicit conversion.
	  id = repstr2ident(predefinedType2ident(predefTypeType2SignalType(vtype)));
	  nexp1 = mkCallExp(id, NILEXPCAL, mkNaryExp(oplesign, exp, NILEXPCAL, NILBACKANNOTATION), vtype, hundef);
	  addPragma(nexp1, newPredefinedPragma(PRAGMA_CONVERSION, NILOBJPRAGMA, NILTREE));
	  addMark(nexp1, INTRINSIC);
	  return nexp1;
	}
      return exp;
      break;

    case oper1:
      vexp1 = ptexpcal_get_unaryArg(exp);
      nexp1 = explicitConversions(vtype, vexp1);
      if (nexp1 != vexp1)
	ptexpcal_set_unaryArg(exp, nexp1);
      return exp;

    case oper2:
      ptexpcal_get_binaryArgs(exp, &vexp1, &vexp2);
      // For the relationnal op. vtype is modified.
      if (isRelationalOp(op))
	{
	  getTypeUpb(vexp1, vexp2, &vtype);
	  if (!TypeTestPredef(vtype))
	    return exp;
	}

      nexp1 = explicitConversions(vtype, vexp1);
      if (nexp1 != vexp1)
	ptexpcal_set_binaryArgstAt(exp, 1, nexp1);
      nexp2 = explicitConversions(vtype, vexp2);
      if (nexp2 != vexp2)
	ptexpcal_set_binaryArgstAt(exp, 2, nexp2);
      return exp;

    case oper3:		// delay
      if (op == opret)
	{
	  vexp1 = ptexpcal_get_ternaryArgstAt(exp, 1);
	  nexp1 = explicitConversions(vtype, vexp1);
	  if (nexp1 != vexp1)
	    ptexpcal_set_ternaryArgstAt(exp, 1, nexp1);
	}
      return exp;
      break;

      // case oper1tp oper3tp: opertyp.
    default:
      return exp;
    }
}

/**
   aWhen is assumed to be an expression such that (E when h).
   @return the definition of the signal S (E is a signal) and
   - S is defined by an equation (not a dynamical expression), 
   - S does not have a direct recursive definition.
*/
ptexpcal ExprManager::rewriteWhen(ptexpcal aWhen, ptexpcal s)
{
  ptexpcal e, ve, vexp;
  ptinsth h;
  ptINode n;
  ptldependence li;
  ptdependence aDep;

  vexp = getWhenOrCellArg(aWhen);
  if (!IsSignal(vexp))
    return aWhen;
  if (vexp == s)
    return aWhen;
  // Direct recursivity is not allowed.
  n = getNode(vexp);
  li = NM->get_successors(n);
  if (li->exist_dependence(n, n, &aDep))
    return aWhen;
  // If vexp is defined by a default expression, it is substituted.
  if (!NM->is_equation(n))
    return aWhen;
  if (!NM->get_rightOperand(n, &e))
    return aWhen;
  if (IsCellExpr(e))
    return aWhen;
  ve = getBasicExpr(e);
  if (getOperator(ve) != opmele)
    return aWhen;
  h = getContributionClock(aWhen);
  if (h == getClock(vexp))
    return ve;
  return distributivityDefaultClock(ve, h);
}

/**
   The operand expression (exp) is assumed to be a default expression (i.e.  (E1 when h1) default ...default (En when hn), Ei are synchronous expressions).
   The expression (exp) is assumed to be a sub expression of the definition of a signal s.
   
   When an expression Ei is reduced to a signal (X) and the signal is also defined by an equation (i.e.  X:=  F1 when k1 default ...Fm when km),
   then (Ei when hi) is replaced by ( F1 when k1 default ...Fm when km) when hi. The resulting expression is then normalized.
   @return the expression produced by such substitutions, or the operand expression (exp) when it it not rewritten.

*/
ptexpcal ExprManager::rewriteDefault(ptexpcal exp, ptexpcal s)
{
  ptexpcal narg1, narg2, arg1, arg2;

  arg1 = getDefaultArgstAt(exp, 1);
  if (getOperator(arg1) == opmele)
    narg1 = rewriteDefault(arg1, s);
  else
    narg1 = rewriteWhen(arg1, s);
  arg2 = getDefaultArgstAt(exp, 2);
  narg2 = rewriteWhen(arg2, s);
  if ((narg1 == arg1) && (narg2 == arg2))
    return exp;
  arg1 = mergeExpressions(narg1, narg2);
  setBackAnnotation(arg1, getBackAnnotation(exp));
  return arg1;
}

/**
   @return "true" if the operand expresseion (e) may be removed from the graph, "false" otherwise.
   - An output signal cannot be removed.
   - A signal used in a placed clock cannot be removed.
   - A signal marked "DontOptimize" cannot be removed
*/
boolean ExprManager::mayBeRemoved(ptexpcal e)
{
  ptinsth vhSub;

  if (IsOutputFlow(e))
    return pk_false;
  if (ptexpcal_get_extraction(e, &vhSub))
    if (!CKM->satisfiesNullDepth(vhSub))
      return pk_false;
  if (isSharedSignalBySeveralGraphs(e))
    return pk_false;
  if (NM->isMarked(getNode(e), DontOptimize))
    return pk_false;

  return pk_true;
}

/**
   Assumed: exp is a when expression (E when H) and lexp is a list of when expressions.
   @return "false" if it exists an expression (E when K) is lexp, it is replaced by (E when H ^+ K), "true" otherwise and exp is added to lexp.
*/
boolean ExprManager::addExpToDefaultList(ptlinstances * lexp, ptexpcal exp)
{
  ptlinstances le;
  ptexpcal ne, elem, e;

  le = *lexp;
  e = getWhenOrCellArg(exp);
  while (ptlinstances_get_next(&le, &elem))
    if (getWhenOrCellArg(elem) == e)
      {
	ne = mkWhenOrCellExp(opextr, e, CKM->makeSup(getContributionClock(elem), getContributionClock(exp)));
	ptlinstances_replace(*lexp, elem, ne);
	return pk_false;
      }
  *lexp = ptlinstances_push_front(*lexp, exp);
  return pk_true;
}

/**
   @return 
   - "true" if the rule, (E when H) default (E when K) --> E when (H ^+ K), is applied on the operand expression (exp). lnexp contains the resulting elements (a set of "Ei when Hi"),
   - "false"otherwise and lnexp is set to NILLINSTANCES.
*/
boolean ExprManager::normalDefaults(ptexpcal exp, ptlinstances * lnexp)
{
  ptexpcal exp2, nexp;
  ptlinstances lexp;
  boolean bmodified;

  lexp = NILLINSTANCES;
  bmodified = pk_false;
  if (getOperator(exp) == opmele)
    {
      // Building the list (lexp) of (Ei when Hi), unicity guaranteed for Ei.
      do
	{
	  getDefaultArgs(exp, &exp, &exp2);
	  if (!addExpToDefaultList(&lexp, exp2))
	    bmodified = pk_true;
	}
      while (getOperator(exp) == opmele);
      if (!addExpToDefaultList(&lexp, exp))
	bmodified = pk_true;
    }
  *lnexp = lexp;
  if (!bmodified)
    ptlinstances_free(lexp);
  return bmodified;
}

/**
   @return a new expression in which the expression e1 is repalced by e2 in an expression (use).
   The references are not rewritten.
*/
ptexpcal ExprManager::replace(ptexpcal e1, ptexpcal e2, ptexpcal use)
{
  cloper ope;
  ptexpcal exp1, exp2, exp3, nexp1, nexp2, nexp3;
  ptident id;

  if (use == NILEXPCAL)
    return use;
  if (e1 == use)
    return e2;

  ope = ptexpcal_get_operatorClass(use);
  switch ((int)ope)
    {
    case oper1:
      exp1 = ptexpcal_get_unaryArg(use);
      nexp1 = replace(e1, e2, exp1);
      if (exp1 == nexp1)
	return use;
      return mkUnaryExp(ptexpcal_get_operator(use), nexp1, getBackAnnotation(use));
      break;

    case oper2:
      ptexpcal_get_binaryArgs(use, &exp1, &exp2);
      nexp1 = replace(e1, e2, exp1);
      nexp2 = replace(e1, e2, exp2);
      if ((nexp1 == exp1) && (nexp2 == exp2))
	return use;
      return mkBinaryExp(ptexpcal_get_operator(use), nexp1, nexp2, getBackAnnotation(use));
      break;

    case oper3:
      ptexpcal_get_ternaryArgs(use, &exp1, &exp2, &exp3);
      nexp1 = replace(e1, e2, exp1);
      nexp2 = replace(e1, e2, exp2);
      nexp3 = replace(e1, e2, exp3);
      if ((nexp1 == exp1) && (nexp2 == exp2) && (nexp3 == exp3))
	return use;
      return mkTernaryExp(ptexpcal_get_operator(use), nexp1, nexp2, nexp3, getBackAnnotation(use));
      break;

    case opereh:
      exp1 = ptexpcal_get_expWhen(use);
      nexp1 = replace(e1, e2, exp1);
      if (nexp1 == exp1)
	return use;
      return mkWhenOrCellExp(ptexpcal_get_operator(use), nexp1, getContributionClock(use));
      break;

    case oper2tp:
      ptexpcal_get_defaultArgs(use, &exp1, &exp2);
      nexp1 = replace(e1, e2, exp1);
      nexp2 = replace(e1, e2, exp2);
      if ((nexp1 == exp1) && (nexp2 == exp2))
	return use;
      nexp1 = mkBinaryExpTp(ptexpcal_get_operator(use), nexp1, nexp2, getBackAnnotation(use));
      setContributionClock(nexp1, getContributionClock(use));
      return nexp1;
      break;

    case call:
      getFunctionCall(use, &id, &exp2, &exp1);
      nexp1 = replace(e1, e2, exp1);	// inputs
      nexp2 = replace(e1, e2, exp2);	// parameters
      if ((nexp1 == exp1) && (nexp2 == exp2))
	return use;
      return mkCallExp(id, nexp2, nexp1, getType(use), getContributionClock(use));
      break;

    default:
      //    case variabl: case constant: case oper1tp: case oper3tp:
      return use;

    }
}

// For PolyEndo
void ExprManager::replaceDefault(ptinsth vh1, ptinsth vh2, ptexpcal exp)
{
  ptexpcal narg1, narg2, arg1, arg2;
  arg1 = getDefaultArgstAt(exp, 1);
  if (getOperator(arg1) == opmele)
    replaceDefault(vh1, vh2, arg1);
  else
    replaceWhen(vh1, vh2, arg1);
  arg2 = getDefaultArgstAt(exp, 2);
  if (getOperator(arg2) == opmele)
    replaceDefault(vh1, vh2, arg2);
  else
    replaceWhen(vh1, vh2, arg2);
}

// For PolyEndo
void ExprManager::replaceWhen(ptinsth vh1, ptinsth vh2, ptexpcal exp)
{
  ptinsth h = getContributionClock(exp);
  if (CKM->getRepresentativeFormula(h) == CKM->getRepresentativeFormula(vh1))
    {
      setContributionClock(exp, vh2);
    }
}

// For PolyEndo
void ExprManager::replace(ptinsth vh1, ptinsth vh2, ptexpcal exp)
{
  ptexpcal e;

  if (IsCellExpr(exp))
    {
      replaceWhen(vh1, vh2, exp );
      replace(vh1, vh2, getWhenOrCellArg(exp));
    }
  else
    {
      e = getBasicExpr(exp);
      if (getOperator(e) == opmele)
	replaceDefault(vh1, vh2, e);
      else
	replaceWhen(vh1, vh2, exp);
    }
}
