/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   @return 
   - "true" if an indice (vindex), with nbDim dimensions, is defined by an arithmetical progression (isArithmeticalProgression()) 
   for its vrank-th dimension.
      - vrank = 1, nbDim = 1:  0..n step 1
      - vrank = 1, nbDim = 2:  tr (0.. n step 1 |* m)
      - vrank = 2, nbDim = 1:  0..n step 1
      - vrank = 2, nbDim = 2: (0.. n step 1 |* m)

   - "false" otherwise

   To be COMPLETED.
*/
  boolean ExprManager::IndiceSatisfiesArithmeticalProgression (ptexpcal vindex, int vrank, int nbDim)
{
  ptexpcal
    vdef,
    vsize;
  ptINode
    vnode;

  if (nbDim == 0)
    return pk_false;
  vnode = getNode (vindex);
  if (NM->is_undefined (vnode))
    return pk_false;
  if (NM->is_intrinsicNode (vnode, getIndicesId ()))
    return pk_true;
  if (nbDim > 2)
    return pk_false;
  // if (vrank > nbDim) return pk_false;
  // assumed: expressions are well typed.
  if (!NM->is_equation (vnode))
    return pk_false;
  if (!NM->get_rightOperand (vnode, &vdef))
    return pk_false;

  getIthDimension (vindex, vrank, &vsize);
  if (vrank == 1)
    {
      if (nbDim == 1)
	return isArithmeticalProgression (vdef, 1, vsize);	// 0..n step 1 admitted
      if (nbDim == 2)
	return isArithmeticalProgression (vdef, 2, vsize);	// tr (0.. n step 1 |* m) admitted
      return pk_false;
    }

  if (vrank == 2)
    {
      if (nbDim == 1)
	return isArithmeticalProgression (vdef, 1, vsize);	// 0..n step 1 admitted
      // First : tr( 0..n |* m) and second (0..m |* n)
      if (nbDim == 2)
	return isArithmeticalProgression (vdef, 3, vsize);	// (0.. n step 1 |* m)
      return pk_false;
    }
  return pk_false;
}

/**
   @return
   - "true" if a signal is defined as an index, i.e. by:
       - a reference to the intrinsic function "indices"
       - an enumerating expression (e1 .. e2 step e3),
       - a matrix transposition (tr (0.. n step 1 |* m)),
       - a repetition ( e1 |* e2 , repeat e1 e2 times)

   -"false" otherwise.
*/
boolean ExprManager::isIndexDefinition (ptexpcal vexp)
{
  ptexpcal
    vdef;
  ptINode
    vnode;
  noperators
    op;

  if (!ptexpcal_IsSignal (vexp))
    return pk_false;
  vnode = getNode (vexp);
  if (NM->is_undefined (vnode))
    return pk_false;
  if (NM->is_intrinsicNode (vnode, getIndicesId ()))
    return pk_true;
  if (!NM->is_equation (vnode))
    return pk_false;
  if (!NM->get_rightOperand (vnode, &vdef))
    return pk_false;

  if (ptexpcal_satisfies_cell (vdef))
    return pk_false;
  vdef = ptexpcal_get_basicExpr (vdef);
  if (ptexpcal_IsSignal (vdef))
    {
      vnode = getNode (vdef);
      if (!NM->is_equation (vnode))
	return pk_false;
      if (!NM->get_rightOperand (vnode, &vdef))
	return pk_false;
      return isIndexDefinition (vdef);
    }
  op = ptexpcal_get_operator (vdef);
  if (op == openum)
    return pk_true;
  if (op == optran)
    return pk_true;
  if (op == oprepe)
    return pk_true;
  return pk_false;
}

/**
   It adds to the synthetized travel indexes associated with an expression (vexp) the list
   of the signals marked as possible indexes (getPossiblyIndexesReferencedIn()).
*/
void
ExprManager::addTravelIndexesToExpression (ptexpcal vexp)
{
  ptlinstances l1, l, ls;
  ptexpcal vsig, e1, e2;
  noperators op;

  if (vexp != NILEXPCAL)
    {
      op = ptexpcal_get_operator (vexp);
      if (op == opsequ)		// E1 next E2 ... next En
	{
	  getBinaryArgs (vexp, &e1, &e2);
	  if (e1 != NILEXPCAL)
	    addTravelIndexesToExpression (e1);
	  if (e2 != NILEXPCAL)
	    addTravelIndexesToExpression (e2);
	}
      else
	{
	  getTravelIndexes (vexp, &l);
	  ls = getPossiblyIndexesReferencedIn (vexp);
	  l1 = ls;
	  while (ptlinstances_get_next (&ls, &vsig))
	    if (!ptlinstances_isin_exp (l, vsig))
	      l = ptlinstances_push_end (l, vsig);
	  setTravelIndexes (vexp, l);
	  ptlinstances_free (l1);
	}
    }
}

/**
   @return the list of the optimized indexes referenced in an expression.
   An index element of the list of the synthetized travel indexes associated with the expression, is declared optimized
   if its definition is an arithmetical progression (IndiceSatisfiesArithmeticalProgression).
   The built list is assigned to the expression (setOptimizedTravelIndexes(), getOptimizedTravelIndexes()).
*/
ptlinstances ExprManager::mkOptimizedIndexes (ptexpcal vdef)
{
  ptlinstances
    l,
    loptimizedIndices;
  ptexpcal
    vsig,
    e1,
    e2;
  noperators
    op;
  int
    i,
    nb;

  if (vdef == NILEXPCAL)
    return NILLINSTANCES;
  op = ptexpcal_get_operator (vdef);
  if (op == opsequ)		// E1 next E2... next En
    {
      getBinaryArgs (vdef, &e1, &e2);
      return ptlinstances_sortedfusion (mkOptimizedIndexes (e1), mkOptimizedIndexes (e2));
    }
  if (!getTravelIndexes (vdef, &l))
    return NILLINSTANCES;
  loptimizedIndices = NILLINSTANCES;
  i = 0;
  while (ptlinstances_get_next (&l, &vsig))
    {
      i++;
      nb = ptexpcal_get_nbDim (vsig);
      if (IndiceSatisfiesArithmeticalProgression (vsig, i, nb))
	loptimizedIndices = ptlinstances_insert (loptimizedIndices, vsig);
    }
  setOptimizedTravelIndexes (vdef, loptimizedIndices);
  return loptimizedIndices;
}
