/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   @return the list of travel indexes of an expression (e) assumed to be an index expression or a list of index expressions.
*/
  ptlinstances ExprManager::mkIndexLst (ptexpcal e)
{
  ptlinstances
    l;
  int
    n;
  ptexpcal
    aIndex;
  TypeType
    TT;
  noperators
    op;

  l = NILLINSTANCES;
  n = ptexpcal_get_nbelem (e);
  while (n > 0)
    {
      aIndex = ptexpcal_getNaryArgs (&e);
      op = ptexpcal_get_operator (aIndex);
      if (op == opident)
	{
	  TT = ptexpcal_getUnamedType (aIndex);
	  if (TypeTestArrayClass (TT))
	    if (!ptlinstances_isin_exp (l, aIndex))
	      l = ptlinstances_push_end (l, aIndex);
	}
      else
	l = getMaxTravelIndexes (l, ptexpcal_get_travelIndexes (aIndex));
      n--;
    }
  return l;
}

/**
   @return the list with the greater size.
   The elements of the shorter list are assumed to be equal with the elements of the other one: an indirection is 
   currently set (it should be a constraint).
*/
ptlinstances ExprManager::getMaxTravelIndexes (ptlinstances l1, ptlinstances l2)
{
  int
    n1,
    n2;
  ptexpcal
    v1,
    v2;
  ptlinstances
    l;

  if (l1 == NILLINSTANCES)
    return l2;
  if (l2 == NILLINSTANCES)
    return l1;
  n1 = ptlinstances_get_size (l1);
  n2 = ptlinstances_get_size (l2);
  if (n2 > n1)
    {
      l = l2;
      l2 = l1;
      l1 = l;
    }
  l = l1;

  // Ajout contraintes d'égalité
  while (ptlinstances_get_next (&l2, &v2))
    {
      ptlinstances_get_next (&l1, &v1);
      setIndirection (v2, v1);	// pour test ( A terminer )
    }
  return l;
}

/**
   @return the getMaxTravelIndexes(ptlinstances, ptlinstances) on the list of travel indexes of each expression.
*/
ptlinstances ExprManager::getMaxTravelIndexes (ptexpcal e1, ptexpcal e2)
{
  ptlinstances
    l1,
    l2;

  if (e1 == NILEXPCAL)
    l1 = NILLINSTANCES;
  else
    l1 = ptexpcal_get_travelIndexes (e1);
  if (e2 == NILEXPCAL)
    l2 = NILLINSTANCES;
  else
    l2 = ptexpcal_get_travelIndexes (e2);

  return getMaxTravelIndexes (l1, l2);
}

/**
   It synthetizes the set of travel indexes for a synchronous unary expression res, defined by op(e1).
   It is defined by:
    - if res= << E >> (cartesian product of indexes),  mkIndexLst() of E
    - the the set of travel indexes of e1, otherwise.
*/
void
ExprManager::SpatialUnary (ptexpcal res, noperators op, ptexpcal e1)
{
  ptlinstances l;
  if (op == oppcar)
    l = mkIndexLst (e1);
  else
    l = ptexpcal_get_travelIndexes (e1);
  ptexpcal_set_travelIndexes (res, l);
}

/**
   It assigns the set of the travel indexes of an expression (e1) to an another one res=op(e1), for op a temporal unary operator.
   \note Loïc Besnard: never called, I think.
*/
void
ExprManager::SpatialUnaryTp (ptexpcal res, noperators op, ptexpcal e1)
{
  ptexpcal_set_travelIndexes (res, ptexpcal_get_travelIndexes (e1));
}


/**
   It synthetizes the set of travel indexes for a synchronous binary expression res, defined by op(e1, e2).
   It is defined by:
    - if res= E1 next E2 next .... En (each component has its own indices), no synthesis
    - if res=x[E] it is the mkIndexLst() of E
    - if res=x:E  it is the mkIndexLst() of x
    - otherwise, the getMaxTravelIndexes(ptexpcal, ptexpcal) of e1, e2.
*/
void
ExprManager::SpatialBinary (ptexpcal res, noperators op, ptexpcal e1, ptexpcal e2)
{
  ptlinstances l;

  if (op == opsequ)
    return;			// E1 next E2 next .... En (each component has its own indices).
  if (op == opelem)
    l = mkIndexLst (e2);
  else if (op == oprest)
    l = mkIndexLst (e1);
  else
    l = getMaxTravelIndexes (e1, e2);
  ptexpcal_set_travelIndexes (res, l);
}

/**
   It synthetizes the set of travel indexes for a temporal-binary  expression (default) res, defined by op(e1, e2).
   It is the getMaxTravelIndexes(ptexpcal, ptexpcal) of set of travel indexes of the operands e1, e2.
*/
void
ExprManager::SpatialBinaryTp (ptexpcal res, noperators op, ptexpcal e1, ptexpcal e2)
{
  ptexpcal_set_travelIndexes (res, getMaxTravelIndexes (e1, e2));
}


/**
   It synthetizes the set of travel indexes for a ternary expression res defined by op(e1,e2,e3)
   Currently, no computation.
*/
void
ExprManager::SpatialTernary (ptexpcal res, noperators op, ptexpcal e1, ptexpcal e2, ptexpcal e3)
{
}


/**
   It synthetizes the set of travel indexes for a n-ary combed expression res, defined by op(e1, e2).
   It is the getMaxTravelIndexes(ptexpcal, ptexpcal) of set of travel indexes of the operands e1, e2.
*/
void
ExprManager::SpatialNary (ptexpcal res, noperators op, ptexpcal e1, ptexpcal e2)
{
  ptexpcal_set_travelIndexes (res, getMaxTravelIndexes (e1, e2));
}

/**
   It synthetizes the set of travel indexes for an extraction (E when H) or a memory (E cell H) expression.
   It is the set of travel indexes of the first argument (E).
*/
void
ExprManager::SpatialExtractOrMemory (ptexpcal res, noperators op, ptexpcal e1, ptinsth vclk)
{
  ptexpcal_set_travelIndexes (res, ptexpcal_get_travelIndexes (e1));
}

/**
   It synthetizes the set of travel indexes for an function call expression (exp) with ins as inputs.
   It is the getMaxTravelIndexes(ptlinstances, ptlinstances) of set of travel indexes of the inputs.
*/
void
ExprManager::SpatialFuncCall (ptexpcal exp, ptexpcal ins)
{
  ptlinstances l;
  ptexpcal arg;

  l = NILLINSTANCES;
  while (getNaryArgs (&ins, &arg))
    l = getMaxTravelIndexes (l, ptexpcal_get_travelIndexes (arg));
  ptexpcal_set_travelIndexes (exp, l);
}

/**
   Assumed: the expressions are evaluated.
   @return 
   - "true" if the expression vexp, with vsize the size of an expression of its dimensions, satisfies
        - codeOp = 1,  0..n step 1  with vsize = n
	- codeOp = 2,  tr (0.. n step 1 |* m)  with vsize = n (i.e. isArithmeticalProgression(getUnaryArg(vexp), 3, vsize))
	- codeOp = 3,  tr (0.. n step 1 |* m)  with vsize = n (i.e. isArithmeticalProgression(getBinaryArgstAt(vexp, 1), 1, vsize))
   - "false" otherwise.
*/
boolean ExprManager::isArithmeticalProgression (ptexpcal vexp, int codeOp, ptexpcal vsize)
{
  ptexpcal
    vdef,
    e1,
    e2,
    e3;
  ptINode
    vnode;
  noperators
    op;
  compilerInteger
    n,
    m;

  if (ptexpcal_satisfies_cell (vexp))
    return pk_false;
  vexp = ptexpcal_get_basicExpr (vexp);
  if (ptexpcal_IsSignal (vexp))
    {
      vnode = getNode (vexp);
      if (!NM->is_equation (vnode))
	return pk_false;
      if (!NM->get_rightOperand (vnode, &vdef))
	return pk_false;
      return isArithmeticalProgression (vdef, codeOp, vsize);
    }
  op = ptexpcal_get_operator (vexp);
  switch (codeOp)
    {
    case 1:			// 0..n step 1 admitted
      if (op != openum)
	return pk_false;
      getTernaryArgs (vexp, &e1, &e2, &e3);
      if (!getIntegerValue (e3, &n))
	return pk_false;
      if (n != 1)
	return pk_false;
      if (!getIntegerValue (e1, &n))
	return pk_false;
      if (n != 0)
	return pk_false;
      if (!getIntegerValue (e2, &n))
	return pk_false;
      if (!getIntegerValue (vsize, &m))
	return pk_false;
      return (m == n + 1);
      break;

    case 2:			// tr (0.. n step 1 |* m)
      if (op != optran)
	return pk_false;
      return isArithmeticalProgression (getUnaryArg (vexp), 3, vsize);
      break;

    case 3:			// (0.. n step 1 |* m)
      if (op != oprepe)
	return pk_false;
      return isArithmeticalProgression (getBinaryArgstAt (vexp, 1), 1, vsize);
      break;
    }
  return pk_false;
}
