/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   @return the normalized array type T, [DimLst,...]Term, representing [DimLst] TT.
   In normal form, Term is not an array. [DimLst,...] shall contain all dimensions.
*/
  TypeType ExprManager::mkArrayType (ptexpcal DimLst, TypeType TT)
{
  ptexpcal r;
  if (TypeTestArrayClass (TT))
    {
      r = ArrayMatrixTypeGetDimLst (TT);
      DimLst = fusionlist (getOperator (DimLst), DimLst, r, NILBACKANNOTATION);
      TT = ArrayMatrixTypeGetTerm (TT);
    }
  return ArrayTypeCreate (DimLst, TT);
}

/**
   @return "false".
   It associates the error type to an expression (vexp) and a message (imessage) to
   the corresponding tree (where).
*/
boolean
ExprManager::setErrorTypeExp (ptexpcal vexp, int imessage, BackAnnotationType where)
{
  pK_messenger->errorTree (imessage, BACK_ANNOT_GET_ABSTRACT_TREE(where));
  setErroneousType (vexp);
  return pk_false;
}

/**
   @return "false".
   It associates the error type to an expression (vexp) and an error message, "unexpected type", 
   to the corresponding tree (where).
*/
boolean
ExprManager::errUnexpectedType (ptexpcal vexp, BackAnnotationType where)
{
  return setErrorTypeExp (vexp, errctUnexpectedType, where);
}

/**
   @return "false".
   It associates the error type to an expression (vexp) and an error message, "error for extension type", 
   to the corresponding tree (where).
*/
boolean
ExprManager::errExtension (ptexpcal vexp, BackAnnotationType where)
{
  return setErrorTypeExp (vexp, errctTypeExtension, where);
}


/**
   It verifies the  Numerical/Logical expressions extensions for an expression (vexp).
   The eventual error is hooked to the where.
*/
void
ExprManager::TypeCheckingExpcalNLExtensions (ptexpcal vexp, BackAnnotationType where)
{
  if (!ExpcalTestNLExtensions (vexp))
    errExtension (vexp, where);
}


/**
   @return 
   - "true" if the operation succeeds, t = greatest type of source and target (unifytctype()).
   - "false" otherwise, t is not modified.
*/
boolean
ExprManager::GetGreaterType (TypeType source, TypeType target, TypeType * t)
{
  if (unifytctype (source, target, t))
    return pk_true;
  return unifytctype (target, source, t);
}


/** 
    @return "true" if the operation succeeds, "false" otherwise.
    It verifies the type of an index expression excluding cartesian product (return pk_true)
    It assumes that the list of travel indexes is available. 
    It verifies that the indexes have the same type.
*/
boolean
ExprManager::TypeCheckingIndexes (ptexpcal vexp)
{
  ptlinstances l;
  ptexpcal v1, v2;
  TypeType t1;
  noperators vop;

  vop = ptexpcal_get_operator (vexp);
  if (vop != oppcar)		// catersian product
    {
      l = ptexpcal_get_travelIndexes (vexp);
      if (!ptlinstances_get_next (&l, &v1))
	return pk_true;
      t1 = ptexpcal_getUnamedType (v1);
      while (ptlinstances_get_next (&l, &v2))
	if (t1 != ptexpcal_getUnamedType (v2))
	  return pk_false;
    }
  return pk_true;
}

/**
   It controls the compatibility of the types for the definition 
   @param vexpcal1 (in) left part of the definition
   @param vexpcal2 (in) the expression of definition of the equation/partial definition.
   @param refTree (in) : reference to the syntactic tree of the program ; used for annotations.
   @param partialDef (in) when "true", it specifies that the equation is a partial one (vexpcal1 ::= vexpcal2).
   @param bextension (out) set to "true" if the expressions are well typed using extension rules, "false" otherwise.
   @return "true" if the expressions are well typed, in this case bextension is fixed, "false" otherwise
*/
boolean
ExprManager::TypeCheckingEquation (ptexpcal vexpcal1, ptexpcal * vexpcal2, BackAnnotationType refTree, boolean * bextension)
{
  TypeType TT;
  noperators op;
  boolean b;

  if (ptexpcal_getUnamedType (*vexpcal2) == PureType)
    *vexpcal2 = ptexpcal_getTrue ();
  op = ptexpcal_get_operator (*vexpcal2);
  if (op == oprest)
    return verifAssignRestructuration (vexpcal1, *vexpcal2, bextension, refTree);

  VerifForEachDimensionOf = pk_true;
  b = TypeCheckingAccept (vexpcal1, *vexpcal2, &TT, refTree, pk_true, pk_true);

  if (b && verifSequentialDef (vexpcal1, *vexpcal2, refTree))
    (*bextension) = pk_false;
  else if (!booleanIntegerConversion (vexpcal1, vexpcal2, refTree))
    {
      if (getTypeUpbForArray (ptexpcal_getUnamedType (vexpcal1), ptexpcal_getUnamedType (*vexpcal2), &TT))
	(*bextension) = pk_true;
      else
	{
	  erreurattachref1nom (errctaff1, refTree, vexpcal1);
	  return pk_false;
	}
    }

  if (!VerifForEachDimensionOf)
    pK_messenger->warningacc (warNoct,  BACK_ANNOT_GET_ABSTRACT_TREE(refTree));

  ptexpcal_belongsDefined (vexpcal1);
  if (!TypeCheckingIndexes (*vexpcal2))
    {
      pK_messenger->errorTree (errTypeIndex,  BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
      return pk_false;
    }
  return pk_true;
}

/**
   For an equation lexp := rexp, if rexp is an event, it is propagated to lexp.
*/
void
ExprManager::TypeCheckingEquationFinalize (ptexpcal lexp, ptexpcal rexp, BackAnnotationType refTree)
{
  noperators op;
  TypeType t;

  if (rexp != NILEXPCAL)
    {
      if (ptexpcal_IsEvent (rexp))
	setPureType (lexp);
      else
	{
	  t = ptexpcal_getUnamedType (rexp);
	  if (TypeTestBundleClass (t))
	    {
	      op = ptexpcal_get_operator (getBasicExpr (rexp));
	      if (!((op == opnom) || (op == opident)))
		pK_messenger->errorTree (errExtensionNotYetImplemented,  BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
	    }
	}
    }
}

/**
   @return "true" if an expression (vexpcalcul) is a scalar expression, "false" otherwise (in this case it is modified: it is set to scalar).
   @param vexpcalcul (in) an expression
   @param refTree (in) : reference to the syntactic tree of the program ; used for annotations.
   \see controlequadim
*/
boolean
ExprManager::TypeCheckingIsScalar (ptexpcal vexpcalcul, BackAnnotationType refTree)
{
  if (controlequadim (vexpcalcul, NILEXPCAL, refTree))
    return pk_true;
  ptexpcal_set_dimension (vexpcalcul, NILEXPCAL);
  return pk_false;
}


/**
   If the type of the expression is not defined, it is fixed to boolean.
   @return 
   - "true" if the operand expression is boolean, 
   - "false" otherwise, an error is hooked to vtree.
*/
boolean
ExprManager::TypeCheckingCondition (ptexpcal vexp, BackAnnotationType vtree)
{
  if (ExpcalTypeSet (vexp, BooleanType))
    return TypeCheckingIsScalar (vexp, vtree);
  pK_messenger->errorTree (errctaff1,  BACK_ANNOT_GET_ABSTRACT_TREE(vtree));
  return pk_false;
}

/** 
    If the type of the expression is defined, it verifies that it is boolean.
    @return 
    - "true" if the operand expression is boolean or not fixed.
    - "false" otherwise, an error is hooked to vtree.
*/
boolean
ExprManager::TypeCheckingVerifyTypeCondition (ptexpcal vexp, BackAnnotationType vtree)
{
  TypeType t;

  t = ptexpcal_getUnamedType (vexp);
  if (TypeTestPredef (t))
    {
      if (t == PureType) return pk_true;
      if (t == BooleanType) return pk_true;    
    }
  else 
    if (TypeTestAnyClass (t)) return pk_true;
  pK_messenger->errorTree (errtypecond,  BACK_ANNOT_GET_ABSTRACT_TREE(vtree));
  return pk_false;
}

/**
   It controls the compatibility of the types of a effective parameter and a formal parameter in a call node.
   @param TT (in) type of the formal element
   @param vexpcal (in) the effectif element
   @param refTree (in) reference to the syntactic tree of the program ; used for annotations.
*/
void
ExprManager::TypeCheckingFormalEffective (TypeType TT, ptexpcal vexpcal, boolean enentree, BackAnnotationType refTree)
{
  TypeType TTeff, vtt;
  boolean b;

  if (TypeTestNamedClass (TT))
    TT = NamedTypeGetTerm (TT);
  if (!setTypeIfFree (vexpcal, TT, refTree))
    {
      TTeff = ptexpcal_getUnamedType (vexpcal);
      if (TT != TTeff)
	{ 
	  if (!TypeTestAbstract (TT))
	    {
	      if (enentree)
		{
		  if (TypeTestArrayClass (TT))
		    b = getTypeUpbForArray (TT, TTeff, &TT);
		  else
		    b = unifytctype (TTeff, TT, &vtt);
		}
	      else
		{
		  if (TypeTestArrayClass (TT))
		    b = getTypeUpbForArray (TTeff, TT, &TT);
		  else
		    b = unifytctype (TT, TTeff, &vtt);
		}
	      if (!b)
		erreurattachnom (errctEffectiveFormal, refTree, vexpcal);
	    }
	  else if (!TypeTestAnyClass (TT))
	    erreurattachnom (errctEffectiveFormal, refTree, vexpcal);
	}
    }
}

/** 
    @return 
    - "true" if the expressions, vexpcal1 and vexpcal2, have the same dimension, 
    - "false" otherwise, an error is hooked to refTree.
*/
boolean
ExprManager::TypeCheckingSameSize (ptexpcal vexpcal1, ptexpcal vexpcal2, BackAnnotationType refTree)
{
  return controlequadim (vexpcal2, ptexpcal_get_dimension (vexpcal1), refTree);
}


/**
   @return 
    - "true" if the type of vexpcal1 is the type event (a conversion boolean->event can be applied).
    - "false" otherwise, an error is hooked to declaration of  vexpcal1.
    
    notdefined is set to "true" if vexpcal1 is not valuated, "false" otherwise.
*/
boolean
ExprManager::TypeCheckingIsEvent (ptexpcal vexpcal1, boolean * notdefined)
{
  boolean res;
  *notdefined = pk_false;
  if (modeNoEvents)
    return pk_false;
  if (ptexpcal_IsEvent (vexpcal1))
    {
      if (ptexpcal_IsAssigned (vexpcal1))
	{
	  if (!ptexpcal_equalTrue (vexpcal1))
	    erreurattachnom (errctren1, ptexpcal_get_treeOfSignalDeclaration (vexpcal1), vexpcal1);
	}
      else
	*notdefined = pk_true;
      return pk_true;
    }
  res = TypeChecking_convert2event (vexpcal1);
  return res;
}

/**
   Type checking for a list expression.
   @param vexp (in) expression defined by vope(vexpcal1, vexpcal2).
   @param vope (in) operator of the expression.
   @param vexpcal1, vexpcal2  (in) operands of the expression.
   @param refTree (in) : reference to the syntactic tree of the program ; used for annotations.
   @return "true" if the expression is well typed, "false" otherwise.
*/
boolean
ExprManager::TypeCheckingList (noperators vope, ptexpcal vexp, ptexpcal vexpcal1, ptexpcal vexpcal2, BackAnnotationType refTree)
{
  switch ((int) vope)
    {
    case opleflot:
      return TypeCheckinglesign (vexp, vexpcal1, vexpcal2, refTree);
    case oplelevec:
      return TypeCheckinglelevec (vexp, vexpcal1, vexpcal2, refTree);
    case oplparc:
      return TypeCheckinglparc (vexp, vexpcal1, vexpcal2, refTree);
    case opchai:
      return TypeCheckingString (vexp, vexpcal1, vexpcal2, refTree);
    case oplident:
    case oplepara:
    case oplopra:
    case oplesign:
      return pk_true;

    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
  return pk_false;
}

/** 
    @return 
    - "true" if a value (val) is an element of an assumed enumerated type (TT); TT must be a enumerated type.
    - "false" otherwise, an error is hooked to refTree.
 */
boolean ExprManager::TypeCheckingControlValEnum (TypeType TT, ptrepstr val, BackAnnotationType refTree)
{
  int
    i;
  if (TypeTestEnumerationClass (TT))
    if (EnumTypeOrdinal (TT, val, &i))
      return pk_true;
  pK_messenger->errorTreeStr (errenumvalue,  BACK_ANNOT_GET_ABSTRACT_TREE(refTree), val);
  return pk_false;
}

/**
   Type checking for an expression (vexp ) defined by a temporal binary expression with vexp1, vexp2 as operands.
   On error, a message is hooked to refTree.
*/
void
ExprManager::TypeCheckingBinaryTemporal (noperators op, ptexpcal vexp, ptexpcal vexp1, ptexpcal vexp2, BackAnnotationType refTree)
{
  TypeType TT;
  compilerInteger bi, bs, bi1, bs1, bi2, bs2;

  switch ((int) op)
    {
    case opmele:
      if (!getTypeUpb (vexp1, vexp2, &TT))
	pK_messenger->errorTree (errctaff1,  BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
      else
	{
	  ptexpcal_set_type (vexp, TT);
	  if (getBounds (vexp1, &bi1, &bs1) && getBounds (vexp2, &bi2, &bs2))
	    {
	      bi = minimum (minimum (minimum (bi1, bi2), bs1), bs2);
	      bs = maximum (maximum (maximum (bi1, bi2), bs1), bs2);
	      ptexpcal_set_integerBoundsHolder (vexp, bi, bs);
	    }

	}
      break;

      // case opwhenLustre: ptexpcal_set_type(vexp, ptexpcal_get_type(vexp1)); break;
    }
}

/**
   Type checking for a when/cell expression (vexp) defined by vexp1.
*/
void
ExprManager::TypeCheckingExtractOrMemory (noperators op, ptexpcal vexp, ptexpcal vexp1, BackAnnotationType refTree)
{
  compilerInteger bi, bs;

  ptexpcal_set_type (vexp, ptexpcal_get_type (vexp1));
  // if (op == opextr)
  if (getBounds (vexp1, &bi, &bs))
    ptexpcal_set_integerBoundsHolder (vexp, bi, bs);
}

/**
   Type checking for delay expression (vexp) defined by "vexpcal1 $ valret window valfen init vinit". On error, a message is hooked to refTree.
   The control on initialisation is realized at the end of the graph production \see{TypeCheckingInit}.
*/
void
ExprManager::TypeCheckingDelay (noperators op, ptexpcal vexp, ptexpcal vexpcal1, ptexpcal valret, ptexpcal valfen, BackAnnotationType refTree)
{
  compilerInteger anInt;
  TypeType t1, tb1;
  ptexpcal vinit, sigret, dim;
  boolean b;

  isGreaterOrEqual_or_Greater_than_0 (valret, refTree, &anInt, &b, pk_true, pk_true);
  // anInt = 0 : The delay must be removed.
  if (ptexpcal_get_operatorClass (vexpcal1) == variabl) 
    {
      sigret = vexpcal1; 
      t1 = ptexpcal_get_type (sigret);
    }
  else
    { // opinit
      sigret = ptexpcal_get_binaryArgstAt (vexpcal1, 1);
      t1 = ptexpcal_get_type (sigret);
      if (t1 == PureType)
	{
	  vinit = ptexpcal_get_binaryArgstAt (vexpcal1, 2);
	  t1 = getDimAndBaseType (ptexpcal_get_type (vinit), &dim);
	  // if (t1 ==  PureType) delay of an event !!!
	}
    }
  if (valfen != NILEXPCAL)
    {
      isGreaterOrEqual_or_Greater_than_0 (valfen, refTree, &anInt, &b, pk_true, pk_true);
      dim = mkNaryExp (oplesign, valfen, NILEXPCAL, refTree);
      t1 = mkArrayType (dim, t1);
    }
  ptexpcal_set_type (vexp, t1);
  if (TypeTestBundleClass (t1))
    pK_messenger->errorTree (errExtensionNotYetImplemented,  BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
}

/**
   Type checking for an initial value (vinit) for a delayed signal (delayedExpr).
   On error, a message is hooked to refTree.

   \attention not implemented for the oplevect operator, only for oplelevec operator (see verifyEachElevec()).
*/
void
ExprManager::TypeCkeckingInitialValue (ptexpcal delayedExpr, ptexpcal vinit, BackAnnotationType refTree)
{
  noperators vop;

  if (vinit != NILEXPCAL)
    {
      vop = ptexpcal_get_operator (vinit);
      switch ((int) vop)
	{
	case oplelevec:
	  verifyEachElevec (delayedExpr, vinit, refTree);
	  break;
	case oplevect:
	  // TO DO
	  break;
	}
    }
}

/**
  Type checking for vexp := lb .. up step st. On error, a message is hooked to refTree.
  @return "true" if the expression is well type, "false" otherwise.
*/
boolean
ExprManager::TypeCheckingIndex (ptexpcal vexp, ptexpcal lb, ptexpcal ub, ptexpcal st, BackAnnotationType refTree)
{
  if (verifyIndexExpr (vexp, lb, ub, st, refTree))
    {
      setTypeIndex (vexp, lb, ub, st, refTree);
      return pk_true;
    }
  setErroneousType (vexp);
  return pk_false;
}


/**
  Verification on value.
  @param vexp (in) an expression.
  @param refTree (in) reference to the syntactic tree of the program ; used for annotations.
  @param strict (in) the value must be strictly positive if strict is pk_true
  @param bval (out) at pk_true if the value of the expression (val) is known, "false" otherwise.
  @param bmess (in) if pk_true a message if hooked to the tree (refTree) is the value (val) is not positive.
  @return 
      - "true" when the expression is undefined (nil), or is an int expression and its value is unknown or known and positive.
      - "false" otherwise.
*/
boolean
ExprManager::hasPositiveIntegerType (ptexpcal vexp, BackAnnotationType refTree, boolean strict, boolean * bval, boolean bmess)
{
  compilerInteger val;
  return isGreaterOrEqual_or_Greater_than_0 (vexp, refTree, &val, bval, bmess, pk_true);
}

/**
  Verification on value.
  @param vexp (in) an expression.
  @param refTree (in) reference to the syntactic tree of the program ; used for annotations.
  @param strict (in) the value must be strictly positive if strict is pk_true
  @param bval (out) at pk_true if the value of the expression (val) is known, "false" otherwise.
  @param bmess (in) if pk_true a message if hooked to the tree (refTree) is the value (val) is not positive.
  @return 
      - "true" when the expression is undefined (nil), or is an int expression and its value is unknown or known and positive.
      - "false" otherwise.
*/
boolean
ExprManager::hasIntegerValueEqualTo1 (ptexpcal vexp)
{
  compilerInteger val;
  boolean bval;
  if (isGreaterOrEqual_or_Greater_than_0 (vexp, NILBACKANNOTATION, &val, &bval, pk_false, pk_true)) return (val == 1);
  return pk_false;
}

/**
   It updates the type of vexp when TT a enumerated type and vexp is an access to a value of TT (updateEnumValue()).
   It is applied when the type of vexp is no fixed.
*/
void
ExprManager::TypePropagate (TypeType TT, ptexpcal vexp)
{
  if (TypeTestEnumerationClass (TT) && ptexpcal_IsFreeType (vexp))
    {
      updateEnumValue (vexp, TT, NILBACKANNOTATION);
    }
}

/**
   It sets (forces) TT as the type of an expression (exp) if its type is free. 
   A message (warning) is hooked to refTree ("no type control done for the exp expression").
*/
void
ExprManager::TypeCheckingForceType (ptexpcal exp, TypeType TT, BackAnnotationType refTree)
{
  if (setTypeIfFree (exp, TT, refTree)) pK_messenger->warningacc (warNoct,  BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
  /*  if (ptexpcal_IsFreeType(exp)) { ptexpcal_set_type( exp, TT); pK_messenger->warningacc(warNoct, refTree); } */
}

/**
    Type checking for ls:= id{lp}(le), the control is achieved for id an intrinsic function identifier (bounds, dim , size, indices).
    @return 
    - "true" if the expression is well typed, or id is not considered ,
    - "false" otherwise, an error message is hooked to vt.
*/
boolean
ExprManager::TypeCheckingIntrinsic (ptident id, ptexpcal lp, ptexpcal le, ptexpcal ls, BackAnnotationType vt)
{
  ptexpcal e1, e2, e3, s;
  TypeType TT, target;
  compilerInteger n;
  boolean b;

  if (isBoundsId (id))
    {
      // bounds : Scalar-type (not complex) or enumarated type
      s = le;
      ptexpcal_get_binaryArgs (le, &e1, &le);
      ptexpcal_get_binaryArgs (le, &e2, &le);
      ptexpcal_get_binaryArgs (le, &e3, &le);
      if (!TypeCheckingAccept (e1, e2, &TT, vt, pk_false, pk_false))
	return errUnexpectedType (s, vt);
      if (!TypeCheckingAccept (e1, e3, &TT, vt, pk_false, pk_false))
	return errUnexpectedType (s, vt);
      getNaryArgs (&ls, &s);
      if (!TypeCheckingAccept (e1, s, &TT, vt, pk_false, pk_false))
	return errUnexpectedType (s, vt);
      if (TypeTestComplexNumber (TT))
	return errUnexpectedType (s, vt);
      // conserver les bornes.
      if (ptexpcal_IsAssigned (e2) && ptexpcal_IsAssigned (e3))
	ptexpcal_set_integerBoundsHolder (s, ptexpcal_get_integerValue (e2), ptexpcal_get_integerValue (e3));
      return pk_true;
    }
  // dim : array ; for Scalar-type, external type or enumarated type 
  if (isDimId (id))
    {
      ptexpcal_get_binaryArgs (le, &e1, &le);
      if (ptexpcal_IsFreeType (e1))
	return pk_true;
      TT = ptexpcal_getUnamedType (e1);
      if (!TypeTestPredef (TT))
	if (!TypeTestArrayClass (TT))
	  return errUnexpectedType (e1, vt);
      return pk_true;
    }
  // size(T,i) defined for  1 <= i <= dim(T)
  if (isSizeId (id))
    {
      ptexpcal_get_binaryArgs (le, &e1, &le);
      // second arg must be a int positive value.
      ptexpcal_get_binaryArgs (le, &e2, &le);
      isGreaterOrEqual_or_Greater_than_0 (e2, vt, &n, &b, pk_true, pk_true);
      //  first arg.
      if (ptexpcal_IsFreeType (e1))
	return pk_true;
      TT = ptexpcal_getUnamedType (e1);
      if (!TypeTestArrayClass (TT))
	return errUnexpectedType (e1, vt);
      return pk_true;
    }
  if (isIndicesId (id))
    {
      TT = mkArrayType (le, IntegerType);
      while (getNaryArgs (&ls, &s))
	if (ptexpcal_IsFreeType (s))
	  setType (s, TT);
	else
	  {
	    target = ptexpcal_getUnamedType (s);
	    if (!TypeTestArrayClass (target))
	      return errUnexpectedType (s, vt);
	    if (!ArrayTypeVerifSize (TT, target))
	      return errUnexpectedType (s, vt);
	  }
    }
  return pk_true;
}

/**  
     Type checking for vexp a synchronous unary expression vope(vexpcal1).
     @return  "true" if the expression is well typed, "false" otherwise, an error message is hooked to refTree.
     It calls TypeCheckingMatrixTr, TypeCheckingnega, TypeCheckingCartesianProduct.
*/
boolean
ExprManager::TypeCheckingUnary (noperators vope, ptexpcal vexp, ptexpcal vexpcal1, BackAnnotationType refTree)
{
  switch ((int) vope)
    {
    case optran:		// tr()
      return TypeCheckingMatrixTr (vexp, vexpcal1, refTree);

    case opvvarin:		// ?x
    case opsdec:		// x[?] (valeur de x à l'iteration precedente) 
      ptexpcal_set_type (vexp, ptexpcal_get_type (vexpcal1));
      break;

    case opnega:
      TypeCheckingnega (vexp, vexpcal1, refTree);
      break;

    case opnon:
      /* pas d'occurrences (filtre avant) */
      break;

    case oppcar:		// Cartesian product
      TypeCheckingCartesianProduct (vexp, vexpcal1, refTree);
      break;

      //    case opCurrentLustre: ptexpcal_set_type(vexp, ptexpcal_get_type(vexpcal1)); break;
    case opformact:
    case opformnode:
    case opformfunc:
    case opformproc:
      break;
    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
  return pk_true;
}

/**
   @return
   - "true" if the definition vTuple := vDef, where vDef is a cartesian product is well typed
   - "false" otherwise, in this case the tree (refTree) is annotated with the error message.   
*/
boolean
ExprManager::TypeCheckingCartesianProductEquation (ptexpcal vTuple, ptexpcal vDef, BackAnnotationType refTree)
{
  ptexpcal v;
  TypeType TT;

  getNaryArgs (&vTuple, &v);
  if (!TypeCheckingAccept (v, vDef, &TT, refTree, pk_false, pk_false))
    return errUnexpectedType (v, refTree);
  if (getNaryArgs (&vTuple, &v))
    if (!TypeCheckingAccept (v, vDef, &TT, refTree, pk_false, pk_false))
      return errUnexpectedType (v, refTree);
  return pk_true;
}

/** 
    It validates the type of vexp defined by a synchronous binary expression vope(vexpcal1, vexpcal2).
    On error, messages are hooked to refTree tree.
    @return vexp or the evaluated equivalent expression. 
    It calls : TypeVerifFieldsAccess, TypeVerifCounter, TypeVerifException, TypeVerifComplex, TypeCheckingRelation1, TypeCheckingRelationEqEq
    TypeCheckingRelationIIEQ, TypeCheckingRelation2, TypeCheckingArithmBinary, evalarithBinary, TypeCheckingelem, TypeCheckingparc,
    TypeCheckingconc, TypeCheckingRepeat, TypeCheckingMatrixMult, TypeCheckingNext.
    
 */
ptexpcal
ExprManager::TypeCheckingBinary (noperators vope, ptexpcal vexp, ptexpcal vexpcal1, ptexpcal vexpcal2, BackAnnotationType refTree)
{
  switch ((int) vope)
    {
    case opcham: // vexpcal1.vexpcal2
      vexp = TypeVerifFieldsAccess (vexp, vexpcal1, vexpcal2, refTree);
      break;

    case opnombm: // vexpcal1 count vexpcal2
      vexp = TypeVerifCounter (vexp, vexpcal1, vexpcal2, refTree);
      break;
    case opinit:
      break;
    case oprecu: // vexpcal1 \\ vexpcal2
      TypeVerifException (vexp, vexpcal1, vexpcal2, refTree);
      break;
    case opcplx: // vexpcal1 @ vexpcal2
      TypeVerifComplex (vexp, vexpcal1, vexpcal2, refTree);
      break;
    case opappe:
      ptexpcal_set_type (vexp, ptexpcal_get_type (vexpcal1));
      break;

    case opdiff:
    case opegal:
      TypeCheckingRelation1 (&vexp, vexpcal1, vexpcal2, refTree);
      break;
    case opegeg:		// ==
      TypeCheckingRelationEqEq (&vexp, vexpcal1, vexpcal2, refTree);
      break;
    case opiieg:		// <<=
      TypeCheckingRelationIIEQ (&vexp, vexpcal1, vexpcal2, refTree);
      break;

    case opiegal:
    case opinfe:
    case opsegal:
    case opsupe:
      TypeCheckingRelation2 (&vexp, vexpcal1, vexpcal2, refTree);
      break;

    case opmod:

    case opdivi:
    case opmult:
    case oppuis:
    case opplus:
    case opmoin:
      if (TypeCheckingArithmBinary (vope, vexp, vexpcal1, vexpcal2, refTree))
	{
	  vexp = evalarithBinary (vope, vexp, vexpcal1, vexpcal2, refTree);
	}
      break;
    case opelem:
      TypeCheckingelem (vexp, vexpcal1, vexpcal2, refTree);
      break;
    case opet:
    case opou:
      break;
    case opparc:
      TypeCheckingparc (vexp, refTree);
      break;
    case opelevec:
      // TypeCheckingelevec(vexp, vexpcal1, vexpcal2, refTree);
      ptexpcal_set_type (vexp, ptexpcal_get_type (vexpcal2));	// control on the list (oplelevec)
      break;
    case opensvec:
      ptexpcal_set_type (vexp, ptexpcal_get_type (vexpcal2));
      break;

    case opleft_tt:
      ptexpcal_set_type (vexp, ptexpcal_get_type (vexpcal1));
      break;

    case opvenum:		// <type> # <valeur_type_enumere>
      ptexpcal_set_type (vexp, ptexpcal_get_exprType (vexpcal1));
      break;

    case opentsor:		// expression créee pour un morphisme
      break;

    case opconc:
      TypeCheckingconc (vexp, vexpcal1, vexpcal2, refTree);
      break;

    case opsequ:		// vexpcal1 next vexpcal2.
      TypeCheckingNext (vexp, vexpcal1, vexpcal2, refTree);
      break;

    case oprepe:		/* e1 |* e2 , repeat e1 e2 times */
      vexp = TypeCheckingRepeat (vexp, vexpcal1, vexpcal2, refTree);
      break;

    case opmmul:		/* Matrix product *. */
      TypeCheckingMatrixMult (vexp, vexpcal1, vexpcal2, refTree);
      break;

    case oprest:		// done in TypeCheckingEquation
      break;

      // case opwhenLustre: ptexpcal_set_type(vexp, ptexpcal_get_exprType(vexpcal1)); break;

    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
  return vexp;
}


/**
   @return the upper bound of an index (vind).

   On error, messages are hooked to refTree tree.
 */
ptexpcal
ExprManager::TypeCheckingGetUpperBoundIterationIndex (ptexpcal vind, BackAnnotationType refTree)
{
  ptexpcal vx, aux, upperBound;
  boolean blist;
  TypeType TT;

  if (satisfiesList (vind))
    {				// Tuple
      aux = vind;
      getNaryArgs (&aux, &vx);
      blist = pk_true;
    }
  else
    {
      vx = vind;
      blist = pk_false;
    }

  if (!getIthDimension (vx, 1, &upperBound))
    {
      pK_messenger->errorTree (errTypeIndex,  BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
      return mkIntegerConstant (0);
    }
  upperBound = mkBinaryExp (opmoin, upperBound, mkIntegerConstant (1), refTree);
  if (ptexpcal_get_nbDim (vx) != 1)
    pK_messenger->errorTree (errTypeIndex,  BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
  else if (blist)
    {
      TT = ptexpcal_getUnamedType (vx);
      while (getNaryArgs (&aux, &vx))
	if (TT != ptexpcal_getUnamedType (vx))
	  pK_messenger->errorTree (errTypeIndex,  BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
    }
  return upperBound;
}


/**
   For an equation vl := vr, it assigns to vl the values of vr taking into account the values of the initial value (vinit).
*/
void
ExprManager::setValuesEq (ptexpcal vl, ptexpcal vr, noperators op)
{
  compilerInteger lb1, ub1, lb2, ub2;
  ptexpcal vinit;
  boolean b;

  if (vr != NILEXPCAL)
    {
      if (op == opnil)
	op = getOperator (vr);
      if ((op == opmemo) || (op == opret))
	{
	  vinit = getInitialValue (vl);
	  vr = getBasicExpr (vr);
	  b = (vinit != NILEXPCAL);
	}
      else
	{
	  b = pk_true;
	  vinit = NILEXPCAL;
	}
      if (b)
	if (getBounds (vr, &lb1, &ub1))
	  {
	    if (getBounds (vinit, &lb2, &ub2))
	      {
		lb1 = (lb1 < lb2 ? lb1 : lb2);
		ub1 = (ub1 > ub2 ? ub1 : ub2);
	      }
	    setBounds (vl, lb1, ub1);
	  }
    }
}
