/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   @return "true" if a not typed expression (exp), "false" otherwise (typed)
   When it is not typed, the type (t) is associated with it.
   On error, a message is hooked to the backannotation trees vt.
*/
  boolean ExprManager::setTypeIfFree (ptexpcal exp, TypeType t, BackAnnotationType vt)
{
  if (!ptexpcal_IsFreeType (exp)) return pk_false;
  ptexpcal_set_type (exp, t);
  TypePropagateEnumValue (exp, t, vt);
  return pk_true;
}

/**
   @return "true" if TT is the predefined int type, "false" otherwise.
*/
boolean
ExprManager::isIntegerType (TypeType TT)
{
  if (TypeTestPredef (TT))
    return TypeTestIntNumber (TT);
  return pk_false;
}

/**
   It assigns the error type to an expression (vexp).
*/
void
ExprManager::setErroneousType (ptexpcal vexp)
{
  if (ptexpcal_get_operatorClass (vexp) != variabl)
    ptexpcal_set_type (vexp, get_ErrorType ());
}

/** 
   Implementation of Numerical/Logical expressions extension. 
   @return "true" if the  operand expression vexp is not a bundle/structure type, "false" otherwise
*/
boolean
ExprManager::acceptNLExtensions (TypeType t)
{
  if (TypeTestBundleClass (t))
    return pk_false;
  if (TypeTestStructureClass (t))
    return pk_false;
  if (TypeTestArrayClass (t))
    return acceptNLExtensions (ArrayMatrixTypeGetTerm (t));
  return pk_true;
}

/** 
   Implementation of Numerical/Logical expressions extension 
   @return "true" if the  operand expression vexp is not a bundle/structure type, "false" otherwise
*/
boolean
ExprManager::ExpcalTestNLExtensions (ptexpcal vexp)
{
  if (ptexpcal_IsFreeType (vexp))
    return pk_true;
  return acceptNLExtensions (ptexpcal_getUnamedType (vexp));
}

/**
   @return "true" if each element of d1 are equivalent to each element of d2.
   @param d1 a list of expressions (assumed)
   @param d2 a list of expressions (assumed). d2 may have more elements than d1.
*/
boolean
ExprManager::EqForEachDimensionOf (ptexpcal d1, ptexpcal d2)
{
  ptexpcal aDim1, aDim2;
  boolean b;

  while (ptexpcal_get_Nary (&d1, &aDim1))
    if (ptexpcal_get_Nary (&d2, &aDim2))
      {
	if (ptexpcal_IsAssigned (aDim1) && ptexpcal_IsAssigned (aDim2))
	  {
	    if (!ptexpcal_equalValues (aDim1, aDim2))
	      return pk_false;
	  }
	else
	  VerifForEachDimensionOf = pk_false;
	// if (!ptexpcal_isEquivalent(aDim1, aDim2, pk_true)) return pk_false;
      }
  return pk_true;
}

/**
   @return "true" if the expression d1, d2 (dimensions) are equivalent, "false" otherwise.
*/
boolean
ExprManager::ptct_belongs_equaltc (ptexpcal d1, ptexpcal d2)
{
  if (ptexpcal_get_nbelem (d1) != ptexpcal_get_nbelem (d2))
    return pk_false;
  return EqForEachDimensionOf (d1, d2);
}

/** It hooks the errnum-th message error, parametrized by nom, to a tree (vtree). */
void
ExprManager::erreurattachnom (int errnum, ntrees vtree, ptexpcal nom)
{
  pK_messenger->errorTreeIdent (errnum, vtree, ptexpcal_get_treeIdentifier (nom));
}

/** It hooks the errnum-th message error, parametrized by nom, to a tree (vtree). */
void
ExprManager::erreurattachnom (int errnum, BackAnnotationType vtree, ptexpcal nom)
{
  erreurattachnom (errnum, BACK_ANNOT_GET_ABSTRACT_TREE(vtree), nom);
  // pK_messenger->errorTreeIdent (errnum, BACK_ANNOT_GET_ABSTRACT_TREE(vtree), ptexpcal_get_treeIdentifier (nom));
}

/**
   It hooks the errnum-th message error, parametrized by a signal (expsig), to a tree (vtree)
   and also to the declaration of expsig (ptexpcal_get_treeOfSignalDeclaration()).
*/
void
ExprManager::erreurattachref1nom (int errnum, BackAnnotationType vtree, ptexpcal expsig)
{
  pK_messenger->errorTreeRefIdent (errnum, BACK_ANNOT_GET_ABSTRACT_TREE(vtree), ptexpcal_get_treeOfSignalDeclaration (expsig), ptexpcal_get_treeIdentifier (expsig));
}

/**
   It hooks to a tree (vtree) the errnum-th message error, parametrized by  two signals (vexpcal1, vexpcal2).
   It is also signaled to the declarations of vexpcal1 and vexpcal2 (ptexpcal_get_treeOfSignalDeclaration()).
*/
void
ExprManager::erreurattachref2nom (int errnum, BackAnnotationType vtree, ptexpcal vexpcal1, ptexpcal vexpcal2)
{
  pK_messenger->errorTree2RefIdent (errnum, BACK_ANNOT_GET_ABSTRACT_TREE(vtree), ptexpcal_get_treeOfSignalDeclaration (vexpcal1),
				    ptexpcal_get_treeOfSignalDeclaration (vexpcal2), ptexpcal_get_treeIdentifier (vexpcal1));
}

/** 
    It prints the warnum-th warning parametrized by a signal (vexpcal).
*/
void
ExprManager::warningdisnom (int warnum, ptexpcal vexpcal)
{
  pK_messenger->warningident (warnum, ptexpcal_get_treeIdentifier (vexpcal));
}



/** It hooks to a tree (vtree) the warnum-th  warning  parametrized by a signal (vexpcal). */
void
ExprManager::warningattachnom (int warnum, BackAnnotationType vtree, ptexpcal vexpcal)
{
  pK_messenger->warningaccnom (warnum, BACK_ANNOT_GET_ABSTRACT_TREE(vtree), ptexpcal_get_treeIdentifier (vexpcal));
}


/**
   @return "true" if an expression (vexp) is a synchronisation expression (logical type), "false" otherwise.
*/
boolean
ExprManager::verifySynchroType (ptexpcal vexp)
{
  if (ptexpcal_IsFreeType (vexp))
    return pk_true;
  return TypeTestLogical (ptexpcal_get_type (vexp));
}

/**
   @return "true" if an expression (vexp) is a numerical expression, "false" otherwise.
*/
boolean
ExprManager::verifyNumericType (ptexpcal vexp)
{
  if (ptexpcal_IsFreeType (vexp))
    return pk_true;
  return TypeTestNumerical (ptexpcal_get_type (vexp));
}

/**
   @return "true" if an expression (vexp) is a text expression, "false" otherwise.
*/
boolean
ExprManager::verifyTextType (ptexpcal vexp)
{
  if (ptexpcal_IsFreeType (vexp))
    return pk_true;
  return TypeTestText (ptexpcal_get_type (vexp));
}

/**
   @return "true" if an expression (vexp) is a complex expression, "false" otherwise.
*/
boolean
ExprManager::verifyComplexType (ptexpcal vexp)
{
  return TypeTestComplexNumber (ptexpcal_get_type (vexp));
}

/**
   @return "true" if T denotes a scalar type, "false" otherwise.
*/
boolean
ExprManager::verifyScalarType (TypeType T)
{
  if (TypeTestLogical (T))
    return pk_true;
  if (TypeTestNumerical (T))
    return pk_true;
  if (TypeTestText (T))
    return pk_true;
  return pk_false;
}

/**
   @return "true" if an expression (vexp) is a scalar expression, "false" otherwise.
*/
boolean
ExprManager::verifyScalarType (ptexpcal vexp)
{
  TypeType T;
  if (ptexpcal_IsFreeType (vexp))
    return pk_true;
  T = ptexpcal_get_type (vexp);
  return verifyScalarType (T);
}


boolean ExprManager::setPureType (ptexpcal vexpcal)
{
  TypeType
    t;
  t = ptexpcal_get_type (vexpcal);
  if (t == PureType)
    return pk_true;
  if ((t == BooleanType) && isMarked (vexpcal, DONT_CHANGE_TYPE))
    return pk_false;
  ptexpcal_set_type (vexpcal, PureType);
  return pk_true;
}

/**
   @return "true" if the type of an expression (vexpcal1) is event, "false" otherwise.
*/
boolean
ExprManager::TypeChecking_convert2event (ptexpcal vexpcal)
{
  if (modeNoEvents)
    return pk_false;
  if (!ptexpcal_IsBoolean (vexpcal))
    return pk_false;
  if (!ptexpcal_IsAssigned (vexpcal))
    return pk_false;
  if (!ptexpcal_equalTrue (vexpcal))
    return pk_false;
  return setPureType (vexpcal);
}

/**
   @return 
   - "true" if an expression is not a reference to a value of an enumerated type (vType), or it is one well typed (TypeCheckingControlValEnum()),
   - "false" otherwise, a message is hooked to the tree ar.
*/
boolean
ExprManager::TypePropagateEnumValue (ptexpcal vexpcal, TypeType vType, BackAnnotationType ar)
{
  ptrepstr val;

  if (ptexpcal_get_operator (vexpcal) != opvenum)
    return pk_true;
  ptexpcal_set_exprType (ptexpcal_get_binaryArgstAt (vexpcal, 1), vType);

  if (!TypeTestEnumerationClass (vType))
    return pk_true;
  ptexpcal_set_type (vexpcal, vType);
  val = ptexpcal_get_constant (ptexpcal_get_binaryArgstAt (vexpcal, 2)); // PLG WARNING should be ptexpcal_get_ident ?
  if (IS_NILBACKANNOTATION (ar)) ar = getBackAnnotation (vexpcal);
  return TypeCheckingControlValEnum (vType, val, ar);	// PLG WARNING val should be ptident
}

/**
   @return 
   - "true" if the operand types (source, target) are compatible, result in t. The implicit conversions are defined in TypeImplicitConversionPredef().
   - "false" otherwise, t is not modified.
*/
boolean
ExprManager::unifytctype (TypeType source, TypeType target, TypeType * t)
{
  int n;

  //  -- A completer--
  *t = target;
  if (source == target)
    return pk_true;
  n = TypeCastType (source, target);
  // n =-1 impossible, 0 not decidable, 
  // 1 source can be converted to target without explicit internal conversions(restrictions for SIGNAL)
  if (n != 1)
    return pk_false;
  if (TypeTestPredef (source) && TypeTestPredef (target))
    return TypeImplicitConversionPredef (source, target);
  return pk_true;		// je suppose
}


/**
    @return 
    - "true" if the assumed types target and source, assumed to be the type of vexpcal2, are compatible, result in vType.
    - "false" otherwise, vType is not modified.  On error, a message is hooked to the tree vt.
*/
boolean
ExprManager::acceptTypeCheckingStructTuple (TypeType target, ptexpcal vexpcal2, TypeType source, TypeType * vType, BackAnnotationType vt)
{
  TypeList l1, l2;
  TypeType t1, t2, t;
  TypeSubList SubLst1, SubLst2;
  ptexpcal velem;

  (*vType) = target;
  if (target == source)
    return pk_true;
  if (ptexpcal_get_operator (vexpcal2) != opleflot)
    return pk_false;

  l1 = TupleTypeGetTypeLst (source);
  l2 = TupleTypeGetTypeLst (target);
  SubLst1 = TypeListCast2SubList (l1);
  SubLst2 = TypeListCast2SubList (l2);
  while (pk_true)
    {
      ptexpcal_get_Nary (&vexpcal2, &velem);
      t1 = TypeSubListGetHead (SubLst1);
      t2 = TypeSubListGetHead (SubLst2);
      if (TypeTestArrayClass (t1) && TypeTestArrayClass (t2))
	acceptTypeCheckingArrayArray (NILEXPCAL, t2, velem, t1, vType, vt);
      else if (!unifytctype (t1, t2, &t))
	{
	  if (!setTypeIfFree (velem, t, vt))
	    return pk_false;
	  /*if (!ptexpcal_IsFreeType(velem)) return pk_false;
	     ptexpcal_set_type(velem, t);
	     TypePropagateEnumValue(velem, t, vt); */
	}
      if (TypeSubListTestLast (SubLst1))
	{
	  if (TypeSubListTestLast (SubLst2))
	    return pk_true;
	  return pk_false;
	}
      else if (TypeSubListTestLast (SubLst2))
	return pk_false;
      SubLst1 = TypeSubListGetTail (SubLst1);
      SubLst2 = TypeSubListGetTail (SubLst2);
    }
}

/**
   @return "true" if the syze of two Array types (source, target) are equal (ptct_belongs_equaltc()), "false" otherwise.
*/
boolean
ExprManager::ArrayTypeVerifSize (TypeType source, TypeType target)
{
  ptexpcal d1, d2;
  d1 = ArrayMatrixTypeGetDimLst (source);
  d2 = ArrayMatrixTypeGetDimLst (target);
  return ptct_belongs_equaltc (d1, d2);
}

/**
   @return 
   - "true" if the type of two expressions (vexpcal1 and vexpcal2), with resp. target/source as type are compatible, result in vType.
   It updates the type of an expression (vexpcal1 or vexpcal2) when it was undefined (any type) and the result is a array of enumerated type 
   (updateEnumValue()).
   - "false" otherwise. vType is not modified. On error, a message is hooked to the tree vt.
*/
boolean
  ExprManager::acceptTypeCheckingArrayArray (ptexpcal vexpcal1, TypeType target, ptexpcal vexpcal2, TypeType source, TypeType * vType, BackAnnotationType vt)
{
  TypeType eSource, eTarget, t;

  if (ArrayTypeVerifSize (source, target))
    {
      eSource = ArrayMatrixTypeGetTerm (source);
      eTarget = ArrayMatrixTypeGetTerm (target);
      if (TypeTestNamedClass (eTarget))
	eTarget = NamedTypeGetTerm (eTarget);
      if (TypeTestNamedClass (eSource))
	eSource = NamedTypeGetTerm (eSource);

      if (unifytctype (eSource, eTarget, &t))
	{
	  if (t == eSource)
	    *vType = source;
	  else
	    *vType = target;
	  return pk_true;
	}

      if (vexpcal2 != NILEXPCAL)
	if (TypeTestAnyClass (eSource))
	  {
	    *vType = target;
	    ptexpcal_set_type (vexpcal2, *vType);
	    if (TypeTestEnumerationClass (eTarget))
	      updateEnumValue (vexpcal2, eTarget, vt);
	    return pk_true;
	  }

      if (vexpcal1 != NILEXPCAL)
	if (TypeTestAnyClass (eTarget))
	  {
	    *vType = source;
	    ptexpcal_set_type (vexpcal1, *vType);
	    if (TypeTestEnumerationClass (eSource))
	      updateEnumValue (vexpcal1, eSource, vt);
	    return pk_true;
	  }
    }
  else				// for lelevec we accept.
  if (ptexpcal_get_operator (vexpcal2) == oplelevec)
    {
      eSource = ArrayMatrixTypeGetTerm (source);
      eTarget = ArrayMatrixTypeGetTerm (target);
      if (TypeTestNamedClass (eTarget))
	eTarget = NamedTypeGetTerm (eTarget);
      if (TypeTestNamedClass (eSource))
	eSource = NamedTypeGetTerm (eSource);

      if (unifytctype (eSource, eTarget, &t))
	{
	  pK_messenger->warningacc (warNoct, BACK_ANNOT_GET_ABSTRACT_TREE(vt));
	  *vType = target;
	  return pk_true;
	}
    }
  return pk_false;
}

/** 
    @return 
    - "true" if the type of vexpcal1 is compatible with the type of vexpcal2, result in vType. 
    - "false" otherwise,  vType is not modified. On error, a message is hooked to the tree vt.

    If one of the types is free, the other is the resulting type and it is set to the untyped expression,
    otherwise the compatibility of  the types is tested (TypeControlExpcal2()).
    The array types are accepted when barray, the structured types when bstruct.
    A type of an expression is updated (TypePropagateEnumValue()) when it is free.
*/
boolean
ExprManager::TypeCheckingAccept (ptexpcal vexpcal1, ptexpcal vexpcal2, TypeType * vType, BackAnnotationType vt, boolean barray, boolean bstruct)
{
  boolean b1, b2;
  TypeType target;

  b1 = ptexpcal_IsFreeType (vexpcal1);
  b2 = ptexpcal_IsFreeType (vexpcal2);
  if (b1)
    {
      if (ptexpcal_IsEvent (vexpcal2))
	*vType = get_BooleanType ();
      else
	*vType = ptexpcal_get_type (vexpcal2);
      ptexpcal_set_type (vexpcal1, (*vType));
      TypePropagateEnumValue (vexpcal1, *vType, vt);
      return pk_true;
    }
  if (b2)
    {
      (*vType) = ptexpcal_get_type (vexpcal1);
      ptexpcal_set_type (vexpcal2, (*vType));
      TypePropagateEnumValue (vexpcal2, *vType, vt);
      return pk_true;
    }
  if (ptexpcal_IsEvent (vexpcal1) && TypeChecking_convert2event (vexpcal2))
    {
      (*vType) = ptexpcal_get_type (vexpcal1);
      return pk_true;
    }
  target = ptexpcal_getUnamedType (vexpcal1);
  return TypeControlExpcal2 (vexpcal1, target, vexpcal2, vType, vt, barray, bstruct);
}

/** 
    @return 
    - "true" if the type of vexpcal1 is compatible with the type of vexpcal2, result in vType. The operand types are assumed not free types.
    The array types are accepted when barray, the structured types when bstruct.
    
    - "false" otherwise,  vType is not modified. On error, a message is hooked to vt.
*/
boolean ExprManager::TypeControlExpcal2 (ptexpcal vexpcal1, TypeType target, ptexpcal vexpcal2, TypeType * vType, BackAnnotationType vt, boolean barray, boolean bstruct)
{
  TypeList l;
  TypeSubList SubLst;
  TypeType source, tupleType;

  source = ptexpcal_getUnamedType (vexpcal2);
  if (target == source)
    {
      (*vType) = target;
      return pk_true;
    }

  if (barray)
    if (TypeTestArrayClass (source) && TypeTestArrayClass (target))
      return acceptTypeCheckingArrayArray (vexpcal1, target, vexpcal2, source, vType, vt);
  if (bstruct)
    if (TypeTestStructureClass (target))
      {
	tupleType = StructureTypeGetArgument (target);
	if (TypeTestTupleClass (source))
	  return acceptTypeCheckingStructTuple (tupleType, vexpcal2, source, vType, vt);
	(*vType) = target;
	// we accept  an element if the structure has one field.
	if (!TypeTestPredef (source))
	  return pk_false;
	l = TupleTypeGetTypeLst (tupleType);
	SubLst = TypeListCast2SubList (l);
	if (!TypeSubListTestLast (SubLst))
	  return pk_false;
	target = TypeSubListGetHead (SubLst);
      }
  return unifytctype (source, target, vType);
}

/**
   @return for a type (T), its base type if T is an array type, T otherwise.
   @param T (in) a type.
   @param r (out) the expression of its dimensions if T is an array type, NILEXPCAL otherwise.
 */
TypeType
ExprManager::getDimAndBaseType (TypeType T, ptexpcal * r)
{
  if (TypeTestArrayClass (T))
    {
      (*r) = ArrayMatrixTypeGetDimLst (T);
      return ArrayMatrixTypeGetTerm (T);
    }
  (*r) = NILEXPCAL;
  return T;
}

/**
   For an expression (exp), it returns 
   - r the expression of its dimensions, 
   - Tr its base type.
 */
void
ExprManager::getDimAndType (ptexpcal exp, ptexpcal * r, TypeType * Tr)
{
  (*Tr) = getDimAndBaseType (ptexpcal_getUnamedType (exp), r);
}


/**   
      @return 
      - "true" if the type of source is compatible with the type of target, when source is an array restructuration.      
      - "false" otherwise. On error, a message is hooked to refTree.
*/
boolean
ExprManager::TypeCheckingElevecEnsvec (ptexpcal target, ptexpcal source, BackAnnotationType refTree)
{
  int nWait, n;
  ptexpcal l, tElemSource, dims;
  noperators vop;
  TypeType tElemTarget, vType;

  vop = ptexpcal_get_operator (source);
  if ((vop != opensvec) && (vop != opelevec))
    return pk_true;

  getDimAndType (target, &dims, &tElemTarget);

  nWait = ptexpcal_get_nbelem (dims);
  l = ptexpcal_get_binaryArgstAt (source, 1);
  n = ptexpcal_get_nbelem (l);
  if (n > nWait)
    {
      erreurattachref1nom (errctaff1, refTree, target);
      return pk_false;
    }
  tElemSource = ptexpcal_get_binaryArgstAt (source, 2);
  if (TypeControlExpcal2 (NILEXPCAL, tElemTarget, tElemSource, &vType, refTree, pk_false, pk_true))
    return pk_true;
  pK_messenger->errorTree (errElemType, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
  return pk_false;
}


/**
   @return 
      - "true" if the type of vexpcal2 is compatible with the type of vexpcal1, when vexpcal2 is a list of array restructurations.
      It calls TypeCheckingElevecEnsvec() for each element.
      - "false" otherwise. On error, a message is hooked to refTree.
*/
boolean
ExprManager::TypeCheckingEquation_lelevec (ptexpcal vexpcal1, ptexpcal vexpcal2, BackAnnotationType refTree)
{
  ptexpcal listindices, valeur;
  boolean vResult;

  vResult = pk_true;
  while ((vexpcal2 != NILEXPCAL) && vResult)
    {
      ptexpcal_get_binaryArgs (vexpcal2, &listindices, &vexpcal2);
      if (ptexpcal_get_operator (listindices) != opelevec)
	{
	  ptexpcal_get_binaryArgs (listindices, &listindices, &valeur);
	  if (ptexpcal_get_operator (valeur) == opelevec)
	    ptexpcal_get_binaryArgs (valeur, &listindices, &valeur);
	}
      vResult = TypeCheckingElevecEnsvec (vexpcal1, listindices, refTree);
    }
  return vResult;
}

/**
   It verifies the compatibility of the types of vexpcal1 (target) and  vexpcal2 (source).
   It calls TypeCheckingEquation_lelevec() if vexpcal2 is a list of array restructurations, otherwise
   if verifies that the dimensions are compatible.  
   On error, a message is hooked to refTree.
   void
   ExprManager::TypeCheckingEquation_extensions(ptexpcal vexpcal1, ptexpcal vexpcal2, BackAnnotationType refTree)
   {
   ptexpcal dimRes, dimDef;
   int n,m;
   noperators op;
   
   op = ptexpcal_get_operator(vexpcal2);
   if (op == oplelevec)
   TypeCheckingEquation_lelevec(vexpcal1, vexpcal2, refTree);
   else 
   {
   dimRes = ptexpcal_get_dimension(vexpcal1);
   dimDef = ptexpcal_get_dimension(vexpcal2);
   n = ptexpcal_get_nbelem(dimDef);
   m = ptexpcal_get_nbelem(dimRes);
   if (n > m ) pK_messenger->errorTree(errctaff2, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
   }
   }
*/


/** @return "true" if an expression (vexpcal) accepts vType as type, "false" otherwise.  */
boolean
ExprManager::ExpcalTypeSet (ptexpcal vexpcal, TypeType vType)
{
  TypeType TT;

  TT = ptexpcal_get_type (vexpcal);
  if (TypeTestAbstract (TT))	// if (AnyTypeTestDefined(TT))  ???
    {
      ptexpcal_set_type (vexpcal, vType);
      return pk_true;
    };
  return unifytctype (TT, vType, &TT);
}


/**
   @return 
   - "true" if an expression (parc), assumed to be a travel expression (lb to ub step st), is well typed, 
   - "false" otherwise. On error, a message is hooked to refTree.
*/
boolean
ExprManager::TypeCheckingparc (ptexpcal parc, BackAnnotationType refTree)
{
  boolean bontype;
  ptexpcal bornes, borneinf, bornesup, pas, indice;

  indice = ptexpcal_get_binaryArgstAt (parc, 1);
  bornes = ptexpcal_get_binaryArgstAt (parc, 2);
  if (bornes == NILEXPCAL)
    {
      bontype = pk_true;
      ptexpcal_set_integerBoundsHolder (indice, 0, 0);
    }
  else
    {
      borneinf = ptexpcal_get_ternaryArgstAt (bornes, 1);
      bornesup = ptexpcal_get_ternaryArgstAt (bornes, 2);
      pas = ptexpcal_get_ternaryArgstAt (bornes, 3);
      bontype = verifyIndexExpr (indice, borneinf, bornesup, pas, refTree);
    }
  if (!bontype)
    {
      setErroneousType (parc);
      return pk_false;
    }
  // Il faudrait : type vecteur[bs] si "bounded"
  ptexpcal_set_type (parc, IntegerType);
  return pk_true;
}

/**
   @return "true" if the type of an expression (e) is a vector, with vsize as 
   size, of integers, "false" otherwise.
 */
boolean
ExprManager::isVectorOfIntegers (ptexpcal e, ptexpcal * vsize)
{
  TypeType TT, baseType;
  ptexpcal ldims;

  TT = ptexpcal_getUnamedType (e);
  if (!TypeTestArrayClass (TT))
    return pk_false;
  baseType = ArrayMatrixTypeGetTerm (TT);
  if (!isIntegerType (baseType))
    return pk_false;
  ldims = ArrayMatrixTypeGetDimLst (TT);
  getNaryArgs (&ldims, vsize);
  return (ldims == NILEXPCAL);
}

/**
   Type Checking for Cartesian Product expression (vres) defined by << voperand >>. 
   On error, a message is hooked to refTree.
*/
void
ExprManager::TypeCheckingCartesianProduct (ptexpcal vres, ptexpcal voperand, BackAnnotationType refTree)
{
  ptexpcal vsize, aVector, csize, ldims;
  TypeType TT;

  vsize = NILEXPCAL;
  while (getNaryArgs (&voperand, &aVector))
    if (!isVectorOfIntegers (aVector, &csize))	// Restrictions
      errUnexpectedType (aVector, refTree);
    else if (vsize == NILEXPCAL)
      vsize = csize;
    else
      vsize = mkBinaryExp (opmult, csize, vsize, refTree);
  ldims = mkNaryExp (oplesign, vsize, NILEXPCAL, NILBACKANNOTATION);
  TT = mkArrayType (ldims, IntegerType);
  ptexpcal_set_type (vres, TT);
}

/**
   @return "true" if the expression (vResult) defined by -vexp1 is well types, "false" otherwise.
   On error, a message is hooked to refTree.
*/
boolean
ExprManager::TypeCheckingnega (ptexpcal vResult, ptexpcal vexp1, BackAnnotationType refTree)
{
  compilerInteger bi, bs, bi1, bs1;

  if (!verifyNumericType (vexp1))
    return errUnexpectedType (vResult, refTree);
  if (!ExpcalTestNLExtensions (vexp1))
    return errExtension (vResult, refTree);
  ptexpcal_set_type (vResult, ptexpcal_get_type (vexp1));
  if (ptexpcal_IsAssigned (vexp1))
    ptexpcal_set_computedIntegerValue (vResult, computeValue (ptexpcal_get_operator (vResult), ptexpcal_get_integerValue (vexp1), 0));
  else if (ptexpcal_IsBounded (vexp1))
    {
      ptexpcal_get_integerBounds (vexp1, &bi1, &bs1);
      if (computeBounds (ptexpcal_get_operator (vResult), bi1, 0, bs1, 0, &bi, &bs))
	ptexpcal_set_integerBoundsHolder (vResult, bi, bs);
      else
	return setErrorTypeExp (vResult, errUnexpectedExpr, refTree);
    }
  return pk_true;
}

/**
   @return 
   - "true" if t1 an array type and t2 a predefined type or an array (extension), the types t2 and the base type of t are compatible, TT = t
   - "false" otherwise, TT is not modified.
*/
boolean
ExprManager::getTypeUpbForArray (TypeType t1, TypeType t2, TypeType * TT)
{
  TypeType tb1, tb2;
  ptexpcal d1, d2;

  tb1 = getDimAndBaseType (t1, &d1);
  if (d1 == NILEXPCAL)
    return pk_false;
  tb2 = getDimAndBaseType (t2, &d2);
  if (d2 != NILEXPCAL)
    {
      if (ptexpcal_get_nbelem (d2) > ptexpcal_get_nbelem (d1))
	return pk_false;
      if (!EqForEachDimensionOf (d2, d1))
	return pk_false;
    }
  if (unifytctype (tb2, tb1, TT))
    {
      *TT = t1;
      return pk_true;
    }
  return pk_false;
}


/**
   @return 
   - "true" if the types of two expressions (vexp1, vexp2) are compatible, result is returned in TT
   - "false" otherwise, TT is not modified.

   It calls TypeCheckingAccept,getTypeUpbForArray.
*/
boolean
ExprManager::getTypeUpb (ptexpcal vexp1, ptexpcal vexp2, TypeType * TT)
{
  TypeType t2, t1;
  if (TypeCheckingAccept (vexp1, vexp2, TT, NILBACKANNOTATION, pk_true, pk_true))
    return pk_true;
  if (TypeCheckingAccept (vexp2, vexp1, TT, NILBACKANNOTATION, pk_true, pk_true))
    return pk_true;
  t1 = ptexpcal_getUnamedType (vexp1);
  t2 = ptexpcal_getUnamedType (vexp2);

  if (getTypeUpbForArray (t1, t2, TT))
    return pk_true;
  return getTypeUpbForArray (t2, t1, TT);
}

/**
   @return 
   - "true" if the types of two expressions (vexp1, vexp2) are compatible (getTypeUpb), result is returned in TT
   - "false" otherwise, TT is not modified. A message is hooked to refTree and the type of vResult  is set to "error type" (errUnexpectedType).
*/
boolean
ExprManager::getTypeUpbSetError (ptexpcal vResult, ptexpcal vexp1, ptexpcal vexp2, BackAnnotationType refTree, TypeType * vTypeType)
{
  if (getTypeUpb (vexp1, vexp2, vTypeType))
    return pk_true;
  errUnexpectedType (vResult, refTree);
  return pk_false;
}

/**
   @return 
   - "true" if the type of vexp is not the complex type, 
   - "false" otherwise. A message is hooked to refTree and the type of vResult is set to "error type" (errUnexpectedType).
*/
boolean
ExprManager::verifComplexSetError (ptexpcal vexp, ptexpcal vResult, BackAnnotationType vtree)
{
  if (ptexpcal_IsFreeType (vexp))
    return pk_true;
  if (!verifyComplexType (vexp))
    return pk_true;
  errUnexpectedType (vResult, vtree);
  return pk_false;
}

/**
   @return 
   - "true" if the expression vResult defined by ope(vexp1, vexp2) is well typed.
   the evaluation of the expression is not achieved here (see evalarithBinary)
   - "false" otherwise. On error, a message is hooked to refTree 
*/
boolean
ExprManager::TypeCheckingArithmBinary (noperators ope, ptexpcal vResult, ptexpcal vexp1, ptexpcal vexp2, BackAnnotationType refTree)
{
  TypeType TT;
  boolean b;

  if (!verifyNumericType (vexp1))
    return errUnexpectedType (vResult, refTree);
  if (!verifyNumericType (vexp2))
    return errUnexpectedType (vResult, refTree);
  if (!ExpcalTestNLExtensions (vexp1))
    return errExtension (vResult, refTree);
  if (!ExpcalTestNLExtensions (vexp2))
    return errExtension (vResult, refTree);
  switch ((int) ope)
    {
    case opmod:
      // Il faudrait que vexp2 soit > 0 et borné supérieurement
      if (!TypeTestIntNumber (ptexpcal_get_type (vexp2)))
	return errUnexpectedType (vResult, refTree);
      if (TypeTestComplexNumber (ptexpcal_get_type (vexp1)))
	return errUnexpectedType (vResult, refTree);
      return pk_true;
      break;

    case oppuis:
      if (!TypeTestIntNumber (ptexpcal_get_type (vexp2)))
	return errUnexpectedType (vResult, refTree);
      return pk_true;
      break;
    }
  b = getTypeUpb (vexp1, vexp2, &TT);
  if (b)
    ptexpcal_set_type (vResult, TT);
  else
    {
      setErroneousType (vResult);
      pK_messenger->errorTree (errctarithm, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
    }
  return b;
}

/**
   @return 
   - "true" if an expression (e) is defined by (vexp1 = vexp2), or (vexp1 != vexp2) and 
            - they have the same external type,
            - they have the same enumeration type,
   - "false" otherwise.

   On error, a message is hooked to ar.
*/
boolean
ExprManager::AcceptEnumOrExternTest (ptexpcal e, ptexpcal vexp1, ptexpcal vexp2, BackAnnotationType ar)
{
  TypeType T2, T1;
  noperators op;
  boolean b1, b2;

  op = ptexpcal_get_operator (e);
  if (!((op == opegal) || (op == opdiff)))
    return pk_false;
  T1 = ptexpcal_getUnamedType (vexp1);
  T2 = ptexpcal_getUnamedType (vexp2);
  b1 = TypeTestEnumerationClass (T1);
  b2 = TypeTestEnumerationClass (T2);
  if ((!b1) && (!b2))
    {
      if (!TypeTestExternClass (T1))
	return pk_false;
      if (!TypeTestExternClass (T2))
	return pk_false;
      return (T1 == T2);
    }
  if (b1)
    if (b2)
      {
	if (T1 != T2)
	  errUnexpectedType (vexp1, ar);
	else;
      }
    else
      updateEnumValue (vexp2, T1, ar);
  else
    updateEnumValue (vexp1, T2, ar);

  return pk_true;
}


/**
   Type checking and evaluation of an expression vResult defined by OP(exp1, exp2), for OP in { =, /=}.
   The operands must have the same scalar-type domain. The evaluation is done for constants expressions: in this case, vResult is modified.
   On error, a message is hooked to refTree.
*/
void
ExprManager::TypeCheckingRelation1 (ptexpcal * vResult, ptexpcal vexp1, ptexpcal vexp2, BackAnnotationType refTree)
{
  TypeType vTypeType, tres;
  boolean b;

  tres = get_BooleanType ();
  if (!AcceptEnumOrExternTest (*vResult, vexp1, vexp2, refTree))
    {
      if (!verifyScalarType (vexp1))
	{
	  errUnexpectedType (*vResult, refTree);
	  return;
	}
      if (!verifyScalarType (vexp2))
	{
	  errUnexpectedType (*vResult, refTree);
	  return;
	}
      if (!getTypeUpbSetError (*vResult, vexp1, vexp2, refTree, &vTypeType))
	return;
      if (TypeTestArrayClass (vTypeType))
	{
	  tres = mkArrayType (ArrayMatrixTypeGetDimLst (vTypeType), tres);
	  b = pk_true;
	}
      else
	b = !testValueRelaOperator (ptexpcal_get_operator (*vResult), vexp1, vexp2, vResult);
    }
  else
    b = !testValueRelaOperator (ptexpcal_get_operator (*vResult), vexp1, vexp2, vResult);
  if (b)
    ptexpcal_set_type (*vResult, tres);
}


/** 
    Type Checking and evaluation  for an expression vResult defined by OP(exp1,exp2), for OP in { <=, <, >, >=}.
    The operands must have the same scalar-type domain but not complex type.
    The evaluation is done for constants expressions: in this case, vResult is modified.
    On error, a message is hooked to refTree.
*/
void
ExprManager::TypeCheckingRelation2 (ptexpcal * vResult, ptexpcal vexp1, ptexpcal vexp2, BackAnnotationType refTree)
{
  TypeCheckingRelation1 (vResult, vexp1, vexp2, refTree);
  if (verifComplexSetError (vexp1, *vResult, refTree))
    verifComplexSetError (vexp2, *vResult, refTree);
}

/** 
    Type Checking and evaluation for an expression vResult defined by OP(exp1,exp2), for OP in { == }.
    The operands must have the same type domain. The type of the vResult is boolean-type.
    On error, a message is hooked to refTree.    
*/
void
ExprManager::TypeCheckingRelationEqEq (ptexpcal * vResult, ptexpcal vexp1, ptexpcal vexp2, BackAnnotationType refTree)
{
  TypeType vTypeType, tres;

  if (!getTypeUpbSetError (*vResult, vexp1, vexp2, refTree, &vTypeType))
    return;
  tres = get_BooleanType ();
  ptexpcal_set_type (*vResult, tres);
}


/** 
    Type Checking and evaluation  for an expression vResult defined by OP(exp1,exp2), for OP in { <<= }.
    The operands must have the same scalar-type domain different of complex type.
    The type of the result is boolean type.
    On error, a message is hooked to refTree.  
*/
void
ExprManager::TypeCheckingRelationIIEQ (ptexpcal * vResult, ptexpcal vexp1, ptexpcal vexp2, BackAnnotationType refTree)
{
  TypeType vTypeType, tres;

  if (!verifyScalarType (vexp1))
    {
      errUnexpectedType (*vResult, refTree);
      return;
    }
  if (!verifyScalarType (vexp2))
    {
      errUnexpectedType (*vResult, refTree);
      return;
    }
  if (!getTypeUpbSetError (*vResult, vexp1, vexp2, refTree, &vTypeType))
    return;
  // The complexes are not allowed!!! 
  if (verifComplexSetError (vexp1, *vResult, refTree))
    verifComplexSetError (vexp2, *vResult, refTree);
  tres = get_BooleanType ();
  ptexpcal_set_type (*vResult, tres);
}

/**
   @return "true" if the dimensions of type of an expression (vexpcal) are equal to those of another expression (vdim1),
   "false" otherwise, a message is hooked to refTree.
*/
boolean
ExprManager::controlequadim (ptexpcal vexpcal, ptexpcal vdim1, BackAnnotationType refTree)
{
  ptexpcal vd;
  vd = ptexpcal_get_dimension (vexpcal);
  if (ptct_belongs_equaltc (vdim1, vd))
    return pk_true;
  pK_messenger->errorTree (errctaff2, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
  return pk_false;
}

/**
   @return "true" if a list of expressions (l) defined by elem1 followed by elem2 is well typed,
   "false" otherwise, a message is hooked to refTree.
*/
boolean
ExprManager::TypeCheckinglesign (ptexpcal l, ptexpcal elem1, ptexpcal elem2, BackAnnotationType refTree)
{
  TypeList TL;
  TypeType TT1, TT;
  ptexpcal e;

  TT1 = ptexpcal_get_type (elem1);
  if (TT1 == ErrorType)
    return pk_false;
  TL = TypeListCreate ();
  TypeListAppendType (TL, TT1);
  while (ptexpcal_get_Nary (&elem2, &e))
    {
      TT1 = ptexpcal_get_type (e);
      if (ErrorType == TT1)
	return pk_false;
      TypeListAppendType (TL, TT1);
    }
  TL = TypeListRegister (TL);
  TT = TupleTypeCreate (TL);
  ptexpcal_set_type (l, TT);
  return pk_true;
}

/**
   @return "true" if the type of elements (ele1, ele2) of a list of vectorial elements (lelevec) have compatible type, "false" otherwise.
   On error, a message is hooked to refTree.
*/
boolean
ExprManager::TypeCheckinglelevec (ptexpcal lelevec, ptexpcal ele1, ptexpcal ele2, BackAnnotationType refTree)
{
  TypeType TT, Tr;
  ptexpcal t1;

  if (ele2 == NILEXPCAL)
    return pk_true;
  TT = ptexpcal_get_type (ele1);
  t1 = ptexpcal_get_binaryArgstAt (ele2, 1);
  return getTypeUpbSetError (lelevec, ele1, t1, refTree, &Tr);
}

/**
   @return "true" if the type of elements (ele1, ele2) of a list of strings (lstr) have compatible type, "false" otherwise.
   On error, a message is hooked to refTree.
*/
boolean
ExprManager::TypeCheckingString (ptexpcal lstr, ptexpcal ele1, ptexpcal ele2, BackAnnotationType refTree)
{
  TypeType TT;

  TT = get_StringType ();
  ptexpcal_set_type (lstr, TT);
  return pk_true;
}


/**
   @return "true" if the type of elements (ele1, ele2) of a list of travelling expressions (l) have compatible type, "false" otherwise.
   On error, a message is hooked to refTree.
*/
boolean
ExprManager::TypeCheckinglparc (ptexpcal l, ptexpcal ele1, ptexpcal ele2, BackAnnotationType refTree)
{
  return pk_true;
  // ---------- TODO
  if (ele2 != NILEXPCAL)
    /* ExpcalListConcat est à ecrire dans le module de representation des types (ExpcalManager_E.c */
    ptexpcal_set_dimension (l, ExpcalListConcat (ptexpcal_get_dimension (ele1), ptexpcal_get_dimension (ele2)));
  else
    ptexpcal_set_dimension (l, ptexpcal_get_dimension (ele1));
}


/** 
    Type Checking for an expression vexp defined by vexpcal1 \\ vexpcal2
    On error, a message is hooked to refTree.
    @return "true" if the expression is well typed, "false" otherwise.
*/
boolean
ExprManager::TypeVerifException (ptexpcal vexp, ptexpcal vexpcal1, ptexpcal vexpcal2, BackAnnotationType refTree)
{
  TypeType TT;

  if (ptexpcal_get_operator (vexpcal1) != opelem)
    return setErrorTypeExp (vexp, errUnexpectedExpr, refTree);
  if (!getTypeUpbSetError (vexp, vexpcal1, vexpcal2, refTree, &TT))
    return pk_false;
  if (!ExpcalTestNLExtensions (vexpcal1))
    return errExtension (vexp, refTree);
  ptexpcal_set_type (vexp, TT);
  return pk_true;
}

/** 
    Type Checking for an expression vexp defined by vexpcal1 @ vexpcal2
    On error, a message is hooked to refTree.
    @return "true" if the expression is well typed, "false" otherwise.
*/
boolean
ExprManager::TypeVerifComplex (ptexpcal vexp, ptexpcal vexpcal1, ptexpcal vexpcal2, BackAnnotationType refTree)
{
  TypeType TT, t1, Tr;

  if (!getTypeUpbSetError (vexp, vexpcal1, vexpcal2, refTree, &TT))
    return pk_false;
  if (!TypeTestRealNumber (TT))
    return errUnexpectedType (vexp, refTree);
  if (TypeTestArrayClass (TT))
    {
      t1 = ArrayMatrixTypeGetTerm (TT);
      if (t1 == RealType)
	Tr = ComplexType;
      else
	Tr = DComplexType;
      Tr = mkArrayType (ArrayMatrixTypeGetDimLst (TT), Tr);
    }
  else if (TT == RealType)
    Tr = ComplexType;
  else
    Tr = DComplexType;
  ptexpcal_set_type (vexp, Tr);
  return pk_true;
}

/** 
    Type Checking for an expression vexp defined by tr(vexpcal1) (matrix transposition).
    If the type of vexpcal1 is a  Matrix [n,m], the type of vexp is Matrix [m,n].
    On error, a message is hooked to refTree.
    @return "true" if the expression is well typed, "false" otherwise.
*/
boolean
ExprManager::TypeCheckingMatrixTr (ptexpcal vexp, ptexpcal vexpcal1, BackAnnotationType refTree)
{
  TypeType Tr;
  ptexpcal r;

  // type(vexpcal1) Matrix [n,m], result Matrix [m,n]
  // il faudrait le garder en tant que non resolu
  if (!ptexpcal_IsFreeType (vexpcal1))
    {
      getDimAndType (vexpcal1, &r, &Tr);
      if (verifyScalarType (Tr) && (ptexpcal_get_nbelem (r) == 2))
	return pk_true;
    }
  return errUnexpectedType (vexp, refTree);
}


/**
   @return 
   - "true" if the propagation of the type TT on vexp is compatible with its type.
   - "false" otherwise; on error, a message is hooked to refTree.

   The expression is travelled taking account it arity, TypePropagateEnumValue() is called when a reference to a enumerated value is found.
*/
boolean
ExprManager::updateEnumValue (ptexpcal vexp, TypeType TT, BackAnnotationType vt)
{
  ptexpcal nexp;
  if (vexp == NILEXPCAL)
    return pk_true;
  if (ptexpcal_get_indir (vexp, &nexp))
    vexp = nexp;		// A virer mais urgence

  switch ((int) (ptexpcal_get_operatorClass (vexp)))
    {
    case variabl:
    case constant:
      return pk_true;
      break;
    case oper1:
    case oper1tp:
      return updateEnumValue (ptexpcal_get_unaryArg (vexp), TT, vt);
      break;
    case oper2:
    case oper2tp:
      if (ptexpcal_get_operator (vexp) == opvenum)
	return TypePropagateEnumValue (vexp, TT, vt);
      return (updateEnumValue (ptexpcal_get_binaryArgstAt (vexp, 1), TT, vt) && updateEnumValue (ptexpcal_get_binaryArgstAt (vexp, 2), TT, vt));
      break;
    case oper3:
    case oper3tp:
      return (updateEnumValue (ptexpcal_get_ternaryArgstAt (vexp, 1), TT, vt) &&
	      updateEnumValue (ptexpcal_get_ternaryArgstAt (vexp, 2), TT, vt) && updateEnumValue (ptexpcal_get_ternaryArgstAt (vexp, 3), TT, vt));
      break;
    case opereh:
      return updateEnumValue (ptexpcal_get_expWhen (vexp), TT, vt);
      break;
    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
  return pk_false;
}


/**
  @return "true" if an expression (lb) is an int expression, "false" otherwise
  For an int expression, it returns 
  the values of the bounds (bi: the lower bound and bs the upper bound), when they are available (veriftailles is set to pk_true).
  Moreover, baffect is set to "true" if veriftailles and bi=bs, false otherwise.
*/
boolean
ExprManager::getInfoValues (ptexpcal lb, boolean * baffect, compilerInteger * bi, compilerInteger * bs, boolean * veriftailles)
{
  boolean b;

  if (lb == NILEXPCAL)
    {
      *veriftailles = pk_false;
      *bi = 0;
      *bs = 0;
      return pk_true;
    }
  b = ptexpcal_hasIntegerType (lb);
  if (ptexpcal_IsAssigned (lb))
    {
      *veriftailles = pk_true;
      *baffect = pk_true;
      *bi = ptexpcal_get_integerValue (lb);
      *bs = *bi;
    }
  else if (ptexpcal_IsBounded (lb))
    {
      *veriftailles = pk_true;
      *baffect = pk_false;
      ptexpcal_get_integerBounds (lb, bi, bs);
    }
  else
    *veriftailles = pk_false;
  return b;
}



/**
   @return 
   - "true" if the expression of a travel, index \b from lb  \b to bs  \b step step, is well typed. It also computes when the
   informations are available the values of the bounds of index.
   - "false" otherwise. On error, a message is hooked to refTree.
*/
boolean
ExprManager::verifyIndexExpr (ptexpcal index, ptexpcal lb, ptexpcal ub, ptexpcal step, BackAnnotationType refTree)
{
  boolean veriftailles, veriftailles1, veriftailles2, bontype, bontype1, bontype2;
  boolean controlable, decroissant, croissant, binfaff, bsupaff;
  compilerInteger bi1, bs1, bi2, bs2, bi3, bs3, n;

  bontype1 = getInfoValues (lb, &binfaff, &bi1, &bs1, &veriftailles1);
  bontype2 = getInfoValues (ub, &bsupaff, &bi2, &bs2, &veriftailles2);
  bontype = bontype1 && bontype2;
  veriftailles = veriftailles1 && veriftailles2;

  if (step != NILEXPCAL)
    {
      veriftailles = veriftailles && (ptexpcal_IsAssigned (step) || ptexpcal_IsBounded (step));
      bontype = bontype && ptexpcal_hasIntegerType (step);
    }

  if (veriftailles && bontype)
    {
      if (step == NILEXPCAL)
	{
	  croissant = pk_true;
	  controlable = pk_true;
	}
      else if (ptexpcal_IsAssigned (step))
	{
	  n = ptexpcal_get_integerValue (step);
	  croissant = n > 0;
	  decroissant = n < 0;
	  controlable = croissant || decroissant;
	  if (!controlable)
	    {
	      bontype = pk_false;
	      pK_messenger->errorTree (errctparc2, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
	    }
	}
      else if (ptexpcal_IsBounded (step))
	{
	  ptexpcal_get_integerBounds (step, &bi3, &bs3);
	  croissant = (bi3 >= 0) && (bs3 >= 0);
	  decroissant = (bi3 < 0) && (bs3 < 0);
	  controlable = croissant || decroissant;
	  if (!controlable)
	    {
	      bontype = pk_false;
	      pK_messenger->errorTree (errctparc2, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
	    }
	}
      else
	controlable = pk_false;
      if (controlable)
	{
	  if (binfaff || bsupaff)
	    /*
	       if ((croissant && (bi2 < bs1)) || ( decroissant && (bs2 > bi1)))
	       { bontype = pk_false; pK_messenger->errorTree(errctparc1, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));} else;
	     */
	    if (croissant && (bi2 < bs1))
	      bi2 = bs1;
	    else
	      {
		if (decroissant && (bs2 > bi1))
		  bs2 = bi1;
	      }
	  else
	    pK_messenger->warningacc (warctparc1, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
	}
      ptexpcal_set_integerBoundsHolder (index, minimum (bi1, bi2), maximum (bs1, bs2));
    }
  else if (bontype)
    {
      pK_messenger->warningacc (warctparc1, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
      // ptexpcal_set_integerBoundsHolder(index, minimum(bi1, bi2), maximum(bs1, bs2)); veriftailles is false.
      ptexpcal_reset_value (index);
    }
  else
    pK_messenger->errorTree (errctIntegerExpected, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
  return bontype;
}



/**
   @return 
   - "true" if a name (nameField) is the name of a field of a structure/bundle type(t), then tr is the type of this field,
   - "false" otherwise, tr is not modified.   
*/
boolean
ExprManager::getTypeField (TypeType t, ptrepstr nameField, TypeType * tr)
{
  NameList NL;
  TypeList fieldsType;
  TypeSubList NSubLstType;
  boolean Goon;
  NameSubList NSubLst;
  ptrepstr CurrentName;

  if (TypeTestBundleClass (t))
    t = BundleTypeGetTerm (t);
  if (!TypeTestStructureClass (t))
    return pk_false;
  NL = StructureTypeGetNameLst (t);	// field Name
  NSubLst = NameListCast2SubList (NL);
  fieldsType = StructureTypeGetTypeLst (t);
  NSubLstType = TypeListCast2SubList (fieldsType);

  Goon = pk_true;
  while (Goon)
    {
      CurrentName = NameSubListGetHead (NSubLst);
      if (streq (CurrentName, nameField))
	{
	  *tr = TypeSubListGetHead (NSubLstType);
	  return pk_true;
	}			// Found

      if (NameSubListTestLast (NSubLst))
	Goon = pk_false;
      else
	{
	  NSubLst = NameSubListGetTail (NSubLst);
	  NSubLstType = TypeSubListGetTail (NSubLstType);
	}
    }
  return pk_false;		// Not found
}

/**
   @return vexp.
   Type Checking for an expression vexp defined by vexp1.vfield
   On error (getTypeField()), a message is hooked to refTree.
*/
ptexpcal
ExprManager::TypeVerifFieldsAccess (ptexpcal vexp, ptexpcal vexp1, ptexpcal vfield, BackAnnotationType refTree)
{
  ptrepstr nameField;
  TypeType TT1, Tr;

  nameField = ptexpcal_get_constant (vfield);
  TT1 = ptexpcal_getUnamedType (vexp1);
  if (!getTypeField (TT1, nameField, &Tr))
    {
      printf ("DEBUG      in  ExprManager::TypeVerifFieldsAccess, ERROR\n");
      fflush (stdout);
      pK_messenger->errorTreeStr (errenumvalue, BACK_ANNOT_GET_ABSTRACT_TREE(refTree), nameField);
      setErroneousType (vexp);
    }
  else
    ptexpcal_set_type (vexp, Tr);
  return vexp;
}

/**
   It verifies that an expression of index (aIndex) is compatible with an expression of dimension (aDim).
   The test is achieved for aIndex denoting an access to a element y[i] (referenced in x [ y[i], ...], with x=vArray)
   Currently vArray is unused.

   On error, a message is hooked to refTree if the controle is not suspended (modeTypeCheckingOutOfBounds attribute of the ExprManager class).
   @return 0 when it is not controlable
           1 when the index is in the bounds
           2 when the index is out of the bounds.
*/
int
ExprManager::verifyAccess (ptexpcal aDim, ptexpcal aIndex, ptexpcal vArray, BackAnnotationType refTree)
{
  compilerInteger n, m, bres;
  boolean bn, bm;
  TypeType TT, typeDeBase;
  noperators op;

  bres = 0;
  op = ptexpcal_get_operator (aIndex);
  if (op != opelem)
    {				// x[ y[i], ...] is accepted.
      TT = ptexpcal_getUnamedType (aIndex);
      if (isIntegerType (TT))
	{			// index :  array of integers.
	  isGreaterOrEqual_or_Greater_than_0 (aIndex, refTree, &m, &bm, modeTypeCheckingOutOfBounds, pk_true);
	  isGreaterOrEqual_or_Greater_than_0 (aDim, refTree, &n, &bn, pk_true, pk_true);
	  if (bn && bm)
	    {
	      if (n <= m)
		{
		  bres = 2;
		  // depending of the context:
		  if (modeTypeCheckingOutOfBounds)
		    pK_messenger->errorTree (errctmotif2, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
		}
	      else if (m < 0)
		bres = 2;
	      else
		bres = 1;
	    }
	}
      else
	// index :  array of integers ::= ident | bi..bs step 
      if (TypeTestArrayClass (TT))
	{
	  typeDeBase = ArrayMatrixTypeGetTerm (TT);
	  if (!isIntegerType (typeDeBase))
	    setErroneousType (aIndex);
	  else
	    {			//   getBounds()
	      // if (op == opident) ; else if (op == openum);
	    }
	}
      else
	setErroneousType (aIndex);
    }
  return bres;
}


/**
   Type Checking for an expression (element) defined by vArray[lindices].
   On error, a message is hooked to refTree.    
*/
void
ExprManager::TypeCheckingelem (ptexpcal element, ptexpcal vArray, ptexpcal lindices, BackAnnotationType refTree)
{
  TypeType TT, TT1, typeDeBase;
  ptexpcal dims, aIndex, aDim;
  int n, m, code;

  TT = ptexpcal_getUnamedType (vArray);
  if (TypeTestBundleClass (TT))
    pK_messenger->errorTree (errBundleType, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
  lastTypeCheckingelem = -1;
  if (lindices != NILEXPCAL)
    {
      if (!TypeTestArrayClass (TT))
	erreurattachref1nom (errctmotif2, refTree, vArray);
      else
	{

	  dims = ArrayMatrixTypeGetDimLst (TT);
	  m = ptexpcal_get_nbelem (dims);
	  n = ptexpcal_get_nbelem (lindices);
	  if (n > m)
	    {
	      erreurattachref1nom (errctmotif2, refTree, vArray);
	      n = m;
	    }
	  while (n > 0)
	    {
	      aDim = ptexpcal_getNaryArgs (&dims);
	      aIndex = ptexpcal_getNaryArgs (&lindices);
	      code = verifyAccess (aDim, aIndex, vArray, refTree);
	      if (lastTypeCheckingelem < 0)
		lastTypeCheckingelem = code;
	      else if (code == 2)
		lastTypeCheckingelem = code;
	      else if (code == 0 && (lastTypeCheckingelem != 2))
		lastTypeCheckingelem = code;
	      n--;
	    }

	  typeDeBase = ArrayMatrixTypeGetTerm (TT);
	  if (dims != NILEXPCAL)
	    TT1 = mkArrayType (dims, typeDeBase);
	  else
	    TT1 = typeDeBase;
	  ptexpcal_set_type (element, TT1);
	}
    }
  if (lastTypeCheckingelem == -1)
    lastTypeCheckingelem = 0;	// not controllable by default.
}

/**
   Type Checking for an expression (vexp) defined by a counter: vexpcal1 count vexpcal2
   On error, a message is hooked to refTree.    
*/
ptexpcal
ExprManager::TypeVerifCounter (ptexpcal vexp, ptexpcal vexpcal1, ptexpcal vexpcal2, BackAnnotationType refTree)
{
  boolean b;
  int v;

  if (!TypeCheckingIsEvent (vexpcal1, &b))
    pK_messenger->errorTree (errctUnexpectedType, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));

  if (!ptexpcal_isNil (vexpcal2))
    {
      if (!ptexpcal_hasIntegerType (vexpcal2))
	pK_messenger->errorTree (errctIntegerExpected, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
      else if (ptexpcal_IsAssigned (vexpcal2))
	{
	  v = (int) ptexpcal_get_integerValue (vexpcal2);
	  if (v == 0)
	    pK_messenger->errorTree (errUnexpectedValue, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
	  else;
	}
    }

  ptexpcal_set_type (vexp, IntegerType);
  return vexp;
}


/**
   Type ckecking for an expression (r) defined by a concatenation operator e1 |+ e2.
   with 
   - type(e1) = [0..m1-1] -> t1, 
   - type(e2) = [0..m2-1] -> t2, 
   - then type(r) = [0..m1+m2-1] -> upper(t1, t2).

   On error, a message is hooked to refTree.  
 */
void
ExprManager::TypeCheckingconc (ptexpcal r, ptexpcal e1, ptexpcal e2, BackAnnotationType refTree)
{
  ptexpcal dim1, fdim1;
  ptexpcal dim2, fdim2;
  ptexpcal fdim, dim;
  TypeType t1, t2, TT;
  boolean bok;

  if (!ptexpcal_IsFreeType (e1) && !ptexpcal_IsFreeType (e2))
    {
      bok = pk_false;
      getDimAndType (e1, &dim1, &t1);
      getDimAndType (e2, &dim2, &t2);
      if (GetGreaterType (t1, t2, &TT))
	{
	  if (dim1 == NILEXPCAL)
	    fdim1 = mkIntegerConstant (1);
	  else
	    fdim1 = ptexpcal_getNaryArgs (&dim1);
	  if (dim2 == NILEXPCAL)
	    fdim2 = mkIntegerConstant (1);
	  else
	    fdim2 = ptexpcal_getNaryArgs (&dim2);
	  if (isEquivalent (dim1, dim2, pk_true))
	    {
	      fdim = mkBinaryExp (opplus, fdim1, fdim2, NILBACKANNOTATION);
	      dim = mkNaryExp (oplesign, fdim, dim1, NILBACKANNOTATION);
	      TT = mkArrayType (dim, TT);
	      bok = pk_true;
	    }
	}
      if (!bok)
	{
	  errUnexpectedType (r, refTree);
	  TT = get_ErrorType ();
	}
      setType (r, TT);
    }
}


/**
   Type ckecking for an expression (r) defined by a sequence operator (e1 next e2)
   with 
   - type(e1) = [0..c1]...[0..cp] -> t1, 
   - type(e2) = [0..b1]...[0..bp] -> t2, 
   - then type(r) = [0..c1]...[0..cp]-> upper(t1, t2), and ci>=bi for each i.

   On error, a message is hooked to refTree.  
 */
void
ExprManager::TypeCheckingNext (ptexpcal r, ptexpcal e1, ptexpcal e2, BackAnnotationType refTree)
{
  ptexpcal dim1, dim2, rdim, v1, v2;
  TypeType t1, t2, TT;
  boolean berr, bverified;

  if (!ptexpcal_IsFreeType (e1) && !ptexpcal_IsFreeType (e2))
    {
      berr = pk_false;
      getDimAndType (e1, &dim1, &t1);
      rdim = dim1;
      getDimAndType (e2, &dim2, &t2);
      if (dim1 == NILEXPCAL)
	berr = pk_true;
      else if (getSize (dim1) != getSize (dim2))
	berr = pk_true;
      else if (!GetGreaterType (t1, t2, &TT))
	berr = pk_true;
      else
	{
	  bverified = pk_true;
	  while (getNaryArgs (&dim1, &v1) && !berr)
	    {
	      getNaryArgs (&dim2, &v2);
	      if (ptexpcal_IsAssigned (v1) && ptexpcal_IsAssigned (v2))
		berr = ptexpcal_get_integerValue (v1) < ptexpcal_get_integerValue (v2);
	      else
		bverified = pk_false;
	    }
	}
      if (berr)
	{
	  errUnexpectedType (r, refTree);
	  TT = get_ErrorType ();
	}
      else if (!bverified)
	{
	  TT = ptexpcal_get_type (e1);
	  pK_messenger->warningacc (warNoct, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
	}
      else if (TT == t1)
	TT = ptexpcal_get_type (e1);
      else
	TT = mkArrayType (rdim, TT);
      setType (r, TT);
    }
}

/**  
     Type checking for vexp a matrix product: first *. second. 
     On error, messages are hooked to vt tree.
     @return vexp expression.
*/
ptexpcal
ExprManager::TypeCheckingMatrixMult (ptexpcal vexp, ptexpcal first, ptexpcal second, BackAnnotationType vt)
{
  ptexpcal r1, r2, d11, d12, d21, d22, nb;
  TypeType Tr1, Tr2, Tr;
  int n1, n2;
  boolean b;

  if (IsFreeType (first) || IsFreeType (second))
    {
      errUnexpectedType (vexp, vt);
      return vexp;
    }
  getDimAndType (first, &r1, &Tr1);
  getDimAndType (second, &r2, &Tr2);
  if (!(TypeTestNumerical (Tr1) && TypeTestNumerical (Tr2) && GetGreaterType (Tr1, Tr2, &Tr)))
    {
      errUnexpectedType (vexp, vt);
      return vexp;
    }

  b = pk_false;
  n1 = getSize (r1);
  n2 = getSize (r2);
  if (n1 == 2)
    {
      d11 = getNaryArgs (&r1);
      d12 = getNaryArgs (&r1);
      if (n2 == 2)
	{
	  d21 = getNaryArgs (&r2);
	  d22 = getNaryArgs (&r2);
	  if (isEquivalent (d12, d21, pk_true))
	    {
	      nb = mkNaryExp (oplesign, d11, mkNaryExp (oplesign, d22, NILEXPCAL, NILBACKANNOTATION), NILBACKANNOTATION);
	      b = pk_true;
	    }
	}
      else if (n2 == 1)
	if (isEquivalent (d12, getNaryArgs (&r2), pk_true))
	  {
	    nb = mkNaryExp (oplesign, d11, NILEXPCAL, NILBACKANNOTATION);
	    b = pk_true;
	  }
    }
  else if (n1 == 1)
    {
      d11 = getNaryArgs (&r1);
      if (n2 == 2)
	{
	  d21 = getNaryArgs (&r2);
	  d22 = getNaryArgs (&r2);
	  if (isEquivalent (d11, d21, pk_true))
	    {
	      nb = mkNaryExp (oplesign, d22, NILEXPCAL, NILBACKANNOTATION);
	      b = pk_true;
	    }
	}
      else if (n2 == 1)
	if (isEquivalent (d11, getNaryArgs (&r2), pk_true))
	  {
	    nb = NILEXPCAL;
	    b = pk_true;
	  }
    }
  if (b)
    if (nb != NILEXPCAL)
      setType (vexp, mkArrayType (nb, Tr));
    else
      setType (vexp, Tr);
  else
    errUnexpectedType (vexp, vt);
  return vexp;
}

/** 
    Type checking for vexp a repetition definition "first |* second".
    On error, messages are hooked to vt tree.
    @return vexp expression.
*/
ptexpcal
ExprManager::TypeCheckingRepeat (ptexpcal vexp, ptexpcal first, ptexpcal second, BackAnnotationType vt)
{
  ptexpcal r1, nb;
  TypeType Tr1, T;
  compilerInteger n;
  boolean b;

  if (isGreaterOrEqual_or_Greater_than_0 (second, vt, &n, &b, pk_true, pk_true))
    {
      getDimAndType (first, &r1, &Tr1);
      nb = mkNaryExp (oplesign, second, r1, NILBACKANNOTATION);
      T = mkArrayType (nb, Tr1);
    }
  else
    T = get_ErrorType ();
  setType (vexp, T);
  return vexp;
}

/** 
    Type checking for lelevec a list of vectorial expressions. Such an element is a travel (# {i to n, ..;}:E #) or 
    an access to an element of an array (#[x,...]:E#).
    E must be a type compatible with the type of target.
    On error, messages are hooked to vt tree.
*/
void
ExprManager::verifyEachElevec (ptexpcal target, ptexpcal lelevec, BackAnnotationType refTree)
{
  ptexpcal velem;

  while (getNaryArgs (&lelevec, &velem))
    switch ((int) (getOperator (velem)))
      {
      case opelevec:
	verifyElevec (target, velem, refTree);
	break;
      case opensvec:
	verifyEnsvec (target, velem, refTree);
	break;
      }
}


/** 
    Type checking for an access (AnElevec) to an element of an array (#[x,...]:E#).
    E must be a type compatible with the type of target.
    On error, messages are hooked to refTree tree.
*/
void
ExprManager::verifyElevec (ptexpcal target, ptexpcal AnElevec, BackAnnotationType refTree)
{
  ptexpcal lexp, vexp;
  TypeType TT;

  getBinaryArgs (AnElevec, &lexp, &vexp);
  TypeCheckingAccept (target, vexp, &TT, refTree, pk_false, pk_true);
}

/** 
    Type checking for  an access (AnElevec) to an element of an array (# { i to ..., ...}:E #).
    E must be a type compatible with the type of target.
    On error, messages are hooked to refTree tree.
*/
void
ExprManager::verifyEnsvec (ptexpcal target, ptexpcal AnElevec, BackAnnotationType refTree)
{

}


/** 
    Type checking for vleft := vrest, vrest a restructuration expression. (extension not yet allowed).
    On error, messages are hooked to refTree tree.
*/
boolean
ExprManager::verifAssignRestructuration (ptexpcal vleft, ptexpcal vrest, boolean * bextension, BackAnnotationType refTree)
{
  ptexpcal lexp, vexp, dims1, dims;
  TypeType t1, tb1, t2, tb2, vtype;
  int nbIndex, s1, s2;

  (*bextension) = pk_false;
  if (ptexpcal_IsFreeType (vleft))
    {
      pK_messenger->warningacc (warNoct, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
      return pk_true;
    }

  t1 = ptexpcal_getUnamedType (vleft);
  if (!TypeTestArrayClass (t1))
    return errUnexpectedType (vleft, refTree);

  tb1 = getDimAndBaseType (t1, &dims1);
  s1 = ptexpcal_get_nbelem (dims1);

  getBinaryArgs (vrest, &lexp, &vexp);
  if (ptexpcal_IsFreeType (vexp))
    pK_messenger->warningacc (warNoct, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
  else
    {
      nbIndex = getSize (lexp);
      t2 = ptexpcal_getUnamedType (vexp);
      tb2 = getDimAndBaseType (t2, &dims);
      s2 = ptexpcal_get_nbelem (dims);
      // if (( s2 + nbIndex) != s1) return errUnexpectedType(vleft, refTree);
      if (!unifytctype (tb2, tb1, &vtype))
	{
	  errUnexpectedType (vleft, refTree);
	  return pk_false;
	}
      verifMaxIndexValuesAndDims (dims1, lexp, refTree);
    }
  return pk_true;
}

/**
   On error, messages are hooked to refTree tree.
   @note {nmax must be initialized before the first call to this method.}
   @return
   - "true" if the max value of a list of expressions (e) is defined, in this case nmax is set.
   - "false" otherwise, nmax is not modified
*/
boolean
ExprManager::getMaxValueOf (ptexpcal e, compilerInteger * nmax, BackAnnotationType refTree)
{
  ptexpcal velem;
  compilerInteger cmax;

  if (satisfiesList (e))
    {
      while (getNaryArgs (&e, &velem))
	{
	  if (!getMaxValueOf (velem, &cmax, refTree))
	    return pk_false;
	  if (cmax > *nmax)
	    (*nmax) = cmax;
	}
      return pk_true;
    }
  // it is an identifier. x:= [5,6,7] [1,2]: E is translated by x:= I1, I2: E with I1:=[5,6,7] I2:=[1,2]
  if (!ptexpcal_IsBounded (e))
    {
      pK_messenger->warningacc (warNoct, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
      return pk_false;
    }
  ptexpcal_get_integerBounds (e, &cmax, nmax);
  return pk_true;
}

/**
   For each element e_i of lexp, it verifies that the max value of e_i must be <= size of the i-th value of dims.
   On error, messages are hooked to refTree tree.
   @return "false" when an error is detected, "true" otherwise.
*/
boolean
ExprManager::verifMaxIndexValuesAndDims (ptexpcal dims, ptexpcal lexp, BackAnnotationType refTree)
{
  ptexpcal e_i, dim_i;
  compilerInteger sizeCurrentDim, nmax;
  boolean bok;

  bok = pk_true;
  while (getNaryArgs (&lexp, &e_i))
    {
      getNaryArgs (&dims, &dim_i);
      if (ptexpcal_IsAssigned (dim_i))
	{
	  sizeCurrentDim = ptexpcal_get_integerValue (dim_i);
	  nmax = 0;
	  if (getMaxValueOf (e_i, &nmax, refTree))
	    if (nmax >= sizeCurrentDim)
	      {
		if (bok)
		  pK_messenger->errorTree (errOutOfBound, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
		bok = pk_false;
	      }
	}
      else
	pK_messenger->warningacc (warNoct, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
    }
  return bok;
}


/*
  void
  ExprManager::setSoftModeArrayTypeChecking(ptexpcal vexp)
  {
  noperators op;
  ptexpcal velem;
  
  op = ptexpcal_get_operator(vexp);
  if (op == oplelevec)
  {
  getNaryArgs(&vexp, &velem);
  softModeArrayTypeChecking = (ptexpcal_get_operator(velem) != opelevec);
  }
  }
*/

/** 
    Type checking for vleft := vexp, vexp a sequential expression. (extension not yet allowed).
    On error, messages are hooked to refTree tree.
    \note Applied for vexp defined by [ [...]:E1, ...., [...]: En]
*/
boolean
ExprManager::verifSequentialDef (ptexpcal vleft, ptexpcal vexp, BackAnnotationType refTree)
{
  ptexpcal dtarget, velem, lexp, vdef;
  TypeType target;
  noperators op;
  boolean b1, bres;

  bres = pk_true;
  op = ptexpcal_get_operator (vexp);
  if (op == oplelevec)
    {
      target = ptexpcal_getUnamedType (vleft);
      dtarget = ArrayMatrixTypeGetDimLst (target);
      b1 = pk_false;
      while (getNaryArgs (&vexp, &velem))
	{
	  if (ptexpcal_get_operator (velem) != opelevec)
	    b1 = pk_true;
	  else
	    {
	      getBinaryArgs (velem, &lexp, &vdef);
	      // refTree is attributed when an error is detected in verifMaxIndexValuesAndDims
	      if (!verifMaxIndexValuesAndDims (dtarget, lexp, refTree))
		bres = pk_false;
	    }
	}
      if (b1)
	pK_messenger->warningacc (warNoct, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
    }
  return bres;
}


/**
   @return  the type of the index (index) built from its operands 
   - lower bound (lb),
   - upper bound (ub),
   - and the step (st)
*/
TypeType
ExprManager::setTypeIndex (ptexpcal index, ptexpcal lb, ptexpcal ub, ptexpcal step, BackAnnotationType refTree)
{
  compilerInteger bi, bs, m, n;
  ptexpcal vsize, ldims;
  TypeType TT;
  boolean sb, croissant, controlable;

  if (!ptexpcal_IsFreeType (index))
    TT = ptexpcal_get_type (index);
  else
    {
      controlable = pk_true;
      if (step == NILEXPCAL)
	{
	  croissant = pk_true;
	  n = 1;
	}
      else if (ptexpcal_IsAssigned (step))
	{
	  n = (int) ptexpcal_get_integerValue (step);
	  croissant = n > 0;
	}
      else			// bounded ?
	controlable = pk_false;

      if (!controlable)
	TT = ptexpcal_get_type (index);
      else
	{
	  if (ptexpcal_IsBounded (index))
	    {
	      ptexpcal_get_integerBounds (index, &bi, &bs);
	      if (bs > bi)
		m = (bs - bi) + 1;
	      else
		m = (bi - bs) + 1;
	      vsize = mkIntegerConstant (m);
	    }
	  else
	    {
	      sb = setArithmeticRuleMode (pk_true);
	      if (croissant)
		vsize = mkBinaryExp (opmoin, ub, lb, refTree);
	      else
		vsize = mkBinaryExp (opmoin, lb, ub, refTree);
	      vsize = mkBinaryExp (opplus, vsize, mkIntegerConstant (1), refTree);
	      sb = setArithmeticRuleMode (sb);
	    }
	  if (!croissant)
	    n = -n;
	  if (n != 1)
	    {
	      if (ptexpcal_IsAssigned (vsize))
		{
		  m = (int) ptexpcal_get_integerValue (vsize);
		  if ((m % n) == 0)
		    vsize = mkIntegerConstant (m / n);
		  else
		    vsize = mkIntegerConstant ((m / n) + 1);
		}
	      else
		return ptexpcal_get_type (index);	// TO BE COMPLETED. use a macro to defined the size.
	    }
	  ldims = mkNaryExp (oplesign, vsize, NILEXPCAL, refTree);
	  TT = mkArrayType (ldims, IntegerType);
	  ptexpcal_set_type (index, TT);
	}
    }
  return TT;
}


/**
   @return true is one of the following conversions is applied, false otherwise.
   - vexpcal2 is a int constant, vexpcal1 a boolean: 0,1 -> pk_true, pk_false;
   - vexpcal2 is a int constant, vexpcal1 an event: 1 -> event;
   - vexpcal2 is a boolean constant, vexpcal1 a boolean: pk_true, pk_false -> 1,0.

   On error, messages are hooked to refTree tree.
*/
boolean
ExprManager::booleanIntegerConversion (ptexpcal vexpcal1, ptexpcal * vexpcal2, BackAnnotationType refTree)
{
  TypeType source, target;
  long int n;
  boolean b;

  if (ptexpcal_IsFreeType (vexpcal1))
    return pk_false;
  if (ptexpcal_IsFreeType (*vexpcal2))
    return pk_false;
  if (!ptexpcal_IsAssigned (*vexpcal2))
    return pk_false;
  n = ptexpcal_get_integerValue (*vexpcal2);
  if ((n != 0) && (n != 1))
    return pk_false;
  b = pk_false;
  target = ptexpcal_get_type (vexpcal1);
  source = ptexpcal_get_type (*vexpcal2);
  if (target == BooleanType)	// 0, 1 --> boolean
    if ((source == LongType) || (source == ShortType) || (source == IntegerType))
      {
	if (n == 0)
	  (*vexpcal2) = ptexpcal_getFalse ();
	else
	  (*vexpcal2) = ptexpcal_getTrue ();
	b = pk_true;
      }
    else
      if (target == PureType)	//  1 --> event
	{ 
	  if ((source == LongType) || (source == ShortType) || (source == IntegerType))
	    { 
	      if (n == 0)
	      pK_messenger->errorTree (errUnexpectedValue, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
	    else
	      {
		(*vexpcal2) = ptexpcal_getTrue ();
		b = pk_true;
	      }
	    }
	  else if ((target == LongType) || (target == ShortType) || (target == IntegerType))
	    // pk_true, pk_false --> int types
	    if ((source == BooleanType) || (source == PureType))
	      {
		(*vexpcal2) = mkIntegerConstant (n);
		b = pk_true;
	      }
	}

  return b;
}

/**
  Verification on value.
  @param vexp (in) an expression.
  @param refTree (in) reference to the syntactic tree of the program ; used for annotations.
  @param val (out) value of the expression when it is known.
  @param bval (out) at pk_true if the value of the expression (val) is known, "false" otherwise.
  @param bmess (in) if pk_true a message if hooked to the tree (refTree) is the value (val) is not positive.
  @param bge (in) if pk_true the value must be greater or equal to 0, and greater than 0 otherwise.
  @return 
   - "true" when the expression is undefined (nil), or is an int expression and its value is unknown or known and >=0 when bge, or > 0 when not bge.
   - "false" otherwise.
*/
boolean
  ExprManager::isGreaterOrEqual_or_Greater_than_0 (ptexpcal vexp, BackAnnotationType refTree, compilerInteger * val, boolean * bval, boolean bmess, boolean bge)
{
  *bval = pk_false;
  if (ptexpcal_isNil (vexp))
    return pk_true;
  if (!ptexpcal_hasIntegerType (vexp))
    {
      pK_messenger->errorTree (errctIntegerExpected, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
      return pk_false;
    }
  if (!ptexpcal_IsAssigned (vexp))
    {
      pK_messenger->warningacc (warctValue, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
      return pk_true;
    }
  *val = ptexpcal_get_integerValue (vexp);
  *bval = pk_true;
  if (bge)
    {
      if (*val >= 0)
	return pk_true;
    }
  else
    {
      if (*val > 0)
	return pk_true;
    }
  if (bmess)
    pK_messenger->errorTree (errctPositiveValue, BACK_ANNOT_GET_ABSTRACT_TREE(refTree));
  return pk_false;
}
