/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   @return 
   - "true" if the expression not(vexpcal1) is evaluated in the expression vres, 
   - "false" otherwise, vres is not modified. 
   The following rules are applied:
   - NOT(NOT(E)) --> E
   - NOT(pk_false) --> pk_true
   - NOT(pk_true) --> false

   vtree: abstract tree information (parser information) used for error managing.   
*/
  boolean ExprManager::evalNot (ptexpcal vexpcal1, BackAnnotationType vtree, ptexpcal * vres)
{
  if (ptexpcal_IsEvent (vexpcal1))
    {
      *vres = ptexpcal_getFalse ();
      return pk_true;
    }
  if (ptexpcal_equalTrue (vexpcal1))
    {
      *vres = ptexpcal_getFalse ();
      return pk_true;
    }
  if (ptexpcal_equalFalse (vexpcal1))
    {
      *vres = ptexpcal_getTrue ();
      return pk_true;
    }
  if (ptexpcal_get_operator (vexpcal1) == opnon)
    {
      *vres = ptexpcal_get_unaryArg (vexpcal1);
      return pk_true;
    }
  TypeCheckingCondition (vexpcal1, vtree);
  return pk_false;
}

/**
   @return 
   - "true" if the expression vexpcal1 and vexpcal2 is evaluated in the expression vres, 
   - "false" otherwise vres is not modified.

   The classical boolean rules on boolean constants are applied.

   vtree: abstract tree information (parser information) used for error managing.
*/
boolean
ExprManager::evalAnd (ptexpcal vexpcal1, ptexpcal vexpcal2, BackAnnotationType vtree, ptexpcal * vres)
{
  if (ptexpcal_IsEvent (vexpcal1))
    {
      if (ptexpcal_IsEvent (vexpcal2))
	{
	  *vres = ptexpcal_getTrue ();
	  return pk_true;
	}
      TypeCheckingCondition (vexpcal2, vtree);
      *vres = vexpcal2;
      return pk_true;
    }

  if (ptexpcal_equalFalse (vexpcal1))
    {
      TypeCheckingCondition (vexpcal2, vtree);
      *vres = ptexpcal_getFalse ();
      return pk_true;
    }

  if (ptexpcal_IsEvent (vexpcal2))
    {
      TypeCheckingCondition (vexpcal1, vtree);
      *vres = vexpcal1;
      return pk_true;
    }

  if (ptexpcal_equalFalse (vexpcal2))
    {
      TypeCheckingCondition (vexpcal1, vtree);
      *vres = ptexpcal_getFalse ();
      return pk_true;
    }
  return pk_false;
}


/**
   @return 
   - "true" if the expression vexpcal1 or vexpcal2 is evaluated in the expression vres, 
   - "false" otherwise, vres is not modified.

   The classical boolean rules on boolean constants are applied.\n
   vtree: abstract tree information (parser information) used for error managing.
*/
boolean
ExprManager::evalOr (ptexpcal vexpcal1, ptexpcal vexpcal2, BackAnnotationType vtree, ptexpcal * vres)
{

  if (ptexpcal_IsEvent (vexpcal1) || ptexpcal_IsEvent (vexpcal2))
    {
      *vres = ptexpcal_getTrue ();
      return pk_true;
    }

  if (ptexpcal_equalFalse (vexpcal1))
    {
      TypeCheckingCondition (vexpcal2, vtree);
      *vres = vexpcal2;
      return pk_true;
    }

  if (ptexpcal_equalFalse (vexpcal2))
    {
      TypeCheckingCondition (vexpcal1, vtree);
      *vres = vexpcal1;
      return pk_true;
    }
  return pk_false;
}



/**
   @return 
   - "true" if the expression vexpcal1 or vexpcal2 is evaluated in the expression vres, 
   - "false" otherwise, vres is not modified.

   The classical boolean rules on boolean constants are applied.\n
   vtree: abstract tree information (parser information) used for error managing.
*/
boolean
ExprManager::evalXor (ptexpcal vexpcal1, ptexpcal vexpcal2, BackAnnotationType vtree, ptexpcal * vres)
{
  ptexpcal vf;

  vf = ptexpcal_getFalse ();
  if (vexpcal1 == vexpcal2)
    {
      *vres = vf;
      return pk_true;
    }

  if (ptexpcal_IsEvent (vexpcal1))
    {
      TypeCheckingCondition (vexpcal2, vtree);
      *vres = mkNot (vexpcal2, vtree);
      return pk_true;
    }

  if (ptexpcal_IsEvent (vexpcal2))
    {
      TypeCheckingCondition (vexpcal1, vtree);
      *vres = mkNot (vexpcal1, vtree);
      return pk_true;
    }

  if (vexpcal1 == vf)
    {
      TypeCheckingCondition (vexpcal2, vtree);
      *vres = vexpcal2;
      return pk_true;
    }

  if (vexpcal2 == vf)
    {
      TypeCheckingCondition (vexpcal1, vtree);
      *vres = vexpcal1;
      return pk_true;
    }
  return pk_false;
}



/**
   It associates to an index (vindex) 
   - a lower bound = 0
   - an upper bound = hbound (0 if undefined)

   vtree: abstract tree information (parser information) used for error managing.
*/
void
ExprManager::setIndexBounds (ptexpcal vindex, ptexpcal hbound, BackAnnotationType vtree)
{
  if (hbound != NILEXPCAL)
    if (ptexpcal_hasIntegerValue (hbound))
      ptexpcal_set_integerBoundsHolder (vindex, 0, ptexpcal_get_integerValue (hbound));
    else
      pK_messenger->warningaccnom (warcontrole, BACK_ANNOT_GET_ABSTRACT_TREE(vtree), ptexpcal_get_treeIdentifier (vindex));
  else
    ptexpcal_set_integerBoundsHolder (vindex, 0, 0);
}


/** @return the value of an assumed constant expression (pt).*/
long long int
ExprManager::getIntegerValueAssignedToConstantExpression (ptexpcal pt)
{
  return repstr2integer (ptexpcal_get_constant (pt));
}


/**
   @return "true" if the intersection [*l,*u] = [l1,u1] and [l2,u2] is not empty, "false" otherwise (empty intersection).
*/
boolean ExprManager::intersection (compilerInteger l1, compilerInteger u1, compilerInteger l2, compilerInteger u2, compilerInteger * l, compilerInteger * u)
{
  if (u1 < l2) return pk_false;
  if (u2 < l1) return pk_false;
  *l = maximum (l1, l2);
  *u = minimum (u1, u2);
  return pk_true;
}



/** 
    @return 
    - "true" if the expression vexpcal1 "vopRela" vexpcal2, where vopRela is a relational operator, is simplified in the expression vresult,
    - "false" otherwise, vresult in not modified.

    The evaluation is restricted to int expressions.
*/
boolean
ExprManager::testValueRelaOperator (noperators vopRela, ptexpcal vexp1, ptexpcal vexp2, ptexpcal * vresult)
{
  compilerInteger l1, l2, u1, u2, u, l;
  boolean b, ctest;

  if (!getBounds (vexp1, &l1, &u1)) return pk_false;
  if (!getBounds (vexp2, &l2, &u2)) return pk_false;

  b = pk_true;
  switch ((int) (vopRela))
    {
    case opdiff:
      if (!intersection (l1, u1, l2, u2, &l, &u))
	ctest = pk_true;
      else if ((l1 == u1) && (l2 == u2))
	ctest = pk_false;
      else
	b = pk_false;
      break;
    case opegal:
      if (!intersection (l1, u1, l2, u2, &l, &u))
	ctest = pk_false;
      else if ((l1 == u1) && (l2 == u2))
	ctest = pk_true;
      else
	b = pk_false;
      break;
    case opiegal:
      if (!intersection (l1, u1, l2, u2, &l, &u))
	ctest = (l1 <= l2);
      else if ((l1 == u1) && (l2 == u2))
	ctest = (l1 <= l2);
      else
	b = pk_false;
      break;
    case opinfe:
      if (!intersection (l1, u1, l2, u2, &l, &u))
	ctest = (l1 < l2);
      else
	b = pk_false;
      break;
    case opsegal:
      if (!intersection (l1, u1, l2, u2, &l, &u))
	ctest = (l1 >= l2);
      else if ((l1 == u1) && (l2 == u2))
	ctest = (l1 >= l2);
      else
	b = pk_false;
      break;
    case opsupe:
      if (!intersection (l1, u1, l2, u2, &l, &u))
	ctest = (l1 > l2);
      else
	b = pk_false;
      break;
    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
  if (!b)
    return pk_false;
  if (ctest)
    *vresult = ptexpcal_getTrue ();
  else
    *vresult = ptexpcal_getFalse ();
  return pk_true;
}


/**
   It sets the bounds of an expression (vresult) by an operation (opRes) on two bounded (bi1, bs1) and (bi2,bs2)
   \see computeBounds().
   vtree: abstract tree information (parser information) used for error managing.
*/
void
ExprManager::setBoundsExpr (ptexpcal vresult, noperators opRes, compilerInteger bi1, compilerInteger bs1, compilerInteger bi2, compilerInteger bs2, BackAnnotationType vtree)
{
  compilerInteger bi, bs;

  if (computeBounds (opRes, bi1, bi2, bs1, bs2, &bi, &bs))
    ptexpcal_set_integerBoundsHolder (vresult, bi, bs);
  else
    pK_messenger->errorTree (errUnexpectedExpr, BACK_ANNOT_GET_ABSTRACT_TREE(vtree));
}


/**
   @return the evaluation of a binary expression (vresult) with vexp1 and vexp2 as operands.
   - The evaluation is restricted to int expressions.
   - operators:  +, *, -, /, modulo, **.

   vtree: abstract tree information (parser information) used for error managing.
*/
ptexpcal
ExprManager::evalarithBinary (noperators opRes, ptexpcal vresult, ptexpcal vexp1, ptexpcal vexp2, BackAnnotationType vtree)
{
  compilerInteger bi1, bs1, bi2, bs2;
  // noperators opRes;

  // opRes = ptexpcal_get_operator(vresult);
  if (ptexpcal_IsAssigned (vexp1))
    {
      bi1 = ptexpcal_get_integerValue (vexp1);
      switch ((int) opRes)
	{ 
	case opdivi:
	  if (vexp1 == ptexpcal_0) return ptexpcal_0;
	  break;
	case opmult:
	  if (vexp1 == ptexpcal_0) return ptexpcal_0;
	  if (vexp1 == ptexpcal_1) return vexp2;
	  break;
	case oppuis:
	  if (vexp1 == ptexpcal_0) return ptexpcal_0;
	  if (vexp1 == ptexpcal_1) return ptexpcal_1;
	  break;
	case opplus:
	  if (vexp1 == ptexpcal_0) return vexp2;
	  break;
	case opmoin:
	  if (bi1 == 0)
	    return mkUnaryExp (opnega, vexp2, vtree);
	  break;
	case opmod:
	  break;
	}
      if (ptexpcal_IsAssigned (vexp2))
	{
	  ptexpcal_set_computedIntegerValue (vresult, computeValue (opRes, bi1, ptexpcal_get_integerValue (vexp2)));
	}
      else if (ptexpcal_IsBounded (vexp2))
	{
	  ptexpcal_get_integerBounds (vexp2, &bi2, &bs2);
	  setBoundsExpr (vresult, opRes, bi1, bi1, bi2, bs2, vtree);
	}
    }
  else
    {
      if (ptexpcal_IsAssigned (vexp2))
	{
	  bi2 = ptexpcal_get_integerValue (vexp2);
	  switch ((int) opRes)
	    {
	    case opdivi:
	      if (bi2 == 0)
		pK_messenger->errorTree (errUnexpectedExpr, BACK_ANNOT_GET_ABSTRACT_TREE(vtree));
	      if (bi2 == 1)
		return vexp1;
	      break;
	    case opmult:
	      if (bi2 == 0)
		return mkIntegerConstant (0);
	      if (bi2 == 1)
		return vexp1;
	      break;
	    case oppuis:
	      if (bi2 == 0)
		return mkIntegerConstant (1);
	      if (bi2 == 1)
		return vexp1;
	      break;
	    case opplus:
	    case opmoin:
	      if (bi2 == 0)
		return vexp1;
	      break;
	    case opmod:
	      break;
	    }
	}
      if (ptexpcal_IsBounded (vexp1))
	if (ptexpcal_IsAssigned (vexp2))
	  {
	    ptexpcal_get_integerBounds (vexp1, &bi1, &bs1);
	    bi2 = ptexpcal_get_integerValue (vexp2);
	    setBoundsExpr (vresult, opRes, bi1, bs1, bi2, bi2, vtree);
	  }
	else
	  {
	    if (ptexpcal_IsBounded (vexp2))
	      {
		ptexpcal_get_integerBounds (vexp1, &bi1, &bs1);
		ptexpcal_get_integerBounds (vexp2, &bi2, &bs2);
		setBoundsExpr (vresult, opRes, bi1, bs1, bi2, bs2, vtree);
	      }
	  }
      else if (ptexpcal_IsBounded (vexp2))
	{
	  ptexpcal_get_integerBounds (vexp2, &bi2, &bs2);
	  if (divparzero (opRes, bi2, bs2))
	    pK_messenger->errorTree (errUnexpectedExpr, BACK_ANNOT_GET_ABSTRACT_TREE(vtree));
	}
    }
  return vresult;
}


/** @return the value of n1**n2. */
int
ExprManager::puissance (int n1, int n2)
{
  int x, i;

  x = 1;
  if (n2 > 0)
    for (i = 1; i <= n2; i++)
      x = x * n1;
  return x;
}

/**
   @return "true" if for a division (ope) the values ent1 and ent2 verify ent1<= 0 and ent2 >= 0, "false" otherwise.
*/
boolean
ExprManager::divparzero (noperators ope, int ent1, int ent2)
{
  if (ope == opdivi)
    return ((ent1 <= 0) && (ent2 >= 0));
  return pk_false;
}


/** @return the value of a binary operation (ope) with ent1 and ent2 as arguments.
    Restricted to operators: -, /, *, +, **; modulo.
 */
compilerInteger
ExprManager::computeValue (noperators ope, compilerInteger ent1, compilerInteger ent2)
{
  switch ((int) ope)
    {
    case opnega:
      return -ent1;
    case opdivi:
      return ent1 / ent2;
    case opmoin:
      return ent1 - ent2;
    case opmult:
      return ent1 * ent2;
    case opplus:
      return ent1 + ent2;
    case oppuis:
      return puissance (ent1, ent2);
    case opmod:
      return ent1 % ent2;
    default:
      pK_messenger->Caseerror (currentLine, currentFile);
      return 0;
    }
}

/**
   It produces the lower bound (bi) and upper bound (bs) of the interval defined by bi1, bi2, bs1, bs2 depending on an operator (ope)
   @return true if the operation is possible on the interval, "false" otherwise (divide by 0)
*/
boolean
  ExprManager::computeBounds (noperators ope, compilerInteger bi1, compilerInteger bi2, compilerInteger bs1, compilerInteger bs2,
			      compilerInteger * bi, compilerInteger * bs)
{
  compilerInteger A[5];
  compilerInteger i, nbele, bs2p;

  if ((ope == opdivi) && (bi2 <= 0) && (bs2 >= 0))
    return pk_false;
  if ((ope == oppuis) && ((bs2 > 0) && (bs2 != bi2) && (bs2 % 2 != 0) && (bi1 <= 0)))
    {
      bs2p = bs2 - 1;
      nbele = 6;
      A[3] = computeValue (ope, bi1, bs2p);
      A[4] = computeValue (ope, bs1, bs2p);
    }
  else
    nbele = 4;
  *bi = computeValue (ope, bi1, bi2);
  *bs = *bi;
  A[0] = computeValue (ope, bi1, bs2);
  A[1] = computeValue (ope, bs1, bi2);
  A[2] = computeValue (ope, bs1, bs2);
  i = 2;
  do
    {
      if ((*bi) > A[i - 2])
	*bi = A[i - 2];
      else if ((*bs) < A[i - 2])
	*bs = A[i - 2];
      i = i + 1;
    }
  while (!(i > nbele));
  return pk_true;
}

/** @return min(a,b) */
compilerInteger
ExprManager::minimum (compilerInteger a, compilerInteger b)
{
  if (a <= b)
    return a;
  return b;
}

/** @return max(a,b) */
compilerInteger
ExprManager::maximum (compilerInteger a, compilerInteger b)
{
  if (a >= b)
    return a;
  return b;
}
