/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   It adds an iteration-precedence vinstrFrom->vinstrTo.
*/
    void
 NodeManager::addIterationPrecedence(ptINode vinstrFrom, ptINode vinstrTo)
{
  ptldependence li;
  ptdependence aDep;

  li = get_iterationPrecedence(vinstrFrom);
  if (!li->exist_dependence(vinstrFrom, vinstrTo, &aDep))
    {
      aDep = new Idependence(vinstrFrom, vinstrTo, hundef);
      li = li->insert(aDep);
      set_iterationPrecedence(vinstrFrom, li);
    }
}

/**
   It adds a dependence aNode -> vto, for (x, vto) an element of lDep.
*/
void NodeManager::addDependenceFromTo(ptINode aNode, ptldependence lDep)
{
  ptldependence li;
  ptinsth h1;
  ptINode n, vfrom, vto;
  ptdependence aDep;
  ptlINode ln;
  ptlinsth lh;

  li = lDep;
  while (li->get_next(&aDep, &li))
    {
      h1 = aDep->getDependence(&vfrom, &vto);
      addDependence(aNode, vto, hundef);	// revoir hundef
      if (is_subgraph(vto))
	{
	  get_subGraphDefinitions(vto, &ln, &lh);
	  while (ptlINode_get_elem(&ln, &n))
	    addDependence(aNode, n, hundef);	// revoir hundef      
	}
    }
}

/**
   It adds a dependence n1 -> aNode, for (n1, x) an element of lDep.
*/
void NodeManager::addDependenceFromTo(ptldependence lDep, ptINode aNode)
{
  ptldependence li;
  ptinsth h1;
  ptINode vfrom, vto;
  ptdependence aDep;

  li = lDep;
  while (li->get_next(&aDep, &li))
    {
      h1 = aDep->getDependence(&vfrom, &vto);
      addDependence(vfrom, aNode, h1);
    }
}

/**
   It adds a dependence h -> aNode, for h an element of lh.
*/
void NodeManager::addDependenceFromTo(ptlinsth lh, ptINode aNode)
{
  ptinsth vh;
  while (ptlinsth_get_next(&lh, &vh))
    addDependence(vh, aNode);
}

/**
   @return "true" if a dependence n -> vinstr when vh, with n the node of a signal (aSig), is added, "false" otherwise.
   It is NOT added when:
   - the signal is a state variable,
   - the expresion (aSig) is a reference to a previous iteration (x[?]): in this case, an iteration-precedence (addIterationPrecedence()) from vinstr to x is added.
   - when the node of  the signal is an element of a "delay cluster"
   - the expresion (aSig) is a reference to an element of an array (x[E]) and E is index independant.
*/
boolean NodeManager::addDependenceSig2Node(ptexpcal aSig, ptINode vinstr, ptinsth vh)
{
  ptINode node;
  noperators op;
  ptlinstances l;
  ptexpcal e, e1;
  boolean b;

  if (EM->IsConstant(aSig))
    return pk_false;
  if (EM->isReferenceToStateVar(aSig))
    return pk_false;		// ref to a state variable x?
  op = EM->getOperator(aSig);
  if (op == opsdec)		/* x[?] */
    {
      e1 = EM->getUnaryArg(aSig);
      node = EM->getNode(e1);
      if (node != vinstr)
	addIterationPrecedence(vinstr, node);
      return pk_false;
    }

  b = pk_false;
  if (op == opelem)
    {
      if (EM->getAttribute(aSig, AccessToArrayElement, (attributeValue *) & node))
	{
	  addDependence(node, vinstr, vh);
	  b = pk_false;
	}
      else
	{
	  e1 = EM->getBinaryArgstAt(aSig, 1);
	  e = EM->getBinaryArgstAt(aSig, 2);
	  l = EM->getSignalsReferencedIn(e);
	  if (l != NILLINSTANCES)
	    {
	      b = addDependenceSig2Node(e1, vinstr, vh);
	      addDependenceSigList2NodeAt(l, vinstr, vh);
	    }
	  else			// constant/index expression.
	  if (EM->isIndexDependent(e))
	    b = addDependenceSig2Node(e1, vinstr, vh);
	  ptlinstances_free(l);
	}
    }
  else if (EM->getOperatorClass(aSig) == variabl)
    {
      node = EM->getNode(aSig);
      if (!existPredefinedPragma(getPragmas(node), PRAGMA_DELAYCLUSTER))
	{
	  addDependence(node, vinstr, vh);
	  b = pk_true;
	}
    }
  return b;
}

/**
   It applies addDependenceSig2Node() for each signal of a list (l).
*/
void NodeManager::addDependenceSigList2NodeAt(ptlinstances l, ptINode n, ptinsth h)
{
  ptexpcal v;
  while (ptlinstances_get_next(&l, &v))
    addDependenceSig2Node(v, n, h);
}

/** 
    @return "true" if it exists a dependence from a node of the subGraph (sg1) to a node of another sub-graph (sg2),
    "false" otherwise.
*/
boolean NodeManager::existDependenceFromToOneNodeOf(ptINode sg1, ptINode sg2)
{
  ptlINode ln1, ln2, lsucc;
  ptlinsth clocks1, clocks2;
  ptinsth h1;
  ptINode n1;

  getNodesClocksDefinedIn(sg1, &ln1, &clocks1);
  getNodesClocksDefinedIn(sg2, &ln2, &clocks2);

  while (ptlINode_get_elem(&ln1, &n1))
    if (existDependenceFromNodeToOneNode(n1, ln2))
      return pk_true;

  while (ptlinsth_get_next(&clocks1, &h1))
    {
      lsucc = CKM->getNodeReferencesTo(h1);
      if (!ptlINode_intersection_isEmpty(lsucc, ln2))
	return pk_true;
    }
  return pk_false;
}

/** 
    @return "true" if it exists a dependence from a node (node) to a node of lnodes,
    "false" otherwise. 
*/
boolean NodeManager::existDependenceFromNodeToOneNode(ptINode node, ptlINode lnodes)
{
  ptINode vfrom, vto;
  ptinsth h1;
  ptldependence lsucc;
  ptdependence aDep;
  ptlinsth lh;
  ptinsth vh;
  ptlINode luse;

  lsucc = get_successors(node);
  while (lsucc->get_next(&aDep, &lsucc))
    {
      h1 = aDep->getDependence(&vfrom, &vto);
      if (ptlINode_isIn(lnodes, vto))
	return pk_true;
    }
  lh = get_clockSuccessors(node);
  while (ptlinsth_get_next(&lh, &vh))
    {
      luse = CKM->getNodeReferencesTo(vh);
      while (ptlINode_get_elem(&luse, &vto))
	if (ptlINode_isIn(lnodes, vto))
	  return pk_true;
    }

  return pk_false;
}

/**
   It may add a dependence from the node of signal (aSig) to a clock (vclk).
   aSig can be:
   - x[?], the dependence is not added
   - when the node of  the signal is an element of a "delay cluster"
   - the expresion (aSig) is a reference to an element of an array (x[E]) and E is index independant.
*/
void NodeManager::addDependenceSignalToClock(ptexpcal aSig, ptinsth vclk)
{
  ptINode node;
  noperators op;
  ptexpcal e;
  ptlinstances l;

  op = EM->getOperator(aSig);
  if (op == opsdec)		/* x[?] */
    return;
  if (op == opelem)
    {
      e = EM->getBinaryArgstAt(aSig, 2);
      l = EM->getSignalsReferencedIn(e);
      if (l != NILLINSTANCES)
	{
	  addDependenceSignalToClock(EM->getBinaryArgstAt(aSig, 1), vclk);
	  addDependenceSignalsToClock(l, vclk);
	}
      else			// constant/index expression.
      if (EM->isIndexDependent(e))
	addDependenceSignalToClock(EM->getBinaryArgstAt(aSig, 1), vclk);
      ptlinstances_free(l);
      return;
    }
  node = EM->getNode(aSig);
  if (!existPredefinedPragma(getPragmas(node), PRAGMA_DELAYCLUSTER))
    addDependence(node, vclk);
}

/**
   It adds a dependence n -> vinsth, for n the node of a signal of linst.
   It applies addDependenceSignalToClock() for each element.
*/
void NodeManager::addDependenceSignalsToClock(ptlinstances linst, ptinsth vinsth)
{
  ptexpcal aSig;
  while (ptlinstances_get_next(&linst, &aSig))
    addDependenceSignalToClock(aSig, vinsth);
}

/**
   For a node (anode) not a sub-graph and a sub-graph (aSubGraph),
   it adds a dependence from each clock successor of anode to aSubGraph, if the clock in not in the 
   sub-graph.
*/
void NodeManager::addDependenceBooleanClockTo(ptINode anode, ptINode aSubGraph)
{
  ptlinsth lsucch, clocks;
  ptlINode lnodes, lsucc;
  ptinsth h1;

  if (ptINode_satisfies_subGraph(anode)) return;
  if (!ptINode_satisfies_subGraph(aSubGraph)) return;

  get_subGraphDefinitions(aSubGraph, &lnodes, &clocks);
  lsucch = get_clockSuccessors(anode);
  while (ptlinsth_get_next(&lsucch, &h1))
    {
      h1 = CKM->getRepresentativeFormula(h1);
      if (!ptlinsth_sorted_isin_elem(clocks, h1))
	{
	  lsucc = CKM->getNodeReferencesTo(h1);
	  if (!ptlINode_intersection_isEmpty(lsucc, lnodes))
	    addDependence(h1, aSubGraph);
	}
    }
}

/**
   @return the fusion of two lists of dependences (ldepref, ldepapp) with side effect on ldepref.
   For (x->y when h1) and (x->y when h2) is the lists, it produces: x->y when (h1 ^+ h2).
*/
ptldependence NodeManager::fusionDependences(ptldependence ldepref, ptldependence ldepapp)
{
  ptldependence ldepres;
  ptdependence aDep, aDep1;
  ptINode vTo, vFrom;
  ptinsth h;

  ldepres = ldepref;
  while (ldepapp->get_next(&aDep, &ldepapp))
    {
      h = aDep->getDependence(&vFrom, &vTo);
      if (ldepres->exist_dependence(vFrom, vTo, &aDep1))
	aDep1->setClockDependence(CKM->makeSup(h, aDep1->getClockDependence()));
      else
	ldepres = ldepres->insert(aDep);
    }
  return ldepres;
}

/**
   For a undersampling (h), it adds depedences from x -> h, for x a signal referenced in the expression associated with the undersampling.
   It calls addDependenceSignalsToClock() for the positive/negative extractions.
*/
boolean NodeManager::setDependencesDownSampling(ptinsth h)
{
  ptlinstances lsig;
  ptinsth vh2;

  if (CKM->getOperator(h) != sub)
    return pk_false;
  lsig = CKM->getSignalsOfCondition(h);
  addDependenceSignalsToClock(lsig, h);
  CKM->getComplExtraction(h, &vh2); /* nosub */
  addDependenceSignalsToClock(lsig, vh2);
  return pk_true;
}

/**
   @return
   - "true" if a node (node) is not a node of nodes (ln), taking into account the fields of the structured types, 
   - "false" otherwise.
   If one of the internal nodes of a bundle/structure is the list (ln), node is assumed to be in ln.
*/
boolean NodeManager::testArrayDependence(ptINode node, ptlINode ln)
{
  ptexpcal outs, vsig;
  ptlinstances lsig;
  ptINode cn, n;
  // UNUSED ---
  if (!is_lfields(node))
    {
      if (ptlINode_isIn(ln, node))
	return pk_false;
      if (get_inComposedNode(node, &cn))
	return testArrayDependence(cn, ln);
      return pk_true;
    }
  outs = get_leftOperand(node);
  if (!EM->getFieldsOf(outs, &lsig))
    return !ptlINode_isIn(ln, node);
  while (ptlinstances_get_next(&lsig, &vsig))
    {
      n = EM->getNode(vsig);
      if (n != node)
	if (testArrayDependence(n, ln))
	  return pk_true;
    }
  return pk_false;
}

boolean NodeManager::isNodeInComposedNode(ptINode aNode, ptINode composedNode)
{
  ptINode vblock;
  if (!get_inComposedNode(aNode, &vblock))
    return pk_false;
  if (vblock == composedNode)
    return pk_true;
  return isNodeInComposedNode(vblock, composedNode);
}

/**
   Internal to addComposedNodesDependences().
   For an array node (arrayNode), the node of its index (IndexNode), and ln, assumed to be the internal nodes of arrayNode, it adds
   a dependence from aNode to arrayNode when it is not IndexNode, arrayNode, and not in ln. Moreover, if arrayNode is in a block, a dependence 
   from  aNode to this block is also added.
*/
void NodeManager::addComposedNodesDependence(ptINode aNode, ptINode arrayNode, ptINode IndexNode, ptINode internalArrayNode, ptlINode ln)
{
  // addArrayDependence renamed addComposedNodesDependence April 2014 LBesnard
  ptlinsth lh;
  ptlINode lnvBlock;
  ptINode vBlock, vBlockInternalArrayNode;

  if ((aNode != IndexNode) && (aNode != arrayNode) && (aNode != internalArrayNode))
    {
      // if (testArrayDependence(aNode, ln))
      if (!isNodeInComposedNode(aNode, arrayNode))	// tester si aNode n'est pas interne au block arrayNode
	{
	  addDependence(aNode, arrayNode, hundef);
	  if (getInArrayNode(arrayNode, &vBlock))
	    {
	      get_array_bodyDefinitions(vBlock, &lnvBlock, &lh);
	      get_array_inputNode(vBlock, &vBlockInternalArrayNode);
	      addComposedNodesDependence(aNode, vBlock, get_indexNode(vBlock), vBlockInternalArrayNode, lnvBlock);
	    }
	  else;			// subgraph ?
	}
    }
}

/**
   For an array node (arrayNode), the node of its index (IndexNode), ln the nodes of its body, lh the clocks of its body, it adds
   dependences from the predecessors (nodes and clocks) of a node of ln, to the arrayNode (addArrayDependence()) for not internal nodes of the array.

   lh is currently unused.
*/
void NodeManager::addComposedNodesDependences(ptINode arrayNode, ptINode IndexNode, ptINode internalArrayNode, ptlINode ln, ptlinsth lh)
{
  // addArrayDependences renamed addComposedNodesDependences April 2014 LBesnard.
  ptINode n, n11, n12;
  ptinsth vh, hdep;
  ptlinsth lp;
  ptlINode lnSave;
  ptldependence lpred;
  ptdependence aDep1;

  lnSave = ln;
  while (ptlINode_get_elem(&ln, &n))
    {
      lpred = get_predecessors(n);
      while (lpred->get_next(&aDep1, &lpred))
	{
	  hdep = aDep1->getDependence(&n11, &n12);	// n12 == n
	  addComposedNodesDependence(n11, arrayNode, IndexNode, internalArrayNode, lnSave);
	}
      lp = get_clockPredecessors(n);
      while (ptlinsth_get_next(&lp, &vh))
	{
	  vh = CKM->getRepresentativeFormula(vh);
	  if (!CKM->satisfiesInitClock(vh))
	    if (!CKM->getInComposedNode(vh, &n11))
	      addDependence(vh, arrayNode);
	    else if (n11 != arrayNode)
	      addDependence(vh, arrayNode);
	}
    }
}

/**
   (arrays, subgraphs).
   It adds a dependence from IndexNode
   - to each node of ln
   - to each clock of lh
   
   vhemb is currently unused.
*/
void NodeManager::addInternalDependences(ptINode IndexNode, ptlINode ln, ptlinsth lh, ptinsth vhemb)
{
  ptINode aNode;
  ptinsth aClock, hdep;

  while (ptlINode_get_elem(&ln, &aNode))
    {
      hdep = hundef;		// hdep = CKM->makeInf(vhemb, get_nodeClock(aNode));        
      addDependence(IndexNode, aNode, hdep);
    }
  while (ptlinsth_get_next(&lh, &aClock))
    addDependence(IndexNode, aClock);
}

/**
   It removes the dependences, from nodes, reaching a node (node).
*/
void NodeManager::removeNodeDependencesReaching(ptINode node)
{
  ptldependence lpred;
  ptdependence aDep;
  ptINode in;

  lpred = get_predecessors(node);
  while (lpred->get_next(&aDep, &lpred))
    // if (!aDep-> satisfiesExplicitDependence()) ?? conservation des explicites ??
    removeDependence(aDep);
  // node->set_predecessors( NILLDEPENDENCE);
  if (get_subGraphInputNode(node, &in))
    removeNodeDependencesReaching(in);	// added Sep 2014
}

/**
   It removes the dependences, from clocks, reaching node.
*/
void NodeManager::removeClockDependencesReaching(ptINode node)
{
  ptlinsth lh;
  ptinsth vh;

  lh = get_clockPredecessors(node);
  while (ptlinsth_get_next(&lh, &vh))
    removeDependence(CKM->getRepresentativeFormula(vh), node);
  set_clockPredecessors(node, NILLINSTH);
}

/** It removes the dependences with vh as weight. */
void NodeManager::removeDependenceWithLabel(ptinsth vh)
{
  ptldependence l;
  ptdependence aDep;
  ptinsth dCk;

  l = GdependencesList;
  while (l->get_next(&aDep, &l))
    {
      dCk = aDep->getClockDependence();
      dCk = CKM->getRepresentativeFormula(dCk);
      if (dCk == vh)
	removeDependence(aDep);
    }
}

/**
   It adds dependences from signals referenced in the selectors used in an access to element of an array.
   For example, for a node n defined by  x[y] and y a signal, a dependence is added from the node of y to the node n, with h as the clock of the 
   dependence.
*/
void NodeManager::addDependencesForOutputs(ptexpcal s, ptINode n, ptinsth h)
{
  ptlinstances ls;
  noperators op;

  op = EM->getOperator(s);
  if (op == opelem)
    {
      ls = EM->getPredecessors(EM->getBinaryArgstAt(s, 2));
      addDependenceSigList2NodeAt(ls, n, h);
    }
}

/**
   It adds dependences from signals referenced in the selectors used in an access to element of an array for output signal of a node n.
   For example, for a node n defined by  x[y] := E and y a signal, a dependence is added from the node of y to the node n, with h as the clock of the 
   dependence.
*/
void NodeManager::addDependencesForOutputs(ptINode n, ptinsth h)
{
  ptexpcal e, s;

  e = get_leftOperand(n);
  if (e != NILEXPCAL)
    if (!EM->satisfiesList(e))
      addDependencesForOutputs(e, n, h);
    else
      while (EM->getNaryArgs(&e, &s))
	addDependencesForOutputs(s, n, h);
}

