/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#define isSignalOfInterface(vexp) (!EM->isSharedSignal(vexp) && !EM->isStateVar(vexp) && !EM->isSharedSignalBySeveralGraphs(vexp))

_PUBLIC_
/** 
    It inserts in an interface (vinterface) the identifiers of a list (lins) taking into account the status of the identifier (signal, parameter, constant).
    For a signal it is added in the inputs of the interface. For the outputs, see ptinterface_insert_outputSignal(), ptinterface_set_outputSignals() functions.

    @return the modified interface (it can be created).
*/
    ptinterface NodeManager::inSplipListToInterface(ptinterface vinterface, ptlinstances lins)
{
  ptexpcal vsig;

  vinterface = ptinterface_set_inputSignals(vinterface, NILLINSTANCES);
  while (ptlinstances_get_next(&lins, &vsig))
    if (!EM->IsSignal(vsig))
      {
	if (EM->IsParameter(vsig))
	  ptinterface_insert_parameter(vinterface, vsig);
	else if (EM->IsConstant(vsig))
	  ptinterface_insert_constant(vinterface, vsig);
	// to complete for: indice, label
      }
    else if (isSignalOfInterface(vsig))
      vinterface = ptinterface_insert_inputSignal(vinterface, vsig);
    else
      vinterface = ptinterface_insert_inOutSignal(vinterface, vsig);

  return vinterface;
}

ptinterface NodeManager::outSplipListToInterface(ptinterface vinterface, ptlinstances louts)
{
  ptexpcal vsig;

  vinterface = ptinterface_set_outputSignals(vinterface, NILLINSTANCES);
  while (ptlinstances_get_next(&louts, &vsig))
    if (isSignalOfInterface(vsig))
      vinterface = ptinterface_insert_outputSignal(vinterface, vsig);
    else
      vinterface = ptinterface_insert_inOutSignal(vinterface, vsig);
  return vinterface;
}

/**
   @return the interface of a node (aNode).
   \see getSet_interface.
*/
ptinterface NodeManager::get_interface(ptINode aNode)
{
  return ptINode_get_interface(aNode);
}

/**
   It assigns an interface (vinter) to a node (aNode).
   It can be a reseting (vinter= NILINTERFACE).
*/
void NodeManager::set_interface(ptINode aNode, ptinterface vinter)
{
  ptlINode ln;
  ptINode n;
  ptlinsth lh;
  boolean b;
  ptgfd vgfd;

  ptINode_set_interface(aNode, vinter);
  if (vinter == NILINTERFACE)
    {
      interface_free(ptINode_get_interface(aNode));
      // reset.
      b = pk_true;
      if (ptINode_satisfies_subGraph(aNode))
	{
	  vgfd = ptINode_get_subGraph(aNode);
	  get_subGraphDefinitions(aNode, &ln, &lh);
	}
      else if (ptINode_satisfies_array(aNode))
	{
	  vgfd = get_array_bodyGraph(aNode);
	  get_array_bodyDefinitions(aNode, &ln, &lh);
	}
      else
	b = pk_false;

      if (b)
	{
	  interface_free(ptgfd_get_interface(vgfd));
	  ptgfd_set_interface(vgfd, NILINTERFACE);
	  while (ptlINode_get_elem(&ln, &n))
	    set_interface(n, NILINTERFACE);
	}
    }
}

/** @return the interface of a node (n). If it is undefined, the interface is computed. */
ptinterface NodeManager::getSet_interface(ptINode n)
{
  ptinterface vinterface;

  vinterface = ptINode_get_interface(n);
  if (vinterface == NILINTERFACE)
    return buildinterface(n);
  return vinterface;
}

/**
   @return the list of the clocks referenced in a node (n).
   The interface of the node is computed if it is undefined.
*/
ptlinsth NodeManager::getInodelInputClocks(ptINode n)
{
  ptinterface vinterface;

  vinterface = getSet_interface(n);
  if (vinterface == NILINTERFACE)
    return NILLINSTH;
  return ptlinsth_duplicate(ptinterface_get_inputClocks(vinterface));
}

/**
   @return the list of the signals referenced in a node (n).
   The interface of the node is computed if it is undefined.
*/
ptlinstances NodeManager::getInodelInputFlows(ptINode n)
{
  ptinterface vinterface;

  vinterface = getSet_interface(n);
  if (vinterface == NILINTERFACE)
    return NILLINSTANCES;
  return ptlinstances_copy(ptinterface_get_inputSignals(vinterface));
}

/**
   @return the list of the signals defined in a node (n).
   The interface of the node is computed if it is undefined.
*/
ptlinstances NodeManager::getInodelOutputFlows(ptINode n)
{
  ptinterface vinterface;

  vinterface = getSet_interface(n);
  if (vinterface == NILINTERFACE)
    return NILLINSTANCES;
  return ptlinstances_copy(ptinterface_get_outputSignals(vinterface));
}

/**
   @return the list of the local signals of a node (n).
   The interface of the node is computed if it is undefined.
*/
ptlinstances NodeManager::getInodelLocalFlows(ptINode n)
{
  ptinterface vinterface;

  vinterface = getSet_interface(n);
  if (vinterface == NILINTERFACE)
    return NILLINSTANCES;
  return ptlinstances_copy(ptinterface_get_localSignals(vinterface));
}

/**
   @return the list of the in-out signals of a node (n).
   The interface of the node is computed if it is undefined.
*/
ptlinstances NodeManager::getInodelInOutFlows(ptINode n)
{
  ptinterface vinterface;

  vinterface = getSet_interface(n);
  if (vinterface == NILINTERFACE)
    return NILLINSTANCES;
  return ptlinstances_copy(ptinterface_get_inOutSignals(vinterface));
}

/**
   @return the list of the constants referenced in a node (n).
   The interface of the node is computed if it is undefined.
*/
ptlinstances NodeManager::getInodelConstants(ptINode n)
{
  ptinterface vinterface;

  vinterface = getSet_interface(n);
  if (vinterface == NILINTERFACE)
    return NILLINSTANCES;
  return ptlinstances_copy(ptinterface_get_constants(vinterface));
}

/** Interface of a subgraph.
    It produces the bDC+ interface of a node (aSubGraph).
    The reference to the clocks are replaced by the references to the boolean signals that represent the clocks.
 */
void NodeManager::SubGraphInterfaceCalculusUndersamplingClocks(ptINode aSubGraph)
{
  ptlinsth lh;
  ptinsth vh;
  ptinterface vinterface;
  ptlinstances l;
  ptexpcal x;

  vinterface = getSet_interface(aSubGraph);
  lh = ptinterface_get_inputClocks(vinterface);
  while (ptlinsth_get_next(&lh, &vh))
    {
      l = CKM->getSignalsOfCondition(vh);
      while (ptlinstances_get_next(&l, &x))
	ptinterface_insert_inputSignal(vinterface, x);
    }
  ptinterface_set_inputClocks(vinterface, NILLINSTH);

  lh = ptinterface_get_localClocks(vinterface);
  while (ptlinsth_get_next(&lh, &vh))
    {
      l = CKM->getSignalsOfCondition(vh);
      while (ptlinstances_get_next(&l, &x))
	ptinterface_insert_localSignal(vinterface, x);
    }
  ptinterface_set_localClocks(vinterface, NILLINSTH);

  lh = ptinterface_get_outputClocks(vinterface);
  while (ptlinsth_get_next(&lh, &vh))
    {
      l = CKM->getSignalsOfCondition(vh);
      while (ptlinstances_get_next(&l, &x))
	if (!EM->IsInputFlow(x))
	  ptinterface_insert_outputSignal(vinterface, x);
    }

  ptinterface_set_outputClocks(vinterface, NILLINSTH);
}

/**
   It calls bdcplusSubGraphInterfaceCalculus(ptINode) for each sub-graph of the node manager.
*/
void NodeManager::bdcplusSubGraphsInterfaceCalculus()
{
  ptINode aSubGraph;
  ptlINode ln;

  if (getSubGraphs(&ln))
    while (ptlINode_get_elem(&ln, &aSubGraph))
      SubGraphInterfaceCalculusUndersamplingClocks(aSubGraph);
}

/**
   Updating the interface of the nodes of the operand list(ln)
*/
void NodeManager::updateInterfaceNodes(ptlINode ln)
{
  ptINode node;

  while (ptlINode_get_elem(&ln, &node)) 
    updateInterfaceNode(node);
}

/**
   Updating the interface of a node.
*/
void NodeManager::updateInterfaceNode(ptINode node)
{
  set_interface(node, NILINTERFACE);
  getSet_interface(node);
}


void NodeManager::mkSubGraphProjectedSystemInfos(ptINode SubGraph)
{
  ptlinsth lh, lhtodeclare;
  ptinsth vhemb, vh;
  ptinterface vinterface;
  ptlINode lnSubGraph;
  ptlinsth lhSubGraph;

  // this method is quite similar to the IGraph::IGraphUpstreamProjectionSubGraph().
  lh = EM->getClockOfSignals(getInodelInputFlows(SubGraph));
  lh = ptlinsth_sorted_fusion(lh, EM->getClockOfSignals(getInodelOutputFlows(SubGraph)));
  lh = ptlinsth_sorted_fusion(lh, getInodelInputClocks(SubGraph));
  vhemb = CKM->getBranchingClocks(lh);
  lhtodeclare = CKM->CompleteUpstreamParametrization(lh, vhemb);
  vh = CKM->getFather(vhemb);
  if (!CKM->satisfiesFictitiousRoot(vh))
    vhemb = vh;
  set_subGraphProjectedSystemInfos(SubGraph, vhemb, lhtodeclare);
  // updating the interface.
  vinterface = getSet_interface(SubGraph);
  while (ptlinsth_get_next(&lhtodeclare, &vh))
    ptinterface_insert_inputClock(vinterface, CKM->getRepresentativeFormula(vh));	// added only if it is not an output/a local.
}

/** 
   It updates the interfaces of the operand nodes by applying the rules
   - x an input (signal/clock) of vinterface1, and local to vinterface2, it becomes an output of vinterface2
   - x input (signal/clock) of vinterface2, and local to vinterface1, it becomes an output of vinterface1
*/
void NodeManager::updateInterface(ptINode aSubGraph1, ptINode aSubGraph2)
{
  ptinterface vinterface1, vinterface2;
  ptlinstances l1, l2;
  ptexpcal s;
  ptlinsth lh1, lh2;
  ptinsth h;

  vinterface1 = getSet_interface(aSubGraph1);
  vinterface2 = getSet_interface(aSubGraph2);

  l1 = ptinterface_get_inputSignals(vinterface1);
  l2 = ptinterface_get_localSignals(vinterface2);
  while (ptlinstances_get_next(&l1, &s))
    if (ptlinstances_isInSortedList(l2, s))
      ptinterface_insert_outputSignal(vinterface2, s);

  l1 = ptinterface_get_inputSignals(vinterface2);
  l2 = ptinterface_get_localSignals(vinterface1);
  while (ptlinstances_get_next(&l1, &s))
    if (ptlinstances_isInSortedList(l2, s))
      ptinterface_insert_outputSignal(vinterface1, s);

  lh1 = ptinterface_get_inputClocks(vinterface1);
  lh2 = ptinterface_get_localClocks(vinterface2);
  while (ptlinsth_get_next(&lh1, &h))
    if (ptlinsth_sorted_isin_elem(lh2, h))
      ptinterface_insert_outputClock(vinterface2, h);

  lh1 = ptinterface_get_inputClocks(vinterface2);
  lh2 = ptinterface_get_localClocks(vinterface1);
  while (ptlinsth_get_next(&lh1, &h))
    if (ptlinsth_sorted_isin_elem(lh2, h))
      ptinterface_insert_outputClock(vinterface1, h);
}

/**
   It switches the outputs (clocks and signals) to the locals of a node (aNode), 
   excepts some signals (lexcept) and some clocks (lhexcept).
*/
void NodeManager::moveOutputs2localsExcept(ptINode aNode, ptlinstances lexcept, ptlinsth lhexcept)
{
  ptinterface vinterface;

  vinterface = getSet_interface(aNode);
  moveOutputs2localsExceptSignals(vinterface, lexcept);
  moveOutputs2localsExceptClocks(vinterface, lhexcept);
}
