/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_ void
 NodeManager::insertNodeToArray(ptINode aNode, ptexpcal vexp)
{
  ptINode nodeArrayIndex, n;
  ptlinstances lsig, lindi;

  lsig = EM->getSignalsReferencedIn(vexp);
  lindi = EM->getIndexesReferencedIn(vexp);
  if (arrayGetMaxLevelIndex(lsig, lindi, &nodeArrayIndex))
    {
      get_inComposedNode(nodeArrayIndex, &n);
      linkInodeToInodeArray(aNode, n);
      arrayBodyInsertNode(n, aNode);
    }
}

/** It creates the equation vsig := (vexp when vhcal); vsig is synchronized with the clock vh; 
    @param vsig (in) the signal defined by the equation.
    @param vexp (in) expression definition of X;
    @param vh (in) the clock of X;
    @param vhcal (in) the contributon clock of vexp to the definition of X.
    @return the node of the defined signal (i.e vsig).
*/
ptINode NodeManager::insertEquation(ptexpcal vsig, ptexpcal vexp, ptinsth vh, ptinsth vhcal)
{
  ptINode aNode, nodeArrayIndex, n;
  ptlinstances lsig, lindi;

  lsig = EM->getPredecessors(vexp);
  aNode = mk_node_equation(vsig, vexp, vhcal, lsig, NILBACKANNOTATION);
  EM->setClock(vsig, vh);
  insertNode(aNode, vh);
  insertNodeToArray(aNode, vexp);
  return aNode;
}

/** It creates a new signal X then the equation X := (vexp when vhcal); X is synchronized with the clock vh; 
   @param id  (in) identifier of the new signal;
   @param TT (in) the type of the new signal;
   @param vexp (in) expression definition of X;
   @param vh (in) the clock of X;
   @param vhcal (in) the contributon clock of vexp to the definition of X.
   @param node (out) the node of X.
   @return the created signal (i.e X) */
ptexpcal NodeManager::insertEquation(ptident id, TypeType TT, ptexpcal vexp, ptinsth vh, ptinsth vhcal, ptINode * node)
{
  ptexpcal newSig;

  newSig = EM->mkIdentExp(id, tsignal, TT, mk_node_undefined(), vh, NILBACKANNOTATION);
  *node = insertEquation(newSig, vexp, vh, vhcal);
  return newSig;
}

/** It creates an equation C1 = leftt(c,bk) where C1 is a new signal with vh as clock.
   @param c (in) a boolean clock expression.
   @param bk (in) a boolean clock expression such that father(bk) = vh on the branching (vh, ^c) 
   @param vh (in) the clock of the expression "leftt(c, bk)"
   @return the created SIGNAL defined by lefttt expression.
 */
ptexpcal NodeManager::createLeftt(ptexpcal c, ptexpcal bk, ptinsth vh)
{
  ptexpcal vexp, newSig;
  ptinsth vhsig;
  ptident id;
  ptINode aNode;

  EM->getSynchronousExprHorl(c, &vhsig);
  if (vhsig == vh)
    return c;
  vexp = EM->mkBinaryExp(opleft_tt, c, bk, NILBACKANNOTATION);
  // if (EM->IsSignal(c)) id =  EM->getOriginalName(c); else id =  getIdentNil();
  id = EM->getNameBooleanExpression(vexp);
  newSig = insertEquation(id, get_BooleanType(), vexp, vh, vh, &aNode);
  return newSig;
}

/** 
    @return 
     - the expression c, when force is "false" and when clock(c) is equal to the clock vh; 
     - oherwise a new signal, defined by the equation c when vh, and synchronized with the clock vh. 
    @param force (in) at pk_true to add a new signal.
    @param c (in) a monochromous expression.
    @param vh (in) the clock of the new signal
*/
ptexpcal NodeManager::createWhen(ptexpcal c, ptinsth vh, boolean force)
{
  ptinsth vhsig;
  ptident id;
  ptINode aNode;
  TypeType t;

  EM->getSynchronousExprHorl(c, &vhsig);
  if (!force)
    if (vhsig == vh)
      return c;
  if (EM->IsFreeType(c))
    t = get_BooleanType();
  else
    t = EM->getType(c);
  if (EM->IsSignal(c))
    id = EM->getOriginalName(c);
  else if (t == BooleanType)
    id = EM->getNameBooleanExpression(c);
  else
    id = getAnonymousId();
  return insertEquation(id, t, c, vh, vh, &aNode);
}

/** @return the input list in which the input node aNode is added if it is an undeclared external model of lExtern.

   @param aNode (in) a  node ; but it works only for external node call;
   @param lExtern (in) : a list of external node call.
   Unicity guaranted using the name of the  external node call.
*/
ptlINode NodeManager::addExternDecl(ptINode aNode, ptlINode lExtern)
{
  ptident nom;
  if (ptINode_get_nameCall(aNode, &nom))
    if (!isDeclaredExternal(nom, lExtern))
      ptlINode_insert_elem(&lExtern, aNode);
  return lExtern;
}

/** @return "true" if the list lExtern contains a external call with n as name, "false" otherwise
    @param n (in) an ident
    @param lExtern (in) : a list of external node call (assumed)
*/
boolean NodeManager::isDeclaredExternal(ptident n, ptlINode lExtern)
{
  ptINode node;
  ptident id;

  while (ptlINode_get_elem(&lExtern, &node))
    if (ptINode_get_nameCall(node, &id))
      if (eq_idents(id, n))
	return pk_true;
  return pk_false;
}

/** @return a new untyped node */
ptINode NodeManager::mk_node_undefined()
{
  ptINode n;
  n = ptINode_create();
  ptINode_init(n, hundef);
  return n;
}

/** It adds (vexpdedef when vhcal) as a partial definition of the signal vexpadef 
    @return the node defining the signal vexpadef.
    The node is not inserted into the graph of the clock of this node, use \ref insertNode for that (when the definition is completed).
*/
ptINode NodeManager::addPartAff(ptexpcal vexpadef, ptexpcal vexpdedef, ptinsth vhcal, ptlinstances instent)
{
  ptINode n, nNode;
  ptlINode ln;
  ptexpcal vexpToadd;

  vexpToadd = EM->mkWhenOrCellExp(opextr, vexpdedef, vhcal);
  nNode = mkPartEquation(vexpadef, vexpToadd, vhcal, instent);
  if (EM->isMarked(vexpadef, PART_DEFAULT_DEFINITION))
    set_partialDefinitionDefaultValue(nNode);
  n = EM->getNode(vexpadef);
  if (is_setOfPartialDefinitions(n))
    ln = get_setOfPartialDefinitions_arg(n);
  else
    ln = NILLINODE;
  ln = ptlINode_append_elem(ln, nNode);
  becomes_setOfPartialDefinitions(n, vexpadef, ln);
  return nNode;
}

/** It adds (vexpdedef when vhcal) as the default-value part of the signal vexpadef.
    @return the node defining the signal vexpadef.
    The node is not inserted into the graph of the clock of this node, use \ref insertNode for that (when the definition is completed).
*/
ptINode NodeManager::addDefaultValuePart(ptexpcal vexpadef, ptexpcal vexpdedef, ptinsth vhcal, ptlinstances instent)
{
  ptINode n;

  EM->addMark(vexpadef, PART_DEFAULT_DEFINITION);
  n = addPartAff(vexpadef, vexpdedef, vhcal, instent);
  EM->removeMark(vexpadef, PART_DEFAULT_DEFINITION);
  return n;
}

/**
  It adds to the node defining the signal expcal1 the expression (expcal2 when vhcal).
  @param expcal1 (in) a signal.
  @param expcal2 (in) a monochronous expression.
  @param vhcal (in) contribution clock of expcal2 to the definition of expcal1.
  @param instent (in) list of referenced signals of the expression expcal2.
  @param treeRef (in) reference to the syntactic tree of the program(tracability).
  @return the node defining the signal vexpcal1.
  The node is not inserted into the graph of the clock of this node, use \ref insertNode for that (when the definition is completed).
*/
ptINode NodeManager::mk_node_equation(ptexpcal expcal1, ptexpcal expcal2, ptinsth vhcal, ptlinstances instent, BackAnnotationType ba)
{
  ptINode n;
  boolean eqPart, bextension;

  n = EM->getNode(expcal1);
  if (CKM->satisfiesNullClock(vhcal))
    {
      nodesMerging(n);
      return n;
    }
  eqPart = (EM->isMarked(expcal1, PART_DEFINITION));
  EM->TypeCheckingEquation(expcal1, &expcal2, ba, &bextension);
  nodesMerging(n);
  if (!eqPart)
    n = addaff(expcal1, expcal2, vhcal, instent);
  else
    n = addPartAff(expcal1, expcal2, vhcal, instent);
  setBackAnnotation(n, ba);
  return n;
}

/**
  It defines the node of an input signal.
  @param i (in) a input signal of the program when NodeCall is undefined or a signal defined in the 
  NodeCall node.
  @param NodeCall (in) a node call 
  @return the node defining the signal i (node operator is opread)

  The node is not inserted into the graph of the clock of this node, use \ref insertNode for that (when the definition is completed).
*/
ptINode NodeManager::mk_node_input(ptexpcal i, ptINode NodeCall, BackAnnotationType ba)
{
  ptINode n;

  n = EM->getNode(i);
  removeInSetCase(n);
  ptINode_becomes_lecenv(n, EM->mkNaryExp(oplident, i, NILEXPCAL, NILBACKANNOTATION), NodeCall);
  setBackAnnotation(n, ba);
  return n;
}

/**
  It makes the writing node of an output signal.
  @param i (in) a output signal of the program.
  @return the writing node of the signal i (node operator is opwrite).
  The node is not inserted into the graph of the clock of this node, use \ref insertNode for that (when the definition is completed).
*/
ptINode NodeManager::mk_node_output(ptexpcal i, BackAnnotationType ba)
{
  ptINode n;

  n = mk_node_undefined();
  ptINode_becomes_ecrenv(n, EM->mkNaryExp(oplident, i, NILEXPCAL, NILBACKANNOTATION));
  setBackAnnotation(n, ba);
  return n;
}

/**
  It defines the node of signal as a delay node (node operator is opret).
  @param s (in) a output signal of the equation.
  @param eDelay (in) the delayed signal.
  @param vdelay (in) the delay expression.
  @param vwindow (in) the window expression.
  @param vinit (in) the initial value.
  @param treeRef (in) the reference to the syntactic tree of the program (tracability).

  @return the node of the signal s.
  The node is not inserted into the graph, use \ref insertNode for that (when the definition is completed).
*/
ptINode NodeManager::mk_node_delay(ptexpcal s, ptexpcal eDelay, ptexpcal vdelay, ptexpcal vwindow, ptexpcal vinit, BackAnnotationType ba)
{
  ptINode n, m;
  ptexpcal x, nSig;
  boolean bextension, b;
  ptinsth vclock;

  n = EM->getNode(s);
  nodesMerging(n);
  setBackAnnotation(n, ba);

  x = EM->mkBinaryExp(opinit, eDelay, vinit, ba);
  x = EM->mkTernaryExp(opret, x, vdelay, vwindow, ba);
  b = pk_true;
  if (EM->TypeCheckingEquation(s, &x, ba, &bextension))
    if (bextension)
      {
	vclock = EM->getClock(s);
	nSig = EM->mkIdentExp(EM->getOriginalName(s), tsignal, EM->getType(x), mk_node_undefined(), vclock, ba);
	m = EM->getNode(nSig);
	ptINode_becomes_delay(m, nSig, x);
	insertNode(m, vclock);
	addaff(s, nSig, vclock, ptlinstances_insert(NILLINSTANCES, nSig));
	return n;
      }
  ptINode_becomes_delay(n, s, x);
  return n;
}

/**
  It defines the node of a set of signals as an external procedure call node (node operator is opappe).
  @param idCall (in) the name of the external procedure;
  @param parameters (in) the set of parameters;
  @param inputs (in) the set of input signals;
  @param outputs (in) the set of output signals.
  @param treeRef (in) the reference to the syntactic tree of the program (tracability).
  @return the procedure node call.
  The node is not inserted into the graph of the clock of this node, use \ref insertNode for that (when the definition is completed).
*/
ptINode NodeManager::mk_procCall(ptident idCall, ptexpcal parameters, ptexpcal inputs, ptexpcal outputs, BackAnnotationType ba)
{
  ptINode n;

  n = mk_node_undefined();
  ptINode_becomes_extern(n, idCall, parameters, inputs, outputs);
  setBackAnnotation(n, ba);
  return n;
}

/**
  It defines the node of a set of signals as a "nodecall" node (node operator is opnodecall)
  @param vgraph (in) graph model.
  @param parameters (in) the set of parameters;
  @param inputs (in) the set of signal inputs;
  @param outputs (in) the set of signal outputs;
  @param treeRef (in) the reference to the syntactic tree of the program (tracability).
  @return the procedure node call.
  The node is not inserted into the graph of the clock of this node, use \ref insertNode for that (when the definition is completed).
*/
ptINode NodeManager::mk_node_call(ptDCG vgraph, ptexpcal parameters, ptexpcal inputs, ptexpcal outputs, BackAnnotationType ba)
{
  ptINode n;

  n = mk_node_undefined();
  ptINode_becomes_nodecall(n, vgraph, parameters, inputs, outputs);
  setBackAnnotation(n, ba);
  return n;
}

/** @return the sub-graph node assigned to a label 
    @param vlabel (in) a label identifier.
    @param treeRef (in) the reference to the syntactic tree of the program (tracability).
    The node is not inserted into the graph of the clock of this node, use \ref insertNode for that (when the definition is completed).
*/
ptINode NodeManager::mk_node_subgraph(ptexpcal vlabel, BackAnnotationType ba)
{
  ptINode n;

  n = EM->getNode(vlabel);
  ptINode_becomes_subGraph(n, vlabel, ptgfd_create());
  setBackAnnotation(n, ba);
  return n;
}

/**
   It defines the node n as "bundle/structure" one which produces the signal s.
   This node is characterized by a list of signals (lsigFields) that must be the signals  s.y1, ....s.yn where yi are the fieds of 
   the "bundle/structure" type of the signal s.
   @param treeRef (in) the reference to the syntactic tree of the program (tracability).
   @return the node n.
   The node is not inserted into the graph of the clock of this node, use \ref insertNode for that (when the definition is completed).
*/
ptINode NodeManager::mk_node_lfields(ptexpcal s, ptINode n, ptlinstances lsigFields, BackAnnotationType ba)
{
  ptINode_set_leftOperand(n, s);
  nodesMerging(n);
  ptINode_becomes_lfields(n, lsigFields);
  insertTable(n, &lBundStruct, &lBundStructEnd);
  return n;
}

/**
   @return a new array node.
   @param indexNode (in) the index of the array, \see {mk_node_index}.
   @param inputs (in) the referenced signals in the array.
   @param outputs (in) the defined signals in the array.
   @param treeRef (in) the reference to the syntactic tree of the program (tracability).
   The node is not inserted into the graph of the clock of this node, use \ref insertNode for that (when the definition is completed).
*/
ptINode NodeManager::mk_node_array(ptINode indexNode, ptexpcal inputs, ptexpcal outputs, BackAnnotationType ba)
{
  ptINode n;

  n = mk_node_undefined();
  ptINode_becomes_array(n, indexNode, ptgfd_create(), ptgfd_create());
  set_inComposedNode(indexNode, n);	// link index --> array

  ptINode_set_leftOperand(n, outputs);
  ptINode_set_rightOperand(n, inputs);
  return n;
}

/**
   @return the node assigned to the index i.
   @param i (in) a index identifier (assumed)
   @param e (in) definition of the index
   \see {mk_node_array}
*/
ptINode NodeManager::mk_node_index(ptexpcal i, ptexpcal e)
{
  ptINode n;
  n = EM->getNode(i);
  ptINode_becomes_assign(n, i, e);
  return n;
}

/** It inserts a node (node) into the AllNodes set, and its typed-node set. */
void NodeManager::insertinSet(ptINode node)
{
  boolean b;
  ptINode n1;

  insertTable(node, &AllNodes, &AllNodesEnd);
  switch (get_oper(node))
    {
    case opnil:		// undefined
      break;

    case opdefpart:		// x ::= E
      break;

    case opldefpart:		// { x::= e1, ... x::= eN}, 
      insertTable(node, &ldefpart, &ldefpartEnd);
      break;

    case opdefsig:		// x:= E
      insertTable(node, &ldefsig, &ldefsigEnd);
      break;

    case opret:		// delay
      insertTable(node, &ldelay, &ldelayEnd);
      break;

    case opappe:		// procedure call
      if (ptINode_satisfies_readNode(node))
	{
	  if (!is_outputNodeCall(node, &n1))
	    b = pk_true;
	  else
	    b = !isNodeCall(n1);
	  if (b)
	    insertTable(node, &lread, &lreadEnd);
	}
      else
	{
	  if (ptINode_satisfies_procCall(node))
	    insertTable(node, &listrefextern, &listrefexternEnd);
	  else if (ptINode_satisfies_writeNode(node))
	    insertTable(node, &lwrite, &lwriteEnd);
	}
      break;

    case opmoti:		// Array
      insertTable(node, &lArray, &lArrayEnd);
      break;

    case opnodecall:		// node call
      insertTable(node, &lNodeCall, &lNodeCallEnd);
      break;

    case opsubgraph:		// sub-graph
      insertTable(node, &lSubGraph, &lSubGraphEnd);
      break;

    case opbund:		// Set of the nodes of a struct/bundle (lfields)
      insertTable(node, &lBundStruct, &lBundStructEnd);
      break;

    }
}

/** It inserts a node (node) 
    - into the graph at the operand clock (vh).
    - in the internal sets ( insertinSet() method) */
void NodeManager::insertNode(ptINode node, ptinsth vh)
{
  ptexpcal le, ls, resret, sigret, vret, vfen, lp, vinit;
  ptlinstances lent;
  ptident nom;
  boolean b, b1;
  ptINode n1;

  insertinSet(node);
  b1 = pk_false;
  switch (get_oper(node))
    {
    case opnil:		// undefined
      set_nodeClock(node, vh);
      break;

    case opdefpart:		// x ::= E
      break;

    case opldefpart:		// { x::= e1, ... x::= eN}, 
      set_nodeClock(node, vh);
      break;

    case opdefsig:		// x:= E
      end_equation(node);
      graph_add_defsig(node, vh);
      b1 = pk_true;
      break;

    case opret:		// delay
      get_delay_arg(node, &resret, &sigret, &vret, &vfen, &vinit);
      lent = ptlinstances_sortedfusion(EM->getSignalsReferencedIn(vret), EM->getSignalsReferencedIn(vfen));
      if (vret == NILEXPCAL)
	lent = ptlinstances_sortedfusion(EM->getPredecessors(sigret), lent);
      putinstr(node, vh, lent);
      EM->changeUndefinedName(resret, sigret, opret);
      if (vinit != NILEXPCAL)
	EM->setValuesEq(resret, sigret, opret);
      b1 = pk_true;
      break;

    case opappe:		// procedure call
      if (ptINode_satisfies_readNode(node))
	putinstr(node, vh, NILLINSTANCES);
      else
	{
	  get_procCall_arg(node, &nom, &lp, &le, &ls);
	  if (is_intrinsicfunc(node) && (isStatusId(nom) || isPresentId(nom)) || isReceiveId(nom))
	    le = NILEXPCAL;
	  addmultinstruction(ls, node, vh, le);
	  b1 = pk_true;
	}
      break;

    case opmoti:		// Array
      get_array_io(node, &le, &ls);
      addmultinstruction(ls, node, vh, le);
      break;

    case opnodecall:		// node call
      addNodeCall(node, vh);
      b1 = pk_true;
      break;

    case opsubgraph:		// sub-graph
      addSubGraph(node, vh);
      break;

    case opbund:		// Set of the nodes of a struct/bundle (lfields)
      putinstr(node, vh, NILLINSTANCES);
      break;

    }
  if (b1)
    addDependencesForOutputs(node, vh);
}

/** It moves a node (node) from its current clock to a new one (h).
 */
void NodeManager::moveNodeToClock(ptINode node, ptinsth h)
{
  ptinsth hn;

  hn = CKM->getRepresentativeFormula(get_nodeClock(node));
  h = CKM->getRepresentativeFormula(h);
  if (hn != h)
    {
      removeDependence(hn, node);
      removegfd(node);
      set_nodeClock(node, h);
      insertgfd(node, h);
      addDependence(h, node);
    }
}

/**
   It removes a node (node) from the graph assigned to its clock. 
*/
void NodeManager::removegfd(ptINode node)
{
  ptgfd gfd;
  ptlINode ln;
  ptINode n;

  gfd = CKM->getClockSubGraph(get_nodeClock(node));
  ptgfd_remove_node(gfd, node);
  if (is_setOfPartialDefinitions(node))
    {
      ln = get_setOfPartialDefinitions_arg(node);
      while (ptlINode_get_elem(&ln, &n))
	removegfd(n);
    }

}

/** It resets a node (node) from the node manager (dependences to the node are deleted). */
void NodeManager::resetNodeForModification(ptINode node)
{
  removeDependenceTo(node);
  removegfd(node);
  ptINode_set_interface(node, NILINTERFACE);
  removeInSet(node);
  removeInArraysSubGraphs(node);
}

/** It removes a node (node) from the node manager (dependences to and from the node are deleted). */
void NodeManager::removeNode(ptINode node)
{
  resetNodeForModification(node);
  removeDependenceFrom(node);
}

/**
   It removes the write node of a signal (vsig).
   For a signal x := E and x an output of a graph, an internal write-node is added into the graph.
   This method removes this internal node.
*/
void NodeManager::remove_node_output(ptexpcal vsig)
{
  ptlINode lnodes;
  ptINode aNode;
  boolean found;

  if (getOutputs(&lnodes))
    {
      found = pk_false;
      while (!found && ptlINode_get_elem(&lnodes, &aNode))
	found = (get_io_arg(aNode) == vsig);
      if (found)
	removeNode(aNode);
    }
}

/** Each node of a sub-graph node or an array node (node) becomes invisible from its clock.
    These nodes are accessible from the sub-graph/array node.
    For an array node, the operation is applied for the body-graph and the init-graph of the array-node.
*/
void NodeManager::unlinkNodesFromClock(ptINode node)
{
  ptlinsth lh;
  ptlINode ln;

  if (get_subGraphDefinitions(node, &ln, &lh))	// SubGraph
    unlinkNodesToClock(ln);
  else if (get_array_bodyDefinitions(node, &ln, &lh))	// array
    {
      unlinkNodesToClock(ln);
      if (get_array_initDefinitions(node, &ln, &lh))
	unlinkNodesToClock(ln);
    }
}

/** 
    Each node of a sub-graph node or an array node (node) becomes visible from its clock.
 */
void NodeManager::linkNodesToClock(ptINode node)
{
  ptlinsth lh;
  ptlINode ln;

  if (get_subGraphDefinitions(node, &ln, &lh))	// SubGraph
    linkNodesToClock(ln);
  else if (get_array_bodyDefinitions(node, &ln, &lh))	// array
    {
      linkNodesToClock(ln);
      if (get_array_initDefinitions(node, &ln, &lh))
	linkNodesToClock(ln);
    }
}

/**
   @return teh created node.
   It creates and inserts an intrinsic unsafe procedure call denoting "louts:= name{lparam}(lins)" at the clock vh in the graph.
*/
ptINode NodeManager::mkIntrinsicUnsafeAt(ptident id, ptexpcal lparam, ptexpcal lins, ptexpcal louts, ptinsth vh, BackAnnotationType ba)
{
  ptINode n;

  n = mk_procCall(id, lparam, lins, louts, ba);
  set_intrinsicfunc(n);
  set_safetyProperty(n, opunsafe);
  insertNode(n, vh);
  return n;
}

/**
   It unifies the nodes assigned to a list of signals (listOfSignals) into only one node (n).
*/
void NodeManager::nodesMerging(ptlinstances ls, ptINode n)
{
  ptexpcal s;
  ptINode n1;
  ptlinstances ls1;
  boolean b;

  ls1 = ls;
  b = pk_true;
  while (ptlinstances_get_next(&ls, &s))
    {
      n1 = EM->getNode(s);
      if (n1 == n)
	b = pk_false;
      else
	nodesMerging(n1);
    }
  if (b)
    unifyNodesForSeveralOutputs(ls1, n);
}

/**
   sub-Graph node.
   The internal nodes and the clocks of a sub-graph (SubGraph) are linked with this sub-graph node.
   After this operation the \ref get_inComposedNode (n, &x) will return "true" and SubGraph in x, for each nodes n of the sub-graph node.
   When the parameter brec is "true" the operation is recursively applied to the internal sub-graphs of SubGraph.
   \see {get_inComposedNode}
*/
void NodeManager::subGraphLinkNodes(ptINode SubGraph, boolean brec)
{
  subGraphLinkNodes(SubGraph, SubGraph, brec);
}

/**
   sub-Graph node.
   The internal nodes and the clocks of a sub-graph (SubGraph) are unlinked with this sub-graph node.
   After this operation the get_inComposedNode(n, ) will return "false", for each nodes n of the subgraph node.
   When brec is "true" the operation is recursively applied to the internal sub-graphs of SubGraph.
   \see {get_inComposedNode}
*/
void NodeManager::subGraphUnlinkNodes(ptINode SubGraph, boolean brec)
{
  // subGraphLinkNodes(SubGraph, NILINODE, brec);
}

/**
   Each node of ln is inserted in the sub-graph of its clock.
*/
void NodeManager::linkNodesToClock(ptlINode ln)
{
  ptINode vblock, aNode;

  while (ptlINode_get_elem(&ln, &aNode))
    {
      insertgfd(aNode, get_nodeClock(aNode));
      if (get_inComposedNode(aNode, &vblock))
	{
	  if (!get_inComposedNode(vblock, &vblock))
	    vblock = NILINODE;
	  set_inComposedNode(aNode, vblock);
	}

      if (is_setOfPartialDefinitions(aNode))
	linkNodesToClock(get_setOfPartialDefinitions_arg(aNode));
    }
}

/**
   Sub-Graph node.
   It replaces a sub-graph (aSubGraph) by its nodes in the graph. The dependences are propagated to/from the internal
   nodes of the sub-graph. The sub-graph (aSubGraph) is removed from the node manager.
*/
boolean NodeManager::subGraphFlattenning(ptINode aSubGraph)
{
  ptINode InternalInputNode, aNode, vfrom, vto;
  ptlINode ln, l;
  ptlinsth lh;
  ptldependence lsucc, lpred, lp;
  ptdependence aDep;
  ptinsth h1;
  boolean bcdelay;
  ptpragma vp;

  if (!ptINode_satisfies_subGraph(aSubGraph))
    return pk_false;
  // removing Internal subGraph dependences.
  get_subGraphInputNode(aSubGraph, &InternalInputNode);
  removeDependenceFrom(InternalInputNode);

  lsucc = get_successors(aSubGraph);
  // lhsucc = get_clockSuccessors(aSubGraph); A COMPLETER
  lpred = get_predecessors(aSubGraph);
  get_subGraphDefinitions(aSubGraph, &ln, &lh);
  // get_subGraphProjectedSystemInfos(aSubGraph, &vhemb, &lh1);

  bcdelay = getPragmaKeyword(getPragmas(aSubGraph), PRAGMA_DELAYCLUSTER, &vp);

  // Dependence Propagating and delay pragmas
  l = ln;
  while (ptlINode_get_elem(&l, &aNode))
    {
      if (bcdelay)
	addPragma(aNode, vp);
      addDependenceFromTo(aNode, lsucc);
    }

  lp = lpred;
  while (lp->get_next(&aDep, &lp))
    {
      h1 = aDep->getDependence(&vfrom, &vto);
      l = ln;
      while (ptlINode_get_elem(&l, &vto))
	addDependence(vfrom, vto, h1 /* hundef */ );
    }

  linkNodesToClock(ln);		// Nodes of subgraph become visible.
  removeNode(aSubGraph);
  return pk_true;
}

/**
   Array node.
   The internal nodes and the clocks of an array-node (arrayNode) are linked with this array node. The method 
   is applied recursively if brec.
   After this operation the get_inComposedNode(n, &x) will return "true" and arrayNode in x, for each nodes n of the array-node.
   This operation is applied for the body graph and the initial graph of an array node.
   \see {get_inComposedNode}
*/
void NodeManager::arrayLinkNodes(ptINode arrayNode, boolean brec)
{
  ptlINode ln;
  ptlinsth lh;
  ptINode aNode, n;
  ptinsth vh;

  // A fusionner avec NodeManager::subGraphLinkNodes
  if (get_array_bodyDefinitions(arrayNode, &ln, &lh))
    {
      while (ptlINode_get_elem(&ln, &aNode))
	{
	  set_inComposedNode(aNode, arrayNode);
	  if (brec && is_array(aNode))
	    arrayLinkNodes(aNode, brec);
	}
      while (ptlinsth_get_next(&lh, &vh))
	CKM->setInComposedNode(vh, arrayNode);
      if (get_inComposedNode(arrayNode, &n))
	if (get_array_initDefinitions(arrayNode, &ln, &lh))
	  while (ptlINode_get_elem(&ln, &aNode))
	    set_inComposedNode(aNode, n);
    }
}

void NodeManager::arrayUnlinkNodes(ptINode arrayNode, boolean brec)
{
  ptlINode ln;
  ptlinsth lh;
  ptINode aNode, n;
  ptinsth vh;

  if (get_array_bodyDefinitions(arrayNode, &ln, &lh))
    {
      while (ptlINode_get_elem(&ln, &aNode))
	{
	  set_inComposedNode(aNode, NILINODE);
	  if (brec && is_array(aNode))
	    arrayUnlinkNodes(aNode, brec);
	}
      while (ptlinsth_get_next(&lh, &vh))
	CKM->setInComposedNode(vh, NILINODE);
      if (get_inComposedNode(arrayNode, &n))
	if (get_array_initDefinitions(arrayNode, &ln, &lh))
	  while (ptlINode_get_elem(&ln, &aNode))
	    set_inComposedNode(aNode, n);
    }
}

/**
   Array node.
   @return the depth of the array block for a node (node).
*/
int NodeManager::arrayGetLevelBlock(ptINode node)
{
  int level;

  if (node == NILINODE)
    return -1;
  level = 0;
  while (node != NILINODE)
    // if (get_inComposedNode(node, &node)) if (is_array(node))
    if (getInArrayNode(node, &node))
      level = level + 1;
  return level;
}

/**
   Array node.
   @return true if the ouptut parameter nodeArrayIndex is defined (i.e in a array block), false otherwise.
   @param nodeArrayIndex (out)  the node of the inner index of the array blocks for a set of nodes
   (nodes of the signals of lsig list, and nodes of the indexes of the lindexes list)
   @param lsig (in) list of signals 
   @param lindexes (in) list of indexes
*/
boolean NodeManager::arrayGetMaxLevelIndex(ptlinstances lsig, ptlinstances lindexes, ptINode * nodeArrayIndex)
{
  ptexpcal vindcour, aSig;
  int numblock, numcour;
  ptINode node, cn;

  numblock = 0;
  while (ptlinstances_get_next(&lindexes, &vindcour))
    {
      node = EM->getNode(vindcour);
      numcour = arrayGetLevelBlock(node);
      if (numcour > numblock)
	{
	  *nodeArrayIndex = node;
	  numblock = numcour;
	}
    }
  while (ptlinstances_get_next(&lsig, &aSig))
    {
      node = EM->getNode(aSig);
      numcour = arrayGetLevelBlock(node);
      if (numcour > numblock)
	if (getInArrayNode(node, &cn))
	  {
	    *nodeArrayIndex = get_indexNode(cn);
	    numblock = numcour;
	  }
    }
  return (numblock != 0);
}

/**
   Sub-Graph node.
   A node (aNode) is inserted into a sub-graph (sg).
*/
boolean NodeManager::subGraphInsertNode(ptINode sg, ptINode aNode)
{
  ptgfd vgfd;

  if (!ptINode_satisfies_subGraph(sg))
    return pk_false;
  vgfd = ptINode_get_subGraph(sg);
  ptgfd_insert_node(vgfd, aNode);
  return pk_true;
}

/**
   Sub-Graph node.
   A node (aNode) is removed from a sub-graph node (node).
*/
boolean NodeManager::subGraphRemoveNode(ptINode node, ptINode aNode)
{
  ptgfd vgfd;

  if (!ptINode_satisfies_subGraph(node))
    return pk_false;
  vgfd = ptINode_get_subGraph(node);
  return ptgfd_remove_node(vgfd, aNode);
}

/**
   Sub-Graph node.
   A clock (vh) is inserted into a sub-graph node (node).
*/
boolean NodeManager::subGraphInsertClock(ptINode node, ptinsth vh)
{
  ptgfd vgfd;

  if (!ptINode_satisfies_subGraph(node))
    return pk_false;
  vgfd = ptINode_get_subGraph(node);
  ptgfd_insert_clock(vgfd, vh);
  return pk_true;
}

/**
   Sub-graph node.
   A clock (vh) is removed from a sub-graph node (node).
*/
boolean NodeManager::subGraphRemoveClock(ptINode node, ptinsth vh)
{
  ptgfd vgfd;

  if (!ptINode_satisfies_subGraph(node))
    return pk_false;
  vgfd = ptINode_get_subGraph(node);
  ptgfd_remove_clock(vgfd, vh);
  return pk_true;
}

/**
   array node.
   The node (aNode) is inserted into the body sub-graph of an array (node).
*/
void NodeManager::arrayBodyInsertNode(ptINode node, ptINode aNode)
{
  ptexpcal lsig, s;
  ptgfd vgfd;

  if (ptINode_satisfies_array(node))
    {
      vgfd = ptINode_get_bodySubGraph(node);
      ptgfd_insert_node(vgfd, aNode);
      if (!is_partialDefinition(aNode))
	{
	  lsig = get_leftOperand(aNode);
	  if (lsig != NILEXPCAL)
	    if (!EM->satisfiesList(lsig))
	      EM->setNodeAccessToArrayElement(lsig, aNode);
	    else
	      while (EM->getNaryArgs(&lsig, &s))
		EM->setNodeAccessToArrayElement(s, aNode);
	}
    }
}

/**
   array node.
   The clock (vh) is inserted into the body sub-graph of an array (node).
*/
void NodeManager::arrayBodyInsertClock(ptINode node, ptinsth vh)
{
  ptgfd vgfd;

  if (ptINode_satisfies_array(node))
    {
      vgfd = ptINode_get_bodySubGraph(node);
      ptgfd_insert_clock(vgfd, vh);
    }
}

/**
   array node.
   The clock (vh) is removed from the body sub-graph of an array (node).
*/
void NodeManager::arrayBodyRemoveClock(ptINode node, ptinsth vh)
{
  ptgfd vgfd;

  if (ptINode_satisfies_array(node))
    {
      vgfd = ptINode_get_bodySubGraph(node);
      ptgfd_remove_clock(vgfd, vh);
    }
}

/**
   array node.
   The node (aNode) is inserted into the initial sub-graph of an array (node).
*/
void NodeManager::arrayInitInsertNode(ptINode node, ptINode aNode)
{
  ptgfd vgfd;

  if (ptINode_satisfies_array(node))
    {
      vgfd = ptINode_get_initSubGraph(node);
      ptgfd_insert_node(vgfd, aNode);
      addMark(aNode, DontOptimize);
      addMark(aNode, InitArrayMark);
    }
}

/**
   array node.
   The node (aNode) is removed from the initial sub-graph of an array (node).
*/
void NodeManager::arrayInitRemoveNode(ptINode node, ptINode aNode)
{
  ptgfd vgfd;

  if (ptINode_satisfies_array(node))
    {
      vgfd = ptINode_get_initSubGraph(node);
      ptgfd_remove_node(vgfd, aNode);
      removeMark(aNode, DontOptimize);
      removeMark(aNode, InitArrayMark);
    }
}

/**
   array node.
   A clock (vclk) is inserted into the initial sub-graph of an array (node).
*/
void NodeManager::arrayInitInsertClock(ptINode node, ptinsth vclk)
{
  ptgfd vgfd;

  if (ptINode_satisfies_array(node))
    {
      vgfd = ptINode_get_initSubGraph(node);
      ptgfd_insert_clock(vgfd, vclk);
    }
}

/**
   array node.
   A clock (vclk) is removed from the initial sub-graph of an array (node).
*/
void NodeManager::arrayInitRemoveClock(ptINode node, ptinsth vclk)
{
  ptgfd vgfd;

  if (ptINode_satisfies_array(node))
    {
      vgfd = ptINode_get_initSubGraph(node);
      ptgfd_remove_clock(vgfd, vclk);
    }
}

/**
   array node.
   The node (n) is removed from the body sub-graph of an array (node).
*/
boolean NodeManager::arrayBodyRemoveNode(ptINode node, ptINode n)
{
  ptgfd vgfd;

  if (!ptINode_satisfies_array(node))
    return pk_false;
  vgfd = ptINode_get_bodySubGraph(node);
  return ptgfd_remove_node(vgfd, n);
}

/**
   array node.
   The nodes of the list (ln) and the clocks of the list (lh) are assigned to the 
   body sub-graph of an array (n).
*/
boolean NodeManager::arrayBodySetDef(ptINode n, ptlINode ln, ptlinsth lh)
{
  ptgfd vgfd;

  if (!ptINode_satisfies_array(n))
    return pk_false;
  vgfd = ptINode_get_bodySubGraph(n);
  ptgfd_set_nodes(vgfd, ln);
  ptgfd_set_clocks(vgfd, lh);
  return pk_true;
}

/**
   array node.
   The nodes of the list (ln) and the clocks of the list (lh) are assigned to the 
   initial sub-graph of an array (n).
*/
boolean NodeManager::arrayInitSetDef(ptINode n, ptlINode ln, ptlinsth lh)
{
  ptgfd vgfd;

  if (!ptINode_satisfies_array(n))
    return pk_false;
  vgfd = ptINode_get_initSubGraph(n);
  ptgfd_set_nodes(vgfd, ln);
  ptgfd_set_clocks(vgfd, lh);
  return pk_true;
}

/**
   Partial definition.
   A partial definition (node) is removed from the set of definitions of a signal (vsig).
   @return "true" is the set of partial definition becomes empty, "false" otherwise.
*/
boolean NodeManager::removeDefPart(ptexpcal vsig, ptINode node)
{
  ptINode n;
  ptlINode ln;

  if (!is_partialDefinition(node))
    return pk_false;
  n = EM->getNode(vsig);	// n is a opldefpart
  ln = get_setOfPartialDefinitions_arg(n);
  ptlINode_remove_elem(&ln, node);
  ptINode_set_SetPartDefinition(n, ln);
  return (ln == NILLINODE);
}

/** @return a one-delayed signal of vsig.
    It creates a new signal (zx) defined as the 1-delayed of vsig. The node is inserted in the 
    graph.
*/
ptexpcal NodeManager::mkDelayOf(ptexpcal vsig)
{
  ptinsth h1;
  ptexpcal zx, vinit, index;
  ptINode n;

  h1 = EM->getClock(vsig);
  zx = EM->mkIdentExp(NILIDENT, EM->getClassIdentifier(vsig), EM->getType(vsig), mk_node_undefined(), h1, NILBACKANNOTATION);
  EM->setClock(zx, h1);

  vinit = EM->getInitialValue(vsig);
  if (vinit == NILEXPCAL)
    EM->mkDefaultInit(vsig, 1, &index, &vinit);
  n = mk_node_delay(zx, vsig, EM->mkIntegerConstant(1), NILEXPCAL, vinit, NILBACKANNOTATION);
  insertNode(n, h1);
  return zx;
}

/**
   The subgraphs are removed from the graph and their dependencies are propagated.
*/
void NodeManager::removingSubGraphs()
{
  ptlINode lSubGraphs_save, lSubGraphs, ln;
  ptINode InternalInputNode, aNode, aSubGraph;
  ptlinsth lh;

  // It would be best to have a pragma to specify the subgraphs that are not to expand.
  if (getSubGraphs(&lSubGraphs))
    {
      lSubGraphs_save = lSubGraphs;
      while (ptlINode_get_elem(&lSubGraphs, &aSubGraph))
	subGraphLinkNodes(aSubGraph, pk_false);
      lSubGraphs = lSubGraphs_save;
      while (ptlINode_get_elem(&lSubGraphs, &aSubGraph))
	{
	  get_subGraphDefinitions(aSubGraph, &ln, &lh);
	  get_subGraphInputNode(aSubGraph, &InternalInputNode);
	  removeDependenceFrom(InternalInputNode);
	  removeDependenceTo(InternalInputNode);

	  while (ptlINode_get_elem(&ln, &aNode))
	    dependencesMove(aSubGraph, aNode);
	  removeBlockSubGraph(aSubGraph);

	  removeDependenceFrom(aSubGraph);
	  removeDependenceTo(aSubGraph);

	  removeNode(aSubGraph);
	}
    }
}

/**
   The implicit conversions defined in the Signal Manual reference are explicited.
*/
void NodeManager::explicitConversions()
{
  ptlINode lNodes;

  if (getEquations(&lNodes))
    explicitConversions(lNodes);
  if (getDelays(&lNodes))
    explicitConversions(lNodes);
  if (getGroupsOfPartialEquations(&lNodes))
    explicitConversions(lNodes);
}

// --------------------------------

/**
   @return the undefined clock if vexpr is "vDefinedSig $1", the contribution clock of vexpr otherwise.
*/
boolean NodeManager::assignmentClockExpr(ptexpcal vDefinedSig, ptexpcal vexpr, ptexpcal * vinit, ptinsth * vhass)
{
  ptexpcal vcurinit;
  ptinsth h1, h2;

  // cell
  if (EM->IsCellExpr(vexpr))
    {
      *vhass = EM->getContributionClock(EM->getWhenOrCellArg(vexpr));
      return pk_true;
    }
  // default
  if (EM->getOperator(vexpr) == opmele)
    {
      if (!assignmentClockExpr(vDefinedSig, EM->getBinaryArgstAt(vexpr, 1), vinit, &h1))
	return pk_false;
      if (!assignmentClockExpr(vDefinedSig, EM->getBinaryArgstAt(vexpr, 2), vinit, &h2))
	return pk_false;
      *vhass = CKM->makeSup(h1, h2);
      return pk_true;
    }
  // E when hc
  *vhass = EM->getContributionClock(vexpr);
  vexpr = EM->getBasicExpr(vexpr);
  if (EM->isDelayOneOf(vexpr, vDefinedSig, &vcurinit))
    {
      if (*vinit == NILEXPCAL)
	(*vinit) = vcurinit;
      else if (!EM->IsCompatibleWith(*vinit, vcurinit))
	return pk_false;
      *vhass = hundef;
      return pk_true;
    }
  if (EM->IsSignal (vexpr))
    {
      *vhass = CKM->makeInf( EM->getSetAssignmentClock(vexpr), *vhass );
    }
  return pk_true;
}

/**
   @return "true" if the assignment clock for the operand node (node) is defined in vhAssignment, "false" otherwise.
   vhAssignment is equal to the undefined clock if the node is x ::= x$1, or  x ::= defaultvalue x$1.
   The assigment clock is assigned to the signal defined in node.
   A node is assumed to be an equation (not a delay).
   \See :setAssignmentClock(), getAssignmentClock().
*/
boolean NodeManager::assignmentClockEquation(ptINode node, ptexpcal * vinit, ptinsth * vhAssignment)
{
  ptexpcal vDefinedSig, vexpdedef, vdedef;

  if (!get_rightOperand(node, &vdedef)) return pk_false;
  vDefinedSig = get_leftOperand(node);
  // vexpdedef = EM->getBasicExpr(vdedef);
  return assignmentClockExpr(vDefinedSig, vdedef /*vexpdedef*/, vinit, vhAssignment);
}

ptinsth NodeManager::assignmentClockPartialEquation(ptINode node)
{
  ptlINode ln;
  ptINode n;
  ptinsth vhAssignment, vh;
  ptexpcal vinit, vDefinedSig;
  boolean b;

  vDefinedSig = get_leftOperand(node);
  ln = get_setOfPartialDefinitions_arg(node);
  vhAssignment = hundef;
  b = pk_true;
  vinit = NILEXPCAL;
  while (b && ptlINode_get_elem(&ln, &n))
    if (assignmentClockEquation(n, &vinit, &vh))
      vhAssignment = CKM->makeSup(vhAssignment, vh);
    else
      b = pk_false;
  if (b)
    {
      EM->setInitialValue(vDefinedSig, vinit);
      EM->setAssignmentClock(vDefinedSig, vhAssignment);
      return vhAssignment;
    }
  return get_nodeClock(node);
}

/**
   It computes the assignment clock of an equation ( node ).
   A node is assumed to be an equation (not a delay).
*/
ptinsth NodeManager::assignmentClockEquation(ptINode node)
{
  ptinsth vhAssignment;
  ptexpcal vDefinedSig, vinit;

  vinit = NILEXPCAL;
  if (assignmentClockEquation(node, &vinit, &vhAssignment))
    {
      vDefinedSig = get_leftOperand(node);
      EM->setInitialValue(vDefinedSig, vinit);
      EM->setAssignmentClock(vDefinedSig, vhAssignment);
      return vhAssignment;
    }
  return get_nodeClock(node);
}



/** It compute the assignment clock of a node (node). */    
ptinsth NodeManager::ComputeAssignmentClock(ptINode node)
{
  ptexpcal le, ls, resret, sigret, vret, vfen, lp, vinit;
  ptlinstances lent;
  ptident nom;
  boolean b, b1;
  ptINode n1;

  switch (get_oper(node))
    {
    case opnil:		// undefined
      return get_nodeClock(node); // for an event.

    case opdefpart:		// x ::= E, -- visited by opldefpart
      return get_nodeClock(node);

    case opldefpart:		// { x::= e1, ... x::= eN}, 
      return assignmentClockPartialEquation(node);

    case opdefsig:		// x:= E
      return assignmentClockEquation(node);

    case opret: // delay
      return get_nodeClock(node);  // to be redefined

    case opappe:		// procedure call
      return get_nodeClock(node);

    case opmoti:		// Array
     return get_nodeClock(node);

    case opnodecall:		// node call
      return get_nodeClock(node);

    case opsubgraph:		// sub-graph
      return get_nodeClock(node);

    case opbund:		// Set of the nodes of a struct/bundle (lfields)
      return get_nodeClock(node);
    }
  return hundef;
}
