/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#include <iostream>
#include <fstream>
using namespace std;

_PRIVATE_
/**
   It resets the dependences reaching a node (node) and the node belongs an equation node.
   The status is not modified for a partial definition.
*/
    void
 NodeManager::resetAsEquation(ptINode node)
{
  // removeInSetCase(node);
  removeDependenceTo(node);
  ptINode_resetAsEquation(node);
}

/**
   @return the node of the defined signal (vexpadef).
   It associates to a signal (vexpadef) the expression (vexpdedef when H) with
   - H=vhcal if vexpadef is not already defined
   - H is the complement of K in vhcal, where K is the clock of the expression already assigned to vexpadef.

   instent: list of the signals referenced in vexpdedef, used to define the dependences.
*/
ptINode NodeManager::addaff(ptexpcal vexpadef, ptexpcal vexpdedef, ptinsth vhcal, ptlinstances instent)
{
  ptINode aNode;
  ptexpcal vexpToadd, e;

  aNode = EM->getNode(vexpadef);
  if (EM->getOperator(vexpdedef) != opmele)
    vexpToadd = EM->mkWhenOrCellExp(opextr, vexpdedef, vhcal);
  else
    vexpToadd = vexpdedef;
  if (ptINode_get_rightOperand(aNode, &e))
    {
      ptINode_remove_luse(aNode, e);
      vexpToadd = EM->mkClockedDefault(e, vexpToadd);
    }
  becomes_assign(aNode, vexpadef, vexpToadd);
  addDependenceSigList2NodeAt(instent, aNode, vhcal);
  return aNode;
}

/**
   It creates a partial definition node in the graph by 
   - "left ::= right when vhcal",  
   - instent: list of the signals referenced in right, used to define the dependences.
   @return the created partial definition node.
*/
ptINode NodeManager::mkPartEquation(ptexpcal left, ptexpcal right, ptinsth vhcal, ptlinstances instent)
{
  ptINode n;

  n = mk_node_undefined();
  becomes_partialDefinition(n, left, right);
  addDependenceSigList2NodeAt(instent, n, vhcal);
  set_nodeClock(n, vhcal);
  insertgfd(n, vhcal);
  insertNode(n, vhcal);
  return n;
}

void NodeManager::addSynchrosFields(ptlinstances lsig1, ptlinstances lsig2)
{
  ptexpcal aSig1, aSig2;
  ptlinstances l1, l2;

  while (ptlinstances_get_next(&lsig1, &aSig1))
    {
      ptlinstances_get_next(&lsig2, &aSig2);
      CKM->makeSynchro(EM->getClock(aSig1), EM->getClock(aSig2));
      addSynchrosBundle(aSig1, aSig2);
    }
}

/**
   Adding the synchronisation for lexp := rexp, for a bundle type.
*/
void NodeManager::addSynchrosBundle(ptexpcal lexp, ptexpcal rexp)
{
  ptlinstances ls1, ls2;

  if (EM->getFieldsOf(lexp, &ls1) && EM->getFieldsOf(rexp, &ls2))
    addSynchrosFields(ls1, ls2);
}

/**
  @param aNode  (in) a node, assumed to be an equation ("opdefsig" operator).
  It simplifies the definition of the signal vexpcal1 when it is a event.
  In this case, the constraint ^vexpca1 ^= when vexpcal is added to the clock system.
  Otherwise, the expression of definition of the signal can be completed as a cell 
  using the clock of the signal and the assignment clock.
  @return none.
*/
void NodeManager::end_equation(ptINode aNode)
{
  boolean notdefined;
  ptinsth vh, vhcal;
  ptexpcal expcal1, vexp;
  int lb, ub;
  ntrees refTree;

  expcal1 = get_leftOperand(aNode);
  ptINode_get_rightOperand(aNode, &vexp);
  EM->TypeCheckingEquationFinalize(expcal1, vexp, getBackAnnotation(aNode));
  addSynchrosBundle(expcal1, EM->getBasicExpr(vexp));
  if (EM->TypeCheckingIsEvent(expcal1, &notdefined))
    if (notdefined)
      {
	vh = EM->getClock(expcal1);
	vh = CKM->makeSynchro(vh, CKM->makeSub(vh, expcal1, EM->getPredecessors(expcal1)));
      }
    else
      {
	resetAsEquation(aNode);
	addaff(expcal1, EM->getTrue(), CKM->getInitClock(), NILLINSTANCES);
	ptINode_get_rightOperand(aNode, &vexp);
      }
  vhcal = get_contributionClockEquation(aNode);
  if (!CKM->satisfiesInitClock(vhcal))
    {
      vh = EM->getClock(expcal1);
      if (vhcal != vh)		// memory 
	{
	  vexp = EM->mkWhenOrCellExp(opmemo, vexp, vh);
	  set_def(aNode, vexp);
	}
    }
  EM->setValuesEq(expcal1, vexp, opnil);
}

/**
   It inserts (putinstr()) a node (aNode), assumed to be an equation, in the graph associated with a clock (vh).
   It adds the dependences (addDependence(ptinsth, ptINode)) from the contribution clocks of the definition to the node. 
   The dependences from the signals referenced in the expressions are not added here, they have been already added in addaff().
*/
void NodeManager::graph_add_defsig(ptINode aNode, ptinsth vh)
{
  ptexpcal vhorlexpcal, vexpcal;
  ptinsth vhsig, vhcontr;
  ptlinsth vlhcontr;

  vexpcal = get_leftOperand(aNode);
  vhsig = CKM->getRepresentativeFormula(EM->getClock(vexpcal));
  vh = CKM->getRepresentativeFormula(vh);

  if ((CKM->satisfiesNullClock(vhsig)) || (CKM->satisfiesNullClock(vh)))
    resetAsEquation(aNode);
  if (ptINode_get_rightOperand(aNode, &vhorlexpcal))
    {
      vlhcontr = EM->getClocksReferencedIn(vhorlexpcal);
      while (ptlinsth_get_next(&vlhcontr, &vhcontr))
	if ((vhcontr != vhsig) && !CKM->satisfiesNullClock(vhcontr))
	  addDependence(vhcontr, aNode);
    }
  putinstr(aNode, vh, NILLINSTANCES);
}

/**
   It inserts a node (aNode) in the graph associated with a clock (vh).
   For each element X of lins, a dependence from the node of X to aNode is added.
*/
void NodeManager::putinstr(ptINode aNode, ptinsth vh, ptlinstances lins)
{
  set_nodeClock(aNode, vh);
  addDependenceSigList2NodeAt(lins, aNode, hundef);
  insertgfd(aNode, vh);
}

/**
   It inserts a node (aNode) in the graph associated with a clock (vh). This node defines a set of signals (louts).
   For each element X of lins, a dependence from the node of X to aNode is added.
*/
void NodeManager::addmultinstruction(ptexpcal louts, ptINode aNode, ptinsth vh, ptexpcal lins)
{
  ptlinstances vlinstdef, instent;

  vlinstdef = EM->getSignalsDefinedIn(louts);
  unifyNodesForSeveralOutputs(vlinstdef, aNode);
  instent = EM->getPredecessors(lins);
  putinstr(aNode, CKM->getRepresentativeFormula(vh), instent);
}

/**
   - aNode is assumed to be a node call with vh as associated clock
   - lout is the signals defined in aNode

   For each element of louts, it adds a node (read node) in the graph with vh as clock.
   The set of created nodes is assigned to the node call.
*/
void NodeManager::addInputNodesForOutputs(ptexpcal lout, ptINode aNode, ptinsth vh)
{
  ptINode inNode;
  ptlINode ln;

  // horloge de la sortie = horloge du node call ?
  ln = NILLINODE;
  while (lout != NILEXPCAL)
    {
      inNode = mk_node_input(EM->getNaryArgs(&lout), aNode, getBackAnnotation(aNode));
      ptlINode_insert_elem(&ln, inNode);
      putinstr(inNode, vh, NILLINSTANCES);
    }
  ptINode_set_nodecallProducedNodes(aNode, ln);
}

/**
   It inserts in the graph a node (aNode) assumed to be a node call with vh as associated clock.
*/
void NodeManager::addNodeCall(ptINode aNode, ptinsth vh)
{
  ptexpcal lin, lout, lparam;
  ptDCG vg;

  get_nodeCall_arg(aNode, &vg, &lparam, &lin, &lout);
  addInputNodesForOutputs(lout, aNode, vh);
  putinstr(aNode, vh, NILLINSTANCES);
}

/**
   It inserts in the graph a node (aSubGraph) assumed to be a node associtaed with a sub-graph with vh as associated clock.
*/
void NodeManager::addSubGraph(ptINode aSubGraph, ptinsth vh)
{
  putinstr(aSubGraph, vh, NILLINSTANCES);
}

/**
   For a sub-graph (sg),
    - it returns ln the list of its nodes and clocks the list of the its clocks 
    - otherwise ln is composed of sg, and clocks the empty list.
*/
void NodeManager::getNodesClocksDefinedIn(ptINode sg, ptlINode * ln, ptlinsth * clocks)
{
  if (ptINode_satisfies_subGraph(sg))
    get_subGraphDefinitions(sg, ln, clocks);
  else
    {
      *ln = NILLINODE;
      ptlINode_insert_elem(ln, sg);
      *clocks = NILLINSTH;
    }
}

/**
   It inserts a node (n) in the sub-graph associated with a clock (h).
*/
void NodeManager::insertgfd(ptINode n, ptinsth h)
{
  ptgfd_insert_node(CKM->getClockSubGraph(h), n);
}

/**
   It removes a node (aNode) from the internal table of the node manager.
*/
void NodeManager::removeInSet(ptINode aNode)
{
  removeTable(aNode, &AllNodes, &AllNodesEnd);
  insertTable(aNode, &deletedNodes, &deletedNodesEnd);
  removeInSetCase(aNode);
}

/**
   For a "bundle/struct" node (lfields) node (n).
   It merges the nodes of the fields in n.
   It is used  when a structure is globaly assigned.
*/
void NodeManager::nodesMerging(ptINode n)
{
  ptlinstances ls;

  if (!is_lfields(n))
    return;
  ls = get_lfields_arg(n);
  removeInSet(n);		//supression de l'ensemble "bundle"
  nodesMerging(ls, n);
}

/**
   A node (n) of a sub-graph is removed from  its sub-graph.
   (It searches in all the sub-graphs).
*/
void NodeManager::removeInArraysSubGraphs(ptINode n)
{
  ptINode aSubGraph;
  ptlINode ln;
  boolean bstop;

  bstop = pk_false;
  // A node may be only in a sub-graph/array.
  if (getSubGraphs(&ln))
    {
      bstop = pk_false;
      while (!bstop && ptlINode_get_elem(&ln, &aSubGraph))
	bstop = subGraphRemoveNode(aSubGraph, n);
    }
  if (!bstop)
    if (getIterates(&ln))
      while (!bstop && ptlINode_get_elem(&ln, &aSubGraph))
	bstop = arrayBodyRemoveNode(aSubGraph, n);
}

/**
   Each node of ln is removed from the sub-graph of its clock.
*/
void NodeManager::unlinkNodesToClock(ptlINode ln)
{
  ptINode aNode;

  while (ptlINode_get_elem(&ln, &aNode))
    {
      ptgfd_remove_node(CKM->getClockSubGraph(get_nodeClock(aNode)), aNode);
      if (is_setOfPartialDefinitions(aNode))
	unlinkNodesToClock(get_setOfPartialDefinitions_arg(aNode));
    }
}

/**
   It inserts a node (node) in a table characterized by its firt element (lbegin) and its last one (lend).
*/
void NodeManager::insertTable(ptINode node, ptlINode * lbegin, ptlINode * lend)
{
  int n, m;
  ptINode x;

  if ((*lend) != NILLINODE)
    {
      x = ptlINode_getElem(*lend);
      n = get_num(node);
      m = get_num(x);
      if (n != m)
	if (n > m)
	  {
	    ptlINode_insert_elem(lend, node);
	    *lend = ptlINode_getNext(*lend);
	  }
	else
	  ptlINode_insert_elem(lbegin, node);
    }
  else
    {
      ptlINode_insert_elem(lbegin, node);
      *lend = *lbegin;
    }
}

/**
   It removes a node (node) from a table characterized by its firt element (lbegin) and its last one (lend).
*/
boolean NodeManager::removeTable(ptINode node, ptlINode * lbegin, ptlINode * lend)
{
  int n, m;
  ptINode x;
  ptlINode l;

  if ((*lend) != NILLINODE)
    {
      x = ptlINode_getElem(*lend);
      n = get_num(node);
      m = get_num(x);
      if (n == m)
	{
	  l = ptlINode_getPred(*lend);
	  *lend = l;
	  if (l == NILLINODE)
	    *lbegin = NILLINODE;
	  else
	    ptlINode_remove_elem(&l, node);
	  return pk_true;
	}
    }
  return ptlINode_remove_elem(lbegin, node);
}

/**
   It removes a node (node) from the node manager. 
   It calls removeTable() taking into account the class of the node (equation, delay,...).
*/
void NodeManager::removeInSetCase(ptINode aNode)
{
  ptexpcal vexp;

  switch (get_oper(aNode))
    {

    case opldefpart:		// A completer
      removeTable(aNode, &ldefpart, &ldefpartEnd);
      break;

    case opdefpart:
      if (ptINode_get_rightOperand(aNode, &vexp))
	ptINode_remove_luse(aNode, vexp);
      // REMOVEINSETLDEFPART() EM->belongsFreeSignal(aNode->get_leftOperand());
      break;

    case opdefsig:
      removeTable(aNode, &ldefsig, &ldefsigEnd);
      if (ptINode_get_rightOperand(aNode, &vexp))
	ptINode_remove_luse(aNode, vexp);
      EM->belongsFreeSignal(ptINode_get_leftOperand(aNode));
      break;

    case opret:
      removeTable(aNode, &ldelay, &ldelayEnd);
      if (ptINode_get_rightOperand(aNode, &vexp))
	ptINode_remove_luse(aNode, vexp);
      break;

    case opappe:
      if (!removeTable(aNode, &lread, &lreadEnd))
	if (!removeTable(aNode, &lwrite, &lwriteEnd))
	  removeTable(aNode, &listrefextern, &listrefexternEnd);
      break;

    case opmoti:
      removeTable(aNode, &lArray, &lArrayEnd);
      break;

    case opnodecall:
      removeTable(aNode, &lNodeCall, &lNodeCallEnd);
      break;

    case opsubgraph:
      removeTable(aNode, &lSubGraph, &lSubGraphEnd);
      if (get_label(aNode, &vexp))
	{
	  CKM->removeFromClock(EM->getClock(vexp), vexp, pk_false);
	  EM->remove(vexp);
	}
      break;

    case opbund:
      removeTable(aNode, &lBundStruct, &lBundStructEnd);
      break;

    case opnil:
      break;

    }
}

/**
   It unifies the nodes assigned to a list of signals (listOfSignals) into only one node (node).
*/
void NodeManager::unifyNodesForSeveralOutputs(ptlinstances listOfSignals, ptINode node)
{
  ptexpcal elem;
  ptINode nodecour, vfrom, vto;
  ptldependence vldep;
  ptlinsth vldephorl;
  ptdependence aDep;
  ptinsth h1;

  vldephorl = NILLINSTH;
  while (ptlinstances_get_next(&listOfSignals, &elem))
    {
      nodecour = EM->getNode(elem);
      if (is_undefined(nodecour))
	{
	  vldep = get_successors(nodecour);
	  while (vldep->get_next(&aDep, &vldep))
	    {
	      h1 = aDep->getDependence(&vfrom, &vto);
	      removeDependence(aDep);
	      addDependence(node, vto, h1);
	    }
	  vldep = get_predecessors(nodecour);
	  while (vldep->get_next(&aDep, &vldep))
	    {
	      h1 = aDep->getDependence(&vfrom, &vto);
	      removeDependence(aDep);
	      addDependence(vfrom, node, h1);
	    }

	  vldephorl = ptlinsth_sorted_fusion(vldephorl, get_clockSuccessors(nodecour));
	}
      EM->setNode(elem, node);
    }
  set_clocksSuccessors(node, vldephorl);
}

/** 
    Array node.
    Adding dependence form the index of an array to an internal node of the array.
    The internal signals (access to element x[i]) are linked with their definition using a "global" attribute.
*/
void NodeManager::linkInodeToInodeArray(ptINode aNode, ptINode ArrayNode)
{
  ptINode n;

  if (ArrayNode == NILINODE)
    return;
  //  indexArrayNode = get_indexNode(ArrayNode);
  get_array_inputNode(ArrayNode, &n);
  addDependence(n, aNode, hundef);
  //  aNode->set_inComposedNode( ArrayNode);
}

/**
   sub-Graph node.
   The internal nodes and the clocks of a sub-graph (SubGraph) are linked with composedNode.
   After this operation the \ref get_inComposedNode (n, &x) will return "true" and SubGraph in x, for each nodes n of the sub-graph node.
   When the parameter bflatInternalSubGraph is "true" the operation is recursively applied to the internal sub-graphs of SubGraph.
*/
void NodeManager::subGraphLinkNodes(ptINode SubGraph, ptINode composedNode, boolean bflatInternalSubGraph)
{
  ptlINode ln;
  ptlinsth lh;
  ptINode aNode;
  ptinsth vh;

  // A fusionner avec NodeManager::arrayLinkNodes
  if (get_subGraphDefinitions(SubGraph, &ln, &lh))
    {
      while (ptlINode_get_elem(&ln, &aNode))
	{
	  set_inComposedNode(aNode, composedNode);
	  if (bflatInternalSubGraph && is_subgraph(aNode))
	    subGraphLinkNodes(aNode, composedNode, bflatInternalSubGraph);
	}
      while (ptlinsth_get_next(&lh, &vh))
	CKM->setInComposedNode(vh, composedNode);
    }
}

/**
   aNode element of aSubGraph becomes a node of a composed node upBlock.
*/
void NodeManager::addInBlock(ptINode aNode, ptINode upBlock, ptINode aSubGraph)
{
  ptlINode ln;
  ptlinsth lh;

  if (upBlock != NILINODE)
    {
      set_inComposedNode(aNode, upBlock);
      if (is_subgraph(upBlock))
	subGraphInsertNode(aNode, upBlock);
      else if (is_array(upBlock))
	{
	  // in the body part or the init part.
	  get_array_bodyDefinitions(upBlock, &ln, &lh);
	  if (ptlINode_isIn(ln, aSubGraph))
	    arrayBodyInsertNode(aNode, upBlock);
	  else
	    arrayBodyInsertNode(aNode, upBlock);
	}
    }
}

/**
   The nodes of a subgraph (aSubGraph) becomes nodes of the subgraph of this subgraph.
*/
void NodeManager::removeBlockSubGraph(ptINode aSubGraph)
{
  ptlINode lnsave, ln;
  ptINode upBlock, aNode;
  ptpragma vp;
  boolean b, bcdelay;
  ptlinsth lh;

  if (is_subgraph(aSubGraph))
    {
      bcdelay = getPragmaKeyword(getPragmas(aSubGraph), PRAGMA_DELAYCLUSTER, &vp);
      get_subGraphDefinitions(aSubGraph, &ln, &lh);
      lnsave = ln;
      b = get_inComposedNode(aSubGraph, &upBlock);
      if (b || bcdelay)
	while (ptlINode_get_elem(&ln, &aNode))
	  {
	    if (bcdelay)
	      addPragma(aNode, vp);
	    if (b)
	      addInBlock(aNode, upBlock, aSubGraph);
	  }
      if (!b)
	linkNodesToClock(lnsave);	// Nodes of subgraph become visible at their clock.
    }

}

/** @return "true" if the output list (list) is not empty, "false" otherwise.
    @param list (out) list of the nodes of the graph  with code as opertor (equation, memorization, ....)
    @param code (in) a SIGNAL operator.
*/
boolean NodeManager::getNodesList(noperators code, ptlINode * list)
{
  *list = NILLINODE;
  switch (code)
    {
    case opnil:
      (*list) = NILLINODE;
      break;
    case opdefsig:
      (*list) = ldefsig;
      break;
    case opret:
      (*list) = ldelay;
      break;
    case opappe:
      (*list) = listrefextern;
      break;
    case oplread:
      (*list) = lread;
      break;
    case oplwrite:
      (*list) = lwrite;
      break;
    case opmoti:
      (*list) = lArray;
      break;
    case opnodecall:
      (*list) = lNodeCall;
      break;
    case opsubgraph:
      (*list) = lSubGraph;
      break;
    case opbund:
      (*list) = lBundStruct;
      break;
    case opldefpart:
      (*list) = ldefpart;
      break;
    }
  return ((*list) != NILLINODE);
}

/**
   The implicit conversions defined in the Signal Manual reference are explicited.
   It calls 
   - ExprManager::ExplicitConversions() followed by
   - ExprManager::EM->rewritingExprEH, then 
*/
void NodeManager::explicitConversions(ptlINode lNodes)
{
  ptINode node;
  ptexpcal nexp, exp;
  TypeType vtype;

  while (ptlINode_get_elem(&lNodes, &node))
    if (is_setOfPartialDefinitions(node))
      explicitConversions(get_setOfPartialDefinitions_arg(node));
    else
      {
	get_rightOperand(node, &exp);
	vtype = EM->getType(get_leftOperand(node));
	if (TypeTestPredef(vtype))
	  {
	    nexp = EM->explicitConversions(vtype, exp);
	    if (nexp != exp)
	      set_rightOperand(node, nexp);
	  }
      }
}
