/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
#define preChangeDef(node) removeInSetCase(node); resetAsEquation(node); ptINode_set_interface(node, NILINTERFACE)
/**
   Initializes the operand node to become a partial definition.
*/
    void
 NodeManager::preRewToPartialDef(ptINode node)
{
  resetNodeForModification(node);
  ptINode_resetAsEquation(node);
  EM->addMark(ptINode_get_leftOperand(node), PART_DEFINITION);
}

/**
   Finalizes the operand node to become a partial definition.
*/
void NodeManager::postRewToPartialDef(ptINode node)
{
  EM->removeMark(ptINode_get_leftOperand(node), PART_DEFINITION);
  setDependencesPartDefinition(node);
  insertinSet(node);
}

/**
   Defines a partial definition of vSig by WhenExpr (assumed to be a When expession, E when H)
*/
void NodeManager::addPartDefinition(ptexpcal vSig, ptexpcal WhenExpr, ptinsth vh)
{
  ptexpcal e;
  ptinsth h;

  h = CKM->makeInf(EM->getContributionClock(WhenExpr), vh);
  if (!CKM->eqClock(h, the_nullClock_NM_copy))
    {
      e = EM->getWhenOrCellArg(WhenExpr);
      addDependence(h, addPartAff(vSig, e, h, EM->getSignalsReferencedIn(e)));
    }
}

/**
   @return "true" if the operation succeeds, "false" otherwise.
   It splits an equation (node) 
   - x:= E1 default .... default En into a set of partial definitions x::=E1, ...x::=En.
   - x:=( E1 default .... default En ) cell ^x into a set of partial definitions x::=E1, ...x::=En, completed by x::= x $1.

   When vh is not the undefined clock, each partial equation is filtered by vh, (x::=E when vh).
   For example, vh may be the utility clock of the defined signal in node.

   RESTRICTION: the node is not th definition of an output signal.
*/
boolean NodeManager::DefaultSpliting(ptINode node, ptinsth vh)
{
  ptexpcal vopg, vopd, exp2, zx, aux;
  ptinsth hContribCell, hDefaultValue, hleftPart, hContrib;
  boolean bcell, bmele;

  if (!ptINode_get_rightOperand(node, &vopd))
    return pk_false;
  vopg = ptINode_get_leftOperand(node);
  if (EM->IsOutputFlow(vopg))
    return pk_false;
  bcell = EM->IsCellExpr(vopd);
  if (bcell)
    {
      vopd = EM->getWhenOrCellArg(vopd);
      aux = vopd;
      hContribCell = EM->getContributionClock(aux);
    }
  vopd = EM->getBasicExpr(vopd);
  bmele = (EM->getOperator(vopd) == opmele);
  if (!bmele)
    return pk_false;

  preRewToPartialDef(node);

  hleftPart = EM->getClock(vopg);
  while (EM->getOperator(vopd) == opmele)
    {
      EM->getDefaultArgs(vopd, &vopd, &exp2);
      addPartDefinition(vopg, exp2, vh);
    }
  addPartDefinition(vopg, vopd, vh);
  // Adding the Default value part.
  hDefaultValue = hundef;
  if (bcell)
    hDefaultValue = CKM->makeMinus(hleftPart, CKM->makeInf(vh, hContribCell));
  else if (vh != hundef)
    hDefaultValue = CKM->makeMinus(hleftPart, vh);

  if (hDefaultValue != hundef)
    {
      zx = mkDelayOf(vopg);
      addDefaultValuePart(vopg, zx, hDefaultValue, ptlinstances_insert(NILLINSTANCES, zx));
    }
  postRewToPartialDef(node);
  EM->setSharedSignal(vopg, pk_true);
  return pk_true;
}

/**
   @retun "true" if the definition of the operande node is rewritten, "false" otherwise.
   The node x := E is rewritten in x::=E when vUtilityClock | x::=defaultvalue x$1
   Assumed node is an equation (but not a delay).

   RESTRICTION: node is not the definition of an output signal.
*/
boolean NodeManager::rewEqUsingProductiveClock(ptINode node, ptinsth vUtilityClock)
{
  ptexpcal vopg, vopd, zx;
  ptinsth hx;
  ptINode DefaultValueNode;

  if (!ptINode_get_rightOperand(node, &vopd))
    return pk_false;
  vopg = ptINode_get_leftOperand(node);
  if (EM->IsOutputFlow(vopg))
    return pk_false;
  if (DefaultSpliting(node, vUtilityClock))
    return pk_true;
  // "simple" equation: x := F when hx, or  x:= (F when hContrib) cell hx
  preRewToPartialDef(node);
  hx = EM->getClock(vopg);
  if (EM->IsCellExpr(vopd))
    vopd = EM->getWhenOrCellArg(vopd);
  addPartDefinition(vopg, vopd, vUtilityClock);
  // The Default value part.
  zx = mkDelayOf(vopg);
  addDefaultValuePart(vopg, zx, CKM->makeMinus(hx, vUtilityClock), ptlinstances_insert(NILLINSTANCES, zx));
  postRewToPartialDef(node);
  EM->setSharedSignal(vopg, pk_true);
  return pk_true;
}

/**
   It reduces the clock of the assigment for a partial definition (node).
   x ::= E1 when h1,...,   x := En when hn is rewritten in x ::= E1 when h1 ^*vUtilityClock ,..., x ::= En when hn^*vUtilityClock
   and the default value is updated for a signal (not a state variable).
*/
boolean NodeManager::rewPartialDefUsingProductiveClock(ptINode node, ptinsth vUtilityClock)
{
  ptlINode ln;
  ptINode DefaultValueNode, n;
  ptexpcal vopg, vopd, zx;
  ptinsth h, hexp, hCell;

  DefaultValueNode = NILINODE;
  zx = NILEXPCAL;
  ln = get_setOfPartialDefinitions_arg(node);
  preRewToPartialDef(node);
  while (ptlINode_get_elem(&ln, &n))
    {
      vopg = ptINode_get_leftOperand(n);
      if (is_partialDefinitionDefaultValue(n))
	DefaultValueNode = n;
      else
	{
	  ptINode_get_rightOperand(n, &vopd);
	  if (EM->IsCellExpr(vopd))	// no occurrence
	    {
	      hCell = EM->getContributionClock(vopd);
	      hexp = EM->getContributionClock(EM->getWhenOrCellArg(vopd));
	      if (zx == NILEXPCAL)
		zx = mkDelayOf(vopg);
	      h = CKM->makeInf(CKM->makeMinus(hCell, hexp), vUtilityClock);
	      addPartAff(vopg, zx, h, ptlinstances_insert(NILLINSTANCES, zx));
	    }
	  vopd = EM->getWhenOrCellArg(vopd);
	  addPartDefinition(vopg, vopd, vUtilityClock);
	}
    }
  if (!EM->isStateVar(vopg))
    {
      // Default value
      if (DefaultValueNode == NILINODE)
	{
	  if (zx == NILEXPCAL)
	    zx = mkDelayOf(vopg);
	}
      else
	{
	  ptINode_get_rightOperand(DefaultValueNode, &vopd);
	  zx = EM->getWhenOrCellArg(vopd);
	}
      addDefaultValuePart(vopg, zx, CKM->makeMinus(EM->getClock(vopg), vUtilityClock), ptlinstances_insert(NILLINSTANCES, zx));
    }
  postRewToPartialDef(node);
  return pk_true;
}

/**
   It applies the rule (E when H) default (E when K) --> E when (H ^+ K) on the operand node (node) assumed to be an equation.
   @return "true" if the node is modified, "false" otherwise.
   @See normalDefaultsEquationPartialDef()
*/
boolean NodeManager::normalDefaultsEquation(ptINode node)
{
  ptexpcal vopd, nexp, exp, aux;
  boolean bcell;
  BackAnnotationType vt;
  ptlinstances lexp, save_lexp;
  ptinsth h;

  // To be completed for the cell operator.
  if (!ptINode_get_rightOperand(node, &vopd))
    return pk_false;
  bcell = EM->IsCellExpr(vopd);
  if (bcell)
    return pk_false;		// vopd =  EM->getWhenOrCellArg(vopd);

  exp = EM->getBasicExpr(vopd);
  if (EM->getOperator(exp) != opmele)
    return pk_false;
  if (!EM->normalDefaults(exp, &lexp))
    return pk_false;

  removeDependenceTo(node);
  set_def(node, NILEXPCAL);
  ptINode_set_interface(node, NILINTERFACE);

  // use EM->isDelayOneOf() to put the $ as the last operand of the default.
  save_lexp = lexp;

  if (ptlinstances_get_next(&lexp, &nexp))
    {
      vt = EM->getBackAnnotation(exp);
      h = EM->getContributionClock(nexp);
      addDependenceSigList2NodeAt(EM->getSignalsReferencedIn(EM->getWhenOrCellArg(nexp)), node, h);
      addDependence(h, node);
      while (ptlinstances_get_next(&lexp, &aux))
	{
	  nexp = EM->mkBinaryExpTp(opmele, nexp, aux, vt);
	  h = EM->getContributionClock(aux);
	  EM->setContributionClock(nexp, CKM->makeSup(EM->getContributionClock(nexp), h));
	  addDependenceSigList2NodeAt(EM->getSignalsReferencedIn(EM->getWhenOrCellArg(aux)), node, h);
	  addDependence(h, node);
	}
      h = EM->getContributionClock(vopd);
      EM->setContributionClock(nexp, h);
      addDependence(h, node);
    }
  ptlinstances_free(save_lexp);
  // if (bcell) nexp = EM->mkWhenOrCellExp(opmemo, nexp, EM->getContributionClock(vopd));
  set_def(node, nexp);
  return pk_true;
}

/**
   Assumed:  node is a partial definition node (set of nodes).
*/
boolean NodeManager::normalDefaultsEquationPartialDef(ptINode node)
{
  ptlINode ln;
  ptINode n;
  ptexpcal LPart, RPart;
  ptlinstances lexp_save, lexp;
  boolean bmodified = pk_false;
  ptINode defaultValueNode;
  ptexpcal exp;
  ptinsth h;

  defaultValueNode = NILINODE;
  lexp = NILLINSTANCES;
  // By construction, there is no node X::= E default F, X::= E cell H (to be verified). 
  ln = get_setOfPartialDefinitions_arg(node);
  while (ptlINode_get_elem(&ln, &n))
    if (!is_partialDefinitionDefaultValue(n))
      {
	ptINode_get_rightOperand(n, &RPart);
	// if (EM->IsCellExpr(vopd)) // no occurrence (?)
	if (!EM->addExpToDefaultList(&lexp, RPart))
	  bmodified = pk_true;
      }
    else
      defaultValueNode = n;

  lexp_save = lexp;
  if (bmodified)
    {
      preRewToPartialDef(node);
      LPart = ptINode_get_leftOperand(node);
      while (ptlinstances_get_next(&lexp, &exp))
	addPartDefinition(LPart, exp, hundef);
      if (defaultValueNode != NILINODE)
	{
	  get_rightOperand(defaultValueNode, &exp);
	  h = EM->getContributionClock(exp);
	  exp = EM->getBasicExpr(exp);
	  addDefaultValuePart(LPart, exp, h, EM->getSignalsReferencedIn(exp));
	}
      postRewToPartialDef(node);
    }
  ptlinstances_free(lexp_save);
  return bmodified;
}

void NodeManager::buildDefaultExpression(ptexpcal vsig, ptlINode ln)
{
  ptexpcal e2, curDef, e22, e221;
  ptinsth h, vhcal;
  ptlinstances lsig;
  ptINode n;

  while (ptlINode_get_elem(&ln, &n))
    {
      if (get_rightOperand(n, &curDef))
	{
	  e2 = EM->getWhenOrCellArg(curDef);
	  vhcal = EM->getContributionClock(curDef);
	  if (EM->getOperator(e2) == opmele)
	    {
	      while (EM->getOperator(e2) == opmele)
		{
		  EM->getDefaultArgs(e2, &e2, &e22);
		  e221 = EM->getWhenOrCellArg(e22);
		  vhcal = EM->getContributionClock(e22);
		  lsig = EM->getPredecessors(e221);
		  addaff(vsig, e221, vhcal, lsig);
		}
	      curDef = e2;
	      e2 = EM->getWhenOrCellArg(curDef);
	      vhcal = EM->getContributionClock(curDef);
	    }
	  lsig = EM->getPredecessors(e2);
	  addaff(vsig, e2, vhcal, lsig);
	}
      removeNode(n);
    }
}

/**
   @return "true" if the operation succeeds, "false otherwise (the node is not a set of partial definitions).
   It merges the partial definitions in an equation: 
   - X::=E1 when h1, .... X::=En xhen hn, X::= defaultvalue Xd 
   is replaced by 
   -X := (E1 when h1) default ... default (En when hn) default Xd
   Note that by construction, the clocks h1,h2... are exclusives.
*/
boolean NodeManager::PartialDefinitionsMergingSharedSignal(ptINode node)
{
  ptlINode ln;
  ptexpcal vsig;

  if (!is_setOfPartialDefinitions(node))
    return pk_false;
  ln = get_setOfPartialDefinitions_arg(node);
  vsig = ptINode_get_leftOperand(node);
  preChangeDef(node);
  EM->setSharedSignal(vsig, pk_false);
  buildDefaultExpression(vsig, ln);
  insertNode(node, get_nodeClock(node));
  return pk_true;
}

/**
   @return "true" if the operand node (node) denotes a definition of a state variable (ie a set of partial definitions), "false" otherwise.
   For y::=E1,...,y::=En and y a state variable, it creates a signal
   X:= E1 default ....default En (contribution clocks are computed), then y ::=X
*/
boolean NodeManager::PartialDefinitionsMergingStateVar(ptINode node)
{
  ptINode m;
  ptlINode ln;
  ptexpcal vsig, newSig;
  ptinsth h;

  if (!is_setOfPartialDefinitions(node))
    return pk_false;
  vsig = ptINode_get_leftOperand(node);
  if (!EM->isStateVar(vsig))
    return pk_false;
  h = get_nodeClock(node);
  // Creating a signal X := E1 default ... default En
  newSig = EM->mkIdentExp(EM->getOriginalName(vsig), tsignal, EM->getType(vsig), mk_node_undefined(), hundef, NILBACKANNOTATION);
  EM->setClock(newSig, h);
  ln = get_setOfPartialDefinitions_arg(node);
  preChangeDef(node);
  buildDefaultExpression(newSig, ln);
  insertNode(EM->getNode(newSig), h);

  // vsig ::= X 
  EM->addMark(vsig, PART_DEFINITION);
  m = mk_node_equation(vsig, newSig, h, EM->getPredecessors(newSig), NILBACKANNOTATION);
  insertNode(m, h);
  EM->removeMark(vsig, PART_DEFINITION);
  return pk_true;
}

/**
   @return "true" if the operand node (node) denotes a definition of a state variable (ie a set of partial definitions), "false" otherwise.
   It replaces (X a state variable) 
   - X::=E1 when h1,...,X::=En when hn 
   by
   - X := E1 default ... default En default X$1.
   and X becomes a signal.
*/
boolean NodeManager::SharedVariableRemoving(ptINode node)
{
  ptlINode ln;
  ptexpcal zx, vsig, vdef;
  ptinsth h, hcontr;

  vsig = ptINode_get_leftOperand(node);
  if (!EM->isStateVar(vsig))
    return pk_false;
  if (!is_setOfPartialDefinitions(node))
    return pk_false;
  ln = get_setOfPartialDefinitions_arg(node);
  h = EM->getClock(vsig);
  preChangeDef(node);
  EM->setStateTag(vsig, pk_false);
  buildDefaultExpression(vsig, ln);
  get_rightOperand(node, &vdef);
  hcontr = EM->getContributionClock(vdef);
  // Adding the delay.
  zx = mkDelayOf(vsig);
  addDefaultValuePart(vsig, zx, CKM->makeMinus(h, hcontr), ptlinstances_insert(NILLINSTANCES, zx));
  insertNode(node, h);
  return pk_true;
}
