/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** @defgroup TravelNodeAttrGrp TravelDepManager:attribute associated with nodes used by the travelling greedy algorithm. */
/**
   @ingroup TravelNodeAttrGrp
   @return the number of predecessors of a node (vi)
*/
  integer TravelDepManager::ptINodeAttr_getNumberOfPred (ptINode vi)
{
  ptTravelINodeAttributes
    vattr;

  vattr = getinstrattr (vi);
  return vattr->getnbpred ();
}

/** 
   @ingroup TravelNodeAttrGrp
   @return "true" is a node (vi) is without predecessors, "false" otherwise.
 */
boolean TravelDepManager::ptINodeAttr_isWithoutPred (ptINode vi)
{
  ptTravelINodeAttributes
    vattr;
  vattr = getinstrattr (vi);
  return vattr->isnullnbpred ();
}

/** 
   @ingroup TravelNodeAttrGrp
   It decreases the number of predecessors of a node (vi).
 */
void
TravelDepManager::ptINodeAttr_decrNumberOfPred (ptINode vi)
{
  ptTravelINodeAttributes vattr;
  vattr = getinstrattr (vi);
  vattr->decrnbpred ();
}


/**
   @ingroup TravelNodeAttrGrp
   It increases the number of predecessors of a node (vi).
*/
void
TravelDepManager::ptINodeAttr_incrNumberOfPred (ptINode vi)
{
  ptTravelINodeAttributes vattr;
  vattr = getinstrattr (vi);
  vattr->incrnbpred ();
}


/**
   @ingroup TravelNodeAttrGrp
   @return "true" if a node (vi) has been already selected, "false" otherwise.
*/
boolean TravelDepManager::ptINodeAttr_isAlreadySelected (ptINode vi)
{
  ptTravelINodeAttributes
    vattr;
  vattr = getinstrattr (vi);
  return vattr->isAlreadySelected ();
}


/**    @ingroup TravelNodeAttrGrp
       @return "true" if a node (vi) is reached by vpred predecessors, "false" otherwise. 
*/
boolean TravelDepManager::ptINodeAttr_allPredSelected (ptINode vi, integer vpred)
{
  ptTravelINodeAttributes
    vattr;
  vattr = getinstrattr (vi);
  return vattr->numberOfSelectedPredEqualTo (vpred);
}


/**    
       @ingroup TravelNodeAttrGrp
       @return the number of waited predecessors for a node (vi).
*/
integer TravelDepManager::ptINodeAttr_getNumberOfWaitedPred (ptINode vi, integer vpred)
{
  ptTravelINodeAttributes
    vattr;
  vattr = getinstrattr (vi);
  return vattr->get_nbWaitedPredecessors (vpred);
}

/**    
       @ingroup TravelNodeAttrGrp
       @return the number of visited predecessors for a node (vi).
*/
integer TravelDepManager::ptINodeAttr_getNbVisitedPred (ptINode vi)
{
  ptTravelINodeAttributes
    vattr;
  vattr = getinstrattr (vi);
  return vattr->getNbVisitedPred ();
}


/**    @ingroup TravelNodeAttrGrp
       It increases the counter of the number of reached predecessors for a node (vi).
 */
void
TravelDepManager::ptINodeAttr_incrNbOfSelectedPred (ptINode vi)
{
  ptTravelINodeAttributes vattr;
  vattr = getinstrattr (vi);
  vattr->incrNbOfSelectedPred ();
}


/**
   @ingroup TravelNodeAttrGrp
   It (re)initializes to 0 the counter of the number of reached predecessors for a node (vi)
   and marks the node (vi) as selected.
*/
void
TravelDepManager::ptINodeAttr_setSelected (ptINode vi)
{
  ptTravelINodeAttributes vattr;
  vattr = getinstrattr (vi);
  vattr->resetCounterSelectedPred ();
  vattr->setSelected ();
}


/**    @ingroup TravelNodeAttrGrp
       It marks a node (vi) as unselected.
 */
void
TravelDepManager::ptINodeAttr_setUnselected (ptINode vi)
{
  ptTravelINodeAttributes vattr;
  vattr = getinstrattr (vi);
  vattr->setUnselected ();
}


/**    @ingroup TravelNodeAttrGrp
       It (re)initializes the "travelling attributes" for a node (vi).
*/
void
TravelDepManager::ptINodeAttr_reset (ptINode vi)
{
  ptTravelINodeAttributes vattr;
  vattr = getinstrattr (vi);
  vattr->reset ();
  vattr->setReached (pk_false);
}

/**
   @ingroup TravelNodeAttrGrp
   It marks a node (n) as reached/unreached when b/not b.
*/
void
TravelDepManager::ptINodeAttr_setReached (ptINode n, boolean b)
{
  ptTravelINodeAttributes vattr;
  vattr = getinstrattr (n);
  vattr->setReached (b);
}

/**
   @ingroup TravelNodeAttrGrp
   @return "true" if a node (n) has been reached, "false" otherwise.
*/
boolean TravelDepManager::ptINodeAttr_isReached (ptINode n)
{
  ptTravelINodeAttributes
    vattr;
  vattr = getinstrattr (n);
  return vattr->isReached ();
}

/** @defgroup TravelClockAttrGrp TravelDepManager:attribute associated with clocks used by the travelling greedy algorithm. */

/** @ingroup TravelClockAttrGrp
    It increases (+1) the number of predecessors of a clock (h)
 */
void
TravelDepManager::ptinsthAttr_incrNumberOfPred (ptinsth h)
{
  ptTravelClockAttributes vattr;
  vattr = getinsthattr (h);
  vattr->incrnbpred ();
}

/** @ingroup TravelClockAttrGrp
    It decreases (-1) the number of predecessors of a clock (h)
 */
void
TravelDepManager::ptinsthAttr_decrNumberOfPred (ptinsth h)
{
  ptTravelClockAttributes vattr;
  vattr = getinsthattr (h);
  vattr->decrnbpred ();
}

/**  @ingroup TravelClockAttrGrp
     @return the number of predecessors of a clock(h)
 */
integer TravelDepManager::ptinsthAttr_getNumberOfPred (ptinsth h)
{
  ptTravelClockAttributes
    vattr;
  vattr = getinsthattr (h);
  return vattr->getnbpred ();
}

/** @ingroup TravelClockAttrGrp
    @return "true" if a clock is without any predecessor "false" otherwise.
 */
boolean TravelDepManager::ptinsthAttr_isWithoutPred (ptinsth h)
{
  ptTravelClockAttributes
    vattr;
  vattr = getinsthattr (h);
  return vattr->isnullnbpred ();
}

/** @ingroup TravelClockAttrGrp
    It specifies that a predecessor of h has been selected by the greedy algorithm
 */
void
TravelDepManager::ptinsthAttr_incrNbOfSelectedPred (ptinsth h)
{
  ptTravelClockAttributes vattr;
  vattr = getinsthattr (h);
  vattr->incrNbOfSelectedPred ();
}

/** @ingroup TravelClockAttrGrp
    It specifies that a clock (pth) has been selected by the greedy algorithm
*/
void
TravelDepManager::ptinsthAttr_setSelected (ptinsth pth)
{
  ptTravelClockAttributes vattr;

  vattr = getinsthattr (pth);
  vattr->setSelected ();
}

/**
   @ingroup TravelClockAttrGrp
   @return "true" if all the predecessors of h have been selected, "false" otherwise.
*/
boolean TravelDepManager::ptinsthAttr_allPredSelected (ptinsth h)
{
  ptTravelClockAttributes
    vattr;
  vattr = getinsthattr (h);
  return vattr->allPredSelected ();
}


/**
   @ingroup TravelClockAttrGrp
   @return "true" if a clock (h) has been selected by the greedy algorithm, "false" otherwise.
*/
boolean TravelDepManager::ptinsthAttr_isAlreadySelected (ptinsth h)
{
  ptTravelClockAttributes
    vattr;

  vattr = getinsthattr (h);
  return vattr->isAlreadySelected ();
}


/**
   @ingroup TravelClockAttrGrp
   It resets the "travelling" attributes of a clock (pth).
*/
void
TravelDepManager::ptinsthAttr_reset (ptinsth pth)
{
  ptTravelClockAttributes vattr;

  vattr = getinsthattr (pth);
  vattr->initialize ();
  vattr->setReached (pk_false);
}

/**
    @ingroup TravelClockAttrGrp
    It specifies that a clock (h) is reached/unreached when b/not b.
*/
void
TravelDepManager::ptinsthAttr_setReached (ptinsth h, boolean b)
{
  ptTravelClockAttributes vattr;

  vattr = getinsthattr (h);
  vattr->setReached (b);
}

/**
   @ingroup TravelClockAttrGrp
   @return "true" if a clock (h) has been reached by the greedy algorithm, "false" otherwise.
*/
boolean TravelDepManager::ptinsthAttr_isReached (ptinsth h)
{
  ptTravelClockAttributes
    vattr;

  vattr = getinsthattr (h);
  return vattr->isReached ();
}

/** @defgroup TravelGraphAttrGrp TravelDepManager:attribute associated with clocks used by the travelling greedy algorithm to select the objects. */

/**  
     @ingroup TravelGraphAttrGrp
     @return the list of the nodes associated with a clock (h) without predecessors 
     (all their predecessors have been already selected by the greedy algorithm).
     The returned elements can be selected by the algorithm.
*/
ptlINode TravelDepManager::ptinsthAttr_getSelectableNodes (ptinsth h)
{
  ptTravelClockAttributes
    vattr;

  vattr = getinsthattr (h);
  return vattr->getFreeNodesInSubTree ();
}

/**  
     @ingroup TravelGraphAttrGrp
     @return the list of the clocks associated with a clock (h) without predecessors 
     (all their predecessors have been already selected by the greedy algorithm).
     The returned elements can be selected by the algorithm.
*/
ptlinsth TravelDepManager::ptinsthAttr_getSelectableClocks (ptinsth h)
{
  ptTravelClockAttributes
    vattr;

  vattr = getinsthattr (h);
  return vattr->getFreeClocksInSubTree ();
}

/**  @ingroup TravelGraphAttrGrp
    It defines by ptlh the set of the clocks that can be selected by the greedy algorithm for a given clock (h).
*/
void
TravelDepManager::ptinsthAttr_setSelectableClocks (ptinsth h, ptlinsth ptlh)
{
  ptTravelClockAttributes vattr;

  vattr = getinsthattr (h);
  vattr->setFreeClocksInSubTree (ptlh);
}

/** 
     @ingroup TravelGraphAttrGrp
     It defines by ptls the set of the nodes that can be selected by the greedy algorithm for a given clock (h).
*/
void
TravelDepManager::ptinsthAttr_setSelectableNodes (ptinsth h, ptlINode ptls)
{
  ptTravelClockAttributes vattr;

  vattr = getinsthattr (h);
  vattr->setFreeNodesInSubTree (ptls);
}


/**
   @ingroup TravelGraphAttrGrp
   It specifies that the node (i) associated with the clock h, is reached by all its predecessors.
   It can been selected by the greedy algorithm.
   ( i is added to the list of the "free" nodes of h).
*/
void
TravelDepManager::ptinsthAttr_insertSelectableNode (ptinsth h, ptINode i)
{
  ptTravelClockAttributes vattr;
  ptlINode linstr;

  vattr = getinsthattr (h);
  linstr = vattr->getFreeNodesInSubTree ();
  if (ptlINode_insert_elem (&linstr, i))
    vattr->setFreeNodesInSubTree (linstr);
}
