/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   @return the "travelling" attributes assigned to a node (vinstr)
   It is created if it does not exist.
*/
  ptTravelINodeAttributes TravelDepManager::getinstrattr (ptINode vinstr)
{
  ptTravelINodeAttributes
    vattr;

  if (!NM->getAttribute (vinstr, TRAVELATTRIBUTE, (attributeValue *) & vattr))
    {
      vattr = new TravelINodeAttributes ();
      NM->addAttribute (vinstr, TRAVELATTRIBUTE, vattr);
    };
  return vattr;
}

/**
   @return the "travelling" attributes assigned to a clock (h).
   It is created if it does not exist.
*/
ptTravelClockAttributes TravelDepManager::getinsthattr (ptinsth h)
{
  ptTravelClockAttributes
    vattr;

  if (!CKM->getAttribute (h, TRAVELATTRIBUTE, (attributeValue *) & vattr))
    {
      vattr = new TravelClockAttributes ();
      CKM->addAttribute (h, TRAVELATTRIBUTE, vattr);
    };
  return vattr;
}


/** @defgroup infoSubTreeGrp TravelDepManager:sub hierachy attribute. */
/**
   @ingroup infoSubTreeGrp
   @return the attribute associated with a clock (pth) giving informations about the 
   hierarchy with pth as root. This informations are specific for the travel algorithm
   (number of nodes reached by all of their predecessors in this hierarchy, ...°
*/
ptinfoGenCodeSubTree TravelDepManager::ptinsthAttr_get_infoGenCodeSubTree (ptinsth pth)
{
  ptTravelClockAttributes
    vattr;
  vattr = getinsthattr (pth);
  return vattr->getInfoGenCodeSubTree ();
}

/**
   @ingroup infoSubTreeGrp
   It associates "an information sub tree" attribute to a clock (pth).
*/
void
TravelDepManager::ptinsthAttr_set_infoGenCodeSubTree (ptinsth pth, ptinfoGenCodeSubTree v)
{
  ptTravelClockAttributes vattr;
  vattr = getinsthattr (pth);
  vattr->setInfoGenCodeSubTree (v);
}


/**
   @ingroup infoSubTreeGrp
  It sets to v the number of nodes, for the  block (bcl), that can 
  be selected ("without predecessors") in the hierarchy with pth as root.
  The bcl block can be undefined (NILINODE).
*/
void
TravelDepManager::ptinsthAttr_set_nbspredtree (ptinsth pth, integer v, ptINode bcl)
{
  ptinfoGenCodeSubTree ptsp, x;

  ptsp = ptinsthAttr_get_infoGenCodeSubTree (pth);
  if (ptsp == NILINFOGENCODESUBTREE)
    ptinsthAttr_set_infoGenCodeSubTree (pth, new infoGenCodeSubTree (v, bcl));
  else if (ptsp->findspredtree (bcl, &x))
    x->setnbspred (v);
  else
    x->setnext (new infoGenCodeSubTree (v, bcl));
}

/**
   @ingroup infoSubTreeGrp
  It increases the number of nodes, for a block (bcl), that can 
  be selected ("without predecessors") in the hierarchy with pth as root.
  The bcl block can be undefined (NILINODE).
*/
void
TravelDepManager::ptinsthAttr_incr_nbspred (ptinsth pth, ptINode bcl)
{
  ptinfoGenCodeSubTree ptsp, x;

  ptsp = ptinsthAttr_get_infoGenCodeSubTree (pth);
  if (ptsp == NILINFOGENCODESUBTREE)
    ptinsthAttr_set_infoGenCodeSubTree (pth, new infoGenCodeSubTree (1, bcl));
  else if (ptsp->findspredtree (bcl, &x))
    x->incrnbspred ();
  else
    x->setnext (new infoGenCodeSubTree (1, bcl));
}


/**
   @ingroup infoSubTreeGrp
   @return the the number of nodes, for a block (bcl), that can 
   be selected ("without predecessors") in the hierarchy with pth as root.
*/
integer TravelDepManager::ptinsthAttr_get_nbspred (ptinsth pth, ptINode bcl)
{
  ptinfoGenCodeSubTree
    ptsp,
    x;

  ptsp = ptinsthAttr_get_infoGenCodeSubTree (pth);
  if (ptsp == NILINFOGENCODESUBTREE)
    return 0;
  if (!ptsp->findspredtree (bcl, &x))
    return 0;
  return x->getnbspred ();
}

/** @defgroup attributesInitGrp TravelDepManager:initialisations of the attributes for the greedy algorithm.
 */

/**
   @ingroup attributesInitGrp
   Global initializing of the "travelling" attributes.
   All clocks (attributesInitialize_clocks()), nodes (attributesInitialize_nodes()) are initialized.
*/
void
TravelDepManager::attributesInitialize ()
{
  ptlINode lnodes;
  typeMark m;

  m = getMark ();
  attributesInitialize_clocks (CKM->getlhroots (), m);
  attributesInitialize_clocks (CKM->getlhplaced (), m);
  if (NM->getAllNodes (&lnodes))
    attributesInitialize_nodes (lnodes, m);
  CKM->resetMarkAll (m);
  NM->resetMarkAll (m);
}

/**
   @ingroup attributesInitGrp
   It initializes the "travelling" attributes for a list of clocks (lh) using a mark (m).
*/
void
TravelDepManager::attributesInitialize_clocks (ptlinsth lh, typeMark m)
{
  ptinsth vh;

  while (ptlinsth_get_next (&lh, &vh))
    {
      vh = CKM->getRepresentativeFormula (vh);
      attributesInitialize_clock (vh, m);
    }
}

/**
   @ingroup attributesInitGrp
   It initializes the "travelling" attributes for a clock (vh) using a mark (m).
*/
void
TravelDepManager::attributesInitialize_clock (ptinsth vh, typeMark m)
{
  ptinsth h;
  ptlinsth lh;
  ptINode node;
  ptlINode ln;

  if (!CKM->satisfiesIsCanonicalRepr (vh))
    return;
  if (CKM->satisfiesMarked (vh, m))
    return;
  CKM->addMark (vh, m);
  lh = CKM->getClockReferencesTo (vh);
  while (ptlinsth_get_next (&lh, &h))
    if (CKM->satisfiesIsCanonicalRepr (h))
      ptinsthAttr_incrNumberOfPred (h);
  ln = CKM->getNodeReferencesTo (vh);
  while (ptlINode_get_elem (&ln, &node))
    ptINodeAttr_incrNumberOfPred (node);
}

/**
   @ingroup attributesInitGrp
   It initializes the "travelling" attributes for a list of nodes (vlinstr) using a mark (m).
*/
void
TravelDepManager::attributesInitialize_nodes (ptlINode vlinstr, typeMark m)
{
  ptINode vinstr;

  while (ptlINode_get_elem (&vlinstr, &vinstr))
    attributesInitialize_node (vinstr, m);
}

/**
   @ingroup attributesInitGrp
   It initializes the "travelling" attributes for a node (node) using a mark (m).
*/
void
TravelDepManager::attributesInitialize_node (ptINode node, typeMark m)
{
  ptlinsth lh;
  ptINode n1, n2, inNode;
  ptlINode ln;
  ptinsth h;
  ptDCG g;
  ptexpcal nodeCallParams, nodeCallInputs, nodeCallOutputs;
  noperators ope;
  ptldependence ldep;
  ptdependence adep;

  if (NM->isMarked (node, m))
    return;
  NM->addMark (node, m);

  ldep = NM->get_successors (node);
  while (ldep->get_next (&adep, &ldep))
    {
      adep->getDependence (&n1, &n2);
      ptINodeAttr_incrNumberOfPred (n2);
    }

  lh = NM->get_clockSuccessors (node);
  while (ptlinsth_get_next (&lh, &h))
    if (CKM->satisfiesIsCanonicalRepr (h))
      ptinsthAttr_incrNumberOfPred (h);
  /* Traitement des sous-graphes */

  ope = NM->get_oper (node);
  switch ((int) (ope))
    {
    case opsubgraph:
      NM->get_subGraphDefinitions (node, &ln, &lh);
      attributesInitialize_nodes (ln, m);
      if (NM->get_subGraphInputNode (node, &n1))
	attributesInitialize_node (n1, m);
      break;
    case opnodecall:
      NM->get_nodeCall_arg (node, &g, &nodeCallParams, &nodeCallInputs, &nodeCallOutputs);
      while (nodeCallOutputs != NILEXPCAL)
	{
	  inNode = EM->getNode (EM->getNaryArgs (&nodeCallOutputs));
	  attributesInitialize_node (inNode, m);
	}
      break;
    case opmoti:
      {
	inNode = NM->get_indexNode (node);
	attributesInitialize_node (inNode, m);
	if (NM->get_array_inputNode (node, &inNode))
	  attributesInitialize_node (inNode, m);
	NM->get_array_bodyDefinitions (node, &ln, &lh);
	attributesInitialize_nodes (ln, m);
	NM->get_array_initDefinitions (node, &ln, &lh);
	attributesInitialize_nodes (ln, m);
      }
      break;
    }
}

/**
   @ingroup attributesInitGrp
   Cleaning the graph: specific "travelling" attributes are reset (reset_clocks() for the clocks, reset_nodes() for the nodes).
*/
void
TravelDepManager::cleanGraph ()
{
  ptlINode lnodes;

  reset_clocks (CKM->getlhroots ());
  reset_clocks (CKM->getlhplaced ());
  if (NM->getAllNodes (&lnodes))
    reset_nodes (lnodes);
}

/**
   It resets the specific "travelling" attributes for a list of clocks (lh).
*/
void
TravelDepManager::reset_clocks (ptlinsth lh)
{
  ptinsth vh, vh1;

  while (ptlinsth_get_next (&lh, &vh))
    {
      vh = CKM->getRepresentativeFormula (vh);
      reset_clock (vh);
      vh1 = CKM->getNextFormula (vh, pk_false);
      while (vh1 != vh)
	{
	  reset_clock (vh1);
	  vh1 = CKM->getNextFormula (vh1, pk_false);
	}
      vh1 = vh;
      while (CKM->getNextAssertion (vh1, &vh1))
	reset_clock (vh1);
    }
}


/**
   It resets the specific "travelling" attributes for a clock (vh).
*/
void
TravelDepManager::reset_clock (ptinsth vh)
{
  ptinsth vh1, vh2;

  ptinsthAttr_reset (vh);
  if (CKM->getInfArgs (vh, &vh1, &vh2))
    {
      if (CKM->getOperator (vh1) == hnot)
	ptinsthAttr_reset (vh1);
      else if (CKM->getOperator (vh2) == hnot)
	ptinsthAttr_reset (vh2);
    }
  ptinsthAttr_setSelectableClocks (vh, NILLINSTH);
  ptinsthAttr_setSelectableNodes (vh, NILLINODE);
}

/**
   It resets the specific "travelling" attributes for a list of nodes (vlinstr).
*/
void
TravelDepManager::reset_nodes (ptlINode vlinstr)
{
  ptINode node;
  while (ptlINode_get_elem (&vlinstr, &node))
    reset_node (node);
}

/**
   It resets the specific "travelling" attributes for a node (node).
*/
void
TravelDepManager::reset_node (ptINode node)
{
  ptINode InternalInputNode, inNode;
  ptlINode lnodes;
  ptlinsth lh;
  ptDCG g;
  ptexpcal nodeCallParams, nodeCallInputs, nodeCallOutputs;
  noperators ope;

  ptINodeAttr_reset (node);
  ope = NM->get_oper (node);
  switch ((int) (ope))
    {
    case opsubgraph:
      NM->get_subGraphDefinitions (node, &lnodes, &lh);
      reset_nodes (lnodes);
      if (NM->get_subGraphInputNode (node, &InternalInputNode))
	ptINodeAttr_setUnselected (InternalInputNode);
      break;
    case opnodecall:
      NM->get_nodeCall_arg (node, &g, &nodeCallParams, &nodeCallInputs, &nodeCallOutputs);
      while (nodeCallOutputs != NILEXPCAL)
	{
	  inNode = EM->getNode (EM->getNaryArgs (&nodeCallOutputs));
	  ptINodeAttr_reset /* ptINodeAttr_setUnselected */ (inNode);
	}
      break;
    case opmoti:
      // inNode = NM->get_indexNode(node);
      if (NM->get_array_inputNode (node, &inNode))
	ptINodeAttr_setUnselected (inNode);
      NM->get_array_bodyDefinitions (node, &lnodes, &lh);
      reset_nodes (lnodes);
      NM->get_array_initDefinitions (node, &lnodes, &lh);
      reset_nodes (lnodes);
      break;
    }
}
