/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_
#define DEBUG_PRINT_LINE_SEP( ) printf("-------------------------------------------\n")
/** Constructor. The travelling is applied to a graph (g).
    It must be 
       - initialized ( initialize()  method)
       - finalized ( finalize() method)

       The constant signals are available after the call to the method initTravelForInitialClock(), the
       other one after the call toi he method initTravelForClock().
*/
  TravelDepManager::TravelDepManager (ptDCG g)
{
  AGRAPH = g;
  unNodes = AGRAPH->getUnused (&unClocks);
  DEBUG_MODE = pk_false;
}

/**
   It initializes the selected graph (constructor) for a travelling.
   The attributes are initialized (attributesInitialize()). In particular, 
   the number of waited predecessors of each node/clock is computed.
*/
void
TravelDepManager::initialize ()
{
  EM = AGRAPH->getExpressionManager ();
  CKM = AGRAPH->getClockManager ();
  NM = AGRAPH->getNodeManager ();
  init ();
  attributesInitialize ();
}

/**
   It finalizes the travelling: the graph is reinitialized for a next travelling.
 */
void
TravelDepManager::finalize ()
{
  init ();
  cleanGraph ();
}

/**
   It Initializes a travelling with the initial clock as entry point (for constant signals). 
 */
void
TravelDepManager::initTravelForInitialClock ()
{
  ptlINode ln;
  hcourante = CKM->getInitClock ();
  //  signalSelectedClock(hcourante);
  if (AGRAPH->getNodesAtInitClock (&ln))
    copierspredinit (ln, hcourante);
  gfdhcourante = CKM->getClockSubGraph (hcourante);
  lancetre = NILLINSTH;
}


/**
   It initializes a travelling with the clock (htop) as entry point. 
   @return "true" if exists a clock without predecessors, this clock is a child of the htop clock.
   (A classical call is when htop is the fictitious root of the clock hierarchy)
*/
boolean TravelDepManager::initTravelForClock (ptinsth htop)
{
  boolean
    entrypoint;
  ptlinsth
    lh;
  ptinsth
    h;
  boolean
    b;

  if (CKM->satisfiesFictitiousRoot (htop))
    signalSelectedClock (htop);
  lh = CKM->getChildren (htop);
  hcourante = htop;
  entrypoint = pk_false;
  gfdhcourante = CKM->getClockSubGraph (hcourante);
  while (lh != NILLINSTH)
    {
      ptlinsth_get_next (&lh, &h);
      if (ptinsthAttr_isWithoutPred (h) && !ptinsthAttr_isAlreadySelected (h))
	{
	  entrypoint = pk_true;
	  addlhspredgen (hcourante, h);
	  ptinsthAttr_setSelected (h);
	}
    }
  lancetre = ptlinsth_sorted_insert_elem (NILLINSTH, hcourante, &b);
  toplancetre = hcourante;
  return entrypoint;
}


/**
   It specifies that the node (inst) is reached by one of its predecessors.
   If it belongs reached by all of its predecessors, it becomes a candidate for the selection for 
   the greedy algorithm.
*/
void
TravelDepManager::NodeReachedByAPredecessor (ptINode inst)
{
  ptinsth hinst;
  ptINode cNode;

  if (!isNodeAlreadySelected (inst))
    {
      if (DEBUG_MODE)
	printf ("     Node reached by a predecessor: %d \n", NM->get_num (inst));
      hinst = NM->get_nodeClock (inst);
      ptINodeAttr_incrNbOfSelectedPred (inst);
      if (ptINodeAttr_allPredSelected (inst, ptINodeAttr_getNumberOfPred (inst)))
	{
	  if (DEBUG_MODE)
	    printf ("     Node may be selected : %d \n", NM->get_num (inst));
	  ptinsthAttr_insertSelectableNode (hinst, inst);
	  ptINodeAttr_setSelected (inst);
	  NM->get_inComposedNode (inst, &cNode);
	  inchancetre (hinst, cNode);
	}
    }
  else if (DEBUG_MODE)
    printf (" !!!!!  NODE REACHED AFTER: %d \n", NM->get_num (inst));
}


/**
  Search for a clock or a node without predecessors (all its predecessors have been visited).
  @return "true" if such object exists in the graph, "false" otherwise.
  @param bType (out) : "true" if node parameter is defined, "false" if vclock is defined.
  @param node (out) : the selected node.
  @param vclock (out) : the selected clock.
*/
boolean TravelDepManager::getNextNode (boolean * bType, ptINode * node, ptinsth * vclock)
{
  ptinsth
    hssarbre;
  boolean
    b;

  if (CKM->satisfiesFictitiousRoot (hcourante))
    hcourante = CKM->getTickClock ();

  while (pk_true)
    {
      if (nextispredgen (node))
	{
	  *bType = pk_true;
	  return pk_true;
	}
      if (nexthspredgen (vclock))
	{
	  *bType = pk_false;
	  return pk_true;
	}
      if (unssarbre (hcourante, &hssarbre))
	{
	  hcourante = hssarbre;
	  gfdhcourante = CKM->getClockSubGraph (hcourante);
	  lancetre = ptlinsth_sorted_insert_elem (lancetre, hssarbre, &b);
	}
      else
	{
	  ptlinsth_sorted_remove_elem (&lancetre, hcourante);
	  if (lancetre == NILLINSTH)
	    return pk_false;	// it does not exist node/clock to be process.
	  ptinsthAttr_set_nbspredtree (hcourante, 0, bclcourante);
	  hcourante = CKM->getFather (hcourante);
	  gfdhcourante = CKM->getClockSubGraph (hcourante);
	}
    }
}

/**
   It signals that the clock (h) has been selected by the greedy algorithm to its successors
   (ClocksReachedByAPredecessor(), NodesReachedByAPredecessor() methods are applied).
*/
void
TravelDepManager::signalSelectedClock (ptinsth h)
{
  if (!CKM->satisfiesInitClock (h))
    {
      copierspredinit (ptgfd_get_nodes (CKM->getClockSubGraph (h)), h);
      ClocksReachedByAPredecessor (CKM->getClockReferencesTo (h));
      NodesReachedByAPredecessor (CKM->getNodeReferencesTo (h));
    }
}

/** @return "true" if a node (vi) has been already selected by the algorithm, "false" otherwise.
 */
boolean TravelDepManager::isNodeAlreadySelected (ptINode vi)
{
  if (ptlINode_isIn (unNodes, vi))
    return pk_true;
  return ptINodeAttr_isAlreadySelected (vi);
}



/** @return "true" if a clock (vh) has been already selected by the algorithm, "false" otherwise.
 */
boolean TravelDepManager::isClockAlreadySelected (ptinsth vh)
{
  if (ptlinsth_isin_elem (unClocks, vh))
    return pk_true;
  return ptinsthAttr_isAlreadySelected (vh);
}


/** @return the current context array-block. */
ptINode TravelDepManager::getCurrentBlock (void)
{
  return bclcourante;
}

/**
   It saves the currnent array-block before setting a node (bcl) as a new one. 
   It will be closed using the closeBlock() method.
*/
void
TravelDepManager::openBlock (ptINode bcl)
{
  savhcour = ptlinsth_insert_elem (savhcour, hcourante);
  savtoplancetre = ptlinsth_insert_elem (savtoplancetre, toplancetre);
  savbclcour = ptlINode_append_head (bclcourante, savbclcour);
  lancetre = ptlinsth_insert_elem (NILLINSTH, hcourante);
  toplancetre = hcourante;
  bclcourante = bcl;
}

/**
   It closes the current array context. It restitutes the previous one saved in openBlock() method.
*/
void
TravelDepManager::closeBlock (void)
{
  boolean b;
  ptinsth h;

  ptlinsth_get_next (&savhcour, &hcourante);
  gfdhcourante = CKM->getClockSubGraph (hcourante);
  ptlinsth_get_next (&savtoplancetre, &toplancetre);
  ptlINode_get_elem (&savbclcour, &bclcourante);
  h = hcourante;
  lancetre = ptlinsth_sorted_insert_elem (NILLINSTH, h, &b);
  while (h != toplancetre)
    {
      h = CKM->getFather (h);
      lancetre = ptlinsth_sorted_insert_elem (lancetre, h, &b);
    }
}

/**
   It signals that a node (node) has been selected by the greedy algorithm to its successors
   (ClocksReachedByAPredecessor(), NodesReachedByAPredecessor() methods are applied).
*/
void
TravelDepManager::signalSelectedNode (ptINode node)
{
  ClocksReachedByAPredecessor (NM->get_clockSuccessors (node));
  NodesReachedByAPredecessor (NM->get_successors (node));
}

/**
   It specifies that a clock (vi) is reached (resp. unreached) when b (resp. not b).
*/
void
TravelDepManager::setReached (ptinsth vi, boolean b)
{
  ptinsthAttr_setReached (vi, b);
}

/**
   @return "true" if a clock (vi) is reached, "false" otherwise.
   A clock is reached if it has been reached by all its predecessors.
*/
boolean TravelDepManager::isReached (ptinsth vi)
{
  return ptinsthAttr_isReached (vi);
}

/**
   It specifies that a node (vi) is reached (resp. unreached) when b (resp. not b).
   A node is reached if it has been reached by all its predecessors.
*/
void
TravelDepManager::setReached (ptINode vi, boolean b)
{
  ptINodeAttr_setReached (vi, b);
}

/**
   @return "true" if a node (vi) is reached, "false" otherwise.
*/
boolean TravelDepManager::isReached (ptINode vi)
{
  return ptINodeAttr_isReached (vi);
}


/**
   @return "true" if a clock (vh) is reachable (it is a useful clock with some predecessors), "false" otherwise.
   
*/
boolean TravelDepManager::isClockReachable (ptinsth vh)
{
  if (ptinsthAttr_isWithoutPred (vh))
    return pk_false;
  if (ptlinsth_isin_elem (unClocks, vh))
    return pk_false;
  return pk_true;
}

/** 
    @return "true" if a node (vi) is reachable (it is a useful clock with some predecessors) , "false" otherwise.
*/
boolean TravelDepManager::isNodeReachable (ptINode vi)
{
  if (ptlINode_isIn (unNodes, vi))
    return pk_false;
  if (ptINodeAttr_isAlreadySelected (vi))
    return pk_false;
  if (NM->is_delay (vi))
    return pk_true;
  return (ptINodeAttr_getNumberOfPred (vi) > 0);
}

/**
   It signals that an array node (node) has been selected by the greedy algorithm.
   Specific actions are applied for the index and the internal array node: it signals (signalSelectedNode()) that
   these nodes have been selected.
*/
void
TravelDepManager::signalArraySelectedNode (ptINode node)
{
  ptINode n;

  // the nodes of the array may depend on the index.
  n = NM->get_indexNode (node);
  signalSelectedNode (n);
  if (DEBUG_MODE)
    printf ("        >>> index Node Num %d \n", NM->get_num (n));
  // the nodes of the sub-graph depends on an internal node(n). 
  NM->get_array_inputNode (node, &n);
  NM->arrayLinkNodes (node, pk_false);
  if (DEBUG_MODE)
    printf ("        >>> array inputNode Node Num %d \n", NM->get_num (n));
  signalSelectedNode (n);
}

/**
   It signals that a subgraph node (node) has been selected by the greedy algorithm: the successors of the 
   internal node associated with a subgraph are visited.
   Assumed: node is a subgraph node.
*/
void
TravelDepManager::signalSubGraphSelectedNode (ptINode node)
{
  ptINode n;

  // the nodes of the sub-graph depends on an internal node(n). 
  NM->get_subGraphInputNode (node, &n);
  signalSelectedNode (n);
}

/**
   Debugging:
*/
void
TravelDepManager::DEBUG_TRACE_NODE (ptINode node)
{
  int nbpred, nbpredAlreadVisited;

  if (DEBUG_MODE)
    {
      nbpred = ptINodeAttr_getNumberOfPred (node);
      nbpredAlreadVisited = ptINodeAttr_getNbVisitedPred (node);
      DEBUG_PRINT_LINE_SEP ();
      if (nbpredAlreadVisited == nbpred)
	printf ("node num= %d has %d predecessors, %d already visited, ----OK \n", NM->get_num (node), nbpred, nbpredAlreadVisited);
      else
	printf ("node num= %d has %d predecessors, %d already visited\n", NM->get_num (node), nbpred, nbpredAlreadVisited);
    }
}


/**
   Debugging:
*/
_VIRTUAL_ void
TravelDepManager::DEBUG_PRINT_NODEINFOS (ptINode node, ptlINode lnreached, ptlinsth lhreached)
{
  ptldependence lpred;
  ptdependence aDep;
  ptinsth hdep;
  ptINode n11, n12;
  ptlinsth lh;
  ptinsth h;

  if (DEBUG_MODE)
    {
      // Predecessors
      printf ("node num= %d waits for \n", NM->get_num (node));
      lpred = NM->get_predecessors (node);
      while (lpred->get_next (&aDep, &lpred))
	{
	  hdep = aDep->getDependence (&n11, &n12);
	  if (!isNodeAlreadySelected (n11))
	    printf ("       node num= %d\n", NM->get_num (n11));
	}

      lh = NM->get_clockPredecessors (node);
      while (ptlinsth_get_next (&lh, &h))
	{
	  if (h == CKM->getRepresentativeFormula (h))
	    if (!ptinsthAttr_isAlreadySelected (h))
	      {
		printf ("       clock name= %s\n", ident2string (CKM->getName (h)));

	      }
	}

      printf ("node num= %d already visited \n", NM->get_num (node));
      lpred = NM->get_predecessors (node);
      while (lpred->get_next (&aDep, &lpred))
	{
	  hdep = aDep->getDependence (&n11, &n12);
	  if (isNodeAlreadySelected (n11))
	    printf ("       node num= %d\n", NM->get_num (n11));
	}

      lh = NM->get_clockPredecessors (node);
      while (ptlinsth_get_next (&lh, &h))
	{
	  if (h == CKM->getRepresentativeFormula (h))
	    if (ptinsthAttr_isAlreadySelected (h))
	      printf ("       clock name= %s\n", ident2string (CKM->getName (h)));
	}
    }
}


/**
   Debugging:
*/
void
TravelDepManager::DEBUG_TRACE_NODE (ptINode node, ptlINode ln, ptlinsth lh)
{
  int nbpred, nbpredAlreadVisited;

  if (DEBUG_MODE)
    {
      DEBUG_TRACE_NODE (node);
      DEBUG_PRINT_NODEINFOS (node, ln, lh);
    }
}

/**
   Debugging:
*/
void
TravelDepManager::DEBUG_TRACE_CLOCK (ptinsth h, ptlINode ln, ptlinsth lh)
{
  if (DEBUG_MODE)
    {
      // ptinsthAttr_getNumberOfPred(ptinsth h)
    }
}

/**
   Debugging:
*/
void
TravelDepManager::DEBUG_TRACE_CLOCK (ptinsth h)
{
  if (DEBUG_MODE)
    {
      // ptinsthAttr_getNumberOfPred(ptinsth h)
    }
}

/**
   Debugging:
*/
void
TravelDepManager::PRINTDEBUG (ptlINode ln, ptlinsth lh)
{
  ptINode node;
  ptinsth h;
  ptlINode lns;
  ptlinsth lnh;

  if (DEBUG_MODE)
    {
      if ((ln != NILLINODE) || (lh != NILLINSTH))
	{
	  printf ("******************** CircuitCalculus::visitReached() BEGIN\n");
	  lnh = lh;
	  lns = ln;
	  while (ptlINode_get_elem (&ln, &node))
	    DEBUG_TRACE_NODE (node, lns, lnh);
	  while (ptlinsth_get_next (&lh, &h))
	    DEBUG_TRACE_CLOCK (h, lns, lnh);
	  printf ("******************** CircuitCalculus::visitReached() END\n");

	}
    }
}

/**
   Set the DEBUG MODE to b.
*/
void
TravelDepManager::setDebugMode (boolean b)
{
  DEBUG_MODE = b;
}
