/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/** 
    It adds a clock (h) in the list of the possible clocks, 
    associated with the clock k, that can be selected by the algorithm.
*/
  void
TravelDepManager::addlhspredgen (ptinsth k, ptinsth h)
{
  boolean b;

  ptinsthAttr_setSelectableClocks (k, ptlinsth_sorted_insert_elem (ptinsthAttr_getSelectableClocks (k), h, &b));
}


/**
  It increases the number of nodes without predecessors, for a array block (i), for each clock on the 
  branch h until the root of its hierarchy.
  (i can be the NILNODE node).
*/
void
TravelDepManager::inchancetre (ptinsth h, ptINode i)
{
  boolean b_fin;
  ptinsth k;

  b_fin = pk_false;
  do
    {
      ptinsthAttr_incr_nbspred (h, i);
      k = CKM->getFather (h);
      if (h == k)
	b_fin = pk_true;
      else
	{
	  h = k;
	  if (CKM->satisfiesFictitiousRoot (h))
	    b_fin = pk_true;
	  else
	    b_fin = (h == CKM->getTickClock ());
	}
    }
  while (!b_fin);
}



/**
   @return "true" if it exists a sub-tree (vhres) of a clock(h) with nodes that can be selected (all of their predecessors have been selected by the 
   algorithm and reached), "false" otherwise.
   The selected clock (vhres) is the child of h that have the greater number of "free" nodes.
   This information is available for each clock (see inchancetre()).
*/
boolean TravelDepManager::unssarbre (ptinsth h, ptinsth * vhres)
{
  ptlinsth
    lhfils;
  ptinsth
    hfils;
  integer
    max,
    n;

  lhfils = CKM->getChildren (h);
  max = 0;
  *vhres = NILINSTH;
  while (ptlinsth_get_next (&lhfils, &hfils))
    {
      hfils = CKM->getRepresentativeFormula (hfils);
      if (ptinsthAttr_isAlreadySelected (hfils))
	{
	  n = ptinsthAttr_get_nbspred (hfils, bclcourante);
	  if (max < n)
	    {
	      *vhres = hfils;
	      max = n;
	    }
	}
    }
  return (*vhres != NILINSTH);
}



/**
   It adds a node (node), associated with the clock (h), if it is without predecessors in the 
   list of the possible nodes, associated with h, that can be selected by the algorithm.
*/
void
TravelDepManager::appendPossibly_generated_inst (ptINode node, ptinsth h)
{
  ptINode cNode;

  if (!isNodeAlreadySelected (node))
    if (ptINodeAttr_isWithoutPred (node))
      {
	ptINodeAttr_setSelected (node);
	ptinsthAttr_insertSelectableNode (h, node);
	NM->get_inComposedNode (node, &cNode);
	inchancetre (h, cNode);
      }
}

/**
   It adds the nodes (appendPossibly_generated_inst()) of the list ln, associated with a clock (h) 
   without predecessors in the list of the possible nodes that can be selected by the algorithm.
   It must be called when a clock has been selected/reached by the algorithm.
*/
void
TravelDepManager::copierspredinit (ptlsucc lnode, ptinsth h)
{
  ptINode node;

  while (ptlINode_get_elem (&lnode, &node))
    appendPossibly_generated_inst (node, h);
}


/**
   @return "true" if a node (vinst) is selected by the greedy algorithm, "false" otherwise.
   The selecting takes into account the current contexts (current selected clock, current array block).
   This node is removed from the set of the possibly nodes to be selected associated with the current selected clock.
   See nexthspredgen() for the nodes.
*/
boolean TravelDepManager::nextispredgen (ptINode * vinstr)
{
  ptlsucc
    linst,
    linstsave;
  ptINode
    cNode;
  boolean
    selected;

  linst = ptinsthAttr_getSelectableNodes (hcourante);
  linstsave = linst;
  while (ptlINode_get_elem (&linst, vinstr))
    {
      if (NM->is_lfields (*vinstr))
	selected = pk_true;
      else
	{
	  NM->get_inComposedNode (*vinstr, &cNode);
	  selected = (cNode == bclcourante);
	}
      if (selected)
	{
	  if (ptlINode_remove_elem (&linstsave, *vinstr))
	    ptinsthAttr_setSelectableNodes (hcourante, linstsave);
	  return pk_true;
	}
    }
  return pk_false;
}

/**
   @return "true" if a clock (horl) is selected by the greedy algorithm, "false" otherwise.
   The selecting takes into account the current contexts (current selected clock, current array block).
   This clock is removed from the set of the possibly nodes to be selected associated with the current selected clock.
   See nextispredgen() for the nodes.
*/
boolean TravelDepManager::nexthspredgen (ptinsth * horl)
{
  ptlinsth
    lhorl,
    lhorlsave;
  boolean
    bdummy;
  ptINode
    cNode;

  lhorl = ptinsthAttr_getSelectableClocks (hcourante);
  lhorlsave = lhorl;
  while (ptlinsth_get_next (&lhorl, horl))
    {
      CKM->getInComposedNode (*horl, &cNode);
      if (cNode == bclcourante)
	{
	  ptinsthAttr_setSelectableClocks (hcourante, ptlinsth_remove_elem (lhorlsave, *horl, &bdummy));
	  return pk_true;
	}
    }
  return pk_false;
}



/**
   It specifies that the operand clock(horl) is reached by one of its predecessors.
   If it belongs reached by all of its predecessors, the clock becomes a candidate 
   for the selection by the greedy algorithm.
*/
void
TravelDepManager::ClockReachedByAPredecessor (ptinsth horl)
{
  ptinsth hpere, h;
  ptINode cNode;

  if (CKM->satisfiesIsCanonicalRepr (horl))
    {
      if (!ptinsthAttr_isAlreadySelected (horl))
	{
	  if (DEBUG_MODE)
	    printf ("         Clock reached by a predecessor: %s (num %d)", ident2string (CKM->getName (horl)), CKM->getNum (horl));
	  ptinsthAttr_incrNbOfSelectedPred (horl);
	  if (ptinsthAttr_allPredSelected (horl))
	    {
	      if (DEBUG_MODE)
		printf (" >>> CLOCK MAY BE SELECTED ");
	      ptinsthAttr_setSelected (horl);
	      if ((CKM->getUnaryArg (horl, &h)) && (CKM->getOperator (horl) == hnot))
		ClocksReachedByAPredecessor (CKM->getClockReferencesTo (horl));
	      else
		{
		  hpere = CKM->getFather (horl);
		  addlhspredgen (hpere, horl);
		  // inchancetre(hpere, bclcourante);
		  CKM->getInComposedNode (horl, &cNode);
		  inchancetre (hpere, cNode);
		}
	    }
	  if (DEBUG_MODE)
	    printf ("\n");
	}
      else if (DEBUG_MODE)
	printf (" !!!!!  CLOCK REACHED AFTER: %s (num %d) \n", ident2string (CKM->getName (horl)), CKM->getNum (horl));
    }
}


/**
   It specifies that the clocks of a list (lh) are reached by one of its predecessors.
   If it belongs reached by all of its predecessors, a clock becomes a candidate for the selection by 
   the greedy algorithm.
*/
void
TravelDepManager::ClocksReachedByAPredecessor (ptlinsth lh)
{
  ptinsth h;

  while (ptlinsth_get_next (&lh, &h))
    ClockReachedByAPredecessor (h);
}


/**
   For each node n of the list of dependences (ldep),
   it applies the NodeReachedByAPredecessor() method for n.
*/
void
TravelDepManager::NodesReachedByAPredecessor (ptlINode linst)
{
  ptINode inst;

  while (ptlINode_get_elem (&linst, &inst))
    NodeReachedByAPredecessor (inst);
}


/**
   For each dependence (n1->n2) of the list of dependences (ldep),
   it applies the NodeReachedByAPredecessor() method for n2.
*/
void
TravelDepManager::NodesReachedByAPredecessor (ptldependence ldep)
{
  ptINode n1, n2;
  ptdependence adep;
  while (ldep->get_next (&adep, &ldep))
    {
      adep->getDependence (&n1, &n2);
      NodeReachedByAPredecessor (n2);
    }
}

/**
   Internal initialisations.
*/
void
TravelDepManager::init ()
{
  hcourante = CKM->getUndefClock ();
  gfdhcourante = NILGFD;
  bclcourante = NILINODE;
  toplancetre = NILINSTH;
  // saving context.
  savhcour = NILLINSTH;
  lancetre = NILLINSTH;
  savtoplancetre = NILLINSTH;
  savbclcour = NILLINODE;
}
