/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** 
    Constructor. It initializes the profile calculus for a graph (g).
 */
  ProfileBuilder::ProfileBuilder (ptDCG g)
{
  init (g);
}

/**
   Constructor. The user must use the init() method to initialize the profile calculus for a graph.
*/
ProfileBuilder::ProfileBuilder ()
{
}

/**  
     It executes the profile calculus for the current graph (see constructor).
     It solves the visibility of the signals/clocks by the branching of their references in the hierarchy:
       - the inputs are visited (See InputsGraphVisibility()).
       - interface calculus is applied for specific objects: arrays (interface_arrays()), subgraphs (interface_subGraphs()),
         node calls (interface_nodeCalls()).
       - the hierarchy is left to right topdown traversed (leftRightTopdownTraversal())
       - the constraints on placed formula and the assertions are visited (ConstrainstPlacedClocks())
       - the explicit dependences are visisted (mkInterfaceDependences()).
*/
void
ProfileBuilder::interfaceCalculus ()
{
  InputsGraphVisibility (inputs);
  interfaceComposedNodes ();
  substituableClocksInterface (rootghdc);
  leftRightTopdownTraversal (rootghdc);
  ConstrainstPlacedClocks ();
  NullClockAssertions();
  mkInterfaceDependences ();
}

/** 
    It finalizes the profile calculus for the current graph.
    A message is printed when the current graph has some signals synchronous with the null clock.
 */
void
ProfileBuilder::finalize ()
{
  boolean b;
  ptlinstances vli, lident, lsig;
  ptexpcal vsig;

  if (AGRAPH->getNullClockSignal (&lsig))
    b = pk_true;
  else
    {				// Useful ?
      b = pk_false;
      if (EM->getExprList (opident, &lident))
	{
	  lsig = CKM->getSynchronousSignals (CKM->getNullClock ());
	  while (b && ptlinstances_get_next (&lsig, &vsig))
	    b = ptlinstances_isin_exp (lident, vsig);
	}
    }
  if (b)
    pK_messenger->message (info_hnull);
}

/** 
    It initializes the profileBuilder for a graph (g).
*/
void
ProfileBuilder::init (ptDCG g)
{
  AGRAPH = g;
  EM = AGRAPH->getExpressionManager ();
  CKM = AGRAPH->getClockManager ();
  NM = AGRAPH->getNodeManager ();
  DCplusGraph = isEventClockedFormat (AGRAPH->getFormatLevel ());
  unNodes = AGRAPH->getUnused (&unClocks);
  inputs = AGRAPH->getInputs ();
  outputs = AGRAPH->getOutputs ();
  rootghdc = AGRAPH->getRootClock ();
  TickClock = CKM->getTickClock ();
}

/**
   A clock (vh) is declared as an output of the graph (g).
*/
void
ProfileBuilder::becomesOutput (ptDCG g, ptinsth vh)
{
  if (AGRAPH != g)
    init (g);
  insert_clock_output_to_except (vh, rootghdc, pk_false);
}


/**
   A shared signal (vsig) is declared as an input/output of the graph.
*/
void
ProfileBuilder::becomesInputOutput (ptDCG g, ptexpcal vsig)
{
  if (AGRAPH != g)
    init (g);
  if (EM->isSharedSignalBySeveralGraphs (vsig))
    stInterface_insert_sort (rootghdc, vsig);
}
