/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#define BTRACE 1

_PRIVATE_
/**
   It suspends the name setting for the built clocks during the current operation.
   The setting clock name operation is expensive for big programs. 
   It is done using a mark called NO_NAME_FOR_OBJ, global attribute defined in CompilerAttributes.h file.
   It is applied to each clock of a list (l).
*/
  void
exclusiveClocks::suspendNameSetting (ptlinsth l)
{
  ptinsth h;

  while (ptlinsth_get_next (&l, &h))
    if (CKM->getNum (h) > clockStartIndex)
      // created here : no name.
      CKM->addMark (h, NO_NAME_FOR_OBJ);
}

/**
   It updates the matrix (triangular) by setting the representative formula of each element.
*/
void
exclusiveClocks::triangularMatrixFinalize ()
{
  int i, j;

  for (i = 0; i < tableSize; i++)
    for (j = i + 1; j < tableSize; j++)
      matrix[i][j] = CKM->getRepresentativeFormula (matrix[i][j]);
}


/**
  It marks the clock vh1 and vh2
  - as exclusive when bexclusive : 
       - the matrix matrix[i][j] is set to null clock (i, j are the index of vh1/vh2 for the matrix calculus).
       - matrix DoneMatrix[i][j] is set to pk_true.	
  - as not exclusive otherwise: the matrix DoneMatrix[i][j] is set to pk_true.
*/
void
exclusiveClocks::setExclusiveInfo (ptinsth vh1, ptinsth vh2, boolean bexclusive)
{
  int *i1, *i2;
  int i, j, tmp;

  if (CKM->getAttribute (vh1, IndexAttr, (attributeValue *) & i1))
    if (CKM->getAttribute (vh2, IndexAttr, (attributeValue *) & i2))
      {
	i = *i1;
	j = *i2;
	if (i > j)
	  {
	    tmp = i;
	    i = j;
	    j = tmp;
	  }
	if (bexclusive)
	  matrix[i][j] = the_nullClock_MC_copy;
	DoneMatrix[i][j] = pk_true;
      }
}

/**
   The clocks vh1 and vh2 are marked as exclusive (setExclusiveInfo()).
   The property is propagated using the hierarchy: all clock H in tree(vh1) and K in tree(vh2) are exclusive.
*/
void
exclusiveClocks::exclusiveForEachClockSubTrees (ptinsth vh1, ptinsth vh2)
{
  ptlinsth lchild;
  ptinsth vh;

  if (vh1 != vh2)
    {
      setExclusiveInfo (vh1, vh2, pk_true);
      lchild = CKM->getChildren (vh1);
      while (ptlinsth_get_next (&lchild, &vh))
	exclusiveForEachClockSubTrees (vh, vh2);

      lchild = CKM->getChildren (vh2);
      while (ptlinsth_get_next (&lchild, &vh))
	exclusiveForEachClockSubTrees (vh, vh1);
    }
}

/**
   The clocks vh1 and vh2 are marked as not exclusive (setExclusiveInfo()).
   The property is propagated using the hierarchy: all clock H in tree(vh1) and K in tree(vh2) are not exclusive.
*/
void
exclusiveClocks::notExclusiveForEachClockSubTrees (ptinsth vh1, ptinsth vh2)
{
  ptlinsth lchild;
  ptinsth vh;

  if (vh1 != vh2)
    {
      setExclusiveInfo (vh1, vh2, pk_false);
      lchild = CKM->getChildren (vh1);
      while (ptlinsth_get_next (&lchild, &vh))
	notExclusiveForEachClockSubTrees (vh, vh2);

      lchild = CKM->getChildren (vh2);
      while (ptlinsth_get_next (&lchild, &vh))
	notExclusiveForEachClockSubTrees (vh, vh1);
    }
}

/**
   Assumed: lh is a list of clocks, sister of vh.
   For vh=[c] and [d] an element of lh undersampling of the same clock, all clock H in tree([c]) and K in tree([d]) cannot be exclusive,
   by definition of the hierarchy. It calls notExclusiveForEachClockSubTrees(ptinsth , ptinsth )for each pair.
*/
void
exclusiveClocks::notExclusiveForEachClockSubTrees (ptinsth vh, ptlinsth lh)
{
  ptinsth vh1;
  ptexpcal e;

  if (CKM->satisfiesSub (vh))
    {
      e = CKM->getConditionSub (vh);
      while (ptlinsth_get_next (&lh, &vh1))
	{
	  vh1 = CKM->getRepresentativeFormula (vh1);
	  if (CKM->satisfiesSub (vh1))
	    if (e != CKM->getConditionSub (vh1))
	      notExclusiveForEachClockSubTrees (vh, vh1);
	}
    }
}

/**
   Assumed: lh is the list of the clocks under vh in the hierarchy.
   For a clock k of lh, k and vh are, by definition of the hierarchy, not exclusive.
*/
void
exclusiveClocks::notExclusiveForSubTree (ptinsth vh, ptlinsth lh)
{
  ptinsth aChild;

  while (ptlinsth_get_next (&lh, &aChild))
    {
      setExclusiveInfo (vh, aChild, pk_false);
      notExclusiveForSubTree (vh, CKM->getChildren (aChild));
    }
}

/**
   It travels the clock hierarchy to extract some obvious exclusive informations, such that:
    - For [c] and [not c], all clock H in tree([c]) and K in tree([not c]) are exclusive.  It calls exclusiveForEachClockSubTrees().
    - For [c] and [d], all clock H in tree([c]) and K in tree([d]) are not exclusive.  It calls  notExclusiveForEachClockSubTrees().
*/
void
exclusiveClocks::extractHierarchyInfo (ptlinsth lh)
{
  ptinsth vother, vh;
  ptlinsth lch;

  while (ptlinsth_get_next (&lh, &vh))
    {
      /* For [c] and [not c], all clock H in tree([c]) and K in tree([not c]) are exclusive. */
      if (CKM->getOperator (vh) == sub)
	if (CKM->getComplExtraction (vh, &vother))
	  exclusiveForEachClockSubTrees (vh, vother);
      /* For [c] and [d], all clock H in tree([c]) and K in tree([d]) are not exclusive. */
      notExclusiveForEachClockSubTrees (vh, lh);
      lch = CKM->getChildren (vh);
      /* k a child of vh, k and vh are not exclusive. */
      notExclusiveForSubTree (vh, lch);
      /* the hierarchy of vh is visited. */
      extractHierarchyInfo (lch);
    }
}



/**
   The elements of the matrix are first initialized by the "undefined" clock.
   Then, the hierarchy is exploited, see extractHierarchyInfo().
*/
void
exclusiveClocks::lowerBoundOpInitialize ()
{
  ptinsth vh1;
  int i, j;

  for (i = 0; i < tableSize; i++)
    for (j = i + 1; j < tableSize; j++)
      {
	matrix[i][j] = hundef;
	DoneMatrix[i][j] = pk_false;
      }

  for (i = 0; i < tableSize; i++)
    {
      vh1 = table[i];
      CKM->addAttribute (vh1, IndexAttr, &tableIndex[i]);
    }

  extractHierarchyInfo (CKM->getlhroots ());
}

/**
   It computes the lower bounds of each pair of selected clocks. Only 1/2 matrix is used.
   It exploits the known information set in extractHierarchyInfo().
*/
void
exclusiveClocks::lowerBoundOp ()
{
  ptinsth vh1, vh2, res;
  int i, j, n;

  if (BTRACE)
    cout << "         Table Size = " << tableSize << ", Possible number of cases: " << (tableSize * tableSize) / 2 << endl;

  lowerBoundOpInitialize ();
  n = 0;
  for (i = 0; i < tableSize; i++)
    {
      vh1 = table[i];
      // if (BTRACE) cout << " Current line:" << i << endl;
      for (j = i + 1; j < tableSize; j++)
	{
	  vh2 = table[j];
	  /* It is useful or already known ? */
	  if (!DoneMatrix[i][j])
	    {
	      n++;
	      if (CKM->treeReduceInf (vh1, vh2, &res))
		{
		  /* Here also, the hierarchy is used. */
		  if (CKM->satisfiesNullClock (res))
		    exclusiveForEachClockSubTrees (vh1, vh2);
		  else
		    notExclusiveForEachClockSubTrees (vh1, vh2);
		}
	      else
		res = CKM->makeInf (vh1, vh2);
	      matrix[i][j] = res;
	    }
	}
    }
  if (BTRACE)
    cout << "         Computed number of cases: " << n << endl;
}

/**
   It adds (selects)
   - all elements of ls when m=UNDEFINED_MARK
   - the elements marked with m otherwise
   in a table used for matrix calculus, a table of indexes is also used.
*/
void
exclusiveClocks::clockListProcessing (ptlinsth ls, typeMark m)
{
  ptinsth vh;
  integer i;

  i = 0;
  if (m == UNDEFINED_MARK)
    while (ptlinsth_get_next (&ls, &vh))
      {
	if (!CKM->satisfiesUndefClock (vh) && !CKM->satisfiesInitClock (vh))
	  {
	    table.push_back (vh);
	    tableIndex.push_back (i);
	    i++;
	  }
      }
  else
    while (ptlinsth_get_next (&ls, &vh))
      if (CKM->satisfiesMarked (vh, m))
	{
	  table.push_back (vh);
	  tableIndex.push_back (i);
	  i++;
	}
}

/**
   It selects the clocks of the list (lh) for a matrix calculus.
   It builds the \e table attribute member. 
   The matrix will be produced using the size of this table.
*/
void
exclusiveClocks::selectClocks (typeMark m)
{
  clockStartIndex = CKM->getNumberOfCreatedFormula ();
  clockListProcessing (CKM->getlhplaced (), m);
  clockListProcessing (CKM->getlhroots (), m);
  tableSize = table.size ();
}

/**
   It creates the matrix from the size of the table (see selectClocks()).
*/
void
exclusiveClocks::makeMatrix ()
{
  integer i;
  matrix.reserve (tableSize);
  DoneMatrix.reserve (tableSize);
  for (i = 0; i < tableSize; i++)
    {
      matrix.push_back (typetab (tableSize));
      DoneMatrix.push_back (BoolTabType (tableSize));
    }
}

/**
   It assigns to each clock its exclusive clocks. It is set as an attribute, this attribute is a predefined one called EXCLUSIVE_CLOCKS.
   It is defined in the file $GraphVM_export/CompilerAttributes.h.
*/
void
exclusiveClocks::setAttribute ()
{
  integer i, j;
  ptinsth prod, h1, h2;
  ptlinsth lh1, lh2;

  // Cleaning.
  CKM->resetMarkAll (EXCLUSIVE_CLOCKS);
  // Setting.
  for (i = 0; i < tableSize; i++)
    {
      h1 = table[i];
      if (!CKM->getAttribute (h1, EXCLUSIVE_CLOCKS, (attributeValue *) & lh1))
	lh1 = NILLINSTH;
      for (j = i + 1; j < tableSize; j++)
	{
	  prod = matrix[i][j];
	  if (CKM->satisfiesNullClock (prod))
	    {
	      h2 = table[j];
	      lh1 = ptlinsth_insert_elem (lh1, h2);
	      if (!CKM->getAttribute (h2, EXCLUSIVE_CLOCKS, (attributeValue *) & lh2))
		lh2 = NILLINSTH;
	      lh2 = ptlinsth_insert_elem (lh2, h1);
	      CKM->addAttribute (h2, EXCLUSIVE_CLOCKS, (attributeValue) lh2);
	    }
	}
      if (lh1 != NILLINSTH)
	CKM->addAttribute (h1, EXCLUSIVE_CLOCKS, (attributeValue) lh1);
    }
}


/**
  Initializing the calculus:
  - It creates a matrix taking account of the number of selected clocks (makeMatrix()),
*/
void
exclusiveClocks::initCalculus ()
{
  hundef = CKM->getUndefClock ();
  the_nullClock_MC_copy = CKM->getNullClock ();
  IndexAttr = getMark ();
  makeMatrix ();
  CKM->systemSolving_init ();
}

/**
   Finalizing the calculus:
   - it applies the clock calculus (IGraphClockCalculus()),
   - it updates the matrix (triangularMatrixFinalize()).
*/
void
exclusiveClocks::finalCalculus ()
{
  CKM->clockCalculus (5);
  triangularMatrixFinalize ();
  suspendNameSetting (CKM->getlhplaced ());
  suspendNameSetting (CKM->getlhroots ());
}

/**
   It computes the lower bound for each pair of selected clocks (selectClocks()). It uses the hierarchy to avoid 
   to compute all the cases.
   It is composed of 
   - an initial step (initCalculus()),
   - the specific computation (lowerBoundOp()),
   - a final step (finalCalculus())
*/
void
exclusiveClocks::lowerBoundPairOfClocksCalculus ()
{
  initCalculus ();
  lowerBoundOp ();
  finalCalculus ();
}
