/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_
/** Constructor.
*/
  Homomorphism::Homomorphism ()
{
}

/**
   Destructor.
*/
_VIRTUAL_ Homomorphism::~Homomorphism ()
{
}

/**
   @return "true" if the operation succeeds, "false" otherwise.

   It produces a graph (gh) by a morphism of another one (g) using a table (homoTable). 
   This table gives for each Signal operator its definition in the new one.
   The clocks are already imported in gh (the hierachy of clocks is kept): the link from a clock of g to its 
   clone in gh is given by an attribute (usedMarkBefore defined by markCopy).
   It is composed of the several calls:
   - initialize()
   - mkAttributeInterface()
   - homomorphismTypeCode()
   - homomorphismDclIdentifiers()
   - homomorphismDclClocks()
   - homomorphismSubGraphs()
   - homomorphismNodes()
   - homomorphismClocks()
   - homomorphismInterface()
   - addInputs()
   - homomorphismDependences()
   - homomorphismEnd()
   - finalize()

   \note
   A clock (clk) of the original graph 
    - is associated with its clone (nclk) in the new graph using the usedMarkBefore attribute. This link is \e assumed available here.
    - will be associated with  the homomorphism signal (x) associated with the clock (nclk) in the new graph.
    .
    The  clock (nclk) of the new graph  will be also associated with the homomorphism signal (x).

    \code
    For example: in profiling studies, the "cost" of a clock, defined by the formula h = h1 ^+ h2, 
    is given by the definition in the morphism table by

    operator ^+ = (?x,y; !z; ) pragmas Morphism "(&h&o, &done) :=CostClkPlus{&t}(&h&i1, &h&i2, &hck&o, &wait)" end pragmas;

    In the new graph, we will have:
    - c(h) = c(h1) ^+ c(h2), c() stands for cloneOf(), the links (h,c(h)), (h1,c(h1)), (h2,c(h2)) are assumed available
    - the homomorphism signal associated with h, cost_h, will be defined by the node:
    (cost_h, done) := CostClkPlus(..)(cost_h1, cost_h2, ...)
    \endcode
    
*/
boolean Homomorphism::apply (ptDCG g, ptDCG gh, typeMark markCopy, ptHomomorphismTable homoTable)
{
  ptlinsth
    lh;
  ptlINode
    ln,
    ln1,
    lio;

  pragmaOperatorTable = homoTable;
  usedMarkBefore = markCopy;
  initialize (g, gh);
  if (isOnError ())
    return pk_false;
  iomark = getMark ();
  currentBlock = sNM->mk_node_undefined ();
  fNM->getAllNodes (&ln);
  ln = ptlINode_Copy (ln);
  ln1 = splitIONode (ln, &lio);
  lh = ptlinsth_duplicate (fCKM->getlhroots ());
  lh = ptlinsth_sorted_fusion (lh, fCKM->getlhplaced ());

  sEM->setUnifyExpressionMode (pk_false);
  mkAttributeInterface ();
  homomorphismTypeCode ();
  homomorphismDclIdentifiers ();
  homomorphismDclClocks ();
  homomorphismSubGraphs ();
  homomorphismNodes (ln1);
  homomorphismClocks (lh);
  homomorphismInterface ();
  addInputs ();
  homomorphismDependences (ln);
  homomorphismEnd ();
  finalize ();
  return !isOnError ();
}
