/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/// PREDEFINED OBJECT OF THE VIRTUAL MACHINE.
UtilityClock *pK_UtilityClock = new UtilityClock ();


_PRIVATE_
#define isApplyOnSubGraph() SelectedObjectMark > 0
  boolean UtilityClock::isSelected (ptINode node)
{
  if (isApplyOnSubGraph ())
    return NM->isMarked (node, SelectedObjectMark);
  return pk_true;
}

/**
   @return "true" if the utility clock of a node (node) is localy known (ie in the TableNodeClock table), this clock is returned in h, "false" otherwise.   
*/
boolean
UtilityClock::getUtilityClock (ptINode node, ptinsth * h)
{
  if (TableNodeClock.find (node) == TableNodeClock.end ())
    return pk_false;
  (*h) = TableNodeClock[node];
  return pk_true;
}

/**
   It assigns (localy in the TableNodeClock table) the utility clock (h) to a node (node).
*/
void
UtilityClock::setUtilityClock (ptINode node, ptinsth h)
{
  TableNodeClock[node] = h;
}


/**
   It computes the utility clock of the operand node (node).
   Assumed : the utility clock of the signals used to defined the undersamplings, the delays, and the signals produced by external calls are already computed.
   (for such a signal, the utility clock is equal to its clock (availibility clock)).

   \note not fully implemented
*/
ptinsth UtilityClock::UtilityClockNode (ptINode node)
{
  ptlinsth
    lh;
  ptinsth
    h,
    h1,
    hdep;
  ptldependence
    lsucc;
  ptdependence
    aDep;
  ptINode
    n1,
    n2;

  if (!getUtilityClock (node, &h))
    {
      if (NM->isMarked (node, aMark))
	h = CKM->new_variFormulaRootClock ();	// cycle.
      else
	{
	  NM->addMark (node, aMark);
	  // UtilityClock(x)= clock(x), for x an output.
	  if (NM->is_writeNode (node) || NM->isProcessCall (node))
	    h = NM->get_nodeClock (node);
	  else
	    {
	      lsucc = NM->get_successors (node);
	      h = the_nullClock_UC_copy;
	      while (lsucc->get_next (&aDep, &lsucc))
		{
		  hdep = aDep->getDependence (&n1, &n2);
		  if (isSelected (n2))
		    h1 = UtilityClockNode (n2);
		  else		// Outside the subgraph.
		  if (!NM->getUtilityClock (n2, &h1))
		    h1 = NM->get_nodeClock (n2);

		  h = CKM->makeSup (h, CKM->makeInf (CKM->getRepresentativeFormula (hdep), h1));
		}
	      NM->removeMark (node, aMark);
	      if (getUtilityClock (node, &h1))
		h = CKM->makeSynchro (h, h1);
	    }
	}
      setUtilityClock (node, h);
    }
  return h;
}

/**
   UtilityClock(x)= availibitlity clock(x), for x a signal referenced in an undersampling (when f(x)).
*/
void
UtilityClock::utilityClockUndersamplings (ptlinsth lh)
{
  ptINode node;
  ptinsth vh;
  ptlinstances lsig;
  ptexpcal aSig;

  while (ptlinsth_get_next (&lh, &vh))
    if (CKM->getOperator (vh) == sub)
      {
	lsig = CKM->getSignalsOfCondition (vh);
	utilityClockEqAvailClock (lsig);
      }
}

/*
  For each signal of the operand list (ls), the utility clock is set to its availability clock.
*/
void
UtilityClock::utilityClockEqAvailClock (ptlinstances ls)
{
  ptexpcal s;

  while (ptlinstances_get_next (&ls, &s))
    setUtilityClock (EM->getNode (s), EM->getClock (s));
}


/**
   UtilityClock(x) = availability clock(x), for x a signal used in an external node, except the functions.
*/
void
UtilityClock::utilityClockExternalNodes ()
{
  ptlINode ln;
  ptINode node;

  if (NM->getProcCalls (&ln))
    while (ptlINode_get_elem (&ln, &node))
      if (isSelected (node))
	{
	  utilityClockEqAvailClock (NM->getInodelInputFlows (node));
	  setUtilityClock (node, NM->get_nodeClock (node));
	}
  if (NM->getNodeCalls (&ln))
    while (ptlINode_get_elem (&ln, &node))
      if (isSelected (node))
	{
	  utilityClockEqAvailClock (NM->getInodelInputFlows (node));
	  setUtilityClock (node, NM->get_nodeClock (node));
	}
}


/**
   It computes the utility clock of a list of nodes(ln).
   \see UtilityClockNode(ptINode)
*/
void
UtilityClock::utilityClockNodes (ptlINode ln)
{
  ptINode node;

  while (ptlINode_get_elem (&ln, &node))
    if (isSelected (node))
      UtilityClockNode (node);
}

/**
   The utility clock of a state variable is set to its clock.
   (Similar to the delay nodes).
*/
void
UtilityClock::utilityClockStateVar ()
{
  ptlinstances lident;
  ptexpcal aSig;
  ptINode node;

  if (EM->getExprList (opident, &lident))
    while (ptlinstances_get_next (&lident, &aSig))
      {
	node = EM->getNode (aSig);
	if (isSelected (node))
	  if (EM->IsSignal (aSig) && EM->isStateVar (aSig))
	    setUtilityClock (node, EM->getClock (aSig));
      }
}


/**
   The utility clock of the delayed signals (delay, cell) is the availability clock of the signal.
*/
void
UtilityClock::utilityClockDelayedSignals ()
{
  ptINode node;
  ptexpcal resret, sigret, vret, vfen, vinit;
  ptlINode ln;

  if (NM->getDelays (&ln))
    while (ptlINode_get_elem (&ln, &node))
      if (isSelected (node))
	{
	  NM->get_delay_arg (node, &resret, &sigret, &vret, &vfen, &vinit);
	  if (EM->IsSignal (sigret))
	    setUtilityClock (EM->getNode (sigret), EM->getClock (sigret));
	}
  if (NM->getEquations (&ln))
    while (ptlINode_get_elem (&ln, &node))
      if (isSelected (node))
	if (NM->is_cellNode (node))
	  setUtilityClock (node, NM->get_nodeClock (node));
}

/**
   It computes the utility clock of the selected signals of the selected graph.
*/
void
UtilityClock::go ()
{
  ptlINode lnodes;

  selectObjects ();
  utilityClockDelayedSignals ();
  utilityClockStateVar ();
  utilityClockExternalNodes ();
  utilityClockUndersamplings (CKM->getlhroots ());
  utilityClockUndersamplings (CKM->getlhplaced ());
  utilityClockNodes (lSelectedNodes);
  finalize ();
}


/**
   End of functionality:
   - The utility clock is associated to the selected signals (transfer from the local table  into the internal structure of the graph).
   - The graph/class are then cleaned.
   \see {storeUtilityClock()}
*/
void
UtilityClock::finalize ()
{
  storeUtilityClock ();
  if (isApplyOnSubGraph ())
    NM->resetMarkAll (SelectedObjectMark);
  NM->resetMarkAll (aMark);
  TableNodeClock.clear ();
  ptlINode_free (lSelectedNodes);
}

/**
   Marks, and recursively for a composed node(array, partial definition, subgraph), by SelectedObjectMark, the operand node (node).
*/
void
UtilityClock::markSubGraphObject (ptINode node)
{
  ptlINode lnodes;
  ptlinsth lh;
  boolean b;

  if (!NM->isMarked (node, SelectedObjectMark))
    {
      NM->addMark (node, SelectedObjectMark);
      b = pk_true;
      if (NM->is_setOfPartialDefinitions (node))
	lnodes = NM->get_setOfPartialDefinitions_arg (node);
      else if (NM->is_subgraph (node))
	NM->get_subGraphDefinitions (node, &lnodes, &lh);
      else if (NM->is_array (node))
	NM->get_array_bodyDefinitions (node, &lnodes, &lh);
      else
	b = pk_false;
      if (b)
	markObjects (lnodes);
    }
}

/**
   Marks the objects of the operand list (ln).
*/
void
UtilityClock::markObjects (ptlINode ln)
{
  ptINode node;

  while (ptlINode_get_elem (&ln, &node))
    markSubGraphObject (node);
}

/**
   It defines the list, lSelectedNodes, of the nodes for which the utility clock must be computed.
   When the functionality is applied on a subgraph, the nodes of the subgraphs are marked (markObjects()).
   The mark is used to filtered th nodes of he graph during the computation (isSelected()).
*/
void
UtilityClock::selectObjects ()
{
  ptlINode ln;
  ptlinsth lh;
  ptlinstances lsig;
  ptexpcal vsig;

  lSelectedNodes = NILLINODE;
  if (isApplyOnSubGraph ())
    {
      lsig = NM->getInodelInputFlows (SubGraph);
      while (ptlinstances_get_next (&lsig, &vsig))
	ptlINode_insert_elem (&lSelectedNodes, EM->getNode (vsig));
      NM->get_subGraphDefinitions (SubGraph, &ln, &lh);
      markObjects (ln);
      markObjects (lSelectedNodes);
    }
  else
    {
      if (NM->getEquations (&ln)) 
	lSelectedNodes = ptlINode_Copy (ln);
      if (NM->getGroupsOfPartialEquations (&ln))
	{
	  if (lSelectedNodes == NILLINODE)
	    lSelectedNodes = ptlINode_Copy (ln);
	  else
	    lSelectedNodes = ptlINode_fusion (lSelectedNodes, ln);
	}
    }
}

/**
   It defines the utility clocks of the selected objects.
   The utility clock of a signal is available by calling the ExprManager::getUtilityClock() method
   or the ExprManager::getUtilityClockInSubGraph() when the SubGraphClockUtility() is applied.
*/
void
UtilityClock::storeUtilityClock ()
{
  ptlinstances lsig;
  ptexpcal vsig;
  ptinsth vh;
  TableNodeClockType::iterator it;

  if (isApplyOnSubGraph ())
    {
      lsig = NM->getInodelInputFlows (SubGraph);
      while (ptlinstances_get_next (&lsig, &vsig))
	if (getUtilityClock (EM->getNode (vsig), &vh))
	  NM->setUtilityClockInSubGraph (SubGraph, vsig, vh);
	else
	  printf ("Internal error: Utility clock (for a subgraph) not defined!");
    }
  else
    for (it = TableNodeClock.begin (); it != TableNodeClock.end (); it++)
      NM->setUtilityClock ((*it).first, (*it).second);
}
