/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


_PUBLIC_
/**
   Constructor. The functionality is applied to a graph (vGraph).
*/
  mkaggregationsAndMultiplexings::mkaggregationsAndMultiplexings (ptDCG vGraph)
{
  EM = vGraph->getExpressionManager ();
  CKM = vGraph->getClockManager ();
  NM = vGraph->getNodeManager ();
  AGRAPH = vGraph;
  PARC = new TravelDepManager (vGraph);
  TYPEAGGREG = getMark ();	// temporary attribute
  markedList = NILLINSTANCES;
  vnoag = noag;
  vagid = agid;
  vagwhen = agwhen;
  vagcell = agcell;
  CircuitDetector = (ptCircuitCalculus) NULL;	// used only for DelaysMemoryOptimisation() method
  htop = AGRAPH->getRootClock ();
}

/**
   Destructor.
*/
mkaggregationsAndMultiplexings::~mkaggregationsAndMultiplexings ()
{
}

/**
   @return "true" is the operation succeeds, "false" otherwise.

     Its realizes optimizations related to the memory size using
     - the "when/cell" aggregations: two signals that can be represented in 
     the same area are annotated using the predefined attribute AGGREGATTRIBUTE. 
     -  the multiplexings of variables: signals built on the same signal using the "window"
     and "delay" operators are represented in the same area (circular array managing) if the initial values are compatible. 

   The graph is travelled taking account the dependences (see \ref GraphTravel documentation). 
   The specific operations are defined when a node is selected by the algorithm (graphTravelling()).
   After the travel, the multiplexing of memories is achieved in setlmemtrees() method.

   When MutliplexDelays, it specifies that the not scalar delays of 1 ($1) must be also multiplexed in the same area, otherwise the
   delays are separatly managed. This option is available in MULTIPLEXINGDELAYS attribute member.

   \note Assumed:
     - the graph is \e free cycle.
     - The multiplexing assumes that the trees of delays is build (mkDelayTree() method has been previously applied).

*/
boolean mkaggregationsAndMultiplexings::aggregationsAndMultiplexing (boolean MutliplexDelays)
{
  boolean
    status;

  if (CKM->isAnEmptySystem ())
    return pk_true;
  MULTIPLEXINGDELAYS = MutliplexDelays;
  EM->resetMarkAll (AGGREGATTRIBUTE);
  initfunctionality ();
  parcoursinstrhinit ();
  status = PARC->initTravelForClock (htop);
  if (status)
    graphTravelling (NILINODE);
  endfunctionality (htop);
  if (status)
    setlmemtrees ();
  EM->resetMark (markedList, TYPEAGGREG);
  return status;
}


/**
   Here is defined a functionality applied to delay operators.

   It tries to merge x and zx:=x$1 in the same memory by reinforcing the dependences
   (a node that references zx becomes a predecessor of the node of x).
   If such operations don't add a circuit in the graph, the signal x and zx can be merge into the same memory.
   Such an optimisation is specified by some global attributes:
   - zx and node(zx) are marked OPTIMIZED_MEMORY_ATTRIBUTE.
   - node(x) is marked BELONG_MEMORY.

   This method applies for each delay of the program this optimisation:
   if it satisfies some conditions (isSingleMemtree()) and if the operation delayMemoryOptimisation() succeeds, 
   the tree of delays is updated (memtreeRemoveScalarDelayOne()).

*/
void
mkaggregationsAndMultiplexings::delaysMemoryOptimisation (boolean bclu)
{
  ptlmemnode vlmemt, lmemtrees;
  ptmemnode vmemtree;
  ptlINode lNodesInitClk;

  lmemtrees = AGRAPH->getDelaysTree ();
  // lNodesInitClk = CKM->getNodeReferencesTo(CKM->getInitClock());
  AGRAPH->getNodesAtInitClock (&lNodesInitClk);
  vlmemt = lmemtrees;
  while (ptlmemnode_getNaryArgs (&vlmemt, &vmemtree))
    if (isSingleMemtree (vmemtree, lmemtrees))
      if (delayMemoryOptimisation (vmemtree, lNodesInitClk, bclu))
	{
	  lmemtrees = memtreeRemoveScalarDelayOne (vmemtree, lmemtrees);
	  AGRAPH->setDelaysTree (lmemtrees);	// must be done here : used by the cycle detection algorithm
	}
  AGRAPH->setDelaysTree (lmemtrees);

}
