/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
#define isEventClockedFormat(t) ((t == DCG_BASIC) || (t == DCG_POLY) || (t == DCG_ENDO))

/**
   @defgroup ColouringClusterGrp Clusters: methods useful for the colouring step.
   
*/
/**
    @ingroup ColouringClusterGrp
    @return "true" if the mark m has been inserted to the list of marks associated with a clock (clk), "false" otherwise.
   The list of marks is managed as an attribute (ClusterAttribute) of the clock for the partionning.
*/
boolean Clusters::ptinsthAttr_append_mark(ptinsth clk, typeMark m)
{
  ptlMark lm;
  boolean b;

  if (!CKM->getAttribute(clk, ClusterAttribute, (attributeValue *) & lm))
    lm = NILLMARK;
  b = ptlMark_insert_elem(&lm, m);
  if (b)
    CKM->addAttribute(clk, ClusterAttribute, lm);
  return b;
}

/**
  @ingroup ColouringClusterGrp
  It removes a mark (m) from the list of marks associated with a clock (clk)
   The list of marks is managed as an attribute (ClusterAttribute) of the clock for the partionning.
*/
void Clusters::ptinsthAttr_removeMark(ptinsth clk, typeMark m)
{
  ptlMark lm;

  if (CKM->getAttribute(clk, ClusterAttribute, (attributeValue *) & lm))
    if (ptlMark_removeMark(&lm, m))
      CKM->addAttribute(clk, ClusterAttribute, lm);
}

/**
    @ingroup ColouringClusterGrp
    @return "true" if the mark (m) is an element of the list of marks associated with a clock (clk), "false" otherwise.
   The list of marks is managed as an attribute (ClusterAttribute) of the clock for the partionning.
*/
boolean Clusters::ptinsthAttr_satisfies_marked(ptinsth clk, typeMark m)
{
  ptlMark lm;

  if (!CKM->getAttribute(clk, ClusterAttribute, (attributeValue *) & lm))
    return pk_false;
  return ptlMark_IsIn(lm, m);
}

/**
   @ingroup ColouringClusterGrp
   @return the list of marks associated with a clock (clk).
   The list of marks is managed as an attribute (ClusterAttribute) of the clock for the partionning.
*/
ptlMark Clusters::ptinsthAttr_get_marks(ptinsth clk)
{
  ptlMark lm;
  if (!CKM->getAttribute(clk, ClusterAttribute, (attributeValue *) & lm))
    return NILLMARK;
  return lm;
}

/**
   @ingroup ColouringClusterGrp
   The list of marks (lm) is associated with a node (aCluster).
   The list of marks is managed as an attribute (ClusterAttribute) of the node for the partionning.
*/
void Clusters::ptINodeAttr_set_marks(ptINode aCluster, ptlMark lm)
{
  NM->addAttribute(aCluster, ClusterAttribute, (attributeValue) lm);
}

/**
    @ingroup ColouringClusterGrp
    A mark (m) is added to the list of marks associated with a node (node).
   The list of marks is managed as an attribute (ClusterAttribute) of the node for the partionning.
*/
boolean Clusters::ptINodeAttr_append_mark(ptINode node, typeMark m)
{
  ptlMark lm;
  lm = ptINodeAttr_get_marks(node);
  if (!ptlMark_insert_elem(&lm, m))
    return pk_false;
  ptINodeAttr_set_marks(node, lm);
  return pk_true;
}

/**
   @ingroup ColouringClusterGrp
   @return "true" if a node (node) is marked with the mark (m) for the partionning.
*/
boolean Clusters::ptINodeAttr_satisfies_marked(ptINode node, typeMark m)
{
  ptlMark lm;
  lm = ptINodeAttr_get_marks(node);
  return ptlMark_IsIn(lm, m);
}

/**
   @ingroup ColouringClusterGrp
   @return the list of the marks associated with a node (aCluster) for the partitionning.
*/
ptlMark Clusters::ptINodeAttr_get_marks(ptINode aCluster)
{
  ptlMark lm;
  if (!NM->getAttribute(aCluster, ClusterAttribute, (attributeValue *) & lm))
    return NILLMARK;
  return lm;
}

/**
   @ingroup ColouringClusterGrp
   It updates the firstMark, maxMark values (attributes member of the class). This values can be used for 
   some partitionning operations (such that boolean/other partitionning, states/other partitionning)
   to allocate vectors of sub-graph.
*/
_VIRTUAL_ void Clusters::setMinMaxMark(typeMark Mark)
{
  if (Mark > maxMark)
    maxMark = Mark;
  if (firstMark < 0)
    firstMark = Mark;
  else if (Mark < firstMark)
    firstMark = Mark;
}

/**
   @ingroup ColouringClusterGrp
   @return a new mark by calling the mark server.
   setMinMaxMark() is called to update internal values.
 */
_VIRTUAL_ typeMark Clusters::getNewMark()
{
  typeMark amark;

  amark = getMark();
  setMinMaxMark(amark);
  return amark;
}

/**
   @ingroup ColouringClusterGrp
   It cleans the objects of the graph by removing the attributes/marks referencing 
   the internal marks.
*/
_VIRTUAL_ void Clusters::resetmarks()
{
  integer i;

  for (i = firstMark; i <= maxMark; i++)
    {
      NM->resetMarkAll(i);
      CKM->resetMarkAll(i);
    }
}

/**
   @defgroup partitionningGrp: useful methods to define the partitionning for sets of objects (nodes, clocks,...) or objects (clock, node) of the 
   graph.   
*/

/**
  @ingroup partitionningGrp
  It defines, for a partitionning functionality, the sub-graph of each element of a list of nodes (lNodes).
   \see assignNode(ptINode).
*/
_VIRTUAL_ void Clusters::assignNodes(ptlINode lNodes)
{
  ptINode cn, aNode;
  while (ptlINode_get_elem(&lNodes, &aNode))
    if (!NM->getInArrayNode(aNode, &cn))
      assignNode(aNode);
}

/**
   @ingroup partitionningGrp
   It defines, for a partitionning functionality, the sub-graph for the composed nodes (list of partial definition, list of fields, sub-graph).
   \see assignNode(ptINode).
*/
_VIRTUAL_ void Clusters::assignComposedNodes()
{
  ptlINode lNodes, ln;
  ptINode aNode;

  if (NM->getGroupsOfPartialEquations(&lNodes))
    {
      assignNodes(lNodes);
      while (ptlINode_get_elem(&lNodes, &aNode))
	{
	  ln = NM->get_setOfPartialDefinitions_arg(aNode);	// nodes of ln must be in same partition!!!
	  assignNodes(ln);
	}
    }
  if (NM->getBundles(&lNodes))
    assignNodes(lNodes);
  if (NM->getSubGraphs(&lNodes))
    assignNodes(lNodes);
  if (NM->getIterates(&lNodes))
    assignNodes(lNodes);
}

/**
   @ingroup partitionningGrp
   It defines, for a partitionning functionality, the sub-graph of all the nodes of the graph.
  \see assignNodes(), assignComposedNodes().
*/
_VIRTUAL_ void Clusters::assignAllNodes()
{
  ptlINode lNodes;

  if (NM->getEquations(&lNodes))
    assignNodes(lNodes);
  if (NM->getDelays(&lNodes))
    assignNodes(lNodes);
  if (NM->getProcCalls(&lNodes))
    assignNodes(lNodes);
  if (NM->getNodeCalls(&lNodes))
    assignNodes(lNodes);
  assignComposedNodes();
}

/**
    @ingroup partitionningGrp
    It defines, for a partitionning functionality, the sub-graph of the clocks of a list (lh).
*/
_VIRTUAL_ void Clusters::assignClocks(ptlinsth lh)
{
  ptinsth vh;

  while (ptlinsth_get_next(&lh, &vh))
    assignClock(vh);
}

/**
   @ingroup partitionningGrp
   It defines, for a partitionning functionality, the sub-graph of all the clocks of the graph.
   <em> Without effect here, must be redefined in derived classes.</em>
*/
_VIRTUAL_ void Clusters::assignAllClocks()
{
}

/**
   @ingroup partitionningGrp
   It defines, for a partitionning functionality, the sub-graph of a node (aNode).
   It uses the list of marks, associated with the node, at the colouring step.
   \see getSetSubGraphOf().
*/
_VIRTUAL_ void Clusters::assignNode(ptINode aNode)
{
  ptlMark lm;
  ptINode aSubGraph;
  ptlinstances vlsort;
  ptexpcal sig;
  noperators opNode;

  lm = ptINodeAttr_get_marks(aNode);
  if (getSetSubGraphOf(lm, &aSubGraph))
    {
      NM->subGraphInsertNode(aSubGraph, aNode);
      opNode = NM->get_oper(aNode);
      switch ((integer) opNode)
	{
	case opnodecall:
	  vlsort = NM->getInodelOutputFlows(aNode);
	  while (ptlinstances_get_next(&vlsort, &sig))
	    assignNode(EM->getNode(sig));
	  break;
	case opldefpart:	// set of ::= node
	  assignNodes(NM->get_setOfPartialDefinitions_arg(aNode));
	  break;
	}
    }
}

/**
    @ingroup partitionningGrp
    It defines, for a partitionning functionality, the sub-graph of a clock (vh).
    It uses the list of marks, associated with the clock, at the colouring step.
    \see getSetSubGraphOf()
*/
_VIRTUAL_ void Clusters::assignClock(ptinsth vh)
{
  ptlMark lm;
  ptINode aSubGraph, aNode;
  ptexpcal exp;

  if (isEventClockedFormat(graph_type))
    lm = ptinsthAttr_get_marks(vh);
  else
    {
      exp = CKM->getConditionSub(vh);
      aNode = EM->getNode(exp);
      lm = ptINodeAttr_get_marks(aNode);
    }
  if (getSetSubGraphOf(lm, &aSubGraph))
    NM->subGraphInsertClock(aSubGraph, vh);
}

/**
   @ingroup partitionningGrp
   @return a sub-graph node in which all the nodes/clocks with lm as associated list of marks will be inserted.
   The sub-graph is added to the list of sug-graphs lSubGraph (attribute member of the class).
*/
_VIRTUAL_ ptINode Clusters::mkNodeCluster(ptlMark lm)
{
  ptINode aCluster;
  ptexpcal vexpcal;

  vexpcal = EM->mkLabelIdent(getLabelId(), NM->mk_node_undefined(), CKM->getUndefClock(), NILBACKANNOTATION, pk_true);
  aCluster = NM->mk_node_subgraph(vexpcal, NILBACKANNOTATION);
  NM->set_trainClusterTag(aCluster);
  ptINodeAttr_set_marks(aCluster, lm);
  ptlINode_insert_elem(&lSubGraph, aCluster);
  return aCluster;
}

/**
   @return "true" if it exists, in the list of clusters (global list lSubGraph)
   a cluster (aCluster) characterized by a list of marks (lm),
   "false" otherwise.
*/
_VIRTUAL_ boolean Clusters::getSubGraphOf(ptlMark lm, ptINode * aCluster)
{
  ptlINode ln;
  ptlMark lmCluster;

  ln = lSubGraph;
  while (ptlINode_get_elem(&ln, aCluster))
    {
      lmCluster = ptINodeAttr_get_marks(*aCluster);
      if (ptlMark_satisfies_equal(lm, lmCluster))
	return pk_true;
    }
  return pk_false;
}

/**
   @ingroup partitionningGrp
   @return "true" if it exists a sub-graph (sg) with lm as associated list of marks, "false" otherwise.
   It searches in the list of sug-graphs lSubGraph (attribute member of the class).
*/
_VIRTUAL_ boolean Clusters::getSetSubGraphOf(ptlMark lm, ptINode * sg)
{

  if (ptlMark_satifies_isEmpty(lm))
    return pk_false;		// possible pour les racines du calcul (selectedNodes, selectedClocks)
  if (!getSubGraphOf(lm, sg))
    *sg = mkNodeCluster(lm);
  return pk_true;
}

/**
   @ingroup partitionningGrp
   @return "true" if the sub-graph (sg) of a node (aNode) for the current partitionning operation is defined, "false" otherwise.
*/
_VIRTUAL_ boolean Clusters::nodeGetSubGraph(ptINode aNode, ptINode * sg)
{
  ptlMark lm;

  lm = ptINodeAttr_get_marks(aNode);	/* lieu de definition */
  return getSetSubGraphOf(lm, sg);
}

/**
    @ingroup partitionningGrp
    @return "true" if the sub-graph (sg) of a clock (vh) for the current partitionning operation is defined, "false" otherwise.
*/
_VIRTUAL_ boolean Clusters::clockGetSubGraph(ptinsth vh, ptINode * sg)
{
  ptlMark lm;

  lm = ptinsthAttr_get_marks(vh);	/* lieu de definition */
  return getSetSubGraphOf(lm, sg);
}

/**
   @defgroup postPartitionningGrp Clusters: useful methods after the partitionning (colouring + partition definition) step.
*/

/** It imports the clocks referenced in the assertions of a subgraph (aSubGraph). */
_VIRTUAL_ void Clusters::importAssertions(ptINode aSubGraph)
{
}

/**
   @defgroup postPartitionningGrp Clusters: useful methods after the partitionning (colouring + partition definition) step.
*/

/**
   @ingroup postPartitionningGrp
   It computes the interface (inputs, outputs,...) of a sub-graph (aSubGraph).
   The outputs of the graph are, here, only the outputs of the graph.
   InterfaceCalculusSubGraphsUpdate() can be called to update the interface.
*/
_VIRTUAL_ void Clusters::InterfaceCalculus(ptINode aSubGraph)
{
  importAssertions(aSubGraph);
  NM->moveOutputs2localsExcept(aSubGraph, AGRAPH->getOutputs(), NILLINSTH);
}

/**
   @ingroup postPartitionningGrp
   If the graph is a bDC+ one, it applies bdcplusSubGraphInterfaceCalculus() 
   to each sub-graph of lSubGraph (attribute member of the class): the referenced to clocks are replaced by the reference to booleans signals.
*/
_VIRTUAL_ void Clusters::bdcplusInterfaceCalculusSubGraphs()
{
  ptINode aSubGraph;
  ptlINode ln;

  if (bDCplusFormat)
    {
      ln = lSubGraph;
      while (ptlINode_get_elem(&ln, &aSubGraph))
	NM->SubGraphInterfaceCalculusUndersamplingClocks(aSubGraph);
    }
}

/**
   @ingroup postPartitionningGrp
   For each sub-graph, it updates its interface taking into account the interface of the other sub-graphs
   of the partitionning operation.
*/
_VIRTUAL_ void Clusters::InterfaceCalculusSubGraphsUpdate()
{
  ptlINode ln, ln1;
  ptINode aSubGraph, aSubGraph_i;

  ln = lSubGraph;
  while (ptlINode_get_elem(&ln, &aSubGraph))
    {
      ln1 = ln;
      while (ptlINode_get_elem(&ln1, &aSubGraph_i))
	{
	  NM->updateInterface(aSubGraph, aSubGraph_i);
	}
    }
}

/**
   @ingroup postPartitionningGrp
   It calls InterfaceCalculus() followed by InterfaceCalculusSubGraphsUpdate().
*/
_VIRTUAL_ void Clusters::InterfaceCalculusSubGraphs()
{
  ptlINode ln;
  ptINode aSubGraph;

  ln = lSubGraph;
  while (ptlINode_get_elem(&ln, &aSubGraph))
    InterfaceCalculus(aSubGraph);
  InterfaceCalculusSubGraphsUpdate();
}

/**
   @ingroup postPartitionningGrp
   It completes the interface of a sub-graph (SubGraph) by computing the branching clock of all the clocks/clocks of the 
   nodes/clocks of the input/output signals of the sub-graph.
   It calls ClockManager::CompleteUpstreamParametrization().
*/
void Clusters::IGraphUpstreamProjectionSubGraph(ptINode SubGraph)
{
  ptlinsth lhtodeclare;
  ptinsth vhemb;
  ptinterface vinterface;

  NM->mkSubGraphProjectedSystemInfos(SubGraph);
  NM->get_subGraphProjectedSystemInfos(SubGraph, &vhemb, &lhtodeclare);
  vinterface = NM->getSet_interface(SubGraph);
  // The root is not an input of the sub-graph.
  if (!isEventClockedFormat(AGRAPH->getFormatLevel()))
    {
      ptinterface_remove_inputClock(vinterface, vhemb);
      ptinterface_remove_inputSignal(vinterface, CKM->getBooleanSignal(vhemb));
    }
}

/**
   @ingroup postPartitionningGrp
   It applies
   - InterfaceCalculusSubGraphs() method to build the interface of the produced partitions (sub-graphs)
   - IGraphUpstreamProjectionSubGraph() method to produce for each subgraph, the projection informations
   - Clusters::bdcplusInterfaceCalculusSubGraphs() method to for bDC+ graph interface 
   - Clusters::InterfaceCalculusSubGraphsUpdate() method to update the interfaces of the sub-graphs..
*/
_VIRTUAL_ void Clusters::subGraphsBuildProfile()
{
  ptINode aSubGraph;
  ptlINode ln;

  InterfaceCalculusSubGraphs();
  ln = lSubGraph;
  while (ptlINode_get_elem(&ln, &aSubGraph))
    IGraphUpstreamProjectionSubGraph(aSubGraph);
  bdcplusInterfaceCalculusSubGraphs();
  InterfaceCalculusSubGraphsUpdate();
}

/**
  @ingroup postPartitionningGrp
  For each sub graph of the partitionning (element of the global list lSubGraph):
   - It builds its interface
   - it inserts the sub-graph at the branching clock of the nodes of the sub-graph (ie the information must be available)
   - the branching clock becomes the clock of the label associated with the sub-graph.
*/
_VIRTUAL_ void Clusters::insertSubGraphs()
{
  ptlINode lsg;
  ptINode aSubGraph;
  ptlinsth lh;
  ptinsth vhemb;
  ptexpcal vlab;

  lsg = lSubGraph;
  while (ptlINode_get_elem(&lsg, &aSubGraph))
    {
      NM->getSet_interface(aSubGraph);
      NM->get_subGraphProjectedSystemInfos(aSubGraph, &vhemb, &lh);
      if (!CKM->satisfiesUndefClock(vhemb))
	{
	  NM->insertNode(aSubGraph, vhemb);
	  if (NM->get_label(aSubGraph, &vlab))
	    EM->setClock(vlab, vhemb);
	}
    }
}

/**
   @defgroup debuggingClusterGrp Clusters: debugging methods
*/
/**
   @ingroup debuggingClusterGrp
   In debug mode (BDEBUG attribute member), it prints a message (message) 
   followed by informations about a clock (clk).
*/
_VIRTUAL_ void Clusters::clockTracing(const char *message, ptinsth clk)
{
  ptident id;
  if (BDEBUG)
    {
      pK_messenger->printString(message);
      id = CKM->getName(clk);
      // pK_messenger->printString(repstr2string( strmakefromid(id)));
    }
}

/**
   Called at the beginning of a partitionning functionality.
*/
_VIRTUAL_ void Clusters::partitionningInitialize()
{
}

/** 
    It cleans the graph by removing the reference to the internal used marks.
*/
void Clusters::graphCleaning()
{
  resetmarks();
}

/**
   Called at the end of a partitionning functionality:
   - the nodes of each subgraph are unlinked to their clock and linked to their partition.
   - the graph is cleaned.
*/
_VIRTUAL_ void Clusters::partitionningFinalize()
{
  ptINode aSubGraph;
  ptlINode lsg;

  lsg = lSubGraph;
  while (ptlINode_get_elem(&lsg, &aSubGraph))
    NM->unlinkNodesFromClock(aSubGraph);
  lsg = lSubGraph;
  while (ptlINode_get_elem(&lsg, &aSubGraph))
    NM->subGraphLinkNodes(aSubGraph, pk_false);
  graphCleaning();
}

void Clusters::attributesSetting()
{
  ptpragma vp;
  ptINode aNode;
  ptlINode lsg;

  if (existPredefinedPragma(AGRAPH->getPragmas(), PRAGMA_DEFINED_CLOCK_HIERARCHY))
    {
      lsg = lSubGraph;
      vp = newPredefinedPragma(PRAGMA_DEFINED_CLOCK_HIERARCHY, NILOBJPRAGMA, NILTREE);
      while (ptlINode_get_elem(&lsg, &aNode))
	NM->addPragma(aNode, vp);	// OK because no objects assigned with!.
    }
}

/**
   It keeps the link node->mark to the markOfSelectedNode table
*/
void Clusters::addTable(ptINode aNode, typeMark aMark)
{
  markOfSelectedNode[aMark] = aNode;
}

/**
   @return "true" if it exists a node assigned with the operand mark (aMark), "false" otherwise.
*/
boolean Clusters::getTableNode(typeMark aMark, ptINode * node)
{
  (*node) = markOfSelectedNode[aMark];
  return ((*node) != NILINODE);
}
