/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   Specialisation of the "abstract" method declared in the separate_selected_others::isSelectedNode().
   @return "true" if a node (aNode) defines a state variable (delay cell) of the program, "false" otherwise.
*/
    _VIRTUAL_ boolean mkStatesOther::isSelectedNode(ptINode aNode)
{
  return pk_true;
}

/**
   @return true if the operand expression (exp) is a boolean one, false otherwise.
*/
_VIRTUAL_ boolean mkStatesOther::isSelectedType(ptexpcal vexp)
{
  return EM->IsBoolean(vexp);
}

/**
   Let A = the set of the boolean delay signals of the program.
   Then for an element e of A, the boolean signals referenced in the equation of e are added to A.

   The colouring operation marks the nodes defining the signals of A.
*/
_VIRTUAL_ boolean mkStatesOther::colouring()
{
  ptexpcal vexp;
  ptINode n;
  ptlINode lNodes;

  initcolouring();

  if (!NM->getDelays(&lNodes))
    lNodes = NILLINODE;
  // Selecting the boolean ones.
  while (ptlINode_get_elem(&lNodes, &n))
    {
      vexp = NM->get_leftOperand(n);
      if (isSelectedType(vexp))
	colouringNode(n);
    }
  // Marks the unmarked nodes to be the "static" part.(the no selected delays are also in this part).
  if (NM->getEquations(&lNodes))
    colouringNodesOtherMark(lNodes);
  if (NM->getDelays(&lNodes))
    colouringNodesOtherMark(lNodes);
  if (NM->getProcCalls(&lNodes))
    colouringNodesOtherMark(lNodes);
  if (NM->getNodeCalls(&lNodes))
    colouringNodesOtherMark(lNodes);
  if (NM->getGroupsOfPartialEquations(&lNodes))
    colouringNodesOtherMark(lNodes);
  if (NM->getIterates(&lNodes))
    colouringNodesOtherMark(lNodes);
  return pk_true;
}

/**
   Marks the boolean signals of a list (lsig).
*/
void mkStatesOther::marklsig(ptlinstances lsig)
{
  ptexpcal vsig;

  while (ptlinstances_get_next(&lsig, &vsig))
    if (isSelectedType(vsig))
      colouringNode(EM->getNode(vsig));
}

/**
   Marks the boolean signals referenced in a formula (formula).
*/
void mkStatesOther::markSignalsReferencedInClock(ptinsth formula)
{
  ptinsth h1, h2;

  if (CKM->satisfiesNullClock(formula))
    return;
  if (CKM->satisfiesUndefClock(formula))
    return;
  switch (CKM->getOperator(formula))
    {
    case nosub:
    case sub:
      marklsig(CKM->getSignalsOfCondition(formula));
      break;
    case infer:
      if (!CKM->getMinusArgs(formula, &h1, &h2))
	CKM->getBinaryArgs(formula, &h1, &h2);
      markSignalsReferencedInClock(h1);
      markSignalsReferencedInClock(h2);
      break;
    case super:
      CKM->getBinaryArgs(formula, &h1, &h2);
      markSignalsReferencedInClock(h1);
      markSignalsReferencedInClock(h2);
      break;

    case hcomp:
    case hnot:
      CKM->getUnaryArg(formula, &h1);
      markSignalsReferencedInClock(h1);
      break;
    case vari:
    case predef:
      break;
    default:
      pK_messenger->Caseerror(currentLine, currentFile);
    }
}

/**
   The operand node (aNode) is marked as an element of the "States" part
   Then, the signals referenced in the definition of this node are also marked as element of this part.
*/
_VIRTUAL_ void mkStatesOther::colouringNode(ptINode aNode)
{
  ptlinstances ls;
  ptexpcal vsig;
  ptinterface vinterface;
  ptlinsth lh;
  ptinsth vh;

  if (!ptINodeAttr_satisfies_marked(aNode, SelectedMark))
    {
      ptINodeAttr_append_mark(aNode, SelectedMark);
      vinterface = NM->get_interface(aNode);
      ls = ptinterface_get_inputSignals(vinterface);
      while (ptlinstances_get_next(&ls, &vsig))
	colouringNode(EM->getNode(vsig));
      lh = ptinterface_get_inputClocks(vinterface);
      while (ptlinsth_get_next(&lh, &vh))
	markSignalsReferencedInClock(vh);
    }
}

/**
   An unmarked node by SelectedMark mark is marked as an element of the "otherpart" (OtherMark mark).
*/
_VIRTUAL_ void mkStatesOther::colouringNodesOtherMark(ptlINode lNodes)
{
  ptINode n;
  while (ptlINode_get_elem(&lNodes, &n))
    {
      if (!ptINodeAttr_satisfies_marked(n, SelectedMark))
	ptINodeAttr_append_mark(n, OtherMark);
    }
}
