/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   Specifies that the dependence (aDep) is inserted in the exported Signal tree model.
*/
    void
 DCG2SignalTree::setVisited(ptdependence aDep)
{
  printedDependences.insert(aDep);
}

/**
   @return "true" if the dependence (aDep) has been already inserted in the Signal tree model, "false" otherwise.
*/
boolean DCG2SignalTree::isVisited(ptdependence aDep)
{
  return (printedDependences.find(aDep) != printedDependences.end());
}

/**
   @return NILTREE if the clock h is the clock of the signal produced by the node vFrom or vTo, the name of the clock otherwise.
 */
ntrees DCG2SignalTree::mkTreeClockDependence(ptinsth h, ptINode vFrom, ptINode vTo)
{
  ptexpcal sig;
  if (!bDependenceWithClock)
    return NILTREE;
  if (CKM->satisfiesUndefClock(h))
    return NILTREE;
  if (vFrom != NILINODE)
    {
      if (NM->getDependenceNodeLabel(vFrom, &sig))
	if (CKM->eqClock(h, EM->getClock(sig)))
	  return NILTREE;
    }
  if (vTo != NILINODE)
    {
      if (NM->getDependenceNodeLabel(vTo, &sig))
	if (CKM->eqClock(h, EM->getClock(sig)))
	  return NILTREE;
    }
  return getClockExprInWhenCell(h);	// getClockExpr0(h);
}

/** 
    @return the SIGNAL Abstract Tree representation of a dependence (v1 -> v2) when vform.
*/
ntrees DCG2SignalTree::mkdependtree(ntrees v1, ntrees v2, ptinsth vform)
{
  return (STree_mkDependenceTree(STree_scopy(v1), STree_scopy(v2), mkTreeClockDependence(vform, NILINODE, NILINODE)));
}

/**
   @return the Signal abstract Tree of the dependence "vFrom->vTo when h".
*/
ntrees DCG2SignalTree::mkTreeDependence(ptINode vFrom, ptINode vTo, ptinsth h)
{
  return STree_mkDependenceTree(mkNameNode(vFrom), mkNameNode(vTo), mkTreeClockDependence(h, vFrom, vTo));
}

/** 
    @return "true" if for a dependence (aDep), the source and the target are nodes of the list of nodes (ln), "false" otherwise.
    @param vt (out) the SIGNAL Abstract Tree representation of a dependence (aDep) when it retruns pk_true, otherwise it is not set.
*/
boolean DCG2SignalTree::mkTreeDependence(ptdependence aDep, ptlINode ln, ntrees * vt)
{
  ptINode vFrom, vTo;
  ptinsth h;

  // if (ln == NILLINODE) return pk_false;
  h = aDep->getDependence(&vFrom, &vTo);
  if (!isSelectedNode(vFrom))
    return pk_false;
  if (!isSelectedNode(vTo))
    return pk_false;
  h = CKM->getRepresentativeFormula(h);
  if (CKM->satisfiesNullClock(h))
    return pk_false;
  if (ln != NILLINODE)
    if (!ptlINode_isIn(ln, vFrom))
      return pk_false;
  if (ln != NILLINODE)
    if (!ptlINode_isIn(ln, vTo))
      return pk_false;

  *vt = mkTreeDependence(vFrom, vTo, h);
  setVisited(aDep);
  return pk_true;
}

/**
   @return "true" if it exists a next dependence (aDepNext: n1 --> n2 when hdep) of a list of dependences (lexplicitDep) such that n1 and n2 are not optimized, "false" otherwise.
*/
boolean DCG2SignalTree::getNextDependence(ptldependence * lexplicitDep, ptdependence * aDepNext)
{
  ptINode vFrom2, vTo2;
  boolean found;

  found = pk_false;
  while (!found && (*lexplicitDep)->get_next(aDepNext, lexplicitDep))
    {
      (*aDepNext)->getDependence(&vFrom2, &vTo2);
      found = isSelectedNode(vFrom2) && isSelectedNode(vTo2);
    }
  return found;
}

/**
   @return "true" if for aDep= n1->n2 when hdep, aDepnext is n1->n3 when hdep, "false" otherwise.
   vToTree (out) is the symbol (an output or a label) associated with the node n3.
*/
boolean DCG2SignalTree::isSimilarDep(ptdependence aDep, ptdependence aDepNext, ntrees * vToTree)
{
  ptINode vFrom1, vTo1, vFrom2, vTo2;
  ptinsth h1, h2;

  h2 = CKM->getRepresentativeFormula(aDepNext->getDependence(&vFrom2, &vTo2));
  h1 = CKM->getRepresentativeFormula(aDep->getDependence(&vFrom1, &vTo1));
  if (vFrom1 != vFrom2)
    return pk_false;
  if (h1 != h2)
    return pk_false;
  setVisited(aDep);
  *vToTree = mkNameNode(vTo2);
  return pk_true;
}

void DCG2SignalTree::addDependences(ntrees vtprocess, ptldependence lDep, integer numMess)
{
  ptdependence aDep;
  ntrees vt, tdependences;

  tdependences = STree_getNilTree();
  while (lDep->get_next(&aDep, &lDep))
    if (!isVisited(aDep))
      {
	SET_NILTREE(vt);
	if (mkTreeDependence(aDep, NILLINODE, &vt))
	  STree_addExprToComposition_Post(&tdependences, vt);
      }
  if (!STree_isSignalNilTree(tdependences))
    {
      if (numMess > 0)
	pK_messenger->messageacc(messReinforcingDep, tdependences);
      STree_addExpressionOfProcessToaModel(vtprocess, tdependences);
    }
}

/**
   It adds the explicit dependences and the reinforcing one of the graph to a model (vtprocess).
*/
void DCG2SignalTree::mkGraphTreeDependences(ntrees vtprocess)
{
  addDependences(vtprocess, AGRAPH->getExplicitDependences(), -1);
  addDependences(vtprocess, AGRAPH->getReinforcedDependences(), messReinforcingDep);
  //  addDependences(vtprocess,  AGRAPH->getDependences(), -1); xxxxxxxxxxxxxx
}

/**
   @return the SIGNAL Abstract Tree representation of some nodes of a cycle (vCycle), composed of
   node objects and clock objects.

   - The returned tree is composed of the list of definitions of the nodes of vCycle not already marked by a mark (vm).
   - The nodes of the cycles are marked by vm. 
   - A name is associated with each node (vm-th attribute). This name will be used later to express the SIGNAL Abstract Tree representation of the cycle.
   
   \note Only node objects are considered, clock objects are ignored.
*/
ntrees DCG2SignalTree::mkDefinitionsReferencedInCycle(pointer vCycle, typeMark vm)
{
  ntrees vt, name, vdef;
  ptlPointer aCycle;
  pointer velem;
  ptObjectCommonInfos x;
  objectType objtype;
  ptINode node;

  SET_NILTREE(vt);
  aCycle = (ptlPointer) vCycle;
  while (ptlPointer_get_elemNext(&aCycle, &velem))
    {
      x = (ptObjectCommonInfos) velem;
      objtype = x->getwhoami();
      if (objtype == nodeObject)
	{
	  node = (ptINode) velem;
	  if (!NM->isMarked(node, vm))
	    {
	      vdef = mkNodeDefinition(node);
	      name = mkNameNode(node);
	      if (IS_NILTREE(name))
		{
		  name = mknumtreewithprefix(NM->get_num(node), "<node>_");
		  vdef = LL_make_petiq_tree(STree_scopy(name), vdef);
		}
	      STree_addExprToComposition_Post(&vt, vdef);
	      NM->addMark(node, vm);
	      NM->addAttribute(node, vm, (attributeValue) name);
	    }
	}
      /* else  ptinsth vclk;
         if (objtype == clockObject)
         {
         vclk = (ptinsth) velem;
         if (!CKM->satisfiesMarked(vclk, vm))
         { STree_addExprToComposition_Post(&vt, clock2Tree(vclk));
         CKM->addMark(vclk, vm);
         }
         } */
    }
  return vt;
}

/**
   @return the name of an object (velem),a node object or a clock object:
   - for a node object, it used the vm-th attribute (assumed) previously set in mkDefinitionsReferencedInCycle().
   - for a clock object, the name of the clock is returned.
*/
ntrees DCG2SignalTree::getName(pointer velem, typeMark vm)
{
  ptObjectCommonInfos x;
  objectType objtype;
  ptINode node;
  ptinsth vclk;
  ntrees name;

  x = (ptObjectCommonInfos) velem;
  objtype = x->getwhoami();
  if (objtype == nodeObject)
    {
      node = (ptINode) velem;
      NM->getAttribute(node, vm, (attributeValue *) & name);
      return STree_scopy(name);
    }
  if (objtype == clockObject)
    {
      vclk = (ptinsth) velem;
      return clock2Tree(vclk);
    }
  return NILTREE;
}

/**
   @return the clock of the dependence prev -> elem.
   prev, elem may be nodes/clocks.
   For clocks, it returns the undefined clock. (why ?)
*/
ptinsth DCG2SignalTree::getDependenceClock(pointer prev, pointer elem)
{
  ptObjectCommonInfos x;
  objectType objtype;
  ptinsth h;
  ptINode n1, n2;
  ptdependence aDep;
  ptldependence li;

  h = CKM->getUndefClock();
  x = (ptObjectCommonInfos) prev;
  objtype = x->getwhoami();
  if (objtype != nodeObject)
    return h;			// (ptinsth) prev;
  n1 = (ptINode) prev;
  x = (ptObjectCommonInfos) elem;
  objtype = x->getwhoami();
  if (objtype != nodeObject)
    return h;
  n2 = (ptINode) elem;
  li = NM->get_successors(n1);
  if (!li->exist_dependence(n1, n2, &aDep))
    return h;
  return aDep->getClockDependence();
}

/**
    @return the Signal Abstract Tree representation (composition) of the cycle (vCycle) using dependence syntax.
    The getName() method provides the name associated with each element of the cycle.
*/
ntrees DCG2SignalTree::enumerateCycle(pointer vCycle, typeMark vm)
{
  ntrees n1, n2, vt;
  ptlPointer aCycle;
  pointer prev, velem;
  ptinsth h;

  aCycle = (ptlPointer) vCycle;
  if (aCycle == NILLPOINTER)
    return NILTREE;
  SET_NILTREE(vt);
  ptlPointer_get_elemNext(&aCycle, &prev);
  n1 = getName(prev, vm);
  while (ptlPointer_get_elemNext(&aCycle, &velem))
    {
      n2 = getName(velem, vm);
      h = getDependenceClock(prev, velem);
      STree_addExprToComposition_Pre(&vt, mkdependtree(n2, n1, h));
      n1 = n2;
      prev = velem;
    }
  return vt;
}

/**
   @return the Signal Abstract Tree representation (a composition) of a list of cycles (cycles) and false-cycles (fcycles)
   For each element of the list, it produces 
   - the definitions of the nodes of the cycle (mkDefinitionsReferencedInCycle()),
   - Signal Abstract Tree representation of the cycle using dependence syntax (enumerateCycle()).
*/
ntrees DCG2SignalTree::buildTreeCycles(ptlPointer cycles, typeMark vm)
{
  ntrees vt;
  pointer velem;
  ptlPointer lcycles;

  SET_NILTREE(vt);
  lcycles = cycles;
  while (ptlPointer_get_elemNext(&lcycles, &velem))
    STree_addExprToComposition_Post(&vt, mkDefinitionsReferencedInCycle(velem, vm));

  lcycles = cycles;
  while (ptlPointer_get_elemNext(&lcycles, &velem))
    STree_addExprToComposition_Post(&vt, enumerateCycle(velem, vm));
  return vt;
}

/**
   @return the Signal Abstract Tree representation (a composition) of a list of cycles (cycles) and false-cycles (fcycles)
   For each element of the list, it produces 
   - the definitions of the nodes of the cycle (mkDefinitionsReferencedInCycle()),
   - Signal Abstract Tree representation of the cycle using dependence syntax (enumerateCycle()).
*/
ntrees DCG2SignalTree::buildTreeCycles(ptlPointer cycles, ptlPointer fcycles)
{
  ntrees vt, vt1, vt2;
  typeMark vm;

  vm = getMark();
  vt = buildTreeCycles(cycles, vm);
  vt1 = buildTreeCycles(fcycles, vm);
  if (IS_NILTREE(vt))
    vt = vt1;
  else if (ISNOT_NILTREE(vt1))
    {
      pK_messenger->warningaccForce(warFalseCycles, vt1);
      SET_NILTREE(vt2);
      STree_addExprToComposition_Post(&vt2, vt);
      STree_addExprToComposition_Post(&vt2, vt1);
      vt = vt2;
    }
  // Cleaning the graph.
  NM->resetMarkAll(vm);
  CKM->resetMarkAll(vm);
  return vt;
}

boolean DCG2SignalTree::flatmodeRequiredForDependencies(ptldependence lDep)
{
  ptdependence aDep;
  ptINode vFrom, vTo;
  ptexpcal sig;

  while (lDep->get_next(&aDep, &lDep))
    {
      aDep->getDependence(&vFrom, &vTo);
      if (isSelectedNode(vFrom) && isSelectedNode(vTo))
	{
	  if (!NM->getDependenceNodeLabel(vFrom, &sig))
	    return pk_true;
	  if (!(EM->IsInputFlow(sig) || EM->IsOutputFlow(sig)))
	    return pk_true;
	  if (!NM->getDependenceNodeLabel(vTo, &sig))
	    return pk_true;
	  if (!(EM->IsInputFlow(sig) || EM->IsOutputFlow(sig)))
	    return pk_true;
	}
    }
  return pk_false;
}

boolean DCG2SignalTree::isSelectedDependenceForAbstraction(ptdependence aDep)
{
  ptINode vFrom, vTo;
  ptexpcal sig;
  ptinsth hdep;

  hdep = aDep->getDependence(&vFrom, &vTo);
  hdep = CKM->getRepresentativeFormula(hdep);
  if (CKM->satisfiesNullClock(hdep))
    return pk_false;
  if (isSelectedNode(vFrom) && isSelectedNode(vTo))
    {
      if (!NM->getDependenceNodeLabel(vFrom, &sig))
	return pk_false;
      if (!(EM->IsInputFlow(sig) || EM->IsOutputFlow(sig)))
	return pk_false;
      if (!NM->getDependenceNodeLabel(vTo, &sig))
	return pk_false;
      if (!(EM->IsInputFlow(sig) || EM->IsOutputFlow(sig)))
	return pk_false;
    }
  return pk_true;
}
