/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/** 
    @return true if a clock (vh) is a selected clock for the operation.
*/
    boolean DCG2SignalTree::isSelectedClock(ptinsth vh)
{
  if (ptlinsth_sorted_isin_elem(unClocks, vh))
    return pk_false;
  return !CKM->satisfiesIsAnEquivalentForm(vh);
}

/**
   @return true if a clock (vClk) is an element of the current block (\see{mkArrayDefinition}). 
*/
_VIRTUAL_ boolean DCG2SignalTree::isClockOfCurrentBlock(ptinsth vClk)
{
  ptINode cn;

  if (!isSelectedClock(vClk))
    return pk_false;
  CKM->getInComposedNode(vClk, &cn);
  return (cn == CurrentBlock);
}

/** @return vt + the local signals to the process with vh as root +  the external signals to the process with vh as root
    \note It assumes that vt is a composition.
*/
_VIRTUAL_ ntrees DCG2SignalTree::addSynchroSignals(ptinsth vh, ntrees vt)
{
  ntrees vtlni, vtli;

  mktreelsigsync(vh, &vtlni, &vtli);
  if (ISNOT_NILTREE(vtlni))
    STree_addExprToComposition_Post(&vt, vtlni);
  if (ISNOT_NILTREE(vtli))
    STree_addExprToComposition_Post(&vt, vtli);
  return vt;
}

/**
   It adds
   - the constraints on a clock (vh) to the global list lequnotprov: 
   for a clock defined by a set of fromula {f1, f2, ....,fn}, it adds the synchronisation of f1 with the explicited expression of the the fi formulas.
   - the assertions associated with this clock to the list lassertions

   It takes into account only the formulas of the clock, the constraints/assertions on placed formula are managed otherwise (appequformplnotprov()).
*/
_VIRTUAL_ void DCG2SignalTree::addConstraintsAndAssertions(ptinsth vh)
{
  boolean b;
  ptinsth vh1;
  ntrees vt, lnames;

  // For a BooleanClock Graph, the constraints are set on placed formulae.
  if (isEventClockedFormat(graph_type))
    {
      vh1 = CKM->getNextFormula(vh, pk_true);
      while (vh1 != vh)
	{
	  SET_NILTREE(vt);
	  STree_addExprToComposition_Post(&vt, formula2ntrees(vh1, NILTREE, &b));
	  lnames = mac_STree_make_initSynchro(getClockExpr(vh));
	  lnames = ntreeList_appendLast(lnames, getClockExpr(vh1));
	  STree_addExprToComposition_Post(&vt, lnames);
	  /* vt = addSynchroSignals(vh1, vt); */
	  STree_addExprToComposition_Post(&lequnotprov, vt);
	  vh1 = CKM->getNextFormula(vh1, pk_true);
	}
      vh = vh1;
      while (CKM->getNextAssertion(vh, &vh1))
	{
	  addAssertion(vh1, vh);
	  vh = vh1;
	}
    }
}

/**
   @return the tree 
    - vh ^= t for a POLY graph (graph of "event" clocks)
    - 'when C ^= t', for boolean graph, vh is defined by when C.
*/
_VIRTUAL_ ntrees DCG2SignalTree::initListSynchro(ptinsth vh, ntrees t)
{
  ntrees vt;
  vt = mkWhenClock1(vh);
  vt = STree_mkSynchroIni(vt);
  return ntreeList_appendLast(vt, t);
}

/**
   It adds in vtlni a synchronisation between vh and vsig.
   first (in) must be set  to "true" value the first call to this method, fisrt (out) is set to "false" value.
*/
_VIRTUAL_ void DCG2SignalTree::addSynchroClockSignal(ptinsth vh, ptexpcal vsig, boolean * first, ntrees * vtlni, boolean bforce)
{
  ntrees t, lastL, vt;
  noperators op;
  int n;
  ptINode node;

  node = EM->getNode(vsig);
  if (bforce || EM->IsInputFlow(vsig) || EM->IsOutputFlow(vsig) || isNodeCurrentBlock(node))
    if (!ptlINode_isIn(unNodes, node))
      // The clock of the label is assigned when the associated sub-graph is built.
      if (!EM->IsLabel(vsig))
	{
	  t = ptexpcal2ntrees(vsig);
	  if (*first)
	    {
	      *first = pk_false;
	      *vtlni = initListSynchro(vh, t);
	    }
	  else
	    {
	      op = noper(*vtlni);
	      if (op != oplequhor)
		lastL = nchild(ntreearity(*vtlni), *vtlni);
	      else
		lastL = *vtlni;
	      n = ntreearity(lastL);
	      if (n >= nb_max_list_elem)
		{
		  vt = initListSynchro(vh, t);
		  if (op == oplequhor)
		    {
		      *vtlni = LL_make_lfeed_list(*vtlni);
		    }
		  *vtlni = ntreeList_appendLast(*vtlni, vt);
		}
	      else
		lastL = ntreeList_appendLast(lastL, t);
	    }
	}
}

/**
   Adding the comment (numMess represents its index in the message table) to a tree vt (if is is defined).
   The name of the clock (aClock) is added at the end of the message.
*/

void DCG2SignalTree::addClockHierarchyComment(ptinsth aClock, ntrees vt, int numMess)
{
  ntrees vt1;

  if (!STree_isSignalNilTree(vt))
    {
      vt1 = STree_mkSignalDeclaration(mkWhenClock(aClock), mac_STree_make_comment(numMess), NILTREE);
      STree_addPrintableTreeAttribute(vt, vt1, pk_true);
    }
}

/**
   It removes duplicate identifiers from the operand list (vtSynchro).
   When a clock is represented by a signal x, x appears twice in the synchronisation expression(vtSynchro).
*/
ntrees DCG2SignalTree::ClockSynchroClean(ntrees vtSynchro)
{
  int n;
  ntrees vt;

  if (!STree_isSignalNilTree(vtSynchro))
    {
      n = ntreearity(vtSynchro);
      if (n > 1)
	{
	  vt = nchild(1, vtSynchro);
	  if (STree_isTreeEq(vt, nchild(2, vtSynchro)))
	    {
	      ndestroyListElem(vtSynchro, 2);
	      if (n == 2)
		SET_NILTREE(vtSynchro);
	    }
	}
    }
  return vtSynchro;
}

/**
   It creates for a clock (vh) two lists:
   - vtli (out) : the list of signals synchronous to vh that are local signals to the process with vh as root.
   - vtlni (out) : the list of signals synchronous to vh that are external signals to the process with vh as root.
*/
_VIRTUAL_ void DCG2SignalTree::mktreelsigsync(ptinsth vh, ntrees * vtli, ntrees * vtlni)
{
  boolean firstli, firstlni;
  ptlinstances lsig, lsiglocaux;
  ptexpcal vexpcal;

  SET_NILTREE(*vtli);
  SET_NILTREE(*vtlni);
  lsig = CKM->getSynchronousSignals(vh);
  if (lsig != NILLINSTANCES)
    {
      lsiglocaux = ptinterface_get_localSignals(CKM->getInterfaceSubHierarchy(vh));
      firstlni = pk_true;
      firstli = pk_true;
      while (ptlinstances_get_next(&lsig, &vexpcal))
	if (!KernelExprMode && ptlinstances_isin_exp(lsiglocaux, vexpcal))
	  addSynchroClockSignal(vh, vexpcal, &firstli, vtli, pk_false);
	else
	  addSynchroClockSignal(vh, vexpcal, &firstlni, vtlni, pk_false);
      *vtlni = ClockSynchroClean(*vtlni);
      *vtli = ClockSynchroClean(*vtli);
    }
}

/**
   It adds the constraints name(vh1) ^= name(vh2) to the global list lequnotprov.
   The expression of vh1, vh2 are assumed to be explicited.
*/
_VIRTUAL_ void DCG2SignalTree::hoaddcontrhopl(ptinsth vh1, ptinsth vh2)
{
  ntrees vt, vt1, vt2, lnames;

  if (!isEventClockedFormat(graph_type))
    {
      vt1 = mkhonametree(vh1);
      vt2 = mkhonametree(vh2);
      if (!CKM->satisfiesTickClock(vh1))
	vt1 = LL_make_when1_tree(vt1);
      if (!CKM->satisfiesTickClock(vh2))
	vt2 = LL_make_when1_tree(vt2);
    }
  vt1 = getClockExpr(vh1);
  vt2 = getClockExpr(vh2);
  SET_NILTREE(vt);
  lnames = mac_STree_make_initSynchro(vt1);
  lnames = ntreeList_appendLast(lnames, vt2);
  STree_addExprToComposition_Post(&vt, lnames);
  STree_addExprToComposition_Post(&lequnotprov, vt);
}

/**
  @return "true" if a SIGNAL Abstract Tree representation is associated with a clock (vh), "false" otherwise.
  Moreover, it is \b useful if the SIGNAL Abstract Tree representation is a model.
*/
_VIRTUAL_ boolean DCG2SignalTree::isAlreadyVisited(ptinsth vh, boolean * useful)
{
  ntrees vtree;

  if (!CKM->satisfiesMarked(vh, VisitedMark))
    return pk_false;
  vtree = ptinsthAttr_get_sigtree(vh);
  // if (IS_NILTREE(vtree)) return pk_false;
  if (IS_NILTREE(vtree))
    *useful = pk_false;
  else
    *useful = (boolean) (noper(vtree) == opprocess);
  return pk_true;
}

/** It masks the local signals of a clock (vh) in the body of its  SIGNAL Abstract Tree representation (vprocessus).
 */
_VIRTUAL_ void DCG2SignalTree::_masksigloc_unused(ptinsth vh, ntrees vprocessus)
{
  ntrees vtreeloc;
  ptlinsth lClocks;

  // unused function --- 
  lClocks = ptinterface_get_localClocks(CKM->getInterfaceSubHierarchy(vh));
  if (lClocks != NILLINSTH)
    vtreeloc = mkClockIdentifierList(lClocks);
  else
    SET_NILTREE(vtreeloc);
  vtreeloc = mklnamessig(ptinterface_get_localSignals(CKM->getInterfaceSubHierarchy(vh)), vtreeloc /*, oplident */ );
  STree_maskSignals(vprocessus, vtreeloc);
}

/**
   @return the name of the model associated with a clock
   (for a clock with \em foo as name, it returns \em SUBTREE_of_foo).
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkClockNameTreeProcess(ptinsth vh)
{
  ntrees t;

  t = mkhonametree(vh);
  if (noper(t) != opident)
    t = CKM->getName(vh);
  if (noper(t) != opident)
    return t;
  return newIdent_concIdentIdent(getSubTreeId(), t);  /** action */
}

/** @return  the SIGNAL Abstract Tree representation of  a reference to the model associated with a clock (vh).
    the name of the model is given by mkClockNameTreeProcess().
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkInstClockTreeProcess(ptinsth vh)
{
  return LL_make_inst_tree(mkClockNameTreeProcess(vh), NILTREE);
}

/** 
    @return the Signal abstract tree representation of a sub-graph (Actions) of a clock (aClock).
    It is a nil tree or a process.
*/
_VIRTUAL_ ntrees DCG2SignalTree::buildGFDTree(ptinsth aClock)
{
  ptlsucc linstr;
  ntrees vtreeinstr, vgfdtree;
  ptgfd vgfd;

  vgfd = CKM->getClockSubGraph(aClock);
  if (ptgfd_satisfies_empty(vgfd)) return NILTREE;
  vgfdtree = STree_mkSkeletonModel(NILIDENT, opprocess);
  linstr = ptgfd_get_nodes(vgfd);
  buildtree(linstr, vgfdtree);

  vtreeinstr = STreeModel_getBody(vgfdtree);
  addClockHierarchyComment(aClock, vtreeinstr, comment_actions);
  return vgfdtree;
}

/**
   It adds to a model (vhierarchy) the not proved equivalences stored in lequnotprov (attribute member of the class).
*/
_VIRTUAL_ void DCG2SignalTree::applequnotprov(ntrees vhierarchy)
{
  ntrees vLabel, vt;

  if (ISNOT_NILTREE(lequnotprov))
    {
      STree_addPrintableTreeAttribute(lequnotprov, mac_STree_make_comment(comment_clockconstraint), pk_true);
      if (flatMode)
	{
	  vLabel = getConstraintsId();
	  STree_addLabelInModel(STree_scopy(vLabel), vhierarchy);
	  vt = LL_make_petiq_tree(STree_scopy(vLabel), lequnotprov);
	}
      else
	vt = lequnotprov;
      STree_addExpressionOfProcessToaModel(vhierarchy, vt);
    }
}

/**
   It initializes the translation of a node of the hierarchy (vh) by the following operations:
   - it adds a reference to the process associated with the clock (vh) to the expression of the process of the father clock (vexpro).
   - it builds a skeleton model for the clock; it is associated, as an attribute, with the clock. The attribute will be used later to complete the translation.
   - it adds the declaration of the model to the father clock (vt).
   - it adds in the expression of process of the created model the vli tree (list of the local signals of the clock synchronized with it)
   .
   It assumes that vexpro is a composition, vt is a model.
*/
_VIRTUAL_ void DCG2SignalTree::initNodeHierarchyTranslation(ntrees * vexpro, ptinsth vh, ntrees vli, ntrees vt)
{
  ntrees vtreetomk;

  // For the  expression of process of the father clock: it adds a reference to the process associated with the clock.
  STree_addExprToComposition_Post(vexpro, mkInstClockTreeProcess(vh));

  // A model is created for the clock
  vtreetomk = STree_mkSkeletonModel(mkClockNameTreeProcess(vh), opprocess);
  ptinsthAttr_set_sigtree(vh, vtreetomk);
  // The declaration of the model is added to the father clock.
  STree_addProcessDeclaration(vt, vtreetomk);
  if (ISNOT_NILTREE(vli))
    STree_addExpressionOfProcessToaModel(vtreetomk, vli);
}

/**
   It adds
   - to lequnotprov (attribute member of the class) the non proved equivalence for placed formula in the hierarchy.
   It calls hoaddcontrhopl() for each constraint.
   - to lassertions (attribute member of the class) the assertions on placed formula in the hierarchy.
   It calls assertion() for each assertion.
   .
   These lists are added at the end of the translation to the built model.
*/
_VIRTUAL_ void DCG2SignalTree::appequformplnotprov()
{
  ptlinsth lh1, lh2;
  ptinsth vh1, vh2;

  CKM->getConstraintsOnPlacedClocks(&lh1, &lh2);
  while (ptlinsth_get_next(&lh1, &vh1))
    {
      ptlinsth_get_next(&lh2, &vh2);
      vh1 = CKM->getRepresentativeFormula(vh1);
      vh2 = CKM->getRepresentativeFormula(vh2);
      if (vh1 != vh2)
	hoaddcontrhopl(vh1, vh2);
    }
  if (isEventClockedFormat(graph_type))
    {
      CKM->getAssertionsOnPlacedClocks(&lh1, &lh2);
      while (ptlinsth_get_next(&lh1, &vh1))
	{
	  ptlinsth_get_next(&lh2, &vh2);
	  vh1 = CKM->getRepresentativeFormula(vh1);
	  vh2 = CKM->getRepresentativeFormula(vh2);
	  if (vh1 != vh2)
	    addAssertion(vh1, vh2);
	}
    }
}

/**
   @return the Signal Abstract Tree representation of a pragma statement (PragmaStatement).
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkPragmaStatement(ntrees PragmaStatement)
{
  ptrepstr repstrPragmaStatement;
  ptrepstr vnewstr;

  if (IS_NILTREE(PragmaStatement))
    return STree_mkAtomNilTree();
  if (noper(PragmaStatement) == opnil)
    return PragmaStatement;

  repstrPragmaStatement = nstringof(nchild(1, PragmaStatement));
								 /** A revoir : on peut avoir une liste */
  if (getDoubleQuotedString(repstrPragmaStatement, &vnewstr))
    return PragmaStatement;
  return LL_make_chai_list(LL_make_cha_atom(vnewstr));
}

/**
   It adds (STree_addPragma()) a pragma (v) to the model (vt). The
   pragma statement is given by mkPragmaStatement().
*/
_VIRTUAL_ void DCG2SignalTree::addPragma(ptpragma v, ntrees vt)
{
  ntrees vtPrag, vtexpComment, tobjComment, vtexp, vtPragmaStatement, PragmaStatement, tobj;
  ptrepstr name;
  ptObjPragma obj;
  ptexpcal exp;
  ptinsth h;
  ptlinstances l;

  getPragma(v, &name, &obj, &PragmaStatement);
  SET_NILTREE(tobj);
  SET_NILTREE(tobjComment);

  l = (ptlinstances) obj;
  if (l != NILLINSTANCES)
    {
      while (ptlinstances_get_next(&l, &exp))
	{
	  SET_NILTREE(vtexp);
	  SET_NILTREE(vtexpComment);
	  if (EM->IsSignal(exp))
	    {
	      if (!ptlINode_isIn(unNodes, EM->getNode(exp)))
		vtexp = ptexpcal2ntrees(exp);
	      else
		vtexpComment = ptexpcal2ntrees(exp);
	    }
	  else
	    {
	      if (EM->getOperator(exp) == opextr)
		if (EM->getOperator(EM->getWhenOrCellArg(exp)) == optrue)
		  {
		    h = EM->getContributionClock(exp);	// unClocks
		    if (isEventClockedFormat(graph_type))
		      {
			if (!ptlinsth_isin_elem(unClocks, h))
			  vtexp = mkhonametree(h);
			else
			  vtexpComment = mkhonametree(h);
		      }
		    else
		      {
			exp = CKM->getBooleanSignal(h);
			if (exp != NILEXPCAL)
			  if (!ptlINode_isIn(unNodes, EM->getNode(exp)))
			    vtexp = ptexpcal2ntrees(exp);
			  else
			    vtexpComment = ptexpcal2ntrees(exp);
			else
			  vtexpComment = mkhonametree(h);
		      }
		  }
		else
		  vtexpComment = ptexpcal2ntrees(exp);
	      else
		vtexpComment = ptexpcal2ntrees(exp);
	    }
	  if (ISNOT_NILTREE(vtexp))
	    {
	      if (IS_NILTREE(tobj))
		tobj = LL_make_lopra_emptyList();
	      tobj = ntreeList_appendLast(tobj, vtexp);
	    }
	  else if (ISNOT_NILTREE(vtexpComment))
	    {
	      if (IS_NILTREE(tobjComment))
		tobjComment = LL_make_lident_emptyList();
	      tobjComment = ntreeList_appendLast(tobjComment, vtexpComment);
	    }
	}

    }

  vtPragmaStatement = mkPragmaStatement(PragmaStatement);
  vtPrag = STree_mkPragma(name, tobj, STree_scopy(vtPragmaStatement));
  if (ISNOT_NILTREE(tobjComment))
    STree_setVMAnnotations(vtPrag, tobjComment);
  STree_addPragma(vt, vtPrag);
}

/** 
    It adds each pragma, of a list (lpragmas), to a model (vtcurrent).
*/
_VIRTUAL_ void DCG2SignalTree::addPragmasModel(ntrees vtcurrent)
{
  ptpragma v, vpragma;
  ptlpragmas lp;
  boolean badd;
  ptlinstances l;
  ptexpcal exp;
  ptrepstr name;
  ptObjPragma obj;
  ptINode n;
  ntrees vmodel, PragmaStatement, vtn;

  lp = lpragmas;
  while (ptlpragmas_get_next(&lpragmas, &v))
    {
      if (!isPragmaKeyword(v, PRAGMA_RUN_ON))
	badd = pk_true;
      else if (RunOnSetOnSubgraphs)
	badd = pk_false;	// the pragma has been assigned with the sub-graph.
      else
	{
	  // getLinkNodeToTree(ptINode vnode, ntrees *vtree) inserer le pragma
	  getPragma(v, &name, &obj, &PragmaStatement);
	  l = (ptlinstances) obj;
	  if (l == NILLINSTANCES)
	    badd = pk_true;
	  else
	    {
	      badd = pk_false;
	      while (ptlinstances_get_next(&l, &exp))
		if (EM->IsSignal(exp))
		  {
		    n = EM->getNode(exp);
		    if (!ptlINode_isIn(unNodes, n))
		      if (getLinkNodeToTree(n, &vtn))
			{
			  vmodel = STree_getModelInWhich(vtn);
			  vpragma = newpragma(name, (ptObjPragma) ptlinstances_insert(NILLINSTANCES, exp), PragmaStatement);
			  addPragma(vpragma, vmodel);
			}
		  }
	    }
	}
      if (badd)
	addPragma(v, vtcurrent);
    }
}

/** 
    It adds each pragma, of a list (lpragmas), to a model (vtcurrent).
*/
_VIRTUAL_ void DCG2SignalTree::addPragmas(ptlpragmas lpragmas, ntrees vtcurrent)
{
  ptpragma v;

  while (ptlpragmas_get_next(&lpragmas, &v))
    addPragma(v, vtcurrent);
}

/** 
    It adds to a model (vhierarchy) as comments the definitions of the nodes of a list of nodes (ln).
    A message (codemess) is added to this comment.
*/
_VIRTUAL_ void DCG2SignalTree::addInfosNodes(ptlINode ln, ntrees vhierarchy, int codemess)
{
  ntrees vt;
  boolean b;

  if (ln != NILLINODE)
    {
      b = setKernelExprMode(pk_false);
      vt = lnodes2trees(ln);
      if (ISNOT_NILTREE(vt))
	{
	  pK_messenger->warningaccForce(codemess, vt);
	  STree_setVMAnnotations(vhierarchy, vt);
	}
      b = setKernelExprMode(b);
    }
}

/** 
    It adds to a model (vhierarchy), the nodes not localized by the "RunOn" partitionning.
    This information is set by the runonpartitionning as an attribute of the graph.
*/
_VIRTUAL_ void DCG2SignalTree::addErrorPartitionning(ntrees vhierarchy)
{
  ptlINode ln;
  if (AGRAPH->getAttribute(RUNON_NODES_NOT_LOCALIZED, (attributeValue *) & ln))
    addInfosNodes(ln, vhierarchy, warnotlocalized);
}

/** 
    It adds to a model (vhierarchy) a comment to specify that the signals (lsighnull) are synchronized with the null clock.
    (The auxillary signals are filtered).
*/
_VIRTUAL_ void DCG2SignalTree::addSignalsSynchronizedWithNullClock(ptlinstances lsighnull, ntrees vhierarchy)
{
  ntrees vt;
  ptlinstances lsave, lsig, dummy;
  ptexpcal aSig;

  lsig = NILLINSTANCES;
  while (ptlinstances_get_next(&lsighnull, &aSig))
    if (!EM->IsAuxiliarySignal(aSig))
      ptlinstances_isin_insert_exp(&lsig, &dummy, aSig);

  if (lsig == NILLINSTANCES)
    return;
  vt = mac_STree_make_initSynchro(mkhonametree(CKM->getNullClock()));
  lsave = lsig;
  while (ptlinstances_get_next(&lsig, &aSig))
    vt = ntreeList_appendLast(vt, ptexpcal2ntrees(aSig));
  pK_messenger->warningaccForce(warsighnull, vt);
  STree_setVMAnnotations(vhierarchy, vt);

  lsig = lsave;
  SET_NILTREE(vt);
  while (ptlinstances_get_next(&lsig, &aSig))
    if (EM->IsOutputFlow(aSig) || EM->IsInputFlow(aSig))
      {
	if (IS_NILTREE(vt))
	  vt = mac_STree_make_initSynchro(mkhonametree(CKM->getNullClock()));
	vt = ntreeList_appendLast(vt, ptexpcal2ntrees(aSig));
      }
  if (ISNOT_NILTREE(vt))
    STree_addSpecification(vhierarchy, vt);
}

// Used only for a DC+ graph.
void DCG2SignalTree::addAssertion(ptinsth vh1, ptinsth vh)
{
  ntrees vt;
  boolean b;

  vt = mac_STree_make_initSynchro(mkhonametree(vh1));
  vt = ntreeList_appendLast(vt, mkhonametree(vh));
  STree_addExprToComposition_Post(&lassertions, vt);
  if (CKM->getRepresentativeFormula(vh1) != vh1)
    STree_addExprToComposition_Post(&lDefForAssertions, formula2ntrees(vh1, NILTREE, &b));
  if (CKM->getRepresentativeFormula(vh) != vh)
    STree_addExprToComposition_Post(&lDefForAssertions, formula2ntrees(vh, NILTREE, &b));
}

/**
   For a DC+ graph.
   For each clock of a list (lh), it adds to the attribute member (lasertions)
   as an assertion the equivalence of the clock with the null clock.
*/
_VIRTUAL_ void DCG2SignalTree::addAssertionsNullClock(ptlinsth lh)
{
  ptinsth vh, hnull;

  hnull = CKM->getNullClock();
  while (ptlinsth_get_next(&lh, &vh))
    addAssertion(vh, hnull);
}

/**
   For each clock of a list (lh), it adds to the attribute member (lequnotprov)
   the synchronisation of the clock (formula2ntrees()) with the null clock.
*/
_VIRTUAL_ void DCG2SignalTree::addConstraintsNullClock(ptlinsth lh)
{
  ptinsth vh, hnull;
  ntrees vt, vtExp, lnames;
  boolean b;

  hnull = CKM->getNullClock();
  SET_NILTREE(vtExp);
  lnames = mac_STree_make_initSynchro(mkhonametree(hnull));
  while (ptlinsth_get_next(&lh, &vh))
    if (!CKM->satisfiesIsAnAssumption(vh))
      {
	vt = formula2ntrees(vh, NILTREE, &b);
	if (!STree_isSignalNilTree(vt))
	  {
	    if (noper(vt) == oplequhor)
	      vt = nchild(2, vt);
	    lnames = ntreeList_appendLast(lnames, vt);
	  }
      }

  STree_addExprToComposition_Post(&vtExp, lnames);
  STree_addExprToComposition_Post(&lequnotprov, vtExp);
}

/**
   It adds, to a model (vtcurrent), the synchronizations with the null clock (addSignalsSynchronizedWithNullClock())
   and the equivalences not proved to the null clock (addConstraintsNullClock()).
*/
_VIRTUAL_ void DCG2SignalTree::insertNullClock(ntrees vtcurrent)
{
  ptlinsth lh;
  ptlinstances lsig;

  if (AGRAPH->getNullClockSignal(&lsig))
    addSignalsSynchronizedWithNullClock(lsig, vtcurrent);
  if (AGRAPH->getNullClockConstraints(&lh))
    addConstraintsNullClock(lh);
  if (isEventClockedFormat(graph_type))
    if (CKM->getNullClockAssertions(&lh))
      addAssertionsNullClock(lh);
}

/**
   @return "true" when the Signal abstract tree representation of the definition of the hierarchy associated with a clock (vh) 
   is integrated to the model (vt), "false" otherwise.

   @param vh (in) a clock
   @param vli (in) local signals to the sub tree with vh as root and with vh as clock
   @param vt (in) the (assumed) Signal abstract tree representation of the model associate with the father clock of vh.
   @param isaextract (in) must be pk_true when vh is an extraction.
   @param vexpro (in/out) the Signal abstract tree representation associated with vh (set when it returns "true")
   \note It is done when vh is a leaf of the hierarchy and when the number of nodes associated with it is equal to 1.   
*/
_VIRTUAL_ boolean DCG2SignalTree::doneForSubTree(ptinsth vh, ntrees * vexpro, ntrees vli, ntrees vt, boolean isaextract)
{
  ntrees vtreeinstr, processTree;
  int nbinstr;

  CKM->addMark(vh, VisitedMark);

  if (flatMode)
    {
      STree_addExprToComposition_Post(vexpro, vli);
      return pk_false;
    }
  if (CKM->getChildren(vh) != NILLINSTH)
    {
      initNodeHierarchyTranslation(vexpro, vh, vli, vt);
      return pk_false;
    }
  vtreeinstr = STree_getNilTree();
  processTree = buildGFDTree(vh);

  vtreeinstr = STreeModel_getBody(processTree);
  nbinstr = ntreearity(vtreeinstr);
  if (genInterface)
    if (nbinstr == 1)
      nbinstr++;
  if (nbinstr > 1)
    {
      initNodeHierarchyTranslation(vexpro, vh, vli, vt);
      return pk_false;
    }
  STree_addExprToComposition_Post(vexpro, vli);
  STree_mergeDeclarations(vt, processTree);
  appendLoc(CKM->getInterfaceSubHierarchy(vh), vt);
  if (nbinstr == 1)
    STree_addExprToComposition_Post(vexpro, STree_getSimplifiedComposition(vtreeinstr));

  if (isaextract)
    ptinsthAttr_set_sigtree(vh, *vexpro);
  addConstraintsAndAssertions(vh);
  return pk_true;
}

/**
   Adds id in the list of clock parameters.
   The global list (ClockParameterTree) is added to the model at the end of the transformation.
*/
void DCG2SignalTree::addToClockParameterTree(ntrees id)
{
  if (noper(id) != opevnt)
    id = LL_make_evnt_tree(id);
  if (STree_isSignalNilTree(ClockParameterTree))
    ClockParameterTree = LL_make_lesign_list(id);
  else
    ClockParameterTree = ntreeList_appendLast(ClockParameterTree, id);
}

/**
    @return "true" if the sub-tree associated with a clock (vh) is not integrated to vt, "false" otherwise.

    It adds to the body of a model (vt) -- vt assumed to be the model associated with the father clock of vh --
    - (1) the composition (vexppro) composed of the Signal abstract tree of the definition of a clock formula (vh)
    and the synchronisation with vh of the signals (non locals to the sub tree with vh as root).
    - (2) and the definition of the sub-tree associated with vh when it satisfies some conditions (see doneForSubTree().
*/
_VIRTUAL_ boolean DCG2SignalTree::buildexprforfath(ptinsth vh, ntrees vt, ntrees * vexpro)
{
  ntrees vtformula, vsigname, vli, vlni;
  boolean isaextract;
  ptexpcal vsig;

  isaextract = pk_false;
  mktreelsigsync(vh, &vli, &vlni);
  if (!getClockNameRepresentedBySignal(vh, &vsigname))
    SET_NILTREE(vsigname);

  if (CKM->satisfiesInitClock(vh))
    {
      if (CKM->getSynchronousSignals(vh) != NILLINSTANCES)
	STree_addExprToComposition_Post(vexpro, formula2ntrees(vh, vsigname, &isaextract));
    }
  else if (!CKM->satisfiesNullClock(vh))
    if (!CKM->isSubstitutable(vh))
      {
	vtformula = formula2ntrees(vh, vsigname, &isaextract);
	if (CKM->isClockRepresentedBySignal(vh, &vsig))
	  addClockHierarchyComment(vh, vtformula, comment_clockrepresentedby);
	if (ISNOT_NILTREE(vtformula))
	  STree_addExprToComposition_Post(vexpro, vtformula);
	else			// Clock parameter.
	  {
	    if (isEventClockedFormat(graph_type))	// Why ???
	      {
		if (IS_NILTREE(vsigname))
		  vsigname = LL_make_evnt_tree(mkhonametree(vh));
		addToClockParameterTree(vsigname);
	      }
	  }
      }

  if (ISNOT_NILTREE(vlni))
    STree_addExprToComposition_Post(vexpro, vlni);
  return !doneForSubTree(vh, vexpro, vli, vt, isaextract);
}

/**
   It adds the definition of the complementary clock of a clock (vh) assumed to be an extraction to the expression of process (vexpro) of a model (vt).
   This operation is achieved if such a clock is useful.
   @return "true" if the operation is achieved, "false" otherwise.
*/
boolean DCG2SignalTree::mkSubTreeComplExtraction(ptinsth vh, ntrees vt, ntrees * vexpro)
{
  boolean b;
  ptinsth vh1;

  b = pk_false;
  CKM->getComplInFather(vh, &vh1);
  vh1 = CKM->getRepresentativeFormula(vh1);
  if (!CKM->satisfiesIsAnAssumption(vh1))
    if (isSelectedClock(vh1) && !isAlreadyVisited(vh1, &b))
      b = buildexprforfath(vh1, vt, vexpro);
  return b;
}

/**
   It defines, if it is not already done and when it is useful, the expression of vh for its father clock: 
   the definition of the formula (buildexprforfath()) is added to the body of the model associated to its father clock (vt).
   Moreover, under some conditions (it is a leaf with one node associated with it) the sub tree of vh is completely 
   integrated to the body of vt. 

   @return "true" if the operand clock (vh) is useful for the graph and the sub-tree associated with vh is not
   integrated to vt, "false" otherwise.
*/
_VIRTUAL_ boolean DCG2SignalTree::mkSubTree(ptinsth vh, ntrees vt)
{
  boolean useful;
  ntrees vexpro;

  if (!isSelectedClock(vh))
    return pk_false;

  if (!isClockOfCurrentBlock(vh))
    {
      if (!flatMode)
	return pk_false;
      vexpro = addSynchroSignals(vh, STree_getNilTree());
      STree_addExpressionOfProcessToaModel(vt, vexpro);
      return pk_true;
    }
  if (flatMode)
    {
      vexpro = STree_getNilTree();
      useful = buildexprforfath(vh, vt, &vexpro);
      STree_addExpressionOfProcessToaModel(vt, vexpro);
      return useful;
    }

  if (!isAlreadyVisited(vh, &useful))
    {
      vexpro = STree_getNilTree();
      if (CKM->getOperator(vh) == sub)
	{
	  useful = buildexprforfath(vh, vt, &vexpro);
	  mkSubTreeComplExtraction(vh, vt, &vexpro);
	}
      else if (CKM->getOperator(vh) == nosub)
	{
	  mkSubTreeComplExtraction(vh, vt, &vexpro);
	  useful = buildexprforfath(vh, vt, &vexpro);
	}
      else
	useful = buildexprforfath(vh, vt, &vexpro);
      STree_addExpressionOfProcessToaModel(vt, vexpro);
    }
  return useful;
}

/**
   Cleaning the composition.
   For a composition (vt), a sub-composition reduced to one element is replaced by this element.
   @return the simplified tree. If vt is reduced to one element, its child is returned.
*/
ntrees DCG2SignalTree::SimplifyComposition(ntrees vt, boolean bRootClock)
{
  int n, j;
  ntrees vti, nvti;

  if (STree_isComposition(vt))
    {
      n = ntreearity(vt);
      if (n == 0)
	return NILTREE;
      if (!bRootClock)
	if (n == 1)
	  return nchild(1, vt);

      for (j = n; j >= 1; j--)
	{
	  vti = nchild(j, vt);
	  nvti = SimplifyComposition(vti, pk_false);
	  if (nvti != vti)
	    {
	      if (IS_NILTREE(nvti))
		ndestroyListElem(vt, j);
	      else
		nchangchild(vt, nvti, j);
	    }
	}
    }
  else if (STree_isConfinedProcess(vt))
    {
      vti = nchild(1, vt);
      if (IS_NILTREE(vti))
	return NILTREE;
      nvti = SimplifyComposition(vti, pk_false);
      if (nvti != vti)
	{
	  if (IS_NILTREE(nvti))
	    return NILTREE;
	  nchangchild(vt, nvti, 1);
	}
    }
  return vt;
}

/**
   The reference to the model (vt) associated with a clock (aClock) is removed from its father clock.
*/
void DCG2SignalTree::TreeRemoveChild(ptinsth aClock, ntrees vt)
{
  ntrees vtFather, vtCall;

  vtFather = ptinsthAttr_get_sigtree(CKM->getFather(aClock));
  if (STreeModel_getCallToIn(STreeModelModule_getName(vt), vtFather, &vtCall))
    {
      STree_Destroy(vtCall);
      STree_Destroy(vt);
    }
}

/**
   It finalizes the translation of a node of the hierarchy (aClock) as a SIGNAL Abstract Tree representation (vt):
   - If the expression of process of vt is undefined, the reference to the model in the body of the father
   clock is removed (ie the representation of the clock is optimized).
   - otherwise, the expression of definition is cleaned: a sub-composition reduced to one element is replaced by this element.
*/
_VIRTUAL_ void DCG2SignalTree::buildHierachyTreeEnd(ptinsth aClock, ntrees vt)
{
  ntrees vBody;
  boolean bRootClock;

  bRootClock = (vRootClock == aClock);
  vBody = STreeModel_getBody(vt);
  if (IS_NILTREE(vBody))
    TreeRemoveChild(aClock, vt);
}

/** 
    It defines the SIGNAL Abstract Tree representation for a node of the clock hierarchy (aClock)
    and visits the hierarchy by a left to right topdown traversal.
    The SIGNAL Abstract Tree representation, associated with the clock, is a model (process). It is added as an attribute to the 
    clock.
    
    The following operations are achieved:
    - addConstraintsAndAssertions() it adds the constraints and the assertions associated with the clock to the model.
    - mtInterfaceTree() it defines the interface of the SIGNAL Abstract Tree model. 
    - mktreeloc() it defines the locals of the SIGNAL Abstract Tree model. 
    - for each child of aClock, it executes "if mkSubTree() then buildHierachyTree()".
    - buildGFDTree() it defines the body of the model
    - buildHierachyTreeEnd() it finalizes the SIGNAL Abstract Tree representation.
*/
_VIRTUAL_ void DCG2SignalTree::buildHierachyTree(ptinsth aClock)
{
  ntrees vt;
  ptlinsth lchild;
  ptinsth vh;

  addConstraintsAndAssertions(aClock);
  vt = ptinsthAttr_get_sigtree(aClock);
  mtInterfaceTree(aClock, vt);
  mktreeloc(aClock, vt);
  lchild = CKM->getChildren(aClock);
  while (ptlinsth_get_next(&lchild, &vh))
    if (mkSubTree(vh, vt))
      buildHierachyTree(vh);

  STree_mergeProcesses(vt, buildGFDTree(aClock));
  buildHierachyTreeEnd(aClock, vt);
}

/**
   For an endochronous model, it adds 
   - a comment when the master clock is represented by a signal
   - a comment when the master clock is not a clock of an i/o signal (oversampling).
 */
_VIRTUAL_ void DCG2SignalTree::annotRootClock()
{
  ntrees vt;
  ptexpcal vsig;
  ptinsth vhComplete, vclk;

  vclk = CKM->getTickClock();
  if (!CKM->satisfiesFictitiousRoot(vclk))
    {
      // endochronous model.
      vt = ptinsthAttr_get_sigtree(vRootClock);
      // Is the root clock represented by a signal ?
      if (CKM->isClockRepresentedBySignal(vclk, &vsig))
	{
	  addClockHierarchyComment(vclk, STreeModel_getBody(vt), comment_endorepresentedby);
	}

      if (AGRAPH->isEndochronousWithInternalClockRoot(&vhComplete))
	{
	  ntrees body_tree;
	  body_tree = STreeModel_getBody(vt);
	  STree_addPrintableTreeAttribute(body_tree, mac_STree_make_comment(comment_endointernalclock), pk_true);
	}
    }
}

void DCG2SignalTree::appClockParameters(ntrees vtcurrent)
{
  if (!STree_isSignalNilTree(ClockParameterTree))
    STree_addPrintableTreeAttribute(STreeModel_getBody(vtcurrent),
				    STree_mkSignalDeclaration(ClockParameterTree, mac_STree_make_comment(comment_clockparameters), NILTREE), pk_true);
}

void DCG2SignalTree::finalizeHierarchy(ntrees vtcurrent)
{
  ntrees vt;

  vt = STreeModel_getBody(vtcurrent);
  if (STree_isSignalNilTree(vt))
    STree_setBody(vtcurrent, STree_mkNullaryTree(opnullproc));

  insertDeclarationsFlatMode();
  appequformplnotprov();
  insertNullClock(vtcurrent);
  applequnotprov(vtcurrent);
  appClockParameters(vtcurrent);
  if (ISNOT_NILTREE(lDefForAssertions))
    STree_addExpressionOfProcessToaModel(vtcurrent, lDefForAssertions);
  if (ISNOT_NILTREE(lassertions))
    {
      lassertions = LL_make_passert_tree(lassertions);
      STree_addExpressionOfProcessToaModel(vtcurrent, lassertions);
    }
  addInfosNodes(unNodes, vtcurrent, warsigunused);
  addPragmasModel(vtcurrent);
  addErrorPartitionning(vtcurrent);
  // Currently, lMemOptimizedAttribute is always empty.
  if (ISNOT_NILTREE(lMemOptimizedAttribute))
    STree_setVMAnnotations(vtcurrent, lMemOptimizedAttribute);
}

/**
   It adds to a model (vtcurrent) the SIGNAL Abstract Tree representation of a graph (aGraph)
   When the parameter bflat 
    - is "true", the clock hierarchy is recursively explicited as a list of subprocesses.
    - is "false", the clock hierarchy is flatenned (for example, if a graph contains some labels, it is necessary to flatten the hierarchy).
    .
    To achieve this translation, it calls
    - flatModeInit() it initializes the "flat mode" translation.
    - buildHierachyTree() it visits the hierarchy left to right, topdown. 
    - mkGraphTreeDependences() it adds the explicit dependences to the SIGNAL Abstract Tree representation
    - insertDeclarationsFlatMode() it adds the declarations in "flatened mode"
    - appequformplnotprov() it adds the not proved equivalences (constrainsts)
    - insertNullClock() it adds the synchronizations with the null clock.
    - addInfosNodes() it adds (comments) the unused objects (signals, clocks) of the graph to the SIGNAL Abstract Tree representation,
*/
_VIRTUAL_ void DCG2SignalTree::IGraph2Process(ptDCG aGraph, ntrees vtcurrent)
{
  ptinsth vhtop;
  boolean o_genInterface;

  DCG2SignalTreeini(aGraph);
  lequnotprov = STree_getNilTree();
  lassertions = STree_getNilTree();
  ClockParameterTree = STree_getNilTree();
  lDefForAssertions = STree_getNilTree();
  lMemOptimizedAttribute = STree_getNilTree();
  vhtop = aGraph->getRootClock();
  vRootClock = vhtop;
  ptinsthAttr_set_sigtree(vhtop, vtcurrent);
  o_genInterface = genInterface;
  flatModeInit(vtcurrent, getHierarchyModePrinting());
  buildHierachyTree(vhtop);
  annotRootClock();
  finalizeHierarchy(vtcurrent);
  DCG2SignalTreeFinal(vtcurrent);
  mkGraphTreeDependences(vtcurrent);	// added after to keep the composition (removed otherwise in DCG2SignalTreeFinal()).
  cleanGraph();
  genInterface = o_genInterface;
}

/**
   Local to DCG2SignalTree::cleanTree(ntrees).
   @return the equations of the composition (compTree) added to the global composition (vt).
*/
ntrees DCG2SignalTree::cleantree(ntrees compTree, ntrees vt)
{
  ntrees t;
  int n, i;

  if (!STree_isUndefinedTree(compTree))
    {
      n = ntreearity(compTree);
      for (i = 1; i <= n; i++)
	{
	  t = nchild(i, compTree);
	  if (STree_isComposition(t))
	    vt = cleantree(t, vt);
	  else
	    STree_addExprToComposition_Post(&vt, t);
	}
    }
  return vt;
}

/**
   For the "flat mode", all the definitions of a model (vmodel) are defined in the same composition.
   The labelled sub-systems are kept but the no labelled ones are flattened.
   \see DCG2SignalTree::cleantree(ntrees, ntrees).
*/
void DCG2SignalTree::cleanTree(ntrees vmodel)
{
  ntrees body, vt;

  if (flatMode)
    {
      SET_NILTREE(vt);
      body = STreeModel_getBody(vmodel);
      if (!STree_isUndefinedTree(body))
	if (STree_isComposition(body))
	  {
	    vt = cleantree(body, vt);
	    if (vt != body)
	      STree_setBody(vmodel, vt);
	  }
    }
}

/**
   When the SpecMode 
   - is ON, it adds the specification part of a graph (aGrap) to its abstract tree representation (vtcurrent).
   - is OFF, the equations of the specification part is added to the body of the abstract tree representation.
*/
void DCG2SignalTree::addSpec(ntrees vtcurrent, ptDCG aGraph)
{
  ntrees leq, vtSpec;
  boolean x;

  x = bDependenceWithClock;
  bDependenceWithClock = pk_false;	// to avoid the declaration of H in (X --> Y when H) in the spec area. (TO BE REVISITED).
  vtSpec = IGraph2AbtractionTree(aGraph);
  bDependenceWithClock = x;

  vtSpec = STree_getSpec(vtSpec);
  if (SpecMode)
    STree_addSpecification(vtcurrent, vtSpec);
  else if (STree_getSpecEquations(vtSpec, &leq))
    STree_addPreExpressionOfProcessToaModel(vtcurrent, leq);	// And the declarations!!!
}

/**
   @return "true" if it exists at least a signal of lsig synchronous with a clock (vh), "false" otherwise.
   It returns in vt the Signal Abstract tree representation of the signals (list of identifiers) of lsig synchronous with vh.
*/
boolean DCG2SignalTree::mkSynchronizedSignalsOfToClock(ptinsth vh, ptlinstances lsig, boolean boolFormat, ntrees * vt)
{
  boolean first, b;
  ptlinstances lsigvh;
  ptexpcal vexpcal, vtick;
  ntrees vtlni, vt1;

  if (boolFormat)
    {
      b = CKM->satisfiesTickClock(vh);
      vtick = CKM->getBooleanSignal(vh);
    }
  else
    b = pk_false;

  lsigvh = CKM->getSynchronousSignals(vh);
  vtlni = STree_getNilTree();
  first = pk_true;
  while (ptlinstances_get_next(&lsigvh, &vexpcal))
    if (ptlinstances_isin_exp(lsig, vexpcal))
      {
	if (b && (vexpcal == vtick))
	  vt1 = mkExtraction(vexpcal, pk_false);
	else
	  vt1 = ptexpcal2ntrees(vexpcal);
	if (first)
	  {
	    first = pk_false;
	    vtlni = STree_mkSynchroIni(vt1);
	  }
	else
	  vtlni = ntreeList_appendLast(vtlni, vt1);
      }
  *vt = vtlni;
  return (ISNOT_NILTREE(vtlni));
}

/*
  For the operand clock (vh), it builds the synchronisation of the signals of lsig with vh as clock. 
  When the parameter bDef is pk_true, the definition of the clock is added to the synchronisation expression.
  The results is added to the vabstract tree. In kernelMode, the clock operators are considered as binary operators.

  Example : [:c] ^= x1 ...  ^= xn is added to vabtract for a clock (vh) defined by [:c] and the signals xi of lsig with [:c] as clock.
*/
void DCG2SignalTree::mkSignalsClockSynchro(ptinsth vh, ptlinstances lsig, boolean bDef, ntrees * vabstract)
{
  ntrees vt1, vt;
  boolean isaextract;

  vt = STree_getNilTree();
  if (mkSynchronizedSignalsOfToClock(vh, lsig, pk_false, &vt))
    {
      if (bDef && mkClockExpression(vh, pk_true, &isaextract, &vt1))
	vt = ntreeList_insertFirst(vt1, vt);
      if (ISNOT_NILTREE(vt) && ntreearity(vt) > 1)
	{
	  if (KernelExprMode)
	    vt = STree_ClockOperatorsListToBinaryOperator(vt);
	  STree_addExprToComposition_Post(vabstract, vt);
	}
    }
  else if (bDef)
    {
      STree_addExprToComposition_Post(vabstract, clock2Tree(vh));
    }
}

/**
   For each clock of lh it applies 
   @return the composition of these synchroniations.
*/
ntrees DCG2SignalTree::mkSignalsClockListSynchro(ptlinsth lh, ptlinstances lsig, boolean bDef)
{
  ptinsth vh;
  ntrees vabstract;

  vabstract = STree_getNilTree();
  while (ptlinsth_get_next(&lh, &vh))
    mkSignalsClockSynchro(vh, lsig, bDef, &vabstract);
  return vabstract;
}

/**
   @return the tree of the definitions of boolean signals such that (C)
   For a clock of the list lh, vh an extraction built on a boolean signal (C), 
   it adds the definition of C is C is defined by E and E is an expression on input/output signals then when C is replaced by E
*/
ntrees DCG2SignalTree::addBooleanDefinitions(ptlinsth lh, ntrees * vtdecl)
{
  ptinsth vh;
  ntrees vtlids, vtres, vt;
  boolean bpos;
  ptexpcal c, vExport;

  vtlids = STree_getNilTree();
  vtres = STree_getNilTree();
  while (ptlinsth_get_next(&lh, &vh))
    if (CKM->satisfiesSub(vh))
      {
	c = CKM->getConditionSub(vh);
	if (EM->IsSignal(c))
	  {
	    vExport = EM->getExprExportBooleanAbtraction(c, NILLINSTANCES, &bpos);
	    if (vExport != c)
	      {
		vt = ptexpcal2ntrees(vExport);
		if (!bpos)
		  vt = LL_make_non_tree(vt);
		vt = LL_make_defsig_tree(mkhonametree(vh), vt);
		STree_addExprToComposition_Post(&vtres, vt);
		if (IS_NILTREE(vtlids))
		  vtlids = LL_make_lident_emptyList();
		vtlids = ntreeList_appendLast(vtlids, mkhonametree(vh));
	      }
	  }
      }
  if (ISNOT_NILTREE(vtlids))
    *vtdecl = LL_make_decsig_tree(STree_mkNullaryTree(oplogical), vtlids);
  else
    *vtdecl = STree_getNilTree();
  return vtres;
}

/**
   @return the Signal Abstract tree representation of a clock table for a graph with boolean clocks (ie endochronous and boolean clocks)
   
   It produces a composition in which each signal of lsig appears with its clock and the definition of this clock. 
   (mkSynchronizedSignalsOfToClock(), mkClockExpression())
   
   @param h (in) assumed to be the branching clock of the clocks of lsig.
   @param lh (in) the clocks of the system
   @param lsig (in) a list of signals 
   @param bWithTick must be "true" when the definition of the local tick (h) must be added to the table, "false" otherwise.

   @note ---UNUSED-----
 */
ntrees DCG2SignalTree::_buildClockTableBool(ptinsth h, ptlinsth lh, ptlinstances lsig, boolean bWithTick, boolean babstract)
{
  ptinsth vh;
  ntrees vt, vabstract;

  // ---UNUSED NOW -----
  vabstract = STree_getNilTree();
  vt = STree_getNilTree();
  if (!CKM->satisfiesFictitiousRoot(h))
    {
      // if (bWithTick) STree_addExprToComposition_Post(&vabstract, tick2Tree(h)); remove Aug 2013 (LBesnard)
      if (mkSynchronizedSignalsOfToClock(h, lsig, pk_true, &vt))
	{
	  /* vt is defined by h ^= x1, ..., xn, with xi in lsig */
	  if (bWithTick && CKM->satisfiesTickClock(h))
	    vt = ntreeList_insertFirst(mkExtraction(CKM->getBooleanSignal(h), pk_false), vt);
	  else if (ntreearity(vt) == 1)
	    vt = ntreeList_insertFirst(STree_copy(nchild(1, vt), pk_false), vt);

	  STree_addExprToComposition_Post(&vabstract, vt);
	}
    }

  /* For each clock of lh : adding the clock definition with it synchronized signals. */
  while (ptlinsth_get_next(&lh, &vh))
    {
      vh = CKM->getRepresentativeFormula(vh);
      if (!CKM->eqClock(vh, h))
	if (CKM->eqClock(CKM->getBranching(h, vh), h))	// why ?
	  mkSignalsClockSynchro(vh, lsig, pk_true, &vabstract);
    }
  return vabstract;
}

/**
   @return the Signal Abstract tree representation of a clock table for a graph.
   It produces a composition in which each signal of lsig appears with its clock and the definition of this clock. 
   @param h (in) assumed to be the branching clock of the clocks of lsig. It may be the fictitious root (polychronous program).
   @param lh (in) the clocks of the system
   @param lsig (in) a list of signals 
*/
ntrees
    DCG2SignalTree::buildClockTable(ptlinsth ExportedClocksBuiltOnIOClocks, ptlinsth ExportedClocksNotBuiltOnIOClocks, ptlinstances lsig,
				    ntrees * vtdecl)
{
  ntrees vtres, synchros;
  ptlinsth lh, lhToCompare;

  lh = EM->getClockOfSignals(lsig);
  lh = ptlinsth_sorted_fusion(lh, ExportedClocksBuiltOnIOClocks);
  lh = ptlinsth_sorted_fusion(lh, ExportedClocksNotBuiltOnIOClocks);

  // G_ClockListToDeclareProjection = markClocksExport(lh, &lhToCompare);

  G_ClockListToDeclareProjection = ExportedClocksBuiltOnIOClocks;
  lhToCompare = ExportedClocksNotBuiltOnIOClocks;

  vtres = STree_getNilTree();
  synchros = mkSignalsClockListSynchro(G_ClockListToDeclareProjection, lsig, pk_true);
  STree_addExprToComposition_Post(&vtres, synchros);

  synchros = mkSignalsClockListSynchro(lhToCompare, lsig, pk_false);
  STree_addExprToComposition_Post(&vtres, synchros);

  // x ^< y for x,y I/O signals and their clocks are not exported.
  synchros = addClockRelation(lhToCompare, lh);
  STree_addExprToComposition_Post(&vtres, synchros);

  // adding boolean definitions.
  synchros = addBooleanDefinitions(lhToCompare, vtdecl);
  STree_addExprToComposition_Post(&vtres, synchros);
  // cleaning
  ptlinsth_free(lh);
  return vtres;
}
