/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/** 
   It adds to the Signal tree model (vtprocess) the explicit dependences of the sub-graph operand(aSubGraph).
*/
    void
 DCG2SignalTree::SubGraphExplicitDependences(ptINode aSubGraph, ntrees vtprocess)
{
  ptldependence lexplicitDep;
  ptdependence aDep;
  ntrees vt, tdependences;
  ptlINode lnodes;
  ptlinsth clocks;

  NM->get_subGraphDefinitions(aSubGraph, &lnodes, &clocks);
  lexplicitDep = AGRAPH->getExplicitDependences();
  tdependences = STree_getNilTree();
  while (lexplicitDep->get_next(&aDep, &lexplicitDep))
    {
      SET_NILTREE(vt);
      if (mkTreeDependence(aDep, lnodes, &vt))
	STree_addExprToComposition_Post(&tdependences, vt);
    }
  STree_addExpressionOfProcessToaModel(vtprocess, tdependences);
}

/** 
    It adds to the Signal Abstract tree representation (process) of a sub-graph (aSubGraph) 
    the declarations of the external models referenced in the external calls (procedures/functions/process) of the sub-graph.
 */
void DCG2SignalTree::SubGraphDeclareExternalCalls(ptINode aSubGraph, ntrees process)
{
  ptlINode ldecl, linstr;
  ptlinsth clocks;
  ptINode vinstr;

  NM->get_subGraphDefinitions(aSubGraph, &linstr, &clocks);
  ldecl = NILLINODE;
  while (ptlINode_get_elem(&linstr, &vinstr))
    ldecl = NM->addExternDecl(vinstr, ldecl);
  appDeclExternProcedures(process, ldecl, NILLINSTANCES);	// To be revisited for the functions.
  appDeclExternNodeCall(process);
}

/**
   @return the Signal Abstract tree representation of the synchronisations of all the signals of 
   the sub-graph. It is a composition, each element is  h ^= x1 ...^= xN where h is a clock 
   and xI signals of the sub-graph.
*/
ntrees DCG2SignalTree::mkClockSynchro(ptINode aSubGraph, boolean bwithTick)
{
  ptlinstances lsig;
  ntrees synchros;

  lsig = ptlinstances_sortedfusion(NM->getInodelInputFlows(aSubGraph), NILLINSTANCES);
  lsig = ptlinstances_sortedfusion(NM->getInodelOutputFlows(aSubGraph), lsig);
  lsig = ptlinstances_sortedfusion(NM->getInodelLocalFlows(aSubGraph), lsig);
  synchros = mkClockSynchro(lsig, bwithTick);
  ptlinstances_free(lsig);
  return synchros;
}

/**
   It adds the Signal Abstract tree representation 
   - of a list of nodes (ln) to a model (process) (buildtree() method)
   - of a list of clocks (lh) to a model clocks2ntrees (clocks2ntrees() method). It is not applied in "flat mode".
*/
void DCG2SignalTree::addNodesAndClocks(ptlINode ln, ptlinsth lh, ntrees process)
{
  ntrees treeClocks;

  buildtree(ln, process);
  SET_NILTREE(treeClocks);
  treeClocks = clocks2ntrees(lh, treeClocks);
  STree_addExpressionOfProcessToaModel(process, treeClocks);
}

/**
   @return "true" when the graph is a DC+ graph, "false" otherwise.
   For a "event-clock graph", it adds to the Signal Abstract tree representation (process) of a sub-graph (aSubGraph)  
   - the synchronisation informations (mkClockSynchro()) to a model (process)
   - the definitions of the clocks/nodes of the sub-graph (addNodesAndClocks()).
   - the explicit dependences associated with the sub-graph (SubGraphExplicitDependences())
*/
boolean DCG2SignalTree::SubGraph2sigTreeEventClock(ptINode aSubGraph, ntrees process, boolean isolated)
{
  ptlINode linstr;
  ptlinsth lh;

  NM->get_subGraphDefinitions(aSubGraph, &linstr, &lh);
  if (isEventClockedFormat(graph_type))
    STree_addExpressionOfProcessToaModel(process, mkClockSynchro(aSubGraph, pk_true));
  else
    {
      lh = NILLINSTH;
      if (isolated)
	STree_addExpressionOfProcessToaModel(process, mkClockSynchro(aSubGraph, pk_false));
    }
  addNodesAndClocks(linstr, lh, process);
  SubGraphExplicitDependences(aSubGraph, process);
  return pk_true;
}

/**
   @return Signal Abstract tree representation of the name of a sub-graph (aSubGraph).
*/
ptident DCG2SignalTree::mknameSubGraph(ptINode aSubGraph)
{
  ptident id, gn;
  ptlpragmas lp;
  ptpragma aPrag;
  ptrepstr EnonceValue;
  ntrees tEnonce;

  if (!NM->get_clusterName(aSubGraph, &id))
    {
      gn = AGRAPH->getName();
      lp = NM->getPragmas(aSubGraph);
      if (getPragmaKeyword(lp, PRAGMA_RUN_ON, &aPrag) || getPragmaKeyword(lp, PRAGMA_SELECTED_OTHERS_PARTITION, &aPrag))
	{
	  tEnonce = getPragmaStatement(aPrag);
	  EnonceValue = nstringof(nchild(1, tEnonce));
	  EnonceValue = ExtractString(EnonceValue);
	  id = createSuffIdent(gn, EnonceValue);
	}
      else
	id = makeIdent_ident_int(gn, NM->get_num(aSubGraph));
    }
  return id;
}

/**
   @return the Signal Abstract tree representation of the reference to a sub-graph (aSubGraph).

   A sub-graph is assumed to be declare as a process (SubGraph2sigTree()). 
   The reference to the sub-graph is a call to this process, with the inputs/outputs 
   of the sub-graph if withIO.
*/
ntrees DCG2SignalTree::mkSubGraphCallTree(ptINode aSubGraph, boolean withIO)
{
  ptident nom;
  ptexpcal lins, louts, lparams;

  nom = mknameSubGraph(aSubGraph);
  lparams = NILEXPCAL;
  if (withIO)
    {
      NM->get_inOutGraphRef(aSubGraph, &lins, &louts);
      return mkCallTree(nom, lparams, lins, louts);
    }
  return mkInstTree(nom, lparams);
}

/**
   It adds to the Signal Abstract tree representation model (process) associated with a
   sub-graph (aSubGraph) the definition of the nodes/clocks elements of the sub-graph.
*/
void DCG2SignalTree::SubGraph2sigTreeBody(ptINode aSubGraph, ntrees process, boolean isolated)
{
  boolean s;
  ptINode x;

  x = setCurrentBlock(aSubGraph);
  s = G_TickIsReferenced;
  G_TickIsReferenced = pk_false;
  SubGraph2sigTreeEventClock(aSubGraph, process, isolated);
  if (G_TickIsReferenced)
    dclBooleanTick(process);
  G_TickIsReferenced = s;
  setCurrentBlock(x);
}

/**
   Imports some pragmas of the graph to the operand model.
   Imported pragmas: DefinedClockHierarchy
*/
void DCG2SignalTree::importGraphPragmas(ntrees process)
{
  ptlpragmas lp;
  ptpragma v;

  lp = lpragmas;
  while (ptlpragmas_get_next(&lpragmas, &v))
    if (isPragmaKeyword(v, PRAGMA_DEFINED_CLOCK_HIERARCHY))
      addPragma(v, process);
}

/**
   It adds to the Signal Abstract tree representation model (process) associated with a
   sub graph (aSubGraph):
   - the pragmas (addPragmas()).
   - when isolated, the declarations of the models of the external references.
*/
void DCG2SignalTree::SubGraph2sigTreeFinalyze(ptINode aSubGraph, ntrees process, boolean isolated)
{
  ptlpragmas lp;

  lp = NM->getPragmas(aSubGraph);
  addPragmas(lp, process);
  if (existPredefinedPragma(lp, PRAGMA_RUN_ON))
    RunOnSetOnSubgraphs = pk_true;
  if (isolated)
    {
      insertDeclarationsFlatMode();
      SubGraphDeclareExternalCalls(aSubGraph, process);
      importGraphPragmas(process);
    }
  addNodeInfos(aSubGraph, STreeModelModule_getName(process));
}

/**
   @return a skeleton of a Signal Abstract tree representation model associated with a
   sub graph (aSubGraph) considered as a program if isolated.
*/
ntrees DCG2SignalTree::SubGraph2sigTreeInitialize(ptINode aSubGraph, boolean isolated)
{
  ntrees process;
  process = STree_mkSkeletonModel(mknameSubGraph(aSubGraph), opprocess);
  if (isolated)
    flatModeInit(process, pk_true);
  setLinkNodeToTree(aSubGraph, process);
  return process;
}

/**
   @return the Signal Abstract tree representation of a sub-graph (aSubGraph)   
   It is represented as a model of process, moreover,
   when isolated, the sub-graph is considered as a program (all the 
   declarations are produced), otherwise it is considered as a sub-process.

   It is produced by the following operations:
   - SubGraph2sigTreeInitialize() that produces a skeleton of a model.
   - SubGraph2sigTreeInterface() that translates the interface
   - SubGraph2sigTreeBody() that translates the definitions. 
   - SubGraph2sigTreeFinalyze() that finalyzes the transaltion.
*/
void DCG2SignalTree::SubGraph2sigTree(ptINode aSubGraph, boolean isolated, ntrees process)
{
  SubGraph2sigTreeInterface(aSubGraph, process, isolated);
  SubGraph2sigTreeBody(aSubGraph, process, isolated);
  SubGraph2sigTreeFinalyze(aSubGraph, process, isolated);
}

/** 
    It adds the Signal Abstract tree representation of the clock table of a sub_graph (aSubGraph) to the
    model (process) associated with the sub-graph.
    It assumes that 
    - the projection informations associated with the sub-graph are available (get_subGraphProjectedSystemInfos()).
    - the graph is not a POLY one; clocks are represented by booleans.
    
    The clock table is produced by buildClockTable().
*/
void DCG2SignalTree::addSubGraphClocksTable(ptINode aSubGraph, ntrees process)
{
  ntrees vtDecl, clocktable;
  ptlinsth lh;
  ptlinstances lsig;

  lh = ptlinsth_duplicate(CKM->getlhroots());
  lh = ptlinsth_sorted_fusion(lh, CKM->getlhplaced());
  lsig = NILLINSTANCES;
  lsig = ptlinstances_sortedfusion(NM->getInodelInputFlows(aSubGraph), lsig);
  lsig = ptlinstances_sortedfusion(NM->getInodelOutputFlows(aSubGraph), lsig);
  lsig = ptlinstances_sortedfusion(NM->getInodelLocalFlows(aSubGraph), lsig);

  clocktable = buildClockTable(NILLINSTH, lh, lsig, &vtDecl);	// A revoir (param 2), il faut mettre Ã  jour l'info de projection avant...
  ptlinstances_free(lsig);
  STree_addExpressionOfProcessToaModel(process, clocktable);
}

/**
   It adds to the Signal Abstract tree representation (process) associated with a sub-graph (aSubGraph)
   the definition of the interface.
   Moreover, when bIsolated, the constants of the sub-graph are also added.
*/
void DCG2SignalTree::SubGraph2sigTreeInterface(ptINode aSubGraph, ntrees process, boolean bIsolated)
{
  ptinterface vinterface;
  ntrees inter;

  vinterface = NM->get_interface(aSubGraph);
  inter = mktreeinterface(vinterface);
  nchangchild(process, inter, 2);
  appdeclsig(process, ptinterface_get_parameters(vinterface), 1, pk_false);
  appendLoc(vinterface, process);
  if (bIsolated)
    mktreeconstants(process, ptinterface_get_constants(vinterface));
}
