/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_

/** 
    Constructor.
 */
DCG2SignalTree::DCG2SignalTree()
{
  DEBUG_MODE = pk_false;
  genInterface = pk_false;
  TraceabilityMode = pk_false;
  KernelExprMode = pk_false;
  flatMode = pk_false;
  SpecMode = pk_true;
  notInitialized = pk_true;
  nb_max_list_elem = CST_NB_MAX_LIST_ELEM;	// limit in the SIGNAL syntactic analyzer.
  innerModel = pk_false;
}

/**  Destructor. */
_VIRTUAL_ DCG2SignalTree::~DCG2SignalTree()
{
}

/**
   The interface (inputs/outputs) tree will be created if b.
   @return the value of the mode before the call.
 */
boolean DCG2SignalTree::setInterfaceMode(boolean b)
{
  boolean o_b;
  o_b = genInterface;
  genInterface = pk_true; // b;
  return o_b;
}

/**
   The hierarchy is flatenned if b.
   @return the value of the mode before the call.
 */
boolean DCG2SignalTree::setFlatMode(boolean b)
{
  boolean o_b;

  // return flatMode;

  o_b = flatMode;
  flatMode = b;
  return o_b;
}

/**
   The expresssions are decomposed if b; new auxillary signals are created.
   Note that the derived processes are not rewritten into Basic Signal.
   @return the value of the mode before the call.
 */
boolean DCG2SignalTree::setKernelExprMode(boolean b)
{
  boolean o_b;

  o_b = KernelExprMode;
  KernelExprMode = b;
  if (b)
    nb_max_list_elem = 2;
  else
    nb_max_list_elem = CST_NB_MAX_LIST_ELEM;
  return o_b;
}

/**
   The expresssions are decomposed if b; new auxillary signals are created.
   Note that the derived processes are not rewritten into Basic Signal.
   @return the value of the mode before the call.
 */
boolean DCG2SignalTree::setSpecMode(boolean b)
{
  boolean o_b;

  o_b = SpecMode;
  SpecMode = b;
  return o_b;
}

/** 
    @return the SIGNAL Abstract Tree representation of a graph (in_DCG).
    When the parameter bflat 
    - is "true", the clock hierarchy is recursively explicited as a list of subprocesses,
    - is "false", the clock hierarchy is flatenned (for example, if a graph contains some labels, it is necessary to flatten the hierarchy).  
    .
    Signals with null clocks, unused signals, not well defined signals in arrays are also explicited. 
*/
ntrees DCG2SignalTree::graph2SignalTree(ptDCG in_DCG, ptident abstractTree_name)
{
  ntrees in_DCG_abstractTree, cur_DCG_abstractTree;
  boolean done;
  ptDCG cur_DCG;
  ptlDCG lvisitedGraphs;
  pointer elem;
  ptlinstances lTypesCst, lt, lfuncs;
  ptlINode lextref;

  DCG2SignalTreeini(in_DCG);
  if (abstractTree_name == NILIDENT)
    abstractTree_name = createSuffIdent(mkGraphNameWithSubFormat(in_DCG), traSuffix);
  setCurrentBlock(NILINODE);
  lIGraph = NILLPOINTER;	// list of the graphs to translate.
  lvisitedGraphs = NILLPOINTER;

  in_DCG_abstractTree = STree_mkSkeletonModel(abstractTree_name, in_DCG->getKind());
  if (!in_DCG->getModuleGraphs(&lIGraph))
    mustBePrinted(in_DCG);

  lTypesCst = ptlinstances_copy(in_DCG->getConstants());
  if (in_DCG->getUserTypesOneLevel(&lt))
    lTypesCst = EM->mergeTypesCst(lt, lTypesCst);

  cur_DCG_abstractTree = in_DCG_abstractTree;
  while (ptlPointer_get_elemNext(&lIGraph, &elem))
    {
      if (!ptlPointer_IsIn(lvisitedGraphs, elem))
	{
	  ptlPointer_insert_elem(&lvisitedGraphs, elem);
	  cur_DCG = (ptDCG) elem;
	  abstractTree_name = cur_DCG->getName();
	  done = pk_false;
	  if (in_DCG != cur_DCG)
	    {
	      done = isKnown_SIGNAL_tree(cur_DCG, &cur_DCG_abstractTree);
	      if (!done)
		cur_DCG_abstractTree = STree_mkSkeletonModel(abstractTree_name, cur_DCG->getKind());
	      STree_addProcessDeclaration(in_DCG_abstractTree, cur_DCG_abstractTree);	// added to the locals of the main.
	    }
	  if (!done)
	    {
	      IGraph2Process(cur_DCG, cur_DCG_abstractTree);
	      addSpec(cur_DCG_abstractTree, cur_DCG);
	      lTypesCst = ptlinstances_sortedfusion(cur_DCG->getConstants(), lTypesCst);
	      if (cur_DCG->getUserTypesOneLevel(&lt))
		lTypesCst = EM->mergeTypesCst(lt, lTypesCst);
	    }
	}
    }
  lextref = in_DCG->getListExternalProcedure(&lfuncs);
  appDeclExternProcedures(in_DCG_abstractTree, lextref, lfuncs);
  mkDeclCstTypes(lTypesCst, in_DCG_abstractTree);
  ptlPointer_free(lvisitedGraphs);
  if (!innerModel)
    STree_addGeneralComments(in_DCG_abstractTree);
  return STree_updateModule(in_DCG_abstractTree);
}

/** 
    @return the SIGNAL Abstract Tree representation of a graph (g).
    \see{ :graph2SignalTree(ptDCG, ptident)}
*/
ntrees DCG2SignalTree::graph2SignalTree(ptDCG g)
{
  return graph2SignalTree(g, NILIDENT);
}

/** 
    @return the SIGNAL Abstract Tree representation of a graph (in_DCG).
    When the parameter bflat 
    - is "true", the clock hierarchy is recursively explicited as a list of subprocesses,
    - is "false", the clock hierarchy is flatenned (for example, if a graph contains some labels, it is necessary to flatten the hierarchy).  
    .
    Signals with null clocks, unused signals, not well defined signals in arrays are also explicited. 
*/
ntrees DCG2SignalTree::graph2SignalTree(ptDCG in_DCG, const char *optionSuffix)
{
  ptident new_DCGId;

  if (optionSuffix == (char *)NULL)
    new_DCGId = NILIDENT;
  else
    new_DCGId = createSuffIdent(mkGraphNameWithSubFormat(in_DCG), string2repstr(optionSuffix));
  return graph2SignalTree(in_DCG, new_DCGId);
}

/** 
    @return the abstraction("Black Box" or "Grey Box") representation of a graph (g).
*/
ntrees DCG2SignalTree::graphAbstraction2SignalTree(ptDCG g)
{
  ntrees vhierarchy, vtcurrent;
  ptDCG aGraph;
  pointer elem;
  ptlDCG lvisitedGraphs;

  DCG2SignalTreeini(g);
  lIGraph = NILLPOINTER;
  lvisitedGraphs = NILLPOINTER;
  if (!g->getModuleGraphs(&lIGraph))
    return IGraph2AbtractionTree(g);
  vhierarchy = STree_mkSkeletonModel(createSuffIdent(g->getName(), abstractSuffix), g->getKind());
  while (ptlPointer_get_elemNext(&lIGraph, &elem))
    if (!ptlPointer_IsIn(lvisitedGraphs, elem))
      {
	ptlPointer_insert_elem(&lvisitedGraphs, elem);
	aGraph = (ptDCG) elem;
	if (g != aGraph)
	  {
	    vtcurrent = IGraph2AbtractionTree(aGraph);
	    STree_addProcessDeclaration(vhierarchy, vtcurrent);
	  }
      }
  ptlPointer_free(lvisitedGraphs);
  return vhierarchy;
}

/** 
    @return "true" if it exists some undeclared external calls in a graph (aGraph), "false" otherwise.
    @param UndeclaredModule (out): a module in which the undeclared externall calls are declared.
    @param aGraph (in): a graph, it can be a graph of a module.

    It is used to generate a module of undefined model declarations. Such object exists in the graph when 
    this graph is produced (by example) by a morphism: for a reference to an external model \em foo in the 
    original graph, a reference to an external model \em foo' is generated, but  \em foo' is undeclared.
        
    
    The user will complete to compile the result of a morphism the .
	
*/
boolean DCG2SignalTree::getTreeUndeclaredExternals(ptDCG aGraph, ntrees * UndeclaredModule)
{
  ptlinstances lfuncs, l2;
  ptlINode lextref, l1;
  ptident id;
  ntrees vt;
  pointer elem;
  ptDCG g;
  ptlDCG lg, lg1, lvisitedGraphs;

  DCG2SignalTreeini(aGraph);
  lextref = NILLINODE;
  lfuncs = NILLINSTANCES;
  lg = NILLDCG;
  lvisitedGraphs = NILLPOINTER;
  if (!aGraph->getModuleGraphs(&lg))
    ptlPointer_insert_elem(&lg, (pointer) aGraph);
  while (ptlPointer_get_elemNext(&lg, &elem))
    if (!ptlPointer_IsIn(lvisitedGraphs, elem))
      {
	ptlPointer_insert_elem(&lvisitedGraphs, elem);
	g = (ptDCG) elem;
	l1 = g->getListExternalProcedure(&l2);
	lextref = fusionExternalProcs(lextref, l1);
	if (g->getReferencedGraphs(&lg1))
	  lg = ptlPointer_fusion(lg, lg1);
	ptlPointer_remove_elem(&lg, elem);
      }
  ptlPointer_free(lvisitedGraphs);
  SET_NILTREE(vt);
  if (!mkUndeclaredDeclarationExternProcedures(lextref, lfuncs, &vt))
    return pk_false;
  id = createSuffIdent(aGraph->getName(), string2repstr("UNDEFINED_EXTERNAL_CALL"));
  *UndeclaredModule = STree_mkSkeletonModel(id, opmodu);
  STree_setLocalDeclarations(*UndeclaredModule, vt);
  return pk_true;
}

/**
   @return the Signal Abstract Tree representation of the cycles and the "false cycles" of a graph (g). 
   The entry points of the cycles are specified by a list of list of nodes (cycles).
   The entry points of the "false cycles" are specified by a list of list of nodes (fcycles).
*/
ntrees DCG2SignalTree::mkTreeCycles(ptDCG g, ptlPointer cycles, ptlPointer fcycles)
{
  DCG2SignalTreeini(g);
  return buildTreeCycles(cycles, fcycles);
}

/**
   @return the SIGNAL Tree representation of a sub-graph (aSubGraph), a node of a graph (g).
 */
ntrees DCG2SignalTree::graphSubGraph2SignalTree(ptDCG g, ptINode aSubGraph)
{
  ntrees vt;
  boolean b;

  DCG2SignalTreeini(g);
  setCurrentBlock(NILINODE);
  lIGraph = NILLPOINTER;	// list of the graphs that should be printed (node Call) */
  b = setFlatMode(pk_true);
  vt = SubGraph2sigTreeInitialize(aSubGraph, pk_true);
  SubGraph2sigTree(aSubGraph, pk_true, vt);
  vt = DCG2SignalTreeFinal(vt);
  setFlatMode(b);
  return vt;
}

/**
   @return the exclusive clocks of a graph (g). 
   The operation is applied for each referenced graph in g. 
   The result is for each of them given by a list of pairs of exclusive clocks ( x ^# y ).
*/
ntrees DCG2SignalTree::exclusiveClocks(ptDCG g)
{
  ntrees vhierarchy, vtcurrent;
  ptDCG aGraph;
  pointer elem;
  ptlDCG lvisitedGraphs;

  DCG2SignalTreeini(g);
  lIGraph = NILLPOINTER;
  lvisitedGraphs = NILLPOINTER;
  if (!g->getModuleGraphs(&lIGraph))
    return mkExclusiveClocks(g);
  vhierarchy = STree_mkSkeletonModel(createSuffIdent(g->getName(), ecSuffix), g->getKind());
  while (ptlPointer_get_elemNext(&lIGraph, &elem))
    if (!ptlPointer_IsIn(lvisitedGraphs, elem))
      {
	ptlPointer_insert_elem(&lvisitedGraphs, elem);
	aGraph = (ptDCG) elem;
	if (g != aGraph)
	  {
	    vtcurrent = mkExclusiveClocks(aGraph);
	    if (ISNOT_NILTREE(vtcurrent))
	      STree_addProcessDeclaration(vhierarchy, vtcurrent);
	  }
      }
  ptlPointer_free(lvisitedGraphs);
  vhierarchy = DCG2SignalTreeFinal(vhierarchy);
  return vhierarchy;
}

/** 
    @return the SIGNAL Abstract Tree representation of the automaton assiciated with a graph (in_DCG).
*/
ntrees DCG2SignalTree::automaton2SignalTree(ptDCG in_DCG, const char *optionSuf)
{
  ptAutomaton vautomaton;
  ntrees vt_vautomaton;
  ptlinstances lt, lTypesCst;
  ptinterface vinterface;

  if (!in_DCG->getAutomaton(&vautomaton))
    return NILTREE;
  initTranslatorFor(vautomaton);
  vt_vautomaton = STree_mkSkeletonModel(vautomaton->getName(), opautomaton);
  vinterface = vautomaton->getInterface();
  mktreeinterface(vinterface, vt_vautomaton);
  appendLoc(vinterface, vt_vautomaton);
  mkBodyAutomaton(vautomaton, vt_vautomaton);

  lTypesCst = ptlinstances_copy(ptinterface_get_constants(vinterface));
  if (vautomaton->getUserTypes(&lt))
    lTypesCst = EM->mergeTypesCst(lt, lTypesCst);
  mkDeclCstTypes(lTypesCst, vt_vautomaton);
  mkStatesDeclaration(vautomaton, vt_vautomaton);
  STree_addGeneralComments(vt_vautomaton);
  return vt_vautomaton;
}
