/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   For a delay vexpcal := e $ val.
   @return val if (val > v), val + upperboundDelay(e) otherwise (by construction e is necessary a delay).

   This method produces also 
   - the signal resexp such that vexpcal=S1 $ N1  init V1, S(n-1)= resexp $ Nn with v <=(N1+..+Nn) 
   - the list of the initial values l_init={V1,...}
*/
  integer downstreamRetiming::upperboundDelay (ptexpcal vexpcal, ptexpcal * resexp, integer v, ptlinstances * l_init)
{
  ptINode nodeSignal;
  ptexpcal expelt, expg, expd, dval, wval, res_int, vinit;
  ptlinstances l_init_int1, l_init_int2;
  integer val, val_int;

  nodeSignal = EM->getNode (vexpcal);
  NM->get_delay_arg (nodeSignal, &expg, &expd, &dval, &wval, &vinit);

  val = EM->getValue (dval);
  if (val >= v)
    {
      mklistinit (expg, val, l_init);	/* increasing */
      *resexp = expd;
      return val;
    }
  l_init_int1 = NILLINSTANCES;
  l_init_int2 = NILLINSTANCES;
  val_int = val;
  mklistinit (expg, val, &l_init_int1);
  l_init_int1 = ptlinstances_reverse (l_init_int1);
  val = upperboundDelay (expd, &res_int, (v - val_int), &l_init_int2);
  while (ptlinstances_get_next (&l_init_int1, &expelt))
    l_init_int2 = ptlinstances_push_front (l_init_int2, expelt);
  *l_init = l_init_int2;
  *resexp = res_int;
  return (val_int + val);
}


/** @return the max delay associated with a signal (vexpcal)
    If the signal is 
    - not defined by a delay, the value is 0,
    - defined by a delay <em>(x:= y $ n) </em>, the value is equal to <em> n+maxDelayValue(y).</em> 
*/
integer
downstreamRetiming::maxDelayValue (ptexpcal vexpcal)
{
  ptINode nodeSignal;
  ptexpcal expg, expd, dval, wval, v;
  integer val;

  if (!EM->IsSignal (vexpcal))
    return 0;
  nodeSignal = EM->getNode (vexpcal);
  if (!NM->is_delay (nodeSignal))
    return 0;

  NM->get_delay_arg (nodeSignal, &expg, &expd, &dval, &wval, &v);
  val = EM->getValue (dval);
  return (val + maxDelayValue (expd));
}


/**
   It adds the operand expression (exp) in front of the list lopres when it is a leaf (constant or a signal),
   otherwise it recursively recalls this method on the operand of exp (for a synchronous one).
*/
void
downstreamRetiming::getLeafsOf (ptexpcal exp, ptlinstances * lopres)
{
  ptexpcal exp1, exp2;
  cloper cl;

  if (EM->isNILEXPCAL (exp))
    return;

  cl = EM->getOperatorClass (exp);
  switch (cl)
    {
    case constant:
    case variabl:
    case call:
      *lopres = ptlinstances_push_front (*lopres, exp);
      break;

    case oper1:
      exp1 = EM->getUnaryArg (exp);
      getLeafsOf (exp1, lopres);
      break;
    case oper2:
      if (EM->getOperator (exp) != opelem)
	{
	  exp1 = EM->getBinaryArgstAt (exp, 1);
	  getLeafsOf (exp1, lopres);

	  exp2 = EM->getBinaryArgstAt (exp, 2);
	  getLeafsOf (exp2, lopres);
	};
      break;
    default:
      ;
    }
}


/**
  @return the list of the leafs of the operand expression (exp) produced by the getLeafsOf() method.
*/
ptlinstances
downstreamRetiming::getLeafsOfExpression (ptexpcal exp)
{
  ptlinstances l_op;

  l_op = NILLINSTANCES;
  getLeafsOf (exp, &l_op);
  l_op = ptlinstances_reverse (l_op);
  return l_op;
}


/**
   @return a new signal defined by the expression 
   <em> vexpcal $ val_delay init vinit </em>, \n
   where \e vinit is produced by mkexpcalinit() then simplifyExpr().\n   
   The definition is inserted in the graph.
*/
ptexpcal
downstreamRetiming::mknewdelay (ptexpcal vexpcal, integer val_delay, ptlinstances * l_init)
{
  ptexpcal expdelay, newexp, expinit;
  ptinsth horl;
  ptINode vinstr;

  newexp = EM->mkIdentExp (NILIDENT, EM->getClassIdentifier (vexpcal), EM->getType (vexpcal), NM->mk_node_undefined (), CKM->getUndefClock (), NILBACKANNOTATION);

  expinit = mkexpcalinit (0, l_init);
  expinit = simplifyExpr (expinit);
  EM->setInitialValue (newexp, expinit);
  vinstr = NM->mk_node_delay (newexp, vexpcal, EM->mkIntegerConstant (val_delay), EM->mkIntegerConstant (0), expinit, NILBACKANNOTATION);
  horl = EM->getClock (vexpcal);
  NM->insertNode (vinstr, horl);
  return newexp;
}


/**
   It builds:
   - the list l_op the list of the signals useful for the rewriting of exp in (exp') $ val_max
   - the list l_list_init of the initial values for builting the previous expression.It will be used to produced the initial value of the delay.
   This method may introduce some new signals (delays) in the graph to build exp'.
*/
void
downstreamRetiming::mklistoperande (ptexpcal exp, integer val_max, ptlinstances * l_op, ptexpcal * l_list_init)
{
  ptlinstances l_oper;
  ptexpcal expint, resexp, nexp, opexp, liste_init_res;
  ptlinstances liste_init, liste_initres, liste_init_int;
  integer val, i;

  liste_init_int = NILLINSTANCES;
  *l_op = NILLINSTANCES;
  l_oper = getLeafsOfExpression (exp);
  while (ptlinstances_get_next (&l_oper, &opexp))
    {
      liste_init = NILLINSTANCES;
      if (EM->getOperatorClass (opexp) == variabl)
	val = upperboundDelay (opexp, &resexp, val_max, &liste_init);
      else
	{
	  val = val_max;
	  resexp = opexp;
	  for (i = 1; i <= val; i++)
	    liste_init = ptlinstances_push_front (liste_init, opexp);
	}

      if (val == val_max)
	{			/* pas besoin de creer un nv sig */
	  *l_op = ptlinstances_push_front (*l_op, resexp);

	  liste_init_res = mkexpcalfromptlinstances (&liste_init);
	  liste_init_int = ptlinstances_push_front (liste_init_int, liste_init_res);
	}
      else
	{			/* besoin de creer un nv sig */
	  liste_initres = NILLINSTANCES;
	  for (i = 1; i <= val_max; i++)
	    {
	      if (ptlinstances_get_next (&liste_init, &expint));
	      liste_initres = ptlinstances_push_front (liste_initres, expint);
	    };
	  liste_initres = ptlinstances_reverse (liste_initres);

	  nexp = mknewdelay (resexp, (val - val_max), &liste_init);
	  *l_op = ptlinstances_push_front (*l_op, nexp);

	  liste_init_res = mkexpcalfromptlinstances (&liste_initres);
	  liste_init_int = ptlinstances_push_front (liste_init_int, liste_init_res);
	};
    };

  liste_init_int = ptlinstances_reverse (liste_init_int);
  *l_list_init = mkexpcalfromptlinstances (&liste_init_int);
  *l_op = ptlinstances_reverse (*l_op);
}



/**


 \note vexpcal is assumed to be defined by an equation node, and the definntion is a synchronous expression.
 
 For a signal (vexpcal), defined by vexpcal := expdef, 
 - it synthetizes the leafs (constant/signals) of expdef (getLeafsOfExpression() method),
 - it calculates the max (n) of the value delay on this set of signals (this value is the common factor) (see maxDelayValue() method)
 - if the max value is greater than 0, the definition of the signal (vexp) is modified.
      - it synthetizes the new leafs and the initial values (mklistoperande() method)
      - it produces a new expression (expdef') in which the current leafs are replaced by the the new one (mkexpcald() method)
      - a new signal (x) is introduced such that x:= expdef' 
      - the definition of vexpcal becomes vexpcal := x $ n
      
 \note this method rewrites the following definition <em> x := (a$1) + (b$1) + c$1 </em> in  <em> x := (a + b + c) $1 </em>

 \todo{To be completed for function calls: y := foo2(a$1 init 6, b $1 init 2) may be rewritten in y := foo2(a,b) $ 1 init foo(6,2)
 but not this one <em> x := (a$1) + (b$1) + c </em>.
*/
void
downstreamRetiming::rewritesig (ptexpcal vexpcal)
{
  ptINode nodeSignal;
  ptinsth horl;
  ptexpcal expg, expd, opexp, l_list_init, newexpd, newexpdelay, newsig, expinit;
  ptlinstances l_oper, loperres;
  integer val, chgt, val_max;
  ptINode aNode;
  cloper vcl;

  nodeSignal = EM->getNode (vexpcal);
  NM->get_equation_arg (nodeSignal, &expg, &expd);

  l_oper = NILLINSTANCES;
  l_oper = getLeafsOfExpression (expd);
  val_max = 9999;
  chgt = 0;

  // Getting the common delay factor (if val_max = 0 the retiming is not applied).
  while ((val_max > 0) && ptlinstances_get_next (&l_oper, &opexp))
    {
      vcl = EM->getOperatorClass (opexp);
      if (vcl == call)
	val_max = 0;
      else if (vcl == variabl)
	{
	  val = maxDelayValue (opexp);
	  chgt = 1;
	  if (val < val_max)
	    val_max = val;
	}
    }
  // modifying the definition?
  if ((val_max > 0) && (chgt == 1))
    {
      // Getting the new leafs (loperres) and the initial values (l_list_init)
      // For x := (a$ init a0 +b$ init b0 +c$ init c0 ), it produces <(a,b,c), a0+b0+c0>
      l_list_init = NILEXPCAL;
      loperres = NILLINSTANCES;
      mklistoperande (expd, val_max, &loperres, &l_list_init);
      // Adding a new node u := (a+b+c)
      newexpd = mkexpcaldinit (expd, &loperres);
      horl = EM->getClock (vexpcal);
      newsig = NM->insertEquation (getIdentNil (), EM->getType (newexpd), newexpd, horl, horl, &aNode);
      // x := u $  val_max init expinit
      expinit = mkinitialisation (expd, &l_list_init);
      newexpdelay = mkexpcaldelay (newsig, expinit, val_max);
      NM->modifiedDefinition (vexpcal, newexpdelay, horl, horl);
      EM->setInitialValue (vexpcal, expinit);
    }
}

/**
   @return "true" if  the expression of definition of the operand signal (vexpcal) is a synchronous 
   expression, "false" otherwise.

   When it returns "true", the output list l contains the signals referenced in the definition of vexpcal.
   \note vexpcal is assumed to be defined by an equation node.
*/
boolean
downstreamRetiming::isSynchronousExpr (ptexpcal vexpcal, ptlinstances * l)
{
  ptINode nodeSignal;
  ptexpcal expg, expd;

  nodeSignal = EM->getNode (vexpcal);
  NM->get_equation_arg (nodeSignal, &expg, &expd);
  if (EM->getOperatorClass (expd) == oper1 || EM->getOperatorClass (expd) == oper2)
    {				/* synchronous expression. */
      *l = EM->getSignalsReferencedIn (expd);
      return pk_true;
    };
  return pk_false;
}

/**
   @return "false" when a (data-flow) cycle  is detected, "true" otherwise;
   The cycle detection is achieved by marking the signal at the beginning of the method (if is is not marked), 

   For a signal (vexp) defined by 
   - a delay, it recalls the operation on the delayed signal (for cycle detection)
   - an equation, it applies the retiming operation if its definition satisfies the criteria (see isSynchronousExpr() method)
   to each signal it references. If there is no cycle, the rewritting is called (rewritesig()) method).

   The operand signal (vexp) is removed from the list unvisitedSignalList and it is unmarked at the end of this method.
*/
boolean
downstreamRetiming::retimingSignal (ptlinstances * unvisitedSignalList, ptexpcal vexp)
{
  ptINode nodeSignal;
  ptlinstances l_oper;
  ptexpcal opexp, expg, expd, dval, wval, v;
  boolean b;

  // if (EM->getOperatorClass(vexp) == call) return pk_false;
  if (!EM->IsSignal (vexp))
    return pk_true;
  if (EM->isMarked (vexp, retimMark))
    return pk_false;

  EM->addMark (vexp, retimMark);
  b = pk_true;
  nodeSignal = EM->getNode (vexp);
  switch (NM->get_oper (nodeSignal))
    {
    case opret:		/* delay */
      NM->get_delay_arg (nodeSignal, &expg, &expd, &dval, &wval, &v);
      if (ptlinstances_isin_exp (*unvisitedSignalList, expd))
	if (EM->getOperator (expd) != opelem)
	  b = retimingSignal (unvisitedSignalList, expd);
      break;

    case opdefsig:		/* equation */
      l_oper = NILLINSTANCES;
      if (isSynchronousExpr (vexp, &l_oper))
	{
	  while (ptlinstances_get_next (&l_oper, &opexp))
	    if (ptlinstances_isin_exp (*unvisitedSignalList, opexp))
	      if (!retimingSignal (unvisitedSignalList, opexp))
		b = pk_false;

	  /* if all the operands satisfy the "retiming": the expression is rewritten */
	  if (b)
	    rewritesig (vexp);
	}
      else
	b = pk_false;
      break;
    };
  (*unvisitedSignalList) = ptlinstances_remove_exp (*unvisitedSignalList, vexp);
  EM->removeMark (vexp, retimMark);
  return b;
}



/**
   It applies the retiming for each signal of a list (unvisitedSignalList).
   The list is modified: it contains the unvisited signals.
*/
void
downstreamRetiming::retimingSignalList (ptlinstances * unvisitedSignalList)
{
  ptexpcal vexpcal;

  while (ptlinstances_get_next (unvisitedSignalList, &vexpcal))
    retimingSignal (unvisitedSignalList, vexpcal);
}


/**
  It applies the retiming for a clock (vclk): all the synchronous signals with the clock are
  visited (retimingSignalList()).
*/
void
downstreamRetiming::retimingClock (ptinsth vclk)
{
  ptlinstances lexp, liste_sig_non_traite;

  lexp = CKM->getSynchronousSignals (vclk);
  liste_sig_non_traite = ptlinstances_copy (lexp);
  retimingSignalList (&liste_sig_non_traite);
}

/**
   TO DO (Jan 2014- LBesnard)
*/
void
downstreamRetiming::applyRetimingFunctionCall ()
{

}
