/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor.
   The functionality is applied to a graph (g).
*/
  statesVariables::statesVariables (ptDCG g)
{
  GraphLevelType t;
  AGRAPH = g;
  EM = AGRAPH->getExpressionManager ();
  CKM = AGRAPH->getClockManager ();
  NM = AGRAPH->getNodeManager ();
  Outputs = AGRAPH->getOutputs ();
  t = AGRAPH->getFormatLevel ();
  dcplusFormat = ((t == DCG_BASIC) || (t == DCG_POLY) || (t == DCG_ENDO));
}

/**
   Destructor.
*/
statesVariables::~statesVariables ()
{
}


/** For each cell of the subtree with vhfrom as root, it creates a cell at the clock vhto.
    The definition of original cell are modified.
    \note It assumes that vhfrom is under vhto in the hierarchy of clocks.

   - y := x cell c is replaced by (| mx := x cell vhto | mx ^= vhto |)
   - and the definition of y becomes y := mx when ^y
*/
boolean statesVariables::extendCellToCellFromClockToClock (ptinsth vhfrom, ptinsth vhto)
{
  cellModeTranslation = pk_false;
  return extendCellFromClockToClock (vhfrom, vhto);
}

/** 
    For each cell of the subtree with vhfrom as root, it creates a kernel-cell at the clock vhto.
    The definition of original cell are modified.
    \note It assumes that vhfrom is under vhto in the hierarchy of clocks.
    
    - y := x cell c is replaced by (| mx := (x when ^x) default zmx | zmx := mx $ 1 | mx ^= vhto |)
    - and the definition of y becomes y := mx when ^y
*/
boolean statesVariables::extendCellToKernelFromClockToClock (ptinsth vhfrom, ptinsth vhto)
{
  cellModeTranslation = pk_true;
  return extendCellFromClockToClock (vhfrom, vhto);
}


/** It creates states variables at a clock (vhto) from the tree with hfrom as root.
    The ExtendDelaysToClock() method is applied to each clocks of the hierarchy dominated by vhfrom.
    \note It assumes that vhfrom is under vhto in the hierarchy of clocks.
*/
boolean statesVariables::extendDelayFromClockToClock (ptinsth vhfrom, ptinsth vhto)
{
  ptinsth
    vh;
  ptlinsth
    lchild;
  boolean
    b;

  /* for a clock (vhfrom) */
  if (vhfrom != vhto)
    b = ExtendDelaysToClock (vhfrom, vhto);
  else
    b = pk_true;
  if (b)
    {
      /* For all children of clock of vhform */
      lchild = CKM->getChildren (vhfrom);
      while (ptlinsth_get_next (&lchild, &vh) && b)
	b = extendDelayFromClockToClock (vh, vhto);
    }
  return b;
}

/** 
    It defines the signals defined by STATUS intrinsic process by state variables.
    For x := STATUS(y), x becomes a state variable defined by x::= y when ^y as partial definition.
    It assumes that x is initialized.
*/
void
statesVariables::status2stateVariables ()
{
  ptlINode lNodes, lexternals;
  ptINode node;
  ptexpcal x, y, vin, vout;
  ptident id;
  ptinsth vclock;
  ptlinstances lfuncs;

  if (NM->getProcCalls (&lNodes))
    {
      lexternals = AGRAPH->getListExternalProcedure (&lfuncs);
      while (ptlINode_get_elem (&lNodes, &node))
	if (NM->is_intrinsicfunc (node))
	  {
	    NM->get_nameCall (node, &id);
	    if (isStatusId (id))
	      {
		x = NM->get_leftOperand (node);
		EM->getNaryArgs (&x, &vout);
		NM->get_rightOperand (node, &y);
		EM->getNaryArgs (&y, &vin);
		if (!EM->getExprClock (vin, &vclock))
		  vclock = EM->getClock (vout);
		NM->resetNodeForModification (node);	// removeNode(node);
		mkStateVariable (vout, vin, vclock);
		ptlINode_remove_elem (&lexternals, node);
	      }
	  }
      AGRAPH->setListExternalProc (lexternals, lfuncs);
    }
}


/** 
    It applies the stateVariable2Status() method for each state variables of the seected graph

    \note It assumes that the state variables are defined by ONE partial definition.
    (ie the PartialDefinitionsMerging() method has been applied on the graph).
*/
void
statesVariables::stateVariables2Status ()
{
  ptexpcal vs, mem;
  ptlinstances ls;

  ls = AGRAPH->getStatesVar ();
  while (ptlinstances_get_next (&ls, &vs))
    {
      mem = stateVariable2Status (vs);
      EM->substituteStateVariablePreviousBy (vs, mem);
    }
  AGRAPH->setStatesVar (NILLINSTANCES);
}
