/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   It adds an intrinsic "status" node defined at the clock vh, such that
   "zmemvx := status{vinit} (zvx)", where
      - zmemvx is the element of lout,
      - zvx is the element of lin ,
      - and vinit the element of lparam.
*/
  ptINode statesVariables::addStatusNodeAt (ptexpcal lparam, ptexpcal lin, ptexpcal lout, ptinsth vh)
{
  return NM->mkIntrinsicUnsafeAt (getStatusId (), lparam, lin, lout, vh, NILBACKANNOTATION);
}


/**
   @return a new signal with same type than x, same class identifier than x and with vclk as clock.
*/
ptexpcal statesVariables::mkCopyIdentExp (ptexpcal x, ptinsth vclk)
{
  return EM->mkIdentExp (NILIDENT, EM->getClassIdentifier (x), EM->getType (x), NM->mk_node_undefined (), vclk, NILBACKANNOTATION);
}

/** @return "true" if the signal is selected, "false" otherwise.
    By default, operation is applied to each signal. 
 */
boolean statesVariables::isSelected (ptexpcal vsig)
{
  return pk_true;
}

/** It adds a state variable with partial definition.
   vy ::= vx when vhx.
*/
void
statesVariables::addStateVar (ptexpcal vy, ptexpcal vx, ptinsth vhx)
{
  ptINode n;

  EM->setStateTag (vy, pk_true);

  EM->addMark (vy, PART_DEFINITION);
  n = NM->mk_node_equation (vy, vx, vhx, NILLINSTANCES, NILBACKANNOTATION);
  n = EM->getNode (vy);
  NM->insertNode (n, vhx);
  EM->removeMark (vy, PART_DEFINITION);

  AGRAPH->addStatesVar (vy);
}

/**
  @return a 1-delay signal zmx of mx synchroniezd to clkDelay clock.
*/
ptexpcal statesVariables::makeDelayOneAt (ptexpcal mx, ptinsth clkDelay, ptexpcal vinit)
{
  ptexpcal
    zmx;			// vpass;
  ptINode
    node;

  zmx = mkCopyIdentExp (mx, clkDelay);
  EM->setClock (zmx, clkDelay);
  EM->setInitialValue (zmx, vinit);
  //  vpass = EM->mkTernaryExp(opret, EM->mkBinaryExp(opinit, mx, vinit, NILBACKANNOTATION), EM->mkIntegerConstant(1) , NILEXPCAL,NILBACKANNOTATION);
  node = NM->mk_node_delay (zmx, mx, EM->mkIntegerConstant (1), NILEXPCAL, vinit, NILBACKANNOTATION);
  NM->insertNode (node, clkDelay);
  return zmx;
}

/** 
   @return a 1-delayed-memory of a given delayed-signal to a given clock for "boolean" format.
   @param vx (in) signal that must be memorized as a state-variable.
   @param vhx (in) the clock of vx.
   @param vinit (in) initial value of the memory.
   @param vh (in) the clock of the memory.
   @param memx (out) cell of vx at vhx.
   The memory (zmemx) is defined by 
   - memx = (vx when vhx) default zmemx
   - zmemx = memx $ VALDELAY  init VINIT
   - synchro{ memx, vh }  

   \note restriction: 1-delay (VALDELAY), no window.
*/
ptexpcal statesVariables::makeMemoryofAt (ptexpcal vx, ptinsth vhx, ptexpcal vinit, ptinsth vh, ptexpcal * memx)
{
  ptexpcal
    mx,
    zmx;
  ptINode
    node;
  ptinsth
    vh1;

  // Creation de la variable d'etat mx 
  mx = mkCopyIdentExp (vx, vh);
  EM->setClock (mx, vh);
  // zmx = mx $ 1 init vinit | synchro{ zmx, mx, vh } 
  zmx = makeDelayOneAt (mx, vh, vinit);
  // mx definition : mx := ( vx when vhx ) default zmx
  node = NM->mk_node_equation (mx, vx, vhx, ptlinstances_push_front (NILLINSTANCES, vx), NILBACKANNOTATION);
  CKM->getComplExtraction (vhx, &vh1);	// tick ^- vhx
  node = NM->mk_node_equation (mx, zmx, vh1, ptlinstances_push_front (NILLINSTANCES, zmx), NILBACKANNOTATION);
  NM->insertNode (node, vh);
  *memx = mx;
  return zmx;
}


/** Creating and returning zmemvx  defined by zmemvx := status{zvx}, zmemvx init init(zvx)
   at the clock vhto.
   "status" is said intrinsic call.
*/
ptexpcal statesVariables::makeStatusCall (ptexpcal zvx, ptexpcal vx, ptinsth clkvx, ptinsth vhto, ptINode * createdNode)
{
  ptexpcal
    lentren,
    lsortren,
    lparam,
    zmemvx,
    e1,
    vinit;

  zmemvx = mkCopyIdentExp (zvx, vhto);
  EM->setClock (zmemvx, vhto);
  vinit = EM->getInitialValue (zvx);
  EM->setInitialValue (zmemvx, vinit);

  lparam = EM->mkNaryExp (oplident, vinit, NILEXPCAL, NILBACKANNOTATION);
  e1 = EM->mkWhenOrCellExp (opextr, vx, clkvx);	// vx can be a constant !
  lentren = EM->mkNaryExp (oplident, e1 /*vx */ , NILEXPCAL, NILBACKANNOTATION);
  lsortren = EM->mkNaryExp (oplident, zmemvx, NILEXPCAL, NILBACKANNOTATION);
  (*createdNode) = addStatusNodeAt (lparam, lentren, lsortren, vhto);
  return zmemvx;
}

/** Defining a state variable S ::= vx when vhx. If vy is an output, an internal state variable (S) is
   build (an output cannot be a state variable) and vy is defined by this statevar.
   Otherwise S is vy and the references to vy are replaced by a state variable reference expression (vy?).
   Assumed : vy is initialized.
*/
void
statesVariables::mkStateVariable (ptexpcal vy, ptexpcal vx, ptinsth vhx)
{
  ptexpcal mx, refStateVar;
  ptinsth clock_y;

  clock_y = EM->getClock (vy);
  if (ptlinstances_isin_exp (Outputs, vy))
    {
      // an output cannot be a statevar.
      mx = mkCopyIdentExp (vy, CKM->getUndefClock ());
      addStateVar (mx, vx, vhx);	// mx ::= vx when vhx
      refStateVar = EM->mkUnaryExp (opvvarin, mx, EM->getBackAnnotation (mx));
      NM->modifiedDefinition (vy, refStateVar, clock_y, clock_y);
    }
  else
    {
      CKM->removeFromClock (clock_y, vy, EM->IsEvent (vy));
      addStateVar (vy, vx, vhx);	// my ::= vx when vhx
      // Reference to vy is replaced by  "vy ?"
      refStateVar = EM->mkUnaryExp (opvvarin, vy, EM->getBackAnnotation (vy));
      EM->expReplace (vy, refStateVar, clock_y);
    }
}

/**
   It copies the partitionnning infos from a node (fnode) to an another one (tnode).
*/
void
statesVariables::movePragmasFromTo (ptINode fnode, ptINode tnode)
{
  ptlpragmas LPrag;
  ptpragma aPragma;

  LPrag = NM->getPragmas (fnode);
  if (getPragmaKeyword (LPrag, PRAGMA_RUN_ON, &aPragma))
    NM->addPragma (tnode, aPragma);
}

/** 
   @return true if the extension to a given clock of a delay node is achieved.
   @param delayNode (in) a node assumed to be a delay.
   @param vhto (in) a clock at which the delay is memorized (clock greater than the clock of the delay)
   The memorization is defined according to the sub-format level:
   - dc+ format : a SIGNAL state variable is created with partial definition assigned to it.
   - other format: the 1-delayed-memory (zmemx) is defined by
     -(| memx = (vx when ^vhx) default zmemx 
     - | zmemx = memx $ VALDELAY  init VINIT
     - | synchro{ memx, vh }  
     -|)

   \note Restriction: 1-delay (VALDELAY), no-window.

*/
boolean statesVariables::ExtendADelayToClock (ptINode delayNode, ptinsth vhto)
{
  ptexpcal
    vy,
    vx,
    vret,
    vfen,
    vinit;
  integer
    valret;
  ptinsth
    vhx;
  ptINode
    n;

  NM->get_delay_arg (delayNode, &vy, &vx, &vret, &vfen, &vinit);	/* <vy> = <vx> $ <vret> window <vfen> */
  if (vfen != NILEXPCAL)
    return pk_false;		/* vy := xxxxx window vfen */
  if (vret == NILEXPCAL)
    return pk_false;		/* blindage : vy := vx $ 0 window 0 */
  if (!EM->IsAssigned (vret))
    return pk_false;
  valret = EM->getValue (vret);
  if (valret == 0)
    return pk_false;		/* case: vy := vx $ 0 window 0 */
  vhx = EM->getClock (vy);
  if (isSelected (vy))
    if (dcplusFormat)
      {
	// DC+ format
	NM->resetNodeForModification (delayNode);
	mkStateVariable (vy, vx, vhx);
      }
    else			// Boolean formats(bDC+,....)
      {
	NM->modifiedDefinition (vy, makeStatusCall (vy, vx, vhx, vhto, &n), vhx, vhx);
	movePragmasFromTo (delayNode, n);
      }

  return pk_true;
}

/** @return true if the extension to a given clock of all the delay-node of a list of nodes is achieved.
   @param ln (in) a list of nodes.
   @param clkTo (in) a clock at which the delays are memorized (clock greater than the clock of such a delay).
*/
boolean statesVariables::ExtendDelaysToClock_loc (ptlINode ln, ptinsth clkTo)
{
  ptINode
    aNode;
  boolean
    b;

  b = pk_true;
  while (ptlINode_get_elem (&ln, &aNode) && b)
    if (NM->is_delay (aNode))
      if (!ExtendADelayToClock (aNode, clkTo))
	b = pk_false;
  return b;
}


/**
   @return the list of the nodes associated with a clock (vclk).
*/
ptlINode statesVariables::getNodesAssignedTo (ptinsth vclk)
{
  ptgfd
    vgfd;

  vgfd = CKM->getClockSubGraph (vclk);
  if (vgfd == NILGFD)
    return NILLINODE;
  return ptgfd_get_nodes (vgfd);
}

/** @return true if the extension to a given clock of all the delay-node assigned to a clock is achieved.
    @param clkFrom (in) a clock for which the extension of delay-nodes is applied.
    @param clkTo (in) a clock at which the delay are memorized (clock greater than clkFrom ).
*/
boolean statesVariables::ExtendDelaysToClock (ptinsth clkFrom, ptinsth clkTo)
{
  return ExtendDelaysToClock_loc (getNodesAssignedTo (clkFrom), clkTo);
}


/** It defines the state variables (aStatevar) by a status call.
    For aStatevar a state variable, mx is defined by mx := STATUS{init(aStatevar)}(aStatevar), aStatevar becomes a signal.
    Assumed: the state variable is defined by ONE partial definition.
    (ie the PartialDefinitionsMerging() method has been applied on the graph).
*/
ptexpcal statesVariables::stateVariable2Status (ptexpcal aStatevar)
{
  ptinsth
    Tick,
    h;
  ptexpcal
    v;
  ptINode
    n;

  Tick = CKM->getTickClock ();
  h = NM->get_nodeClock (EM->getNode (aStatevar));
  EM->setStateTag (aStatevar, pk_false);
  if (h == Tick)
    v = makeDelayOneAt (aStatevar, h, EM->getInitialValue (aStatevar));
  else
    v = makeStatusCall (aStatevar, aStatevar, h, Tick, &n);
  return v;
}
