/* 
   Authors: LoÃÂ¯c Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
   http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_ DCG_basic::DCG_basic()
{
}

DCG_basic::DCG_basic(ptDCG_basic gb)
{
  vDCG = gb->getDCG();
}

DCG_basic::DCG_basic(ptDCG g)
{
  vDCG = g;
  if (g != NILDCG)
    {
      setArrayClocks();
      computeControlClocks();
    }
}

DCG_basic::DCG_basic(ptDCG g, boolean bstrong, boolean bComputeControlClocks)
{
  vDCG = g;
  if (g != NILDCG)
    {
      if (bstrong) setArrayClocks();
      if (bComputeControlClocks) computeControlClocks();
    }
}

ptDCG DCG_basic::getDCG()
{
  return vDCG;
}

void DCG_basic::setDCG(ptDCG g)
{
  vDCG = g;
}

/** 
    It creates states variables at the clock \b hto from the tree of root \b hfrom.
    \warning it Assumes that \b hfrom is under \b hto in the hierarchy of clocks.
    @return "true" if the operation succeeds, "false" otherwise.
*/
boolean DCG_basic::extendDelayClock(ptinsth hfrom, ptinsth hto)
{
  ptstatesVariables stateBuilder;

  // boolean extendDelayClock ( ptinsth hfrom, ptinsth hto ) is a DCG_basic function
  // comment apply extendDelayClock ( asignal, ptinsth hto ) for all signals the clock of which is hfrom
  stateBuilder = new statesVariables(getDCG());
  return stateBuilder->extendDelayFromClockToClock(hfrom, hto);
}

/** 
    sub-graph to graph translation.
    @return the new graph obtained by the copy of the nodes and the clock hierarchy of the sub-graph (aSubGraph).
    The name of the extacted graph is fixed by taking into account the pragmas assigned to the subgraph (cluster, RunOn ,...)
*/
ptDCG_basic DCG_basic::copySubgraph2Graph(ptINode aSubGraph)
{
  return _copySubgraph2Graph_(aSubGraph, pk_false, pk_false, pk_true);
}

/**
   @return the new graph obtained by the copy of the nodes and the sub-clock hierarchy (or the complete hierarchy if bExportClockSystem) of a sub-graph (aSubGraph) of the global graph.
   The name of the extacted graph is fixed by taking into account the pragmas assigned to the subgraph (cluster, RunOn ,...)
   When the parameter bExportClockSystem is "true" all the clock system is exported, otherwise only the clocks of the subgraph.
   When the parameter breduce is "true" the new formula may be simplified in construction (elementary simplifications are applied.
*/
ptDCG_basic DCG_basic::cloneSubgraph2Graph(ptINode aSubGraph, boolean bExportClockSystem, boolean breduce)
{
  return _copySubgraph2Graph_(aSubGraph, pk_true, bExportClockSystem, breduce);
}

/** 
    It replaces the inputs of the graph by calls.
    Signals declared in the interface of the program becomes local signals. 
*/
void DCG_basic::makeCallsForInputs()
{
  //LPT should be replaced by building a container CP that calls P

  ptlinstances l;
  ptexpcal vsig;
  ptINode node;

  require_expr_manager(EM);
  require_node_manager(NM);

  l = getInputs();
  while (ptlinstances_get_next(&l, &vsig))
    {
      EM->belongsLocalFlow(vsig);
      node = EM->getNode(vsig);
      NM->makeInputByInstrinsicRead(node);
    }
  setInputs(NILLINSTANCES);
}

/**
   For an input I of the graph, it adds the equation I:= New_I and New_I is a new input signal.
   It allows to isolated the "read" without changing the expressions in the graph.
   The name of the new signal New_I is the name of the original signal, and a new name is assigned to I.
 */
void DCG_basic::makeNewInputs()
{
  // LPT should be replaced by building a container CP that calls P

  ptlinstances nl, l;
  ptexpcal vsig, vNewInput;
  ptINode node;
  ptinsth h;
  ptident vsigId, vNewInputId;

  require_expr_manager(EM);
  require_node_manager(NM);
  l = getInputs();
  setInputs(NILLINSTANCES);
  nl = NILLINSTANCES;
  while (ptlinstances_get_next(&l, &vsig))
    {
      h = EM->getClock (vsig);
      vsigId =  EM->getName(vsig);
      vNewInput = EM->mkIdentCopy (vsig, NM->mk_node_undefined (), h);
      vNewInputId = EM->getName(vNewInput);
      EM->setName(vNewInput, vsigId);
      EM->setName(vsig, vNewInputId);
      EM->belongsInFlow(vNewInput);
      node = NM->insertEquation(vsig, vNewInput, h, h);
      NM->addMark (node, DontOptimize);

      EM->belongsLocalFlow(vsig);
      node = EM->getNode(vsig);
      
      NM->mk_node_input(vNewInput, NILINODE, NM->getBackAnnotation(node));
      nl = ptlinstances_insert(nl, vNewInput);
      changePragmas(vsig, vNewInput, PRAGMA_TOPOLOGY);
    }
  setInputs(nl);
}


/** 
    It replaces the inputs/outputs of the graph by calls.
    Signals declared in the interface of the program becomes local signals. 
*/
void DCG_basic::makeCallsForInputsOutputs()
{
  //LPT should be replaced by building a container CP that calls P

  ptlinstances l;
  ptexpcal vsig;
  ptlINode ln;
  ptINode node;

  makeCallsForInputs();
  require_expr_manager(EM);
  require_node_manager(NM);

  l = getOutputs();
  while (ptlinstances_get_next(&l, &vsig)) EM->belongsLocalFlow(vsig);
  setOutputs(NILLINSTANCES);

  if (NM->getOutputs(&ln))
    while (ptlINode_get_elem(&ln, &node))
      NM->makeOutputByInstrinsicWrite(node);
}

/**
   Specific colouring for sequential code generator (optimisations).
   The assert nodes are visible for code generation when bCheck (code will be generated for verifying assumptions at run time).   
   It uses the \b GencodeMark predefined mark.
   \see ColouringGenCode
*/
void DCG_basic::CodeGenColouring(boolean bCheck)
{
  ptDCGColouringGenCode x;

  x = new ColouringGenCode(getDCG());
  x->apply(GencodeMark, bCheck);
  delete x;
}

/** 
    @return a partial copy of the current graph by copying selected objects.
    @param vMark (in) : mark used for selecting.
    @param id (in) : suffix used to built the name of the new (duplicte) graph. When equal to NILIDENT 
    the new name is equal to the name of the original graph.
    @param usedMark (out) : mark used to keep feedback information between new and original graph. 
*/
ptDCG_basic DCG_basic::copySelected(boolean exportHierarchy, typeMark vMark, ptident id, typeMark * usedMark)
{
  ptDuplicator aDuplicator;
  ptDCG g;

  // TO BE COMPLETED with transitive closure,...
  aDuplicator = new Duplicator();
  g = aDuplicator->copySelected(getDCG(), vMark, exportHierarchy);
  ptident idg;
  idg = getName();
  if (eq_idents(id, NILIDENT))
    g->setName(copy_ident(idg));
  else
    g->setName(makeIdent_ident_ident(idg, id));
  (*usedMark) = aDuplicator->getUsedMark();
  delete aDuplicator;
  return new DCG_basic(g, pk_false, pk_true);
}

/**
   It performs some constant expression evaluations 
   See:
   - NodeManager::spatialProcessingIntrinsicEvaluation(),
   - delaysRewriting()
   - ExprManager::intrinsicEvaluation()
   - substituteEvaluatedExpression()
   
   The expressions are then updated (ExprManager::rewritingExprEH())
*/
void DCG_basic::constantEvaluation()
{
  require_node_manager(NM);
  require_expr_manager(EM);
  NM->spatialProcessingIntrinsicEvaluation();
  EM->intrinsicEvaluation();
  getDCG()->makeConstantSignals();
  delaysRewriting();
  substituteEvaluatedExpression();
}

/** 
    @return a copy of the current graph (defined as a morphism).
    \warning (\e restriction) This method can be applied to a graph of a process, not to a graph of a module.
*/
ptDCG_basic DCG_basic::copy()
{
  ptDCG gCopy;
  ptDuplicator o;

  // --Take care of shared signal (no reference to ?)
  o = new Duplicator();
  gCopy = o->copy(getDCG());
  gCopy->setName(newIdent_stringSuffixed(getName(), "COPY"));
  delete o;
  return new DCG_basic(gCopy, pk_false, pk_true);
}

/** 
    Restriction: The graph must be in DCGBool sub-format.
 
    It performs the user-partitionning by analyzing the pragma "RunOn i " where i is a string.
    Nodes annotated by the same value are in the same class or sub-graph.
    @param ln (out) is a list of "sub-graph" nodes produced by the partitionning.
    @return the value "false" if the operation failed, true otherwise.
    \warning (\e restriction) This method can be applied to a graph of a process, 
    not to a graph of a module (set of processes).
 
*/
boolean DCG_basic::partitionningRunOn(ptlINode * ln)
{
  // Take care of shared signal (no reference to ?)
  pK_messenger->globalMessageStr(codeNotYetImplemented, string2repstr("DCG_basic::userPartitionningRunOn"));
  return pk_false;
}

/**
   Similar to produce_tree(char *) with TRA suffix for the name of generated model.
*/
ntrees DCG_basic::produce_tree()
{
  return produce_tree((char *)NULL);
}

void  DCG_basic::setKeepTrace(boolean b)
{
  ptTraceabilityManager tm;  // For test LBesnard
  tm = getTraceabilityManager ();
  tm ->setModeKeepSubstitutions(b);
}

/**
   This  method can be applied to a graph of a module (set of processes) or a process.
   It can be applied to any format. 
   It builds an external representation of the graph as an SIGNAL abstract tree using some internal options.
   @return the Signal abstract tree which can be printed using the \see{decsig} function.

   \see{unparse, graph2SignalTree}.

   Assumed: the options of the translator are set before this call.
*/
ntrees DCG_basic::produce_tree(const char *optionSuf)
{
  ntrees vt;
  int date;
  ptTraceabilityManager tm;  // For test LBesnard

  tm = getTraceabilityManager ();
  tm->beginFunctionality();

  tm->printmodifiedSinceLastMarkedFunction();

  date = pK_messenger->messageBeginOperVerbose(infoGraph2SignalTree);
  vt = pK_Graph2SignalTreeBuilder->graph2SignalTree(getDCG(), optionSuf);
  tm->setCurrentAsReference();

  pK_messenger->messageEndOperVerbose(date);
  return vt;
}

/** 
    Similar to unparse(char *optionSuf), with NULL string as parameter.
*/
boolean DCG_basic::unparse()
{
  return unparse((char *)NULL);
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    This is the entry point of Signal code generation. It can be applied for any format.
    It produces code in the file foo_X.SIG, where X denotes a codification of the graph level, completed with the optionsuff suffix.

    The trees memory is cleaned after the printing.
    \see{produce_tree}
*/
boolean DCG_basic::unparse(const char *optionSuf)
{
  ntrees DCGAbstractTree;
  boolean b;

  ntrees_memoryStoreIndex();
  DCGAbstractTree = produce_tree(optionSuf);
  
  b = pK_SignalPrinter->setPrinterContext(pk_false);	// to print the annotations of the trees.
  SignalUnparse(DCGAbstractTree, infotra);
  pK_SignalPrinter->setPrinterContext(b);

  ntrees_memoryRestoreIndex();
  return pk_true;
}

/** 
    It builds an external representation of a sub-graph of the current graph as an SIGNAL abstract tree.
    
    @param  aSubGraph (in) a sub-graph node.
    @return the Signal abstract tree. This tree can be printed using the \see{decsig} function.
*/
ntrees DCG_basic::produceSubgraph(ptINode aSubGraph)
{
  return pK_Graph2SignalTreeBuilder->graphSubGraph2SignalTree(getDCG(), aSubGraph);
}

/** 
    It returns informations on the clock hierarchy.
    For a DCG_basic graph, such information is not available.
*/
boolean DCG_basic::printCHStatistics()
{
  // No infos about the hierarchy for such a graph.
  return pk_true;
}

/**
   Goal: Building the containers around the processes.
   
   It adds the communication nodes in the extracted graphs of sub-graphs.
   For two subgraphs SG1,SG2 elements of lSubGraphs, and the associated extracted graph (G1, G2) elements of lsg,
   for each signal, output of SG1 and input of SG2:
   - a communication node (Send) is inserted in G1,
   - a communication node (Receive) is inserted in G2.
   
   Assumed: lsg is an ordered list of the graphs, extractions of the subgraphs of lsubGraphs: the i-th 
   element of lsg is the graph associated with the i-th element of lSubGraphs.

   \see addCommunicationsFromTo
*/
boolean DCG_basic::communicationContainment(ptlINode lSubGraphs, ptlDCG lsg)
{
  ptCommManager comObj;
  boolean b;

  comObj = new CommManager(getDCG());
  b = comObj->AddCommContainers(lSubGraphs, lsg);
  delete comObj;
  return b;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    \copydoc expressionExpandOneLevel
*/
boolean DCG_basic::expressionExpand()
{
  groupPartialDefinitions();
  splitDataStructure();
  for_all_graphs_do_br(expressionExpandOneLevel(), breturn);
  return breturn;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    \copydoc transformBooleanClocks2EventsOneLevel
*/
boolean DCG_basic::transformBooleanClocks2Events()
{
  for_all_graphs_do_br(transformBooleanClocks2EventsOneLevel(), breturn);
  return breturn;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    \copydoc transformAssertions2EventsOneLevel
*/
boolean DCG_basic::transformAssertions2Events()
{
  for_all_graphs_do_br(transformAssertions2EventsOneLevel(), breturn);
  return breturn;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    \copydoc hasUnsolvedClockConstraintsOneLevel
*/
boolean DCG_basic::hasUnsolvedClockConstraints()
{
  for_all_graphs_do_br(hasUnsolvedClockConstraintsOneLevel(), breturn);
  return breturn;
}

/** 
    \copydoc removeSignalEqualitiesOneLevel
*/
boolean DCG_basic::removeSignalEqualities()
{
  for_all_graphs_do_br(removeSignalEqualitiesOneLevel(), b);
  return b;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    \copydoc removeSignalClones
*/
boolean DCG_basic::removeSignalClones()
{
  for_all_graphs_do_br(removeSignalClonesOneLevel(), apply);
  return apply;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    \copydoc substituteExpr2VarOneLevel
*/
boolean DCG_basic::substituteExpr2Var(int maxRef)
{
  for_all_graphs_do_br(substituteExpr2VarOneLevel(maxRef), apply);
  return apply;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    It realizes elementary optimizations by deleting objects of the graph which are 
    not used to produce (resp. consume) outputs (resp. inputs) of the program. 
*/
boolean DCG_basic::nonProductiveSignalElimination()
{
  for_all_graphs_do(nonProductiveSignalEliminationOneLevel(pk_true));
  buildClockProcessInterface();
  return pk_true;
}

/** 
    \copydoc moveIn2OutDelaysOneLevel
*/
boolean DCG_basic::moveIn2OutDelays()
{
  for_all_graphs_do(moveIn2OutDelaysOneLevel());
  return pk_true;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    \copydoc expandModelReferenceOneLevel
*/
boolean DCG_basic::expandModelReference(boolean bAsSubGraph)
{
  pointer elem;
  boolean recomm, result;
  ptlDCG lg;
  ptDCG aGraph;
  ptINode aNodeCall;
  ptlINode lNodes;

  lg = NILLDCG;
  result = pk_false;
  if (!getModuleGraphs(&lg))
    ptlPointer_insert_elem(&lg, (pointer) getDCG());
  while (ptlPointer_get_elemNext(&lg, &elem))
    {
      aGraph = (ptDCG) elem;
      pK_messenger->messageIdent(mexpandNodeCall, getName());
      require_node_manager_of(aGraph, NM);
      recomm = NM->getNodeCalls(&lNodes);

      while (recomm)
	{
	  recomm = pk_false;
	  while (ptlINode_get_elem(&lNodes, &aNodeCall))
	    if (expandModelReferenceOneLevel(aNodeCall, !bAsSubGraph))
	      {
		recomm = pk_true;
		result = pk_true;
	      }
	  if (recomm)
	    NM->getNodeCalls(&lNodes);
	}
    }
  removeSignalEqualities();
  return result;
}

/** 
    This  method can be applied to a graph or a module.
    \copydoc groupPartialDefinitionsOneLevel
    \see {splitDefaultDefinitions}
*/
boolean DCG_basic::groupPartialDefinitions()
{
  for_all_graphs_do(groupPartialDefinitionsOneLevel());
  return pk_true;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    \copydoc splitDefaultDefinitionsOneLevel
*/
boolean DCG_basic::splitDefaultDefinitions()
{
  for_all_graphs_do_br(splitDefaultDefinitionsOneLevel(), b);
  if (b)
    buildClockProcessInterface();
  return b;
}

/** 
    This  method can be applied to a graph or a module.
    \copydoc dependenceReinforcingOneLevel
*/
boolean DCG_basic::dependenceReinforcing()
{
  for_all_graphs_do_br(dependenceReinforcingOneLevel(), b);
  return b;
}

/**
   This  method can be applied to a graph or a module.
    \copydoc ModelReference2NodeReferenceOneLevel
*/
boolean DCG_basic::transformModelReferences2NodeReferences()
{
  for_all_graphs_do_br(transformModelReferences2NodeReferencesOneLevel(), b);
  return b;
}

/** 
    It produces the morphim of the current graph according to the definitions assigned of the Signal operators 
    given in the "ht" table. 
    When bSimul is "true", the current graph is modified (selected booleans become outputs).
   
    @return true if the operation succeeds, false otherwise.
    \warning (\em restriction) this operation is available for a graph (not for a module).
*/
boolean DCG_basic::applyTabularMorphism(ptHomomorphismTable ht, ptDCG_basic * gresult, boolean bSimul)
{
  // -- TO DEFINE
  pK_messenger->globalMessageStr(codeNotYetImplemented, string2repstr("DCG_basic::applyTabularMorphism"));
  return pk_false;
}

/** 
    \copydoc breakCPGOneLevel
*/
boolean DCG_basic::breakCPG()
{
  for_all_graphs_do(breakCPGOneLevel());
  return pk_true;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    \copydoc changeEnumerated2IntegersOneLevel
*/
boolean DCG_basic::changeEnumerated2Integers()
{
  for_all_graphs_do_br(changeEnumerated2IntegersOneLevel(), b);
  return b;
}

/**
   This  method can be applied to a graph of a module (set of processes) or a process.
   \copydoc unparseCyclesOneLevel ()
*/
boolean DCG_basic::unparseCycles()
{
  for_all_graphs_do_br(unparseCyclesOneLevel(), b);
  return b;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    For each graph of the application,

    \copydoc buildClockProcessInterfaceOneLevel
*/
boolean DCG_basic::buildClockProcessInterface()
{
  for_all_graphs_do(buildClockProcessInterfaceOneLevel());
  setExternalReferences();
  return pk_true;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    \copydoc ClockCalculusOneLevel
*/
boolean DCG_basic::ClockCalculus(int ilevelmax)
{
  for_all_graphs_do(ClockCalculusOneLevel(ilevelmax));
  if (getFormatLevel() == DCG_BASIC)
    setFormatLevel(DCG_POLY);
  return pk_true;
}

/** 
    It calls ClockCalculus(int) with a default value for the depth of solving.
*/
boolean DCG_basic::ClockCalculus()
{
  return ClockCalculus(5);
}

/**  
     @return "true" if the operation succeeds, "false" otherwise.
     It calls suitable marking and extraction functions to get 
     This is the entry point of the code generation for the SIGALI prover.
     The code is generated in the file foo.z3z (foo is the name of the process).
     Moreover, if some specific intrinsic functions of SIGALI (see library $SignalLib_Sigali/SIGALI.SIG),
     are referenced in the program, the file foo_CMD.z3z is generated. It contains : 
     - some initialization commands of SIGALI tool and 
     - the set of properties that have to be proved and/or synthesized.
     \warning (\e restriction) This method can be applied to a graph of a process, not to a graph of a module.
*/
boolean DCG_basic::buildF3Abstraction(ptDCG_basic * F3Abstraction)
{
  ptmkBoolOther mkparc;
  ptlINode ln, lSubGraphs;
  ptINode sigaliPart, otherPart;

  // Partitionning in booleans+sigali nodes (sigaliPart) and others (otherPart).
  mkparc = new mkBoolOther(getDCG());
  ln = getTaggedNodes(PRAGMA_SIGALI);
  mkparc->addSelected(ln);
  lSubGraphs = mkparc->apply();
  ptlINode_free(ln);
  finalyzeBinPart(lSubGraphs, &sigaliPart, &otherPart);
  // extracting the "sigali" part
  (*F3Abstraction) = cloneSubgraph2Graph(sigaliPart, pk_true, pk_true);
  (*F3Abstraction)->setName(newIdent_stringSuffixed(getName(), "F3"));
  // Cleaning the graph (added Jan 2014)
  subGraphFlattenning(sigaliPart);
  subGraphFlattenning(otherPart);
  delete mkparc;
  return pk_true;
}


/**  
     @return "true" if the operation succeeds, "false" otherwise.
     It calls suitable marking and extraction functions to get 
     This is the entry point of the code generation for the SIGALI prover.
     The code is generated in the file foo.z3z (foo is the name of the process).
     Moreover, if some specific intrinsic functions of SIGALI (see library $SignalLib_Sigali/SIGALI.SIG),
     are referenced in the program, the file foo_CMD.z3z is generated. It contains : 
     - some initialization commands of SIGALI tool and 
     - the set of properties that have to be proved and/or synthesized.
     \warning (\e restriction) This method can be applied to a graph of a process, not to a graph of a module.
*/
boolean DCG_basic::buildSMTAbstraction(ptDCG_basic * smt2Abstraction)
{
  ptmkBoolOther mkparc;
  ptlINode lSubGraphs;
  ptINode smt2Part , otherPart;

  // Partitionning in booleans+ integers nodes (smt2Part) and others (otherPart).
  mkparc = new mkBooleanBoundedIntegerOther(getDCG());
  // ln = getTaggedNodes(PRAGMA_SIGALI); mkparc->addSelected(ln);  // export  ??
  lSubGraphs = mkparc->apply();
  // ptlINode_free(ln);
  finalyzeBinPart(lSubGraphs, &smt2Part, &otherPart);
  // extracting the "smt2" part
  (*smt2Abstraction) = cloneSubgraph2Graph(smt2Part, pk_true, pk_false);
  (*smt2Abstraction)->setName(newIdent_stringSuffixed(getName(), "SMT2"));
  // Cleaning the graph (added Jan 2014)
  subGraphFlattenning(smt2Part);
  subGraphFlattenning(otherPart);
  delete mkparc;
  return pk_true;
}


/**
   It analyzes the detected static cycles (computed by the isStaticCycleFree() method).
   The null clock cycles (pseudo-cycles) are isolated. The method getPseudoCycles() gives them after the analyzeCycles()  method.
*/
void DCG_basic::analyzeCycles()
{
  pK_messenger->message(infoCycleAnalysis);
  for_all_graphs_do(analyzeCyclesOneLevel());
}

/**
   It builds the transitive closure reduced to the input/output signals of the graph.
   It calls DCG_basic::buildTransitiveClosure(ptlinstances) with the I/O signal.
   The dependences are conditionned by clocks. These clocks are not placed in the hierarchy.
*/
void DCG_basic::buildIOTransitiveClosure()
{
  ptlinstances ls;
  if (!isGraphOfExternalProcess())
    {
      ls = ptlinstances_sortedfusion(getOutputs(), ptlinstances_copy(getInputs()));
      buildTransitiveClosure(ls);
      ptlinstances_free(ls);
    }
}

/**
   It builds the transitive closure for the signals of the operand list (lsig).
   The dependences are conditionned by clocks. These clocks are not placed in the hierarchy.
*/
void DCG_basic::buildTransitiveClosure(ptlinstances lsig)
{
  ptClosureCalculus mkparc;

  mkparc = new ClosureCalculus(getDCG());
  mkparc->transitiveClosure(lsig);
  delete mkparc;
}

/**
   It initializes the graph before modifications.
*/
void DCG_basic::initGraphForModifications()
{
  require_clock_manager(CKM);
  CKM->initSystemForModification();
}

/**
   Each node of a subgraph (aSubGraph) is inserted in the sub-graph of its clock.
*/
void DCG_basic::subGraphFlattenning(ptINode aSubGraph)
{
  require_node_manager(NM);
  NM->subGraphFlattenning(aSubGraph);
}

/**
   @return "true" if the formula (vh) is built on clock of I/O signals clocks, "
*/
boolean DCG_basic::isExprOnIOClocks(ptinsth vh, ptinsth vhSubForm, ptlinsth lh, typeMark vMark)
{
  ptinsth vharg1, vharg2;
  ptlinstances ls;
  ptexpcal aSig;
  boolean b, bres;

  // To avoid loops on recursive formulae, the mark vMark is used.
  require_expr_manager(EM);
  require_clock_manager(CKM);

  if (vh != vhSubForm)
    if (ptlinsth_sorted_isin_elem(lh, vhSubForm))
      return pk_true;
  if (CKM->satisfiesMarked(vhSubForm, vMark))
    return (vh == vhSubForm);
  CKM->addMark(vhSubForm, vMark);

  switch ((int)(CKM->getOperator(vhSubForm)))
    {
    case sub:
    case nosub:
      ls = CKM->getSignalsOfCondition(vhSubForm);
      // if all S in ls are I/O -> pk_true
      bres = pk_true;
      while (bres && ptlinstances_get_next(&ls, &aSig))
	if (!EM->IsInputFlow(aSig) && !EM->IsOutputFlow(aSig))
	  bres = pk_false;
      break;

    case infer:
      if (!CKM->getMinusArgs(vhSubForm, &vharg1, &vharg2))
	CKM->getInfArgs(vhSubForm, &vharg1, &vharg2);
      bres = (isExprOnIOClocks(vh, vharg1, lh, vMark) && isExprOnIOClocks(vh, vharg2, lh, vMark));
      break;

    case super:
      CKM->getSupArgs(vhSubForm, &vharg1, &vharg2);
      bres = (isExprOnIOClocks(vh, vharg1, lh, vMark) && isExprOnIOClocks(vh, vharg2, lh, vMark));
      break;

    case vari:
      ls = CKM->getSynchronousSignals(vhSubForm);
      bres = pk_false;
      while (!bres && ptlinstances_get_next(&ls, &aSig))
	bres = (EM->IsInputFlow(aSig) || EM->IsOutputFlow(aSig));
      break;

    case hcomp:
    case predef:
      bres = pk_false;
      break;

    default:
      pK_messenger->Caseerror(currentLine, currentFile);
    }
  CKM->removeMark(vhSubForm, vMark);
  return bres;
}

ptlinsth DCG_basic::getClocksToExportFor(ptinsth vh, ptlinsth lhToExport)
{
  boolean b;
  pkClockType vop;
  ptinsth vharg1, vharg2;

  require_expr_manager(EM);
  require_clock_manager(CKM);

  if (!CKM->satisfiesSub(vh))
    {
      b = pk_true;
      vop = CKM->getOperator(vh);
      if (vop == infer)
	{
	  if (!CKM->getMinusArgs(vh, &vharg1, &vharg2))
	    CKM->getInfArgs(vh, &vharg1, &vharg2);
	}
      else if (vop == super)
	CKM->getSupArgs(vh, &vharg1, &vharg2);
      else
	b = pk_false;
      if (b)
	{
	  if (!CKM->satisfiesSub(vharg1))
	    {
	      lhToExport = ptlinsth_sorted_insert_elem(lhToExport, vharg1, &b);
	      if (b)
		lhToExport = getClocksToExportFor(vharg1, lhToExport);
	    }
	  if (!CKM->satisfiesSub(vharg2))
	    {
	      lhToExport = ptlinsth_sorted_insert_elem(lhToExport, vharg2, &b);
	      if (b)
		lhToExport = getClocksToExportFor(vharg2, lhToExport);
	    }
	}
    }
  return lhToExport;
}

/**
   @return the sub-list of the clocks of lh built on I/O clocks.
   The list (lotherclock) contains the element of lh not built on I/O clocks.
*/
ptlinsth DCG_basic::IOSplitExportedClocks(ptlinsth lh, ptlinsth * lotherclock)
{
  ptinsth vh;
  ptlinsth lhToExport, lhSave;
  boolean b;
  typeMark vMark;

  vMark = getMark();
  lhSave = lh;
  lhToExport = NILLINSTH;
  (*lotherclock) = NILLINSTH;
  while (ptlinsth_get_next(&lh, &vh))
    {
      if (isExprOnIOClocks(vh, vh, lhSave, vMark))
	{
	  lhToExport = ptlinsth_sorted_insert_elem(lhToExport, vh, &b);
	  lhToExport = getClocksToExportFor(vh, lhToExport);
	}
      else
	(*lotherclock) = ptlinsth_sorted_insert_elem(*lotherclock, vh, &b);
    }
  return lhToExport;
}

/** Assumes: The signals x1 and x2 are synchronous and x1 ^= x2 ^= vh.
   It adds an exception node using the intrinsic process pK_exception((x1 = x2)) at the clock vh.
   @return the created node.
*/
ptINode DCG_basic::addException(ptexpcal x1, ptexpcal x2, ptinsth vh)
{
  ptexpcal ins;

  require_node_manager(NM);
  require_expr_manager(EM);
  ins = EM->mkNaryExp(oplesign, x1, NILEXPCAL, NILBACKANNOTATION);
  ins = EM->mkNaryExp(oplesign, x2, ins, NILBACKANNOTATION);
  return NM->mkIntrinsicUnsafeAt(getExceptionId(), NILEXPCAL, ins, NILEXPCAL, vh, NILBACKANNOTATION);
}

/**
   It normalises the default definitions and the partial definitions.
   \see DCG_poly::normalDefaultsOneLevel()
*/
boolean DCG_basic::normalDefaults()
{
  boolean bm;
  bm = pK_messenger->setSilentModeMessagesTo(pk_true);
  for_all_graphs_do_br(normalDefaultsOneLevel(), b);
  pK_messenger->setSilentModeMessagesTo(bm);
  return b;
}

/** 
   -- Testing: the utility clock used a s availity clock may introduce cycles.
   This  method can be applied to a graph or a module.
    \copydoc solvingUsingUtilityClocksOnelevel()
*/
boolean DCG_basic::setUtilityClocksAsClock()
{
  for_all_graphs_do_br(setUtilityClocksAsClockOneLevel(), b);
  return b;
}

/**
   It computes the utility clocks of the node of the graph.
   It can be used to reduce the clock of the computation.
*/
void DCG_basic::computeProductiveClock()
{
  for_all_graphs_do(computeProductiveClockOneLevel());
}

/**
   It computes the utility clocks of the inputs of the operand (aSubgraph), element of the current graph.
   For an input x of the subgraph,  x:= x_in becomes x::= in_x when UtilityClk(in_x) | x::=defaultvalue x$1
*/
void DCG_basic::computeProductiveClock(ptINode aSubgraph)
{
  pK_UtilityClock->SubGraphClockUtility(getDCG(), aSubgraph);
}

/**
   It computes the assignment clocks of the node of the graph.
   It can be used to reduce the clock of the computation.
*/
void DCG_basic::computeAssignmentClock()
{
  for_all_graphs_do(computeAssignmentClockOneLevel());
}

/**
   It rewrites the definitions using the utility clock (assumed available)
   \see DCG_poly::computeProductiveClock()
*/
boolean DCG_basic::rewUsingProductiveClock()
{
  boolean bm;
  bm = pK_messenger->setSilentModeMessagesTo(pk_true);
  for_all_graphs_do_br(rewUsingProductiveClockOneLevel(), b);
  pK_messenger->setSilentModeMessagesTo(bm);
  return b;
}

void DCG_basic::computeClockWithoutDefaultConstantValue()
{
  for_all_graphs_do(computeClockWithoutDefaultConstantValueClockOneLevel());
}

/**
   This  method can be applied to a graph of a module (set of processes) or a process.
   It can be applied to any format. 
   It builds an external representation of the graph as an SIGNAL abstract tree using some internal options.
   @return the Signal abstract tree which can be printed using the \see{decsig} function.

   \see{unparse, graph2SignalTree}.

   Assumed: the options of the translator are set before this call.
*/
ntrees DCG_basic::produce_automaton_tree(const char *optionSuf)
{
  ntrees vt;
  int date;

  date = pK_messenger->messageBeginOperVerbose(infoGraph2SignalTree);
  vt = pK_Graph2SignalTreeBuilder->automaton2SignalTree(getDCG(), optionSuf);
  pK_messenger->messageEndOperVerbose(date);
  return vt;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    This is the entry point of Signal code generation. It can be applied for any format.
    It produces code in the file foo_X.SIG, where X denotes a codification of the graph level, completed with the optionsuff suffix.

    The trees memory is cleaned after the printing.
    \see{produce_tree}
*/
boolean DCG_basic::unparseAutomaton(const char *optionSuf)
{
  ntrees DCGAbstractTree;
  boolean b;

  ntrees_memoryStoreIndex();
  DCGAbstractTree = produce_automaton_tree(optionSuf);
  if (DCGAbstractTree != NILTREE)
    {
      b = pK_SignalPrinter->setPrinterContext(pk_false);	// to print the annotations of the trees.
      SignalUnparse(DCGAbstractTree, infotra);
      pK_SignalPrinter->setPrinterContext(b);
    }
  ntrees_memoryRestoreIndex();
  return pk_true;
}

/**
   It updates the interfaces of the sub-graph sg1 and sg2, elements of the current graph.
*/
void DCG_basic::updateInterfacesSubGraphs(ptINode sg1, ptINode sg2)
{
  require_node_manager(NM);
  NM->updateInterface(sg1, sg2);
}


