/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PRIVATE_ static ptDCG_bool gw = NILDCG_BOOL;

#include "DCGs_Macros.c"

void
DCG_bool::init (ptDCG g)
{
  if (gw == NILDCG_BOOL)
    gw = new DCG_bool ();
  gw->setDCG (g);
}


/** 
    DCG_bool graph to DCG_sch graph translation.
    
    This method supposes that the method checkExplicitTyping() has been applied.
    It performs the interformat bDC+ (a tree of boolean clocks) to sbDC+ (Â«sequentialized boolean \e dc+Â») translation. 
    The nodes of the graph are totally ordered (static schedulling) when btotally, otherwise only the clusters are ordered and the schedulling of the clusters is dynamical. 
    The informations for sbDC+ are stored in graph structure for code generators.
    The assert nodes are visible for code generation when bCheck (code will be generated for verifying assumptions at run time).
    \see{transformToBooleanControl,checkExplicitTyping, setDCGSchInfos}
    
    \ref InterFormats "For more details..."
*/
void
DCG_bool::OrderingOneLevel (boolean btotally, boolean bCheck)
{
  ptDCGBool2DCGSchTranslator gen;

  if (!isGraphOfExternalProcess ())
    {
      ptDCG vdcg = getDCG ();
      vdcg->setsbDCplusDependencesForComplementaryClocks ();
      if (spatialOptimisations ())
	NodesInterfaceCalculus ();
      CodeGenColouring (bCheck);
      vdcg->markReferencedMaxTime (substituteGenCodeMark, 1, pk_true);
      if (!bCheck && vdcg->existAssumptions ())
	pK_messenger->warning (warningAssumptions);
      gen = new DCGBool2DCGSchTranslator (vdcg);
      gen->Scheduling (btotally);
      delete gen;
    }
  setFormatLevel (DCG_SCH);
}

/**  
     DCG_bool graph to DCG_flat graph translation.

     It performs the interformat bDC+ (a tree of boolean clocks) to STS 
    (a tree of boolean clocks reduces to one level) translation.
    \see{transformToBooleanControl}

    \ref InterFormats "For more details..."
 */
void
DCG_bool::flattenCHOneLevel ()
{
  ptDCGBool2DCGFlatTranslator x;
  ptinsth tick;
  boolean b;

  tick = getClockManager ()->getTickClock ();
  /* All the delays are replaced by state variables defined at the master clock of the program. */
  b = extendBooleanDelayClockOneLevel (tick, tick);
  x = new DCGBool2DCGFlatTranslator (getDCG ());
  x->apply ();
  delete x;
  if (!b)
    pK_messenger->printString (" *** extendBooleanDelayClock not applied.");
}


/** 
    It adds exceptions for clock constraints. 
    It allows to have code generating with clock constraints.
    The equivalence of clock formula will be, for example, verified dynamically during simulation (at run time).
*/
void
DCG_bool::addExceptionsForConstraintsOneLevel ()
{
  ptlinsth lh1, lh2;
  ptinsth vh1, vh2, vhemb, vh;
  ptexpcal e1, e2;

  require_managers (EM, NM, CKM);
  pK_messenger->messageIdent (infoExceptions, getName ());
  if (CKM->getConstraintsOnPlacedClocks (&lh1, &lh2))
    while (ptlinsth_get_next (&lh1, &vh1))
      {
	ptlinsth_get_next (&lh2, &vh2);
	if (CKM->satisfiesNullClock (vh1))
	  {
	    vhemb = CKM->getTickClock ();
	    e1 = EM->getFalse ();
	    e2 = CKM->getConditionSub (mkBooleanClockScopeExtand (vh2, vhemb));
	  }
	else if (CKM->satisfiesNullClock (vh2))
	  {
	    vhemb = CKM->getTickClock ();
	    e2 = EM->getFalse ();
	    e1 = CKM->getConditionSub (mkBooleanClockScopeExtand (vh1, vhemb));
	  }
	else
	  {
	    vhemb = CKM->getBranching (vh1, vh2);
	    vh = mkBooleanClockScopeExtand (vh1, vhemb);
	    if (CKM->satisfiesSub (vh))
	      e1 = CKM->getConditionSub (vh);
	    else
	      e1 = EM->getTrue ();

	    vh = mkBooleanClockScopeExtand (vh2, vhemb);
	    if (CKM->satisfiesSub (vh))
	      e2 = CKM->getConditionSub (vh);
	    else
	      e2 = EM->getTrue ();
	  }
	addException (e1, e2, vhemb);
      }
  if (getNullClockConstraints (&lh1))
    {
      vhemb = CKM->getTickClock ();
      while (ptlinsth_get_next (&lh1, &vh1))
	{
	  e1 = CKM->getConditionSub (mkBooleanClockScopeExtand (vh1, vhemb));
	  addException (e1, EM->getFalse (), vhemb);
	}
    }
}

/**
   For each undersampling clock ([C]) of lhtodeclare, 
   if C is not an input of the graph, C is added in the outputs of the graph; it is also added to lappSignals
   - otherwise, C is added to lappSignals

   It can be used for exporting some booleans in an abstration of a model.
   @return "true" if the list lappSignals is modified, "false" otherwise.
*/
boolean
DCG_bool::buildClockExportation (ptlinsth lhtodeclare, ptlinstances * lappSignals)
{
  ptinsth vh;
  ptexpcal exp, vExport, vsig;
  ptINode vinstr;
  ptlinstances l, louts, ls, curlist, lins;
  boolean bpos, badded;

  require_managers (EM, NM, CKM);
  louts = getOutputs ();
  lins = getInputs ();
  badded = pk_false;

  while (ptlinsth_get_next (&lhtodeclare, &vh))
    if (!CKM->satisfiesTickClock (vh))
      if (CKM->satisfiesSub (vh))
	{
	  vExport = EM->getExprExportBooleanAbtraction (CKM->getBooleanSignal (vh), *lappSignals, &bpos);
	  ls = EM->getSignalsReferencedIn (vExport);
	  l = ls;
	  while (ptlinstances_get_next (&ls, &vsig))
	    if (!ptlinstances_isin_exp (lins, vsig))
	      if (ptlinstances_isin_insert_exp (&louts, &curlist, vsig))
		{ // not already in the outputs.
		  ptlinstances_isin_insert_exp (lappSignals, &curlist, vsig);
		  vinstr = NM->mk_node_output (vsig, EM->getBackAnnotation(vsig) );
		  NM->insertNode (vinstr, EM->getClock (vsig));
		  EM->belongsOutFlow (vsig);
		  badded = pk_true;
		}
	  ptlinstances_free (l);
	}
  if (badded)
    setOutputs (louts);
  return badded;
}
