_PUBLIC_

/**
   @return "true" if the graph is not a graph of a module and it is endochronous, "false" otherwise.
*/
boolean
DCG::isEndochronous ()
{
  if (isModule ())
    return pk_false;
  // return (aClockManager->getNbFreeVariables () == 1); -- Nov 2017 replaced by 
  return (aClockManager->getNbRoots () == 1);
}


/**
   It unmarks the objects of the graph (objects of the clock manager, the node manager and the expressions manager) 
   marked by m.
*/
void
DCG::unmarkObjects (typeMark m)
{
  getClockManager ()->resetMarkAll (m);
  getNodeManager ()->resetMarkAll (m);
  getExpressionManager ()->resetMarkAll (m);
}


/** 
    @return the clocks of lh  + the clocks assigned to the dependences (lclos). 
*/
ptlinsth
DCG::getDependencesClocks (ptlclosure lclos, ptlinsth lh)
{
  ptinsth hdep;
  boolean b;
  pticlosure iclos;
  pthclosure hclos;

  iclos = ptlclosure_get_iclos (lclos);
  while (iclos != NILICLOSURE)
    {
      hdep = pticlosure_get_hdep (iclos);
      hdep = aClockManager->getRepresentativeFormula (hdep);
      lh = ptlinsth_sorted_insert_elem (lh, hdep, &b);
      iclos = pticlosure_get_next (iclos);
    };
  hclos = ptlclosure_get_hclos (lclos);
  while (hclos != NILHCLOSURE)
    {
      hdep = pthclosure_get_hdep (hclos);
      hdep = aClockManager->getRepresentativeFormula (hdep);
      lh = ptlinsth_sorted_insert_elem (lh, hdep, &b);
      hclos = pthclosure_get_next (hclos);
    };
  return lh;
}

/** @return the list of clocks assigned to the i/o dependences.
    \warning : it assumes that the closure calculus has been applied.
*/
ptlinsth
DCG::getIODependencesClocks ()
{
  ptlinsth lh;
  ptdependence aDep;
  ptldependence lDep;

  require_clock_manager (CKM);
  require_expr_manager (EM);

  lh = NILLINSTH;
  lDep = getExplicitDependences ();
  while (lDep->get_next (&aDep, &lDep))
    lh = getIODependencesClocks (aDep, lh);

  lDep = getReinforcedDependences ();
  while (lDep->get_next (&aDep, &lDep))
    lh = getIODependencesClocks (aDep, lh);
  return lh;
}

/**
   @return "true" if the list of the graphs (lg) referenced in the current graph, at one level only, 
   is not empty, "false" otherwise.
*/
boolean
DCG::getReferencedGraphs (ptlDCG * lg)
{
  ptlINode lNodes;
  ptINode aNode;
  ptDCG g;

  require_node_manager (NM);
  (*lg) = NILLDCG;

  if (!NM->getNodeCalls (&lNodes))
    return pk_false;
  while (ptlINode_get_elem (&lNodes, &aNode))
    {
      g = NM->get_nodeCall_graph (aNode);
      ptlPointer_insert_elem (lg, (attributeValue) g);
    };
  return pk_true;
}

/**
   @return the safety property of the graph (safe, unsafe, deterministic).
   See the reference manual for the definitions.
*/
safetyPropertyType
DCG::getSafetyProperty ()
{
  ptINode node;
  ptlINode ln;
  safetyPropertyType cc, code;

  require_node_manager (NM);
  code = opsafe;
  if (NM->getAllNodes (&ln))
    while (ptlINode_get_elem (&ln, &node))
      {
	if (NM->is_delayNodeOrDelayClusterNode (node) || NM->is_cellNode (node))
	  {
	    cc = opunsafe;
	    NM->set_safetyProperty (node, cc);
	  }
	else
	  // default opsafe, procall, nodecall mis a jour a la construction.
	  cc = NM->get_safetyProperty (node);
	if (code != cc)
	  if ((code == opdeterm) && (cc == opunsafe)) code = opunsafe;
	  else 
	    { if ((code == opsafe) && (cc == opdeterm)) code = opdeterm;}
      }
  return code;
}

/**
   Spatial processing.

   For x := f ( y[?] ), the expression y[?] specifies the value of y at the previous iteration.
   The "iteration precedence" x -> y has been assigned to the node during the production of the graph.
   This method tests that the dependence x -> y does not introduce circuit in the graph. 
   In this case:
   - an auxillary signals yy is created and defined by yy := y[?],
   - then y[?] is replaced by yy in other expressions,
   - some dependences are added : yy -> y.

   @return "true" if the graph is modified, "false" otherwise.
*/
boolean
DCG::iterationPrecedences ()
{
  ptldependence iterationPrecedenceLst;
  ptdependence aDep;
  ptINode node, aNode, vfrom, vto, yyNode;
  ptlINode lNodes, ArrayNodeLst, ln;
  ptlinsth ClockLst;
  boolean b, modifForThisNode, bm;
  ptinsth hundef, vh;
  ptexpcal y, yprev, yy;

  require_expr_manager (EM);
  require_node_manager (NM);
  if (!NM->getIterates (&lNodes))
    return pk_false;

  b = pk_false;
  hundef = getClockManager ()->getUndefClock ();
  bm = pK_messenger->setSilentModeMessagesTo (pk_true);	// Silent for messages
  while (ptlINode_get_elem (&lNodes, &node))
    {
      NM->get_array_bodyDefinitions (node, &ArrayNodeLst, &ClockLst);
      modifForThisNode = pk_false;
      ln = ArrayNodeLst;
      while (ptlINode_get_elem (&ArrayNodeLst, &aNode))
	{
	  iterationPrecedenceLst = NM->get_iterationPrecedence (aNode);
	  // aNode must be generated before the nodes of iterationPrecedenceLst
	  while (iterationPrecedenceLst->get_next (&aDep, &iterationPrecedenceLst))
	    {
	      aDep->getDependence (&vfrom, &vto);
	      if (vfrom != vto)
		if (ptlINode_isIn (ln, vto))
		  {
		    NM->addDependence (vfrom, vto, hundef);
		    if (!isStaticCycleFree (pk_false, pk_false))
		      {
			pK_messenger->globalMessage (ErrCycleAddedByReinforcing);
			NM->removeDependence (aDep);
			/*
			 * For aDep : x->y ; an auxilary variable is created yy := y[?], then y[?] is replaced by yy and the dependence x --> y. 
			 * is added. 
			 */
			y = NM->get_leftOperand (vto);
			vh = EM->getClock (y);
			yprev = EM->mkUnaryExp (opsdec, y, EM->getBackAnnotation (y));
			yy = EM->mkIdentExp (EM->getOriginalName (y), tsignal, EM->getType (y), NM->mk_node_undefined (), vh, NILBACKANNOTATION);
			EM->substitute (yprev, yy, vh);
			yyNode = NM->insertEquation (yy, yprev, vh, vh);
			NM->arrayBodyInsertNode (node, yyNode);
			NM->set_inComposedNode (yyNode, node);

			NM->addDependence (yyNode, vto, hundef);
			NM->addDependence (yyNode, vfrom, hundef);
			modifForThisNode = pk_true;
			// ptinterface_insert_localSignal(NM->get_interface(node), yy);
			b = pk_true;
		      }
		  }
	    }
	}
      if (modifForThisNode)
	{
	  NM->set_interface (node, NILINTERFACE);
	  NM->getSet_interface (node);
	}
    }
  pK_messenger->setSilentModeMessagesTo (bm);
  return b;
}

/**
   It synthesis the set of indices referenced in the equations and the partial definitions.
*/
void
DCG::spatialIndicesSynthesis ()
{
  ptlINode ln;

  require_node_manager (NM);
  if (NM->getEquations (&ln))
    spatialIndicesSynthesis (ln);
  // if (NM->getNodesList(opdefpart, &ln)) spatialIndicesSynthesis(ln);
  if (NM->getGroupsOfPartialEquations (&ln))
    spatialIndicesSynthesis (ln);
}

/**
   For each node (equation, partial definition) of ln, 
   it synthesis the set of indices referenced in the node.
*/
void
DCG::spatialIndicesSynthesis (ptlINode ln)
{
  ptINode node;
  ptexpcal vdef;

  require_expr_manager (EM);
  require_node_manager (NM);
  while (ptlINode_get_elem (&ln, &node))
    if (NM->is_setOfPartialDefinitions (node))
      spatialIndicesSynthesis (NM->get_setOfPartialDefinitions_arg (node));
    else if (NM->is_equation (node) || NM->is_partialDefinition (node))
      if (NM->get_rightOperand (node, &vdef))
	if (!EM->IsCellExpr (vdef))
	  {
	    vdef = EM->getBasicExpr (vdef);
	    EM->addTravelIndexesToExpression (vdef);
	  }
}

/**
   It substitutes each expression such that x[vce] by x, when vce is an indice, x is a enumerator) 
   and x is marked previously by OPTIMIZED_ATTRIBUTE (see arithmeticalProgressionOptimizing(), sameDefinitionExpressionOptimizing))
   and x and vce are equivalent.\n
   Example : I:= 0..5, I[k] with k an index defined form 0 to 5, can be replaced by I.
   @return "true" is some substitutions have been applied, "false" otherwise.
*/
boolean
DCG::spatialSubstitutions ()
{
  boolean b, bres;
  ptexpcal x, vce, e1, e2;
  ptlinstances lst, l;

  require_expr_manager (EM);
  bres = pk_false;
  do
    {
      b = pk_false;
      if (EM->getExprList (opelem, &lst))
	while (ptlinstances_get_next (&lst, &x))
	  if (!EM->getIndirection (x, &vce))
	    if (EM->getAttribute (x, OPTIMIZED_ATTRIBUTE, (attributeValue *) & vce))
	      {
		e1 = EM->getBinaryArgstAt (x, 1);
		e2 = EM->getBinaryArgstAt (x, 2);	// list of indexes.
		if (EM->getSize (e2) == 1)
		  {
		    EM->getNaryArgs (&e2, &vce);
		    if (EM->isEquivalent (e1, vce, pk_false))	// x[x] is replaced by x
		      {
			if (EM->getTravelIndexes (x, &l))
			  EM->setTravelIndexes (e1, l);
			EM->replace (x, e1);
			b = pk_true;
			bres = pk_true;
		      }
		  }
	      }
    }
  while (b);
  return bres;
}

/**
   It applies some optimisations on spatial processing.
   @return "true" if some optimisations have been applied, "false" otherwise.
*/
boolean
DCG::spatialOptimisations ()
{
  boolean b, bres;

  sameDefinitionExpressionOptimizing ();
  indexDefinitionsOptimizing ();

  bres = pk_false;
  b = pk_true;
  do
    if (arithmeticalProgressionOptimizing ())
      bres = pk_true;
    else if (spatialSubstitutions ())
      bres = pk_true;
    else
      b = pk_false;
  while (b);
  spatialIndicesSynthesis ();
  return bres;
}

/**
   It marks the set of identifiers that are definitions of indices.
   It uses the predefined mark : POSSIBLY_INDEX.
*/
void
DCG::indexDefinitionsOptimizing ()
{
  ptlinstances lident;
  ptexpcal aSig;

  require_expr_manager (EM);
  if (EM->getExprList (opident, &lident))
    while (ptlinstances_get_next (&lident, &aSig))
      if (EM->IsSignal (aSig))
	if (!EM->isMarked (aSig, POSSIBLY_INDEX))
	  if (EM->isIndexDefinition (aSig))
	    EM->addMark (aSig, POSSIBLY_INDEX);
}

/**
   Signals defined by the same syntactic expression (taking account value of  expression) are marked.
   Restriction : only equations (X:=E) are considered.
   It uses the predefined mark : OPTIMIZED_ATTRIBUTE.
*/
boolean
DCG::sameDefinitionExpressionOptimizing ()
{
  ptlINode lNodes;
  ptINode node;
  boolean b, bapply, bres;
  ptexpcal x, vdef, vce, vattr;

  require_expr_manager (EM);
  require_node_manager (NM);
  bres = pk_false;
  do
    {
      b = pk_false;
      if (NM->getEquations (&lNodes))
	while (ptlINode_get_elem (&lNodes, &node))
	  if (!NM->isMarked (node, DontOptimize))
	    if (NM->get_rightOperand (node, &vdef))
	      {
		x = NM->get_leftOperand (node);
		if (EM->IsSignal (x))
		  if (!EM->getAttribute (x, OPTIMIZED_ATTRIBUTE, (attributeValue *) & vce))
		    if (!EM->IsCellExpr (vdef))
		      {
			if (!EM->IsOutputFlow (x))
			  bapply = pk_true;
			else
			  bapply = EM->IsTypeExprPredefinedType (x);
			if (bapply)
			  {
			    vdef = EM->getBasicExpr (vdef);
			    if (!EM->getAttribute (vdef, OPTIMIZED_ATTRIBUTE, (attributeValue *) & vattr))
			      // sinon peut avoir le cas zx := x | zx := x $1 and optimized delay.
			      if (EM->IsSameExpressionForEachElement (vdef, &vce))
				{
				  b = pk_true;
				  bres = pk_true;
				  EM->addAttribute (x, OPTIMIZED_ATTRIBUTE, (attributeValue) vce);
				}
			  }
		      }
	      }
    }
  while (b);
  return bres;
}

/**
   Indices defined by an arithmetical progression are marked.
   It uses the predefined mark: OPTIMIZED_ATTRIBUTE.
*/
boolean
DCG::arithmeticalProgressionOptimizing ()
{
  boolean b, bres;
  ptexpcal x, vce, e1, e2;
  ptlinstances lst;

  require_expr_manager (EM);
  bres = pk_false;
  do
    {
      b = pk_false;
      if (EM->getExprList (opelem, &lst))
	while (ptlinstances_get_next (&lst, &x))
	  if (!EM->getIndirection (x, &vce))
	    if (!EM->getAttribute (x, OPTIMIZED_ATTRIBUTE, (attributeValue *) & vce))
	      {
		e1 = EM->getBinaryArgstAt (x, 1);
		if (EM->IsSignal (e1))
		  {
		    e2 = EM->getBinaryArgstAt (x, 2);	// list of indexes.
		    if (EM->getSize (e2) == 1)
		      {
			EM->getNaryArgs (&e2, &vce);
			if (EM->IndiceSatisfiesArithmeticalProgression (e1, 1, 1))
			  {
			    EM->addAttribute (x, OPTIMIZED_ATTRIBUTE, (attributeValue) vce);
			    b = pk_true;
			    bres = pk_true;
			    // x[vce] and vce is an indice, x is a enumerator.
			    if (EM->IsIndex (vce))
			      EM->addAttribute (e1, OPTIMIZED_ATTRIBUTE, (attributeValue) vce);
			  }
		      }
		  }
	      }
    }
  while (b);
  return bres;
}

/**
   It unifies the constants defined by the same syntactic expression.
*/
boolean
DCG::ConstantsUnifying ()
{
  ptlinstances lc, completelist;
  ptexpcal aConst;
  boolean modif, bres;

  completelist = getConstants ();
  bres = pk_false;
  do
    {
      modif = pk_false;
      lc = completelist;
      while (!modif && ptlinstances_get_next (&lc, &aConst))
	if (UnifyConstant (aConst, lc, &completelist))
	  {
	    modif = pk_true;
	    bres = pk_true;
	  }
    }
  while (modif);
  setConstants (completelist);
  return bres;
}

/**
   It eliminates the expressions "when e", for e an event.
   when e becomes e, when not e becomes the null clock.
*/
boolean
DCG::eliminateWhenEvent ()
{
  ptbooleans2events arew;
  boolean b;

  arew = new booleans2events (this);
  b = arew->apply (pk_false, pk_false, pk_false);
  delete arew;
  return b;
}

/** 
    Removing, for the current graph, the signals without definitions. 
*/
void
DCG::removeUndefinedSignals ()
{
  ptlinstances lident;
  ptexpcal aSig;
  ptINode node;

  require_managers (EM, NM, CKM);
  if (!isGraphOfExternalProcess ())
    {
      if (EM->getExprList (opident, &lident))
	while (ptlinstances_get_next (&lident, &aSig))
	  if (EM->IsSignal (aSig))
	    {
	      node = EM->getNode (aSig);
	      if (NM->is_undefined (node))
		removeSignal (aSig);
	    }
    }
}

/**
   It rebuilts the interface of all the nodes of the graph.
   The clock hierarchy is not considered.
*/
void
DCG::NodesInterfaceCalculus ()
{
  ptlINode lnodes;
  ptINode node;

  require_node_manager (NM);
  if (NM->getAllNodes (&lnodes))
    while (ptlINode_get_elem (&lnodes, &node))
      {
	NM->set_interface (node, NILINTERFACE);
	NM->getSet_interface (node);
      }
}

/** 
    Delay Rewriting: It applies the reductions defined in delaysRewriting1() and delaysRewriting2() methods.
    @return "true" if the graph is modified, "false" otherwise.
*/
boolean
DCG::delaysRewriting ()
{
  boolean b;

  b = delaysRewriting1 ();
  // if (delaysUnifying()) b = pk_true; moved 
  return b;
}

/**
   It substitutes the evaluated expressions by their values.
   It calls 
   - ExprManager::substituteEvaluatedExpression() followed by
   - ExprManager::EM->rewritingExprEH, then 
   - updatePragmas() to update the objects assigned to the pragmas.
*/
void
DCG::substituteEvaluatedExpression ()
{
  require_expr_manager (EM);
  EM->substituteEvaluatedExpression ();
  EM->rewritingExprEH ();
  updatePragmas ();
}

/** It cleans the structures used to build the interface of the graph (ie each node of the clock hierarchy).
    \see buildClockProcessInterfaceOneLevel 
*/
void
DCG::interfaceReset ()
{
  ptinsth vh;
  ptinterface vinter;

  require_clock_manager (CKM);
  // Cleaning the interface of the tick: inputs, outputs are not removed.
  vh = getRootClock ();
  if (!CKM->satisfiesUndefClock (vh))
    {
      vinter = CKM->getInterfaceSubHierarchy (vh);
      interface_free_clocks (vinter, isEventClockedFormat ());
      ptinterface_set_localSignals (vinter, NILLINSTANCES);
    }
  // Cleaning the interface of the roots, and the placed clocks.
  interfaceReset (CKM->getlhroots ());
  interfaceReset (CKM->getlhplaced ());
}

/**
   It imports the result of the transitive closure into the graph.
   Informations stored in specific structures are translated into dependences.
   \see computeIOTransitiveClosure
*/
void
DCG::importTransitiveClosure ()
{
  ptlinsth lhclos;
  ptinsth h, hdep;
  ptlinstances lexpclos, levents;
  ptexpcal exp, expFrom;
  ptlclosure lclos;
  pticlosure iclos;

  require_managers (EM, NM, CKM);
  lexpclos = NILLINSTANCES;
  levents = NILLINSTANCES;
  EM->splitEvents (getInputs (), &lexpclos, &levents);
  EM->splitEvents (getOutputs (), &lexpclos, &levents);
  lhclos = EM->getClockOfSignals (levents);

  h = CKM->getUndefClock ();
  while (ptlinstances_get_next (&lexpclos, &exp))
    if (EM->getAttribute (exp, CLOSUREATTRIBUTE, (attributeValue *) & lclos))
      {
	iclos = ptlclosure_get_iclos (lclos);
	while (iclos != NILICLOSURE)
	  {
	    expFrom = pticlosure_get_from (iclos);
	    if (exp != expFrom)
	      {
		hdep = CKM->getRepresentativeFormula (pticlosure_get_hdep (iclos));
		if (!CKM->satisfiesNullClock (hdep))
		  NM->addDependence (expFrom, exp, hdep, pk_true);
	      }
	    iclos = pticlosure_get_next (iclos);
	  }
      }

  while (ptlinsth_get_next (&lhclos, &h))
    {
      h = CKM->getRepresentativeFormula (h);
      if (CKM->getAttribute (h, CLOSUREATTRIBUTE, (attributeValue *) & lclos))
	{
	  iclos = ptlclosure_get_iclos (lclos);
	  while (iclos != NILICLOSURE)
	    {
	      expFrom = pticlosure_get_from (iclos);
	      hdep = CKM->getRepresentativeFormula (pticlosure_get_hdep (iclos));
	      if (!CKM->satisfiesNullClock (hdep))
		NM->addDependence (EM->getNode (expFrom), h);
	      iclos = pticlosure_get_next (iclos);
	    }
	}
    }
}

/**
   It sets the clocks that may be substitutable in the graph.
   \see setSubstitutable()
*/
void
DCG::setSubstituableClocks ()
{
  ptlinsth lh;
  ptinsth h;

  require_clock_manager (CKM);
  // The roots ones.
  lh = CKM->getlhroots ();
  while (ptlinsth_get_next (&lh, &h))
    setSubstitutable (h);
  // The placed ones.
  lh = CKM->getlhplaced ();
  while (ptlinsth_get_next (&lh, &h))
    setSubstitutable (h);
}

/**
   Builts and stores the lists of external procedures and functions.
   They are respectively stored in list_extern_proc and list_extern_funcs attributes.
   \see setExternalReferencesUsingMark, setListExternalProc, getListExternalProcedure.
*/
void
DCG::setExternalReferences ()
{
  typeMark myMark;

  myMark = getMark ();
  setExternalReferencesUsingMark (myMark);
  resetMark (myMark);
}

/**
   @return the clock system as a sub-graph node.
*/
ptINode
DCG::encapsulatingClockSystem ()
{
  ptinsth vh;
  ptlinsth lhorl;
  ptexpcal vlabel;
  ptINode sg;
  ptinterface vinterface;

  require_managers (EM, NM, CKM);
  vlabel = EM->mkLabelIdent (getLabelId (), NM->mk_node_undefined (), CKM->getUndefClock (), NILBACKANNOTATION, pk_true);
  sg = NM->mk_node_subgraph (vlabel, NILBACKANNOTATION);
  lhorl = aClockManager->getlhroots ();
  while (ptlinsth_get_next (&lhorl, &vh))
    NM->subGraphInsertClock (sg, vh);
  lhorl = aClockManager->getlhplaced ();
  while (ptlinsth_get_next (&lhorl, &vh))
    NM->subGraphInsertClock (sg, vh);

  vinterface = NM->getSet_interface (sg);
  upstreamProjectionSubGraph (sg);
  NM->addPragma (sg, newPredefinedPragma (PRAGMA_SELECTED_OTHERS_PARTITION, NILOBJPRAGMA, repstr2PragmaStatement (string2repstr ("ClockSystem"))));
  return sg;
}

/**
   Storing informations about the null clock (universal object) for the current graph.
   It assigns
   - the list of  the signals of the graph at null clock to \b lSigNullClock attribute,
   - the list of clocks (set of formula) not proved equivalent to null clock, to \b lConstraintsNullClock attribute.

   \see getNullClockSignal, getNullClockConstraints
*/
void
DCG::storeNullClockInfo ()
{
  ptlinstances vli, lident, lsig;
  ptexpcal vsig;

  require_expr_manager (EM);
  require_clock_manager (CKM);
  if (EM->getExprList (opident, &lident))
    {
      lsig = CKM->getSynchronousSignals (CKM->getNullClock ());
      while (ptlinstances_get_next (&lsig, &vsig))
	if (ptlinstances_isin_exp (lident, vsig))
	  ptlinstances_isin_insert_exp (&lSigNullClock, &vli, vsig);
    }
  CKM->getNullClockConstraints (&lConstraintsNullClock);
}

/** 
    It builds the list of the nodes of the current graph that are at the initial clock.
    To get this list use the getNodesAtInitClock() method.
*/
void
DCG::storeInitClockInfo ()
{
  ptexpcal vopd;
  ptINode node;
  ptinsth hini, hd;
  ptlINode ln;

  // ***************** A virer : hinit n'est pas une horloge universelle.
  require_managers (EM, NM, CKM);
  // free_linode(lNodesAtInitClock);
  lNodesAtInitClock = NILLINODE;
  if (NM->getEquations (&ln))
    {
      hini = CKM->getInitClock ();
      while (ptlINode_get_elem (&ln, &node))
	if (NM->get_rightOperand (node, &vopd))
	  if (!EM->IsCellExpr (vopd))
	    {
	      hd = CKM->getRepresentativeFormula (EM->getContributionClock (vopd));
	      if (hd == hini)
		ptlINode_insert_elem (&lNodesAtInitClock, node);
	    }
    }
}

/** 
    Unmarks the current graph, marked by m, and recursively the graphs it references.
    All objects of the managers are visited (clocks, nodes, expressions).
*/
void
DCG::resetMark (int m)
{
  ptlINode lNodes;
  ptINode aNode;
  ptDCG g;
  ptlDCG lGraph;
  pointer ptr;

  require_node_manager (NM);
  removeMark (m);

  if (NM->getNodeCalls (&lNodes))
    while (ptlINode_get_elem (&lNodes, &aNode))
      {
	g = NM->get_nodeCall_graph (aNode);
	g->resetMark (m);
      };
  lGraph = getLocalGraphs ();
  while (ptlPointer_get_elemNext (&lGraph, &ptr))
    {
      g = (ptDCG) ptr;
      g->resetMark (m);
    }
}

/**
   This method is used for completing the clock hierarchy to build an endochronous program.
   It adds:
   - an input boolean (name \b b based on \b id) synchronized to the clock clk,
   - the constraint vh ^= when \b b

   @return the created boolean signal.
*/
ptexpcal
DCG::addBooleanInput (ptinsth vh, ptinsth clk, ptident id)
{
  ptinsth vhsub;
  ptexpcal bhsig;

  require_managers (EM, NM, CKM);
  bhsig = EM->mkIdentExp (id, tsignal, get_BooleanType (), NM->mk_node_undefined (), CKM->getUndefClock (), NILBACKANNOTATION);
  EM->setClock (bhsig, clk);
  NM->insertNode (NM->mk_node_input (bhsig, NILINODE, NILBACKANNOTATION), clk);
  vhsub = CKM->makeSub (clk, bhsig, EM->getPredecessors (bhsig));
  vhsub = CKM->makeSynchro (vh, vhsub);
  setInputs (ptlinstances_insert (getInputs (), bhsig));
  EM->setAuxiliarySignal (bhsig);
  EM->belongsInFlow (bhsig);
  return bhsig;
}

/**
   This method is used for completing the clock hierarchy to build an endochronous program.
   It adds a node call idClk := _PRESENT_id() at the clk clock.
   - an input boolean (name \b b based on \b id) synchronized to the clock clk,
   - the constraint vh ^= when \b b

   @return the created boolean signal.
*/
ptexpcal
DCG::addPresentNode (ptinsth vh, ptinsth clk, ptexpcal vsig)
{
  ptinsth vhsub;
  ptexpcal bhsig;
  ptINode n;
  ptident idSig, idClk;

  require_managers (EM, NM, CKM);
  idSig = EM->getOriginalName (vsig);
  idClk = namBooleanClockfrom (idSig);
  if (idSig == NILIDENT) idSig = idClk;
  bhsig = EM->mkIdentExp (idClk, tsignal, get_BooleanType (), NM->mk_node_undefined (), CKM->getUndefClock (), NILBACKANNOTATION);
  EM->setClock (bhsig, clk);
  n = NM->mkIntrinsicUnsafeAt(getPresentId (), NILEXPCAL, EM->mkNaryExp (oplesign, vsig, NILBACKANNOTATION), EM->mkNaryExp (oplesign, bhsig, NILBACKANNOTATION), clk, NILBACKANNOTATION);
  //  NM->insertNode (n, clk);
  vhsub = CKM->makeSub (clk, bhsig, EM->getPredecessors (bhsig));
  vhsub = CKM->makeSynchro (vh, vhsub);
  EM->setAuxiliarySignal (bhsig);
  return bhsig;
}

/**
   It updates the objects associated with the pragmas.
   Only signal identifiers are kept.
*/
void
DCG::updatePragmas ()
{
  ptpragma vp;
  ptlpragmas lp;
  ptlinstances lsave, l;
  ptexpcal aSig;
  boolean bremove;

  require_expr_manager (EM);
  lp = getPragmas ();
  while (ptlpragmas_get_next (&lp, &vp))
    {
      l = (ptlinstances) getPragmaObjects (vp);
      lsave = l;
      while (ptlinstances_get_next (&l, &aSig))
	{
	  if (EM->IsSignal (aSig))
	    bremove = ptlINode_isIn (unusedNodes, EM->getNode (aSig));
	  else
	    bremove = pk_true;
	  if (bremove)
	    lsave = ptlinstances_remove_exp (lsave, aSig);
	}
      setPragmaObjects (vp, (ptObjPragma) lsave);
    }
}

/**
   Removing nodes (ln) and clocks (lh) from the graph.
*/
void
DCG::removeNodes (ptlINode ln, ptlinsth lh)
{
  typeMark myMark;
  ptlinsth lh1;
  boolean b;
  ptinsth vh, vh1;
  pkClockType vop;

  require_node_manager (NM);
  require_clock_manager (CKM);
  myMark = getMark ();
  NM->marklINodes (ln, myMark);
  // The extractions are kept.
  lh1 = lh;
  while (ptlinsth_get_next (&lh1, &vh))
    {
      vop = CKM->getOperator (vh);
      if ((vop == sub) || (vop == nosub))
	lh = ptlinsth_remove_elem (lh, vh, &b);
    }

  CKM->marklClocks (lh, myMark);
  removeMarkedNodes (myMark);
  unmarkObjects (myMark);
}

/** Removing a node. */
void
DCG::removeNode (ptINode node)
{
  ptexpcal sig;
  ptlinstances lsig;

  require_node_manager (NM);
  lsig = NM->getInodelOutputFlows (node);
  while (ptlinstances_get_next (&lsig, &sig))
    removeSignal (sig);
  NM->removeNode (node);
}

/**
   The implicit conversions defined in the Signal Manual reference are explicited.
*/
void
DCG::explicitConversions ()
{
  require_node_manager (NM);
  NM->explicitConversions ();
}

/**
   Garbaging types.
   Not implemented.
*/
void
DCG::garbageTypes ()
{

}

/** 
    Misc.
    
    Masking outputs of the graph. Outputs becomes local signals.
*/
void
DCG::maskOutputs ()
{
  ptlinstances l;
  ptexpcal vsig;

  require_expr_manager (EM);
  l = getOutputs ();
  while (ptlinstances_get_next (&l, &vsig))
    EM->belongsLocalFlow (vsig);
  setOutputs (NILLINSTANCES);
}

/** 
    @return true if it exists non predefined types (returned in lTypes) for the current object(DCG, Automaton), 
    false otherwise.
*/
boolean
DCG::getUserTypesOneLevel (ptlinstances * lTypes)
{
  return Common_DCG_Automaton::getUserTypes (lTypes);
}

/** 
    This method can be applied to a graph of a module (set of processes) or a process.
    \copydoc checkExplicitTypingOneLevel
*/
boolean
DCG::checkExplicitTyping ()
{
  for_all_graphs_do_br (checkExplicitTypingOneLevel (), apply);
  return apply;
}

/**
   \copydoc setArrayClocksOneLevel
*/
boolean
DCG::setArrayClocks ()
{
  for_all_graphs_do_br (setArrayClocksOneLevel (), b);
  return b;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    \copydoc splitDataStructureOneLevel
*/
boolean
DCG::splitDataStructure ()
{
  for_all_graphs_do (splitDataStructureOneLevel ());
  return pk_true;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    \copydoc setClockDependencesOneLevel
*/
boolean
DCG::setComposedNodesDependences ()
{
  for_all_graphs_do (setComposedNodesDependencesOneLevel ());
  return pk_true;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    \copydoc setClockDependencesOneLevel
*/
boolean
DCG::setClockDependences ()
{
  for_all_graphs_do (setClockDependencesOneLevel ());
  return pk_true;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    It detects cycles in all the graphs of the application.
    When bKeepCycles is "true", the cycles are kept in the graph sructures.
    \copydoc hasStaticCycleOneLevel
    \See getCycles(), getPseudoCycles()
*/
boolean
DCG::isStaticCycleFree (boolean bmess, boolean bKeepCycles)
{
  for_all_graphs_do_br (hasStaticCycleOneLevel (bmess, bKeepCycles), breturn);
  return !breturn;
}

/**
   @return the list of the nodes of the graph tagged with the pragma code (pragmaName).
*/
ptlINode
DCG::getTaggedNodes (int pragmaName)
{
  require_node_manager (NM);
  return NM->getTaggedNodes (pragmaName);
}

/**
   It finalyzes a binary partitionning functionality. 
   @param lSubGraphs (in) the result of the operation, composed of two subgraphs.
   @param part1 (out): the selected part.
   @param part2 (out): the other part.
   
   The nodes of the subgraphs are unlinked from their clock in the graph. The nodes are accessible by the subgraph.
*/
void
DCG::finalyzeBinPart (ptlINode lSubGraphs, ptINode * part1, ptINode * part2)
{
  require_node_manager_of (this, NM);
  // NM->subGraphUnlinkNodesFromClocks(lSubGraphs); 
  ptlINode_get_elem (&lSubGraphs, part1);
  ptlINode_get_elem (&lSubGraphs, part2);
}

/**
   The signals of the operand list (ls) are marked by the reserved mark DontOptimize.
*/
void
DCG::DontOptimizeSignals (ptlinstances ls)
{
  ptexpcal id;

  require_node_manager (NM);
  require_expr_manager (EM);
  while (ptlinstances_get_next (&ls, &id))
    NM->addMark (EM->getNode (id), DontOptimize);
}

/**
   It merges the equations defined by the same function call: (| x:=f(E) | y := f(E) |) and f a function is rewritten (| x:=f(E) | y := x |).
   @return "true" if the graph is modified, "false" otherwise.
   
   It calls functionUnification(ptlINode) with the list of procedure call nodes, and the list of node call node.
*/
boolean
DCG::functionUnification ()
{
  boolean b;
  ptlINode ln;

  require_node_manager (NM);
  b = pk_false;
  if (NM->getProcCalls (&ln))
    if (functionUnification (ln))
      b = pk_true;
  if (NM->getNodeCalls (&ln))
    if (functionUnification (ln))
      b = pk_true;
  return b;
}

/** 
    It unifies the nodes defined by the same expression except for the renamings (see removeSignalEqualitiesOneLevel()) :
    If it exists a set of nodes Ni (i > 1)defined such that Xi := E, then one node is choosen 
    to be the representative element.
    @return "true" if some nodes have been selected, "false" otherwise.
*/
boolean
DCG::nodesUnification ()
{
  ptexpcal kept, replaced, vexp;
  ptlinstances liste;
  int i;
  ptlINode ln;
  boolean keptClass, b, apply;
  ptINode n, selectedNode;
  ptinsth h;

  require_managers (EM, NM, CKM);
  b = pk_false;
  for (i = 0; i < SIG_NB_OPERATORS; i++)
    if (!((i == opnom) || (i == opident)))	// For the renaming, see removeSignalEqualitiesOneLevel()
      if (EM->getExprList (i, &liste))
	while (ptlinstances_get_next (&liste, &vexp))
	  {
	    apply = pk_true;
	    /*
	     * Removed 2013 July lbesnard. Scalar constants are filtered (they will be substituted). if (i == opextr) if
	     * (CKM->satisfiesInitClock(EM->getContributionClock(vexp))) apply = !EM->IsTypeExprPredefinedType(vexp); 
	     */

	    if (apply)
	      {
		ln = EM->getNodesFeedBy (vexp);	// Set of the nodes connected to the expression vexp.
		if (selectNodeForUnification (ln, &selectedNode))
		  {
		    kept = NM->get_leftOperand (selectedNode);
		    keptClass = EM->IsSignal (kept);
		    while (ptlINode_get_elem (&ln, &n))
		      if ((n != selectedNode) && CKM->eqClock(NM->get_nodeClock(n), NM->get_nodeClock(selectedNode)))
			  {
			    replaced = NM->get_leftOperand (n);
			    if (keptClass == EM->IsSignal (replaced))	// not for the constants
			      if (EM->mayBeRemoved (replaced))
				{
				  EM->substitute (replaced, kept, CKM->getUndefClock ());
				  NM->dependencesPasteOn (n, selectedNode);
				  NM->set_successors (n, NILLDEPENDENCE);
				  removeNode (n);
				  b = pk_true;
				}
			      else
				{ 
				  if (EM->IsOutputFlow (replaced))	// X:=E, Y:=E and X,Y are outputs Y:=E is replaced by Y:=X
				    {
				      h = EM->getClock (kept);
				      NM->modifiedDefinition (replaced, kept, h, h);
				    }
				}
			  }
		  }
	      }
	  }
  if (functionUnification ())
    b = pk_true;
  if (delayUnification ())
    b = pk_true;
  return b;
}

/** 
    It marks nodes X:=E (E is not a delay expression) for which X is not referenced more than maxRef 
    time in the  graph.
    The node marked by the predefined mark \b DontOptimize are not considered.
    @return "true" if some node are selected, "false" otherwise.
*/
boolean
DCG::markReferencedMaxTime (typeMark mark, int maxRef, boolean applyToClock)
{
  int n;
  noperators op;
  boolean b, isaClock, bapply;
  ptlINode lNodes, ln;
  ptINode node, cn;
  ptexpcal vopg, vopd, vexpdedef, vdedef, del, win, vinit;
  ptinsth vh, vhs;
  ptlinsth lh;

  require_managers (EM, NM, CKM);
  /*
   * First, the signals that cannot be optimized are marked : - signals operand of the delay nodes, - assigments of the initial part of an array. 
   */
  if (NM->getDelays (&lNodes))
    while (ptlINode_get_elem (&lNodes, &node))
      {
	NM->get_delay_arg (node, &vopg, &vopd, &del, &win, &vinit);
	if (EM->IsSignal (vopd))
	  NM->addMark (EM->getNode (vopd), -mark);
      }

  if (NM->getIterates (&lNodes))
    while (ptlINode_get_elem (&lNodes, &node))
      if (NM->get_array_initDefinitions (node, &ln, &lh))
	while (ptlINode_get_elem (&ln, &cn))
	  {
	    vopg = NM->get_leftOperand (cn);
	    if (vopg != NILEXPCAL)
	      if (EM->satisfiesList (vopg))
		{
		  while (EM->getNaryArgs (&vopg, &vopd))
		    NM->addMark (EM->getNode (vopd), -mark);
		}
	      else 
		{
		  if (EM->IsSignal (vopg))
		    NM->addMark (EM->getNode (vopg), -mark);
		}
	  }

  /*
   * Second : optimizing the equations (X:=E, excluding the delays), some restrictions are applied : - excluding the nodes marked by the predefined 
   * mark "DontOptimize", - excluding the cells, the defaults, the left_tt, - internal nodes of the arrays, - definitions of unpredefined types. 
   */
  b = pk_false;
  if (NM->getEquations (&lNodes))
    while (ptlINode_get_elem (&lNodes, &node))
      if (!NM->isMarked (node, DontOptimize))
	if (!NM->isMarked (node, -mark))
	  {
	    vopg = NM->get_leftOperand (node);
	    if (NM->getInArrayNode (node, &cn))
	      bapply = pk_false;
	    else
	      {
		if (EM->getOperatorClass (vopg) != variabl)
		  bapply = pk_false;
		else if (EM->IsTypeExprPredefinedType (vopg))
		  bapply = pk_true;
		else if (EM->getNumberOfDimensions (vopg) == 0)
		  bapply = TypeTestExternClass (EM->getType (vopg));
		else
		  bapply = pk_false;
	      }
	    if (bapply)
	      if (!EM->IsOutputFlow (vopg))	// pas les sorties.
		if (NM->get_rightOperand (node, &vdedef))
		  if (EM->getType (vopg) == EM->getType (vdedef))	// Meme type 
		    if (!EM->IsCellExpr (vdedef))	// pas les mÃ©moires.
		      {
			vexpdedef = EM->getBasicExpr (vdedef);
			op = EM->getOperator (vexpdedef);
			bapply = pk_false;
			if ((op != opleft_tt) && (op != opmele))	// les default, leftTT sont conservÃ©s.
			  {
			    // x := e(x1,..xn) when vh, cas traitÃ© ^x1 = vh
			    vh = CKM->getRepresentativeFormula (EM->getContributionClock (vdedef));
			    if (CKM->satisfiesInitClock (vh))
			      bapply = pk_true;
			    else if (EM->getSynchronousExprHorl (vexpdedef, &vhs))
			      {
				vhs = CKM->getRepresentativeFormula (vhs);
				bapply = (vh == vhs);
			      }
			  }
			if (bapply)
			  {
			    isaClock = EM->getExtractionLink (vopg, &vh);
			    if (isaClock && !applyToClock)
			      bapply = pk_false;
			  }
			if (bapply)
			  {
			    n = NM->get_successors (node)->getSize ();
			    if (isaClock && (n <= maxRef))
			      {
				n = n + ptlINode_get_size (CKM->getNodeReferencesTo (vh));
				if (n <= maxRef)
				  if (CKM->getComplExtraction (vh, &vh))
				    n = n + ptlINode_get_size (CKM->getNodeReferencesTo (vh));
			      }
			    if (n <= maxRef)	// and the clocks ?.
			      {
				NM->addMark (node, mark);
				b = pk_true;
			      }
			  }
		      }
	  }
  NM->resetMarkAll (-mark);
  return b;
}

/** 
    For n a node  of ln, if n is marked by mark, it is removed from the graph, otherwise
    the dependences (n -> m) with m marked by mark are removed.
*/
void
DCG::cleaningNodes (ptlINode ln, typeMark mark)
{
  ptINode node;

  require_node_manager (NM);
  while (ptlINode_get_elem (&ln, &node))
    if (!NM->isMarked (node, mark))
      NM->removeDependenceMarkedSuccessors (node, mark);
    else
      removeNode (node);
}

/** 
    For h a clock of lh, not marked by mark: 
    - the reference to h in a formula k, marked by mark, is removed, 
    - the nodes assigned to h marked by mark are removed. 
*/
void
DCG::cleaningClocks (ptlinsth lh, typeMark mark)
{
  ptinsth h;

  require_node_manager (NM);
  require_clock_manager (CKM);
  NM = getNodeManager ();
  while (ptlinsth_get_next (&lh, &h))
    if (CKM->satisfiesMarked (h, mark))
      {
	NM->removeDependenceTo (h);
	NM->removeDependenceFrom (h);
	CKM->removeClock (h);
      }
}

/**
   Cleaning the objects marked by mark or referencing objects marked by mark.
   \see cleaningNodes, cleaningClocks
*/
void
DCG::removeMarkedNodes (typeMark mark)
{
  ptlINode lNodes;
  ptlinsth lhorl;

  /* Cleaning nodes */
  require_node_manager (NM);
  if (NM->getEquations (&lNodes))
    cleaningNodes (lNodes, mark);
  if (NM->getDelays (&lNodes))
    cleaningNodes (lNodes, mark);
  if (NM->getProcCalls (&lNodes))
    cleaningNodes (lNodes, mark);
  if (NM->getNodeCalls (&lNodes))
    cleaningNodes (lNodes, mark);
  if (NM->getSubGraphs (&lNodes))
    cleaningNodes (lNodes, mark);
  if (NM->getGroupsOfPartialEquations (&lNodes))
    cleaningNodes (lNodes, mark);

  /* Cleaning clocks */
  require_clock_manager (CKM);
  if (isEventClockedFormat ()) CKM->initSystemForModification (); // Added july 2016.
  lhorl = CKM->getlhroots ();
  cleaningClocks (lhorl, mark);
  lhorl = CKM->getlhplaced ();
  cleaningClocks (lhorl, mark);
  CKM->finalizeClockHierarchy (); 

  removeUndefinedSignals (); // Don't remove this line.
}

/**
   @return "true" two constants are unified, "false" otherwise.
   For c a constant, element of the list lform,
   if c and aConst are defined by the same syntatic expression, c and aConst are unified (keeping aConst), in  this case
   c is then added to the list completelist (in/out  parameter).
*/
boolean
DCG::UnifyConstant (ptexpcal aConst, ptlinstances lfrom, ptlinstances * completelist)
{
  ptexpcal vopg, vopd, currentCst, vopg2, vopd2;
  ptINode nConst, n;
  TypeType aConstType;

  require_managers (EM, NM, CKM);
  // Voir si ce n'est pas un cas particulier ()
  if (EM->IsExternalConstant (aConst))
    return pk_false;
  nConst = EM->getNode (aConst);
  vopg = NM->get_leftOperand (nConst);
  if (!NM->get_rightOperand (nConst, &vopd))
    return pk_false;
  aConstType = EM->getType (vopg);
  while (ptlinstances_get_next (&lfrom, &currentCst))
    if (!EM->IsExternalConstant (currentCst))
      {
	n = EM->getNode (currentCst);
	vopg2 = NM->get_leftOperand (n);
	if (aConstType == EM->getType (vopg2))
	  if (NM->get_rightOperand (n, &vopd2))
	    if (vopd2 == vopd)
	      {			/* unicity of the expressions */
		EM->substitute (currentCst, aConst, CKM->getUndefClock ());
		NM->dependencesPropagate (n);
		removeNode (n);
		EM->setIndirection (currentCst, aConst);
		*completelist = ptlinstances_remove_exp (*completelist, currentCst);
		return pk_true;
	      }
      }
  return pk_false;
}

/**
   For an equation x:= E, marked with a mark (mark), x is replaced by E in the graph 
   and the equation is removed from the graph.
*/
void
DCG::substituteMarkedVariables (typeMark mark)
{
  ptlINode lnodes;
  ptINode node;
  ptexpcal vopg, vopd;
  ptlinstances ls;

  require_managers (EM, NM, CKM);
  if (NM->getEquations (&lnodes))
    {
      while (ptlINode_get_elem (&lnodes, &node))
	{
	  if (NM->isMarked (node, mark))
	    {
	      NM->get_equation_arg (node, &vopg, &vopd);
	      if (EM->IsOutputFlow (vopg))
		{
		  ls = getOutputs ();
		  ls = ptlinstances_remove_exp (ls, vopg);
		  ptlinstances_insert (ls, vopd);	// I hope that vopd is a signal!!
		  setOutputs (ls);
		  EM->belongsLocalFlow (vopg);
		}
	      if (EM->IsAssigned (vopg))
		EM->getEquivConstant (vopg, &vopd, pk_false);
	      if (EM->mayBeRemoved (vopg))
		{
		  ptinsth vopg_clockFormula, vopg_clock;
		  vopg_clockFormula = EM->getClock (vopg);
		  vopg_clock = CKM->getRepresentativeFormula (vopg_clockFormula);
		  EM->substitute (vopg, vopd, CKM->getRepresentativeFormula (EM->getClock (vopg)));
		  NM->dependencesPropagate (node);
		  removeNode (node);
		}
	    }
	}
    }
  // To be completed ?
}

/** 
    It unifies the nodes {N1,....,Nn} equivalent (defined by the same syntactic expression)
    to a node (node parameter) marked by mark keeping node. {N1,....,Nn} are removed from the graph.
*/
void
DCG::unifyEquivalentTo (ptINode node)
{
  ptexpcal vexp, kept, replaced;
  ptINode n;
  ptlINode ln;
  boolean keptClass;

  require_managers (EM, NM, CKM);
  kept = NM->get_leftOperand (node);
  keptClass = EM->IsSignal (kept);
  if (NM->get_rightOperand (node, &vexp))
    {
      ln = EM->getNodesFeedBy (vexp);
      while (ptlINode_get_elem (&ln, &n))
	if (n != node)
	  {
	    replaced = NM->get_leftOperand (n);
	    if (keptClass == EM->IsSignal (replaced))
	      // LBesnard: the constants are filtered, Why ?
	      if (EM->mayBeRemoved (replaced))
		{
		  EM->substitute (replaced, kept, CKM->getUndefClock ());
		  NM->dependencesPasteOn (n, node);
		  NM->set_successors (n, NILLDEPENDENCE);
		  removeNode (n);
		}
	  }
    }
}

/**
   @return "true" if a node (selectedNode, Xi:=E) is selected in the operand list (ln). Xi becomes the 
   representative signal of a set of signals {X1,...,Xn} such that nodes X1:= E, .... Xn:= E (n>1)
   "false" otherwise.
   The priority is given to an output signal, and the expression E can be a delay expression.
*/
boolean
DCG::selectNodeForUnification (ptlINode ln, ptINode * selectedNode)
{
  ptINode n, other;
  ptexpcal vexp;

  require_managers (EM, NM, CKM);
  if (ptlINode_get_size (ln) < 2)
    return pk_false;
  other = NILINODE;
  while (ptlINode_get_elem (&ln, &n))
    if (NM->is_equation (n) || NM->is_delay (n))
      {
	vexp = NM->get_leftOperand (n);
	if (EM->IsOutputFlow (vexp))
	  {
	    *selectedNode = n;
	    return pk_true;
	  }
	if (EM->IsSignal (vexp))
	  other = n;
      }
    else
      return pk_false;

  if (other == NILINODE)
    return pk_false;
  *selectedNode = other;
  return pk_true;
}

/** 
    It marks nodes defined by x := y where y is a signal and x is not an output.
    The node marked by the predefined mark \b DontOptimize are not considered.
    @return "true" if some nodes are selected, "false" otherwise.
*/
boolean
DCG::markRenamming (typeMark mark)
{
  ptlINode lNodes;
  ptINode node, n;
  ptexpcal vopg, vopd;
  boolean b, ok;

  require_managers (EM, NM, CKM);

  b = pk_false;
  if (NM->getEquations (&lNodes))
    while (ptlINode_get_elem (&lNodes, &node))
      if (!NM->isMarked (node, DontOptimize) && !NM->isMarked (node, mark))
	{
	  vopg = NM->get_leftOperand (node);
	  if (EM->getOperator (vopg) == opident)	// not for the fields.
	    if (NM->get_rightOperand (node, &vopd))
	      if ((EM->getType (vopg) == EM->getType (vopd)) && !EM->IsCellExpr (vopd))
		{
		  ok = pk_false;
		  vopd = EM->getBasicExpr (vopd);
		  if ((vopd != vopg) && (EM->getOperator (vopd) == opident) && CKM->eqClock (EM->getClock (vopg), EM->getClock (vopd)))
		    {
		      // vopd = EM->getBasicExpr(vopd);
		      if (EM->mayBeRemoved (vopg))
			ok = pk_true;
		      else
			/*
			 * (| x := y | y := E |) x an output, and y not an output, is replaced by (| x := E | y := x |). 
			 */
		      if (EM->mayBeRemoved (vopd))
			{
			  n = EM->getNode (vopd);
			  ok = NM->switchingNodes (node, n);
			  node = n;
			}
		      if (ok)
			{
			  NM->addMark (node, mark);
			  b = pk_true;
			}
		    }
		}
	}
  return b;
}

/** 
    Marking by the operand mark (mark) of the scalar constant signal for substitution .
    @return "true" if it exists such signals, "false" otherwise.
*/
boolean
DCG::markScalarConstantSignal (typeMark mark)
{
  ptlINode lNodes;
  ptINode node;
  boolean b;
  ptexpcal vopg;

  require_expr_manager (EM);
  require_node_manager (NM);
  b = pk_false;
  if (NM->getEquations (&lNodes))
    while (ptlINode_get_elem (&lNodes, &node))
      if (!NM->isMarked (node, DontOptimize))
	{
	  vopg = NM->get_leftOperand (node);
	  if (!EM->IsOutputFlow (vopg))
	    if (EM->getOperator (vopg) == opident)
	      if (EM->IsTypeExprPredefinedType (vopg))
		if (EM->IsAssigned (vopg))
		  {
		    b = pk_true;
		    NM->addMark (node, mark);
		  }
	}

  /*
   * if (NM->getProcCalls(&lNodes)) while (ptlINode_get_elem(&lNodes,&node)) if (is_constantCall(node)) { b = pk_true; NM->addMark(node,mark); } 
   */
  return b;
}

/**
   Assumed : The operator of the node argument is "opdefsig" (equation, not partial, not a delay).
   @return 
   - "true" if a node (node) is now defined by a constant,
   - "false" otherwise.
*/
boolean
DCG::makeConstantDefsig (ptINode node)
{
  ptexpcal nexp, vopg, vopd, vinit;
  compilerInteger lb1, ub1, v1, v2;
  boolean bcell;

  // printf("DEBUG entering DCG::makeConstantDefsig\n"); fflush(stdout);
  require_managers (EM, NM, CKM);
  vopg = NM->get_leftOperand (node);
  // printf("DEBUG in DCG::makeConstantDefsig 1 vopg adr is %li\n", (long int) vopg); fflush(stdout);
  if (EM->getOperator (vopg) != opident)
    return pk_false;
  if (!EM->IsTypeExprPredefinedType (vopg))
    return pk_false;
  if (!NM->get_rightOperand (node, &vopd))
    return pk_false;
  if (EM->IsAssigned (vopg))
    return pk_false;
  if (EM->IsBounded (vopg))
    return pk_false;
  bcell = EM->IsCellExpr (vopd);
  vopd = EM->getBasicExpr (vopd);
  // printf("DEBUG in DCG::makeConstantDefsig 2 vopd adr is %li\n", (long int) vopd); fflush(stdout);
  if (!EM->IsAssigned (vopd))
    {
      // printf("DEBUG in DCG::makeConstantDefsig 3 vopd is not assigned\n"); fflush(stdout);
      if (EM->getBounds (vopd, &lb1, &ub1))
	{
	  if (bcell)
	    {
	      vinit = EM->getInitialValue (vopg);
	      if (!EM->getBounds (vinit, &v2, &v2))
		return pk_false;
	      lb1 = (lb1 < v2 ? lb1 : v2);
	      ub1 = (ub1 > v2 ? ub1 : v2);
	    }
	  EM->setBounds (vopg, lb1, ub1);
	}
      // printf("DEBUG EXITING DCG::makeConstantDefsig 3 vopd is not assigned\n"); fflush(stdout);
      return pk_false;
    }
  v1 = EM->getValue (vopd);
  // printf("DEBUG in DCG::makeConstantDefsig 4 the value of vopd is %li\n", v1); fflush(stdout);
  if (bcell)
    {
      // printf("DEBUG in DCG::makeConstantDefsig 5\n"); fflush(stdout);
      // constant memory ?
      vinit = EM->getInitialValue (vopg);
      if (!EM->getBounds (vinit, &v2, &v2))
	return pk_false;
      if (v1 != v2)
	return pk_false;
    }
  // printf("DEBUG in DCG::makeConstantDefsig 6\n"); fflush(stdout);
  EM->set_integerValueHolder (vopg, v1);
  // printf("DEBUG in DCG::makeConstantDefsig 7 the value %li has been assigned to %li\n", v1, (long int) vopg); fflush(stdout);
  if (EM->getOperatorClass (vopd) != constant)
    {
      // printf("DEBUG in DCG::makeConstantDefsig 16\n"); fflush(stdout);
      NM->resetNodeForModification (node);
      // printf("DEBUG in DCG::makeConstantDefsig 17\n"); fflush(stdout);
      EM->getExprOfConstant (vopd, &nexp);
      // printf("DEBUG in DCG::makeConstantDefsig 18\n"); fflush(stdout);
      NM->modifiedDefinition (vopg, nexp, CKM->getInitClock (), EM->getClock (vopg));
      // printf("DEBUG in DCG::makeConstantDefsig 19\n"); fflush(stdout);
      EM->setInitialValue (vopg, NILEXPCAL);
      // printf("DEBUG in DCG::makeConstantDefsig 20\n"); fflush(stdout);
      ptlINode_insert_elem (&lNodesAtInitClock, node);
      // printf("DEBUG in DCG::makeConstantDefsig 21\n"); fflush(stdout);
    }
  // printf("DEBUG exiting DCG::makeConstantDefsig *\n"); fflush(stdout);
  return pk_true;
}

/**
   It defines the constant signals of the program.
   For each equation (not partial, not delay), it calls makeConstantDefsig().
*/
boolean
DCG::makeConstantSignals ()
{
  ptlINode lNodes;
  ptINode node;
  boolean b;

  // printf("DEBUG entering DCG::makeConstantSignals\n"); fflush(stdout);
  require_node_manager (NM);
  b = pk_false;
  // printf("DEBUG in DCG::makeConstantSignals 1\n"); fflush(stdout);
  if (NM->getEquations (&lNodes))
    while (ptlINode_get_elem (&lNodes, &node))
      {
	// printf("DEBUG in DCG::makeConstantSignals in while\n"); fflush(stdout);
	if (makeConstantDefsig (node))
	  b = pk_true;
      }
  // printf("DEBUG exiting DCG::makeConstantSignals\n"); fflush(stdout);
  return b;
}

/**
   Adding boolean signal X defined by X:=E, for E used in "when/cell E" and E is not reduced to a signal.
   When keepCondIO is pk_true, boolean expression build on input/outputs signals are not considered.
   When keepLogicalOperators is pk_true, boolean expression build on {or, and, not} are not considered.
   @return "true" if some signals are added, "false" otherwise.
*/
boolean
DCG::undersamplingOnSignal (boolean keepCondIO, boolean keepLogicalOperators)
{
  noperators op;
  ptexpcal s, e, ne;
  ptlinsth l;
  ptinsth h;
  boolean b, bstop, bapply;
  ptlinstances ls, lins, louts;

  require_managers (EM, NM, CKM);
  l = ptlinsth_duplicate (CKM->getlhroots ());
  l = ptlinsth_sorted_fusion (l, CKM->getlhplaced ());
  b = pk_false;

  lins = getInputs ();
  louts = getOutputs ();

  while (ptlinsth_get_next (&l, &h))
    if (CKM->getRepresentativeFormula (h) == h)
      if (CKM->getOperator (h) == sub)
	{
	  e = CKM->getConditionSub (h);
	  op = EM->getOperator (e);
	  if (op != opident)
	    {
	      if (keepCondIO)
		{
		  ls = CKM->getSignalsOfCondition (h);
		  bstop = pk_false;
		  while (bstop && ptlinstances_get_next (&ls, &s))
		    bstop = !(ptlinstances_isin_exp (lins, s) || ptlinstances_isin_exp (louts, s));
		  bapply = !bstop;
		}
	      else if (keepLogicalOperators)
		bapply = ((op != opnon) && (op != opet) && (op != opou) && (op != opleft_tt));
	      else
		bapply = pk_true;

	      if (bapply)
		{
		  b = pk_true;
		  ne = NM->createWhen (e, CKM->getClockSub (h), pk_true);
		  CKM->rewriteCondSub (h, ne, EM->getSignalsReferencedIn (ne));
		}
	    }
	}

  return b;
}

ptlinstances
DCG::getSignalsRefinClocks (ptlinsth lh, ptlinstances lsig)
{
  ptinsth h;

  require_managers (EM, NM, CKM);
  while (ptlinsth_get_next (&lh, &h))
    {
      h = CKM->getRepresentativeFormula (h);
      if (CKM->getOperator (h) == sub)
	lsig = ptlinstances_sortedfusion (CKM->getSignalsOfCondition (h), lsig);
    }
  return lsig;
}

/**
   It selects, using the mark m, the equations X := E , where type(X) is boolean and E is build on a boolean operator (or, and, not).
   The selected signals are removed from the operand list (lsig). The resulting list is returned.
*/
_VIRTUAL_ ptlinstances
DCG::booleanNodeSelecting (ptlinstances lsig, typeMark m)
{
  ptINode node;
  ptexpcal aSig, vopg, vopd;
  ptlinstances l;
  boolean b;

  require_managers (EM, NM, CKM);
  b = pk_true;
  while (b)
    {
      b = pk_false;
      l = lsig;
      while (ptlinstances_get_next (&l, &aSig))
	if (!EM->isMarked (aSig, m))
	  {
	    node = EM->getNode (aSig);
	    if (!ptlINode_isIn (unusedNodes, node))
	      if (NM->is_equation (node))
		{
		  NM->get_equation_arg (node, &vopg, &vopd);
		  if (EM->isLogicalBooleanExpression (vopd))
		    {
		      EM->addMark (aSig, m);	// aSig = vopg
		      NM->addMark (node, m);
		      lsig = ptlinstances_remove_exp (lsig, aSig);
		      lsig = ptlinstances_sortedfusion (EM->getSignalsReferencedIn (vopd), lsig);
		      b = pk_true;
		    }
		}
	  }
	else
	  lsig = ptlinstances_remove_exp (lsig, aSig);
    }
  return lsig;
}

/**
   The useful signals become outputs of the graph.
*/
_VIRTUAL_ void
DCG::declareSignalsAsOutputs (ptlinstances lsig)
{
  ptexpcal aSig;
  ptINode n;

  require_managers (EM, NM, CKM);
  while (ptlinstances_get_next (&lsig, &aSig))
    if (EM->IsSignal (aSig))
      if (!EM->IsInputFlow (aSig))
	{
	  n = EM->getNode (aSig);
	  if (!ptlINode_isIn (unusedNodes, n))
	    addOutput (aSig);
	}
}

/**
   @return "true" if aSig is a declared I/O signal, "false" otherwise.
*/
boolean
DCG::isIOSignal (ptexpcal aSig)
{
  if (ptlinstances_isin_exp (getInputs (), aSig))
    return pk_true;
  if (ptlinstances_isin_exp (getOutputs (), aSig))
    return pk_true;
  return pk_false;
}

/**
   @return "true" if all the graph of lg are marked with mark value, "false" otherwise.
*/
boolean
DCG::satisfiesMarkedGraphs (ptlDCG lg, typeMark mark)
{
  pointer elem;
  // ptDCG g;
  ptDCG g;

  while (ptlPointer_get_elemNext (&lg, &elem))
    {
      g = (ptDCG) elem;
      if (!g->satisfiesMarked (mark))
	return pk_false;
    }
  return pk_true;
}

/** 
    Builts the lists of external proc/func. They are respectively stored in list_extern_proc and list_extern_funcs attributes.
    @param mark (in) temporary mark used to optimize the algorithm.
    \see setListExternalProc, getListExternalProcedure
*/
void
DCG::setExternalReferencesUsingMark (typeMark mark)
{
  ptlDCG lgraphs;
  ptDCG g;
  ptlINode lNodes, lExtern, lext;
  pointer elem;
  ptINode aNode;
  ptlinstances lExtern2;
  ptident name;
  ptexpcal aCall, lp, lin;
  ptlinstances lfuncs, x;

  require_expr_manager (EM);
  require_node_manager (NM);
  if (satisfiesMarked (mark))
    return;
  addMark (mark);
  if (!getModuleGraphs (&lgraphs))
    if (!getReferencedGraphs (&lgraphs))
      lgraphs = NILLDCG;
  while (ptlPointer_get_elemNext (&lgraphs, &elem))
    {
      g = (ptDCG) elem;
      g->setExternalReferencesUsingMark (mark);
    }

  lExtern = NILLINODE;
  lExtern2 = NILLINSTANCES;

  setListExternalProc (lExtern, lExtern2);
  if (NM->getProcCalls (&lNodes))
    while (ptlINode_get_elem (&lNodes, &aNode))
      lExtern = NM->addExternDecl (aNode, lExtern);

  if (EM->getExprList (opappe, &lfuncs))
    while (ptlinstances_get_next (&lfuncs, &aCall))
      {
	EM->getFunctionCall (aCall, &name, &lp, &lin);
	if (!NM->isDeclaredExternal (name, lExtern))
	  ptlinstances_isin_insert_exp (&lExtern2, &x, aCall);
      }

  if (getReferencedGraphs (&lgraphs))
    while (ptlPointer_get_elemNext (&lgraphs, &elem))
      {
	g = (ptDCG) elem;
	lext = g->getListExternalProcedure (&lfuncs);
	while (ptlINode_get_elem (&lext, &aNode))
	  lExtern = NM->addExternDecl (aNode, lExtern);
	while (ptlinstances_get_next (&lfuncs, &aCall))
	  {
	    EM->getFunctionCall (aCall, &name, &lp, &lin);
	    if (!NM->isDeclaredExternal (name, lExtern))
	      ptlinstances_isin_insert_exp (&lExtern2, &x, aCall);
	  }
      }
  setListExternalProc (lExtern, lExtern2);
}



  /** 
    Update the link Signal/clock and node/clock of the current graph.
*/
void
DCG::updateGraphClock (ptlinsth lh)
{
  ptinsth vh;
  ptlinstances lsig;
  ptexpcal vsig;
  ptlINode ln;
  ptINode node;

  require_managers (EM, NM, CKM);
  while (ptlinsth_get_next (&lh, &vh))
    if (CKM->getRepresentativeFormula (vh) == vh)
      {
	lsig = CKM->getSynchronousSignals (vh);
	while (ptlinstances_get_next (&lsig, &vsig))
	  EM->setClock (vsig, vh);
	ln = ptgfd_get_nodes (CKM->getClockSubGraph (vh));
	while (ptlINode_get_elem (&ln, &node))
	  NM->set_nodeClock (node, vh);
      }
}

/** 
    Updates the references to clocks in the graph.
    (Recall: Clock calculus doesn't manage the graph!).
*/
void
DCG::updateClocksGraph ()
{
  ptinsth vh1, vh2, x;
  ptlinsth lh1, lh2;
  ptlINode ln, ln1;
  ptINode node;
  boolean b;

  require_managers (EM, NM, CKM);
  // updating clock of the nodes for assertions.
  if (CKM->getAssertionsOnPlacedClocks (&lh1, &lh2))
    {
      NM->getAllNodes (&ln1);
      while (ptlinsth_get_next (&lh1, &vh1))
	{
	  ptlinsth_get_next (&lh2, &vh2);
	  // vh1 = CKM->getRepresentativeFormula(vh1); vh2 = CKM->getRepresentativeFormula(vh2);
	  b = pk_false;
	  if (CKM->satisfiesIsAnAssumption (vh1))
	    if (!CKM->satisfiesIsAnAssumption (vh2))
	      b = pk_true;
	    else;
	  else if (CKM->satisfiesIsAnAssumption (vh2))
	    if (!CKM->satisfiesIsAnAssumption (vh1))
	      {
		x = vh2;
		vh2 = vh1;
		vh1 = x;
		b = pk_true;
	      }
	  if (b)
	    {
	      ln = ln1;
	      while (ptlINode_get_elem (&ln, &node))
		if (NM->get_nodeClock (node) == vh1)
		  NM->set_nodeClock (node, vh2);
	    }
	}
    }

  updateGraphClock (CKM->getlhroots ());
  updateGraphClock (CKM->getlhplaced ());
}

/**
   @return "true" if the graph contains assumptions, "false" oherwise.
*/
boolean
DCG::existAssumptions ()
{
  ptlINode ln;
  ptINode n;
  ptident name;

  require_node_manager (NM);
  require_clock_manager (CKM);
  if (isEventClockedFormat ())
    return CKM->isSystemWithAssertions ();
  // Assumptions have been translated in nodes for DCGBool format. (faster than isSystemWithAssertions used for DCGPoly ?)
  NM = getNodeManager ();
  if (NM->getProcCalls (&ln))
    while (ptlINode_get_elem (&ln, &n))
      if (NM->get_nameCall (n, &name))
	if (NM->is_intrinsicfunc (n) && isAssertionId (name))
	  return pk_true;
  return pk_false;
}

/** 
    @return an extraction clock (undersampling by an expression) that represents a clock (clk) at another clock (hAt).
    If the clock hAt is the undefined clock, the clock clk is returned.
    
    The following order is used:
    - searching an equivalent
    - searching an equivalent placed above hAt (hAt is included in such a clock), then creating a when. 
    - Creating a signal defined by a left_tt that represent the clock to hAt.

    \note clk and hAT are assumed to be clocks such that the uper placed bound (in the hierarchy) is hAt.
*/
ptinsth
DCG::mkBooleanClockScopeExtand (ptinsth clk, ptinsth hAT)
{
  ptinsth vhresult;
  ptexpcal vsig, bclk;
  int rcode;

  require_managers (EM, NM, CKM);

  if (CKM->satisfiesUndefClock (hAT))
    return clk;

  rcode = CKM->getBooleanClockScopeExtandAt (clk, hAT, &vhresult);
  if (rcode == 0)
    return vhresult;
  bclk = CKM->getBooleanSignal (clk);
  if (rcode == 1)
    vsig = NM->createWhen (bclk, hAT, pk_false);
  else
    vsig = NM->createLeftt (bclk, CKM->getBooleanSignal (vhresult), hAT);
  return CKM->makeEquivForm (clk, vsig, EM->getPredecessors (vsig), hAT);
}

/** 
    @return the name of the clock for DCGBool sub-format.
*/
ptident
DCG::ChoiceNameBooleanClock (ptinsth clk)
{
  ptlinstances lsig;
  ptident id;

  require_clock_manager (CKM);
  require_expr_manager (EM);
  if (CKM->satisfiesTickClock (clk))
    return string2ident ("Tick");
  lsig = CKM->getSynchronousSignals (clk);
  if (!EM->getident (lsig, &id))
    id = CKM->getName (clk);
  if (id == NILIDENT)
    return getPrefixBooleanClock ();
  return namBooleanClockfrom (id);
}

void
DCG::removeSignal (ptexpcal sig)
{
  ptlinstances vli;
  ptinsth vh;

  require_managers (EM, NM, CKM);
  vh = EM->getClock (sig);
  if (CKM->satisfiesNullClock (vh))
    ptlinstances_isin_insert_exp (&lSigNullClock, &vli, sig);
  else
    CKM->removeFromClock (vh, sig, EM->IsEvent (sig));
  EM->remove (sig);
}

/**
   It completes the interface of a sub-graph (SubGraph) by computing the branching clock of all the clocks/clocks of the 
   nodes/clocks of the input/output signals of the sub-graph.
   It calls ClockManager::CompleteUpstreamParametrization().
*/
void
DCG::upstreamProjectionSubGraph (ptINode SubGraph)
{
  ptlinsth lh, lhtodeclare;
  ptinsth vhemb, vh;
  ptinterface vinterface;

  require_managers (EM, NM, CKM);
  // ptident id; id = getName(); pK_messenger->messageIdent(mbegprojection, id);
  lh = EM->getClockOfSignals (NM->getInodelInputFlows (SubGraph));
  lh = ptlinsth_sorted_fusion (lh, EM->getClockOfSignals (NM->getInodelOutputFlows (SubGraph)));
  lh = ptlinsth_sorted_fusion (lh, NM->getInodelInputClocks (SubGraph));
  vhemb = CKM->getBranchingClocks (lh);
  lhtodeclare = CKM->CompleteUpstreamParametrization (lh, vhemb);
  vh = CKM->getFather (vhemb);
  if (!CKM->satisfiesFictitiousRoot (vh))
    vhemb = vh;
  NM->set_subGraphProjectedSystemInfos (SubGraph, vhemb, lhtodeclare);
  // Modifying the interface.
  vinterface = NM->getSet_interface (SubGraph);
  while (ptlinsth_get_next (&lhtodeclare, &vh))
    ptinterface_insert_inputClock (vinterface, CKM->getRepresentativeFormula (vh));
  // The root is not an input of the sub-graph.
  if (!isEventClockedFormat ())
    {
      ptinterface_remove_inputClock (vinterface, vhemb);
      ptinterface_remove_inputSignal (vinterface, CKM->getBooleanSignal (vhemb));
    }
}

/** 
    Setting dependencies for code generators.
    It sets dependencies for complementary clock.
    @return none.
*/
void
DCG::setsbDCplusDependencesForComplementaryClocks ()
{
  ptinsth vh;
  ptlinsth lh;

  require_clock_manager (CKM);
  lh = CKM->getFormulaList (hnot);
  while (ptlinsth_get_next (&lh, &vh))
    CKM->modifyDependenceComplementaryClock (vh);
}


/**
   Check if all the signals have an explicit type.
   @return "true" if the signals have a explicit type, "false" otherwise.
*/
boolean
DCG::checkExplicitTypingOneLevel ()
{
  boolean b;
  ptlinstances ls;
  ptexpcal id;
  ptlINode ln;
  ptlinsth lh;

  if (checkExplicitTypingApplied) return pk_true;
  checkExplicitTypingApplied= true;
  require_expr_manager (EM);
  if (isGraphOfExternalProcess ()) return pk_true;
  b = pk_true;
  if (EM->getListNotTypedId (&ls))
    {
      ln = getUnused (&lh);
      while (ptlinstances_get_next (&ls, &id))
	// sans doute inutile (LB)
	if (!ptlINode_isIn (ln, EM->getNode (id)))
	  b = pk_false;
    }
  if (!b) pK_messenger->errorIdent (errNotTypedId, getName ());

  // Solving conflict names (enum/variables).
  if (EM->enumNameConflicts()) pK_messenger->message (infoEnumNameModified);

  return b;
}

/**
   Set the clocks of process arrays(see arrayClockUtl(), fixArrayClocks()).
   \ref GraphBuilder "For more details..."
*/
boolean
DCG::setArrayClocksOneLevel ()
{
  ptarrayClockUtl obj;
  boolean b;
  ptlINode lArrayNodes;

  b = pk_true;
  if (getNodeManager ()->getIterates (&lArrayNodes))
    {
      obj = new arrayClockUtl (this);
      b = obj->fixArrayClocks ();
      delete obj;
    }
  return b;
}

/** 
    It splits the structured signals into a set of signals (one for each type of the structured type).

    \ref GraphBuilder "For more details..."
*/
void
DCG::splitDataStructureOneLevel ()
{
  ptsplitting splittingObject;

  splittingObject = new splitting (this);
  splittingObject->splitStructures ();
  delete splittingObject;
}

/**
   Setting dependencies for the current graph for composed nodes ( array, subgraph, nodecall, partial definition)
*/
void
DCG::setComposedNodesDependencesOneLevel ()
{
  require_node_manager (NM);

  NM->setDependencesNodeCall ();
  NM->setDependencesLabels ();
  NM->setDependencesPartDefinitions ();
  NM->setDependencesArrays ();
  // NM->setDependencesContributionClocks();
}

/**
   Setting dependencies for the current graph.
   Cross references between clocks/nodes.
*/
void
DCG::setClockDependencesOneLevel ()
{
  ptlinsth lr, lp;

  require_clock_manager (CKM);
  require_node_manager (NM);

  NM->removeDependenceNullClock ();
  lr = CKM->getlhroots ();
  lp = CKM->getlhplaced ();
  NM->setDependencesINodeToClocks (lp);
  NM->setDependencesINodeToClocks (lr);
  NM->setDependencesClocksToINodes (lr);
  NM->setDependencesClocksToINodes (lp);
}

/**
   It replaces the leftt() operator by a default.
*/
void
DCG::lefttts2defaults ()
{
  getNodeManager ()->lefttts2defaults ();
}

/**
   @return the subgraph node of the hierarchy with rootclk as root.
*/
ptINode
DCG::mkSubGraphTreeHierarchy (ptinsth rootclk)
{
  ptlINode lns, ln;
  ptlinsth lhs, lh;
  ptINode aSubGraph, node;
  ptinsth h;
  ptexpcal vlabel;

  require_managers (EM, NM, CKM);
  ln = NILLINODE;
  lh = NILLINSTH;
  getDCG ()->getHierarchyNodesClocks (rootclk, &ln, &lh);
  vlabel = EM->mkLabelIdent (getLabelId (), NM->mk_node_undefined (), CKM->getUndefClock (), NILBACKANNOTATION, pk_true);
  aSubGraph = NM->mk_node_subgraph (vlabel, NILBACKANNOTATION);

  lhs = lh;
  while (ptlinsth_get_next (&lh, &h))
    NM->subGraphInsertClock (aSubGraph, h);
  lns = ln;
  while (ptlINode_get_elem (&ln, &node))
    NM->subGraphInsertNode (aSubGraph, node);

  NM->unlinkNodesFromClock (aSubGraph);

  ptlinsth_free (lhs);
  ptlINode_free (lns);
  return aSubGraph;
}

/**
   It applies the NodeManager::rewDefinition(ptINode) method to the non dynamical equations.
   @return "true" if a node is rewritten, "false" otherwise.

   \todo: to be completed for the partial definitions.
*/
boolean
DCG::rewDefinitions ()
{
  ptlINode lNodes;
  ptINode node;
  boolean b;

  require_node_manager (NM);
  b = pk_false;
  if (NM->getEquations (&lNodes))
    while (ptlINode_get_elem (&lNodes, &node))
      if (NM->rewDefinition (node))
	b = pk_true;

  // Partial definitions.
  if (NM->getGroupsOfPartialEquations (&lNodes))
    while (ptlINode_get_elem (&lNodes, &node))
      if (NM->rewDefinitionPartialDef (node))
	b = pk_true;
  return b;
}

/**
   @return "true" if the verbose mode is set,"false" otherwise.
*/
boolean
DCG::isVerboseMode ()
{
  return modeVerbose;
}

/**
   @return the value of the verbose mode before calling this method.
   It sets the verbose mode to b.
*/
boolean
DCG::setVerboseMode (boolean b)
{
  boolean x;

  x = modeVerbose;
  modeVerbose = b;
  return x;
}

/**
   It partitions the nodes of a subgraph by clocks.
*/
void
DCG::nodeGroupingByClocks (ptINode aSubGraph)
{
  ptlinsth lh1, lh;
  ptlINode lSubGraphs, ln, ln1;
  ptINode aNode;
  ptinsth vh1;

  require_node_manager (NM);
  if (NM->get_subGraphDefinitions (aSubGraph, &ln, &lh))
    {
      lSubGraphs = NILLINODE;
      nodeGroupingByClocks (ln, &lSubGraphs);
      // Replacing the previous nodes of ln by the nodes of lSubGraphs.
      while (ptlINode_get_elem (&ln, &aNode))
	NM->subGraphRemoveNode (aSubGraph, aNode);
      while (ptlINode_get_elem (&lSubGraphs, &aNode))
	{
	  if (NM->get_subGraphDefinitions (aNode, &ln1, &lh1))
	    if (ptlinsth_get_next (&lh1, &vh1))
	      NM->subGraphRemoveClock (aNode, vh1);
	  NM->subGraphInsertNode (aSubGraph, aNode);
	  NM->subGraphLinkNodes (aNode, pk_false);
	}
      NM->subGraphLinkNodes (aSubGraph, pk_false);
    }
}

/**
   Each (partial) definition x:=E becomes x::=E when hu(x) | x::=defaultvalue x$1 where hu(x) denotes the utility clock of x.
*/
boolean
DCG::rewUsingProductiveClock ()
{
  ptlINode lNodes;
  ptINode aNode;
  boolean bres;

  require_managers (EM, NM, CKM);
  bres = pk_false;
  // keep this order, because the rewUsingProductiveClock() for an equation produces partial definitions.
  if (NM->getGroupsOfPartialEquations (&lNodes))
    while (ptlINode_get_elem (&lNodes, &aNode))
      if (NM->rewUsingProductiveClock (aNode))
	bres = pk_true;

  if (NM->getEquations (&lNodes))
    while (ptlINode_get_elem (&lNodes, &aNode))
      if (NM->rewUsingProductiveClock (aNode))
	bres = pk_true;

  return bres;
}

/**
   Default/partial definition normalisation. 
   It applies, by construction H and K are exclusive so the default operator is commutative, the rules:
   - X:= (E when H) default (E when K) --> (E when H ^+ K)
   - X::= ((E when H) | X::= ((E when K) -->  X::= (E when H ^+ K)

   @return "true" if a definition is modified, "false" otherwise.
*/
boolean
DCG::normalDefaults ()
{
  ptlINode lNodes;
  ptINode aNode;
  boolean bres;

  require_managers (EM, NM, CKM);
  bres = pk_false;

  if (NM->getEquations (&lNodes))
    while (ptlINode_get_elem (&lNodes, &aNode))
      if (NM->normalDefaults (aNode))
	bres = pk_true;

  if (NM->getGroupsOfPartialEquations (&lNodes))
    while (ptlINode_get_elem (&lNodes, &aNode))
      if (NM->normalDefaults (aNode))
	bres = pk_true;

  return bres;
}

/**
   @return "true" if the graph is an endochronous with an internal root clock one, "false" otherwise (may be not computed).
   vhComplete : the "complete" clock (io clock, or the complete clock of a Signal thread,...)
*/
boolean
DCG::isEndochronousWithInternalClockRoot (ptinsth *vhComplete)
{
  ptinsth vclk;

  require_clock_manager (CKM);
  vclk = CKM->getTickClock ();
  if (CKM->satisfiesFictitiousRoot (vclk)) return pk_false;
  *vhComplete = getCompleteClock ();  
  if (CKM->eqClock (*vhComplete, CKM->getUndefClock ())) return pk_false; // not computed !
  return (!CKM->eqClock (*vhComplete, vclk));
}


/**
   NOT YET IMPLEMENTED (LBesnard).
   x := E  | y :=F  and x ^= y
   If we use the utility clock as the availibility clock, the signals must be desynchronized.
   MoreOver, it may introduce cycles!!!.
*/
boolean
DCG::setUtilityClocksAsClock ()
{
  ptlinstances lident;
  ptexpcal aSig;
  boolean b;
  ptinsth availClk, utilityClk;
  ptlinstances ls;

  require_clock_manager (CKM);
  require_expr_manager (EM);
  b = pk_false;
  if (EM->getExprList (opident, &lident))
    while (ptlinstances_get_next (&lident, &aSig))
      if (EM->IsSignal (aSig))
	if (EM->getUtilityClock (aSig, &utilityClk))
	  {
	    availClk = EM->getClock (aSig);
	    if (!CKM->eqClock (utilityClk, availClk))
	      {
		printf (" TESTING--setUtilityClocksAsClock, Availibility clock != Utility clock for %s \n", ident2string (EM->getName (aSig)));
		ls = CKM->getSynchronousSignals (availClk);
		if (ptlinstances_get_size (ls) == 1)
		  {
		    CKM->replaceForm (availClk, utilityClk, CKM->getClockReferencesTo (availClk));
		    printf
		      (" TESTING--setUtilityClocksAsClock, Availibility clock REPLACED by Utility clock for %s \n",
		       ident2string (EM->getName (aSig)));
		    EM->setClock (aSig, utilityClk);
		    b = pk_true;
		  }
	      }
	  }
  return b;
}

void
DCG::markNodesFreeConditions ()
{
  ptlinsth l;
  ptinsth h;
  ptlinstances ls;
  ptINode n;
  ptexpcal s;

  // Unused
  require_clock_manager (CKM);

  l = CKM->getlhplaced ();
  while (ptlinsth_get_next (&l, &h))
    if (CKM->eqClock (CKM->getRepresentativeFormula (h), h))
      if (CKM->getOperator (h) == sub)
	DontOptimizeSignals (CKM->getSignalsOfCondition (h));
}


/**
   It adds at the begining/at the end (depending on the value of the   b_inner parameter) of lStatesVar the boolean state variables assigned to a clock (vclk).
*/
void
DCG::getDelayedSignalsAt (ptinsth vclk, boolean b_inner, ptlinstances * lStatesVar)
{
  ptgfd vgfd;
  ptlINode lnodes;
  ptINode aNode;
  ptexpcal vsig;

  require_managers (EM, NM, CKM);

  vgfd = CKM->getClockSubGraph (vclk);
  if (!ptgfd_satisfies_empty (vgfd))
    {
      lnodes = ptgfd_get_nodes (vgfd);
      while (ptlINode_get_elem (&lnodes, &aNode))
	if (NM->is_delay (aNode))
	  {
	    vsig = NM->get_leftOperand (aNode);
	    if (EM->IsBoolean (vsig))
	      {
		if (b_inner)
		  *lStatesVar = ptlinstances_push_front (*lStatesVar, vsig);
		else
		  *lStatesVar = ptlinstances_push_end (*lStatesVar, vsig);
	      }
	  }
    }
}


/**
   @return "true" if it exists a boolean state signal in the hierarchy with vclk as root, "false" otherwise.
   It selects (one of) the nearest signal of the root (vclk) of the hierarchy.
   b_inner used for the order on state variable, pk_true when the leafs have a hight priority.
*/
void
DCG::mkListOfStatesVariables (ptinsth vclk, boolean b_inner, ptlinstances * lStatesVar)
{
  ptlinsth lchild;
  ptinsth vChild;
  require_clock_manager (CKM);

  getDelayedSignalsAt (vclk, b_inner, lStatesVar);
  lchild = CKM->getChildren (vclk);
  while (ptlinsth_get_next (&lchild, &vChild))
    mkListOfStatesVariables (vChild, b_inner, lStatesVar);
}


/**
   @return the number assigned to the graph.
*/
integer
DCG::getGraphNumber ()
{
  return getnum ();
}

/**
   Suspend (when b is pk_true) or Resume (when b is pk_false) the setClocksName() action.
   @return the mode before the call.
 */
boolean
DCG::getSetSuspendClockNameAction (boolean b)
{
  boolean b1;
  b1 = bSuspendClockNameAction;
  bSuspendClockNameAction = b;
  return b1;
}

boolean DCG::getIthRoot(integer n, ptinsth *vr)
{
  ptlinsth lh;
  integer i;

  if (n<=0) return pk_false;
  require_clock_manager(CKM);
  lh = CKM->getlhroots();
  i =1;
  while (ptlinsth_get_next(&lh, vr))
    {
      if ( i==n ) return pk_true;
      i++;
    }
  return pk_false;
}


boolean  DCG::isHierarchyWithNodes(ptinsth vh)
{
  ptlinsth lh;
  ptinsth vhchild;

  require_clock_manager(CKM);
  if ( CKM->getNodesAssignedTo(vh) != NILLINODE) return pk_true;
  lh =  CKM -> getChildren (vh);
  while (ptlinsth_get_next(&lh, &vhchild))
    if (isHierarchyWithNodes(vhchild))  return pk_true;
  return pk_false;
}

boolean DCG::getNextFreeRoot(ptinsth vh, ptinsth *vr)
{
  ptlinsth lh;
  boolean found;

  require_clock_manager(CKM);
 
  lh = CKM ->getChildren (CKM->getTickClock ()); //  lh = CKM->getlhroots(); list of variables
  if (vh != NILINSTH)
    {
      while ( !found && ptlinsth_get_next(&lh, vr)) found = CKM->eqClock(*vr, vh);
      if (!found) return pk_false;
    }
  
  while (ptlinsth_get_next(&lh, vr))
    {
      if (isHierarchyWithNodes(*vr)) return pk_true;
      // if (CKM->getNodesAssignedTo(*vr) != NILLINODE) return pk_true;
    }
  return pk_false;
}

/**
   It adds a pragma (vpragma) to all the objects (clocks/nodes) of the hierarchy with vclk as root.
*/
void
DCG::addPragmaToAllObjectsOfHierarchy (ptinsth vclk, ptpragma vpragma)
{
  ptlinsth lchild;
  ptinsth vChild;
  ptlINode ln;
  ptINode aNode;
  
  require_clock_manager (CKM);
  require_node_manager (NM);
  
  CKM->addPragma(vclk, vpragma);
  ln = CKM->getNodesAssignedTo(vclk);
  while (ptlINode_get_elem (&ln, &aNode)) 
    {
      NM->addPragma(aNode, vpragma);
    }

  lchild = CKM->getChildren (vclk);
  while (ptlinsth_get_next (&lchild, &vChild))
    addPragmaToAllObjectsOfHierarchy (vChild, vpragma);
}


