/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

///  representation of the expressions on signals in the kernel tree representation.
typedef struct klExprSig
{
  ntrees vtreeBackAnnotation;	///< back annotation to the abstract tree produced by the parser
  noperators operat;		///< operator of the expression
  cloper opera;			///< classification
  union
  {
    ptklExprProc opposi;	///< Call
    ptlrepid opvari;		///< variabl
    ptrepstr opconst;		///< constant expression
    ptklExprSig op;		///< unary expression
    ptreptyp optyp;		///< reference to a type
    struct
    {
      ptklExprSig op21, op22;	///< Binary Expression (+,*,...) or temporal Binary Expression (default, when)
    } BinExp;
    struct
    {
      ptklExprSig op31, op32, op33;	///< Ternary expression or temporal Ternary expression
    } TerExp;

  } U;
}
tklExprSig;



///  representation of the expressions on processes in the kernel tree representation.
typedef struct klExprProc
{
  boolean assertion;		///< Only for composition and clock operators (^=, ^#, ^<,^>).
  ntrees vtreeBackAnnotation;	///< back annotation to the abstract tree produced by the parser
  ptklExprSig etiq;		///< label assigned to the expression
  ptklExprSig On;		///< Expression that controls the expression. (on E::(|...|)

  ptlrepid inList,		///< list of flows referenced in the expression
    outList,			///< list of flows defined in the expression
    param;			///< list of parameters referenced in the expression
  ptlrepid llabels,		///< list of labels referenced in the expression
    cst_typ_loc,		///< list of constants and types referenced in the expression
    sig_loc;			///<  list of locals flows (where.... end)
  noperators oper;		///< operator of the expression.
  union
  {
    ptklExprSig opltl;		///< list of synchronized  expressions (^=),  ^#, ^< expressions or ^> expressions
    ntrees vlanguage;		///< external model implemented in vlanguage language

    struct
    {
      ptklExprProc opmodel;	///< call
      ptklExprSig largs;	///< inputs signals
    } call;
    struct
    {
      ptlrepid nom;		///< name of a referencesd model.
      ptklExprSig lparam;	///< effective parameters 
    } modelRef;
    struct
    {
      ptklExprSig opaff1, opaff2;	///< equation, partial equation, defaultvalue equation
    } equation;

    struct
    {
      ptklExprProc opcol1, opcol2;	///< composition of processes
    } composition;
    struct			///<  Array of processes
    {
      ptklExprSig indice;	///< index
      ptklExprSig bsup;		///< upper bound of the iteration
      ptklExprProc lvarmodi;	///< Initial part of the array (with (|...|)
      ptklExprProc corpsdo;	///< Definition of the array
    } array;
    struct			///< case structure
    {
      ptklExprSig ident;	///< switch case expression
      ptlcase l_case;		///< list of case-statement  case c : (| ... |)
      ptklExprProc p_else;	///< otherwise part
    } acase;
    struct			///< dependances X,Y->Z, W when H
    {
      ptklExprSig vfrom,	///< left part of the -> operator
        vto,			///< right part of the -> operator
        vat;			///< when part of the dependence
    } dependence;

  } U;
}
tklExprProc;



/**
   Contructor of an expression of processes.\n
   @return a new expression of processes characterized by
   - an operator operfil (opnil, oplfeed, opmoti, opiter, opinst, oppcas, opdepd, opappe, opexte, opdefsig, opdefpart, opdefpartdef)
   - the list of the signals referenced in the expression but not defined in it (inputs)
   - the list of the parameters (lp)
   - the list of the signals defined in the expression (outputs)
   - the back annotation to the abstract tree model (vtree)
*/
static ptklExprProc
mkKernelExprProc (ptlrepid inputs, ptlrepid outputs, ptlrepid lp, ntrees vtree, noperators operfil)
{
  ptklExprProc pt;

  pt = (ptklExprProc) malloc ((unsigned) (sizeof *pt));
  pt->etiq = NILEXPRSIG;
  pt->On = NILEXPRSIG;
  klset_inout (pt, inputs, outputs);
  pt->param = insertlist (lp, NILLREPID);
  pt->llabels = NILLREPID;
  pt->cst_typ_loc = NILLREPID;
  pt->sig_loc = NILLREPID;
  pt->vtreeBackAnnotation = vtree;
  pt->oper = operfil;
  pt->assertion = pk_false;
  return pt;
}

/**
    Contructor of an expression of signals.\n
    @return a new expression characterized by its operator (ope) and its class (clop). 
    For tracability, a back annotation to the abstract tree (parsing) is kept (vt).
*/
static ptklExprSig
mkKernelExprSig (noperators ope, cloper clop, ntrees vt)
{
  ptklExprSig pt;
  pt = (ptklExprSig) malloc ((unsigned) (sizeof *pt));
  pt->opera = clop;
  pt->operat = ope;
  pt->vtreeBackAnnotation = vt;
  return pt;
}
