/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
   @return the expression on processes associated with an expression assumed to be an invocation of a process model.
*/
extern ptklExprProc
kl_getFlowExprModelProcCall (ptklExprSig op)
{
  return op->U.opposi;
}


/**
   For an assumed expression of composition (vfi), it returns 
   - the first element (pgcol) of the composition
   - the tail (pdcol) of the composition, 
   - and the back annotation (vtree) to the abstract tree.
*/
extern void
kl_getComposition (ptklExprProc vfi, ptklExprProc * pgcol, ptklExprProc * pdcol, ntrees * vtree)
{
  *pgcol = vfi->U.composition.opcol1;
  *pdcol = vfi->U.composition.opcol2;
  *vtree = vfi->vtreeBackAnnotation;
}

/**
   @return "true" if the expression (eproc) is a composition of an assertion, "false" otherwise.
*/
extern boolean
kl_isCompositionAssertion (ptklExprProc eproc)
{
  return eproc->assertion;
}

/**
   @return the list of the expressions associated with an assumed expression of synchronization (vfi).
   It returns also the back annotation (vtree) to the abstract tree.
*/
extern ptklExprSig
kl_getlExprClockOp (ptklExprProc vfi, ntrees * vtree)
{
  *vtree = vfi->vtreeBackAnnotation;
  return vfi->U.opltl;
}

/** 
    @return the operator of an expression on processes (see mkKernelExprProc()).
 */
extern noperators
kl_getExprProcType (ptklExprProc vfi)
{
  return vfi->oper;
}


/**
   For an expression of processes (vfi), it returns its inputs (li), its outputs (lo) and its parameters (lp).
*/
extern void
kl_getInsOutsParams (ptklExprProc vfi, ptlrepid * le, ptlrepid * lo, ptlrepid * lp)
{
  *le = vfi->inList;
  *lo = vfi->outList;
  *lp = vfi->param;
}

/**
   For an expression of processes (vfi), it returns its inputs.
*/
extern ptlrepid
kl_getInputs (ptklExprProc vfi)
{
  return vfi->inList;
}

/**
   For an expression of processes (vfi), it returns its inputs.
*/
extern ptlrepid
kl_getOutputs (ptklExprProc vfi)
{
  return vfi->outList;
}

/** 
    For an assumed array expression of processes (vfi), it returns 
    - the index (index)
    - the upper bound of the index (ub)
    - the initial part (withPart)
    - the body part (bodyPart)
    - and the back annotation (vtree) to the abstract tree.
*/
extern void
kl_getArray (ptklExprProc vfi, ptklExprSig * index, ptklExprSig * ub, ptklExprProc * withPart, ptklExprProc * bodyPart, ntrees * vtree)
{
  kl_getIterate (vfi, index, withPart, bodyPart, vtree);
  *ub = vfi->U.array.bsup;
}

/** 
    For an assumed iterate expression of processes (vfi), it returns 
    - the index (index)
    - the initial part (withPart)
    - the body part (bodyPart)
    - and the back annotation (vtree) to the abstract tree.
*/
extern void
kl_getIterate (ptklExprProc vfi, ptklExprSig * index, ptklExprProc * withPart, ptklExprProc * bodyPart, ntrees * vtree)
{
  *index = vfi->U.array.indice;
  *withPart = vfi->U.array.lvarmodi;
  *bodyPart = vfi->U.array.corpsdo;
  *vtree = vfi->vtreeBackAnnotation;
}



/**
   @return the invocation of a model associated with an assumed "proc call" expression of processes (vfi).
   Moreover, it returns
   - the ordered list of the effective parameters.
   - and the back annotation (vtree) to the abstract tree.
*/
extern ptklExprProc
kl_getModelProcCall (ptklExprProc vfi, ptklExprSig * listpar, ntrees * vtree)
{
  *listpar = vfi->U.call.largs;
  *vtree = vfi->vtreeBackAnnotation;
  return vfi->U.call.opmodel;
}

/**
   It returns for an expression of processes (vklExprProc):
   - its constants/types (const_typ_loc),
   - its local signals 
*/
extern void
kl_getLocals (ptklExprProc vklExprProc, ptlrepid * const_typ_loc, ptlrepid * sig_loc, ptlrepid * llabels)
{
  *const_typ_loc = vklExprProc->cst_typ_loc;
  *sig_loc = vklExprProc->sig_loc;
  *llabels = vklExprProc->llabels;
}


/**
   @return "true" it an expression on procecesses (vklExprProc) is labelled (lab), "false" otherwise
*/
extern boolean
kl_getLabelExpr (ptklExprProc vklExprProc, ptklExprSig * lab)
{
  (*lab) = vklExprProc->etiq;
  return ((*lab) != NILEXPRSIG);
}


/**
   @return "true" it an expression on procecesses (vklExprProc) is controlled by an expression vOn, "false" otherwise
*/
extern boolean
kl_getControlExpr (ptklExprProc vklExprProc, ptklExprSig * vOn)
{
  (*vOn) = vklExprProc->On;
  return ((*vOn) != NILEXPRSIG);
}



/**
  For an assumed invocation of a model of process (vfi), it returns
  - the model of  the process instanciated in vfi
  - the list of the effective parameters
  - and the back annotation (vtree) to the abstract tree.
*/
extern void
kl_getMacro (ptklExprProc vfi, ptlrepid * nomfi, ptklExprSig * listpar, ntrees * vtree)
{
  *nomfi = vfi->U.modelRef.nom;
  *listpar = vfi->U.modelRef.lparam;
  *vtree = vfi->vtreeBackAnnotation;
}

/**
   @return the effective parameters associated with an an assumed invocation of a model of process (vfi).
*/
extern ptklExprSig
kl_getEffParameterCall (ptklExprProc vfi)
{
  return vfi->U.modelRef.lparam;
}


/**
   For an assumed equation, partial or not, definition (vfi), it returns
   - the left part (X) of the definition (X:= E, or X::=E)
   - the expression (E) of the definition (X:= E, or X::=E)
   - and the back annotation (vtree) to the abstract tree.
*/
extern void
kl_getEquation (ptklExprProc vfi, ptklExprSig * pgaff, ptklExprSig * pdaff, ntrees * vtree)
{
  *pgaff = vfi->U.equation.opaff1;
  *pdaff = vfi->U.equation.opaff2;
  *vtree = vfi->vtreeBackAnnotation;
}


/**
   For an assumed dependence definition (vfi), it returns
   - the left part (X) of the definition (X -> Y when C)
   - the successor part (Y) of the definition  (X -> Y when C)
   - and the condition (C) associated with the dependence.
*/
extern void
kl_getDependence (ptklExprProc vklExprProc, ptklExprSig * vfrom, ptklExprSig * vto, ptklExprSig * vat)
{
  *vfrom = vklExprProc->U.dependence.vfrom;
  *vto = vklExprProc->U.dependence.vto;
  *vat = vklExprProc->U.dependence.vat;
}

/**
   Expression on signals.\n
   It returns the operand (op1) of an assumed unary expression on signals (pt).
*/
extern void
kl_getUnaryExpr (ptklExprSig pt, ptklExprSig * op1)
{
  *op1 = pt->U.op;
}

/**
   Expression on signals.\n
   It returns the operands (op1, op2) of an assumed binary expression on signals (pt).
*/
extern void
kl_getBinaryExpr (ptklExprSig pt, ptklExprSig * op1, ptklExprSig * op2)
{
  *op1 = pt->U.BinExp.op21;
  *op2 = pt->U.BinExp.op22;
}

/**
   Expression on signals.\n
   It returns the operands (op1, op2, op3) of an assumed ternary expression on signals (pt).
*/
extern void
kl_getTernaryExpr (ptklExprSig pt, ptklExprSig * op1, ptklExprSig * op2, ptklExprSig * op3)
{
  *op1 = pt->U.TerExp.op31;
  *op2 = pt->U.TerExp.op32;
  *op3 = pt->U.TerExp.op33;
}

/**
   Expression on signals.\n
   @return the type expression denoted by an assumede type expression on signals (pt).
*/
extern ptreptyp
kl_getTypeExpr (ptklExprSig pt)
{
  return (pt->U.optyp);
}

/**
   Expression on signals.\n
   @return the invocation of a model of process denoted by an expression on signals (op).
   The expression is assumed well formed.
*/
extern ptklExprProc
kl_getExpfilAppel (ptklExprSig op)
{
  return op->U.opposi->U.call.opmodel;
}


/**
   @return the representation of the model of process referenced in an expression on signals (op).
   The expression is assumed well formed.   
*/
extern ptlrepid
kl_getFlowExprModelProcCallname (ptklExprSig op)
{
  return op->U.opposi->U.call.opmodel->U.modelRef.nom;
}

/**
   @return  the class of an expression on  signals (pt).
 */
extern cloper
kl_getFlowExprClassType (ptklExprSig pt)
{
  return pt->opera;
}

/**
   @return the value of an assumed constant (op).
*/
extern ptrepstr
kl_getConstExpr (ptklExprSig op)
{
  return op->U.opconst;
}


/**
   @return the operator of the operand expression (pt).
*/
extern noperators
kl_getFlowExprOperator (ptklExprSig pt)
{
  return pt->operat;
}

/**
   @return the first element of pt and it sets the operand to the next element.
   pt is assumed to be a list of expression.
*/
extern ptklExprSig
kl_getNextExpr (ptklExprSig * pt)
{
  ptklExprSig res;
  res = (*pt)->U.BinExp.op21;
  *pt = (*pt)->U.BinExp.op22;
  return res;
}



/**
   @return the model prototype of a flow/constant/index/parameter denoted by op (assumed to be such an element).
*/
extern ptlrepid
kl_getIdentifierExpr (ptklExprSig op)
{
  return op->U.opvari;
}



/**
   For a case expression on processes (vklExprProc), it returns
   - the selector identifier (id),
   - the list of cases (l_case),
   - the "otherwise" part (p_else),
   - and the back annotation (vtree) to the abstract tree.
*/
extern void
kl_getCase (ptklExprProc vklExprProc, ptklExprSig * id, ptlcase * l_case, ptklExprProc * p_else, ntrees * vtree)
{
  *id = vklExprProc->U.acase.ident;
  *l_case = vklExprProc->U.acase.l_case;
  *p_else = vklExprProc->U.acase.p_else;
  *vtree = vklExprProc->vtreeBackAnnotation;
}


/**
   @return the back annotation (vtree) to the abstract tree associated with an expression on processes (vklExprProc).
   It is the link between two representations (abtract tree, kernel tree) used for annotations.
*/
extern ntrees
kl_getAbstractTree (ptklExprProc vklExprProc)
{
  return vklExprProc->vtreeBackAnnotation;
}

/**
   @return the language implementation associated with an assumed "external" body of a model of process.
*/
extern ntrees
kl_getLangExternal (ptklExprProc vklExprProc)
{
  return vklExprProc->U.vlanguage;	// external "LanguageName" (opexte("string"))
}

/**
   @return the size of a list of expressions (list).
*/
extern integer
kl_getSizeExpr (ptklExprSig list)
{
  integer nb;

  nb = 0;
  while (list != NILEXPRSIG)
    {
      nb++;
      list = list->U.BinExp.op22;
    }
  return nb;
}

/**
   @return the back annotation (vtree) to the abstract tree associated with an expression on signals (vklExprSig).
   It is the link between two representations (abtract tree, kernel tree) used for annotations.
*/
extern ntrees
kl_getTreeExprSig (ptklExprSig vklExprSig)
{
  // new 2007 december, tracability; --not used-- (LBesnard)
  return vklExprSig->vtreeBackAnnotation;
}
