/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** Constructor.
    @return a new representation of a predefined base type with vope as operator (opimplicit, opinteger, opreal, ...)
*/
extern ptreptyp
kl_mkBasicType (noperators vope)
{
  return kl_mkType (tbase, vope, NILEXPRSIG, NILLREPID, NILREPFIELD, NILLPOINTER);
}


/**
   @return a new type definition characterized with:
   - t : type of the type (extern, structure, bundle,...). According to t, the other parameters are not always defined.
   - vope : basic type  (optvect,...)
   - tab : dimension expression for array
   - nom_typ : name of the referenced type
   - field:  set of fields for a structure/bundle type.
   - lenum : set of values for an enumerated type.
*/
extern ptreptyp
kl_mkType (typ_type t, noperators vope, ptklExprSig tab, ptlrepid nom_typ, ptrepfield fields, ptlPointer lenum)
{
  ptreptyp vtyp;

  vtyp = (struct reptyp *) malloc ((unsigned) (sizeof *vtyp));
  SET_NILTREE (vtyp->vdecl);
  vtyp->t = t;
  vtyp->trep = tab;
  switch (t)
    {
    case tbase:
      vtyp->U.ope = vope;
      break;
    case tstruct:
    case tbund:
      vtyp->U.StructBundle.field_def = fields;
      vtyp->U.StructBundle.vspec = NILEXPFIL;
      vtyp->U.StructBundle.lfields = NILLREPID;
      break;
    case textern:
    case tnom:
      vtyp->U.nom_typ_ref = nom_typ;
      break;
    case tenum:
      vtyp->U.lenum = lenum;
    }
  return vtyp;
}


/**
   @return a new structure associated with a field of a structure/bundle type. 
   It is characterized by:
   - nom  : name of the field
   - vtyp : type of the field
   - field_prec : the previous field of the strucutre/bundle
*/
extern ptrepfield
kl_mkFieldType (ptrepstr nom, ptreptyp vtyp, ptrepfield field_prec)
{
  ptrepfield vfield;

  vfield = (struct repfield *) malloc ((unsigned) (sizeof *vfield));
  vfield->nom_field = nom;
  vfield->deftyp = vtyp;
  vfield->suivant = NILREPFIELD;
  if (field_prec != NILREPFIELD)
    field_prec->suivant = vfield;
  return (vfield);
}


/**
   @return the last element of a list of fields (lf)
*/
extern ptrepfield
kl_getLastField (ptrepfield lf)
{
  if (lf == NILREPFIELD)
    return NILREPFIELD;
  while (lf->suivant != NILREPFIELD)
    lf = lf->suivant;
  return lf;
}

/** 
   For a type (vtyp), it returns all the elements of its definition
   - t : type of the type
   - tab : dimension expression (array type)
   and following t:
   - ope : basic type 
   - nom_typ : name of the type
   - fields: set of fields of a strucutre/bundle type
   - lenum list of the values of a enumerated type
*/
extern void
kl_getFieldType (ptreptyp vtyp, typ_type * t, noperators * vope, ptklExprSig * tab, ptlrepid * nom_typ, ptrepfield * fields, ptlPointer * lenum)
{
  *t = vtyp->t;
  *tab = vtyp->trep;
  if (*t == tbase)
    *vope = vtyp->U.ope;
  else if ((*t == tbund) || (*t == tstruct))
    *fields = vtyp->U.StructBundle.field_def;
  else if (*t == tenum)
    *lenum = vtyp->U.lenum;
  else
    *nom_typ = vtyp->U.nom_typ_ref;	/* tnom, textern */
}


/**
   @return the type of a type (vtyp) (i.e. external, structure, predefined, ...)
*/
extern typ_type
kl_getTypeClass (ptreptyp vtyp)
{
  return (vtyp->t);
}


/**
   @return the basic type (Signal operator) associated with vtytp.
*/
extern noperators
kl_getBasicType (ptreptyp vtyp)
{
  typ_type ty;

  ty = vtyp->t;
  if ((ty == tnom) || (ty == textern))
    return (opident);
  if (ty == tstruct)
    return (opstru);
  return (vtyp->U.ope);
}

/**
   @return the dimension expression of a type (vtyp).
*/
extern ptklExprSig
kl_getDimType (ptreptyp vtyp)
{
  return (vtyp->trep);
}

/** 
    @return 
    - "true" if a field (vfield) is defined, then it returns the name of the field (nom) and its type (vtyp), and
    vfields is set to the next element.
*/
extern boolean
kl_get_elemNext_Field (ptrepfield * vfield, ptrepstr * nom, ptreptyp * vtyp)
{
  if (*vfield == NILREPFIELD)
    return pk_false;

  (*nom) = (*vfield)->nom_field;
  (*vtyp) = (*vfield)->deftyp;
  (*vfield) = (*vfield)->suivant;
  return pk_true;
}

/**
   A bundle type (vtyp) is characterized by a synchronisation expression (vsp).
 */
extern void
kl_setSpecBundleType (ptreptyp vtyp, ptklExprProc vsp, ptlrepid lrepidfields)
{
  if (vtyp->t == tbund)
    {
      vtyp->U.StructBundle.vspec = vsp;
      vtyp->U.StructBundle.lfields = lrepidfields;
    }
}

/**
   @return "true" if a type (vtyp) is a bundle type, then the synchronisation expression (vsp) is returned.
 */
extern boolean
kl_getSpecBundleType (ptreptyp vtyp, ptklExprProc * vsp)
{
  if (vtyp->t != tbund)
    return pk_false;
  *vsp = vtyp->U.StructBundle.vspec;
  return pk_true;
}

/**
   @return "true" if a type (vtyp) is a bundle type, then the  list (lrepidfields) of the identifiers of the fields are returned.
 */
extern boolean
kl_getRepidFields (ptreptyp vtyp, ptlrepid * lrepidfields)
{
  if (vtyp->t != tbund)
    return pk_false;
  *lrepidfields = vtyp->U.StructBundle.lfields;
  return pk_true;
}

extern void
kl_setTypeTreeDecl (ptreptyp vtyp, ntrees vd)
{
  vtyp->vdecl = vd;
}

extern ntrees
kl_getTypeTreeDecl (ptreptyp vtyp)
{
  return vtyp->vdecl;
}

/**
   it assigns to vreptyp1 the values of vreptyp2
*/
extern void
kl_typereplace (ptreptyp vreptyp1, ptreptyp vreptyp2)
{
  vreptyp1->t = vreptyp2->t;
  vreptyp1->trep = vreptyp2->trep;
  switch (vreptyp2->t)
    {
    case tbase:
      vreptyp1->U.ope = vreptyp2->U.ope;
      break;
    case tstruct:
    case tbund:
      vreptyp1->U.StructBundle.field_def = vreptyp2->U.StructBundle.field_def;
      vreptyp1->U.StructBundle.vspec = vreptyp2->U.StructBundle.vspec;
      vreptyp1->U.StructBundle.lfields = vreptyp2->U.StructBundle.lfields;
      break;
    case textern:
    case tnom:
      vreptyp1->U.nom_typ_ref = vreptyp2->U.nom_typ_ref;
      break;
    case tenum:
      vreptyp1->U.lenum = vreptyp2->U.lenum;
    }

}
