/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** 
    @return "true" if the Signal Format specified in the file parsin is syntacticly correct.
    @param parsin (in) : name of the input file to analyze.
    @param mpp (out) : abstract tree representation of the analyzed file.
 */
extern boolean
sysFormatparser (char *parsin, ntrees * mpp)
{
  return ntrees_formatLoad (parsin, mpp);
}

/** 
    @return "true" if the Signal text specified in the file parsin
    and with  the waited phylum phylum is syntacticly correct.
    @param phylum (in) : phylum (without_phylum if no phylum, see phylaSignal.h file)
    @param parsin (in) : name of the input file to analyze.
    @param mpp (out) : abstract tree representation of the analyzed file.
 */
extern boolean
sysparser (integer phylum, char *parsin, ntrees * mpp)
{
  return sysparserLoc (phylum, parsin, mpp, pk_true);
}

/** 
    Similar to sysparser for the GUI. But, the errors are not printed on the stderr.
    The errors should be printed by the caller.
 */
extern boolean
sysparserGUI (integer phylum, char *parsin, ntrees * mpp)
{
  return sysparserLoc (phylum, parsin, mpp, pk_false);
}


/** 
    Parsing a Signal process or a Signal module in Intermediate Signal Format
    @return "true" if the Signal program specified in the file parsin is syntacticly correct.
    @param parsin (in) : name of the input file to analyze.
    @param mpp (out) : abstract tree representation of the analyzed file.
    @param compContinue (out) : set to "true" if the returned tree (mpp) is a model or a modul, "false" otherwise.
 */
extern boolean
SignalFormatParseSrcfile (char *parsin, ntrees * mpp, boolean * compContinue)
{
  SET_NILTREE (*mpp);
  pK_messenger->globalMessageString (subinfo_PARSING, parsin);
  if (!exist_file (string2repstr (parsin)))
    {
      *compContinue = pk_false;
      return pk_false;
    }
  if (sysFormatparser (parsin, mpp))
    {
      if (IS_NILTREE (*mpp))
	*compContinue = pk_false;
      else
	*compContinue = LL_isCompilableOp (noper (*mpp));
      return pk_true;
    }
  pK_messenger->globalMessageString (error_PARSING, parsin);
  return pk_false;
}

/** 
    Parsing a Signal program or a Signal module.
    @return "true" if the Signal program specified in the file parsin is syntacticly correct.
    @param parsin (in) : name of the input file to analyze.
    @param mpp (out) : abstract tree representation of the analyzed file.
    @param compContinue (out) : set to "true" if the returned tree (mpp) is a model or a modul, "false" otherwise.
 */
extern boolean
SIGNALParseSrcfile (char *parsin, ntrees * mpp, boolean * compContinue)
{
  noperators vop;
  SET_NILTREE (*mpp);

  pK_messenger->globalMessageString (subinfo_PARSING, parsin);
  if (!exist_file (string2repstr (parsin)))
    {
      *compContinue = pk_false;
      return pk_false;
    }
  if (sysparser (without_phylum, parsin, mpp))
    {
      if (IS_NILTREE (*mpp))
	*compContinue = pk_false;
      else
	{
	  vop = noper (*mpp);
	  *compContinue = LL_isCompilableOp (vop);
	  if (vop == oplmodu)
	    *mpp = nchild (1, *mpp);	// Added March 2015 LBesnard. One component in a file (restriction).
	}
      return pk_true;
    }
  pK_messenger->globalMessageString (error_PARSING, parsin);
  return pk_false;
}

/** 
    Parsing Signal parameters.
    @return "true" if the file paramfile contains Signal parameters.
    @param paramfile (in) : name of the input file to analyze.
    @param mparam (out) : abstract tree representation of the analyzed file.
 */
extern boolean
SIGNALParseParamfile (char *paramfile, ntrees * mparam)
{
  SET_NILTREE (*mparam);
  if (strlen (paramfile) == 0)
    return pk_true;
  pK_messenger->globalMessageString (parameterPARSING_subAnnounce, paramfile);
  if (!exist_file (string2repstr (paramfile)))
    return pk_false;
  sysUpdateParametersFile (paramfile);
  if (sysparser (ph_ol_e_para, paramfile, mparam))
    if (LL_isActualParametersOp (noper (*mparam)))
      return pk_true;
  pK_messenger->globalMessageString (error_parameterPARSING, paramfile);
  return pk_false;
}
