/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


/** @return "true" if a pragma (repstring) exists in a list of pragmas (lprag), "false" otherwise. 
 */
extern boolean
existPragmaKeyword (ptlpragmas lprag, ptrepstr repString)
{
  ptpragma aPragma;

  while (ptlpragmas_get_next (&lprag, &aPragma))
    if (getPragmaName (aPragma) == repString)
      return pk_true;
  return pk_false;
}

/**
   @param vpragmas (in) a list of pragmas.
   @param aPragma (in) a pragma.
   @return "true" if it exists a pragma in vpragmas with the same name than aPragma, without assigned objects, and with the same statement, "false" otherwise.
*/
extern boolean
existPragmaLike (ptlpragmas vpragmas, ptpragma aPragma)
{
  ptrepstr name;
  ptObjPragma obj;
  ntrees stmts;
  ptpragma vp;

  getPragma (aPragma, &name, &obj, &stmts);
  if (obj != NILOBJPRAGMA)
    return pk_false;
  if (getPragmaStrKeyword (vpragmas, name, &vp))
    return STree_isTreeEq (stmts, getPragmaStatement (vp));
  return pk_false;
}

/** 
    @return "true" if a pragma (aPragma) is named pragmaKeyword, "false" otherwise.
*/
extern boolean
isPragmaKeyword (ptpragma aPragma, int pragmaKeyword)
{
  ptrepstr repString;
  repString = getPredefinedPragma (pragmaKeyword);
  return (getPragmaName (aPragma) == repString);
}

/** 
    @return a new pragma "name" "obj" "stmts".
*/
extern ptpragma
newpragma (ptrepstr name, ptObjPragma obj, ntrees stmts)
{
  ptpragma prag;
  prag = new pragma ();
  prag->name = name;
  prag->objects = obj;
  prag->stmts = stmts;
  return prag;
}


/** 
    @return a new predefined pragma where the name is given by the i-th element of the predefined pragma table.
    \see getPredefinedPragma().
*/
extern ptpragma
newPredefinedPragma (int i, ptObjPragma obj, ntrees stmts)
{
  return newpragma (getPredefinedPragma (i), obj, stmts);
}

/** It associates the objects (vobj) with a pragma (aPragma). */
extern void
setPragmaObjects (ptpragma aPragma, ptObjPragma vobj)
{
  aPragma->objects = vobj;
}

/** It assigns the statement (stmts) associated with a pragma (aPragma). */
extern void
setPragmaStatement (ptpragma aPragma, ntrees stmts)
{
  aPragma->stmts = stmts;
}


/** @return the name of a pragma (aPragma). */
extern ptrepstr
getPragmaName (ptpragma aPragma)
{
  return aPragma->name;
}

/** @return the objects associated with a pragma (aPragma). */
extern ptObjPragma
getPragmaObjects (ptpragma aPragma)
{
  return aPragma->objects;
}

/** @return the statement associated with a pragma (aPragma). */
extern ntrees
getPragmaStatement (ptpragma aPragma)
{
  return aPragma->stmts;
}

/** 
    @param aPragma (in) a pragma
    @param name (out) name of the pragma;
    @param obj (out)  set of objects assigned to the pragma;
    @param stmts (out) the text (set of strings) assigned to the pragma.
*/
extern void
getPragma (ptpragma aPragma, ptrepstr * name, ptObjPragma * obj, ntrees * stmts)
{
  (*name) = aPragma->name;
  (*obj) = aPragma->objects;
  (*stmts) = aPragma->stmts;
}


/** @return  "true" if a ragma (pragmakeyword) is in the list of pragmas (lprag), "false" otherwise.
    The output parameter (valPragma) is the founded pragma.
    \see{getPragma}
*/
extern boolean
getPragmaKeyword (ptlpragmas lprag, int pragmaKeyword, ptpragma * valPragma)
{
  return getPragmaStrKeyword (lprag, getPredefinedPragma (pragmaKeyword), valPragma);
}

/** @return  "true" if the pragma (repString) is in the list of pragmas (lprag), "false" otherwise.
    The output parameter (valPragma) is the founded pragma.
   \see{getPragma, getPragmaKeyword}
*/
extern boolean
getPragmaStrKeyword (ptlpragmas lprag, ptrepstr repString, ptpragma * valPragma)
{
  boolean found;
  pointer vpt;
  ptpragma aPragma;

  found = pk_false;
  while (!found && ptlPointer_get_elemNext (&lprag, &vpt))
    {
      aPragma = (ptpragma) vpt;
      if (getPragmaName (aPragma) == repString)
	{
	  found = pk_true;
	  (*valPragma) = aPragma;
	}
    };
  return found;
}


/** @return "true" is the length of a pragma value (a string) is equal to i, "false" otherwise. */
extern boolean
endOfPragmaValue (pragmaValue value, int i)
{
  return (((int) strlen (value)) == i);
}


/**
   @return "true" if the list of pragmas (lprag) is not empty, "false" otherwise.
   @param lprag (in/out) list of pragmas, set to the next element if it exists.
   @param vpt (out) the element of the list, if it exists
*/
extern boolean
ptlpragmas_get_next (ptlpragmas * lprag, ptpragma * vpt)
{
  pointer vpx;
  if (!ptlPointer_get_elemNext (lprag, &vpx))
    return pk_false;
  (*vpt) = (ptpragma) vpx;
  return pk_true;
}

/**
   @return the first string of the value of a pragma (valPragma)
*/
extern ptrepstr
strFirstStmtsValue (ptpragma valPragma)
{
  return getPragmaTerm (getPragmaStatement (valPragma), 1);
}

/**
   @return if it exists, the value of a pragma named codeop of a list of pragmas (tableOpCode), niltree otherwise.
*/
extern ntrees
getEntryCodeOp (ptlpragmas tableOpCode, ptrepstr codeop)
{
  ptpragma valPragma;

  // Could be optimized (using hash code technic for example)
  if (!getPragmaStrKeyword (tableOpCode, codeop, &valPragma))
    return NILTREE;
  return getPragmaStatement (valPragma);
}

/**
   It translates a list of pragmas represented as an abstract tree into the internal representation (lp) of pragma.
*/
extern boolean
ntreesPragmas2pragmas (ntrees p, ptlpragmas * lp)
{
  int i, n;
  ptrepstr name;
  ntrees stmts;
  ptlpragmas l_prag;
  ntrees p_obj, aPragma;
  noperators op;

  l_prag = NILLPRAGMAS;
  if (ISNOT_NILTREE (p))
    {
      op = noper (p);
      if (op != oplprag)
	return pk_false;
      n = ntreearity (p);
      for (i = 1; i <= n; i++)
	{
	  aPragma = nchild (i, p);
	  STree_getPragmaValues (aPragma, &name, &p_obj, &stmts);
	  if (ptlPointer_insert_elem (&l_prag, newpragma (name, NILOBJPRAGMA, stmts)));
	}
    }
  *lp = l_prag;
  return pk_true;
}


extern boolean
ptlpragmas_remove_elem (ptlpragmas * lprag, ptpragma elem)
{
  return ptlPointer_remove_elem ((ptlPointer *) lprag, (pointer) elem);
}


/**
   @return "true" if the predefined pragma, given by its index (see SignalEnvPramas.h file), is in a list of pragma (lprag), "false" otherwise.
*/
extern boolean
existPredefinedPragma (ptlpragmas lprag, int i)
{
  return existPragmaKeyword (lprag, getPredefinedPragma (i));
}

/**
   @return "true" is a name (vn) is an element of the statement list of a pragma (p),
   "false" otherwise.
*/
boolean
isNameInPragmaStatementList (ptrepstr vn, ptpragma p)
{
  ntrees stmts;
  ptrepstr EnonceValue;
  int i, n;

  if (p == NILPRAGMA)
    return pk_false;
  stmts = getPragmaStatement (p);
  if (IS_NILTREE (stmts))
    return pk_false;
  n = ntreearity (stmts);
  for (i = 1; i <= n; i++)
    {
      EnonceValue = ExtractString (nstringof (nchild (i, stmts)));
      if (streq (EnonceValue, vn))
	return pk_true;
    }
  return pk_false;
}
