/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor.
   Such an object offers some functionalities (unparse(), print(), fileUnparse()) for printing a Signal abstract tree
   into a text.
   See initPrinter() for the initialisations.
   By default the expression "when C" is printed [:c], "when not C" is printed [/:c]. See setPrintOptionUnaryWhen().
*/
  SignalPrinter::SignalPrinter ()
{
  notannounced = pk_true;
  stdMode = pk_true;
  shorten_clocks = pk_true;
}


/**
   Destructor.
*/
SignalPrinter::~SignalPrinter ()
{
}

/**
  Printing a tree in a textual form with depth-printing.
  @param aTree (in) the tree to print.
  @param holoph (in) printing depth; when it is null, the tree is not printed it is replaced by a holophraste.
*/
_VIRTUAL_ void
SignalPrinter::unparse (ntrees aTree, int holoph)
{
  initPrinter ();
  unparse_tree (aTree, holoph);
}

/**
   Similar to unparse() method. The text is not terminated by a line feed.
   Printing a tree in a textual form. 
   @param aTree (in) the tree to print.
*/
_VIRTUAL_ void
SignalPrinter::unparse1 (ntrees aTree)
{
  // Added for printing SSME/SME textual form (lbesnard).
  initPrinter ();
  unparse_tree1 (aTree, 9999);
}

/**
  Printing a tree in a textual form with depth-printing.
  It realises the travel of the operand tree, calls indentation functions, prints the comment attribute.

  @param p (in) the tree to print.
  @param cbuf (in) printing buffer.
  @param name (in) prefix of the output file name. 
         It will be completed by a prefix given by the output directory (user option)
	 and completed by the signal predefined suffix.
  @param holoph (in) printing depth. When it is null, the tree is not printed, it is replaced by a holophraste.
  @param outputMode (in) : output mode (file, window, ...).
  @param effName (out) : the effective name of the file (file mode) or name.
  @return true, if the text has been printed, false otherwise (permissions problem, a keyword table is missing,...).
*/
_VIRTUAL_ boolean
SignalPrinter::fileUnparse (ntrees p, ptrepbuf cbuf, ptrepstr name, int holoph, printModeType outputMode, ptrepstr * filePathName)
{
  if (!initPrinter ())
    return pk_false;
  if (outputMode == ofile)
    {
      if (!this_intePrinter->open_bufferMakePathName (cbuf, name, "SIG", filePathName))
	return pk_false;
    }
  else
    {
      if (!this_intePrinter->open_bufferDevice (cbuf, outputMode))
	return pk_false;
    }
  unparse_tree (p, holoph);
  intpriclose ();
  return pk_true;
}

/**
  Printing a tree in a textual form with depth-printing.
  It realises the travel of the operand tree, calls indentation functions, prints the comment attribute.

  @param p (in) the tree to print.
  @param cbuf (in) printing buffer.
  @param name (in) prefix of the output file name. 
         It will be completed by a prefix given by the output directory (user option)
	 and completed by the signal predefined suffix.
  @param holoph (in) printing depth. When it is null, the tree is not printed, it is replaced by a holophraste.
  @param outputMode (in) : output mode (file, window, ...).
  @return true, if the text has been printed, false otherwise (permissions problem, a keyword table is missing,...).
*/
_VIRTUAL_ boolean
SignalPrinter::fileUnparse (ntrees p, ptrepbuf cbuf, ptrepstr name, int holoph, printModeType outputMode)
{
  ptrepstr filePathName;
  return fileUnparse (p, cbuf, name, holoph, outputMode, &filePathName);
}

/**
  Printing a tree in a textual form into a file.
  @param p (in) the tree to print.
  @param cbuf (in) printing buffer.
  @param name (in) output file name.
  @return true, if the text has been printed, false otherwise (permission problems,...) 
*/
_VIRTUAL_ boolean
SignalPrinter::fileUnparse (ntrees p, ptrepbuf cbuf, ptrepstr fileNamePath)
{
  if (fileNamePath == nilstr)
    return pk_false;
  initPrinter ();
  if (!inteprint_open_buffer (cbuf, fileNamePath))
    return pk_false;
  unparse_tree (p, 9999);
  intpriclose ();
  return pk_true;
}

/**
   It sets the printer to "textual" (resp. "graphical") mode when not graphiContext (resp. when graphiContext)
   @return the value of the printer mode before the call to this method, with 
   - "true" for a graphical mode (the comments are not printed).
   - "false" otherwise, this is the textual mode and the comments are printed.
*/
_VIRTUAL_ boolean
SignalPrinter::setPrinterContext (boolean graphiContext)
{
  initPrinter ();
  return this_intePrinter->setPrinterContext (graphiContext);
}

/** Signal printing finalizing. */
_VIRTUAL_ void
SignalPrinter::end ()
{
  this_intePrinter->intpriend ();
}


void
SignalPrinter::reset ()
{
  notannounced = pk_true;
}

/**
   When b is "true", the unary when is printed when C.
   Otherwise the shorten forms [:c] for when C, and  [/:c] for when not C, are used.
*/
void
SignalPrinter::setPrintOptionUnaryWhen (boolean b)
{
  shorten_clocks = !b;
}
