/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** Constructor.
   @param vt (in) value of a pragma (list of "strings"). It is the text to analyze. 
   See \ref StrScanner.h file for more details.
   @param b (in)  must be "true" to skip space during scanning.
*/
  StrScanner::StrScanner (ntrees vt, boolean b)
{
  initKwTable ();
  StrScannerIni (vt, ntreearity (vt), ExtractString (getPragmaTerm (vt, 1)), b);
}

/** Constructor.
    @return an uninitialized scanner. The method StrScannerIni() may be used for this operation.
*/
StrScanner::StrScanner ()
{
  initKwTable ();
}


/**
   Scanning the statements of a pragma. For a string, use StrScannerStr().
   
   @param vt (in) it is the representation of a pragma. The value of a pragma is a list (nbl parameter) of strings.
   @param b (in)  must be "true" to skip space during scanning.

   Assumed : vt is a list of strings.
*/
void
StrScanner::StrScannerPragma (ntrees vt, boolean b)
{
  StrScannerIni (vt, ntreearity (vt), ExtractString (getPragmaTerm (vt, 1)), b);
}

/**
   Scanning a string.

   @param vstr (in) the string to scan.
   @param b (in) the scanner will skip space during scanning when this parameter is "true".
*/
void
StrScanner::StrScannerStr (ptrepstr vstr, boolean b)
{
  StrScannerIni (NILTREE, 1, vstr, b);
}


/**
   Destructor.
*/
StrScanner::~StrScanner ()
{
}

/**
   It sets the value of the spacing mode to b. If b is "true", space will be skipped during scanning.
*/
boolean
StrScanner::setMode (boolean b)
{
  boolean x;
  x = modeSkipSpace;
  modeSkipSpace = b;
  return x;
}

/**
   It scans from the current position of the text to analyse until a separator, taking into account the "escape separator".
   The returned token is also stored in the current object, to get it, use the function \ref getCurrentToken().
   @return a token defined in StrScanner.h
   @return tk_error in case of memory allocation failure
*/
ScanPragmaToken
StrScanner::getToken ()
{
  char c;
  boolean sepFound, stop;
  ptreg strRegister;

  currentToken = tk_error;	// Default
  sepFound = pk_false;
  stop = pk_false;
  if (strregnew (&strRegister))
    {
      while ((!stop) && nextChar (&c))
	{
	  switch (c)
	    {
	    case escapeSep:
	      if (nextChar (&c))
		strregconccha (strRegister, c);
	      else
		stop = pk_true;
	      break;
	    case SEPARATOR:
	      sepFound = pk_true;
	      stop = pk_true;
	      break;
	    case ':':
	      if (nextCharGlue (&c))
		if (c == '=')
		  currentToken = tk_2pe;
	      stop = pk_true;
	      break;
	    case ',':
	      currentToken = tk_virg;
	      stop = pk_true;
	      break;
	    case '(':
	      currentToken = tk_po;
	      stop = pk_true;
	      break;
	    case ')':
	      currentToken = tk_pf;
	      stop = pk_true;
	      break;
	    case '{':
	      currentToken = tk_ao;
	      stop = pk_true;
	      break;
	    case '}':
	      currentToken = tk_af;
	      stop = pk_true;
	      break;

	    default:		/* Appendthe read character c at the end of the string register strRegister */
	      strregconccha (strRegister, c);
	    }
	}
      if (nregGet_usedSize (strRegister) != 0)
	{
	  unput (1);
	  if (!strregstore (strRegister, &this_currentIdent))
	    {
	      strregfree (strRegister);
	      return tk_error;
	    }
	  currentToken = tk_ident;
	}
      else if (sepFound)
	{
	  if (nextCharGlue (&c))
	    {
	      unput (1);
	      if (c == SEPARATOR)
		currentToken = tk_adr;
	      else
		currentToken = getTokenAfterSEPARATOR ();
	    }
	}
      else if (!stop)
	currentToken = tk_end;
      strregfree (strRegister);
      return currentToken;
    }
  /* Memory allocation error */
  return tk_error;
}

/** It scans the analyzed text from the current position until the next separator (c).
    @return "true" if such a text exists (vstr) , "false" otherwise.
    @return false in case of memory allocation failure
*/
boolean
StrScanner::getStrUntilSeparator (char c, ptrepstr * vstr)
{
  boolean stop, omode, b;
  ptreg strRegister;
  char cc, c1;

  omode = setMode (pk_false);
  if (strregnew (&strRegister))
    {
      stop = pk_false;
      while ((!stop) && nextChar (&cc))
	if (c != cc)
	  if (cc != escapeSep)
	    strregconccha (strRegister, cc);
	  else
	    {
	      c1 = cc;
	      if (nextChar (&cc))
		{
		  if (c == cc)
		    strregconccha (strRegister, cc);
		  else
		    {
		      strregconccha (strRegister, c1);
		      strregconccha (strRegister, cc);
		    }
		}
	      else
		{
		  stop = pk_true;
		  strregconccha (strRegister, c1);
		}
	    }
	else
	  {
	    if (nextChar (&cc))
	      {
		if (c == cc)
		  strregconccha (strRegister, cc);
		else
		  unput (1);
		unput (1);
	      }
	    stop = pk_true;
	  }

      if (nregGet_usedSize (strRegister) != 0)
	{
	  b = strregstore (strRegister, vstr);
	}
      else
	b = pk_false;
      strregfree (strRegister);
      setMode (omode);
      return b;
    }
  /* Memory allocation error */
  return pk_false;
}

/** 
    It skips the analyzed text until a character (cc).
    @return "true" if operation succeeds, "false" otherwise (char not found).
 */
boolean
StrScanner::Goto (char cc)
{
  char c;
  while (nextChar (&c))
    if (c == cc)
      {
	return pk_true;
      }
  return pk_false;
}

/** 
    @return "true" if it exists a next char from the current position in the analyzed text (cr), "false" otherwise.
    It takes into account the "spacing mode".
 */
boolean
StrScanner::nextChar (char *cr)
{
  char c;
  if (modeSkipSpace)
    {
      do
	if (!nextCharGlue (&c))
	  return pk_false;
      while (c == ' ');
    }
  else if (!nextCharGlue (&c))
    return pk_false;
  *cr = c;
  return pk_true;
}

/**
   It unputs of n (n>0) characters in the analyse of the current text.
*/
void
StrScanner::unput (int n)
{
  int i;
  // A completer (retour ligne precedente)
  if (n > 0)
    for (i = 1; i <= n; i++)
      indexCurStr--;
}

/** @returns "true" if the text from the current position until the next separator is an identifier (vstr), "false" otherwise.
 */
boolean
StrScanner::getIdent (ptrepstr * vstr)
{
  if (getToken () != tk_ident)
    return pk_false;
  *vstr = this_currentIdent;
  return pk_true;
}


/**
   @return the last scanned identifier.
 */
ptrepstr
StrScanner::getIdent ()
{
  return this_currentIdent;
}

/** @returns "true" if the text from the current position until the next separator is a long int (i), "false" otherwise.
 */
boolean
StrScanner::getInt (int *i)
{
  ScanPragmaToken tk;

  tk = getIntToken ();
  if (tk != tk_ident)
    {
      if (tk == tk_error)
	currentToken = tk_virg;	// new new
      return pk_false;
    }
  *i = (int) (repstr2integer (this_currentIdent));
  return pk_true;
}

/**
   @return the last scanned token.
 */
ScanPragmaToken
StrScanner::getCurrentToken ()
{
  return currentToken;
}
