/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.
   
   Contact: polychronycontact@listes.irisa.fr 
   
   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).
   
   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.
   
   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.
   
   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
   http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


/**
   if 
   - i = 0 it replaces the tree p1 by p2 
   - i !=0 it replaces the i-th son of p1 by p2
*/
extern void
STree_replaceSon (ntrees p1, ntrees p2, int i)
{
  ntrees pere;

  if (i != 0)
    nchangchild (p1, p2, i);
  else
    {
      pere = nfather (p1);
      i = nposit (p1);
      if (!STree_isList (pere))
	{
	  if (i != 0)
	    nchangchild (pere, p2, i);
	}
      else
	{
	  ntreesDestroy (p1);
	  if (i != 0)
	    ninsertright (i - 1, p2, pere);
	}
    }
}

/**
   @return an empty tree
*/
extern ntrees
STree_mkEmptyTree (void)
{
  return LL_make_nil_atom ();
}

/**
   @return the calling process expression to name without parameters.
*/
extern ntrees
STree_mkModelCall (ptrepstr name)
{
  return LL_make_inst_tree (LL_make_nom_atom (name), STree_mkEmptyTree ());
}

/**
   @return an empty tree
*/
extern ntrees
STree_mkCstTree (integer n)
{
  if (n >= 0)
    return LL_make_ent_atom (integer2repstr (n));
  return LL_make_nega_tree (LL_make_ent_atom (integer2repstr (-n)));
}


/**
   @return the skeleton of an interface with 
   - empty lists for parameters, inputs, outputs
   - empty tree for the specification.
   \bsg <interface> ::= <o_suite_parametres_formels> <entrees_sorties> <graphe_externe>.\esg
*/
extern ntrees
STree_mkSkeletonInterface (void)
{
  ntrees s1, s2, p_inte;

  s1 = LL_make_ldecs_emptyList ();	// parameters
  s2 = LL_make_entsor_tree (LL_make_ldecs_emptyList (), LL_make_ldecs_emptyList ());	// inputs, outputs
  p_inte = nmaketernary (opinte, s1, s2, STree_mkEmptyTree ());	// last parameter: spec 
  return (p_inte);
}

/**
   @return an empty body tree (empty list of pragma, empty body).
   \bsg <description> ::= <o_l_pragmas> <o_corps>. \esg
*/
extern ntrees
STree_mkEmptyBody (void)
{
  return (LL_make_desc_tree (STree_mkEmptyTree (), STree_mkEmptyConfinedExpr ()));
}

/**
   @return the model of a process, node, action or function with 
   - p_nom  its name
   - empty interface
   - empty description (expression + local declarations)
*/
extern ntrees
STree_mkEmptyModel (int oper, ntrees p_nom)
{
  return nmaketernary (oper, p_nom, STree_mkSkeletonInterface (), STree_mkEmptyBody ());
}

/**
   @return the model of a process with 
   - p_nom  its name
   - empty interface
   - empty description (expression + local declarations)
*/
extern ntrees
STree_mkEmptyProcess (ntrees p_nom)
{
  return STree_mkEmptyModel (opprocess, p_nom);
}

/**
   @return the model of a node with 
   - p_nom  its name
   - empty interface
   - empty description (expression + local declarations)
*/
extern ntrees
STree_mkEmptyNode (ntrees p_nom)
{
  return STree_mkEmptyModel (opnode, p_nom);
}

/**
   @return the model of a function with 
   - p_nom  its name
   - empty interface
   - empty description (expression + local declarations)
*/
extern ntrees
STree_mkEmptyFunction (ntrees p_nom)
{
  return STree_mkEmptyModel (opfunc, p_nom);
}

/**
   @return the model of an action with 
   - p_nom  its name
   - empty interface
   - empty description (expression + local declarations)
*/
extern ntrees
STree_mkEmptyAction (ntrees p_nom)
{
  return STree_mkEmptyModel (opact, p_nom);
}
/**
   @return the model of an automaton with 
   - p_nom  its name
   - empty interface
   - empty description (expression + local declarations)
*/
extern ntrees
STree_mkEmptyAutomaton (ntrees p_nom)
{
  return STree_mkEmptyModel (opautomaton, p_nom);
}


/**
   @return the model of a process with 
   - p_nom  its name
   - p_inte  its interface
   - p_desc its description (expression + local declarations)
*/
extern ntrees
STree_mkProcess (ntrees p_nom, ntrees p_inte, ntrees p_desc)
{
  return nmaketernary (opprocess, p_nom, p_inte, p_desc);
}

/**
   @return an empty confined process.
   \bsg <processus_confine> ::= <primaire_processus> <bloc_declarations> ;\esg
*/
extern ntrees
STree_mkEmptyConfinedExpr (void)
{
  ntrees p_expr, p_loc;

  p_expr = LL_make_lfeed_emptyList ();
  p_loc = LL_make_ldecl_emptyList ();
  return (LL_make_pconf_tree (p_expr, p_loc));
}

/**
   @return an empty module with name as name.
*/
extern ntrees STree_mkSkeletonModule /* STree_mkModule */ (ntrees name)
{
  return (LL_make_modu_tree (name, LL_make_descmodu_tree (STree_mkEmptyTree (), LL_make_ldecl_emptyList ())));
}


/**
   It replaces the inputs/outputs of a model (proc) by inOut.
   The first child of inOut are the inputs.
   The second child of inOut are the outputs.
*/
extern void
STree_replaceInputsOutputs (ntrees proc, ntrees inOut)
{
  ntrees p_es;

  p_es = nchild (2, nchild (2, proc));
  nchangchild (p_es, nchild (1, inOut), 1);
  nchangchild (p_es, nchild (2, inOut), 2);
}


/**
   It assigns an interface (interface) to a model (proc)
*/
extern void
STree_replaceInterface (ntrees proc, ntrees interface)
{
  nchangchild (proc, interface, 2);
}


/**
   It replaces the parameters of a model (proc) by param.
*/
extern void
STree_setParameters (ntrees proc, ntrees param)
{
  nchangchild (STree_getInterface (proc), param, 1);
}



/**
   It replaces the "external graph" of a model (proc) by eg.
*/
extern void
STree_setExternalGraph (ntrees proc, ntrees eg)
{
  nchangchild (STree_getInterface (proc), eg, 3);
}


/**
   It replaces:
   - when binputs is "true", the inputs of a model (proc) by lsig.
   - when binputs is "false", the outputs of a model (proc) by lsig.
*/
extern void
STree_replaceInputsOutputsModel (ntrees proc, ntrees l_sig, boolean binputs)
{
  ntrees p_es;

  p_es = nchild (2, nchild (2, proc));
  if (binputs)
    nchangchild (p_es, l_sig, 1);
  else
    nchangchild (p_es, l_sig, 2);
}

/**
   It inserts the dlist of declarations (l_decl) to the local declarations of a model (p).
*/
extern void
STree_addDeclarations (ntrees p, ntrees l_decl)
{
  ntrees l_decl_pro;
  int i, nbfils;

  l_decl_pro = STreeModel_getLocalDeclarations (p);
  nbfils = ntreearity (l_decl);
  for (i = 1; i <= nbfils; i++)
    STree_addDeclarationToList (l_decl_pro, nchild (i, l_decl));
}

/**
   It inserts in a model (proc1) a copy of all the local declarations of another model (proc2).
*/
extern void
STree_importDeclarations (ntrees proc1, ntrees proc2)
{
  ntrees l_decl_a_inserer, newTree;

  if (proc1 != NILTREE)
    {
      l_decl_a_inserer = STreeModel_getLocalDeclarations (proc2);
      newTree = ncopytree (l_decl_a_inserer, pk_true);
      if (newTree != NILTREE)
	STree_addDeclarations (proc1, newTree);
    }
}


/**
   Deletes all the types, constants, signals declarations of the model or modul p (in)
   and replaces them by the elements of the declaration list l_decl (in). 
   The new declarations are inserted at the beginning of the list.
*/
extern void
STree_replaceLocalDecl (ntrees p, ntrees l_decl)
{
  ntrees p_loc, fils;
  int nb_elem, i, j;

  if (STree_isLocalDeclList (l_decl))	/* en principe test inutile */
    {
      p_loc = STreeModel_getLocalDeclarations (p);
      nb_elem = ntreearity (p_loc);
      j = 1;
      /* on detruit les anciennes declarations de signaux, types */
      for (i = 1; i <= nb_elem; i++)
	{
	  fils = nchild (j, p_loc);
	  if (!STree_isModel (fils))
	    fils = ndestroyListElem (p_loc, j);
	  else
	    j++;
	}
      /* on insère les nouvelles declarations de signaux, types, cstes */
      nb_elem = ntreearity (l_decl);
      if (nb_elem != 0)
	ntreeList_insertFirst (nchild (1, l_decl), p_loc);	/* on insère le 1 ier élément en tête */
      for (i = 2; i <= nb_elem; i++)	/* on insère les autres à la suite */
	ninsertright (i - 1, nchild (i, l_decl), p_loc);
    }
}

/**
   It assigns a description (desc) to an assumed model (process).
*/
extern void
STree_setDescription (ntrees process, ntrees desc)
{
  nchangchild (process, desc, 3);
}

/**
   It assigns the pragmas (pragmas) to a model (proc).
*/
extern void
STree_setPragmas (ntrees proc, ntrees pragmas)
{
  ntrees p_desc;
  p_desc = nchild (3, proc);
  nchangchild (p_desc, pragmas, 1);
}

/**
   If it exists a pragma "vname OBJ statement" and OBJ is not empty, associated with a model (p), the object (vobj) is added to it.
   @return "true" if the operation succeeds, "false" otherwise.
   assumed: p is a model.
*/
extern boolean
STree_addObjToPragmaOf (ntrees p, ptrepstr vname, ntrees vobj, ntrees statement)
{
  ntrees lp, ap, obj, stmt;
  ptrepstr id;
  int i, n;

  if (STree_isUndefinedTree (vobj))
    return pk_false;
  lp = STree_getPragmas (p);
  n = ntreearity (lp);
  for (i = 1; i <= n; i++)
    {
      ap = nchild (i, lp);
      STree_getPragmaValues (ap, &id, &obj, &stmt);
      if (!STree_isUndefinedTree (obj))
	if (streq (id, vname))
	  if (STree_isTreeEq (stmt, statement))
	    {
	      obj = ntreeList_appendLast (obj, vobj);
	      return pk_true;
	    }
	  else
	    id = nilstr;
    }
  return pk_false;
}

/**
   It inserts in a list of declarations (l_liste) the declaration of a signal/constant (arb_def) with arb_typ as type.
   The unicity is guaranteed.
*/
extern void
STree_addFlowDeclaration (ntrees l_liste, ntrees arb_typ, ntrees arb_def)
{
  int i, j, nblp, nbp, typ, typ_list, ope;
  boolean trouver, typexist;
  ntrees liste, sous_liste, lsign_liste, idc, p, arb_typ2, nom_cour, nom;

  nom = STree_getNameOfDeclaration (arb_def);
  ope = noper (arb_def);
  nblp = ntreearity (l_liste);
  trouver = pk_false;
  typexist = pk_false;
  typ = noper (arb_typ);
  i = 1;
  if (noper (l_liste) == opnil)
    {
      i = nposit (l_liste);
      p = LL_make_ldecs_emptyList ();
      nchangchild (nfather (l_liste), p, i);
      l_liste = p;
    }
  else
    while ((i <= nblp) && !trouver)
      {
	sous_liste = nchild (i, l_liste);
	if ((STree_isSignalDecl (sous_liste) && (ope != opcons)) || (STree_isConstantDecl (sous_liste) && (ope == opcons)))
	  {
	    typ_list = STreeDecl_getType (sous_liste, &arb_typ2);
	    liste = nchild (2, sous_liste);
	    if (typ == typ_list)
	      if (STree_isTreeEq (ntrees_getPrevious (liste), arb_typ))
		{
		  typexist = pk_true;
		  lsign_liste = liste;
		}
	    nbp = ntreearity (liste);
	    j = 1;
	    while ((j <= nbp) && !trouver)
	      {
		idc = nchild (j, liste);
		nom_cour = STree_getNameOfDeclaration (idc);
		trouver = STree_isIdentEq (nom, nom_cour);
		j++;
	      }
	  }
	i++;
      }
  if (!trouver)
    if (typexist)
      lsign_liste = ntreeList_appendLast (lsign_liste, arb_def);
    else if (ope == opcons)
      l_liste = ntreeList_appendLast (l_liste, STree_mkConstantDeclaration (arb_def, arb_typ));
    else
      l_liste = ntreeList_appendLast (l_liste, STree_mkSignalDeclaration (arb_def, arb_typ, NILTREE));
}


/**
   It inserts the declarations of a list of declarations (l_decl_a_inserer) 
   in another list of declarations (ldecl). The lists can contain models or/and signal/constant declaration.
   The unicity is not guaranteed.
*/
extern void
STree_mergeLocalDecl (ntrees l_decl, ntrees l_decl_a_inserer)
{
  ntrees p_typ, elem, sous_liste, l_sig;
  int i, j, nb, nb_elem, typ;

  if (ISNOT_NILTREE ((l_decl_a_inserer)) && (noper (l_decl_a_inserer) != opnil))
    {
      nb = ntreearity (l_decl_a_inserer);
      for (i = 1; i <= nb; i++)
	{
	  sous_liste = nchild (i, l_decl_a_inserer);
	  if (STree_isModel (sous_liste))
	    {			/* declaration of a model. */
	      if (!STree_isModelInList (STreeModelModule_getName (sous_liste), l_decl, &elem))
		STree_addDeclarationToList (l_decl, ncopytree (sous_liste, pk_true));
	    }
	  else if (STree_isSignalDecl (sous_liste) || STree_isConstantDecl (sous_liste))
	    {			/* cas des declarations de signaux */
	      typ = STreeDecl_getType (sous_liste, &p_typ);
	      l_sig = nchild (2, sous_liste);
	      nb_elem = ntreearity (l_sig);
	      for (j = 1; j <= nb_elem; j++)
		{
		  elem = nchild (j, l_sig);
		  STree_addFlowDeclaration (l_decl, ncopytree (p_typ, pk_true), ncopytree (elem, pk_true));
		}
	    }
	}
    }
}


/**
   @return a new SIGNAL tree skeleton of a model/module.
   @param id (in) name of the created object.
   @param op (in) a Signal operator assumed to be in {opprocess, opprocessp, opfunc, opfuncp, opmodu, oplmodu, opact}
   @see {STree_mkSkeletonExternProcess}
*/
extern ntrees
STree_mkSkeletonModel (ptident id, noperators op)
{
  ntrees vtree2, vtree3;

  if (op == opmodu)
    {
      vtree2 = LL_make_descmodu_tree (STree_mkAtomNilTree (), LL_make_ldecl_emptyList ());
      return LL_make_modu_tree (copy_ident (id), vtree2);
    }
  vtree2 = STree_mkInterface (NILTREE, NILTREE, NILTREE, NILTREE);
  vtree3 = LL_make_desc_tree (NILTREE, NILTREE);
  return nmaketernary (op, copy_ident (id), vtree2, vtree3);
}

/**
   @return the Signal skeleton tree of an external process.
   @param nameprocess (in) the name of the model.
   @param vlang (in) tree denoting language implementation of the external process.
   @param ope (in) a Signal operator assumed to be in {opprocess, opprocessp, opfunc, opfuncp, opact}
   @see {STree_mkSkeleton}
*/
extern ntrees
STree_mkSkeletonExternProcess (ptident nameprocess, ntrees vlang, noperators ope)
{
  ntrees vtree2, vtree3;

  vtree2 = nmaketernary (opinte, NILTREE, LL_make_entsor_tree (NILTREE, NILTREE), NILTREE);
  if (IS_NILTREE (vlang))
    vlang = STree_mkEmptyTree ();
  else
    vlang = STree_copy (vlang, pk_false);
  vtree3 = LL_make_desc_tree (NILTREE, vlang);
  //  if ( aProcess) ope = opprocess; else ope = opfunc;
  return nmaketernary (ope, copy_ident (nameprocess), vtree2, vtree3);
}


/**
   @return a list of synchronizations initialized with e (i.e. e ^= ).
   It must be completed to have a syntactic correct tree.
*/
extern ntrees
STree_mkSynchroIni (ntrees e)
{
  return LL_make_lequhor_list (e);
}

/**
   It adds an expression (t2) to a list of synchronizations (l).
   @return the new list.
*/
extern ntrees
STree_addSynchro (ntrees l, ntrees e)
{
  if (IS_NILTREE (l))
    return STree_mkSynchroIni (e);
  return ntreeList_appendLast (l, e);
}


/** 
    @return an "TYPE_NOT_IMPLEMENTED" atomic tree.  
*/
extern ntrees
STree_mkNOTIMPLEMENTED ()
{
  return LL_make_nom_atom (string2repstr ("TYPE_NOT_IMPLEMENTED"));
}

/**
   @return an atomic tree with op as operator.
*/
extern ntrees
STree_mkNullaryTree (noperators op)
{
  return natomstring (op, nilstr);
}

/**  
     @return an "empty string" atomic tree. 
*/
extern ntrees
STree_mkEmptyString ()
{
  ptrepstr emptystr;

  strconcchacha ('"', '"', &emptystr);
  return STree_mkStringFromStr (emptystr);
}

/**
   @return a string atomic tree with vstr as value.
*/
extern ntrees
STree_mkStringFromStr (ptrepstr vstr)
{
  return LL_make_chai_list (LL_make_cha_atom (vstr));
}

/**
   @return a nil tree. 
*/
extern ntrees
STree_mkAtomNilTree ()
{
  return STree_mkNullaryTree (opnil);
}


/** @return an "external" keyword tree. */
extern ntrees
STree_mkUndefinedExternalValue ()
{
  return LL_make_exte_tree (STree_mkAtomNilTree ());
}

/** @return an  "external" vstring keyword tree. */
extern ntrees
STree_mkExternalString (ntrees vstring)
{
  return LL_make_exte_tree (vstring);
}

/** @return a  process call to id. */
extern ntrees
STree_mkReferenceToModel (ptident id)
{
  return LL_make_inst_tree (copy_ident (id), STree_mkNullaryTree (opnil));
}

/** @return an abtract tree representing a Signal confined process
    \bsg (| expr |) where <decl> end;  \esg
*/
extern ntrees
STree_mkConfinedExpression (ntrees expr, ntrees decl)
{
  return LL_make_pconf_tree (expr, decl);
}

/**
   It adds a "local declaration (where)" area in second child of the p_dec tree.
   @return the created tree.
*/
extern ntrees
STree_addLocalBlock (ntrees p_desc)
{
  ntrees p, p_conf;
  // cree un arbre d'operateur oppconf comme deuxieme fils de l'arbre ?p_desc.
  p = nchild (2, p_desc);
  p_conf = STree_mkConfinedExpression (p, NILTREE);
  nchangchild (p_desc, p_conf, 2);
  return p_conf;
}

/**
   It adds an empty interface for a process.
   @param p_proc (in) : process tree (assumed).
   @return the interface created tree.
*/
extern ntrees
STree_addEmptyInterface (ntrees p_proc)
{
  ntrees p_inte;

  p_inte = nmaketernary (opinte, NILTREE, LL_make_entsor_tree (NILTREE, NILTREE), NILTREE);	/* PLG */
  nchangchild (p_proc, p_inte, 2);
  return p_inte;
}


/**
   It adds an empty pragmas area for a model (process).
   @return the created tree (empty list of pragmas).
*/
extern ntrees
STree_mkPragmasList (ntrees process)
{
  ntrees lprag;

  lprag = LL_make_lprag_emptyList ();
  nchangchild (STree_getDescription (process), lprag, 1);

  return lprag;
}


/** \bsg  <pragma> ::=<identificateur> <o_l_objets_pragma> <enonce_pragma>\esg
    @return a new pragma with kw as \<identificateur\>, obj as \<o_l_objets_pragma\> and stmt as \<enonce_pragma\>.
*/
extern ntrees
STree_mkPragma (ptrepstr kw, ntrees obj, ntrees stmt)
{
  return STree_mkPragmaIdent (LL_make_nom_atom (kw), obj, stmt);
}


/** \bsg  <pragma> ::=<identificateur> <o_l_objets_pragma> <enonce_pragma>\esg
    @return a new pragma with kw as \<identificateur\>, obj as \<o_l_objets_pragma\> and stmt as \<enonce_pragma\>.
*/
extern ntrees
STree_mkPragmaIdent (ptident id, ntrees obj, ntrees stmt)
{
  if (STree_isUndefinedTree (obj))
    obj = STree_mkAtomNilTree ();
  else if (noper (obj) != oplopra)
    obj = LL_make_lopra_list (obj);

  if (STree_isUndefinedTree (stmt))
    stmt = STree_mkAtomNilTree ();
  else if (noper (stmt) != opchai)
    stmt = LL_make_chai_list (stmt);

  return nmaketernary (opprag, id, obj, stmt);
  // stmt:  LL_make_chai_list( LL_make_cha_atom( enonce));
}

/** It adds a pragma (apragma) to a process (aProcess). */
extern void
STree_addPragma (ntrees aProcess, ntrees apragma)
{
  ntrees descr, prag, vobjCurrentElem, vobjToAdd, velem;
  boolean b;
  int i, nbelem;

  if (STree_isModel (aProcess))
    {
      descr = nchild (3, aProcess);
      prag = nchild (1, descr);
      if (IS_NILTREE (prag))
	b = pk_true;
      else
	b = (noper (prag) == opnil);
      if (b)
	{
	  prag = LL_make_lprag_list (apragma);
	  nchangchild (descr, prag, 1);
	}
      else
	{
	  nbelem = ntreearity (prag);
	  i = 1;
	  b = pk_false;
	  while (!b && (i <= nbelem))
	    {
	      velem = nchild (i, prag);
	      if (STree_isTreeEq (nchild (1, velem), nchild (1, apragma)))
		if (STree_isTreeEq (nchild (3, velem), nchild (3, apragma)))
		  {
		    vobjCurrentElem = nchild (2, velem);
		    vobjToAdd = nchild (2, apragma);
		    if (IS_NILTREE (vobjCurrentElem))
		      if (IS_NILTREE (vobjToAdd))
			b = pk_true;
		      else;
		    else if (!IS_NILTREE (vobjToAdd))
		      {
			STreeList_conc (vobjCurrentElem, vobjToAdd);
			b = pk_true;
		      }
		  }
	      i++;
	    }
	  if (!b)
	    prag = ntreeList_appendLast (prag, apragma);
	}
    }
  else;				// To be completed for a module.
}

/**
   \bsg  <def_type> ::= <identificateur> "=" <description_type> "init" <expr_signal>. \esg
   @return a type declaration with name as \<identificateur\>, description as \<description_type\> and vinit as \<expr_signal\>.
*/
extern ntrees
STree_mkDclType (ntrees name, ntrees description, ntrees vinit)
{
  if (IS_NILTREE (description))
    return name;
  if (IS_NILTREE (vinit))
    vinit = STree_mkAtomNilTree ();
  return nmaketernary (optype, name, description, vinit);
}


/**
   It adds a declaration (vdecl) to a list of declarations (ldecl).
   For typed identifier, it searches the existency of such type for adding, 
   otherwise, it adds at the head of the list.
   @return the new list of declarations.
*/
extern ntrees
STree_addDeclarationToList (ntrees ldecl, ntrees vdecl)
{
  ntrees vtreeType, declWithType;
  noperators vop;
  
  if (IS_NILTREE (vdecl)) return ldecl;
  if (STree_isModel (vdecl)) return ntreeList_appendLast (ldecl, vdecl);
  vop = noper(vdecl);
  if ((vop != opdecsig) && ( vop != opdecpar)) return insertConstantOrType (ldecl, vdecl);
  vtreeType = nchild (1, vdecl);
  if (!existDeclWithType (ldecl, vtreeType, vop, &declWithType))
    return ntreeList_insertFirst (vdecl, ldecl);
  TreeAddList (nchild (2, declWithType), nchild (2, vdecl));
  return ldecl;
}

/**
   It adds a declaration (vdecl) to the end of the list of declarations (ldecl).
   @return the new list of declarations.
*/
extern ntrees
STree_addEndDeclarationToList (ntrees ldecl, ntrees vdecl)
{
  ntrees vtreeType, declWithType, lastDecl;
  int n;

  if (ISNOT_NILTREE (vdecl))
    {
      n = ntreearity (ldecl);
      if (n == 0)
	return ntreeList_appendLast (ldecl, vdecl);
      if (noper (vdecl) != opdecsig)
	return ntreeList_appendLast (ldecl, vdecl);
      vtreeType = nchild (1, vdecl);
      lastDecl = nchild (n, ldecl);
      if (!isDeclWithType (lastDecl, vtreeType, &declWithType))
	ldecl = ntreeList_appendLast (ldecl, vdecl);
      else
	TreeAddList (nchild (2, lastDecl), nchild (2, vdecl));
    }
  return ldecl;
}

/**
   It adds the declarations(vdecl) to an assumed confined expression (p_conf).
*/
extern void
STree_addDeclarationToConfinedExpr (ntrees p_conf, ntrees vdecl)
{
  ntrees ldecl;

  ldecl = nchild (2, p_conf);
  if (IS_NILTREE (ldecl))
    {
      ldecl = LL_make_ldecl_list (vdecl);
      nchangchild (p_conf, ldecl, 2);
    }
  else				/* Adding after the constants, and before the signals (to be completed) */
    STree_addDeclarationToList (ldecl, vdecl);
}

/**
   It adds the declaration (vdecl) as local to a model or a module (aProcess) or a confined process.
*/
extern void
STree_addLocalDeclaration (ntrees aProcess, ntrees vdecl)
{
  ntrees vcorps, ldecl, p_conf;

  if (IS_NILTREE (vdecl))
    return;
  if (!STree_isModule (aProcess))
    {
      if (noper (aProcess) == oppconf)
	p_conf = aProcess;
      else
	{
	  vcorps = nchild (3, aProcess);
	  p_conf = nchild (2, vcorps);
	  if (IS_NILTREE (p_conf))
	    p_conf = STree_addLocalBlock (vcorps);
	  else if (noper (p_conf) != oppconf)
	    p_conf = STree_addLocalBlock (vcorps);
	}
      STree_addDeclarationToConfinedExpr (p_conf, vdecl);
    }
  else
    {				// module
      ldecl = STreeModel_getLocalDeclarations (aProcess);
      if (noper (vdecl) == oplimpo)
	ldecl = ntreeList_appendLast (ldecl, vdecl);
      else
	STree_addDeclarationToList (ldecl, vdecl);
    }
}

/**
   It adds a process declaration (vtreeproapp)to a model/module (vtreeproc).
*/
extern void
STree_addProcessDeclaration (ntrees vtreeproc, ntrees vtreeproapp)
{
  ntrees vcorps, ldecpro, p_conf;
  int ope;

  if (!STree_isModule (vtreeproc))
    {				// process; node; func, action
      vcorps = nchild (3, vtreeproc);
      p_conf = nchild (2, vcorps);
      if (ISNOT_NILTREE (p_conf))
	ope = noper (p_conf);
      else
	ope = 0;
      if (ope == oppconf)
	{
	  ldecpro = nchild (2, p_conf);
	  if (IS_NILTREE (ldecpro))
	    {
	      ldecpro = LL_make_ldecl_list (vtreeproapp);
	      nchangchild (p_conf, ldecpro, 2);
	    }
	  else
	    ldecpro = ntreeList_appendLast (ldecpro, vtreeproapp);
	}
      else
	{
	  p_conf = STree_addLocalBlock (vcorps);
	  nchangchild (p_conf, LL_make_ldecl_list (vtreeproapp), 2);
	}
    }
  else
    {				// module
      ldecpro = STreeModel_getLocalDeclarations (vtreeproc);
      ldecpro = ntreeList_appendLast (ldecpro, vtreeproapp);
    }
}


/** 
    @return the list of type declaration vtypes + {vtype}.
*/
extern ntrees
STree_addTypeToTypes (ntrees vtypes, ntrees vtype)
{
  if (IS_NILTREE (vtypes))
    return LL_make_ltype_list (vtype);
  return ntreeList_appendLast (vtypes, vtype);
}

/** 
    It adds the declarations (vtDecl) to the specification part of a process model (process).
    @param process (in) : process model (assumed)
    @param vtDecl (in) : a list of declarations(assumed).
*/
extern void
STree_addSpecificationDeclaration (ntrees process, ntrees vtDecl)
{
  ntrees vt, vt1, vt2;

  vt = STree_getSpec (process);
  if (IS_NILTREE (vt))
    {
      vt1 = STree_mkEmptyConfinedExpr ();
      nchangchild (vt1, vtDecl, 2);
      STree_addSpecification (process, vt1);
    }
  else
    {
      vt2 = nchild (2, vt);
      if (IS_NILTREE (vt2))
	{
	  vt1 = STree_mkEmptyConfinedExpr ();
	  nchangchild (vt1, vtDecl, 2);
	}
      else
	vt1 = STree_mkConfinedExpression (vt2, vtDecl);
      nchangchild (vt, vt1, 2);
    }
}

/**
   It adds an element (vexpapp) at the head of a composition list (vlfeed).
   @param vlfeed (in/out) feed list, created when empty.
   @param vexpapp (in) element to add (assumed : process expression).
*/
extern void
STree_Composition_ini (ntrees * vlfeed, ntrees vexpapp)
{
  if (STree_isEmptyList (vexpapp))
    return;
  (*vlfeed) = INIT_COMPOSITION (vexpapp);
}

/**
   It adds an element (vexpapp) at the head of a composition list (vlfeed).
   @param vlfeed (in/out) feed list, created when empty.
   @param vexpapp (in) element to add (assumed : process expression).
*/
extern void
STree_addExprToComposition_Pre (ntrees * vlfeed, ntrees vexpapp)
{
  if (IS_NILTREE (*vlfeed))
    STree_Composition_ini (vlfeed, vexpapp);
  else
    {
      if (STree_isEmptyList (vexpapp))
	return;
      *vlfeed = ntreeList_insertFirst (vexpapp, *vlfeed);
    }
}

/**
   It adds an element (vexpapp) at the tail of a composition list (vlfeed).
   @param vlfeed (in/out) feed list, created when empty.
   @param vexpapp (in) element to add (assumed: process expression).
*/
extern void
STree_addExprToComposition_Post (ntrees * vlfeed, ntrees vexpapp)
{
  if (IS_NILTREE (*vlfeed))
    STree_Composition_ini (vlfeed, vexpapp);
  else
    {
      if (STree_isEmptyList (vexpapp))
	return;
      *vlfeed = ntreeList_appendLast (*vlfeed, vexpapp);
    }
}

/** 
    It adds a specification object (vtreeSpecif) to a process model (process).
    @param process (in) : process model (assumed)
    @param vtreeSpecif ( in) : specification expression(assumed).
*/
extern void
STree_addSpecification (ntrees process, ntrees vtreeSpecif)
{
  ntrees vattr, aux, compo, vinterface, vSpec, vt;
  int i, n;
  noperators op;

  if (!STree_isUndefinedTree (vtreeSpecif))
    if (!STree_isUndefinedTree (nchild (1, vtreeSpecif)) || !STree_isUndefinedTree (nchild (2, vtreeSpecif)))
      {
	vinterface = nchild (2, process);
	vSpec = nchild (3, vinterface);	/* external graph */
	if (IS_NILTREE (vSpec))
	  {
	    if (noper (vtreeSpecif) == opspec)
	      vSpec = vtreeSpecif;
	    else
	      {
		SET_NILTREE (vt);
		STree_Composition_ini (&vt, vtreeSpecif);
		vSpec = LL_make_spec_tree (STree_mkAtomNilTree (), vt);
	      }
	    nchangchild (vinterface, vSpec, 3);
	  }
	else
	  {
	    SET_NILTREE (aux);
	    SET_NILTREE (compo);
	    op = noper (vtreeSpecif);
	    if (op == opspec)
	      {
		aux = nchild (1, vtreeSpecif);
		compo = nchild (2, vtreeSpecif);
	      }
	    else if (op == oplattr)
	      aux = vtreeSpecif;
	    else
	      compo = vtreeSpecif;

	    if (!STree_isSignalNilTree (aux))
	      {
		vattr = nchild (1, vSpec);
		if (STree_isSignalNilTree (vattr))
		  nchangchild (vSpec, aux, 1);
		else
		  {
		    n = ntreearity (aux);
		    for (i = 1; i <= n; i++)
		      vattr = ntreeList_appendLast (vattr, nchild (i, aux));
		  }
	      }
	    if (!STree_isSignalNilTree (compo))
	      {
		vt = nchild (2, vSpec);
		STree_addExprToComposition_Post (&vt, compo);
	      }
	  }
      }
}



/**
   It adds (tail) a process expression (vexprpro) to a process model (vtreeproc).
*/
extern void
STree_addExpressionOfProcessToaModel (ntrees vtreeproc, ntrees vexprpro)
{
  STree_addExpressionOfProcessToaModel_loc (vtreeproc, vexprpro, pk_true);
}

/**
   It adds (head) a process expression (vexprpro) to a process model (vtreeproc).
*/
extern void
STree_addPreExpressionOfProcessToaModel (ntrees vtreeproc, ntrees vexprpro)
{
  STree_addExpressionOfProcessToaModel_loc (vtreeproc, vexprpro, pk_false);
}


/** 
    It replaces the body of the process (vtreeproc) by a new expression (vexprpro).
    @param vtreeproc (in) main object (assumed to be a process).
    @param vexprpro (in) a body expression (assumed).
*/
extern void
STree_setBody (ntrees vtreeproc, ntrees vexprpro)
{
  ntrees p_desc, p;
  noperators op;

  if (IS_NILTREE (vexprpro))
    return;
  op = noper (vexprpro);
  if ((op == oplfeed) || (op == opnullproc))
    {
      p_desc = nchild (3, vtreeproc);
      p = nchild (2, p_desc);
      if (IS_NILTREE (p))
	nchangchild (p_desc, vexprpro, 2);
      else if (noper (p) == oppconf)
	nchangchild (p, vexprpro, 1);
      else
	nchangchild (p_desc, vexprpro, 2);
    }
}


/** 
    \bsg  <interface> ::= <o_suite_parametres_formels> <entrees_sorties> <graphe_externe>. \esg
    
    @return the interface with param as \<o_suite_parametres_formels\>, ins + outs  \<entrees_sorties\> and spec as \<graphe_externe\>
*/
extern ntrees
STree_mkInterface (ntrees param, ntrees ins, ntrees outs, ntrees spec)
{
  return nmaketernary (opinte, param, LL_make_entsor_tree (ins, outs), spec);
}

/**
   @return a copy of the interface of the model without specifications.
   @param process (in) a process model (assumed).
*/
extern ntrees
STree_duplicateInterfaceWithoutSpec (ntrees process)
{
  ntrees inter, param, es;

  inter = STree_getInterface (process);
  param = STree_copy (nchild (1, inter), pk_false);
  es = STree_copy (nchild (2, inter), pk_false);
  return STree_mkInterface (param, nchild (1, es), nchild (2, es), NILTREE);
}

/**
   It adds a declaration to a model. 
   @param vtreeproc (in) a process model (assumed).
   @param vdecsig (in) a signal/parameter declaration.
   @param where (in) specify the area : 1 for parameters, 2 for inputs, 3 for outputs.
*/
extern void
STree_addDeclarationToInterface (ntrees vtreeproc, ntrees vdecsig, int where)
{
  ntrees vinterf, ldecsig, entsor;

  if (IS_NILTREE (vdecsig))
    return;
  vinterf = nchild (2, vtreeproc);
  if (where == 1)
    {
      ldecsig = nchild (1, vinterf);
      if (STree_isSignalNilTree (ldecsig))
	{
	  ldecsig = LL_make_ldecs_list (vdecsig);
	  nchangchild (vinterf, ldecsig, 1);
	}
      else
	ldecsig = ntreeList_insertFirst (vdecsig, ldecsig);
    }
  else
    {
      entsor = nchild (2, vinterf);
      ldecsig = nchild (where - 1, entsor);
      if (STree_isSignalNilTree (ldecsig))
	{
	  ldecsig = LL_make_ldecs_list (vdecsig);
	  nchangchild (entsor, ldecsig, where - 1);
	}
      else
	ldecsig = ntreeList_insertFirst (vdecsig, ldecsig);
    }
}





/**
   \bsg  <production> ::= <identificateur> { <o_l_expr_param> }. \esg
   @return a "process call" tree to a model (pro) with parameters (lpa).
*/
extern ntrees
STree_mkReferenceToModelWithParameter (ntrees lpa, ntrees pro)
{
  ntrees pronom;

  pronom = nchild (1, pro);
  if (ISNOT_NILTREE (pronom))
    if (noper (pronom) == opnom)
      pronom = LL_make_nom_atom (nstringof (pronom));
    else
      pronom = copy_ident (pronom);
  if (IS_NILTREE (lpa))
    return LL_make_inst_tree (pronom, NILTREE);
  if (noper (lpa) == oplepara)
    return LL_make_inst_tree (pronom, lpa);
  return LL_make_inst_tree (pronom, NILTREE);
}


/** 
    It modifies the body of the process (vprocessus) by adding the masking of some signals (SignalList).
    @param vprocessus (in) a process model (assumed).
    @param SignalList (in) a list of signals (assumed) to be masked.
*/
extern void
STree_maskSignals (ntrees vprocessus, ntrees SignalList)
{
  ntrees vtcorps, vt, p;
  int no;

  if (IS_NILTREE (SignalList))
    return;
  vtcorps = nchild (3, vprocessus);
  vt = nchild (2, vtcorps);
  if (noper (vt) == oppconf)
    {
      p = vt;
      vt = nchild (1, vt);
      no = 1;
    }
  else
    no = 2;
  vt = LL_make_obli_tree (vt, SignalList);
  nchangchild (vtcorps, vt, no);
}


/**
   @return the abstract tree denoting an equation.
   @param op (in) Signal operator ( :=, ::= code).
   @param leftExp (in) left operand of the equation.
   @param rightExp (in) right operand of the equation.
*/
extern ntrees
STree_mkEquation (noperators op, ntrees leftExp, ntrees rightExp)
{
  return nmakebinary (op, leftExp, rightExp);
}



/** 
    @return a dependence "vfrom -> vto" or "vfrom -> vto when tClock"
*/
extern ntrees
STree_mkDependenceTree (ntrees vfrom, ntrees vto, ntrees tClock)
{
  ntrees v;
  if (!STree_isUndefinedTree (tClock))
    return nmaketernary (opdepd, vfrom, vto, tClock);
  v = LL_make_ldep_emptyList ();
  ntreeList_appendLast (v, vfrom);
  ntreeList_appendLast (v, vto);
  return v;
}

/** 
    It adds a signal (vto) to the list {y1,...yp} of a dependence (aDep) {x1,...xn} -> {y1,...,yp} when H.
*/
extern void
STree_addDependencePost (ntrees aDep, ntrees vto)
{
  ntrees v, t;

  if (!STree_isSignalNilTree (vto))
    {
      v = nchild (2, aDep);
      t = insertToList (v, vto, oplesign);
      if (t != v)
	nchangchild (aDep, t, 2);
    }
}


/** 
    It adds a signal (vfrom) to the list {x1,...xn} of a dependence (aDep) {x1,...xn} -> {y1,...,yp} when H.
*/
extern void
STree_addDependencePre (ntrees aDep, ntrees vfrom)
{
  ntrees v, t;

  if (!STree_isSignalNilTree (vfrom))
    {
      v = nchild (1, aDep);
      t = insertToList (v, vfrom, oplesign);
      if (t != v)
	nchangchild (aDep, t, 1);
    }
}


/** 
    It adds a dependence (aDep) to a list of dependences (ldep).
    @return the list ldep with the added dependence.
    
    Assumed : ldep is a composition or niltree.
*/
extern ntrees
STree_addDependenceToDependences (ntrees aDep, ntrees ldep)
{
  int n, j;
  boolean bcontinue;
  ntrees elem, vfrom, vto, vclk_elem, vclk;
  noperators op;

  if (STree_isSignalNilTree (aDep))
    return ldep;
  if (STree_isSignalNilTree (ldep))
    return INIT_COMPOSITION (aDep);
  n = ntreearity (ldep);
  j = 1;
  bcontinue = pk_true;
  vfrom = nchild (1, aDep);
  vto = nchild (2, aDep);
  vclk = nchild (3, aDep);
  while ((j <= n) && bcontinue)
    {
      elem = nchild (j, ldep);
      op = noper (elem);
      if (op == opldep)
	SET_NILTREE (vclk_elem);
      else
	vclk_elem = nchild (3, elem);
      if (STree_isTreeEq (vclk, vclk_elem))
	bcontinue = !STree_areSameSets (nchild (1, elem), vfrom);
      j++;
    }
  bcontinue = pk_true;
  if (bcontinue)
    {
      j = 1;
      while ((j <= n) && bcontinue)
	{
	  elem = nchild (j, ldep);
	  op = noper (elem);
	  if (op == opldep)
	    SET_NILTREE (vclk_elem);
	  else
	    vclk_elem = nchild (3, elem);
	  if (STree_isTreeEq (vclk, vclk_elem))
	    bcontinue = !STree_areSameSets (nchild (2, elem), vto);
	  j++;
	}
      bcontinue = pk_true;
      if (bcontinue)
	STree_addExprToComposition_Post (&ldep, aDep);
      else if (STree_isList (vfrom))
	{
	  n = ntreearity (vfrom);
	  for (j = 1; j <= n; j++)
	    STree_addDependencePre (elem, nchild (j, vfrom));
	}
      else
	STree_addDependencePre (elem, vfrom);
    }
  else if (STree_isList (vto))
    {
      n = ntreearity (vto);
      for (j = 1; j <= n; j++)
	STree_addDependencePost (elem, nchild (j, vto));
    }
  else
    STree_addDependencePost (elem, vto);
  return ldep;
}

/**
   @return an empty tree local declarations of a model (p_proc).
*/
extern ntrees
STree_mkLocalDeclaration (ntrees p_proc)
{
  ntrees p, p_fils, p_conf, p_loc;

  p_fils = nchild (3, p_proc);
  if (noper (p_fils) == opdesc)
    p = nchild (2, p_fils);
  else
    p = p_fils;
  p_loc = LL_make_ldecl_emptyList ();
  p_conf = LL_make_pconf_tree (p, p_loc);
  if (p_fils == p)
    nchangchild (p_proc, p_conf, 3);
  else
    nchangchild (p_fils, p_conf, 2);
  return (p_loc);
}


/**
   It removes the parameters tree of a model (p_proc).
*/
extern void
STree_deleteParameters (ntrees p_proc)
{
  ntrees p_inte;

  p_inte = STree_getInterface (p_proc);
  nchangchild (p_inte, STree_mkNullaryTree (opnil), 1);
}


/**
   @return a tree which represents a signal declaration (p_typ name "init" p_init).
   @param p_nom (in) a name
   @param p_typ (in) type
   @param p_init (in) init value.
   
*/
extern ntrees
STree_mkSignalDeclaration (ntrees p_nom, ntrees p_typ, ntrees p_init)
{
  ntrees p, p_ret;

  if (IS_NILTREE (p_init))
    p = p_nom;
  else
    p = LL_make_init_tree (p_nom, p_init);
  p_ret = LL_make_decsig_tree (p_typ, LL_make_lsign_list (p));
  return (p_ret);
}


/**
   @return a tree which represents the signal declaration of the signal (name) 
   with an implicit type and without initialisation.
*/
extern ntrees
STree_mkImplicitSignalDeclaration (ptrepstr name)
{
  ntrees p_ret;

  p_ret = STree_mkSignalDeclaration (LL_make_nom_atom (name), STree_getImplicitTypeTree (), NILTREE);
  return (p_ret);
}


/**
   @returns a constant declaration  with p_typ as type and p_def a list of definitions of constants.
   \bsg <decl_constantes> ::= "constant" <type> <l_def_constantes>. \esg
*/
extern ntrees
STree_mkConstantDeclaration (ntrees p_def, ntrees p_typ)
{
  ntrees p_ret;

  p_ret = LL_make_deccon_tree (p_typ, LL_make_lcons_list (p_def));
  return (p_ret);
}

/** 
    It adds the body of a process (vtreegfd) to the body of another one (vprocessus).
*/
extern void
STree_mergePragmas (ntrees vprocessus, ntrees vtreegfd)
{
  ntrees p, p_desc;
  int n, i;

  if (IS_NILTREE (vtreegfd))
    return;
  p_desc = nchild (3, vtreegfd);
  p = nchild (1, p_desc);
  n = ntreearity (p);
  for (i = 1; i <= n; i++)
    STree_addPragma (vprocessus, nchild (i, p));
}

/** 
    It adds the body of a process (vtreegfd) to the body of another one (vprocessus).
*/
extern void
STree_mergeExprProc (ntrees vprocessus, ntrees vtreegfd)
{
  ntrees p, p_desc;

  if (ISNOT_NILTREE (vtreegfd))
    {
      p_desc = nchild (3, vtreegfd);
      p = nchild (2, p_desc);
      if (ISNOT_NILTREE (p))
	{
	  if (noper (p) == oppconf)
	    p = nchild (1, p);
	  STree_addExpressionOfProcessToaModel (vprocessus, p);
	}
    }
}


/**
   It appends the declarations of a process (vtreegfd) to the declarations of another one (vprocessus).
*/
extern void
STree_mergeDeclarations (ntrees vprocessus, ntrees vtreegfd)
{
  ntrees p, p_desc;
  int n, i;

  if (ISNOT_NILTREE (vtreegfd))
    {
      p_desc = nchild (3, vtreegfd);
      p = nchild (2, p_desc);
      if (ISNOT_NILTREE (p))
	if (noper (p) == oppconf)
	  {
	    p = nchild (2, p);
	    if (ISNOT_NILTREE (p))
	      {
		n = ntreearity (p);
		for (i = 1; i <= n; i++)
		  STree_addProcessDeclaration (vprocessus, nchild (i, p));
	      }
	  }
    }
}


/** 
    It appends the body(resp. declarations) of a process (vtreegfd) to the body(resp. declarations) of another one (vprocessus).
*/
extern void
STree_mergeProcesses (ntrees vprocessus, ntrees vtreegfd)
{
  if (ISNOT_NILTREE (vtreegfd))
    {
      STree_mergePragmas (vprocessus, vtreegfd);
      STree_mergeExprProc (vprocessus, vtreegfd);
      STree_mergeDeclarations (vprocessus, vtreegfd);
    }
}


/**
   It adds the declaration of identifiers (lids) of a process (p) to another process (m).
   Not general : only Type declaration identifier.
*/
extern void
STree_importDeclsFromTo (ntrees lids, ntrees p, ntrees m)
{
  int i, n;
  n = ntreearity (lids);
  for (i = 1; i <= n; i++)
    importDeclFromTo (nchild (i, lids), p, m);
}


/**
   It removes the declaration of identifiers (lids) from a process (p).
   Not general: only type declaration identifier.
*/
extern void
STree_removeDeclsFrom (ntrees lids, ntrees p)
{
  int i, n;
  n = ntreearity (lids);
  for (i = 1; i <= n; i++)
    removeDeclFrom (nchild (i, lids), p);
}


/**
   It sets a name (vident) to a model (m).
*/
extern void
STree_setNameModel (ntrees m, ptident vident)
{
  boolean b;
  if (STree_isModel (m))
    b = pk_true;
  else
    b = STree_isModule (m);
  if (b)
    nchangchild (m, vident, 1);
}


/** 
    @return a copy of a tree (srcTree) with (resp. without) a copy of its attributes when copyattr is true (resp. false).
*/
extern ntrees
STree_copy (ntrees srcTree, boolean copyattr)
{
  return ncopytree (srcTree, copyattr);
}

/** 
    @return a copy of a tree (p) without a copy of its attributes.
    It is equivalent to STree_copy(p ,false)
*/
extern ntrees
STree_scopy (ntrees p)
{
  return ncopytree (p, pk_false);
}

/** 
    \bsg <indication_externe> ::= "external" <chaine>. \esg
    
    @return the abstract tree external language.
    @param vstr (in) code for a language.
*/
extern ntrees
STree_mkExternalImplLanguage (ptrepstr vstr)
{
  ptrepstr v;
  getDoubleQuotedString (vstr, &v);
  return STree_mkExternalString (STree_mkStringFromStr (v));
}



/**
   \bsg   <pragma> ::=<identificateur> <o_l_objets_pragma> <enonce_pragma>.\esg
   
   @return 
   -"true" a pragma with vstrcomment as name exists in a list of pragmas (prag) and the 
   associated \<enonce_pragma\> is not empty. In this case, vt is a copy of \<enonce_pragma\> value.
   - "false" otherwise.
*/
extern boolean
STree_mkSelectedPragma (ntrees prag, ptrepstr vstrcomment, ntrees * vt)
{
  noperators op;
  ntrees enonce, p_obj, aPragma, y, x;
  int i, j, n;
  ptrepstr name;

  if (IS_NILTREE (prag))
    return pk_false;
  op = noper (prag);
  if (op != oplprag)
    return pk_false;
  if (op == opnil)
    return pk_false;
  SET_NILTREE (x);
  n = ntreearity (prag);
  for (i = 1; i <= n; i++)
    {
      aPragma = nchild (i, prag);
      STree_getPragmaValues (aPragma, &name, &p_obj, &enonce);
      if (streq (name, vstrcomment))
	{
	  for (j = 1; j <= ntreearity (enonce); j++)
	    {
	      y = nchild (j, enonce);
	      y = LL_make_cha_atom (nstringof (y));
	      if (IS_NILTREE (x))
		x = LL_make_chai_list (y);
	      else
		x = ntreeList_appendLast (x, y);
	    }
	}
    }
  *vt = x;
  return (ISNOT_NILTREE (x));
}


/**
   It verifies the correction of the form of a module (aModule).
   @return the correct module tree.
*/
extern ntrees
STree_updateModule (ntrees aModule)
{
  noperators op;
  ntrees x, ldecl;

  op = noper (aModule);
  if (op != opmodu)
    return aModule;

  x = nchild (2, aModule);
  if (IS_NILTREE (x))
    {
      x = LL_make_descmodu_tree (NILTREE, NILTREE);
      nchangchild (aModule, x, 2);
      return aModule;
    }
  x = nchild (2, x);
  ldecl = nchild (2, x);
  if (IS_NILTREE (ldecl))
    return aModule;
  if (ntreearity (ldecl) == 0)
    nchangchild (x, NILTREE, 2);
  return aModule;
}

/**
   @return a process call to a model (process).
   The names of the formal declaration (parameters, intputs, outputs) are taken as effective names for the call.
*/
extern ntrees
STree_mkCompleteModelCall (ntrees process)
{
  ntrees inputs, params, outputs, pcall, x;

  x = STreeModelModule_getName (process);
  pcall = STree_copy (x, pk_false);
  params = mkIdListFromDeclarationList (STree_getParameters (process), oplesign);
  if (ISNOT_NILTREE (params))
    pcall = LL_make_inst_tree (pcall, params);
  inputs = mkIdListFromDeclarationList (STree_getInputs (process), oplesign);
  pcall = LL_make_appe_tree (pcall, inputs);
  outputs = mkIdListFromDeclarationList (STree_getOutputs (process), opleflot);
  if (ISNOT_NILTREE (outputs))
    pcall = LL_make_defsig_tree (outputs, pcall);
  return pcall;
}


/**
   @return the l_proc tree without NIL tree sons.
*/
extern ntrees
STree_removeNilTree (ntrees l_proc)
{
  boolean isFinished, noProblem;
  int i, nb;

  isFinished = pk_false;
  // this loop is necessary only if there is problem in the trees; 
  // it deletes the sons of a list which are empty trees.
  while (!isFinished)
    {
      i = 1;
      nb = ntreearity (l_proc);
      noProblem = pk_true;
      while ((i <= nb) && noProblem)
	{
	  if (IS_NILTREE (nchild (i, l_proc)))
	    {
	      ndestroyListElem (l_proc, i);
	      noProblem = pk_false;
	    }
	  else
	    i++;
	}
      isFinished = noProblem;
    }
  return l_proc;
}


/**
   It declares, in p, the list of identifiers (lids) as labels.
   \note It assumes that p is the Signal abstract tree representation of a process.
*/
extern void
STree_addLabelsInModel (ntrees lids, ntrees p)
{
  ntrees vdecl, vt, vtchild;
  int i, where;

  if (ISNOT_NILTREE (lids))
    {
      if (!STree_isList (lids))
	lids = LL_make_lident_list (lids);
      vdecl = LL_make_lab_tree (lids);
      i = 0;
      vt = STree_getDeclaration (p, i, &where);
      vtchild = nchild (where, vt);
      if (IS_NILTREE (vtchild))
	nchangchild (vt, LL_make_ldecl_list (vdecl), where);
      else
	vtchild = ntreeList_insertFirst (vdecl, vtchild);
    }
}


/**
   It declares, in p, a label (vlabel).
   \note It assumes that p is the Signal abstract tree representation of a process.
*/
extern void
STree_addLabelInModel (ntrees vlabel, ntrees p)
{
  ntrees vdecllabel, vdecl, vt, subList, vtchild;
  int i, n, where;

  if (ISNOT_NILTREE (vlabel))
    {
      vt = STree_getDeclaration (p, 0, &where);
      vtchild = nchild (where, vt);
      if (IS_NILTREE (vtchild))
	{
	  vdecl = LL_make_lab_tree (LL_make_lident_list (vlabel));
	  nchangchild (vt, LL_make_ldecl_list (vdecl), where);
	}
      else
	{
	  n = ntreearity (vtchild);
	  SET_NILTREE (vdecllabel);
	  i = 1;
	  while ((IS_NILTREE (vdecllabel)) && i <= n)
	    {
	      subList = nchild (i, vtchild);
	      if (noper (subList) == oplab)
		vdecllabel = subList;
	      else
		i++;
	    }
	  if (IS_NILTREE (vdecllabel))
	    vtchild = ntreeList_appendLast (vtchild, LL_make_lab_tree (LL_make_lident_list (vlabel)));
	  else
	    {
	      subList = nchild (1, vdecllabel);
	      subList = ntreeList_appendLast (subList, vlabel);
	    }

	}
    }
}

/**
   It moves the tyAttr-th (in) attribute of p1 (in) to p2 (in).
*/
extern void
STree_moveAttribute (ntrees p1, ntrees p2, short tyAttr)
{
  ntrees pAttr;

  // printf("DEBUG entering STree_moveAttribute, numattr is %i\n", tyAttr);
  pAttr = nattrntrees (p1, tyAttr);
  ndestroyattr (p1, tyAttr);
  nmakeattrntrees (p2, tyAttr, pAttr);
}


/**
   It sets local declarations (ndecls) to  a model (p).
   @return the previous local declarations of p.
*/
extern ntrees
STree_setLocalDeclarations (ntrees p, ntrees ndecls)
{
  ntrees l, vtBody, vt;

  l = STreeModel_getLocalDeclarations (p);
  if (ISNOT_NILTREE (ndecls))
    if (noper (p) == opmodu)
      nchangchild (nchild (2, p), ndecls, 2);
    else
      {
	vtBody = nchild (3, p);
	vt = nchild (2, vtBody);
	if (IS_NILTREE (vt))
	  {
	    vt = STree_addLocalBlock (vtBody);
	    nchangchild (vt, ndecls, 2);
	  }
	else if (noper (vt) == oppconf)
	  nchangchild (vt, ndecls, 2);
	else
	  {
	    vt = STree_mkConfinedExpression (vt, ndecls);
	    nchangchild (vtBody, vt, 2);
	  }

      }
  return l;
}

/**
   It replaces t by nt in p.
*/
extern boolean
STree_replace (ntrees p, ntrees t, ntrees nt)
{
  noperators op;
  int i, n;
  boolean r;

  if (IS_NILTREE (p))
    return pk_false;
  if (STree_isTreeEq (p, t))
    {
      STree_replaceSon (p, STree_scopy (nt), 0);
      return pk_true;
    }
  op = noper (p);
  if (LL_is_atomOp (op))
    return pk_false;
  if (op == opident)
    return pk_false;
  n = ntreearity (p);
  r = pk_false;
  for (i = 1; i <= n; i++)
    r = r || STree_replace (nchild (i, p), t, nt);
  return r;
}


/**
   For a model (p), it moves the declaration of models defined in the body to the local declarations of p.
   
   \note : The compiler (restriction) does not accept the declaration of models in a confined expression of the body part.
*/
// PLG STree_moveBodyToDecl ==> STreeModel_moveBodyToDecl
extern void
STreeModel_moveBodyToDecl (ntrees p)
{
  ntrees b;

  b = STreeModel_getBody (p);
  STree_moveModelsFrom (p, b);
}

/**
   It imports to a model (p) the processes declared in the tree b.
   It is done for the confined expressions of b, not recursively.
*/
extern void
STree_moveModelsFrom (ntrees p, ntrees b)
{
  int i, n;
  noperators op;
  ntrees ldecl, vdecl;

  if (ISNOT_NILTREE (b))
    {
      op = noper (b);
      switch ((int) op)
	{
	case oplfeed:
	  n = ntreearity (b);
	  for (i = 1; i <= n; i++)
	    STree_moveModelsFrom (p, nchild (i, b));
	  break;

	case opcas:		// cas  -> ENU_VAL E_PROC
	case oppetiq:		// processus etiqueté
	  // case  opspec: spec  -> OL_ATTR O_E_PROC ; 

	  STree_moveModelsFrom (p, nchild (2, b));
	  break;
	case oppconf:
	  ldecl = nchild (2, b);
	  if (ISNOT_NILTREE (ldecl))
	    {
	      n = ntreearity (ldecl);
	      for (i = n; i != 1; i--)
		{
		  vdecl = nchild (i, ldecl);
		  if (STree_isModel (vdecl))
		    {
		      ndestroyListElem (ldecl, i);
		      STree_addLocalDeclaration (p, vdecl);
		    }
		}
	    }
	  break;

	case oppcas:		// p_cas  -> IDENT L_CAS E_PROC ;
	case opmoti:
	case opiter:
	  // iter  -> E_SIGN E_PROC O_E_PROC ;  moti  -> IND_MOT E_PROC O_E_PROC ;
	  STree_moveModelsFrom (p, nchild (2, b));
	  STree_moveModelsFrom (p, nchild (3, b));
	  break;

	  //    case  oppassert: p_assert  -> E_PROC ;
	case opobli:		// obli  -> E_PROC L_IDENT ;
	  STree_moveModelsFrom (p, nchild (1, b));

	}
    }
}

/**
   build a copy of a model P that contains only the name and the interface of P
   @return "true" if it is applied, "false" otherwise (p is not a model).
*/
extern boolean
STree_extractModelHeader (ntrees srcAbstractTree, ntrees * trgtAbstractTree)
{
  if (STree_isModel (srcAbstractTree))
    {
      (*trgtAbstractTree) = STree_mkSkeletonModel (copy_ident (STreeModelModule_getName (srcAbstractTree)), noper (srcAbstractTree));
      nchangchild ((*trgtAbstractTree), STree_duplicateInterfaceWithoutSpec (srcAbstractTree), 2);
      return pk_true;
    }
  return pk_false;
}

/**
   For a module (p), it removes  the definitions of the models to have only the interface.
   @return "true" if it is applied, "false" otherwise (p is not a module).
*/
extern boolean
STree_extractModuleHeaders (ntrees srcAbstractTree, ntrees * trgtAbstractTree)
{
  ntrees src_ldecl, trgt_ldecl, srcDeclarationItem, trgtDeclarationItem, descr;
  int i, n;

  if (!STree_isModule (srcAbstractTree))
    return pk_false;

  (*trgtAbstractTree) = STree_mkSkeletonModule (copy_ident (STreeModelModule_getName (srcAbstractTree)));
  src_ldecl = STreeModel_getLocalDeclarations (srcAbstractTree);
  trgt_ldecl = STreeModel_getLocalDeclarations (*trgtAbstractTree);
  if (ISNOT_NILTREE (src_ldecl))
    {
      n = ntreearity (src_ldecl);
      for (i = 1; i <= n; i++)
	{
	  srcDeclarationItem = nchild (i, src_ldecl);
	  if (STree_extractModelHeader (srcDeclarationItem, &trgtDeclarationItem))
	    STree_addProcessDeclaration (*trgtAbstractTree, trgtDeclarationItem);
	}
    }
  return pk_true;
}


/**
   For a module (p), it removes  the definitions of the models to have only the interface.
   @return "true" if it is applied, "false" otherwise (p is not a module).
*/
extern boolean
STree_mkheaders (ntrees p)
{
  ntrees ldecls, ap, descr;
  int i, n;

  if (!STree_isModule (p))
    return pk_false;

  ldecls = STreeModel_getLocalDeclarations (p);
  if (ISNOT_NILTREE (ldecls))
    {
      n = ntreearity (ldecls);
      for (i = n; i >= 1; i--)
	{
	  ap = nchild (i, ldecls);
	  if (STree_isModel (ap))
	    {
	      descr = nchild (3, ap);
	      nchangchild (descr, STree_mkAtomNilTree (), 2);
	    }
	}
    }
  return pk_true;
}

/**
   Flat mode: there is no composition as child of a composition.
   Warning: the declarations must be moved also !!! (assumed: no pb of visibility)
*/
// PLG STree_flatComposition ==> STreeModel_flatComposition
// PLG void STreeModel_flatComposition(ntrees vmodel)
// PLG {
// PLG   ntrees vt, vdecl;
// PLG
// PLG  if ( STree_isModel(vmodel))
// PLG    {
// PLG      // Body
// PLG      vt = STreeModel_getBody(vmodel);
// PLG      if ( ! STree_isUndefinedTree(vt))
// PLG  if ( STree_isConfinedProcess(vt))
// PLG    STree_flatComposition1(nchild(1, vt), nchild(2, vt));
// PLG  else 
// PLG    STree_setLocalDeclarations(vmodel, STree_flatComposition1(vt, NILTREE));
// PLG      // Spec
// PLG      vt = STree_getSpec(vmodel);
// PLG      if ( ! STree_isUndefinedTree(vt))
// PLG  if ( STree_isConfinedProcess(vt))
// PLG    STree_flatComposition1(nchild(1, vt), nchild(2, vt));
// PLG  else
// PLG    STree_addSpecificationDeclaration(vmodel, STree_flatComposition1(vt, NILTREE));
// PLG    }
// PLG }


/**
   The clock operators (^=, ^<, ^>, ^#) are considered as binary operators: a composition of equations is returned 
   when the arity of the operand (vtlist) is greater than 2.
*/
ntrees
STree_ClockOperatorsListToBinaryOperator (ntrees vtlist)
{
  noperators op;
  integer i, j, n;
  ntrees vtres, l;

  if (STree_isUndefinedTree (vtlist))
    return vtlist;
  if (ntreearity (vtlist) <= 2)
    return vtlist;
  op = noper (vtlist);
  // clock operators: oplcrohor opldechor oplequhor oplexchor
  if (op == oplexchor)
    {
      // all the pairs.
      vtres = STree_getNilTree ();
      i = 1;
      n = ntreearity (vtlist);
      for (i = 1; i <= n - 1; i++)
	for (j = i + 1; j <= n; j++)
	  {
	    l = nmakelist (op, STree_scopy (nchild (i, vtlist)));
	    l = ntreeList_appendLast (l, STree_scopy (nchild (j, vtlist)));
	    STree_addExprToComposition_Post (&vtres, l);
	  }
      return vtres;
    }
  if ((op == oplcrohor) || (op == opldechor) || (op == oplequhor))
    {
      vtres = STree_getNilTree ();
      n = ntreearity (vtlist);
      for (i = 1; i <= n - 1; i++)
	{
	  l = nmakelist (op, STree_scopy (nchild (i, vtlist)));
	  l = ntreeList_appendLast (l, STree_scopy (nchild (i + 1, vtlist)));
	  STree_addExprToComposition_Post (&vtres, l);
	}
      return vtres;
    }
  return vtlist;
}

/**
   If the global comments (Polychrony version, command line) are assigned, they are added  to the operand tree (vt).
   To initialize the global comments, use STree_initPolychronyComment()).
*/
extern void
STree_addGeneralComments (ntrees vt)
{
  STree_addPrintableTreeAttribute (vt, LL_make_nom_atom (string2repstr (" ")), pk_true);
  if (vPolychronyVersion != nilstr)
    STree_addPrintableTreeAttribute (vt, LL_make_nom_atom (vPolychronyVersion), pk_true);
  if (vPolychronyVersion != nilstr)
    STree_addPrintableTreeAttribute (vt, LL_make_nom_atom (vCommandLine), pk_true);
  STree_addPrintableTreeAttribute (vt, LL_make_nom_atom (string2repstr (" ")), pk_true);
}

/**
   Initialises the global comments (Polychrony version, command line).
*/
extern void
STree_initPolychronyComments (ptrepstr messVersion, ptrepstr CommandLine)
{
  vPolychronyVersion = messVersion;
  vCommandLine = CommandLine;
}

/**
   Destroy a tree (vt) if it belongs to a list.
   Recursively to its father if this father becomes empty
*/
extern void
STree_Destroy (ntrees vt)
{
  ntrees vtFather;
  vtFather = nfather (vt);
  if (STree_isList (vtFather))
    {
      ntreesDestroy (vt);
      if (STree_isEmptyList (vtFather))
	STree_Destroy (vtFather);
    }
}
