/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#define INIT_COMPOSITION(v) LL_make_lfeed_list( v)


static ptrepstr vPolychronyVersion = nilstr;
static ptrepstr vCommandLine = nilstr;

/*
  In the documentation  \bsg stands for "begin Signal Grammar" and \esg  "end Signal grammar"
*/
static boolean getDeclarationType (ptident id, ntrees p, ntrees * vdecl);

/**
   @return the number of elements of an assumed list of flow declarations (aList).
   Such a list is a list of lists.

   Example: For int a,b,c; real d; boolean f,g, it returns 6 elements.
*/
static int
get_nb_elem_decl (ntrees aList)
{
  int nb_elem, nb, i;

  if (IS_NILTREE (aList))
    return 0;
  if (noper (aList) == opnil)
    return 0;
  nb_elem = 0;
  nb = ntreearity (aList);
  for (i = 1; i <= nb; i++)
    nb_elem = nb_elem + ntreearity (nchild (2, nchild (i, aList)));
  return (nb_elem);
}


/**
   @return "true" if an identifier (id) is in list of type declarations (ldectypes), "false" otherwise.
*/
static boolean
is_in_listOfTypes (ntrees id, ntrees ldectypes)
{
  int i, n;
  ntrees currentDecl;
  ptident idc;
  noperators voperator;

  n = ntreearity (ldectypes);
  for (i = 1; i <= n; i++)
    {
      currentDecl = nchild (i, ldectypes);
      voperator = noper (currentDecl);
      switch (voperator)
	{
	case opnom:
	case opident:
	  idc = currentDecl;
	  break;
	case optype:
	  idc = nchild (1, currentDecl);
	  break;
	default:
	  idc = NILIDENT;
	}
      if (eq_idents (idc, id))
	return pk_true;
    }
  return pk_false;
}

/**
   @return
   - "true" if the type of a declaration (decl) is equal to arb_typ, declWithType is defined by decl;
   - "false" otherwise (declWithType is not modified).
*/
static boolean
isDeclWithType (ntrees decl, ntrees arb_typ, ntrees * declWithType)
{
  ntrees c_arb_typ;

  if (STree_isSignalDecl (decl) || STree_isSharedSignalDecl (decl))
    {
      STreeDecl_getType (decl, &c_arb_typ);
      if (STree_isTreeEq (c_arb_typ, arb_typ))
	{
	  *declWithType = decl;
	  return pk_true;
	}
    }
  return pk_false;
}


/**
   @return 
   - "true" if it exists a declaration (declWithType) in a list of declarations (ldecl) with a particular type (arb_typ),
   - "false" otherwise (declWithType is not modified).
*/
static boolean
existDeclWithType (ntrees ldecl, ntrees arb_typ, noperators vop, ntrees * declWithType)
{
  int i, nblp;
  ntrees subList;

  nblp = ntreearity (ldecl);
  if (nblp == 0)
    return pk_false;
  for (i = 1; i <= nblp; i++)
    {
      subList = nchild (i, ldecl);
      if (noper(subList) == vop) // vop in { opdecsig, opdecpar,...}
	if (isDeclWithType (subList, arb_typ, declWithType))
	  return pk_true;
    }
  return pk_false;
}

/**
   @return the list l in which a declaration of an assumed constant or type declaration (vdecl) is added.
*/
static ntrees
insertConstantOrType (ntrees l, ntrees vdecl)
{
  int i, nblp, first, last, mid;
  ntrees f;
  boolean bstop;

  nblp = ntreearity (l);
  if (nblp == 0)
    return l = ntreeList_insertFirst (vdecl, l);

  first = 1;
  last = nblp;
  i = first;
  while (first < last)
    {
      mid = first + (last - first) / 2;
      f = nchild (mid, l);
      if (STree_isConstantDecl (f) || STree_isTypeDecl (f))
	{
	  i = first;
	  first = mid + 1;
	}
      else
	last = mid - 1;
    }

  bstop = pk_false;
  while ((!bstop) && (i <= nblp))
    {
      f = nchild (i, l);
      if (STree_isConstantDecl (f))
	i++;
      else if (STree_isTypeDecl (f))
	i++;
      else
	bstop = pk_true;
    }
  if (!bstop)
    l = ntreeList_appendLast (l, vdecl);
  else
    ninsertright (i - 1, vdecl, l);
  return l;
}




/** It imports the declaration of an identifier (id) from a model (p) to an another one (m) as a local declaration.
    Not general: only type declaration identifier.
*/
static void
importDeclFromTo (ntrees id, ntrees p, ntrees m)
{
  ntrees el, elcopy;
  if (getDeclarationType (id, p, &el))
    {
      elcopy = STree_copy (el, pk_false);
      STree_addLocalDeclaration (m, elcopy);
    }
}

/**
   It removes the declaration of an identifier (id) from a model (p).
   Not general: only type declaration identifier.
*/
static void
removeDeclFrom (ntrees id, ntrees p)
{
  ntrees el;
  if (getDeclarationType (id, p, &el))
    ntreesDestroy (el);
}


/**
   @return r in which the identifiers of a list of declarations (e) are added.
*/
static ntrees
addIdsDeclaration (ntrees l, ntrees r)
{
  ntrees e, x;

  e = nchild (1, l);
  while (ISNOT_NILTREE (e))
    {
      if (noper (e) == opinit)
	x = nchild (1, e);
      else
	x = e;
      r = ntreeList_appendLast (r, STree_copy (x, pk_false));
      e = ntrees_getNext (e);
    }
  return r;
}

/**
   @return the list of the identifiers of a list of declarations (l).
*/
static ntrees
mkIdListFromDeclarationList (ntrees aListTree, noperators aListOp)
{
  ntrees e, r;
  if (IS_NILTREE (aListTree))
    return NILTREE;
  if (noper (aListTree) == opnil)
    return NILTREE;
  if (ntreearity (aListTree) == 0)
    return NILTREE;
  r = nmakemptylist (aListOp);
  e = nchild (1, aListTree);
  while (ISNOT_NILTREE (e))
    {
      r = addIdsDeclaration (nchild (2, e), r);
      e = ntrees_getNext (e);
    }
  return r;
}

/**
   @return  "true" if the identifier (id) is found as type in the local declarations of a model (p) and the 
   declaration is returned (vdecl), "false" otherwise.
*/
static boolean
getDeclarationType (ptident id, ntrees p, ntrees * vdecl)
{
  ntrees ldecl, el;
  noperators voperator;
  int i, n;
  boolean trouver;

  ldecl = STreeModel_getLocalDeclarations (p);
  if (IS_NILTREE (ldecl))
    return pk_false;
  n = ntreearity (ldecl);
  i = 1;
  trouver = pk_false;
  while (!trouver && i <= n)
    {
      el = nchild (i, ldecl);
      voperator = noper (el);
      switch (voperator)
	{
	case opltype:
	case opltypep:
	  trouver = is_in_listOfTypes (id, el);
	  break;
	  // not implemented
	  /* case opdeccon: case opdecconp: case opdecsig: case oplimpo:
	     case opact: case opactp: case opnode: case opnodep: case opprocessp: case opfuncp: 
	     case opprocess: case opfunc: case opoper: */
	}
      if (!trouver)
	i++;
    }
  if (!trouver)
    return pk_false;
  *vdecl = el;
  return pk_true;
}


/** Adding lists.
   @param lp (in/out) list of expressions(assumed)
   @param ls (in) list to add (assumed to be a list expression).
   @return lp + ls without copy of elements of ls (side effect on lp list).
*/
static ntrees
TreeAddList (ntrees lp, ntrees ls)
{
  int j, n;
  if (IS_NILTREE (lp))
    return ls;
  n = ntreearity (ls);
  for (j = 1; j <= n; j++)
    lp = ntreeList_appendLast (lp, nchild (j, ls));
  return lp;
}


static boolean
getIthDecl (ntrees aList, int rank, ntrees * s)
{
  int cpt, i, j;
  ntrees l;

  if (rank <= 0)
    return pk_false;
  if (IS_NILTREE (aList))
    return pk_false;
  if (noper (aList) == opnil)
    return pk_false;
  cpt = 0;
  for (i = 1; i <= ntreearity (aList); i++)
    {
      l = nchild (2, nchild (i, aList));
      for (j = 1; j <= ntreearity (l); j++)
	{
	  cpt++;
	  if (cpt == rank)
	    {
	      (*s) = nchild (j, l);
	      return pk_true;
	    }
	}
    }
  return pk_false;
}

/**
   It adds a process expression (vexprpro) to a process model (vtreeproc).
*/
static void
STree_addExpressionOfProcessToaModel_loc (ntrees vtreeproc, ntrees vexprpro, boolean bpost)
{
  ntrees vcorps, exprtree, p;
  int no;

  if (!STree_isEmptyList (vexprpro))
    {
      vcorps = nchild (3, vtreeproc);
      p = nchild (2, vcorps);
      if (IS_NILTREE (p))
	{
	  p = vcorps;
	  no = 2;
	  SET_NILTREE (exprtree);
	}
      else if (noper (p) == oppconf)
	{
	  exprtree = nchild (1, p);
	  no = 1;
	}
      else
	{			/* cas ou on a directement une expression de processus */
	  exprtree = p;
	  p = vcorps;
	  no = 2;
	}
      if (IS_NILTREE (exprtree))
	{
	  exprtree = INIT_COMPOSITION (vexprpro);
	  nchangchild (p, exprtree, no);
	}
      else
	{
	  if (noper (exprtree) == opobli)
	    exprtree = nchild (1, exprtree);
	  if (bpost)
	    exprtree = ntreeList_appendLast (exprtree, vexprpro);
	  else
	    exprtree = ntreeList_insertFirst (vexprpro, exprtree);
	}
    }
}


static boolean
getDeclIdentifierInList (ntrees id, ntrees ls, ntrees * vdecl)
{
  int j, m;
  boolean found;
  noperators op;
  ntrees curId;

  m = ntreearity (ls);
  found = pk_false;
  for (j = 1; j <= m && !found; j++)
    {
      curId = nchild (j, ls);
      op = noper (curId);
      if (op == opinit)
	curId = nchild (1, curId);
      else if (op == opcons)
	curId = nchild (1, curId);
      else if (op == optype)
	curId = nchild (1, curId);
      if (eq_idents (id, curId))
	{
	  found = pk_true;
	  (*vdecl) = curId;
	}
    }
  return found;
}

/**
   @return true if elem and vlist are the same sets, false otherwise.
*/
static boolean
STree_areSameSets (ntrees vlist, ntrees elem)
{
  int j, m, n;

  if (!STree_isList (vlist))
    {
      if (!STree_isList (elem))
	return STree_isTreeEq (elem, vlist);
      n = ntreearity (elem);
      if (n > 1)
	return pk_false;
      return STree_isTreeEq (nchild (1, elem), vlist);
    }
  n = ntreearity (vlist);
  if (!STree_isList (elem))
    {
      if (n > 1)
	return pk_false;
      return STree_isTreeEq (nchild (1, vlist), elem);
    }
  m = ntreearity (elem);
  if (n != m)
    return pk_false;

  for (j = 1; j <= n; j++)
    if (!STree_isInList (elem, nchild (j, vlist)))
      return pk_false;
  return pk_true;
}


/**
   xxxx;
*/
static ntrees
insertToList (ntrees v, ntrees e, noperators aListOp)
{
  ntrees t;
  if (noper (v) == aListOp)
    {
      ntreeList_appendLast (v, e);
      return v;
    }
  t = nmakemptylist (aListOp);
  ntreeList_appendLast (t, v);
  ntreeList_appendLast (t, e);
  return t;
}

static boolean
STreeListDecl_getId (ptident id, ntrees ldecl, noperators op, ntrees * vdecl)
{
  ntrees el;
  noperators voperator;
  int i, n;
  boolean found, b;

  if (IS_NILTREE (ldecl))
    return pk_false;
  n = ntreearity (ldecl);
  i = 1;
  found = pk_false;
  while (!found && i <= n)
    {
      el = nchild (i, ldecl);
      voperator = noper (el);
      b = pk_false;
      if (op == opnil)
	b = pk_true;
      else if (op == voperator)
	b = pk_true;
      else if ((op == opdeccon) && (voperator == opdecconp))
	b = pk_true;
      else if ((op == optype) && ((voperator == opltype) || ((voperator == opltypep))))
	b = pk_true;

      if (b)
	{
	  switch (voperator)
	    {
	    case opltype:
	    case opltypep:
	      found = getDeclIdentifierInList (id, el, vdecl);
	      break;
	    case opdecvar:
	    case opdecpar:
	    case opdecsig:
	    case opdeccon:
	      found = getDeclIdentifierInList (id, nchild (2, el), vdecl);
	      break;
	    case oplab:
	      found = getDeclIdentifierInList (id, el, vdecl);
	      break;
	    default:
	      break;
	    }
	}
      if (found)
	(*vdecl) = el;
      else
	i++;
    }
  return found;
}

/**
   The elements of lo are added to lp (if lp is empty, lo is returned).
   Assume: lo, lp are lists.
*/
ntrees
STreeList_conc (ntrees lp, ntrees lo)
{
  int n, i;

  if (IS_NILTREE (lo))
    return lp;
  if (IS_NILTREE (lp))
    return lo;
  n = ntreearity (lo);
  for (i = 1; i <= n; i++)
    ntreeList_appendLast (lp, nchild (i, lo));
  return lp;
}
