/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
   http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor: it creates an internal option manager (class OptionManager).
   Example: 
   \include main_m.c
*/
  BatchCompiler::BatchCompiler ()
{
  // Command analysis infos
  OM = new OptionManager ();
  recallMode = pk_false;
  this_commandRepstr = nilstr;
  Testing_reducingCommunications_b = pk_false;
  Testing_reducingCommunications_assClock = pk_false;
}

/**
   Destructor.
*/
BatchCompiler::~BatchCompiler ()
{
}

/**
   @return the option manager associated with the batch compiler.
*/
ptOptionManager
BatchCompiler::getOptionManager ()
{
  return OM;
}

/**
   It re-initializes the batch compiler.
*/
void
BatchCompiler::reset ()
{
  pKServicesAPI::reset ();
}

/** 
    It finalyzes the batch compiler : sends some statistics, message. 
*/
void
BatchCompiler::finalyze ()
{
  pK_messenger->CompilingTime ();
  pK_messenger->gmessend ();
  pKServicesAPI::finalyze ();
}

/**
  Application of a set of functionalities on a graph by testing of compiling options.
  API entry for graphical interface and batch compiler
  @param basicGraph  an existong DCG_basic graph
  @return a status code (See polychronyExitStatus.h file):
  - equal to polychronyNoError when the operation succeeds,
  - equal to polychronyTypeError when the program is not well type
  - equal to  polychronyCycle if the program contains circuits.
*/
integer
BatchCompiler::basic_compiling (ptDCG_basic basicGraph)
{
  integer status;
  ptDCG_poly polyGraph;
  boolean traceAll;

  traceAll = OM->req_traceALL ();
  setKeepTrace(basicGraph, OM->req_keeprew_opt ());

  /* ---- basicGraph: DCG_basic ----- */
  status = polychronyNoError;	/* No error */
  basicGraph->setVerboseMode (OM->req_verbose_opt ());
  pK_messenger->announce_BASIC ();

  basicGraph->transformAssertions2Events ();
  common_clean (basicGraph, pk_false);
  genCodeSignal (basicGraph, traceAll || OM->req_trace_basic (), "");
  if (OM->req_run_retiming_opt ())
    {				/* order with unification */
      basicGraph->moveIn2OutDelays ();
      common_clean (basicGraph, pk_false);
      genCodeSignal (basicGraph, traceAll, "retiming");
    }
  if (OM->req_kernel_opt ())
    {				/* signalSubstitution # kernel */
      basicGraph->expressionExpand ();
      genCodeSignal (basicGraph, traceAll, "sigExpand");
    }
  if (OM->req_ctrl2event_opt ())
    {				/* cannot be applied after -poly  */
      basicGraph->transformBooleanClocks2Events ();
      common_clean (basicGraph, pk_false);
      genCodeSignal (basicGraph, traceAll, "ctrl2event");
    }
  if (OM->req_multiple2singleDefinition_opt ())
    {				/* done when syndex; multiple2singleDefinition # single2multipleDefinition */
      groupPartialDefinitions (basicGraph);
      genCodeSignal (basicGraph, traceAll, "uniqueDef");
    }
  else if (OM->req_single2multipleDefinition_opt ())
    {				/* multiple2singleDefinition # single2multipleDefinition, single2multipleDefinition # syndex */
      splitDefaultDefinitions (basicGraph);
      genCodeSignal (basicGraph, traceAll, "multipleDef");
    }
  if (OM->req_normalDefault_opt ())
    {
      normalDefaults (basicGraph);
      genCodeSignal (basicGraph, traceAll, "normalDefault");
    }
  if (OM->req_splitDataStructure_opt ())
    {				/* done when Lustre */
      basicGraph->splitDataStructure ();	/* The structure/bundle types are removed. */
      genCodeSignal (basicGraph, traceAll, "splitStructure");
    }
  if (OM->req_enum2Integer_opt ())
    {
      basicGraph->changeEnumerated2Integers ();	// The enum types are removed (replaced by constants) 
      genCodeSignal (basicGraph, traceAll, "enum2Integer");
    }
  if (OM->req_wiz_testFunctionality_opt ())
    {
      if (basic_wizard_testFunctionality (basicGraph, &status))
	return status;
    }
  if (basicGraph->isEmptyGraph ())
    {
      pK_messenger->globalMessage (InfoEmptyGraph);
      return status;
    }

  if (OM->req_polyGraph ())
    {
      /* -- basicGraph:DCG_Basic --> polyGraph:DCG_Poly ----- WARNING basicGraph and polyGraph share the same GCD data structure */
      if (!DCGBasic2DCGPoly (basicGraph, OM->get_solvingDepth_opt (), &polyGraph))
	{
	  genCodeSignal (polyGraph, pk_true, "erroredPoly");
	  return polychronyInternalError;
	}
      return poly_compiling (polyGraph, traceAll);
    }

  /* ---- DCG_poly is not required thus one define 
     polyGraph a DCG_Basic completed with mandatory data structures ----- */
  polyGraph = new DCG_poly (basicGraph);	/* clock calculus is not processed */

  if (OM->req_run_booleanSeparation_opt ())
    {				/*  basic or poly TOCHECK Poly file generation */
      separateBooleansOthers (polyGraph);
      genCodeSignal (polyGraph, pk_true, "boolSeparation");
    }

  if (OM->req_export_z3z_opt ())
    {				/* basic or poly TODO code generators are exclusive, TOCHECK Poly file generation */
      SigaliPrintExtraction (polyGraph, traceAll);
      genCodeSignal (polyGraph, traceAll, "z3z");
    }

  if (OM->req_run_stateSeparation_opt ())
    {				/* basic or poly TOCHECK Poly file generation */
      separateStatesOthers (polyGraph);
      genCodeSignal (polyGraph, pk_true, "stateSeparation");
    }

  if (OM->req_export_dot_opt ())	/* can be used but may be useless in different paths */
    export_Dot (polyGraph);

  if (OM->req_run_abstraction_opt ())
    {				/* -spec: basic or poly */
      if (!InterfaceAbstraction (polyGraph))
	return polychronyAbstractionErr;
      genCodeSignal (polyGraph, pk_true, "abstraction");
      return polychronyNoError;
    }

  if (OM->req_export_profiling_opt ())
    {
      if (!applyTabularMorphism (polyGraph))
	return polychronyProfilingErr;
    }

  return polychronyNoError;
}

/** 
    It is the entry point for the Signal batch compiler.
    It transforms the Signal Textual form to a graph. 
    @return a status code (See polychronyExitStatus.h file)
    - equal to polychronyNoError when the graph is built.
    - different to polychronyNoError when an error occured.

    It calls SIGNALParseSrcfile() SIGNALParseParamfile() SIGNALTree2DCG() basic_compiling()
*/
integer
BatchCompiler::SignalBatchCompiling ()
{
  boolean cc1, cc2;
  ntrees termAbstractTree, paramAbstractTree;
  integer status;
  ptrepstr debug1, debug2;

  status = polychronyNoError;

  /* ===> log file initialisation */
  if (OM->is_outputDirectoryDefined ())
    sys_set_userOutputDir (string2repstr (OM->get_outputDirectory ()));
  pK_messenger->init_batchLogFile (this_commandRepstr);

  /* ===> command argument check */
  if (!OM->is_sourceFile_defined ())
    {				// no source file!!
      pK_messenger->error (ERR_NO_SRC_FILE);
      session_logFileClose (polychronyNoFile);
      return polychronyNoFile;
    }

  /* ===> output file initialisation */
  pK_SignalPrinter->setPrintOptionUnaryWhen (OM->req_gen_whenFormat_subopt ());
  initSignalTreeComments (this_commandRepstr);

  pK_messenger->announce_POLYCHRONY ();

  /* ===> source file input */
  if (OM->is_inIsf ())
    {				/* read Intermediate Signal Format */
      pK_messenger->announce_PARSING ();
      status = IntermediateFormat2AbstractTree (OM->getSourceFile (), &termAbstractTree);
      if (!(status == polychronyNoError))
	{
	  status = polychronySyntaxErrorProg;	/* PLG TO BE CHANGED */
	  session_logFileClose (status);
	  return status;
	}
    }
  else
    {				/* read input file */
      pK_messenger->announce_PARSING ();
      status = File2AbstractTree (OM->getSourceFile (), &termAbstractTree);
      if (!(status == polychronyNoError))
	{
	  session_logFileClose (status);
	  return status;
	}
    }

  /* ===> parameter file parsing */
  if (OM->req_readParamFile ())
    {
      if (OM->is_parameterFileDefined ())
	{
	  status = parameterFile2abstractTree (termAbstractTree, OM->get_parameterFile (), &paramAbstractTree);
	  if (!(status == polychronyNoError))
	    {
	      session_logFileClose (status);
	      return status;
	    }
	}
      else
	{
	  pK_messenger->globalMessageStr (codeNotYetImplemented, string2repstr ("automatic param file research"));
	  SET_NILTREE (paramAbstractTree);
	}
    }
  else
    SET_NILTREE (paramAbstractTree);
  /* ===> output directory modification */
  if (OM->is_outputDirectoryDefined ())
    {
      if (!(OM->is_targetDirectoryDefined ()))
	{
	  sys_setInner_userOutputDir (STreeModelModule_getNameStr (termAbstractTree));
	}
    }
  else
    sys_set_userOutputDir (STreeModelModule_getNameStr (termAbstractTree));

  /* ===> abstract syntax actions */
  /*      the output directory is set; do not change it */

  if (OM->req_outIsf ())
    unparse_SignalIsf (termAbstractTree);	/* DBUG */

  if (OM->req_listFile ())
    unparse_SignalLis (termAbstractTree, NILTREE, pk_false);

  if (OM->req_wiz_testFunctionality_opt ())
    if (syntax_wizard_testFunctionality (termAbstractTree, paramAbstractTree, &status))
      {
	session_logFileClose (status);
	return status;
      }
  /* ---> abstract syntax action: SME generation */
  if (OM->req_gen_sme_opt ())	// SIGNAL to SME translator.
    status = unparse_SME (termAbstractTree);

  /* ---> abstract syntax action: SSME generation */
  if (OM->req_gen_ssme_opt ())	// SIGNAL to SSME translator.
    signalTree2SSME (termAbstractTree);

  /* ---> abstract syntax action: process header generation */
  if (OM->req_gen_header_opt ())	// It exports the Signal interfaces defined in a module.
    status = exportHeader (termAbstractTree);

  if (OM->req_basicGraph ())
    {
      ptDCG DCG_ = NILDCG;
      setTraceabilityMode (OM->req_gen_cosimuTable_opt ());
      status = SIGNALTree2DCG (termAbstractTree, paramAbstractTree, OM->req_listFile (), &DCG_);
      if (status == polychronyNoError)
	status = basic_compiling (new DCG_basic (DCG_));
    }

  session_logFileClose (status);
  return status;
}

/**
  Analyzes the option and  suboptions of the batch compiler command line
  The options are in strArray, arc is the size of  strArray.
  The options are defined in polychronyOptions.h file.
  @return true if the arguments are correct and compatibles
*/
integer
BatchCompiler::analyzeCommandOptions (int argc, char *strArray[])
{
  this_commandIndex = 1;
  this_strArrayLength = argc;
  this_erroredParsing = pk_false;
  while (addArgument (strArray));
  /* every step can read one argument -xxx and an undefined number of sub arguments */
  if (this_erroredParsing)
    return polychronyOptionError;
  set_implicit_opt ();
  return polychronyNoError;
}

/**
   It initializes the batch compiler for a use from a GUI when GUICall, or
   as a command line otherwise.
*/
void
BatchCompiler::BC_initialize (pKContextType modeCall)
{
  pKServicesAPI::pK_initialize (modeCall);
}

/**
   Entry point of the batch compiler for the polychrony GUI and Pop (Polychrony on Polarsys, Eclipse).
*/
integer
BatchCompiler::SignalBatchCompilerGUIPop (ptrepstr defaultOutputDirName, ptDCG_basic basicGraph, char *voptions)
{
  int rc, nb;
  char *vopt[1024];

  pK_messenger->gmessload (systmess);
  pK_messenger->announce_POLYCHRONY ();
  string2ArgcArgv (voptions, &nb, vopt);	// produce argc, argv for the option command analyser of the batch compiler.
  OM->reset ();
  rc = analyzeCommandOptions (nb, vopt);
  if (rc == polychronyNoError)
    {
      sys_set_userOutputDir (nilstr);
      sys_setInner_userOutputDir (nilstr);
      if (OM->is_outputDirectoryDefined ())
	{
	  sys_set_userOutputDir (string2repstr (OM->get_outputDirectory ()));
	  if (!(OM->is_targetDirectoryDefined ()))
	    sys_setInner_userOutputDir (defaultOutputDirName);
	}
      else
	sys_set_userOutputDir (defaultOutputDirName);

      pK_messenger->gmessload (mkfgmess);
      rc = basic_compiling (basicGraph);
    }
  return rc;
}
