
/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#define eqStrings(s1,s2) strcmp(s1,s2)==0

void
BatchCompiler::initSignalTreeComments (ptrepstr command_line)
{
  ptrepstr vcmd, vPolychronyVersion;
  makeRepstr_repstr_string (pK_messenger->GlobalRecupmessStr (Comment_SIGNALVersion), sysGetSignalVersion (), &vPolychronyVersion);
  strconc2str (pK_messenger->GlobalRecupmessStr (Comment_commandline), command_line, pk_true, &vcmd);
  STree_initPolychronyComments (vPolychronyVersion, vcmd);
}

/** Entry point for command line reporting
 */
void
BatchCompiler::appendCommandLineSimpleOption (char *commandName)
{
  makeRepstr_repstr_string_char (this_commandRepstr, commandName, ' ', &this_commandRepstr);
}

/** Entry point for command line reporting
 */
void
BatchCompiler::appendCommandLineValuedOption (char *commandName, char *valueArgument)
{
  makeRepstr_repstr_string_char (this_commandRepstr, commandName, '=', &this_commandRepstr);
  makeRepstr_repstr_string_char (this_commandRepstr, valueArgument, ' ', &this_commandRepstr);
}

/**
   It applies pKServicesAPI::Sequentializing() functionality 
   with the current specific options of the code generation functionality.
*/
boolean
BatchCompiler::serializing (ptDCG_bool vgraph, ptDCG_sch * gSeq)
{
  boolean b = Serializing (vgraph, !OM->req_export_threads_subopt (), OM->req_force_subopt (), OM->req_export_check_subopt (), OM->req_run_delayMultiplexing_opt (), gSeq);
  return b;
}

/**
   Cycle detection operation.
   @param vgraph (in) : the graph on which the operation is applied.
   @param b (out) : at pk_true if the graph contains some clock constraints(equivalences not proved).
   @param balwaysCode (in) : must be pk_true when the sequentializing operation is asked even if the graph contains constraints.
   @param baddExceptionsForConstraints (in) : when "true", a node "exception" is added in the graph for each clock constraints.
   @return the value "true" if the graph can be sequentialized, "false" otherwise.
*/
boolean
BatchCompiler::isOrderRelation (ptDCG_bool vgraph, boolean * b, boolean balwaysCode, boolean baddExceptionsForConstraints)
{
  boolean res;
  (*b) = hasUnsolvedClockConstraints (vgraph);
  if (*b)
    {
      if (!balwaysCode)
	{
	  return pk_false;
	}
      if (baddExceptionsForConstraints)
	addExceptionsForConstraints (vgraph);
    }
  nonProductiveSignalElimination (vgraph);
  res = isStaticCycleFree (vgraph, pk_true, pk_true);
  return res;
}

/**
   It calls:
   - the restitution of a graph (vgraph) in Signal code (unparse()) if bPrettyPrint. 
      The file name is FOO_LEVEL_SUF_TRA.SIG where
          FOO is the name of the process (or is a user defined name)
          LEVEL is one of the strings "BASIC", "POLY",...
          SUF is optionSuffix value
	  TRA is the string "TRA"
   - the restitution of a graph (vgraph) in dot format (DCG2dotCode()) if the option is enable
*/
void
BatchCompiler::genCodeSignal (ptDCG_basic vgraph, boolean bPrettyPrint, const char *optionSuffix)
{
  if (bPrettyPrint) unparse (vgraph, optionSuffix);
  if (bPrettyPrint && OM->req_export_smt_opt ())  SMTPrintExtraction (vgraph, pk_true);
  if (bPrettyPrint && OM->req_gen_geneauto_opt ()) export_Geneauto(vgraph);
  if (bPrettyPrint && OM->req_export_dot_opt()) export_Dot(vgraph);
}

/**
   Apply common signal unification and signal substitution operations after some transformations.
   Trace the result if genTra is pk_true
 */
void
BatchCompiler::common_clean (ptDCG_basic basicGraph, boolean genTra)
{
  integer v;
  basicGraph->removeSignalEqualities ();
  if (OM->req_run_signalUnification_opt ())
    basicGraph->removeSignalClones ();
  if (OM->req_run_signalSubstitution_opt (&v))
    basicGraph->substituteExpr2Var (v);
}

/**
   It calls the C code generator for a given graph (vgraph), depending on compiling options (sequential, separated, distributed, threaded).
   @return "true" if the operation succeeds, "false" otherwise.
*/
boolean
BatchCompiler::optGencodeC (ptDCG_sch gSeq)
{
  ptCCodeGenerator aCCodeGenerator;

  aCCodeGenerator = new CCodeGenerator (gSeq);
  if (OM->req_export_distrib_subopt ())
    aCCodeGenerator->dist ();
  else if (OM->req_export_threads_subopt ())
    aCCodeGenerator->tasks ();
  else if (OM->req_run_abstraction_opt ())
    aCCodeGenerator->separate ();
  else
    aCCodeGenerator->seq (OM->req_export_c_main_opt (), OM->req_export_c_body_opt (), OM->req_export_c_io_opt ());
  delete aCCodeGenerator;
  return pk_true;
}

/**
   It calls the C++ code generator for a DCG_bool graph (vgraph) depending on compiling options (sequential, separated, distributed, threaded).
   @return "true" if the operation succeeds, "false" otherwise.
*/
boolean
BatchCompiler::optGencodeCPP (ptDCG_sch gSeq)
{
  ptCPPCodeGenerator aCPPCodeGenerator;

  aCPPCodeGenerator = new CPPCodeGenerator (gSeq);
  if (OM->req_export_distrib_subopt ())
    {
      aCPPCodeGenerator->dist ();
    }
  else if (OM->req_export_threads_subopt ())
    {
      aCPPCodeGenerator->tasks ();
    }
  else if (OM->req_run_abstraction_opt ())
    {
      aCPPCodeGenerator->separate ();
    }
  else
    {
      aCPPCodeGenerator->seq (OM->req_export_cpp_main_opt (), OM->req_export_cpp_body_opt (), OM->req_export_cpp_io_opt ());
    }
  delete aCPPCodeGenerator;
  return pk_true;
}

/**
   It calls the Java code generator for a DCG_sch graph (gSeq) depending on compiling options (sequential, separated, distributed, threaded).
   @return "true" if the operation succeeds, "false" otherwise.
*/
boolean
BatchCompiler::optGencodeJava ( ptDCG_sch gSeq)
{
  ptjavaCodeGenerator aJavaCodeGenerator;

  aJavaCodeGenerator = new javaCodeGenerator (gSeq);
  if (OM->req_export_distrib_subopt ())
    aJavaCodeGenerator->dist ();
  else if (OM->req_export_threads_subopt ())
    aJavaCodeGenerator->tasks (OM->req_export_java_main_opt (), OM->req_export_java_body_opt (), OM->req_export_java_io_opt ());
  else if (OM->req_run_abstraction_opt ())
    aJavaCodeGenerator->separate ();
  else
    aJavaCodeGenerator->seq (OM->req_export_java_main_opt (), OM->req_export_java_body_opt (), OM->req_export_java_io_opt ());
  delete aJavaCodeGenerator;
  return pk_true;
}


/**
   It calls the code generator (optGencodeC() for C , optGencodeCPP() for C++, optGencodeJava() for Java) for a scheduled graph (gSeq).
   @return "true" if the operation succeeds, "false" otherwise.
*/
boolean
BatchCompiler::callGenerators (ptDCG_sch gSeq)
{
  if (OM->req_export_c_modules_opt ()) return optGencodeC (gSeq);
  if (OM->req_export_cpp_modules_opt ()) return optGencodeCPP (gSeq);
  if (OM->req_export_java_modules_opt ()) return optGencodeJava (gSeq);
  return pk_true;
}



/**
   It calls the code generator (optGencodeC() for C , optGencodeCPP() for C++, optGencodeJava() for Java) for a graph (vgraph).
   @return "true" if the operation succeeds, "false" otherwise.
*/
boolean
BatchCompiler::callGenerators (ptDCG_bool vgraph)
{
  boolean requireSeq, b;
  ptDCG_sch gSeq;

  if (OM->req_run_abstraction_opt ()) vgraph->DontOptimizeSignals (vgraph->getOutputs ());
  requireSeq = OM->req_export_c_modules_opt () || OM->req_export_cpp_modules_opt () || OM->req_export_java_modules_opt () || OM->req_sequentialGraph ();
  b = pk_true;
  if (requireSeq)
    {
      b = serializing (vgraph, &gSeq);
      if (b) b = callGenerators(gSeq);
    }
  if (OM->req_run_abstraction_opt ()) unparseAbstraction (vgraph);
  return b;
}

void
BatchCompiler::genConfigFile (ptlDCG lsOfGraphs)
{
  ptCCodeGenerator aCCodeGenerator;

  if (OM->req_export_distrib_subopt ())
    {
      if (OM->req_export_c_modules_opt () || OM->req_export_cpp_modules_opt ())
	{
	  aCCodeGenerator = new CCodeGenerator ();
	  aCCodeGenerator->generateConfigFileDist (lsOfGraphs);
	  delete aCCodeGenerator;
	  pK_messenger->globalMessageStrStr (InfoRunConfigFile, string2repstr ("runConfig.h"), sys_get_userOutputDir ());
	}
      // to be completed for Java.
    }
}


/**
   It applies the basic_compiling(DCG_basic, char *) method for each graph of the list of graphs (lsOfGraphs)
   If OUTPUTDIR is the current directory, output files for each element FOO of the list are created in the subdirectory FOO of OUTPUTDIR
*/
integer
BatchCompiler::basic_compiling (ptlDCG lsOfGraphs)
{
  ptrepstr enteringOutputDir;
  ptident localGraphId;
  ptDCG localGraph;
  integer status;

  genConfigFile (lsOfGraphs);
  enteringOutputDir = sys_get_userOutputDir ();	// get the output directory path. Assumes that the output directory exists
  recallMode = pk_true;
  status = polychronyNoError;
  while ((status == polychronyNoError) && ptlPointer_get_elemNext (&lsOfGraphs, (pointer *) & localGraph))
    {
      sys_setInner_userOutputDir (ident2repstr (localGraph->getName ()));
      status = basic_compiling (new DCG_basic (localGraph, pk_false, pk_true));
      sys_restore_userOutputDir (enteringOutputDir);	// reset the output directory.Assumes that the directory exists
    }
  recallMode = pk_false;
  return status;
}

/**
   Called when the options -dist -syndex are used.
   
   It generates distributed code for a graph (vgraph) partitionned into a set of sub-graphs (lSubgraph).
   It is achieved by the following operations:
   - it produces a list of graphs (extractGraphsFromSubGraphs()) from the sub-graphs
   - then, it (re)calls the compiler on each new built graph (basic_compiling()).
*/
integer
BatchCompiler::export_distributedSyndex (ptDCG_bool vgraph, ptlINode lSubgraph)
{
  ptlDCG lsOfGraphs;

  if (OM->req_run_tagBasedSeparation_opt ())
    OM->switch_run_tagBasedSeparation_opt ();
  lsOfGraphs = extractGraphsFromSubGraphs (vgraph, lSubgraph);
  return basic_compiling (lsOfGraphs);
}

/**
   It (re)calls the compiler on each element of lsOfGraphs (basic_compiling()).
 */
integer
BatchCompiler::export_distributedCode (ptlDCG lsOfGraphs)
{
  boolean b1, b2, b3;
  integer status;

  if (OM->req_run_tagBasedSeparation_opt ()) OM->switch_run_tagBasedSeparation_opt ();
  // push the options
  b1 = OM->req_export_distrib_subopt ();
  b2 = OM->req_clustering_opt ();
  b3 = OM->req_export_threads_subopt ();

  OM->setON_export_distrib_subopt ();
  OM->setON_clustering_opt ();
  OM->setON_export_threads_subopt ();
  status = basic_compiling (lsOfGraphs);

  // pop the options
  if (!b1) OM->switch_export_distrib_subopt ();
  if (!b2) OM->switch_clustering_opt ();
  if (!b3) OM->switch_export_threads_subopt ();

  return status;
}



/**
   It generates distributed code for a graph (vgraph) partitionned into a set of sub-graphs (lSubgraph).
   It is achieved by the following operations:
   - it produces a list of graphs (extractGraphsFromSubGraphs()) from the sub-graphs
   - it adds the informations about the communication (communicationContainment()) between each extracted graphs.
   - then, it (re)calls the compiler on each new built graph (basic_compiling()).
*/
integer
BatchCompiler::export_distributedCode (ptDCG_bool vgraph, ptlINode lSubgraph)
{
  ptlDCG lsOfGraphs;

  lsOfGraphs = extractGraphsFromSubGraphs (vgraph, lSubgraph);
  vgraph->communicationContainment (lSubgraph, lsOfGraphs);
  return export_distributedCode(lsOfGraphs);

}

/**
   @return a list of new graphs, one graph associated with an element of a list of subgraph (lSubgraph) of a graph (vgraph).
   For each subgraph, the method, copySubgraph2Graph(), is applied.
*/
ptlDCG
BatchCompiler::extractGraphsFromSubGraphs (ptDCG_poly vgraph, ptlINode lSubgraph)
{
  ptINode aSubGraph;
  ptDCG_basic currentGraph;
  ptlDCG lsOfGraphs;
  ptDCG_bool gb;

  lsOfGraphs = NILLDCG;
  while (ptlINode_get_elem (&lSubgraph, &aSubGraph))
    {
      currentGraph = vgraph->copySubgraph2Graph (aSubGraph);
      ptlPointer_insert_elem (&lsOfGraphs, (pointer) (currentGraph->getDCG ()));
    }
  return lsOfGraphs;
}

/**
   Testing the reducing of the communications by adding containers.
   ?? to be moved in pKService ??
*/
integer
BatchCompiler::Testing_reducingCommunications (ptDCG_endo gEndo, boolean traceAll)
{
  ptlINode ln, lSubgraph;
  ptINode aSubGraph;

  // Testing_reducingCommunications_b/ Testing_reducingCommunications_assClock are managed in BatchCompilerWizard.
  // They are used for tests. TO BE CLEANED. LBesnard, Jul 2014.
  if (Testing_reducingCommunications_b)
    {
      if (!gEndo->partitionningRunOn (&lSubgraph))
	return polychronyInternalError;
      genCodeSignal (gEndo, traceAll, "preReduceComs");
      printSubGraphs (gEndo, lSubgraph);
      reducingCommunications (gEndo, lSubgraph, Testing_reducingCommunications_assClock);
      genCodeSignal (gEndo, traceAll, "postReduceComs");

      // To be conform with the code generation without optimisations, the subgraphs are removed. 
      ln = lSubgraph;
      while (ptlINode_get_elem (&ln, &aSubGraph))
	gEndo->subGraphFlattenning (aSubGraph);
      gEndo->buildClockProcessInterface ();
      // genCodeSignal(gEndo, traceAll, "cleaned");
    }
  return polychronyNoError;
}

/** 
    It applies the partitionning according to user specification (RunOn pragma).
    It applies several functionalities on the graph:
    - It computes the transitive closure of the graph (computeIOTransitiveClosure())
    - it produces the Boolean form of the graph.
    - then it partitions the graph using the RunOn pragma (partitionningRunOn())
    - recalls the compiler for each generated sub-graph (genCodeSubGraphs())

    After some functionalities, it prints the resulting graph in Signal text if bPrettyPrint.

    @return the status of the operation (polychronyNoError if the operation succeeds). 

*/
integer
BatchCompiler::PartitionningRunOn (ptDCG_poly polyGraph, boolean traceAll, ptDCG_bool * gBool, ptlINode * lSubgraph)
{
  boolean b;
  ptDCG_endo gEndo;
  ptINode aSubGraph;
  integer status;

  polyGraph->computeIOTransitiveClosure ();
  genCodeSignal (polyGraph, traceAll, "closure");

  if (!DCGPoly2DCGEndo (polyGraph, &gEndo))
    {
      return polychronyInternalError;
    }
  genCodeSignal (gEndo, traceAll, "");

  status = Testing_reducingCommunications (gEndo, traceAll);
  if (status != polychronyNoError)
    {
      return status;
    }

  if (!DCGEndo2DCGBool (gEndo, gBool))
    {
      return polychronyInternalError;
    }
  genCodeSignal (*gBool, traceAll, "");

  (*gBool)->nonProductiveSignalElimination ();
  (*gBool)->removeSignalClones ();
  (*gBool)->setClockDependences ();
  if (!isOrderRelation (*gBool, &b, OM->req_force_subopt (), pk_false))
    {
      return polychronyCycle;
    }
  genCodeSignal (*gBool, traceAll, "cycleFree");
  *lSubgraph = NILLINODE;
  if (!(*gBool)->partitionningRunOn (lSubgraph))
    {
      return polychronyInternalError;
    }
  genCodeSignal (*gBool, traceAll, "tagSeparation");
  printSubGraphs (*gBool, *lSubgraph);
  return polychronyNoError;
}

/**
   The graph is partitionned using the user "RunOn" pragma.
   @return the status of the operation (polychronyNoError if the operation succeeds). 
 */
integer
BatchCompiler::tagBased_partitionAndExport (ptDCG_poly polyGraph, boolean traceAll)
{
  boolean b;
  ptlINode lSubgraph;
  integer status;
  ptDCG_bool gBool;

  if (polyGraph->isModule ())
    {
      pK_messenger->globalMessageString (infoRestrictionModule1, "Tag partitionning");
      return -1;
    }

  status = polychronyNoError;
  b = polyGraph->hasUnsolvedClockConstraints ();
  if (!b || OM->req_force_subopt ())
    {
      status = PartitionningRunOn (polyGraph, traceAll, &gBool, &lSubgraph);
      if (status == polychronyNoError)
	{
	  if (OM->req_export_syndex_opt ())
	    export_distributedSyndex (gBool, lSubgraph);
	  if (OM->req_export_c_modules_opt ())
	    return export_distributedCode (gBool, lSubgraph);
	}
    }
  return status;
}

boolean
BatchCompiler::preGencode (ptDCG_bool gBool, boolean btrace, boolean gBoolAlreadyPrinted)
{
  boolean bx, balwaysCode;

  balwaysCode = OM->req_force_subopt ();
  if (!OM->req_run_utilityClock_opt ())
    if (!balwaysCode)
      {
	gBool->nonProductiveSignalElimination ();
	// genCodeSignal(gBool, btrace, "cleaned");
      }

  gBool->setClockDependences ();
  if (!isOrderRelation (gBool, &bx, balwaysCode, pk_true)) return pk_false;

  gBool->dependenceReinforcing ();	// internal dep for the arrays
  if (gBool->transformModelReferences2NodeReferences ())
    if (!isStaticCycleFree (gBool, pk_false, pk_true))
      {
	pK_messenger->globalMessage (ErrCycleAddedByReinforcing);
	return pk_false;
      }

  if (OM->req_run_delayOptimizing_opt ())
    gBool->setDelayMemoryOverloading (OM->req_clustering_opt ());
  if (OM->req_clustering_opt ())
    {				// Clusters
      gBool->nonProductiveSignalElimination ();
      gBool->breakCPG ();
      if (!isOrderRelation (gBool, &bx, balwaysCode, pk_false)) return pk_false;

      if (!gBool->phylumClustering ()) return pk_false;
      // Not fully implemented 
      gBool->nonProductiveSignalElimination ();
      genCodeSignal (gBool, btrace, "cluster");
      // verifying 
      if (!isOrderRelation (gBool, &bx, balwaysCode, pk_false))	return pk_false;
    }
  return pk_true;
}

integer
BatchCompiler::flat_compiling (ptDCG_flat flatGraph, boolean traceAll)
{
  integer status;
  boolean b;

  /* ---- flatGraph: DCG_flat ----- */
  status = polychronyNoError;	/* No error */
  pK_messenger->announce_FLAT ();
  genCodeSignal (flatGraph, traceAll || OM->req_trace_flat (), "");

  if (OM->req_wiz_testFunctionality_opt ())
    if (flat_wizard_testFunctionality (flatGraph, &status))
      {
	return status;
      }

  if (OM->req_export_syndex_opt ())
    {
      if (export_SynDEx (flatGraph, !recallMode && !OM->req_run_abstraction_opt ()))
	{
	  // export_SynDEx changes the DCG .  genCodeSignal(flatGraph, traceAll,"SynDEx");
	  return polychronyNoError;
	}
      genCodeSignal (flatGraph, pk_true, "SynDEx_error");
      return polychronyExportError;
    }

  if (OM->req_export_lustre_opt ())
    {
      if (export_Lustre (flatGraph))
	{
	  genCodeSignal (flatGraph, traceAll, "Lustre");
	  return polychronyNoError;
	}
      genCodeSignal (flatGraph, pk_true, "Lustre_error");
      return polychronyExportError;
    }

  if (OM->req_run_abstraction_opt ())
    {				/* -spec */
      /* should be compatible with -c,...  */
      if (!InterfaceAbstraction (flatGraph))
	{
	  return polychronyAbstractionErr;
	}
      genCodeSignal (flatGraph, pk_true, "abstraction");
    }

  if (preGencode (flatGraph, traceAll, traceAll))
    {
      if (callGenerators (flatGraph))
	genCodeSignal (flatGraph, traceAll || OM->req_trace_flat (), "seq");
      return status;
    }
  return polychronyInternalError;
}

integer
BatchCompiler::bool_compiling (ptDCG_bool boolGraph, boolean traceAll)
{
  integer status;
  boolean b;

  /* ---- boolGraph: DCG_bool ----- */
  status = polychronyNoError;	/* No error */
  pK_messenger->announce_BOOL ();
  genCodeSignal (boolGraph, traceAll || OM->req_trace_bool (), "");

  if (OM->req_wiz_testFunctionality_opt ())
    if (bool_wizard_testFunctionality (boolGraph, &status))
      return status;

  if (OM->req_flatGraph ())
    {
      ptDCG_flat gFlat;
      if (!DCGBool2DCGFlat (boolGraph, &gFlat))
	return polychronyInternalError;
      return flat_compiling (gFlat, traceAll);
    }

  /* ---- DCG_flat is not required for those actions ----- */
  if (OM->req_export_lustre_opt ())
    {
      if (export_Lustre (boolGraph))
	{
	  genCodeSignal (boolGraph, traceAll, "Lustre");
	  return polychronyNoError;
	}
      genCodeSignal (boolGraph, pk_true, "Lustre_error");
      return polychronyExportError;
    }

  if (OM->req_run_abstraction_opt ())
    {				/* -spec */
      /* should be compatible with -c,...  */
      if (!InterfaceAbstraction (boolGraph))
	{
	  return polychronyAbstractionErr;
	}
      genCodeSignal (boolGraph, pk_true, "abstraction");
    }

  if (preGencode (boolGraph, traceAll, traceAll))
    {
      if (callGenerators (boolGraph))
	{
	  genCodeSignal (boolGraph, traceAll || OM->req_trace_bool (), "");
	}
      return status;
    }
  return polychronyInternalError;
}

/**
   compiles an endochronous
   applies polychronous level transformations.  
*/
integer
BatchCompiler::endo_compiling (ptDCG_endo endoGraph, boolean traceAll)
{
  ptDCG_bool gBool;
  integer status;

  /* ---- endoGraph: DCG_endo ----- */
  status = polychronyNoError;	/* No error */
  pK_messenger->announce_ENDO ();
  genCodeSignal (endoGraph, traceAll || OM->req_trace_endo (), "");

  if (OM->req_wiz_testFunctionality_opt ())
    if (endo_wizard_testFunctionality (endoGraph, &status))
      return status;

  if (OM->req_boolGraph ())
    {
      if (!endoGraph->is_makeBool_safe ())
	{
	  if (OM->req_force_subopt ())
	    pK_messenger->printString
	      ("      WARNING: boolean clocks are added or substituted to event clocks to endow the process with boolean control\n");
	  else
	    {
	      pK_messenger->printString ("      WARNING: the process cannot be endowed with boolean control\n");
	      return status;
	    }
	}
      if (!DCGEndo2DCGBool (endoGraph, &gBool))
	return polychronyInternalError;
      return bool_compiling (gBool, traceAll);
    }

  /* ---- DCG_bool is not required for those actions ----- */
  if (OM->req_run_abstraction_opt ())
    {				/* -spec: basic or poly or endo or bool or sch */
      if (!InterfaceAbstraction (endoGraph))
	return polychronyAbstractionErr;
      genCodeSignal (endoGraph, pk_true, "abstraction");
      return polychronyNoError;
    }
}


/*
  Using distribution algorithms to compile a poly-endochronous program after adding of "rendez-vous" for placed constraints.
*/
integer BatchCompiler::polyEndoCompiling(ptDCG_poly gPoly)
{
  ptlDCG lsOfGraphs;
  ptDCG_bool gBool;
  ptlINode lSubgraph;
  int status;
  ptinsth vh, vhroot;
  ptpragma vpragma;

  // Defining the rendez-vous
  gPoly->addRendezVousForPlacedConstraints();
  genCodeSignal (gPoly, pk_true, "AfterRDV");
  gPoly->nonProductiveSignalElimination ();
  
  // Using the distributed code generation;
  vh = NILINSTH;
  int cpt = 1;
  while (gPoly->getNextFreeRoot(vh, &vhroot))
    {
      vpragma = newPredefinedPragma (PRAGMA_RUN_ON, NILOBJPRAGMA, repstr2PragmaStatement (integer2repstr(cpt)));
      gPoly->getDCG()->addPragmaToAllObjectsOfHierarchy(vhroot, vpragma);
      vh = vhroot; // Next step
      cpt++;
    }

  // Using the distributed code generation;
  OM->setON_run_tagBasedSeparation_opt ();
  OM->switch_polyendo_opt ();
  // OM->set_wiz_testFunctionality_opt (pk_false);
  status= PartitionningRunOn(gPoly, pk_true, &gBool, &lSubgraph);
  if (status == polychronyNoError)
    {
      if (OM->req_export_c_modules_opt ())
	status = export_distributedCode(gBool, lSubgraph);
    }
  else
    printf("BatchCompiler::polyEndoCompiling, status != polychronyNoError !!!\n");
  return status;
}

/**
   
*/
integer
BatchCompiler::PolyEndoChronous (ptDCG_poly polyGraph, boolean btrace)
{
  ptlINode lSubgraph;
  ptlDCG lsOfGraphs;
  ptDCG currentGraph;
  boolean b, reqController;
  integer status;

  status = polychronyNoError;
  // polyGraph->nonProductiveSignalElimination (); -- done before the call.
  if (polyGraph->isEndochronous())
    printf(" The graph is endochronous\n");
  else
    {
      b = polyGraph->isPolyEndoChronous( &reqController);
      genCodeSignal (polyGraph, btrace, "polyendo");      
      if (b)
	{
	  printf(" The graph is a polyendochronous\n");
	  status = polyEndoCompiling(polyGraph);
	}
      else
	{
	  printf(" The graph is NOT polyendochronous ...)\n");
	}
    }
  return status;
}


/**
  compiles a polychronous process, the clock of which has been achieved
  applies polychronous level transformations.  
*/
integer
BatchCompiler::poly_compiling (ptDCG_poly polyGraph, boolean traceAll)
{
  ptDCG_endo gEndo;
  integer status;
  integer ecLevel;
  boolean b1;

  /* ---- polyGraph: DCG_poly ----- */
  status = polychronyNoError;	/* No error */

  if (polyGraph->isEmptyGraph ())
    {
      pK_messenger->globalMessage (InfoEmptyGraph);
      return status;
    }
  pK_messenger->announce_POLY ();
  genCodeSignal (polyGraph, traceAll || OM->req_trace_poly (), "");


  if (OM->req_run_tagBasedSeparation_opt ())
    return tagBased_partitionAndExport (polyGraph, traceAll);	// Using the "RunOn" or Binding pragmas.



  if (OM->req_crewDefinition_opt ())
    {
      polyGraph->rewDefinitions ();
      genCodeSignal (polyGraph, traceAll, "bool2event");
    }

  { // added June 2016 LBesnard
    polyGraph->nonProductiveSignalElimination(); 
    // genCodeSignal (polyGraph, traceAll, "afterClean");
  }

  if (OM->req_run_utilityClock_opt ())
    {
      computeProductiveClock (polyGraph);
      rewUsingProductiveClock (polyGraph); //  each (partial) definition x:=E becomes x::=E when hu(x) | x::=defaultvalue x$1
      genCodeSignal (polyGraph, traceAll, "utlityOptimisation");
    }

  if (OM->req_stat_opt ()) /* clock calculus statistics */
    polyGraph->printCHStatistics ();

  if (OM->req_gen_exclusiveClock_opt (&ecLevel))
    {  /* generate a file FOO_clockExclusion.SIG that contains (| x^#y ... |) and the corresponding FOO_POLY_clockExclusion.SIG */
      polyGraph->computeClockExclusions (ecLevel);
      genCodeSignal (polyGraph, pk_true, "exlusion");
    }

  // can be used but may be useless in different paths
  /*  if (OM->req_export_dot_opt ()) export_Dot (polyGraph); */

  if (OM->req_run_booleanSeparation_opt ())
    { /*  basic or poly TOCHECK Poly file generation */
      separateBooleansOthers (polyGraph);
      genCodeSignal (polyGraph, pk_true, "boolSeparation");
    }

  if (OM->req_export_z3z_opt ())
    {  /* basic or poly TODO code generators are exclusive, TOCHECK Poly file generation */
      SigaliPrintExtraction (polyGraph, traceAll);
      genCodeSignal (polyGraph, traceAll, "z3z");
    }

  if (OM->req_run_stateSeparation_opt ())
    { /* basic or poly TOCHECK Poly file generation */
      separateStatesOthers (polyGraph);
      genCodeSignal (polyGraph, pk_true, "stateSeparation");
    }

  if (OM->req_export_profiling_opt ())
    if (!applyTabularMorphism (polyGraph))
      status = polychronyProfilingErr;

  b1 = polyGraph->hasUnsolvedClockConstraints ();
  genCodeSignal (polyGraph, b1, "constraint");

  if (OM->req_wiz_testFunctionality_opt ())
    if (poly_wizard_testFunctionality (polyGraph, &status))
      return status;


  if (OM->req_polyendo_opt())
    {
      return PolyEndoChronous(polyGraph, traceAll);
    }

  if (OM->req_endoGraph ())
    {
      if (OM->req_run_abstraction_opt ())	// -spec + ...
	{
	  if (!InterfaceAbstraction (polyGraph))
	    return polychronyAbstractionErr;
	  genCodeSignal (polyGraph, pk_true, "abstraction");
	}
      if (!polyGraph->isEndochronous ())
	if (OM->req_force_subopt ())
	  pK_messenger->printString ("      WARNING: boolean clocks are added to make the process endochronous\n");
	else
	  {
	    pK_messenger->printString ("      ERROR: the process is not endochronous use -force option\n");
	    return polychronyNotEndochronous;
	  }
      if (!DCGPoly2DCGEndo (polyGraph, &gEndo))
	return polychronyInternalError;
      return endo_compiling (gEndo, traceAll);
    }

  /* ---- DCG_endo is not required for those actions ----- */
  if (OM->req_run_abstraction_opt ())
    {
      if (!InterfaceAbstraction (polyGraph))
	return polychronyAbstractionErr;
      genCodeSignal (polyGraph, pk_true, "abstraction");
      return polychronyNoError;
    }
}
