/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** @return "true" if a list (vlist) is an empty list, "false" otherwise. */
extern boolean
ptlMark_satifies_isEmpty (ptlMark vlist)
{
  return (vlist == NILLMARK);
}


/** @return the size of a list (vlist). */
extern integer
ptlMark_get_size (ptlMark vlist)
{
  int nb;
  typeMark m;

  nb = 0;
  while (ptlMark_get_elemNext (&vlist, &m))
    nb++;
  return nb;
}

/** 
    @return "true" if a list (vlist) is not empty, "false" otherwise.
    @param vlist (in/out) the head of a list, it is sets to the next element (if it exists).
    @param mark (out)  the value of the head of the (in) list.
*/
extern boolean
ptlMark_get_elemNext (ptlMark * vlist, typeMark * mark)
{
  if ((*vlist) == NILLMARK)
    return pk_false;
  (*mark) = (*vlist)->elem;
  *vlist = (*vlist)->next;
  return pk_true;
}

/**
  It inserts an element (vmark) in a list of marks (deblist). The unicity is guaranteed.
  @return "false" if the element is yet in the list, "true" otherwise.
*/
extern boolean
ptlMark_insert_elem (ptlMark * deblist, typeMark vmark)
{
  boolean vres;
  typeMark cnum;
  ptlMark curlist, succ;

  if (ptlMark_satifies_isEmpty (*deblist))
    {
      (*deblist) = ptlMark_RIEN_create (vmark, NILLMARK);
      return pk_true;
    }

  cnum = ptlMark_get_num (*deblist);
  if (vmark < cnum)
    {
      *deblist = ptlMark_RIEN_create (vmark, *deblist);
      curlist = *deblist;
      return pk_true;
    }

  curlist = *deblist;
  while (pk_true)
    {
      if (cnum == vmark)
	return pk_false;
      succ = ptlMark_get_next (curlist);
      if (ptlMark_satifies_isEmpty (succ))
	{
	  succ = ptlMark_RIEN_create (vmark, NILLMARK);
	  curlist->next = succ;
	  return pk_true;
	}
      cnum = ptlMark_get_num (succ);
      if (cnum > vmark)
	{
	  succ = ptlMark_RIEN_create (vmark, succ);
	  curlist->next = succ;
	  return pk_true;
	}
      curlist = succ;
    }
  return vres;
}



/** It removes a mark (num) from a list (li).
    @return "true" if the element has been removed from the list, "false" otherwise.
*/
extern boolean
ptlMark_removeMark (ptlMark * li, typeMark num)
{
  boolean res, continuer;
  ptlMark linst, pred;
  typeMark cnum;

  if (ptlMark_satifies_isEmpty (*li))
    return pk_false;

  cnum = ptlMark_get_num (*li);
  if (num < cnum)
    return pk_false;
  if (num == cnum)
    {
      (*li) = (*li)->next;
      return pk_true;
    };

  pred = *li;
  linst = (*li)->next;
  res = pk_false;
  continuer = pk_true;
  while (continuer)
    {
      if (ptlMark_satifies_isEmpty (linst))
	continuer = pk_false;
      else if (ptlMark_get_num (linst) == num)
	{
	  continuer = pk_false;
	  res = pk_true;
	  pred->next = linst->next;
	}
      else
	{
	  pred = linst;
	  linst = linst->next;
	}
    }
  return res;
}


/**
   @return "true" if a mark (num) is in a list (li), "false" otherwise.
 */
extern boolean
ptlMark_IsIn (ptlMark li, typeMark num)
{
  boolean res, continuer;
  typeMark cnum;

  res = pk_false;
  continuer = pk_true;
  while (continuer)
    {
      if (ptlMark_satifies_isEmpty (li))
	continuer = pk_false;
      else
	{
	  cnum = ptlMark_get_num (li);
	  if (num <= cnum)
	    {
	      continuer = pk_false;
	      res = (num == cnum);
	    }
	  else
	    li = li->next;
	}
    }
  return res;
}

/** It tests the equality of two lists of marks (lmark1, lmark2).
    @return "true" if they are equal, "false" otherwise.*/
extern boolean
ptlMark_satisfies_equal (ptlMark lmark1, ptlMark lmark2)
{
  typeMark mark2, mark1;

  while (ptlMark_get_elemNext (&lmark1, &mark1))
    {
      if (!ptlMark_get_elemNext (&lmark2, &mark2))
	return pk_false;
      if (mark1 != mark2)
	return pk_false;
    }
  if (ptlMark_get_elemNext (&lmark2, &mark2))
    return pk_false;
  return pk_true;
}

/** It merges two lists of marks (lprinc,lsec). It modifies lprinc.
    @return lprinc.
*/
extern ptlMark
ptlMark_fusion (ptlMark lprinc, ptlMark lsec)
{
  typeMark elem;
  while (ptlMark_get_elemNext (&lsec, &elem))
    ptlMark_insert_elem (&lprinc, elem);
  return lprinc;
}

/** It adds a mark (mark) at the head of a list (pt).
    @return the new list.
*/
extern ptlMark
ptlMark_push (ptlMark pt, typeMark mark)
{
  return ptlMark_RIEN_create (mark, pt);
}

/** @return the list (pt) without its first element.
 */
extern ptlMark
ptlMark_pop (ptlMark pt)
{
  return pt->next;
}

/** @return the first mark of a list (pt).
 */
extern typeMark
ptlMark_top (ptlMark pt)
{
  return pt->elem;
}
