/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/*
  The following convention is applied:
  - the methods pri* guarantee the operation by printing the buffer if there is not enough place in the buffer;
  - the methods out* return "false" value when the text cannot be inserted in the buffer;
  - the methods wri* don't guarantee the operation.
*/

_PUBLIC_
/**
  Constructor.
  It creates a buffer with a maxima left margin (maxindent) and line size (lineSize)
  To use it, it must be opened (\ref open()).
*/
  buffer::buffer (integer leftMargin, integer lineSize)
{
  indent = ptlMark_push (NILLMARK, 0);
  tabul = ptlMark_push (NILLMARK, 0);
  curl = 0;
  setsize (leftMargin, lineSize);
  this_file = NILFIL;
  charEncodingMode = pk_true;
}

/** destructor.
 */
buffer::~buffer ()
{
}


/**
   @return the associated register.
*/
ptreg
buffer::getRegister ()
{
  return GETBUF_REGISTER ();
}

/**
   It defines the left margin (leftMargin) and the line size (lineSize) of the buffer.
*/
void
buffer::setsize (integer leftMargin, integer lineSize)
{
  maxl = leftMargin;
  maxr = lineSize;
}

void
buffer::newline ()
{
  sysscreenln ();
}

/**
   create a new buffering memory
   if filePathName is a defined value, 
      open the file filePathName in writing mode
   if filePathName is nil open for writing 
       - standard output if vcode = 0
       - graphical windows otherwise (vcode = 1,2)
*/
boolean
buffer::buffer_open (ptrepstr filePathName, integer vcode)
{
  if (CREATEBUF_ ())
    {
      if (filePathName != nilstr)
	{
	  this_file = sysopen (filePathName, pk_true);
	  LINKBUF_ ();
	}
      else
	this_file = sysopenttyout (vcode);
      return pk_true;
    }
  return pk_false;
}

/** close the current session of the buffer:
       the file, if any, is closed.
       the buffering memory is deleted
 */
void
buffer::buffer_close ()
{
  if (!(ISBUF_EMPTY ()))
    empty ();

  if (this_file != outgraph)
    if ((this_file == ttygraph) || (this_file == outfil))
      sysclosettyout (this_file);
    else
      sysclose (this_file);
  this_file = NILFIL;
  DELETEBUF_ ();
}

/** It inserts a character (c) in the buffer.  The insertion is not guaranteed.
 */
void
buffer::writeChar (char c)
{
  WRITEBUF_CHAR (c);
}


/** 
    It inserts l-times a character (c) in the buffer. The insertion is not guaranteed.
*/
void
buffer::writeNChars (char c, integer l)
{
  WRITEBUF_NCHAR (c, l);
}

/** It inserts the l-first characters of a string  (c) in the buffer. The insertion is not guaranteed.
 */
void
buffer::writeAlfa (alfa v, integer l)
{
  WRITEBUF_ALFA (v, l);
}

/** 
    It inserts the l-first characters of a string  (c) in the buffer. The insertion is guaranteed. 
*/
void
buffer::printAlfa (alfa v, integer l)
{
  fit (l);
  WRITEBUF_ALFA (v, l);
}


/** It inserts a string  (str) in the buffer. The insertion is not guaranteed.
 */
void
buffer::writeString (char *aString)
{
  WRITEBUF_STRING (aString);
}

/** It inserts a string  (str) in the buffer. The insertion is not guaranteed.
 */
void
buffer::writeRepstr (ptrepstr str)
{
  WRITEBUF_REPSTR (str);
}

/** @return "true" if the buffer is empty.
 */
boolean
buffer::isEmpty ()
{
  return (boolean) (ISBUF_EMPTY ());
}

/**
   @return "true" if the operation succeeds, "false" otherwise.
   It inserts a character (c) in the buffer.
*/
boolean
buffer::outChar (char c)
{
  if (isOver (1))
    return pk_false;
  return WRITEBUF_CHAR (c);
}

/**
   @return "true" if the operation succeeds, "false" otherwise.
   It inserts l-times a character (c) in the buffer.
*/
boolean
buffer::outNChars (char c, integer l)
{
  if (isOver (l))
    return pk_false;
  return WRITEBUF_NCHAR (c, l);
}

/**
   @return "true" if the operation succeeds, "false" otherwise.
   It inserts a string (str) in the buffer.
*/
boolean
buffer::outRepstr (ptrepstr str)
{
  if (isOver (strsize (str)))
    return pk_false;
  return WRITEBUF_REPSTR (str);
}


/**
   @return "true" if the operation succeeds, "false" otherwise.
   It inserts the text associated with a register (reg) in the buffer.
*/
boolean
buffer::outRegister (ptreg reg)
{
  if (isOver (nregGet_usedSize (reg)))
    return pk_false;
  return WRITEBUF_REG (reg);
}

/**
   It adds a space in the buffer if it is not empty and the last char is a space character.
*/
void
buffer::space ()
{
  integer i;

  i = GETBUF_SIZE ();
  if (i <= 0)
    return;
  if (GETBUF_CHAR (i) != ' ')
    if (!outChar (' '))
      empty ();
}


/** It increases of l characters, the tabulation position in the buffer.
 */
void
buffer::tabulation (integer l)
{
  integer i, j;

  tabul = ptlMark_push (tabul, getCurrentTabulValue () + l);
  i = GETBUF_SIZE ();
  j = getMaxIndentTabul ();
  if (i == 0)
    curl = j;
  else if (j > curl)
    {
      i = i + curl;
      if (j > i)
	{
	  if (j > maxl)
	    j = maxl;
	  WRITEBUF_NCHAR (' ', j - i);
	}
      else
	empty ();
    }
}

/** It sets the tabulation position to its previous value.*/
void
buffer::backTabulation ()
{
  if (tabul != NILLMARK)
    {
      tabul = ptlMark_pop (tabul);
      ladjust ();
    }
}



/** It sets a mark to the current position in the buffer + l.
    It is used for indentation.
*/
void
buffer::mark (integer l)
{
  indent = ptlMark_push (indent, curl + GETBUF_SIZE () + l);
  ladjust ();
}



/** It sets the mark to its previous value.
 */
void
buffer::backMark ()
{
  if (indent != NILINTSTACK)
    {
      indent = ptlMark_pop (indent);
      ladjust ();
    }
}

/** @return "true" if there is at least valent places in the buffer, "false" otherwise
 */
boolean
buffer::isOver (integer valent)
{
  integer i;
  i = GETBUF_SIZE () + valent;
  if (curl > maxl)
    return (boolean) (maxl + i > maxr);
  return (boolean) (curl + i > maxr);
}

/**
   It resets the buffer.
*/
void
buffer::reset ()
{
  RESETBUF_ ();
  ladjust ();
}

/** It writes the buffer in the associated file or window.
 */
void
buffer::empty ()
{
  if (this_file == outfil)
    if (curl > maxl)
      DISPLAYBUF_SCREEN (maxl);
    else
      DISPLAYBUF_SCREEN (curl);
  else if (this_file == outgraph)
    if (curl > maxl)
      DISPLAYBUF_GRAPH (maxl);
    else
      DISPLAYBUF_GRAPH (curl);
  else if (this_file != ttygraph)
    {
      if (curl > 0)
	if (curl > maxl)
	  sysprmcha (this_file, ' ', maxl);
	else
	  sysprmcha (this_file, ' ', curl);
      DISPLAYBUF_FILE ();
    }
  reset ();
}

/** It empties the buffer if there is not lg places in it.
 */
void
buffer::fit (integer lg)
{
  if (isOver (lg))
    empty ();
}

/** It inserts a character (c) in the buffer. The insertion is guaranteed. */
void
buffer::printChar (char c)
{
  fit (1);
  WRITEBUF_CHAR (c);
}

/** It inserts two characters (c1, c2) in the buffer. The insertion is guaranteed. */
void
buffer::print2Chars (char c1, char c2)
{
  fit (2);
  WRITEBUF_CHAR (c1);
  WRITEBUF_CHAR (c2);
}

/**
   It inserts l-times a character (c) in the buffer. The insertion is guaranteed.
*/
void
buffer::printNChars (char c, integer l)
{
  fit (l);
  WRITEBUF_NCHAR (c, l);
}



void
buffer::printXMLChar (unsigned char acar)
{
  unsigned int icar;
  char tabChar[4];

  icar = acar;
  if ((icar == 60)		// <  but '>' and '=' are allowed ((icar >= 60) && icar <=62)
      || icar < 14		// tab, return,...
      || icar > 126		// accented characters ...
      || icar == 34		// "
      || icar == 38		// &
    )
    {
      if (isOver (6))
	empty ();
      outChar ('&');
      outChar ('#');
      sprintf (tabChar, "%i", icar);
      outChar (tabChar[0]);
      if (tabChar[1] != '\0')
	{
	  outChar (tabChar[1]);
	  if (tabChar[2] != '\0')
	    outChar (tabChar[2]);
	}
      outChar (';');
    }
  else if (!outChar (acar))
    {
      empty ();
      outChar (acar);
    }
}

/**  It inserts a string (str) in the buffer taking into account the char encoding mode (XML or standard).
     The insertion is guaranteed.
     It is the only one method that manages the XML encoding (to be generalized ?).
 */
void
buffer::printString (const char *str)
{
  integer lg;
  integer i;
  unsigned char acar;


  lg = strlen (str);
  if (lg > 0)
    {
      fit (lg);
      if (charEncodingMode)
	for (i = 0; i < lg; i++)
	  {
	    acar = str[i];
	    if (!outChar (acar))
	      {
		empty ();
		outChar (acar);
	      }
	  }
      else
	for (i = 0; i < lg; i++)
	  printXMLChar (str[i]);
    }
}


/**  It inserts a string (str) in the buffer taking into account the char encoding mode (XML or standard).
     The insertion is guaranteed.
     It is the only one method that manages the XML encoding (to be generalized ?).
 */
void
buffer::printString (ptrepstr str)
{
  integer lg;
  integer i;
  unsigned char acar;


  lg = strsize (str);
  if (lg > 0)
    {
      fit (lg);
      if (charEncodingMode)
	for (i = 1; i <= lg; i++)
	  {
	    acar = strgetchar (str, i);
	    if (!outChar (acar))
	      {
		empty ();
		outChar (acar);
	      }
	  }
      else
	for (i = 1; i <= lg; i++)
	  printXMLChar (strgetchar (str, i));
    }
}

/**  It inserts a string (str) in the buffer taking into account the char encoding mode (XML or standard).
     The insertion is guaranteed.
     It is the only one method that manages the XML encoding (to be generalized ?).
 */
void
buffer::printInteger (long int intValue)
{
  printString (integer2repstr (intValue));
}

/**
   It sets to b the charEncodingMode :
   b must be "false" for XML, "true" otherwise.
   It is set to "true" in the constructor (default).
*/
void
buffer::setCharEncodingMode (boolean b)
{
  charEncodingMode = b;
}
