/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** @defgroup messglomanaging messenger:global managing.
    Global messenger managing methods.

    @defgroup messload  messenger: loading table of messages.
    Loading table of messages

    @defgroup messcode  messenger: managing error code condition.
    When an error is displayed, an internal error code is set.

   @defgroup messengerTime  messenger: cpu time information
   About time.

    @defgroup messageinfo   messenger: displaying information.
    In "silent mode", these methods are without effects.
    The following methods display message, they differ by the number of parameters 
    (it is the number of substitutions in the message).

    @defgroup glomessage   messenger: displaying global message.
    In "silent mode", these methods are without effects.
    The following methods display message of the global table ($SignalBin_data/messignal.0). 
    They differ by the number of parameters (it is the number of substitutions in the message).

    @defgroup messerror   messenger: displaying error message or annotating trees.
    In  "silent mode", these methods are without effects.
    The following methods 
       - display an error message of the current loaded table, 
       - or annotate a tree by the error.

    They differ by the number of parameters (it is the number of substitutions in the message).

    @defgroup messwarning   messenger: displaying/annotating warnings. 
    In "silent mode", these methods are without effects.
    The following methods display a warning message of the current loaded table, or annotate a tree by the selected message.
    They differ by the number of parameters (it is the number of substitutions in the message).

    @defgroup verbosemode   messenger: tracing operations in debug mode.
    The following methods are without effects if the debug mode is not activated.
    They are used to trace operations.

    @defgroup miscmessage   messenger: some misc methods

*/
/** @ingroup messglomanaging
    Constructor. It must be initialized (\ref gmessini()).
    \note A predefined object exists, called pK_messenger.
*/
  messenger::messenger ()
{
  option_stat = pk_false;
  option_warn = pk_false;
  option_verbose = pk_false;
  isParserLogFileOpen = pk_false;
  isLogFileOpen = pk_false;
}

/*  @ingroup messglomanaging
    Destructor. */
messenger::~messenger ()
{
}

/**  @ingroup messglomanaging
     Initializing.
     - Silent mode if off, messages are displayed.
     - Warning mode is off, they will not be printed.
     - Statistic mode (time informations) is off.
     - Debug mode is off.
     See setOptionMessenger() and setSilentModeMessagesTo() to modify this attributes.
     @return false when a memory allocation error occurs
*/
boolean messenger::gmessini ()
{
  int
    i;

  curtab = -1;
  oldtab = -1;
  beginAt = sysgetcputime ();
  xtime = 0;
  if (strregnew (&this_currentMessage))
    {
      setnulltab ();
      errorCounter = 0;
      annotnone = pk_false;
      for (i = (int) texte; i < (int) interne; i++)
	messcpt[i] = 0;
      silentMode = pk_false;
      gmessopen ();
      LogFileIni ();
      tabtoolsIni ();
      return pk_true;
    }
  return pk_false;
}

/**
   @ingroup messglomanaging
   - for code=1, it defines the "statistic mode"  to val (ie .when it is pk_true, time infos are visible).
   - for code=2, it defines the "warning mode" to val (ie when it is pk_true, warnings are visible).
   - for code=3, it defines the "debug mode" to val (ie when it is pk_true, tracing operations are enable).
*/
void
messenger::setOptionMessenger (int code, boolean val)
{
  switch (code)
    {
    case 1:
      option_stat = val;
      break;
    case 2:
      option_warn = val;
      break;
    case 3:
      option_verbose = val;
      break;
    }
}

/** @ingroup messglomanaging
    Set the silent mode to b.
    When it is "true", the warnings, errors are not displayed.
    Initialy, it is "false".
    @return the mode before the call.
*/
boolean
messenger::setSilentModeMessagesTo (boolean b)
{
  boolean x;
  x = silentMode;
  silentMode = b;
  return x;
}

/** @ingroup messglomanaging
    Finalyzing message manager. */
void
messenger::gmessend (void)
{
  LogFileDelete ();
  gmessclose ();
  strregfree (this_currentMessage);
}

/** @ingroup messload
    Loading  a table of messages (messignal.XXX) according newtab value.
    @return the index of the previous loaded table.
*/
ptindtab
messenger::gmessload (ptindtab newtab)
{
  if (newtab == curtab)
    return oldtab;
  oldtab = curtab;
  curtab = newtab;
  tabdiaload (signalMessages, curtab);
  return oldtab;
}

/** @ingroup messload
    Restoring a table (oldtab), value previously returned by \ref gmessload method.
*/
void
messenger::gmessrestore (ptindtab oldtab)
{
  if (oldtab != -1)
    gmessload (oldtab);
}


/** @ingroup messcode
    @return the value of the error code condition. */
boolean
messenger::getErrorCode (void)
{
  return cc;
}

/** @ingroup messcode
    Sets error code condition tag to scc value. 
*/
void
messenger::setErrorCode (boolean scc)
{
  cc = scc;
}

/**  @ingroup  messengerTime
     Printing compiling cpu-time.
*/
void
messenger::CompilingTime ()
{
  int vtime;

  if (!silentMode)
    if (option_stat)
      {
	vtime = (int) (sysgetcputime () - beginAt);
	printTime (vtime);
      }
}

/**  @ingroup  messengerTime
     Printing a compiling cpu-time (vtime).
*/
void
messenger::printTime (int vtime)
{
  diagprelwrite ();
  gmessTime (vtime);
  writeline ();
}

/** @ingroup messengerTime 
    Printing cpu-time from the previous call. 
*/
void
messenger::setcputime ()
{
  double x;

  if (!silentMode)
    {
      x = sysgetcputime ();
      gmessTime ((int) (x - xtime));
      xtime = x;
    }
}



/** @ingroup messageinfo
    It displays  the mnum-th information message.
 */
void
messenger::message (int mnum)
{
  if (!silentMode)
    {
      infostart (mnum);
      makemess0arg (pk_true);
      eom ();
    }
}

/** @ingroup messageinfo
    It annotates a tree (t) by the mnum-th information message.
 */
void
messenger::messageacc (int mnum, ntrees t)
{
  if (!silentMode)
    {
      diagprelannot (t);
      infoannotstart (t, mnum);
      makemess0arg (pk_true);
      eom ();
    }
}

/** @ingroup messageinfo
    It displays  the mnum-th information message, parametrized by a string (str). 
    @return the current cpu time.
*/
int
messenger::messagestr (int mnum, ptrepstr str)
{
  // printf("DEBUG ===== messenger::messagestr\n"); fflush(stdout);
  if (!silentMode)
    messgener (mnum, str);
  return ((int) sysgetcputime ());
}

/** @ingroup messageinfo
    It displays  the mnum-th information message, parametrized by anIdent. 
    @return the current cpu time.
*/
int
messenger::messageIdent (int mnum, ptident anIdent)
{
  return messagestr (mnum, ident2repstr (anIdent));
}

/** @ingroup messageinfo
    It displays  the mnum-th information message ; 
    this message is parametrized by str1 and str2. */
void
messenger::messagestr2 (int mnum, ptrepstr str1, ptrepstr str2)
{
  if (!silentMode)
    {
      infostart (mnum);
      makemess2arg (str1, str2);
      eom ();
    }
}

/** @ingroup messageinfo
    It displays  the mnum-th information message ; 
    this message is parametrized by aRepstr and anIdent. */
void
messenger::messagestrIdent (int mnum, ptrepstr aRepstr, ptident anIdent)
{
  messagestr2 (mnum, aRepstr, ident2repstr (anIdent));
}

/** @ingroup messageinfo
    It displays  the mnum-th information message ; 
    this message is parametrized by aRepstr and anIdent. */
void
messenger::messageStringString (int mnum, const char *string1, const char *string2)
{
  messagestr2 (mnum, string2repstr (string1), string2repstr (string2));
}

/**
   @ingroup messageinfo
   Send to the console the mnum-th information message;
   this message is parametrized by 3 strings str1, str2 and str3 */
void
messenger::messagestr3 (int mnum, ptrepstr str1, ptrepstr str2, ptrepstr str3)
{
  if (!silentMode)
    {
      infostart (mnum);
      makemess3arg (str1, str2, str3);
      eom ();
    }
}


/** @ingroup miscmessage
    @return the n-th entry of the "comment" message table.
    It is used for example by the code generators.
    The output string is inserted in the produced code with prefix-cooment and postfix-comment.
*/
ptrepstr
messenger::GlobalRecupmessStr (int n)
{
  ptrepstr vstr;

  savesetindpass ();
  vstr = getfullmess (texte, n);
  restoreindpass ();
  return vstr;
}

/**
   @ingroup glomessage
   It displays the numinfo-th message of the global table.
*/
void
messenger::globalMessage (int numinfo)
{
  if (!silentMode)
    {
      savesetindpass ();
      infostart (numinfo);
      makemess0arg (pk_true);
      eom ();
      restoreindpass ();
    }
}

/**
   @ingroup glomessage
   It displays the numinfo-th message of the global table with one parameter (str).
*/
void
messenger::globalMessageStr (int numinfo, ptrepstr str)
{
  if (!silentMode)
    {
      savesetindpass ();
      messgener (numinfo, str);
      restoreindpass ();
    }
}

/**
   @ingroup glomessage
   It displays the numinfo-th message of the global table with one identifier.
*/
void
messenger::globalMessageString (int numinfo, const char *aString)
{
  globalMessageStr (numinfo, string2repstr (aString));
}

/**
   @ingroup glomessage
   It displays the numinfo-th message of the global table with one integer value.
*/
void
messenger::globalMessageInt (int numinfo, int anInt)
{
  globalMessageStr (numinfo, integer2repstr (anInt));
}

/**
   @ingroup glomessage
   It displays the numinfo-th message of the global table with one identifier.
*/
void
messenger::globalMessageIdent (int numinfo, ptident anIdent)
{
  globalMessageStr (numinfo, ident2repstr (anIdent));
}

/**
   @ingroup glomessage
   It displays the numinfo-th message of the global table with  two parameters (str1,str2).
*/
void
messenger::globalMessageStrStr (int numinfo, ptrepstr str1, ptrepstr str2)
{
  if (!silentMode)
    {
      savesetindpass ();
      infostart (numinfo);
      makemess2arg (str1, str2);
      eom ();
      restoreindpass ();
    }
}


/** @ingroup messerror
    Send to the console the mnum-th error message.
*/
/**
   It prints the announce about Signal includig BDD infromation
*/
void
messenger::announce_POLYCHRONY ()
{
  /* Signal specific information lines */
  messagestr (polychronySignalToolbox_announce, string2repstr (sysGetSignalVersion ()));
  message (polychronyCopyright);
  /* BDD announce */
  message (usebddInfo);
  message (bddCopyright);
  message (announcesepline);
}

/*
 */
void
messenger::announce_PARSING ()
{
  globalMessage (PARSING_announce);
}

/*
 */
void
messenger::announce_FORMATPARSING ()
{
  globalMessage (FORMATPARSING_announce);
}

/*
 */
void
messenger::announce_BASIC ()
{
  globalMessage (BASIC_announce);
}

/*
 */
void
messenger::announce_POLY ()
{
  globalMessage (POLY_announce);
}

/*
 */
void
messenger::announce_ENDO ()
{
  globalMessage (ENDO_announce);
}

/*
 */
void
messenger::announce_BOOL ()
{
  globalMessage (BOOL_announce);
}

/*
 */
void
messenger::announce_FLAT ()
{
  globalMessage (FLAT_announce);
}

/*
 */
void
messenger::announce_MONO ()
{
  globalMessage (MONO_announce);
}

void
messenger::announce_SCHED ()
{
  globalMessage (SCHED_announce);
}


void
messenger::error (int mnum)
{
  if (!silentMode)
    {
      diagstarterror (mnum);
      makemess0arg (pk_true);
      eom ();
    }
}

/**  @ingroup messerror
     Send to the console the mnum-th error message parametrized by a string (str).
*/
void
messenger::errorStr (int mnum, ptrepstr str)
{
  if (!silentMode)
    {
      diagstarterror (mnum);
      makemess1arg (str);
      eom ();
    }
}


/**  @ingroup messerror
     Send to the console the mnum-th error message parametrized by a string (str).
*/
void
messenger::errorString (int mnum, const char *str)
{
  errorStr (mnum, string2repstr (str));
}

/** @ingroup messwarning
    It displays the mnum-th warning message, parametrized by an identifier (id).
*/
void
messenger::errorIdent (int mnum, ptident id)
{
  errorStr (mnum, ident2repstr (id));
}

/**  @ingroup messerror
     Send to the console the mnum-th error message parametrized by two strings (str1,str2).
*/
void
messenger::errorStrStr (int mnum, ptrepstr str1, ptrepstr str2)
{
  if (!silentMode)
    {
      diagstarterror (mnum);
      makemess2arg (str1, str2);
      eom ();
    }
}

/**  @ingroup messerror
     Send to the console the mnum-th error message parametrized by an ident and a repstr
*/
void
messenger::errorIdentStr (int mnum, ptident anIdent, ptrepstr anStr)
{
  errorStrStr (mnum, ident2repstr (anIdent), anStr);
}

/**  @ingroup messerror
     Sets the mnum-th error message as an annotation for the tree t.
     If the tree (t) is undefined, the message is sent to the console.
*/
void
messenger::errorTree (int mnum, ntrees t)
{
  if (!silentMode)
    {
      errorannotstart (t, pk_false, mnum);
      makemess0arg (pk_true);
      eom ();
    }
}

/**  @ingroup messerror
     Sets the mnum-th error message, parametrized by a string (str), as an annotation for the tree t.
     If the tree (t) is undefined, the message is sent to the console.
*/
void
messenger::errorTreeStr (int mnum, ntrees t, ptrepstr str)
{
  if (!silentMode)
    {
      errorannotstart (t, pk_false, mnum);
      makemess1arg (str);
      eom ();
    }
}


/**  @ingroup messerror
     Sets the mnum-th error message,  parametrized by an identifierstring (id), as an annotation for the tree t.
     If  the tree (t) is undefined, the message is sent to the console.
*/
void
messenger::errorTreeIdent (int mnum, ntrees t, ptident id)
{
  if (!silentMode)
    errorTreeStr (mnum, t, ident2repstr (id));
}

/**  @ingroup messerror
     Sets the mnum-th error message, parametrized by a string (str),  as an annotation for the tree t.
     When the tree refTo is defined, refTo is also annotated by a message.
     If the tree (t) is undefined, the message is sent to the console.
*/
void
messenger::errorTreeRefStr (int mnum, ntrees t, ntrees refTo, ptrepstr str)
{
  if (!silentMode)
    if (IS_NILTREE (refTo))
      errorTreeStr (mnum, t, str);
    else
      errref1str (mnum, t, refTo, str);
}

/**
   @ingroup messerror
   Similar to errorTreeRefStr() with an identifier (id) as parameter.
*/
void
messenger::errorTreeRefIdent (int mnum, ntrees p1, ntrees p2, ptident id)
{
  if (!silentMode)
    errorTreeRefStr (mnum, p1, p2, ident2repstr (id));
}


/** @ingroup messerror
    Sets the mnum-th error message, parametrized by a string (str), as an annotation for a tree (t).
    It also adds a reference to some trees (p2,p3) when they are defined.
    If the tree (t) is undefined, the message is sent to the console.
*/
void
messenger::errorTree2RefStr (int mnum, ntrees p1, ntrees p2, ntrees p3, ptrepstr str)
{
  if (!silentMode)
    if (IS_NILTREE (p2))
      errorTreeRefStr (mnum, p1, p3, str);
    else
      {
	errref1str (mnum, p1, p2, str);
	if (ISNOT_NILTREE (p3))
	  putmesscf (p3);
      }
}

/**  @ingroup messerror
     Similar to errorTree2RefStr() with an identifier (id) as parameter.
*/
void
messenger::errorTree2RefIdent (int mnum, ntrees p1, ntrees p2, ntrees p3, ptident id)
{
  if (!silentMode)
    errorTree2RefStr (mnum, p1, p2, p3, ident2repstr (id));
}


/** @ingroup messwarning
    It displays  the mnum-th warning message.
*/
void
messenger::warning (int mnum)
{
  if (!silentMode)
    {
      diagprelwrite ();
      diagstartwarn1 (mnum);
      makemess0arg (pk_true);
      eom ();
    }
}

/**
   @ingroup messwarning
    It displays  the mnum-th warning message, parametrized by a string (str).
*/
void
messenger::warningstr (int mnum, ptrepstr str, boolean bforce)
{
  if (!silentMode)
    {
      diagprelwrite ();
      diagstartwarn1 (mnum);
      makemess1arg (str);
      eom ();
    }
}

/**
   @ingroup messwarning
    It displays  the mnum-th warning message, parametrized by a string (str).
*/
void
messenger::warningstr (int mnum, ptrepstr str)
{
  if (option_warn)
    warningstr (mnum, str, pk_true);
}

/**
   @ingroup messwarning
    It displays  the mnum-th warning message, parametrized by two strings (str1, str2).
*/
void
messenger::warningstr2 (int mnum, ptrepstr str1, ptrepstr str2)
{
  if (!silentMode)
    if (option_warn)
      {
	diagprelwrite ();
	diagstartwarn1 (mnum);
	makemess2arg (str1, str2);
	eom ();
      }
}

/** @ingroup messwarning
    It displays the mnum-th warning message, parametrized by an identifier (id).
*/
void
messenger::warningident (int mnum, ptident id)
{
  if (!silentMode)
    warningstr (mnum, ident2repstr (id));
}

/** @ingroup messwarning
    It annotates a tree (t) by the mnum-th warning message even if the warning mode is disabled.
    If the tree (t) is undefined, the message is displayed.
*/
void
messenger::warningaccForce (int mnum, ntrees t)
{
  if (!silentMode)
    {
      diagprelannot (t);
      diagstartwarn1 (mnum);
      makemess0arg (pk_true);
      eom ();
    }
}

/** @ingroup messwarning
    If the warning mode is activated, it annotates a tree (t) by the mnum-th warning message.
    If the tree (t) is undefined, the message is displayed.
*/
void
messenger::warningacc (int mnum, ntrees t)
{
  if (!silentMode)
    if (option_warn)
      warningaccForce (mnum, t);
}

/** @ingroup messwarning
    If the warning mode is activated, it annotates a tree (t) by the mnum-th warning message, parametrized by a string (str).
    If the tree (t) is undefined, the message is displayed.
*/
void
messenger::warningaccstr (int mnum, ntrees t, ptrepstr str)
{
  if (!silentMode)
    if (option_warn)
      {
	diagprelannot (t);
	diagstartwarn1 (mnum);
	makemess1arg (str);
	eom ();
      }
}

/** @ingroup messwarning
    If the warning mode is activated, it annotates a tree (t) by the mnum-th warning message, parametrized by an identifier (id).
    If the tree (t) is undefined, the message is displayed.
*/
void
messenger::warningaccnom (int mnum, ntrees t, ptident id)
{
  if (!silentMode)
    warningaccstr (mnum, t, ident2repstr (id));
}

/**
   @ingroup verbosemode
*/
void
messenger::printStringDebug (ptrepstr vstr)
{
  if (!silentMode)
    if (option_verbose)
      globalMessageStr (traceMess, vstr);
}

/**
   @ingroup verbosemode
*/
void
messenger::printStringDebug (ptident vstr)
{
  printStringDebug (ident2repstr (vstr));
}



/** @ingroup verbosemode
    In verbose mode, it displays the numop-th message. 
    @return the current date.
*/
int
messenger::messageBeginOperVerbose (int numop)
{
  if (silentMode || !option_verbose)
    return -1;
  sntab (2);
  message (numop);
  return ((int) sysgetcputime ());
}

/** @ingroup verbosemode
    In verbose mode, it displays the current time - debtime.
    debtime is obtained by a previous call to messageBeginOperVerbose() method.
*/
void
messenger::messageEndOperVerbose (int debtime)
{
  if (silentMode || !option_verbose)
    return;
  messageEndOper (debtime);
  snbacktab ();
}

/** @ingroup verbosemode
    In verbose mode, it displays the  mnum-th message, parametrized by an int (val).
*/
void
messenger::messageIntVerbose (int mnum, int val)
{
  if (silentMode || !option_verbose)
    return;
  sntab (3);
  messageint (mnum, val);
}

/** @ingroup verbosemode
     In verbose mode, it displays the  mnum-th message, parametrized by a string (str).
     @return the current cpu time or -1 (not in verbose mode).
*/
int
messenger::tracing (int mnum, ptrepstr str)
{
  if (option_verbose && !silentMode)
    return messagestr (mnum, str);
  return -1;
}

/** @ingroup verbosemode
    In verbose mode, it displays the  mnum-th message, parametrized by  two strings (str1, str2).
*/
void
messenger::tracing (int mnum, ptrepstr str1, ptrepstr str2)
{
  if (option_verbose && !silentMode)
    messagestr2 (mnum, str1, str2);
}

/** @ingroup verbosemode
    In verbose mode, it displays the  mnum-th message, parametrized by three strings (str1, str2, str3).
*/
void
messenger::tracing (int mnum, ptrepstr str1, ptrepstr str2, ptrepstr str3)
{
  if (option_verbose && !silentMode)
    messagestr3 (mnum, str1, str2, str3);
}

/**
   @ingroup verbosemode
   In verbose mode, it displays the C-string message (message).
*/
void
messenger::printString (const char *message)
{
  if (!silentMode)
    globalMessageStr (traceMess, string2repstr (message));
}

/**
   @ingroup verbosemode
   In verbose mode, display aMessage
*/
void
messenger::printString (const ptrepstr aMessage)
{
  printString (repstr2string (aMessage));
}

/**
   @ingroup verbosemode
   In verbose mode, display anIdent
*/
void
messenger::printIdent (const ptident anIdent)
{
  printString (ident2string (anIdent));
}

/** @ingroup miscmessage
    @return the n-th entry of the "comment" message table.
    It is used for example by the code generators.
    The output string is inserted in the produced code with prefix-cooment and postfix-comment.
*/
ptrepstr
messenger::recupmessStr (int n)
{
  return getfullmess (texte, n);
}

/** @ingroup miscmessage
    It displays the INTERNAL ERROR message parametrized by the name of a function/method methodName) and the 
    filename (infile) where an exception occurrs.
    \note IT ABORTS AFTER THIS CALL !!!.
*/
void
messenger::errorMethodFile (const char *methodName, const char *infile)
{
  globalMessageStrStr (abort1, string2repstr (methodName), string2repstr (infile));
  PKExit ();
}

/** @ingroup miscmessage
    It displays the INTERNAL ERROR message parametrized by a line number (n) of the filename (infile) where an exception occurrs.
    \note IT ABORTS AFTER THIS CALL !!!
*/
void
messenger::Caseerror (int n, const char *infile)
{
  globalMessageStrStr (abort2, integer2repstr (n), string2repstr (infile));
  PKExit ();
}

/**
   @ingroup messglomanaging
   It initialises the Signal.log file with the "commandline" string.
*/
void
messenger::init_batchLogFile (ptrepstr optionLine)
{
  ptrepstr vstr, vFile;
  if (!isLogFileOpen)
    {
      sys_filePathInUserOutputDir (string2repstr ("Signal.log"), &vFile);
      BufferSignalLogFile = new buffer (0, 1024);
      if (BufferSignalLogFile->buffer_open (vFile, 0 /* 0->stdout */ ))
	{
	  isLogFileOpen = pk_true;
	  BufferSignalLogFile->printString ("Options: ");
	  BufferSignalLogFile->printString (optionLine);
	}
    }
}

/**
   It closes a Signal.log-session.
*/
void
messenger::session_logFileClose (int vstatus)
{
  ptrepstr vstr;
  int i;

  if (isLogFileOpen)
    {
      i = -1;
      if (vstatus == polychronyUnexpected)
	i = 1;
      else if (vstatus == polychronyOptionError)
	i = 2;
      else if (vstatus == polychronyNoArgs)
	i = 3;
      else if (vstatus == polychronyNoFile)
	i = 4;
      else if (vstatus == polychronyInitError)
	i = 5;
      else if (vstatus == polychronySyntaxErrorProg)
	i = 6;
      else if (vstatus == polychronySyntaxErrorParam)
	i = 7;
      else if (vstatus == polychronySolvingRefError)
	i = 8;
      else if (vstatus == polychronyTypeError)
	i = 9;
      else if (vstatus == polychronyCycle)
	i = 10;
      else if (vstatus == polychronyProfilingErr)
	i = 11;
      else if (vstatus == polychronyAbstractionErr)
	i = 12;
      else if (vstatus == polychronyArrayError)
	i = 13;
      else if (vstatus == polychronyInternalError)
	i = 14;
      else if (vstatus == polychronyNoError)
	i = 15;			// success

      if (i > 0)
	{
	  vstr = recupmessStr (i);
	  BufferSignalLogFile->empty ();
	  BufferSignalLogFile->printString (vstr);
	  BufferSignalLogFile->empty ();
	}
      BufferSignalLogFile->buffer_close ();
      isLogFileOpen = pk_false;
    }
}
