/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/* ------------------------------------------------------------------------------------
    messages management(errors, informations, log file).
    Messages can be tree annotations.
   ------------------------------------------------------------------------------------*/

#include <cstring>

/// PREDEFINED OBJECT OF THE VIRTUAL MACHINE.
messenger *pK_messenger = new messenger ();

_PRIVATE_
/* annotation mode : 
   - CST_STDOUT_MODE for standard output
   - CST_TREEANNOT_MODE for tree annotation. 
*/
#define CST_TREEANNOT_MODE   0
#define CST_STDOUT_MODE      1
// #define errsyntax 1
#define warningalfa "**WARNING "
#define warninglg 9
#define erroralfa "**ERROR "
#define errorlg 7
#define cfalfa " cf no- "
#define cflg 7
#define refalfa " (ref-  "
#define reflg 6
#define internalalfa "**BUG   "
#define internallg 5
#define signalMessages "messignal"
#define polychronyBeginCodeOp "/==> "
#define polychronyEndCodeOp "/< "
#define polychronyLineParser "/line "
#define polychronyColumnParser "/column "
#define polychronyErrorMessParser  "/error "
#define polychronyErrorMark '^'
/** The current message (mess) becomes an annotation for the current tree (annottree).
   These parameters must be set before.
*/
  void
messenger::annotline ()
{
  ptrepstr vstr;

  if (strregstore (this_currentMessage, &vstr))
    STree_addStrComment (annottree, vstr, pk_false);	// STree_setVMAnnotationsStr(annottree, vstr);
}

/** It prints the current message (this_currentMessage) ; this parameter must be set before.
 */
void
messenger::writeline ()
{
  ptrepstr vstr;

  if (strregstore (this_currentMessage, &vstr))
    {
      sigpristr (vstr);
      snline ();

      // Writing the same message in the Signal.log file.
      if (isLogFileOpen)
	{
	  BufferSignalLogFile->empty ();
	  BufferSignalLogFile->printString (vstr);
	}
    }
}

/** Sets the lenght of the current message (this_currentMessage) to lg characters. */
void
messenger::scratchline (int lg)
{
  strregreducesize (this_currentMessage, lg);
}


/** 
    It displays the current message (this_currentMessage) if it cannot accept l characters.
*/
void
messenger::messprifit (int l)
{
  if (nregGet_usedSize (this_currentMessage) + l > pagesize)
    switch ((int) annotmode)
      {
      case CST_TREEANNOT_MODE:
	annotline ();
	break;
      case CST_STDOUT_MODE:
	writeline ();
	break;
      default:
	Caseerror (currentLine, currentFile);
      }
}



/**
   It adds to the current message (mess) a charater (c).
*/
void
messenger::putmesschar (char c)
{
  messprifit (1);
  strregconccha (this_currentMessage, c);
}



/** 
    It displays, taking account the current mode (annotation, sttdout) , the current message (this_currentMessage).
*/
void
messenger::putmesseol (void)
{
  switch ((int) annotmode)
    {
    case CST_TREEANNOT_MODE:
      annotline ();
      break;
    case CST_STDOUT_MODE:
      writeline ();
      break;
    default:
      Caseerror (currentLine, currentFile);
    }
}



/**
   It adds to the current message a string (vstr).
*/
void
messenger::putmessstr (ptrepstr vstr)
{
  messprifit (strsize (vstr));
  strregconcstr (this_currentMessage, vstr, pk_false /* pk_true */ );	// modified  LBesnard March 2015
  // strregconccha (this_currentMessage, ' '); Removed LBesnard March 2015
}



/**
   It adds to the current message (this_currentMessage) an integer (n).
*/
void
messenger::putmessint (int n)
{
  putmessstr (integer2repstr (n));
}



/**
   It adds to the current message (this_currentMessage) the l first characters of b.
*/
void
messenger::putmessalfa (alfa b, int l)
{
  putmessstr (strmakewithalfa (b, l));
}



/**
   It adds to the current message (this_currentMessage) the l first characters of alfatype.
   And if numerote is pk_true, it inserts a reference to the value of an internal counter (errorCounter) to the message.
   This counter is incremented.
*/
void
messenger::putmessref (alfa alfatype, int lg, boolean numerote)
{
  putmessalfa (alfatype, lg);
  if (numerote)
    {
      putmessalfa (refalfa, reflg);
      errorCounter++;
      putmessint (errorCounter);
      putmesschar (')');
      putmesschar (':');
    }
  else
    putmesschar (':');
}

void
messenger::initMode (int mode)
{
  annotmode = mode;
  strregreset (this_currentMessage);
}

/** 
    It initializes the manager for annotating a tree (t) 
    or the standard output is the tree is undefined.
*/
void
messenger::diagprelannot (ntrees t)
{
  if ((IS_NILTREE (t)) || annotnone)
    diagprelwrite ();
  else
    {
      annottree = t;
      initMode (CST_TREEANNOT_MODE);
    }
}



/**
   It initializes the manager for displaying  a message on the standard output.
*/
void
messenger::diagprelwrite (void)
{
  initMode (CST_STDOUT_MODE);
}



/** 
    It initializes the manager for displaying the mnum-th information message on standard output.
*/
void
messenger::infostart (int mnum)
{
  diagprelwrite ();
  messcpt[(int) info] = messcpt[(int) info] + 1;
  initmess (info, mnum);
}

/**
   It initializes the manager for annotating a tree (t) by the mnum-th information message.
*/
void
messenger::infoannotstart (ntrees t, int mnum)
{
  diagprelannot (t);
  messcpt[(int) info] = messcpt[(int) info] + 1;
  initmess (info, mnum);
}



/**
   It initializes the manager for annotating a tree (t) by the mnum-th error message.
   if numerote, the message is precedeed by a reference message to the value of the error counter.
*/
void
messenger::errorannotstart (ntrees t, boolean numerote, int mnum)
{
  diagprelannot (t);
  messcpt[(int) err] = messcpt[(int) err] + 1;
  cc = pk_false;
  putmessref (erroralfa, errorlg, numerote);
  if (mnum >= 0)
    initmess (err, mnum);
}



/**
   It initializes the manager for displaying the num-th error message to the standard output.
*/
void
messenger::diagstarterror (int mnum)
{
  diagprelwrite ();
  messcpt[(int) err] = messcpt[(int) err] + 1;
  cc = pk_false;
  putmessref (erroralfa, errorlg, pk_false);
  initmess (err, mnum);
}



/**   It initializes the manager for displaying the num-th warning message to the standard output.
 */
void
messenger::diagstartwarn1 (int mnum)
{
  messcpt[(int) warn1] = messcpt[(int) warn1] + 1;
  putmessref (warningalfa, warninglg, pk_false);
  initmess (warn1, mnum);
}



/**   It initializes the manager for displaying the num-th internal  error message to the standard output.
 */
void
messenger::diagstartinte (int mnum)
{
  messcpt[(int) interne] = messcpt[(int) interne] + 1;
  cc = pk_false;
  putmessref (internalalfa, internallg, pk_false);
  initmess (interne, mnum);
}


/** 
    It  annotates a tree (t)  by a reference to another error (current value of the attribute member errorCounter).
 */
void
messenger::putmesscf (ntrees t)
{
  errorannotstart (t, pk_false, -1);
  putmessalfa (cfalfa, cflg);
  putmessint (errorCounter);
  putmesseol ();
}



/**
   It inserts in the current message (mess) the text of the message of the current loaded table and
   previously selected \ref initmess() method, until the first argument (an argument begins by a & character).
   @return "true" if such arg is found, "false" otherwise.
*/
boolean
messenger::putuntilarg (char *ccar)
{
  boolean cont;

  cont = pk_true;
  while (((*ccar) != 0) && cont)
    {
      if ((*ccar) != '&')
	putmesschar (*ccar);
      else
	{
	  *ccar = getmesschar ();
	  if ((*ccar) == '&')
	    putmesschar (*ccar);
	  else if ((*ccar) == '!')
	    putmesseol ();
	  else
	    cont = pk_false;
	}
      if (cont)
	*ccar = getmesschar ();
    }
  return (boolean) (!cont);
}



/**
    It builds in mess a message without arguments.
    Assumed: the message is pre-selected.
*/
void
messenger::makemess0arg (boolean bemit)
{
  char ccar;

  ccar = startmess ();
  while (putuntilarg (&ccar))
    ccar = getmesschar ();
  if (bemit)
    putmesseol ();
}

/**
   It builds in mess a message with an argument (vstr).
    Assumed: the message is pre-selected.
*/
void
messenger::makemess1arg (ptrepstr vstr)
{
  char ccar;

  ccar = startmess ();
  while (putuntilarg (&ccar))
    {
      putmessstr (vstr);
      ccar = getmesschar ();
    }
  putmesseol ();
}



/**
   It builds in mess a message with 2 arguments (vstr1, vstr2).
   Assumed: the message is pre-selected.
*/
void
messenger::makemess2arg (ptrepstr vstr1, ptrepstr vstr2)
{
  char ccar;

  ccar = startmess ();
  while (putuntilarg (&ccar))
    {
      if (ccar == '1')
	putmessstr (vstr1);
      else
	putmessstr (vstr2);
      ccar = getmesschar ();
    }
  putmesseol ();
}


/**
   It builds in mess a message with 3 arguments (vstr1, vstr2, vstr3).
   Assumed: the message is pre-selected.
*/
void
messenger::makemess3arg (ptrepstr vstr1, ptrepstr vstr2, ptrepstr vstr3)
{
  char ccar;

  ccar = startmess ();
  while (putuntilarg (&ccar))
    {
      if (ccar == '1')
	putmessstr (vstr1);
      else if (ccar == '2')
	putmessstr (vstr2);
      else
	putmessstr (vstr3);
      ccar = getmesschar ();
    }
  putmesseol ();
}

/**
   It initiliazes for an annotation of tree (t).
   It builds in mess a error message (mn) with an argument (str).
   The message is completed by a reference to another tree (t2).
*/
void
messenger::errref1str (int mn, ntrees t, ntrees t2, ptrepstr str)
{
  errorannotstart (t, pk_true, mn);
  makemess1arg (str);
  putmesscf (t2);
}

/**
   It adds to the current message (mess) an integer (val) followed by a character (car).
*/
void
messenger::putmesstimeunit (int val, char car)
{
  if (val < 10)
    putmesschar ('0');
  putmessint (val);
  putmesschar (' ');
  putmesschar (car);
  putmesschar (' ');
}

/**
   Convertion from vtime-integer in hour (hh), minutes(mn), seconds(sec), hundreth of seconds (cs).
*/
void
messenger::intTimeConversion (int vtime, int *hh, int *mn, int *sec, int *cs)
{
  int ts, tm;

  ts = vtime / 100;
  tm = ts % 3600;
  *hh = ts / 3600;
  *mn = tm / 60;
  *sec = tm % 60;
  *cs = vtime % 100;
}


/**
   It adds in the current message (mess) a integer (vtime) converted in hours/mn/sec.
*/
void
messenger::gmessTime (int vtime)
{
  int hh, mn, sec, cs;

  intTimeConversion (vtime, &hh, &mn, &sec, &cs);
  putmessalfa ("        ", 8);
  putmessalfa ("time   :", 8);
  if (hh > 0)
    putmesstimeunit (hh, 'h');
  if (mn > 0)
    putmesstimeunit (mn, 'm');
  putmesstimeunit (sec, 's');
  putmesstimeunit (cs, ' ');
}

/**
   It displays, on the standard output, the end of an operation message began at debtime time.
*/
void
messenger::messageEndOper (double debtime)
{
  savesetindpass ();
  infostart (numinfoEndoper);
  if (debtime > 0)
    {
      makemess0arg (pk_false);
      gmessTime ((int) (sysgetcputime () - debtime));
      putmesseol ();
    }
  else
    makemess0arg (pk_true);
  eom ();
  restoreindpass ();
}

/**
   It displays, on the standard output, the numinfo-th information message, parametrized by a string (vstr).
*/
void
messenger::messgener (int numinfo, ptrepstr vstr)
{
  infostart (numinfo);
  makemess1arg (vstr);
  eom ();
}

/**
   It displays, on the standard output, the nmess-th information message, parametrized by an integer (val).
*/
void
messenger::messageint (int nmess, int val)
{
  messgener (nmess, integer2repstr (val));
}

/** It finished the printing of a message.
    (a new line is sent on the standard output)
 */
void
messenger::eom ()
{
  if (annotmode == CST_STDOUT_MODE)
    newline ();
}

/**
   It Initialises an output session: it links the standard output to a buffer (internal attribute member).
   @return "true" if the operation succeeds, "false" otherwise.
*/
boolean
messenger::gmessopen (void)
{
  messbuf = new buffer (80, 140);
  inteprint_open_stdOut (messbuf, 0);
  return pk_true;
}



/** Finalyze the output session.
 */
void
messenger::gmessclose (void)
{
  intpriclose ();
}

/**
   Aborting the current execution (called when an internal error is detected).
 */
void
messenger::PKExit ()
{
  exit (1);
}


/**
   Printing the log-file.
*/
void
messenger::LogPutAndClose ()
{
  ptrepfil f;
  char c;
  boolean stop;

  f = sysopen (vParserLogFile, pk_false);
  stop = pk_false;
  while (!stop)
    {
      if (syseoline (f))
	{
	  LogFileFilterPrint ();
	  sysgetcha (f, &c);
	}
      if (sysgetcha (f, &c))
	aBuffer->printChar (c);
      else
	stop = pk_true;
    }
  sysclose (f);
  aBuffer->buffer_close ();
}

/**
   It initializes a log-session for writing in a file (POLYCHRONY_STDERR.log).
 */
void
messenger::LogFileIni ()
{
  // Parser
  vParserLogFile = string2repstr ("POLYCHRONY_STDERR.log");
  aBuffer = new buffer (0, 1024);
  isParserLogFileOpen = pk_false;
  prefixBeginCodeop = string2repstr (polychronyBeginCodeOp);
  prefixEndCodeop = string2repstr (polychronyEndCodeOp);
  // Signal.log
  isLogFileOpen = pk_false;
}

/**
   It finalyzes a log-session previously opened by LogFileIni() method.
*/
/* 
void 
messenger::LogFileFinalyze()
{
  parsing_logFileClose();
}
*/

/**
   It deletes the log-session file previously opened by LogFileIni() method.
*/
void
messenger::LogFileDelete ()
{
  sysdelete (vParserLogFile);
}


/**
   It displays the log-file associated with the current log-session, filtering some 
   predefined marks.
*/
void
messenger::LogFileFilterPrint ()
{
  boolean bprint;
  ptreg buf;

  // polychronyBeginCodeOp/polychronyEndCodeOp are filtered
  if (!aBuffer->isEmpty ())
    {
      buf = aBuffer->getRegister ();
      if (strregTestHead (buf, prefixBeginCodeop))
	bprint = pk_false;
      else
	bprint = !strregTestHead (buf, prefixEndCodeop);
      if (bprint)
	{
	  aBuffer->empty ();
	  aBuffer->newline ();
	}
      else
	aBuffer->reset ();
    }
}
