/**
This module is the Polychrony tool set interface to external bdd package.
All fucnctions defined here must be defined for an other BDD implementation
It is written as a front end to Berkeley BDD package.
This interface requires the definition of (see export/BDDPackApi_type.h)
     - bdd_type seen as (void *) in Polychrony, to denote a BDD
     - bdd_id_boolean_variable_type seen as UnsignedInteger in Polychrony, to denote a boolean variable
     - bdd_rank_in_order_type seen as UnsignedInteger in Polychrony, to denote the rank of a boolean variable in the current variable order
     - UnsignedInteger is an unsigned integer the length of which is defined in export/BDD_StdTypes.h depending on the target architecture
*/

/** BDD Initialize
   Initialize the bdd manager with the maximal number of variables.
   It is necessary for some BDD package (called with the value 0 currently).
*/
extern void
pK_bdd_init (int max_var)
{
  manager = bdd_start (max_var);
  max_id = 0;
  id_server = 0;
}

/** BDD Finalize

   @return none.
*/
extern void
pK_bdd_end (void)
{
  bdd_end (manager);
  max_id = 0;
}

/** BDD Formulas. BDD constant true.

   @return the representation of the bdd BDD constant true.
*/
extern bdd_type
pK_bdd_one ()
{
  return (bdd_type) bdd_one (manager);
}

/** BDD Formulas. BDD constant false.

   @return the representation of the BDD constant false.
*/
extern bdd_type
pK_bdd_zero ()
{
  return (bdd_type) bdd_zero (manager);
}

/** BDD Formulas. BDD AND

   @param first (in) a bdd.
   @param second (in) a bdd.
   @return the unique bdd that denotes the conjonction of first and second.
*/
extern bdd_type
pK_bdd_and (bdd_type first, bdd_type second)
{
  boolean phase_first;
  boolean phase_second;

  phase_first = pk_true;
  phase_second = pk_true;
  return (bdd_type) bdd_and ((bdd_t_pointer) first, (bdd_t_pointer) second, phase_first, phase_second);
}

/** BDD Formulas. BDD OR

   @param first (in) a bdd.
   @param second (in) a bdd.
   @return the unique bdd that denotes the disjonction of first and second. 
*/
extern bdd_type
pK_bdd_or (bdd_type first, bdd_type second)
{
  boolean phase_first;
  boolean phase_second;

  phase_first = pk_true;
  phase_second = pk_true;
  return (bdd_type) bdd_or ((bdd_t_pointer) first, (bdd_t_pointer) second, phase_first, phase_second);
}

/** BDD Formulas. BDD NOT

   @param arg (in) : a bdd .
   @return the unique bdd that denotes the negation of arg.
*/
extern bdd_type
pK_bdd_not (bdd_type arg)
{
  return (bdd_type) bdd_not ((bdd_t_pointer) arg);
}

/** BDD operation - equality test-.

   @param first (in) a bdd.
   @param second (in) a bdd.
   @return "true" if the two bdds are equals, false otherwise.
*/
extern boolean
pK_bdd_equal (bdd_type first, bdd_type second)
{
  return (boolean) bdd_equal ((bdd_t_pointer) first, (bdd_t_pointer) second);
}

/** BDD operation - cofactor -.

   @param  first (in) a bdd.
   @param  second (in) a bdd.
   @return the cofactor first by second.
*/
extern bdd_type
pK_bdd_cofactor (bdd_type first, bdd_type second)
{
  return (bdd_type) bdd_cofactor ((bdd_t_pointer) first, (bdd_t_pointer) second);
}

/** BDD access -then-side- .

   @param  arg (in) a bdd.
   @return the then-side of bdd operand.
*/
extern bdd_type
pK_bdd_then (bdd_type arg)
{
  return (bdd_type) bdd_then ((bdd_t_pointer) arg);
}

/** bdd access -else-side-.

   @param  arg (in) a bdd.
   @return the else-side bdd of the bdd operand.
*/
extern bdd_type
pK_bdd_else (bdd_type arg)
{
  return (bdd_type) bdd_else ((bdd_t_pointer) arg);
}

/** bdd access - top variable-.

   @param arg (in)a bdd.
   @return the (new) bdd of the top variable of the bdd operand.
*/
extern bdd_type
pK_bdd_topvar (bdd_type arg)
{
  return (bdd_type) bdd_top_var ((bdd_t_pointer) arg);
}

/** BDD Formula Management:duplicate.

   @param  bdd  (in) : a  bdd descriptor.
   @returns a copy of the given bdd.
*/
extern bdd_type
pK_bdd_dup (bdd_type bdd)
{
  return (bdd_type) bdd_dup ((bdd_t_pointer) bdd);
}

/** BDD Formula Management: free.
   This function frees the given bdd.
   @return none.
*/
extern void
pK_bdd_free (bdd_type bdd)
{
  bdd_free ((bdd_t_pointer) bdd);
}

/** BDD Variable Allocation.
   It links boolean variable/bdd.
   @param id (in) a boolean variable of the bdd "package".
   @return the bdd of the variable id.
*/
extern bdd_type
pK_bdd_get (bdd_id_boolean_variable_type id)
{
  UnsignedInteger i;

  bdd_t_pointer result;
  if (id < max_id)
    return (bdd_type) bdd_get_variable (manager, (UnsignedInteger) id);
  for (i = max_id; i <= id; i++)
    result = bdd_create_variable (manager);
  max_id = id + 1;
  return (bdd_type) result;
}

/**
  Access to boolean variable of a bdd.

 @param  bdd (in) : a bdd.
 @return the first variable of the bdd operand. 
*/
extern bdd_id_boolean_variable_type
pK_bdd_topvar_id (bdd_type bdd)
{
  /* WARNING : access to the structure of a bdd. */
  // return (bdd_id_boolean_variable_type) BDD_REGULAR(((bdd_t_pointer ) bdd)->node)->id;
  return (bdd_id_boolean_variable_type) topvar_id ((bdd_t_pointer) bdd);
}

/**
  Access to boolean variable of a bdd.

 @param  bdd (in): a bdd.
 @return the last variable of the bdd operand.
*/
extern bdd_id_boolean_variable_type
pK_bdd_lastvar_id (bdd_type bdd)
{
  /* WARNING : access to the structure of a bdd. */
  // return (bdd_id_boolean_variable_type) BDD_REGULAR(((bdd_t_pointer ) bdd)->node)->last;
  return (bdd_id_boolean_variable_type) lastvar_id ((bdd_t_pointer) bdd);
}

/** Order on variables.

   @return the first element for the order.
*/
extern bdd_rank_in_order_type
pK_bdd_get_order_first ()
{
  return (bdd_rank_in_order_type) 0;
}

/** Order on variables.

 @return the element following ord th.
*/
extern bdd_rank_in_order_type
pK_bdd_get_order_next (bdd_rank_in_order_type ord)
{
  return (bdd_rank_in_order_type) ((UnsignedInteger) ord + 1);
}

/**    Order on variables. 
      It Assigns the rank of boolean variable.
      @param id (in) : a boolean variable identifier of a bdd.
      @param ord (in): a rank number.
      @return none.
*/
extern void
pK_bdd_set_ord (bdd_id_boolean_variable_type id, bdd_rank_in_order_type ord)
{
  bdd_set_ord (id, (UnsignedInteger) ord);
}



/** 
    new bdd boolean variable.
    @return a new descriptor of boolean variable. 
*/
extern bdd_id_boolean_variable_type
pK_bdd_new_id ()
{
  /* append by l Besnard */
  return id_server++;
}
