
/** List constructor 
   @return a new empty list 
*/
extern list_type
list_create (void)
{
  list_type result;
  result = (list_type) malloc (sizeof (list_struct));
  ASSERT_MALLOC_OK (result);
  result->current = NILITEM;
  result->first = NILITEM;
  result->last = NILITEM;
  return result;
}

/** list Destructor.
    @return none */
extern void
list_destroy (list_type list)
{
  listItem_type aux1, aux2;
  aux1 = list->first;
  while (aux1 != NILITEM)
    {
      aux2 = aux1;
      aux1 = aux1->next;
      FREE (aux2);
    }
  FREE (list);
}

/** @param  list (in) a list.
    @return "true" if the list operand is empty.
*/
extern boolean
list_is_empty (list_type list)
{
  return (list->first == NILITEM);
}

/** Inserting an element at the end of the list operand (unicity not guaranted).
   @param  list (in) a list.
   @param  item (in) an element.
   @return none
*/
extern void
list_insert_last (list_type list, itemValue_type item)
{
  listItem_type aux;
  aux = MALLOC (listItem_type, sizeof (item_struct));
  ASSERT_MALLOC_OK (aux);
  aux->item = item;
  aux->next = NILITEM;
  if (list->last == NILITEM)
    {
      list->first = aux;
      list->last = aux;
      aux->prev = NILITEM;
    }
  else
    {
      aux->prev = list->last;
      list->last->next = aux;
      list->last = aux;
    }
}

/** Inserting an element at the head of the list operand (unicity not guaranted).
   @param  list (in) a list.
   @param  item (in) an element.
   @return none
*/
extern void
list_insert_first (list_type list, itemValue_type item)
{
  listItem_type aux;
  aux = MALLOC (listItem_type, sizeof (item_struct));
  ASSERT_MALLOC_OK (aux);
  aux->item = item;
  aux->prev = NILITEM;
  if (list->first == NILITEM)
    {
      list->first = aux;
      list->last = aux;
      aux->next = NILITEM;
    }
  else
    {
      aux->next = list->first;
      list->first->prev = aux;
      list->first = aux;
    }
}


/** Inserting an element before the current element of a list (unicity not guaranted).
   @param  list (in) a list.
   @param  item (in) an element.
   @return none
*/
extern void
list_insert_before (list_type list, itemValue_type item)
{
  listItem_type prev, aux;

  if (list->current == NILITEM)
    {
      list_insert_last (list, item);
      return;
    }
  prev = list->current->prev;
  aux = MALLOC (listItem_type, sizeof (item_struct));
  ASSERT_MALLOC_OK (aux);
  aux->item = item;
  aux->next = list->current;
  aux->prev = prev;
  list->current->prev = aux;
  if (prev == NILITEM)
    list->first = aux;
  else
    prev->next = aux;
}

/** Inserting an element after the current element of a list (unicity not guaranted).
   @param  list (in) a list.
   @param  item (in) an element.
   @return none
*/
extern void
list_insert_after (list_type list, itemValue_type item)
{
  listItem_type aux;

  if (list->current == NILITEM)
    {
      list_insert_first (list, item);
      return;
    }
  aux = MALLOC (listItem_type, sizeof (item_struct));
  ASSERT_MALLOC_OK (aux);
  aux->item = item;
  aux->prev = list->current;
  aux->next = list->current->next;
  list->current->next = aux;
  if (aux->next == NILITEM)
    list->last = aux;
  else
    aux->next->prev = aux->next;
}

/** Deleting the current element of a list. After this call, the current element 
   becomes the next of the list or the NULL element if the current was the last element.
   @param  list (in) a list.
   @return none
*/
extern void
list_remove_current (list_type list)
{
  /* this fuction is to be rewritten to be more efficient    */
  listItem_type p, n;

  if (list->current == NILITEM)
    return;
  p = list->current->prev;
  n = list->current->next;
  if (n != NILITEM)
    n->prev = p;
  else
    list->last = p;
  if (p != NILITEM)
    p->next = n;
  else
    list->first = n;
  FREE (list->current);
  list->current = n;
}

/** Setting the current element of the list operand to the first element.
   @param  list (in) a list.
   @return none
*/
extern void
list_set_to_first (list_type list)
{
  list->current = list->first;
}

/** Setting the current element of the list operand to the last element.
   @param  list (in) a list.
   @return none
*/
extern void
list_set_to_last (list_type list)
{
  list->current = list->last;
}

/** Testing current element of a list.
   @param  list (in) a list.
   @return "true" if the current element of the list operand is a the undefined element.
*/
extern boolean
list_is_at_end (list_type list)
{
  return (list->current == NILITEM);
}

/** Getting the value of the current element of a list.
   @param  list (in) a list.
   @return the value assigned to the current element of the list operand ; 
   it can be undefined element.
*/
extern itemValue_type
list_get_current (list_type list)
{
  if (list->current == NILITEM)
    return (itemValue_type) NULL;
  return (list->current)->item;
}

/** Modifying the value of the current element of a list ;side effect on the list.
   @param  list (in) a list.
   @param  item (in) assigned to the current element of the list operand.
   @return none
*/
extern void
list_set_current (list_type list, itemValue_type item)
{
  if (list->current != NILITEM)
    (list->current)->item = item;
}

/**
  Setting the current element of a list to the next one.
  Without effect for a list whose current element is the last.
  @param  list (in) a list.
  @return none
*/
extern void
list_forward (list_type list)
{
  if (list->current == NILITEM)
    return;
  list->current = (list->current)->next;
}

/**
  Setting the current element of a list to the previous one.
  Without effect for a list whose current element is the last.
  @param  list (in) a list.
  @return none
*/
extern void
list_backward (list_type list)
{
  if (list->current == NILITEM)
    return;
  list->current = (list->current)->prev;
}


/** Getting the current element of a list.
   @param  list (in) a list.
   @return the current element of the list operand ; it can be undefined element.
*/
extern listItem_type
list_get_position (list_type list)
{
  return list->current;
}

/** Setting the current element of a list (list) to a specific value (position) with control according (nocheck) value..
   @param  list (in) a list.
   @param  position (in) an element of a list.
   @param  nocheck (in) must be at "true" to control the existency of the position element in the list operand.
               With "false" value, side effect.
   @return none.
*/
extern void
list_set_position (list_type list, listItem_type position, boolean nocheck)
{
  listItem_type aux;
  boolean ok;
  if (!nocheck)
    {
      ok = pk_false;
      aux = list->first;
      while (!ok && (aux != list->last))
	{
	  if (aux == position)
	    ok = pk_true;
	  aux = aux->next;
	}
      if (ok)
	list->current = position;
      else
	list->current = NILITEM;
    }
  else
    list->current = position;
}

/** debugging.
   Print to the "error standard file" the list operand.
   @param  list (in) a list.
   @return none.
*/
extern void
list_debug (list_type list)
{
  listItem_type aux;
  aux = list->first;
  fprintf (stderr, "\n");
  while (aux != NILITEM)
    {
      fprintf (stderr, "%p ", aux->item);
      aux = aux->next;
    }
  fprintf (stderr, "\n");
}
