/*
 * This file is strongly host system dependent
 */

/* ========== Access to system commands
 */
/* sets the current command line buffer empty
 */
extern void
sysSetEmptyCommandLine ()
{
  if (commandBuffer == NIL_String)
    init_Command ();
  else
    commandBuffer[0] = NULL_char;
}

/* sets the current command line equal to <str>
 *      NIL_String is considered as empty string
 * returns false if command buffer is too small 
 */
extern boolean
sysSetStringInCommandLine (const char *str)
{
  if (commandBuffer == NIL_String)
    init_Command ();
  if (str == NIL_String)
    commandBuffer[0] = NULL_char;
  else
    {
      if (strlen (str) > commandLineSize)
	{
	  sys_lineErr ("command buffer too small to contain following string");
	  sys_stderrLine ("str");
	  return pk_false;
	}
      commandBuffer = strcpy (commandBuffer, str);
    }
  return pk_true;
}

/* appends str to the current command line 
 *      NIL_String is considered as empty string
 * returns false if command buffer is too small
 */
extern boolean
sysAppendStringInCommandLine (const char *str)
{
  if (commandBuffer == NIL_String)
    init_Command ();
  if (str != NIL_String)
    {
      if ((strlen (commandBuffer) + strlen (str)) > commandLineSize)
	{
	  sys_lineErr ("command buffer too small to append following string");
	  sys_stderrLine ("str");
	  return pk_false;
	}
      commandBuffer = strcat (commandBuffer, str);
    }
  return pk_true;
}

/* sets the current command line equal to the concatenation of <str1> and <str2> (no space included)
 *      NIL_String is considered as empty string
 * NO SPACE IS INSERTED between prefix and <commandArgs>
 * returns false if command buffer is too small 
 */
extern boolean
sysSetCommandLine (const char *str1, const char *str2)
{
  if (str1 == NIL_String)
    return sysSetStringInCommandLine (str2);
  if (str2 == NIL_String)
    return sysSetStringInCommandLine (str1);
  if (commandBuffer == NIL_String)
    init_Command ();
  if ((strlen (str1) + strlen (str2)) > commandLineSize)
    {
      sys_lineErr ("command buffer too small to concatenate following strings");
      sys_stderrLine ("str1");
      sys_stderrLine ("str2");
      return pk_false;
    }
  commandBuffer = strcpy (commandBuffer, str1);
  commandBuffer = strcat (commandBuffer, str2);
  return pk_true;
}

/* calls system for executing the current content of global command buffer "commandBuffer"
 * failure: commandBuffer is not set or
 *          command execution failed (results is not 0)
 */
extern boolean
sysCurrentExec ()
{
  if (commandBuffer == NIL_String)
    {
      sys_lineErr ("command buffer is empty");
      return pk_false;
    }
  return (system (commandBuffer) == 0);
}

/* calls system for executing <commandString>
 * returns false if command execution failed (results is not 0)
 */
extern boolean
sysExec (const char *commandString)
{
  if (commandString == NIL_String)
    {
      sys_lineErr ("try to execute NIL command");
      return pk_false;
    }
  return (system (commandString) == 0);
}

/* calls system command for creating a directory <filePath>
 * returns false if creation failed (results is not 0)
 */
extern boolean
mkdirectory (const char *filePath)
{
  if (filePath == NIL_String)
    sys_lineErr ("make directory empty string, directory name expected");
  else if (sysSetCommandLine (_COM_MKDIR, filePath))
    return sysCurrentExec ();
  return pk_false;
}

/* sets <filePath> user permission executable
 */
extern boolean
sysSetExecutable (const char *filePath)
{
  if (filePath == NIL_String)
    sys_lineErr ("set executable empty string, file name expected");
  else if (sysSetCommandLine (_COM_SETEXECUTABLE, filePath))
    return sysCurrentExec ();
  return pk_false;
}

/* ========== Elementary path name transformations
 */
/*
 * fileRootName is a non symbolic file name
 * returns fileRootName host dependent suffixed by suffix
 * in a new c'allocated string
 */
extern char *
getSuffixedFileName (const char *fileRootName, const char *suffix)
{
  char *fullFileName;
  int lgs = strlen (fileRootName);

  fullFileName = (char *) calloc (_SUFFIXEDFILENAMELENGTH (lgs, strlen (suffix)), 1);
  fullFileName = strcpy (fullFileName, fileRootName);
  _MAKESUFFIXEDFILENAME (fullFileName, lgs, suffix);
  return fullFileName;
}

/* This function deletes in <filename> the suffix following the first Host separator
 * 
 */
extern boolean
removeFileSuffix (char *filename)
{
  char *suffixPos = _FILENAMEPREFIXEND (filename);
  if (suffixPos == NULL)
    return pk_false;
  _REMOVEFILENAMESUFFIX (filename, suffixPos - filename);
  return pk_true;
}

/**
 * <filePathname> is a variable free pathname
 * filePathName = DDD/FFFcSSS, where FFF is a C identifier, c is the first non C identifier character returns FFF in a new malloc'd string
 * returns NIL_String if memory allocation failed or if FFF is empty
 */
extern char *
extractCFileName (const char *filePathName)
{
  int iii, length;
  char *fileName;
  const char *fullFileName = _GETDIRECTORYSUFFIX (filePathName);
  // fprintf(stderr, "PLG DEBUG extractCFileName fullFileName 1 is <<%s>>\n", fullFileName);
  if (fullFileName == NULL)
    fullFileName = filePathName;
  else
    fullFileName = (char *) (fullFileName + 1);
  length = strlen (fullFileName);
  if (length == 0)
    return NIL_String;
  fileName = (char *) malloc (strlen (fullFileName) + 1);
  fileName = strcpy (fileName, fullFileName);
  if (isdigit (fileName[0]))
    return NIL_String;
  for (iii = 0; iii < length; iii++)
    if (!(isalnum (fileName[iii]) || (fileName[iii] == '_')))
      {
	fileName[iii] = '\0';
	return fileName;
      }
  return fileName;
}

/*
 * <filename> is a variable free pathname
 * fileName = XX/nnn returns XX in a new malloc'd string
 * returns NIL_string if memory allocation failed or if file separator is not found
 */
extern char *
extractContainerPathName (const char *fileName)
{
  char *separatPos;
  char *container = (char *) malloc (strlen (fileName) + 1);
  if (container == (char *) NULL)
    return NIL_String;
  container = strcpy (container, fileName);
  separatPos = _GETDIRECTORYSUFFIX (container);
  if (separatPos == (char *) NULL)
    return NIL_String;
  separatPos[0] = NULL_char;
  return container;
}

/*
 * container is a variable free directory pathname
 * returns the full path name of fileName wrt container prefix in a new malloc'd string
 * the result may be an invalid filename
 */
extern char *
getFilePathName (char *container, const char *fileName)
{
  char *filePathName;
  int lgs = strlen (container);
  if (_ISDIRECTORYPREFIX (container, lgs))
    {
      filePathName = (char *) calloc (lgs + strlen (fileName) + 1, 1);
      filePathName = strcpy (filePathName, container);
    }
  else
    {
      filePathName = (char *) calloc (lgs + strlen (fileName) + 2, 1);
      filePathName = strcpy (filePathName, container);
      _MAKEDIRECTORYPREFIX (filePathName, lgs);
    }
  filePathName = strcat (filePathName, fileName);
  return filePathName;
}

/* checks if <dirPath> is a directory path
 * success: returns the directory prefix associated with <dirpath> in a new malloc'd string
 * failure: returns NIL_String
 */
extern char *
getTaggedDirPath (const char *dirPath)
{
  int length;
  char *newDirPath;
  if (!file_exist (dirPath))
    if (!mkdirectory (dirPath))
      return NIL_String;
  length = strlen (dirPath);
  if (_ISNOTDIRECTORYPREFIX (dirPath, length - 1))
    {
      newDirPath = Cstrmalloc (length + 1);
      strcpy (newDirPath, dirPath);
      _MAKEDIRECTORYPREFIX (newDirPath, length);
    }
  else
    {
      newDirPath = Cstrmalloc (length);
      strcpy (newDirPath, dirPath);
    }
  return newDirPath;
}

/* ========== Access to system environment variables
 */

/*
 * varName is NIL_String or a host variable name
 * success: not empty var name is defined in the environnement
 *              returns its value; it must remain unchanged
 *          not empty var name is not defined in the environnement
 *              prints an error message on stderr
 *              return NIL_String
 * failure: returns NIL_String
 *          prints error message
 */
extern char *
get_string_HostValue_ (const char *varName)
{
  if (varName != NIL_String)
    {
      int nameLength = strlen (varName);
      if (nameLength > 0)
	{
	  ref2HostVariable refVariable;
	  if (strchr (_NOTENVNAMECHARSET, varName[0]) == NULL)
	    refVariable = find_HostVariable_ (varName);
	  else if (nameLength > 1)
	    refVariable = find_HostVariable_ ((varName + 1));
	  else
	    {
	      sys_lineErr ("host variable name is empty");
	      return NIL_String;
	    }
	  if (refVariable == NIL_HostVariable)
	    return NIL_String;
	  return refVariable->value;
	}
    }
  sys_lineErr ("host variable name is empty");
  return NIL_String;
}

/*
 * varName is NIL_String or a host variable name
 * success: not empty var name is defined in the environnement
 *              returns its value in <str_res> ; it must remain unchanged
 *          not empty var name is not defined in the environnement
 *              prints an error message on stderr
 *              return <str_res> NIL_String
 * failure: returns NIL_String
 *          error message is printed
 */
extern boolean
getIf_string_HostValue_ (const char *varName, char **str_res)
{
  if (varName != NIL_String)
    {
      int nameLength = strlen (varName);
      if (nameLength > 0)
	{
	  ref2HostVariable refVariable;
	  if (strchr (_NOTENVNAMECHARSET, varName[0]) == NULL)
	    refVariable = find_HostVariable_ (varName);
	  else if (nameLength > 1)
	    refVariable = find_HostVariable_ ((varName + 1));
	  else
	    {
	      sys_lineErr ("host variable name is empty");
	      return pk_false;
	    }
	  if (refVariable == NIL_HostVariable)
	    {
	      sys_lineErr ("unkown host variable");
	      return pk_false;
	    }
	  (*str_res) = refVariable->value;
	  return pk_true;
	}
    }
  sys_lineErr ("empty host variable name");
  return pk_false;
}

/*
 * source is NIL_String or a NUL_char terminated string (not checked !)
 * source begins with a var name equal to the longest string, free of ' ', '!', '\n'
 * success: not empty var name is defined in the environnement
 *              returns its value; it must remain unchanged
 *          not empty var name is not defined in the environnement
 *              prints an error message on stderr
 *              return NIL_String
 * failure: var name is empty
 *          prints an error message on stderr
 *          returns NIL_String
 */
extern char *
get_string_HostValue_heading_ (const char *source)
{
  ref2HostVariable refVariable;
  refVariable = find_HostVariable_heading_ (source);
  if (refVariable == NIL_HostVariable)
    return NIL_String;
  return refVariable->value;
}

/* returns the integer value of the environment variable nomvar 
 * results is undefined if nomvar value is not a correct integer
 */
extern int
get_int_HostValue (const char *nomvar)
{
  char *str = get_string_HostValue_ (nomvar);
  if (str == NIL_String)
    {
      return 0;
    }
  return (atoi (str));
}

/*
 * Non ANSI <stdio> facility 
 * returns the name of the user who owns the current process
 */
extern char *
getEnv_UserName ()
{
  char *userName = (char *) malloc (L_cusername + 1);
  userName = get_string_HostValue_ (_ENV_USERNAME);
  if (userName == NIL_String)
    return (char *) "unknownUser";
  return userName;
}

/* ===================================================================== 
 */
/* set working directory variable to host current directory
 */
extern void
setWorkingDirectory ()
{
  if (workingDirectoryPrefix == NIL_String)
    init_WD ();
  else
    init_WDContent ();
}

extern void
resetSystemEnv ()
{
  setWorkingDirectory ();
  reset_VariableList ();
}

extern void
initFileSystemEnv ()
{
  init_Command ();
  setWorkingDirectory ();
  init_Variable ();
  curErrMess = (char **) calloc (sizeof (char *), 1);
}

/* ============================= variable substitutions
 */
/*
 * if <source> is variable free returns <source>
 * if variables occur in <source>
 *    uses variableBuffer
 *    returns a new malloc'd string equal to the substitution of variable occurrences in <source>
 *     an environnement variable occurrence is the longest string free of _NOTENVVARCHARSET 
 *     following an _ENVNAMETAG occurrence
 * NULL terminated string source; source can be empty string
 * returns NIL_String if some environnement variable is missing
 */
/* <source> cannot be a const or the function should return const */
extern char *
getEnv_metaStringActualValue (char *source)
{
  char *actualValue;
  int sourceIndex = 0, fullLength = 0;

  size_t sourceLength;
  size_t targetIndex = 0;
  int nameLength;
  int lg;
  boolean symbolic = pk_false;

  if (variableBuffer == NIL_String)
    initFileSystemEnv ();
  if (source == NIL_String)
    return NIL_String;
  sourceLength = strlen (source);
  if (sourceLength <= 0)
    return NIL_String;
  while ((sourceIndex < sourceLength) && (fullLength < FILENAME_MAX))
    {
      if (_ISENVNAME (source, sourceIndex))
	{
	  if (!(symbolic))
	    {
	      targetIndex = sourceIndex;
	      resetPathBuffer (source, sourceIndex);
	      symbolic = pk_true;
	    }
	  sourceIndex = sourceIndex + 1;
	  actualValue = get_string_HostValue_heading_length_ ((char *) (source + sourceIndex), &nameLength);
	  if (actualValue == NIL_String)
	    return NIL_String;
	  sourceIndex = sourceIndex + nameLength;
	  lg = append2Target (actualValue, targetIndex);
	  targetIndex = targetIndex + lg;
	}
      if (symbolic)
	{
	  variableBuffer[targetIndex] = source[sourceIndex];
	  targetIndex = targetIndex + 1;
	}
      sourceIndex = sourceIndex + 1;
    }
  variableBuffer[targetIndex] = NULL_char;
  if (symbolic)
    return newCstr_String (variableBuffer);
  return source;
}

/*
 * Non ANSI <stdio> facility 
 * uses workingDirectoryPrefix
 * if <aPathname> is rooted and variable free, returns <aPathname>
 * if <aPathname> is not rooted or has variable occurrences,
 *          returns in a new malloc'd string its vraiable free, rooted pathname
 */
extern char *
getEnv_RootedPathName (char *aPathName)
{
  char *fullPathName;
  char *freePathName;

  if (workingDirectoryPrefix == NIL_String)
    init_WD ();
  freePathName = getEnv_metaStringActualValue (aPathName);
  if (freePathName == NIL_String)
    return NIL_String;
  if (_ISROOTEDPATHNAME (freePathName))
    return freePathName;
  fullPathName = (char *) malloc (WDLength + strlen (freePathName) + 1);
  fullPathName = strcpy (fullPathName, workingDirectoryPrefix);
  fullPathName = strcat (fullPathName, freePathName);
  if (freePathName != aPathName)
    free (freePathName);
  return fullPathName;
}

/* 
 * <dirName> is a string that has enough space to fully contain the found directory path name
 * <pathList> is a sequence of path references separated by spaces (' ')
 * a path reference is a "pathSeparatorInPathList" character (':' for instance) immediately followed by a pathname
 * a pathname is the longuest sequence of not "pathSeparatorInPathList", not ' ' characters
 * returns in <dirName the first pathname after <pos> in <pathList>>
 * if <dirName> is not a directory prefix makes it a directory prefix
 * set <pos> to the position following pathname in path
 * returns pk_true if some pathName has been read in <pathList>
 */
extern boolean
extractDirname (const char *pathList, char *dirName, int *pos)
{
  int iPathList = *pos, iBuf = 0;
  char c = pathList[*pos];

  while (c == ' ')
    {
      iPathList = iPathList + 1;
      c = pathList[iPathList];
    }
  if (c == pathSeparatorInPathList)
    {
      iPathList = iPathList + 1;
      c = pathList[iPathList];
    }
  while ((c != NULL_char) && (c != pathSeparatorInPathList) && (c != ' '))
    {
      dirName[iBuf] = c;
      iBuf = iBuf + 1;
      iPathList = iPathList + 1;
      c = pathList[iPathList];
    }
  *pos = iPathList;
  if (iBuf > 0)
    {
      if (_ISNOTDIRECTORYPREFIX (dirName, iBuf - 1))
	{
	  _MAKEDIRECTORYPREFIX (dirName, iBuf);
	}
      else
	dirName[iBuf] = NULL_char;
      return pk_true;
    }
  dirName[0] = NULL_char;
  return pk_false;
}

/**
  varName is a 0 terminated string denoting a variable environnement
    - success returns the environnement variable value
      the resulting string must not be modified,
    - failure returns NIL_String
 */
extern char *
getEnvDef (const char *varName)
{
  return get_string_HostValue_heading_ (varName);
}

/**
   It defines an environment variable (eVar) by an associated data (vVar).
   It can be used to avoid the access to the execution environment (shell variables).
*/
extern void
setEnvDef (const char *eVar, const char *vVar)
{
  ref2VariableType refVariable;

  refVariable = newVariable (strduplicate (eVar), strduplicate (vVar));
  registerVariable (refVariable);
}

/**
   @return true if an environment variable (eVar) is already defined, false otherwise.
*/
extern boolean
isEnvDef (const char *eVar)
{
  ref2VariableType refVariable;
  return isDefinedVariable (eVar, &refVariable);
}
