/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#include <stdio.h>
#include "boolean.h"
#include "StdTypeDefs.h"
#include "typesbase.h"
#include "gm_arbre.h"
#include "lta_cbase.h"
#include "li_attri.h"

// #include "m_base.h"

/** find in aNode the first attribute tagged by tagAttribute
   @return true if this attribute is found and is typed by attributeType
   (*attribute) is assigned attribute when the fucntion returns true
    @param aNode is not NIL (not checked)
    assume correct parameters (not checked)
 */
extern boolean
getCondTM_typedAttribute (ptnode aNode, short tagAttribute, short attributeType, ptattribute * attributeRes)
{
  // printf("DEBUG entering getCondTM_typedAttribute, tagAttribute is %i\n", tagAttribute); fflush(stdout);
  (*attributeRes) = NODE_GET_ATTRIBUTE (aNode);
  if ((*attributeRes) == (ptattribute) NIL)
    return pk_false;
  while ((ATTRIBUTE_GET_TAG (*attributeRes)) < tagAttribute)
    {
      (*attributeRes) = ATTRIBUTE_GET_SUCC (*attributeRes);
      if ((*attributeRes) == (ptattribute) NIL)
	return pk_false;
    }
  if ((ATTRIBUTE_GET_TAG (*attributeRes)) == tagAttribute)
    {
      return (ATTRIBUTE_GET_INNEROPER ((*attributeRes)) == attributeType);
    }
  return pk_false;
}

/** search in aNode the attribute tagged by tagAttribute
   @return true: attribute is found; it is assigned to attrRes
   @return false: attribute is not found; 
       if a predecessor exists it is assigned to attrRes
       if aNode does not have attribute or if num is smaller than the smallest tag attribute in aNode, NIL is assigned to attrRes
    @param aNode is not NIL (not checked)
    assume correct parameters (not checked)
 */
extern boolean
getCondTM_taggedAttribute (ptnode aNode, short tagAttribute, ptattribute * attrRes)
{
  ptattribute curAttr, predAttr;

  curAttr = NODE_GET_ATTRIBUTE (aNode);
  if (curAttr == NIL_attribute)
    {
      (*attrRes) = NIL_attribute;
      return pk_false;
    }
  if (tagAttribute == (ATTRIBUTE_GET_TAG (curAttr)))
    {
      (*attrRes) = curAttr;
      return pk_true;
    }
  predAttr = NIL_attribute;
  while (tagAttribute > (ATTRIBUTE_GET_TAG (curAttr)))
    {
      predAttr = curAttr;
      curAttr = ATTRIBUTE_GET_SUCC (curAttr);
      if (curAttr == NIL)
	{
	  (*attrRes) = predAttr;
	  return pk_false;
	}
    }
  if (tagAttribute == (ATTRIBUTE_GET_TAG (curAttr)))
    {
      (*attrRes) = curAttr;
      return pk_true;
    }
  (*attrRes) = predAttr;
  return pk_false;
}

/** create and return a new LTA attribute ATT (tagged by 0) attached to aNode as single attribute
    set type and value of ATT respectivly as LTAType and value
    @param aNode is not NIL (not checked)
    assume correct parameters (not checked)
 */
extern ptattribute
attachTM_newAttribute0 (ptnode aNode, short LTAType, untypedValueType value)
{
  ptattribute newAttribute;

  newAttribute = newTM_attribute (0);
  ATTRIBUTE_SET_INNEROPER_VALUE (newAttribute, LTAType, value);
  NODE_LINK_ATTRIBUTE (aNode, newAttribute);
  return newAttribute;
}

/** create and return a new attribute attached to aNode in first position
    @param aNode is not NIL (not checked)
    @param tagAttribute is a correct tag attribute value (not checked)
 */
extern ptattribute
insertTM_newFirstAttribute (ptnode aNode, short tagAttribute)
{
  ptattribute nextAttrSucc;
  ptattribute newAttribute;

  nextAttrSucc = NODE_GET_ATTRIBUTE (aNode);

  newAttribute = newTM_attribute (tagAttribute);
  NODE_LINK_ATTRIBUTE (aNode, newAttribute);

  ATTRIBUTE_SET_PRED (newAttribute, NIL_attribute);
  ATTRIBUTE_SET_SUCC (newAttribute, nextAttrSucc);
  if (nextAttrSucc != NIL_attribute)
    ATTRIBUTE_SET_PRED (nextAttrSucc, newAttribute);

  return newAttribute;
}

/** create and return a new attribute attached to aNode after attrPred
    @param attrPred is not NIL (not checked)
    @param tagAttribute is a correct tag attribute value (not checked)
 */
extern ptattribute
insertTM_newAttributeAfter (ptattribute attrPred, short tagAttribute)
{
  ptattribute nextAttrSucc;
  ptattribute newAttribute;

  nextAttrSucc = ATTRIBUTE_GET_SUCC (attrPred);
  newAttribute = newTM_attribute (tagAttribute);

  ATTRIBUTE_SET_OWNER (newAttribute, ATTRIBUTE_GET_NODE_OWNER (attrPred));
  ATTRIBUTE_SET_PRED (newAttribute, attrPred);
  ATTRIBUTE_SET_SUCC (newAttribute, nextAttrSucc);

  ATTRIBUTE_SET_SUCC (attrPred, newAttribute);
  if (nextAttrSucc != NIL)
    ATTRIBUTE_SET_PRED (nextAttrSucc, newAttribute);
  return newAttribute;
}

/** set aSon as OPARBRE value of attribute newAttr
    @param newAttr is not NIL (not checked)
    @param aSon is a correct filsd root (not checked)
 */
void
attachTM_treeAttribute (ptattribute newAttr, ptfilsd aSon)
{
  ptnode aNode;

  ATTRIBUTE_MAKE_TREE (newAttr, aSon);
  aNode = FILSD_GET_NODE_CONTAINER_OF_FIRST (aSon);
  NODE_SET_ATTRIBUTE_CONTAINER (aNode, newAttr);
}

/** remove the attribute "attr_removed" from the attribute list of the node "nodeContainer"
    @param nodeContainer is not NIL (not checked)
    @param attr_removed is not NIL (not checked)
 */
extern void
detachTM_attribute (ptnode nodeContainer, ptattribute attr_removed)
{
  ptattribute attr_pred, attr_succ;

  attr_pred = ATTRIBUTE_GET_PRED (attr_removed);
  attr_succ = ATTRIBUTE_GET_SUCC (attr_removed);

  ATTRIBUTE_ISOLATE (attr_removed);

  if (attr_pred != NIL)
    ATTRIBUTE_SET_SUCC (attr_pred, attr_succ);
  else
    NODE_SET_ATTRIBUTE (nodeContainer, attr_succ);

  if (attr_succ != NIL)
    ATTRIBUTE_SET_PRED (attr_succ, attr_pred);

}

/** @return the attribute tagged by tagAttribute after creating it if it does not exist
    @param aSubtree should not be nil (checked)
    @param tagAttribute is a correct attribute number (checked)
    @return NIL_attribute if aSubtree is NIL or if tagAttribute is not a correct attribute number
*/
extern ptattribute
getOrCreateTM_attribute (trees aSubtree, short tagAttribute)
{
  ptattribute newAttr, anAttr;
  if (aSubtree == NIL_trees)
    return NIL_attribute;
  if (!(CHECKATTRIBUTENUMBER (tagAttribute)))
    {
      printf ("## ERROR ## in getOrCreateTM_attribute %i is not a correct attribute number\n", tagAttribute);
      fflush (stdout);
      return NIL_attribute;
    }
  if (getCondTM_taggedAttribute (aSubtree, tagAttribute, &anAttr))
    newAttr = anAttr;
  else
    {
      if (anAttr == NIL_attribute)
	newAttr = insertTM_newFirstAttribute ((ptnode) aSubtree, tagAttribute);
      else
	newAttr = insertTM_newAttributeAfter (anAttr, tagAttribute);
    }
  return newAttr;
}
