/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*****************************************************************************/
/*									     */
/*	    recuperer.c							     */
/*									     */
/*									     */
/* Ce module contient les primitives de recuperation de memoire dans la zone */
/* des arbres.								     */
/*									     */
/*****************************************************************************/

#include "StdTypeDefs.h"
#include "varglobal.h"
#include "gm_pile.h"
#include "lta_ctree.h"
#include "marquer.h"
#include "recuperer.h"

#define POS_OPER 8
#define SHORT_SET_UNMARKEDVALUE(pt) *pt = SHORT_GET_UNMARKEDVALUE(pt)

/*****************************************************************************/
/*									     */
/*	 majadresses							     */
/*									     */
/*  Mise a jour, par l'adresse nouv, des pointeurs avant sur la cellule	     */
/* d'adresse old. Cette mise a jour s'effectue par parcours du chainage      */
/* inverse jusqu'a trouver la cellule contenant la valeur initiale de old    */
/* cellule qui est marquee par SHORT_MARQFINCHAINE				     */
/*									     */
/*****************************************************************************/
// premiere_passe, deuxieme_pass call
static void
majadresses (ptchar old, ptchar nouv)
{
  ptLONG pt;
  long int tamp;

  pt = (ptLONG) (LONG_GET_UNMARKEDVALUE ((ptLONG) old));

  while (SHORT_TESTMARQUE ((ptSHORT) pt) != SHORT_MARQFINCHAINE)
    {
      tamp = *pt;
      *pt = (longInteger) nouv;
      pt = (ptLONG) tamp;
      pt = (ptLONG) ((longInteger) pt & LONG_VALUEFILTER);
    }

  SHORT_SET_INITIALMARK ((ptSHORT) pt);
  tamp = *pt;
  *pt = (longInteger) nouv;
  pt = (ptLONG) old;
  *pt = tamp;
}

/*****************************************************************************/
/*									     */
/*	chainer_inverse							     */
/*									     */
/*  Fonction de creation des liens inverses de tous les pointeurs contenus   */
/* dans la cellule pointee par pt					     */
/*									     */
/*****************************************************************************/
// premiere_passe call
static short
chainer_inverse (ptchar pt)
{
  short taille;
  ptnode e;
  ptfilslist fl;
  ptattribute att;
  ptrepere r;

  switch (cellCategory ((ptSHORT) pt))
    {
    case 0:			/* filslist */
      fl = (ptfilslist) pt;
      pt = pt + 8;		/* positionnement sur adfils */
      if (ANYFILS_GET_INNEROPER (fl) >= 6)
	chaineinverse ((ptptSHORT) pt);
      if (ANYFILS_ISIN_LIST (fl))
	{
	  pt = pt + 4;
	  if (FILSLIST_GET_NODE_CONTAINER (fl) != NIL_node)
	    chaineinverse ((ptptSHORT) pt);
	  pt = pt + 4;
	  if (FILSLIST_GET_PRED (fl) != (ptfilslist) NIL)
	    chaineinverse ((ptptSHORT) pt);
	  pt = pt + 4;
	  if (FILSLIST_GET_SUCC (fl) != (ptfilslist) NIL)
	    chaineinverse ((ptptSHORT) pt);
	  taille = FILSLISTSIZE;
	}
      else
	taille = FILSDSIZE;
      break;
    case 1:			/* attribut */
      taille = ATTRIBUTESIZE;
      att = (ptattribute) pt;
      pt = pt + 4;
      if ((ATTRIBUTE_GET_UNTYPED_VALUE (att) != NIL_untypedValueType) && (ATTRIBUTE_GET_INNEROPER (att) >= 6))
	chaineinverse ((ptptSHORT) pt);
      pt = pt + 4;
      if (ATTRIBUTE_GET_NODE_OWNER (att) != NIL_node)
	chaineinverse ((ptptSHORT) pt);
      pt = pt + 4;
      if (ATTRIBUTE_GET_PRED (att) != (ptattribute) NIL)
	chaineinverse ((ptptSHORT) pt);
      pt = pt + 4;
      if (ATTRIBUTE_GET_SUCC (att) != (ptattribute) NIL)
	chaineinverse ((ptptSHORT) pt);
      break;
    case 2:			/* repere */
      taille = REPERESIZE;
      r = (ptrepere) pt;
      pt = pt + 4;
      if (estarbre ((ptSHORT) (REPERE_GET_FILSD_REFERED (r))))
	chaineinverse ((ptptSHORT) pt);
      break;
    case 3:			/* node */
      e = (ptnode) pt;
      pt = pt + 4;
      if (NODE_GET_FILSD_DESCRIPTOR (e) != (ptfilsd) NIL)
	chaineinverse ((ptptSHORT) pt);
      pt = pt + 4;
      if (NODE_GET_ATTRIBUTE (e) != (ptattribute) NIL)
	chaineinverse ((ptptSHORT) pt);
      taille = NODESIZE;
      break;
    }
  return taille;
}

/*****************************************************************************/
/*									     */
/*	premiere_passe							     */
/*									     */
/*  Fonction qui effectue un premier passage lineaire de la memoire. Elle    */
/* traite les trous rencontres pour preparation de la deuxieme passe. Elle   */
/* travaille selon l'algorithme de Jonkers sur chaque cellule occupee :	     */
/*	- calcul de la nouvelle adresse					     */
/*	- mise a jour des pointeurs en avant sur cette cellule par parcours  */
/*	  du chainage inverse.						     */
/*	- chainage inverse des pointeurs contenus dans la cellule.	     */
/*									     */
/*****************************************************************************/
// recuperation call
static int
premiere_passe (void)
{
  ptchar pt, nouvadr;
  int TTROU;			/* taille du trou courant */
  int TRECUP;			/* taille totale des trous rencontres */
  ptLONG dernier_trou;		/* adresse de debut du dernier trou rencontres */
  short booleantrou;
  short taille;
  ptnode e;

  booleantrou = 2;
  pt = GET_TREEMEMORY_FIRSTCELL ();
  TTROU = 0;
  TRECUP = 0;
  dernier_trou = (ptLONG) NIL;

  while (pt < LIB_ARBRE)
    {
      switch (SHORT_TESTMARQUE ((ptSHORT) pt))
	{
	case SHORT_MARQLIBRE:	/* traiter trou */
	  if (booleantrou != 1)	/* debut d'un trou */
	    {
	      dernier_trou = (ptLONG) pt;
	      booleantrou = 1;
	    }
	  switch (cellCategory ((ptSHORT) pt))
	    {
	    case 0:		/* filslist */
	      taille = FILSLISTSIZE;
	      break;
	    case 1:		/* attribut */
	      taille = ATTRIBUTESIZE;
	      break;
	    case 2:		/* repere */
	      taille = REPERESIZE;
	      break;
	    case 3:		/* node */
	      e = (ptnode) pt;
	      taille = NODE_GET_ARITY (e);
	      switch (taille)
		{
		case -1:
		  taille = NODESIZE + FILSDSIZE;
		  break;
		case 0:
		  taille = NODESIZE;
		  break;
		default:
		  taille = NODESIZE + taille * FILSDSIZE;
		  break;
		}
	    }
	  TTROU = TTROU + taille;
	  TRECUP = TRECUP + taille;
	  pt = pt + taille;	/* cellule suivante */
	  break;

	case SHORT_MOVEDMARK:	/* cellule marquee et liee */
	  nouvadr = pt - TRECUP;	/* calcul de la futur adresse */
	  majadresses (pt, nouvadr);	/* mise a jour des pointeurs */
	  /* avant avec nouvadr */
	case SHORT_INITIALMARK:	/* cellule marquee mais non liee */
	case SHORT_MARQFINCHAINE:	/* cellule marque et fin de chainage inverse */
	  taille = chainer_inverse (pt);	/* chainage inverse des */
	  /* pointeurs contenus dans la cellule */
	  pt = pt + taille;
	  if (booleantrou == 1)	/* 1ere cellule marquee apres un trou */
	    {
	      *dernier_trou = TTROU;
	      TTROU = 0;
	      booleantrou = pk_false;
	    }
	  break;
	}
    }
  if (booleantrou == 1)
    {
      /* la me'moire se termine par un trou */
      TRECUP = TRECUP - (LIB_ARBRE - (ptchar) dernier_trou);
      if (TRECUP == 0)
	TRECUP = -1;
      LIB_ARBRE = (ptchar) dernier_trou;
    }
  return (TRECUP);
}

/*****************************************************************************/
/*									     */
/*	recopiercell							     */
/*									     */
/*  Fonction qui recopie la cellule d'adresse old a l'adresse new en enlevant*/
/* le marquage de la cellule						     */
/*									     */
/*****************************************************************************/
// deuxieme_passe call
static short
recopiercell (ptchar old1, ptchar new1)
{
  short taille;
  ptnode e1, e2;
  ptrepere r1, r2;
  ptfilslist fl1, fl2;
  ptattribute att1, att2;
  ptfilsd f1, f2;

  SHORT_SET_UNMARKEDVALUE ((ptSHORT) old1);
  switch (cellCategory ((ptSHORT) old1))
    {
    case 0:			/* filslist ou ptfilsd */
      f2 = (ptfilsd) old1;
      if (ANYFILS_ISIN_LIST (f2))
	{
	  fl1 = (ptfilslist) new1;
	  fl2 = (ptfilslist) old1;
	  *fl1 = *fl2;
	  taille = FILSLISTSIZE;
	}
      else
	{
	  f1 = (ptfilsd) new1;
	  *f1 = *f2;
	  taille = FILSDSIZE;
	}
      break;
    case 1:			/* attribut */
      att1 = (ptattribute) new1;
      att2 = (ptattribute) old1;
      *att1 = *att2;
      taille = ATTRIBUTESIZE;
      break;
    case 2:			/* repere */
      r1 = (ptrepere) new1;
      r2 = (ptrepere) old1;
      *r1 = *r2;
      taille = REPERESIZE;
      break;
    case 3:			/* node */
      e1 = (ptnode) new1;
      e2 = (ptnode) old1;
      *e1 = *e2;
      taille = NODESIZE;
      break;
    }
  return taille;
}

/*****************************************************************************/
/*									     */
/*	deuxieme_passe							     */
/*									     */
/*  Fonction qui effectue le deuxieme parcours lineaire de la memoire.	     */
/* Elle corrige les pointeurs en arriere et deplace les cellules occupees    */
/* suivant l'algorithme de Jonkers :					     */
/*	- calcul de la nouvelle adresse					     */
/*	- mise a jour des pointeurs arrieres sur la cellule		     */
/*	- recopie de la cellule a sa nouvelle adresse			     */
/*									     */
/*****************************************************************************/
// recuperation call
static void
deuxieme_passe (int TRECUP)
{
  ptLONG pi;
  short taille;
  ptchar old1, new1;

  old1 = GET_TREEMEMORY_FIRSTCELL ();
  new1 = GET_TREEMEMORY_FIRSTCELL ();

  while (old1 < LIB_ARBRE)
    {
      switch (SHORT_TESTMARQUE ((ptSHORT) old1))
	{
	case SHORT_MARQLIBRE:
	  pi = (ptLONG) old1;
	  old1 = old1 + *pi;
	  break;
	case SHORT_MOVEDMARK:
	  majadresses (old1, new1);
	case SHORT_INITIALMARK:
	  taille = recopiercell (old1, new1);
	  new1 = new1 + taille;
	  old1 = old1 + taille;
	  break;
	}
    }
  if (TRECUP != -1)
    LIB_ARBRE = LIB_ARBRE - TRECUP;
}

void
recuperation (void)
{
  int i;
  // PLG following return is temporary until garbage is checked 
  return;
  // printf("DEBUG recuperation\n"); fflush(stdout);
  marquerpile ();
  i = premiere_passe ();
  deuxieme_passe (i);
}
