/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/****************************************************************************/
/*                                                                          */
/*        Module d'interface (cote langage C) de gestion des chaines        */
/*                                                                          */
/****************************************************************************/

#include <string.h>
#include "system.h"
#include "varglobal.h"
#include "stringmemStr_type.h"
#include "stringmemRegister_type.h"

#include "gm_chaine.h"
#include "chaine.h"
#include "stringme.h"
#include "TM_DEBUG.h"

#define CSTH 63

/** assign an empty free register to aRegister
    failure: no more available register
 */
extern boolean
new_reg (ptreg * aRegister)
{
  *aRegister = allouerregchaine ();
  // DEBUG_regIndex("stringme::new_reg return ", *aRegister);
  return (*aRegister != -1);
}

/** free register aRegister such that it may be reused
    assume aRegister is a correct string register
 */
extern void
regFree (ptreg aRegister)
{
  // DEBUG_regIndex("entering stringme::regFree return ", aRegister);
  libererregchaine (aRegister);
  // DEBUG_regIndex(" exiting stringme::regFree return ", aRegister);
}

/** clear the register aRegister
    assume aRegister is a correct string register
 */
extern void
regClear (ptreg aRegister)
{
  ptStringRegister aStringRegister;

  // DEBUG_regIndex("entering stringme::regClear return ", aRegister);
  aStringRegister = indexRegisterCAST_stringRegister (aRegister);
  stringRegister_SET_USEDSIZE (aStringRegister, 0);
  stringRegister_INIT_STARTINDEX (aStringRegister);
  stringRegister_INIT_STRINGBUF (aStringRegister);
  // DEBUG_regIndex(" exiting stringme::regClear return ", aRegister);
}

/** @return the adress of aRegister buffer
    assume aRegister is a correct string register
 */
extern char *
regGet_bufAdr (ptreg aRegister)
{
  return stringRegister_GET_BUF (indexRegisterCAST_stringRegister (aRegister));
}

/** @return the rank of the first character in aRegister
    assume aRegister is a correct string register
 */
extern int
regGet_startRank (ptreg aRegister)
{
  return stringRegister_GET_STARTINDEX (indexRegisterCAST_stringRegister (aRegister)) + 1;
}

/** @return the char at rank rank in register aRegister
    assume aRegister is a correct string register
    assume rank is a correct rank in aRegister
 */
extern char
regGet_char (ptreg aRegister, int rank)
{
  ptStringRegister aStringRegister;

  aStringRegister = indexRegisterCAST_stringRegister (aRegister);
  return (stringRegister_GET_BUF (aStringRegister)[stringRegister_GET_STARTINDEX (aStringRegister) + rank - 1]);
}

/** @return the number of usefull characters in aRegister
    assume aRegister is a correct string register
 */
extern int
regGet_usedSize (ptreg aRegister)
{
  return stringRegister_GET_USEDSIZE (indexRegisterCAST_stringRegister (aRegister));
}

/** reset  the number of usefull characters in aRegister
    complete by '\0' the string content of aRegister
    assume aRegister is a correct string register
    assume regLg is a correct length in aRegister
 */
extern void
regSet_usedSize (ptreg aRegister, int regLg)
{
  ptStringRegister aStringRegister = indexRegisterCAST_stringRegister (aRegister);
  stringRegister_SET_USEDSIZE (aStringRegister, regLg);
  stringRegister_COMPLETE_STRING (aStringRegister);
}

/** append char c to the end of aRegister
    return false if aRegister has not enough available space
    assume aRegister is a correct string register
*/
extern boolean
regConc_char (ptreg aRegister, char c)
{
  ptStringRegister tgtStringRegister;
  int tgtFreeCharIndex;

  // DEBUG_regIndex("entering stringme::regConc_char aRegister",aRegister);
  // printf("DEBUG entering stringme::regConc_char char is <<%c>>\n", c); fflush(stdout);
  tgtStringRegister = indexRegisterCAST_stringRegister (aRegister);
  tgtFreeCharIndex = stringRegister_GET_FREECHARINDEX (tgtStringRegister);
  if (tgtFreeCharIndex + 1 > MAXLGCHAINE)
    {
      printf ("##INTERNAL ERROR## in regConc_char string register is full\n");
      fflush (stdout);
      return pk_false;
    }
  stringRegister_GET_BUF (tgtStringRegister)[tgtFreeCharIndex] = c;
  stringRegister_ADD_USEDSIZE (tgtStringRegister, 1);
  stringRegister_COMPLETE_STRING (tgtStringRegister);
  // DEBUG_regIndex(" exiting stringme::regConc_char aRegister",aRegister);
  return pk_true;
}

/** success: assign a free register to aRegister, char c is loaded in aRegister
    failure: no more available register or initial loading failed: 
 */
extern boolean
new_reg_char (ptreg * aRegister, char c)
{
  *aRegister = allouerregchaine ();
  if (*aRegister == -1)
    return pk_false;
  if (regConc_char (*aRegister, c))
    return pk_true;
  libererregchaine (*aRegister);
  return pk_false;
}

/** srcLg times append char c to the end of aRegister
    return false if aRegister has not enough available space
    assume aRegister is a correct string register
*/
extern boolean
regConcNtimes_char (ptreg aRegister, char c, int srcLg)
{
  ptStringRegister tgtStringRegister;
  int i, tgtFreeCharIndex;
  char *tgtString;

  // DEBUG_regIndex("entering stringme::regConcNtimes_char aRegister",aRegister);
  // printf("DEBUG entering stringme::regConcNtimes_char char is <<%c>>, iterator is %i\n", c, srcLg); fflush(stdout);
  if (srcLg > 0)
    {
      tgtStringRegister = indexRegisterCAST_stringRegister (aRegister);
      tgtFreeCharIndex = stringRegister_GET_FREECHARINDEX (tgtStringRegister);
      if (tgtFreeCharIndex + srcLg > MAXLGCHAINE)
	{
	  printf ("##INTERNAL ERROR## in regConcNtimes_char: not enough space in target string register\n");
	  fflush (stdout);
	  return pk_false;
	}
      tgtString = (stringRegister_GET_BUF (tgtStringRegister)) + tgtFreeCharIndex;
      for (i = 0; i < srcLg; i++)
	{
	  tgtString[i] = c;
	}
      stringRegister_ADD_USEDSIZE (tgtStringRegister, srcLg);
      stringRegister_COMPLETE_STRING (tgtStringRegister);
    }
  // DEBUG_regIndex(" exiting stringme::regConcNtimes_char aRegister",aRegister);
  return pk_true;
}


/** assign a free register to aRegister
    char c is loaded srcLg times in aRegister
    failure: no more available register or initial loading failed
 */
extern boolean
new_regNtimes_char (ptreg * aRegister, char c, int srcLg)
{
  *aRegister = allouerregchaine ();
  if (*aRegister == -1)
    return pk_false;
  if (regConcNtimes_char (*aRegister, c, srcLg))
    return pk_true;
  libererregchaine (*aRegister);
  return pk_false;
}

/** append srcLg characters from srcString to the end of aRegister
    srcString is assumed to have more than srcLg not null character
    return false if aRegister has not enough available space
*/
extern boolean
regConcNchar_CString (ptreg aRegister, const char *srcString, int srcLg, boolean separate)
{
  ptStringRegister tgtStringRegister;
  int tgtFreeCharIndex;
  char *tgtString;

  // DEBUG_regIndex("entering stringme::regConcNchar_CString aRegister",aRegister);
  // printf("DEBUG entering stringme::regConcNchar_CString char, bound is %i\n", srcLg); fflush(stdout);
  if (srcLg > 0)
    {
      tgtStringRegister = indexRegisterCAST_stringRegister (aRegister);
      tgtFreeCharIndex = stringRegister_GET_FREECHARINDEX (tgtStringRegister);
      if (separate)
	{
	  if (regConc_char (aRegister, ' '))
	    tgtFreeCharIndex = tgtFreeCharIndex + 1;
	  else
	    return pk_false;
	}
      if (tgtFreeCharIndex + srcLg > MAXLGCHAINE)
	{
	  printf ("##INTERNAL ERROR## in regConcNchar_CString: not enough space in target string register\n");
	  fflush (stdout);
	  return pk_false;
	}
      tgtString = (stringRegister_GET_BUF (tgtStringRegister)) + tgtFreeCharIndex;
      (void) strncat (tgtString, srcString, srcLg);
      stringRegister_ADD_USEDSIZE (tgtStringRegister, srcLg);
      stringRegister_COMPLETE_STRING (tgtStringRegister);
    }
  // DEBUG_regIndex(" exiting stringme::regConcNchar_CString aRegister",aRegister);
  return pk_true;
}

/** assign a free register to aRegister
    srcLg first characters of srcString (WARNING can contain '\0') are loaded in aRegister
    failure: no more available register or initial loading failed
 */
extern boolean
new_regNchar_CString (ptreg * aRegister, const char *srcString, int srcLg)
{
  *aRegister = allouerregchaine ();
  if (*aRegister == -1)
    return pk_false;
  if (regConcNchar_CString (*aRegister, srcString, srcLg, pk_false))
    return pk_true;
  libererregchaine (*aRegister);
  return pk_false;
}

/** append srcString to the end of aRegister
    return false if aRegister has not enough available space
*/
extern boolean
regConc_CString (ptreg aRegister, const char *srcString, boolean separate)
{
  if (srcString == (char *) 0)
    return pk_true;
  return regConcNchar_CString (aRegister, srcString, strlen (srcString), separate);
}

/** assign a free register to aRegister
    srcString is loaded in aRegister
    failure: no more available register or initial loading failed
 */
extern boolean
new_reg_CString (ptreg * aRegister, const char *srcString)
{
  *aRegister = allouerregchaine ();
  if (*aRegister == -1)
    return pk_false;
  if (regConcNchar_CString (*aRegister, srcString, strlen (srcString), pk_false))
    return pk_true;
  libererregchaine (*aRegister);
  return pk_false;
}

/** append srcRepstr CString value to the end of aRegister
    return false if aRegister has not enough available space
*/
extern boolean
regConc_repstr (ptreg aRegister, ptrepstr srcRepstr, boolean separate)
{
  ptStringRegister tgtReg;
  char *tgtString;
  TMStringCharAdr srcTMString;
  char *srcString;
  int srcLg;
  int tgtFreeCharIndex;

  // DEBUG_regIndex("entering stringme::regConc_repstr target is aRegister",aRegister);
  srcTMString = TMSTRING_CAST_TMSTRINGCHARADR (srcRepstr);
  srcLg = TMSTRINGCHARADR_GET_LENGTH (srcTMString);
  srcString = TMSTRINGCHARADR_GET_CSTRING (srcTMString);

  tgtReg = indexRegisterCAST_stringRegister (aRegister);
  tgtFreeCharIndex = stringRegister_GET_FREECHARINDEX (tgtReg);
  if (srcLg > 0)
    {
      if (separate)
	{
	  if (regConc_char (aRegister, ' '))
	    tgtFreeCharIndex = tgtFreeCharIndex + 1;
	  else
	    return pk_false;
	}
      if (tgtFreeCharIndex + srcLg > MAXLGCHAINE)
	{
	  printf ("##INTERNAL ERROR## in regConc_repstr: not enough space in target string register\n");
	  fflush (stdout);
	  return pk_false;
	}
      tgtString = stringRegister_GET_CSTRINGSTART (tgtReg);
      (void) strcat (tgtString, srcString);
      stringRegister_ADD_USEDSIZE (tgtReg, srcLg);
      stringRegister_COMPLETE_STRING (tgtReg);
    }
  // DEBUG_regIndex(" exiting stringme::regConc_repstr aRegister",aRegister);
  return pk_true;
}

/** assign a free register to aRegister
    srcRepstr is loaded in aRegister
    failure: no more available register or initial loading failed
 */
extern boolean
new_reg_repstr (ptreg * aRegister, ptrepstr srcRepstr)
{
  *aRegister = allouerregchaine ();
  if (*aRegister == -1)
    return pk_false;
  if (regConc_repstr (*aRegister, srcRepstr, pk_false))
    return pk_true;
  libererregchaine (*aRegister);
  return pk_false;
}

/** append r2 CString value to the end of r1
    return false if r1 has not enough available space
*/
extern boolean
regConc_reg (ptreg tgtRegIndex, ptreg srcRegIndex, boolean separate)
{
  ptStringRegister tgtStringRegister, srcStringRegister;
  int i, tgtLg, srcLg, tgtFreeCharIndex, srcCurrentCharIndex;

  // DEBUG_regIndex("entering stringme::regConc_reg target register",tgtRegIndex);
  // DEBUG_regIndex("entering stringme::regConc_reg source register",srcRegIndex);
  srcStringRegister = indexRegisterCAST_stringRegister (srcRegIndex);
  srcLg = stringRegister_GET_LG (srcStringRegister);
  if ((srcLg > 0))
    {
      tgtStringRegister = indexRegisterCAST_stringRegister (tgtRegIndex);
      tgtLg = stringRegister_GET_LG (tgtStringRegister);
      tgtFreeCharIndex = tgtLg + stringRegister_GET_STARTINDEX (tgtStringRegister);
      if ((tgtLg > 0))
	{
	  if (separate)
	    {
	      if (regConc_char (tgtRegIndex, ' '))
		tgtFreeCharIndex = tgtFreeCharIndex + 1;
	      else
		return pk_false;
	    }
	  if ((tgtFreeCharIndex + tgtLg + srcLg + 1) > MAXLGCHAINE)
	    {
	      printf ("##INTERNAL ERROR## in regConc_reg not enough space in target string register\n");
	      fflush (stdout);
	      return pk_false;
	    }
	}
      srcCurrentCharIndex = stringRegister_GET_STARTINDEX (srcStringRegister);
      for (i = 0; i < srcLg; i++)
	tgtStringRegister->buf[tgtFreeCharIndex + i] = srcStringRegister->buf[srcCurrentCharIndex + i];
      stringRegister_SET_USEDSIZE (tgtStringRegister, tgtLg + srcLg);
      stringRegister_COMPLETE_STRING (tgtStringRegister);
    }
  // DEBUG_regIndex(" exiting stringme::regConc_reg tgtRegIndex",tgtRegIndex);
  return pk_true;
}

/** split aRegister content by replacing each separator occurrence by '\0'
    complete the string by '\0' at index lg;
    @return the number of resulting strings
*/
extern int
regSplit (ptreg aRegister, char separator, int lg)
{
  ptStringRegister reg;
  char *stringEnd;
  int i = 1;
  char *stringStart;

  // DEBUG_regIndex("entering stringme::regSplit aRegister",aRegister);
  // printf("DEBUG entering stringme::regSplit, char separator is %c, lg is %i\n", separator, lg); fflush(stdout);
  reg = indexRegisterCAST_stringRegister (aRegister);
  stringStart = stringRegister_GET_CSTRINGSTART (reg);
  stringStart[lg] = '\0';
  while (lg > 0)
    {
      stringEnd = (char *) memchr (stringStart, '_', lg);	// pointer to the first occurrence of '_' or null pointer
      if (stringEnd == (char *) NULL)
	{
	  // DEBUG_regIndex(" exiting stringme::regSplit aRegister 1",aRegister);
	  return i;
	}
      (*stringEnd) = '\0';
      lg = lg - (stringEnd - stringStart + 1);
      stringStart = stringEnd + 1;
      i = i + 1;
    }
  stringStart = stringRegister_GET_CSTRINGSTART (indexRegisterCAST_stringRegister (aRegister));	// DEBUG
  // DEBUG_regIndex(" exiting stringme::regSplit aRegister 2",aRegister);
  return i;
}

/** @return the substring index-th substring of aRegister
    assume that such a substring exist
 */
extern ptrepstr
regGet_repstrSubstring (ptreg aRegister, int index)
{
  ptStringRegister reg;
  char *stringStart;
  int i;
  char *ch;

  reg = indexRegisterCAST_stringRegister (aRegister);
  stringStart = stringRegister_GET_CSTRINGSTART (reg);
  // printf("DEBUG entering stringme::regGet_repstrSubstring, index is %i, stringStart adress is %li, stringStart is <<%s>>\n", index, (long)stringStart, stringStart); fflush(stdout);
  for (i = 0; i < index; i++)
    stringStart = stringStart + strlen (stringStart) + 1;
  if ((TMSTRINGAREA_FREE_ptchar + strlen (stringStart) + 4) > (char *) COLLAREA_FREE_ptcoll)
    return (ptrepstr) NULL;
  ch = newTM_CSubstring2TMStringCharAdr (strlen (stringStart), stringStart);
  // printf("DEBUG  exiting stringme::regGet_repstrSubstring, ch is <<%s>>\n", ch); fflush(stdout);
  return TMSTRINGCHARADR_CAST_TMSTRING (ch);
}

extern boolean
regStore_repstr (ptreg aRegister, ptrepstr * ic)
{
  ptStringRegister reg;
  char *ch;

  reg = indexRegisterCAST_stringRegister (aRegister);
  if ((TMSTRINGAREA_FREE_ptchar + stringRegister_GET_LG (reg) + 4) > (char *) COLLAREA_FREE_ptcoll)
    return pk_false;
  ch = regGet_bufAdr (aRegister);
  ch = ch + stringRegister_GET_STARTINDEX (reg);
  ch = newTM_CSubstring2TMStringCharAdr (stringRegister_GET_LG (reg), ch);
  *ic = TMSTRINGCHARADR_CAST_TMSTRING (ch);
  return pk_true;
}


/** @return the hash value of ich
 */
extern short
ptrepstr_get_short_hashValue (ptrepstr ich)
{
  TMStringCharAdr ch;

  ch = TMSTRING_CAST_TMSTRINGCHARADR (ich);
  return (HASHCH (TMSTRINGCHARADR_GET_LENGTH (ch), TMSTRINGCHARADR_GET_CSTRING (ch)));	/* de 0 a TMStringTable_entryNumber -1 */
}

extern char
cstrgetchar (int icc, int i)
{
  char *cc;

  cc = TMSTRING_CAST_TMSTRINGCHARADR (icc);
  cc = TMSTRINGCHARADR_GET_CSTRING (cc);
  cc = cc + i - 1;
  return (*cc);
}



/*****************************************************************************/
/*									     */
/*	CSTREQ								     */
/*									     */
/* Rend vrai si chaines ic1 et ic2 sont e'gales				     */
/*									     */
/*****************************************************************************/

extern boolean
cstreq (ptrepstr ic1, ptrepstr ic2)
{
  return (ic1 == ic2);
}




/*****************************************************************************/
/*									     */
/*	CSTRCOMPAR							     */
/*									     */
/* Rend : -1 si ic1 < ic2 ; 0 si ic1 = ic2 ; 1 si ic1 > ic2		     */
/*									     */
/*****************************************************************************/

extern int
cstrcompar (ptrepstr ic1, ptrepstr ic2)
{
  return (comparchaine (ic1, ic2));
}



/** @return ths size of ic string
 */
extern int
cstrsize (ptrepstr ic)
{
  TMStringCharAdr cc;

  cc = TMSTRING_CAST_TMSTRINGCHARADR (ic);
  return (TMSTRINGCHARADR_GET_LENGTH (cc));
}

extern boolean
cstrinttostr (long long int aLLIValue, ptreg aRegister)
{
  long long int jLLIValue;
  int n, tgtFreeCharIndex, j;
  ptStringRegister reg;

  reg = indexRegisterCAST_stringRegister (aRegister);

  if (aLLIValue > 0)
    {
      n = 0;
      jLLIValue = aLLIValue;
      while (jLLIValue != 0)
	{
	  jLLIValue = (jLLIValue - jLLIValue % 10) / 10;
	  n++;
	}
    }
  else
    n = 1;

  tgtFreeCharIndex = stringRegister_GET_FREECHARINDEX (reg);
  if (tgtFreeCharIndex + n > MAXLGCHAINE)
    return pk_false;
  for (j = tgtFreeCharIndex + n - 1; j >= tgtFreeCharIndex; j--)
    {
      stringRegister_GET_BUF (reg)[j] = (char) ((aLLIValue % 10) + 48);
      aLLIValue = (aLLIValue - (aLLIValue % 10)) / 10;
    }
  stringRegister_ADD_USEDSIZE (reg, n);
  stringRegister_COMPLETE_STRING (reg);
  return pk_true;
}

extern long long int
crepstr2integer (ptrepstr ich)
{
  char *ch;

  ch = TMSTRING_CAST_TMSTRINGCHARADR (ich);
  ch = TMSTRINGCHARADR_GET_CSTRING (ch);
#ifdef __PKWINDOWS__
  return (atol ((char *) ch));	// atoll not yet implemented in visual c++ express ...
#else
  return (atoll ((char *) ch));
#endif
}
