/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


#include "tree_error.h"


// ==================================
//  MACROS AND VARIABLES DECLARATION
// ==================================
// Extern and special instanciation of c_error.

errorClass fatal_tree (et_fatal);
errorClass bug_tree (et_bug);
errorClass error_tree (et_error);
errorClass warn_tree (et_warn);
errorClass msg_tree (et_msg);


// Error management.
static int __Total__ = 0;	// errors number (expect msg).
static int __ErrorFound__ = 0;	// Found a error.

/*Sets up the error manager.*/
void
InitManager ()
{
}

/*Calls ClearErrLog().*/
void
ClearManager ()
{
  ClearErrLog ();
}

/*Sets the number of errors found at the value 0.*/
void
ClearErrLog ()
{
  __Total__ = 0;
  __ErrorFound__ = 0;
}

/*Flushes all the predefined errors.*/
void
FlushError ()
{
  msg_tree.Flush ();
  error_tree.Flush ();
  fatal_tree.Flush ();
  bug_tree.Flush ();
  warn_tree.Flush ();
}

/*Returns non zero value if at least one times error++ have been called, else returns 0.*/
int
ErrorFound ()
{
  return __ErrorFound__;
}

/*Initializes the printer informations at NULL for the file, at 0 for the line and the column. Clears the buffer.*/
void
errorClass::Clean ()
{
  SetPosition (NULL, -1, -1);
  _string[0] = '\0';
}

/*Calls FlushError().
Prints "exiting..." with the code c. Stops the execution of the program.*/
void
FATALEND (int c)
{
  FlushError ();
  msg_tree << "exiting code " << c << " ...";
  msg_tree++;
  // exit(c);
}

// =================
//  C_ERROR CLASS
// =================
/*Instanciates a new class errorClass with the code errortype et. Calls the method /SA{M(errorClass,Clean)}.*/
errorClass::errorClass (t_errortype et)
{
  _errortype = et;
  Clean ();
}

/*Sets the printer informations. The file, where the informations are, is aFile, these informations can be retrieved at the line aLine and the column aColumn.*/
void
errorClass::SetPosition (char *aFile, int aLine, int aColumn)
// /SA{M(errorClass,GetFilename)}, /SA{M(errorClass,GetLine)}, /SA{M(errorClass,GetColumn)}
{
  _filename = aFile;
  _line = aLine;
  _column = aColumn;
}

/*Returns the file name where the informations can be found.*/
char *
errorClass::GetFilename ()
{
  return _filename;
}

/*Returns the line where the informations can be found.*/
int
errorClass::GetLine ()
{
  return _line;
}

/*Returns the column where the informations can be found.*/
int
errorClass::GetColumn ()
{
  return _column;
}

/*Returns the buffer.*/
char *
errorClass::GetString ()
{
  return _string;
}

/*Puts the character c in the buffer associated to the errorClass.*/
errorClass & errorClass::operator << (const char c)
{
  char *
    ptr;

  for (ptr = _string; *ptr != '\0'; ptr++);
  sprintf (ptr, "%c", c);
  return *this;
}

/*Puts the string str in the buffer associated to the errorClass.*/
errorClass & errorClass::operator << (const char *str)
{
  char *
    ptr;

  for (ptr = _string; *ptr != '\0'; ptr++);
  sprintf (ptr, "%s", str);
  return *this;
}

/*Puts the integer i in the buffer associated to the errorClass.*/
errorClass & errorClass::operator << (const int i)
{
  char
    s[30];

  sprintf (s, "%d", i);
  return *this << s;
}

/*Puts the unsigned integer i in the buffer associated to the errorClass.*/
errorClass & errorClass::operator << (const unsigned int i)
{
  return *this << (int) i;
}

/*Puts the short integer i in the buffer associated to the errorClass.*/
errorClass & errorClass::operator << (const short i)
{
  return *this << (int) i;
}

/*Calls first the method /SA{M(errorClass,Display)}.
Up to date the compounents of the instance of the class errorClass according to its type:
- et_msg and et_warn : does nothing,
- et_error : increments the number of errors by 1 and tests if the number of errors is lesser than MAXERRORNB,
- et_fatal : prints the number of errors
- et_bug : 
Calls, if it can, the method /SA{M(errorClass,Clean)}.*/
void
errorClass::operator ++ (int)
{
  // Display the error.
  Display ();
  switch (GetErrorType ())
    {
    case et_msg:
    case et_warn:
      break;
    case et_error:
      __Total__++;
      __ErrorFound__ = 1;
      break;
    case et_fatal:
      if (__Total__)
	(msg_tree << __Total__ << " errors.")++;
      Clean ();
      FATALEND (2);
      break;
    case et_bug:
      Clean ();
      FATALEND (1);
    }
  Clean ();
}

/*Uses the informations from the method /SA{M(errorClass,SetPosition)} to print from a file the selected line and insert a cursor under the selected column.
Prints the buffer associated at the errorClass instanciation.*/
void
errorClass::Display ()
{
  int j;
  char c;
  FILE *f;

  if (GetFilename ())
    {
      cerr << "\"" << GetFilename () << "\"";
      if (GetLine () != -1)
	cerr << ", " << "Line " << GetLine () << " : ";
      else
	cerr << " : ";
    }

  switch (GetErrorType ())
    {
    case et_bug:
      cerr << "*** BUG";
      break;
    case et_fatal:
      cerr << "*** FATAL ERROR";
      break;
    case et_error:
      cerr << "*** ERROR";
      break;
    case et_warn:
      cerr << "*** WARNING";
      break;
    case et_msg:
      cerr << GetString () << "\n";
      return;
    }
  if (!GetFilename ())
    {
      cerr << " : " << GetString () << "\n";
      return;
    }
  if (GetLine () == -1)
    {
      cerr << " : " << GetString () << "\n";
      return;
    }
  cerr << " : " << GetString () << "\n";

  // Display the line file on the output.
  f = fopen (GetFilename (), "r");

  if (!f)
    {
      cerr << "(can not open " << GetFilename () << ")\n";
      fclose (f);
      return;
    }

  // Go to the line of the error.
  for (j = 1; j < GetLine ();)
    {
      if (feof (f))
	{
	  cerr << "(end of file encountered, no line " << GetLine ();
	  cerr << " in " << GetFilename () << ")\n";
	  fclose (f);
	  return;
	}
      if (fgetc (f) == '\n')
	j++;
    }
  // Output of the error line.
  for (; !feof (f);)
    {
      if ((c = fgetc (f)) == '\n')
	break;
      if (c == '\t')
	cerr << ' ';
      else
	cerr << c;
    }
  cerr << '\n';
  for (j = 0; j < GetColumn (); j++)
    cerr << ' ';
  cerr << "^\n";
  fclose (f);
}

/*Flushes the buffer. Looks at if it stays some informations in the buffer.*/
void
errorClass::Flush ()
{
  if (_string[0] == '\0')
    return;
  cerr << " bug - bug - bug - bug - bug \n";
  Display ();
}
