/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/*
  Tree Memory API built over a Tree Memory package.
*/

#define NIL_String ( (char *) NULL )

static char *this_language;	//NIL if this_language is **Unknown language**
static char *this_version;
static boolean is_errored;

typedef long int pointerAsInteger;

void
set_errored (const char *mess)
{
  printf ("ERROR during isf file loading: %s\n", mess);
  is_errored = pk_true;
}

void
set_charErrored (const char *mess, char nextChar)
{
  printf ("ERROR during isf file loading: %s. Unexpected character is %c\n", mess, nextChar);
  is_errored = pk_true;
}

void
set_operatorErrored (const char *mess, int operatorCode)
{
  printf ("ERROR during isf file loading: %s. Operator is %i\n", mess, operatorCode);
  is_errored = pk_true;
}

void
set_arityErrored (char arity)
{
  printf ("ERROR during isf file loading: illegal arity symbol <<%c>> in fix size node\n", arity);
  is_errored = pk_true;
}

int subtree_formatSave (FILE * this_file, ntrees aTree, int number, int prof);

void
ntreesFormatIndent (FILE * this_file, int prof)
{
  int i;
  for (i = 0; i <= prof; i++)
    fprintf (this_file, "  ");
}

void
attributes_formatSave (FILE * this_file, ntrees aTree, int prof)
{
  nattribute aTreeAttributes;
  if (IS_NILTREE (aTree))
    return;
  aTreeAttributes = (nattribute) getTM_firstAttribute ((trees) aTree);
  while (aTreeAttributes != NIL_nattribute)
    {
      fprintf (this_file, "{ %i ", getTagAttribute (aTreeAttributes));
      if (isIntAttribute (aTreeAttributes))
	fprintf (this_file, "%li", getIntValueAttribute (aTreeAttributes));
      else if (isTreeAttribute (aTreeAttributes))
	subtree_formatSave (this_file, getTreeValueAttribute (aTreeAttributes), 0, prof);
      else
	fprintf (this_file, "??");
      fprintf (this_file, "}\n");
      aTreeAttributes = (nattribute) getTM_succAttribute ((ptattribute) aTreeAttributes);
    }
}

/** Format saving a fix size Tree into current file. 
   @param aTree (in) a fix size tree nod with arity >0
   @param number is the number to be assigned to this node
   return the next free number
*/
int
fixSize_formatSave (FILE * this_file, ntrees aTree, int number, int prof)
{
  int arity = ntreearity (aTree);
  int i = 1;
  fprintf (this_file, "%i:[%i %i ", number, arity, (int) (noper (aTree)));
  number++;
  for (i = 1; i <= arity; i++)
    {
      fprintf (this_file, "\n");
      ntreesFormatIndent (this_file, prof);
      number = subtree_formatSave (this_file, nchild (i, aTree), number, prof + 1);
    }
  attributes_formatSave (this_file, aTree, prof + 1);
  fprintf (this_file, "]");
  return number;
}

/** Format saving a list Tree into current file. 
   @param aTree (in) a list tree node.
   @param number is the number to be assigned to this node
   return the next free number
*/
int
list_formatSave (FILE * this_file, ntrees aTree, int number, int prof)
{
  ntrees currentSon;
  int arity = ntreearity (aTree);
  fprintf (this_file, "%i:< %i ", number, (int) (noper (aTree)));
  number++;
  if (arity <= 0)
    {
      fprintf (this_file, "~>");
      return number;
    }
  currentSon = nchild (1, aTree);
  while (arity > 0)
    {
      fprintf (this_file, "\n");
      ntreesFormatIndent (this_file, prof);
      number = subtree_formatSave (this_file, currentSon, number, prof + 1);
      currentSon = ntrees_getNext (currentSon);
      arity = arity - 1;
    }
  attributes_formatSave (this_file, aTree, prof + 1);
  fprintf (this_file, ">");
  return number;
}

/** Format saving a Tree into current file. 
   @param aTree (in) a tree node.
   @param number is the number to be assigned to this node
   return the next free number
*/
void
putString_formatSave (FILE * this_file, const char *str)
{
  // printf("DEBUG entering string_formatSave\n");
  int strLg = strlen (str);
  int i;
  fputc ('"', this_file);
  for (i = 0; i < strLg; i++)
    {
      fputc (str[i], this_file);
      if (str[i] == '"')
	fputc (str[i], this_file);
    }
  fputc ('"', this_file);
}

/** Format saving a Tree into current file. 
   @param aTree (in) a tree node.
   @param number is the number to be assigned to this node
   return the next free number
*/
int
subtree_formatSave (FILE * this_file, ntrees aTree, int number, int prof)
{
  // printf("DEBUG entering subtree_formatSave\n");
  if (IS_NILTREE (aTree))
    {
      fprintf (this_file, "%i:~", number);
      return number + 1;
    }
  if (isTM_hole ((trees) aTree))
    {
      fprintf (this_file, "%i:~", number);
      return number + 1;
    }
  if (isTM_nilAtom ((trees) aTree))
    {
      fprintf (this_file, "%i:[0 %i]", number, noper (aTree));
      return number + 1;
    }
  if (isTM_charAtom ((trees) aTree))
    {
      char cc = ncharof (aTree);
      if (cc == '\'')		/* the character "'" is unparsed as '' */
	fprintf (this_file, "%i:[0 %i '']", number, noper (aTree));
      else			/* a character cc but "'" is unparsed as 'cc' */
	fprintf (this_file, "%i:[0 %i '%c']", number, noper (aTree), cc);
      return number + 1;
    }
  if (isTM_TMStringAtom ((trees) aTree))
    {				/* the characters '"' and EOL should be handled differently */
      char *str = repstr2string (nstringof (aTree));
      fprintf (this_file, "%i:[0 %i ", number, noper (aTree));
      putString_formatSave (this_file, str);
      fprintf (this_file, "]");
      return number + 1;
    }
  if (isTM_intAtom ((trees) aTree))
    {
      fprintf (this_file, "%i:[0 %i %li]", number, noper (aTree), nintof (aTree));
      return number + 1;
    }
  if (isTM_listSubtree ((trees) aTree))
    return list_formatSave (this_file, aTree, number, prof);
  return fixSize_formatSave (this_file, aTree, number, prof);
}

/* LOADING */
boolean subtree_formatLoad (FILE * this_file, ntrees * aTree, boolean inner);

/** Skip spaces in this_file 
   @param this_file contains this_file=PREFIX ?! char* EOF, PREFIX has been read
   @return nextChar when this_file=PREFIX ? space* ! nextChar char* EOF, nextChar is not a space
   @return EOF     when this_file=PREFIX ? space* ! EOF
 */
char
skipSpaces (FILE * this_file)	/* PLG checked */
{
  char nextChar;
  do
    {
      nextChar = fgetc (this_file);
    }
  while ((nextChar == ' ') || (nextChar == '\n'));
  ungetc (nextChar, this_file);
  return nextChar;
}

/** Skip to end of char constant: a correct char representation is assumed to correctly recover error nestings 
   @param this_file contains this_file=PREFIX ' ?! char* EOF
   @return true when this_file=PREFIX ' ? ' ! char* EOF
   @return true when this_file=PREFIX ' ? char1 ' ! char* EOF, char1 is not "'"
   @return false when this_file=PREFIX ' ? ! EOF
   @return false when this_file=PREFIX ' ? char1 ! EOF char1 is not "'"
   @return false when this_file=PREFIX ' ? char1 ! char2 char* EOF, none of char1, char2 is "'"
 */
boolean
skipToEndOfCharAtom (FILE * this_file)	/* PLG checked */
{
  char curChar;
  curChar = fgetc (this_file);
  if (curChar == '\'')
    return pk_true;
  if (curChar == EOF)
    return pk_false;
  curChar = fgetc (this_file);
  if (curChar == '\'')
    return pk_true;
  if (curChar == EOF)
    return pk_false;
  ungetc (curChar, this_file);
  return pk_false;
}

/** Skip to end of string constant
   @param this_file contains this_file=PREFIX " ?! char* EOF
   @return true when this_file=PREFIX " ?  char* "  char* !EOF, '"' does not occur or occurs doubled in first occurrence of char* 
   @return false when this_file=PREFIX " ? char* !EOF, '"' does not occur or occurs doubled in char* 
 */
boolean
skipToEndOfStringAtom (FILE * this_file)	/* PLG checked */
{
  char curChar;
  curChar = fgetc (this_file);
  while (curChar != EOF)
    if ((curChar == '"'))
      {
	curChar = fgetc (this_file);
	if ((curChar != '"'))
	  {
	    ungetc (curChar, this_file);
	    return pk_true;
	  }
      }
    else
      curChar = fgetc (this_file);
  return pk_false;
}

boolean skipToEndOfListFormatNode (FILE * this_file);


/** Skip to end of fix format node
   @param this_file=PREFIX space* digit+ : [ char* ? ! char* EOF
   @return true when this_file=PREFIX space* digit+ : [ char* ? SKIPPED ] ! char* EOF,
                                          where parenthesis are well balanced in SKIPPED=char*
   @return false otherwise, then this_file=this_file=PREFIX  space* digit+ : [ char* ? char* ! EOF 
                                          where parenthesis are not well balanced in "[ char* ? char*"
 */
boolean
skipToEndOfFixFormatNode (FILE * this_file)	/* PLG check occurrences */
{
  char curChar;
  do
    {
      curChar = fgetc (this_file);
      if (curChar == ']')
	return pk_true;
      if (curChar == '>')
	{
	  ungetc (curChar, this_file);
	  return pk_false;
	}
      else if (curChar == '\'')
	{			/* skip '[', ']',  '<', '>', */
	  if (skipToEndOfCharAtom (this_file));
	}			/* this_file=PREFIX ' ? char1 ' ! char* EOF, char1 is not "'" */
      else if (curChar == '"')
	{			/* skip occurrences of [, ],  <, > in the string constant */
	  if (skipToEndOfStringAtom (this_file));
	}
      else if (curChar == '<')
	{
	  if (!skipToEndOfListFormatNode (this_file))
	    {
	      curChar = fgetc (this_file);
	      return (curChar == ']');
	    }
	}
      else if (curChar == '[')
	{
	  if (!skipToEndOfFixFormatNode (this_file))
	    return pk_false;
	}
    }
  while ((curChar != EOF));
  return pk_false;
}

/** Skip to end of list format node
   @param this_file=PREFIX space* digit+ : < char* ? ! char* EOF
   @return true when this_file=PREFIX space* digit+ : < char* ? SKIPPED > ! char* EOF, 
                                        where parenthesis are well balanced in SKIPPED=char*
   @return false otherwise, then this_file=PREFIX  space* digit+ : < char* ? char* ! EOF
                                          where parenthesis are not well balanced in "< char* ? char*"
 */
boolean
skipToEndOfListFormatNode (FILE * this_file)	/* PLG check occurrences */
{
  char curChar;
  do
    {
      curChar = fgetc (this_file);
      if (curChar == '>')
	return pk_true;
      if (curChar == ']')
	{
	  ungetc (curChar, this_file);
	  return pk_false;
	}
      if (curChar == '\'')
	{			/* skip '[', ']',  '<', '>', */
	  if (skipToEndOfCharAtom (this_file));
	}			/* this_file=PREFIX ' ? char1 ' ! char* EOF, char1 is not "'" */
      else if (curChar == '"')
	{			/* skip occurrences of [, ],  <, > in the string constant */
	  if (skipToEndOfStringAtom (this_file));
	}
      else if (curChar == '<')
	{
	  if (!skipToEndOfListFormatNode (this_file))
	    return pk_false;
	}
      /* ( curChar=='>' ) is ignored */
      else if (curChar == '[')
	{
	  if (!skipToEndOfFixFormatNode (this_file))
	    {
	      curChar = fgetc (this_file);
	      return (curChar == '>');
	    }
	}
    }
  while ((curChar != EOF));
  return pk_false;
}

/** Skip to end of fix format node
   @param this_file=PREFIX space* digit+ : [ char* ? ! char* EOF
   @return true when this_file=PREFIX space* digit+ : [ char* ? SKIPPED ] ! char* EOF,
                                          where parenthesis are well balanced in SKIPPED=char*
   @return false otherwise, then this_file=this_file=PREFIX  space* digit+ : [ char* ? char* ! EOF 
                                          where parenthesis are not well balanced in "[ char* ? char*"
 */
boolean
skipToEndOfFormatNode (FILE * this_file)	/* PLG check occurrences */
{
  char curChar;
  do
    {
      curChar = fgetc (this_file);
      if ((curChar == ']') || (curChar == '>'))
	return pk_true;
      if (curChar == '\'')
	{			/* skip '[', ']',  '<', '>', */
	  if (skipToEndOfCharAtom (this_file));
	}			/* this_file=PREFIX ' ? char1 ' ! char* EOF, char1 is not "'" */
      else if (curChar == '"')
	{			/* skip occurrences of [, ],  <, > in the string constant */
	  if (skipToEndOfStringAtom (this_file));
	}
      else if (curChar == '<')
	{
	  if (!skipToEndOfListFormatNode (this_file));
	}
      else if (curChar == '[')
	{
	  if (!skipToEndOfFixFormatNode (this_file));
	}
    }
  while ((curChar != EOF));
  return pk_false;
}

/** Format atom completion from current this_file. 
   @param this_file contains PREFIX ?! char* EOF, 
   @return true iff this_file=PREFIX ? space* ] ! char* EOF
 */
boolean
node_atomLoadCompletion (FILE * this_file)	/* PLG  occurrences to be checked */
{
  char nextChar;

  nextChar = skipSpaces (this_file);	/* this_file=PREFIX ? space* ! nextChar char* EOF, nextChar is not a space */
  if (nextChar == ']')
    {
      nextChar = fgetc (this_file);
      return pk_true;
    }
  set_charErrored ("atom node completion", nextChar);
  return skipToEndOfFixFormatNode (this_file);
}

/** Format loading a long int atom completion from current this_file. 
   @param this_file=PREFIX space* digit+ : [ 0 space* digit+ space* ?! digit char* EOF
   assign the parsed tree to aTree
   @return true if analysis succeeds or if an error is recovered
       this_file contains PREFIX space* digit+ : [ 0 space* digit+ space* ? digit+ space* ] ! char* EOF
       where VALUE is a sequence of digits
*/
boolean
int_formatLoadCompletion (noperators curOperator, FILE * this_file, ntrees * aTree)
{
  long int intValue;

  // printf("DEBUG entering int_formatLoadCompletion\n");
  /* this_file=PREFIX space* digit+ : [ 0 ? space* digit+ space* ! digit char* SUFFIX1 EOF */
  if ((fscanf (this_file, "%li", &intValue)) == 1)
    {				/* this_file=PREFIX space* digit+ : [ 0 ? space* digit+ space* digit+ ! char* SUFFIX1 EOF */
      (*aTree) = natomint (curOperator, intValue);
      return node_atomLoadCompletion (this_file);
    }				/* this_file=PREFIX space* digit+ : [ 0 ? space* digit+ space* digit+ space* ] ! SUFFIX1 EOF */
  set_errored ("integer atom syntax");
  (*aTree) = (ntrees) makeTM_hole ();
  return skipToEndOfFixFormatNode (this_file);	/* pk_true if this_file=PREFIX space* digit+ : [ char* ? SKIPPED ] ! char* EOF */
}

/** Format loading a char atom completion from current this_file. 
   @param this_file contains PREFIX NUMBER:[0 space* curOperator space* 'SUFFIX EOF, the character "'" has been read
   assign the parsed tree to aTree
   @return true if analysis succeeds
       this_file contains PREFIX NUMBER:[0 space* curOperator space* 'C' space* ] ?! SUFFIX EOF
                       or PREFIX NUMBER:[0 space* curOperator space* '' space* ] ?! SUFFIX EOF
       where C is a character
       */
boolean
char_formatLoadCompletion (noperators curOperator, FILE * this_file, ntrees * aTree)
{
  /* this_file=PREFIX space* digit+ : [ 0 space* digit+ space* ' ?! char* SUFFIX1 EOF */
  char curChar = fgetc (this_file);
  /* this_file=PREFIX space* digit+ : [ 0 space* digit+ space* ' ? char ! char* SUFFIX1 EOF */
  // printf("DEBUG char_formatLoadCompletion\n");
  if (curChar == EOF)
    {
      set_errored ("unexpected EOF when reading a char atom");
      (*aTree) = natomchar (curOperator, '\0');
      return pk_false;
    }
  (*aTree) = natomchar (curOperator, curChar);
  if (curChar == '\'')		/* the value is the character "'" */
    return node_atomLoadCompletion (this_file);	/* this_file=PREFIX space* digit+ : [ 0 space* digit+ space* '' space* ] ?! SUFFIX1 EOF */
  curChar = fgetc (this_file);
  if (curChar == '\'')		/* this_file=PREFIX space* digit+ : [ 0 space* digit+ space* ' ? char ' ! SUFFIX1 EOF */
    return node_atomLoadCompletion (this_file);	/* this_file=PREFIX space* digit+ : [ 0 space* digit+ space* 'char' space* ] ?! SUFFIX1 EOF */
  set_charErrored ("char atom value completion", curChar);
  (*aTree) = natomchar (curOperator, '\0');
  return skipToEndOfFixFormatNode (this_file);	/* pk_true if this_file=PREFIX space* digit+ : [ char* ? SKIPPED ] ! char* EOF */
}

/** Format loading a string atom completion from current this_file. 
   @param this_file contains PREFIX NUMBER:[0 space* curOperator space* "SUFFIX EOF, the character '"' has been read
   assign the parsed tree to aTree
   @return true if analysis succeeds
       this_file contains PREFIX NUMBER:[0 space* curOperator space* "STRING" space* ] SUFFIX1 EOF, ']' has been read
       where VALUE is a '"' free sequence of characters 
       */
boolean
getString_formatLoad (FILE * this_file, ptrepstr * loadedString)
{
  int i = 0;
  char curStringBuf[1024];
  char curChar;
  /* this_file=PREFIX space* digit+ : [ 0 space* digit+ space* " ?! char* space* ] SUFFIX1 EOF */
  // printf("DEBUG string_formatLoadCompletion\n");
  curChar = fgetc (this_file);
  while (curChar != EOF)
    {				/* this_file=PREFIX space* digit+ : [ 0 space* digit+ space* " ? char* char ! char* space* ] SUFFIX1 EOF */
      if (curChar == '"')
	{			/* this_file=PREFIX space* digit+ : [ 0 space* digit+ space* " ? char* " ! char* space* ] SUFFIX1 EOF */
	  curChar = fgetc (this_file);
	  if (curChar == EOF)
	    {
	      curStringBuf[i] = '\0';
	      (*loadedString) = string2repstr (curStringBuf);
	      return pk_false;
	    }
	  if (curChar != '"')
	    {			/* this_file=PREFIX space* digit+ : [ 0 space* digit+ space* " ? char* " char ! char* space* ] SUFFIX1 EOF */
	      ungetc (curChar, this_file);
	      curStringBuf[i] = '\0';
	      (*loadedString) = string2repstr (curStringBuf);
	      return pk_true;
	    }
	}			/* this_file=PREFIX space* digit+ : [ 0 space* digit+ space* STRING space* ] ?! SUFFIX1 EOF */
      curStringBuf[i] = curChar;
      curChar = fgetc (this_file);
      i++;
    }
  return pk_false;
}

/** Format loading a string atom completion from current this_file. 
   @param this_file contains PREFIX NUMBER:[0 space* curOperator space* "SUFFIX EOF, the character '"' has been read
   assign the parsed tree to aTree
   @return true if analysis succeeds
       this_file contains PREFIX NUMBER:[0 space* curOperator space* "STRING" space* ] SUFFIX1 EOF, ']' has been read
       where VALUE is a '"' free sequence of characters 
       */
boolean
string_formatLoadCompletion (noperators curOperator, FILE * this_file, ntrees * aTree)
{
  ptrepstr loadedString;
  if (getString_formatLoad (this_file, &loadedString))
    {
      (*aTree) = natomstring (curOperator, loadedString);
      return node_atomLoadCompletion (this_file);
    }				/* this_file=PREFIX space* digit+ : [ 0 space* digit+ space* STRING space* ] ?! SUFFIX1 EOF */
  (*aTree) = (ntrees) makeTM_hole ();
  return pk_false;
}

/** Format loading completion of a fix size Tree from current this_file. 
   @param this_file=PREFIX space* digit+ :[ char* ?! char* EOF
   @param aTree Not implemented: aTree is completed with attributes
   @return true : this_file=PREFIX space* digit+ :[ char* ? space* ] ! char* EOF
*/
boolean
fixSize_formatLoadCompletion (FILE * this_file, ntrees aTree)	/* PLG skipToEndOfFixFormatNode  to be checked, occurrences to be checked */
{
  char nextChar;
  /* read attributes and then */
  // printf("DEBUG fixSize_formatLoadCompletion\n");
  nextChar = skipSpaces (this_file);
  if (nextChar == ']')
    {				/* this_file=PREFIX  space* digit+ :[ char* ? space* ! ] char* EOF */
      nextChar = fgetc (this_file);	/* actual reading of nextChar */
      return pk_true;
    }
  set_charErrored ("wrong completion of fix size node", nextChar);
  return skipToEndOfFixFormatNode (this_file);	/* pk_true if this_file=PREFIX space* digit+ : [I char* ? SKIPPED ] ! char* EOF */
}

/** Format loading completion of a list Tree from current this_file. 
   @param this_file=PREFIX space* digit+ :< char* ?! char* EOF
   @param aTree Not implemented: aTree is completed with attributes
   @return true : this_file=PREFIX space* digit+ :< char* ? space* > ! char* EOF
*/
boolean
list_formatLoadCompletion (FILE * this_file, ntrees aTree)	/* PLG skipToEndOfListFormatNode to be checked, occurrences to be checked */
{
  char nextChar;
  /* read attributes and then */
  // printf("DEBUG list_formatLoadCompletion \n");
  nextChar = skipSpaces (this_file);
  if (nextChar == '>')
    {				/* this_file=PREFIX space* digit+ :< char* ? space* ! > char* EOF */
      nextChar = fgetc (this_file);	/* actual reading of nextChar */
      return pk_true;
    }
  set_charErrored ("wrong completion of list node", nextChar);
  return skipToEndOfListFormatNode (this_file);
}

/** Format loading a nullary Tree from current this_file. 
   @param this_file=PREFIX space* digit+ :[0 ?! char* EOF
   assign the parsed tree to aTree
   @return true if analysis succeeds
       this_file contains PREFIX space* digit+ :[0 ? space* digit+ space* VALUE space*] ! char* EOF
       where VALUE can be 
           empty
           a character 'c'
           a string "char*"
           a long integer digit+

*/
boolean
nullary_formatLoad (char nextChar, noperators curOperator, FILE * this_file, ntrees * aTree)	/* PLG skipToEndOfFixFormatNode to be checked, occurrences to be checked */
{
  /* this_file=PREFIX space* digit+ :[0 ?! char* EOF */
  if (nextChar == EOF)
    {				/* this_file=PREFIX space* digit+ : [ 0 ? space* digit+ space* ! EOF */
      set_errored ("unexpected EOF when reading an atom");
      (*aTree) = (ntrees) makeTM_hole ();
      return pk_false;
    }
  /* this_file=PREFIX space* digit+ : [ 0 ? space* digit+ space* ! nextChar char* EOF, nextChar is not a space */
  if (nextChar == ']')
    {				/* this_file=PREFIX space* digit+ : [ 0 ? space* digit+ space* ! ]  char* EOF */
      nextChar = fgetc (this_file);
      (*aTree) = natomnil (curOperator);
      /* this_file=PREFIX space* digit+ : [ 0 ? space*  digit+ space* ] ! char* EOF */
      return pk_true;
    }
  if (isdigit (nextChar))
    {				/* this_file=PREFIX space* digit+ : [ 0 ? space* digit+ space* ! digit char* EOF */
      return int_formatLoadCompletion (curOperator, this_file, aTree);
    }
  if (nextChar == '\'')
    {				/* this_file=PREFIX space* digit+ : [ 0 ? space* digit+ space* ! ' char* EOF */
      nextChar = fgetc (this_file);	/* to get ' character */
      /* this_file=PREFIX space* digit+ : [ 0 ? space* digit+ space* ' ! char* EOF */
      return char_formatLoadCompletion (curOperator, this_file, aTree);
    }
  if (nextChar == '"')
    {				/* this_file=PREFIX space* digit+ : [ 0 ? space* digit+ space* ! " char* EOF */
      nextChar = fgetc (this_file);	/* to get " character */
      /* this_file=PREFIX space* digit+ : [ 0 ? space* digit+ space* " ! char* EOF */
      return string_formatLoadCompletion (curOperator, this_file, aTree);
    }				/* this_file=PREFIX space* digit+ : [ 0 space* digit+ space* "char*" space*]?! SUFFIX1 EOF */
  /* this_file=PREFIX space* digit+ : [ 0 ? space* digit+ space* ! char+ EOF, first char in char+ EOF is none of a space, a digit, ] , ' character, " character */
  set_charErrored ("atom value reading", nextChar);
  (*aTree) = (ntrees) makeTM_hole ();
  return skipToEndOfFixFormatNode (this_file);
}

/** Format loading a non nul arity fix size Tree from current this_file. 
   @param this_file contains PREFIX NUMBER:[CURCHAR SUFFIX EOF, CURCHAR has been read
   assign the parsed tree to aTree
   @return true if analysis succeeds
       this_file contains PREFIX NUMBER:[CURCHAR OPERATOR SUBTREES] SUFFIX1 EOF, ] has been read
       where SUBTREES is
           SUBTREE if CURCHAR is the character 1
           ...
           4 occurrences of  SUBTREE  if CURCHAR is the character 4
*/
boolean
unary_formatLoad (int curOperator, FILE * this_file, ntrees * aTree)
{
  ntrees sonTree;

  // printf("DEBUG unary_formatLoad,curOperator  is %i\n", curOperator);
  if (subtree_formatLoad (this_file, &sonTree, pk_true))
    {
      (*aTree) = nmakeunary (curOperator, sonTree);
      return fixSize_formatLoadCompletion (this_file, *aTree);
    }
  set_operatorErrored ("missing a son in unary node", curOperator);
  (*aTree) = nmakeunary (curOperator, (ntrees) makeTM_hole ());
  return skipToEndOfFixFormatNode (this_file);	/* pk_true if this_file=PREFIX space* digit+ : [I char* ? SKIPPED ] ! char* EOF */
}

/** Format loading a non nul arity fix size Tree from current this_file. 
   @param this_file contains PREFIX NUMBER:[CURCHAR SUFFIX EOF, CURCHAR has been read
   assign the parsed tree to aTree
   @return true if analysis succeeds
       this_file contains PREFIX NUMBER:[CURCHAR OPERATOR SUBTREES] SUFFIX1 EOF, ] has been read
       where SUBTREES is
           SUBTREE if CURCHAR is the character 1
           ...
           4 occurrences of  SUBTREE  if CURCHAR is the character 4
*/
boolean
binary_formatLoad (int curOperator, FILE * this_file, ntrees * aTree)
{
  ntrees sonTree1, sonTree2;

  // printf("DEBUG binary_formatLoad,curOperator  is %i\n", curOperator);
  if (subtree_formatLoad (this_file, &sonTree1, pk_true))
    {
      if (subtree_formatLoad (this_file, &sonTree2, pk_true))
	{
	  // nDEBUG_subtree(4,2,"     in  intetrees_m::binary_formatLoad 1, son subtree 1 is", sonTree1); 
	  // nDEBUG_subtree(4,2,"     in  intetrees_m::binary_formatLoad 1, son subtree 2 is", sonTree2); 
	  (*aTree) = makeTM_binaryTree (curOperator, sonTree1, sonTree2);
	  return fixSize_formatLoadCompletion (this_file, *aTree);
	}
      else
	sonTree2 = (ntrees) makeTM_hole ();
    }
  else
    {
      sonTree1 = (ntrees) makeTM_hole ();
      sonTree2 = (ntrees) makeTM_hole ();
    }
  set_operatorErrored ("missing a son in binary node", curOperator);
  // nDEBUG_subtree(4,2,"     in  intetrees_m::binary_formatLoad 2, son subtree 1 is", sonTree1); 
  // nDEBUG_subtree(4,2,"     in  intetrees_m::binary_formatLoad 2, son subtree 2 is", sonTree2); 
  (*aTree) = makeTM_binaryTree (curOperator, sonTree1, sonTree2);
  return skipToEndOfFixFormatNode (this_file);	/* pk_true if this_file=PREFIX space* digit+ : [I char* ? SKIPPED ] ! char* EOF */
}

/** Format loading a non nul arity fix size Tree from current this_file. 
   @param this_file contains PREFIX NUMBER:[CURCHAR SUFFIX EOF, CURCHAR has been read
   assign the parsed tree to aTree
   @return true if analysis succeeds
       this_file contains PREFIX NUMBER:[CURCHAR OPERATOR SUBTREES] SUFFIX1 EOF, ] has been read
       where SUBTREES is
           SUBTREE if CURCHAR is the character 1
           ...
           4 occurrences of  SUBTREE  if CURCHAR is the character 4
*/
boolean
ternary_formatLoad (int curOperator, FILE * this_file, ntrees * aTree)
{
  ntrees sonTree1, sonTree2, sonTree3;

  // printf("DEBUG ternary_formatLoad,curOperator  is %i\n", curOperator);
  if (subtree_formatLoad (this_file, &sonTree1, pk_true))
    {
      if (subtree_formatLoad (this_file, &sonTree2, pk_true))
	{
	  if (subtree_formatLoad (this_file, &sonTree3, pk_true))
	    {
	      (*aTree) = nmaketernary (curOperator, sonTree1, sonTree2, sonTree3);
	      return fixSize_formatLoadCompletion (this_file, *aTree);
	    }
	  else
	    sonTree3 = (ntrees) makeTM_hole ();
	}
      else
	{
	  sonTree2 = (ntrees) makeTM_hole ();
	  sonTree3 = (ntrees) makeTM_hole ();
	}
    }
  else
    {
      sonTree1 = (ntrees) makeTM_hole ();
      sonTree2 = (ntrees) makeTM_hole ();
      sonTree3 = (ntrees) makeTM_hole ();
    }
  set_operatorErrored ("missing a son in ternary node", curOperator);
  (*aTree) = nmaketernary (curOperator, sonTree1, sonTree2, sonTree3);
  return skipToEndOfFixFormatNode (this_file);	/* pk_true if this_file=PREFIX space* digit+ : [I char* ? SKIPPED ] ! char* EOF */
}

/** Format loading a non nul arity fix size Tree from current this_file. 
   @param this_file contains PREFIX NUMBER:[CURCHAR SUFFIX EOF, CURCHAR has been read
   assign the parsed tree to aTree
   @return true if analysis succeeds
       this_file contains PREFIX NUMBER:[CURCHAR OPERATOR SUBTREES] SUFFIX1 EOF, ] has been read
       where SUBTREES is
           SUBTREE if CURCHAR is the character 1
           ...
           4 occurrences of  SUBTREE  if CURCHAR is the character 4
*/
boolean
quaternary_formatLoad (int curOperator, FILE * this_file, ntrees * aTree)
{
  ntrees sonTree1, sonTree2, sonTree3, sonTree4;

  // printf("DEBUG quaternary_formatLoad,curOperator  is %i\n", curOperator);
  if (subtree_formatLoad (this_file, &sonTree1, pk_true))
    {
      if (subtree_formatLoad (this_file, &sonTree2, pk_true))
	{
	  if (subtree_formatLoad (this_file, &sonTree3, pk_true))
	    {
	      if (subtree_formatLoad (this_file, &sonTree4, pk_true))
		{
		  (*aTree) = nmakequaternary (curOperator, sonTree1, sonTree2, sonTree3, sonTree4);
		  return fixSize_formatLoadCompletion (this_file, *aTree);
		}
	      else
		sonTree4 = (ntrees) makeTM_hole ();
	    }
	  else
	    {
	      sonTree3 = (ntrees) makeTM_hole ();
	      sonTree4 = (ntrees) makeTM_hole ();
	    }
	}
      else
	{
	  sonTree2 = (ntrees) makeTM_hole ();
	  sonTree3 = (ntrees) makeTM_hole ();
	  sonTree4 = (ntrees) makeTM_hole ();
	}
    }
  else
    {
      sonTree1 = (ntrees) makeTM_hole ();
      sonTree2 = (ntrees) makeTM_hole ();
      sonTree3 = (ntrees) makeTM_hole ();
      sonTree4 = (ntrees) makeTM_hole ();
    }
  set_operatorErrored ("missing a son in quaternary node", curOperator);
  (*aTree) = nmakequaternary (curOperator, sonTree1, sonTree2, sonTree3, sonTree4);
  return skipToEndOfFixFormatNode (this_file);	/* pk_true if this_file=PREFIX space* digit+ : [I char* ? SKIPPED ] ! char* EOF */
}

/** Format loading a non nul arity fix size Tree from current this_file. 
   @param this_file contains PREFIX NUMBER:[CURCHAR SUFFIX EOF, CURCHAR has been read
   assign the parsed tree to aTree
   @return true if analysis succeeds
       this_file contains PREFIX NUMBER:[CURCHAR OPERATOR SUBTREES] SUFFIX1 EOF, ] has been read
       where SUBTREES is
           SUBTREE if CURCHAR is the character 1
           ...
           4 occurrences of  SUBTREE  if CURCHAR is the character 4
*/
boolean
fixSize_formatLoad (FILE * this_file, ntrees * aTree)
{
  char arityChar;
  int curOperator;
  char nextChar;

  // printf("DEBUG entering fixSize_formatLoad\n");
  nextChar = skipSpaces (this_file);
  if (isdigit (nextChar))
    {
      arityChar = fgetc (this_file);
      nextChar = skipSpaces (this_file);
      if ((fscanf (this_file, "%i", &curOperator)) == 1)
	{
	  nextChar = skipSpaces (this_file);	/* this_file=PREFIX space* digit+ : [ 0 ? space* digit+ space* ! char* EOF, first char in char* EOF is not a space */
	  if (arityChar == '0')	/* this_file=PREFIX ? space* digit+ : [ 0 ! FIX-NODE ] SUFFIX1 EOF */
	    return nullary_formatLoad (nextChar, curOperator, this_file, aTree);	/* this_file=PREFIX space* digit+ :[0 space* OPERATOR space* VALUE space*] SUFFIX1 EOF */
	  if (arityChar == '1')
	    return unary_formatLoad (curOperator, this_file, aTree);
	  if (arityChar == '2')
	    return binary_formatLoad (curOperator, this_file, aTree);
	  if (arityChar == '3')
	    return ternary_formatLoad (curOperator, this_file, aTree);
	  if (arityChar == '4')
	    return quaternary_formatLoad (curOperator, this_file, aTree);
	  set_arityErrored (arityChar);
	}
      else
	set_errored ("fix node. Expected operator is missing");
    }
  else
    set_arityErrored (nextChar);
  (*aTree) = (ntrees) makeTM_hole ();
  return skipToEndOfFixFormatNode (this_file);	/* pk_true if this_file=PREFIX space* digit+ : [I char* ? SKIPPED ] ! char* EOF */
}

/** Format loading a list Tree from current this_file. 
   @param this_file contains PREFIX NUMBER:<SUFFIX EOF, CURCHAR has been read
   assign the parsed tree to aTree
   @return true if analysis succeeds
       this_file contains PREFIX NUMBER:[OPERATOR SUBTREES> SUFFIX1 EOF, > has been read
       where SUBTREES is a possibly empty sequence of SUBTREE
*/
boolean
list_formatLoad (FILE * this_file, ntrees * aTree)
{
  int curOperator;
  char nextChar;
  ntrees curSon;

  // printf("DEBUG list_formatLoad\n");
  nextChar = skipSpaces (this_file);
  if ((fscanf (this_file, "%i", &curOperator)) == 1)
    {
      if (subtree_formatLoad (this_file, &curSon, pk_true))
	{
	  (*aTree) = nmakelist (curOperator, curSon);
	  while (subtree_formatLoad (this_file, &curSon, pk_true))
	    {
	      (*aTree) = ntreeList_appendLast (*aTree, curSon);
	    }
	}
      else
	(*aTree) = nmakemptylist (curOperator);
      return list_formatLoadCompletion (this_file, *aTree);
    }

  set_errored ("list node operator");
  (*aTree) = (ntrees) makeTM_hole ();
  return skipToEndOfListFormatNode (this_file);
}

/** Format loading a non empty Tree from this_file. 
   @param this_file=PREFIX ?! SUFFIX EOF
   @param aTree is assigned the loaded tree 
   success: this_file=PREFIX space* digit+:NODE ?! SUFFIX1 EOF
        or  this_file=PREFIX space* ?! EOF
   where NODE=~
      or NODE=[O ATOM]
      or NODE=[ARITY FIX-NODE], with ARITY = 1, 2, 3, or 4
      or NODE=<LIST-NODE>
   return pk_true when a node has been read;
   return pk_true when a node reading started, inner is pk_true and the read content is well balanced and file loading is not completed
*/
boolean
subtree_formatLoad (FILE * this_file, ntrees * aTree, boolean inner)
{
  int ordinal;
  char curChar, nextChar;
  boolean started = pk_false;
  // printf("DEBUG entering subtree_formatLoad\n");
  nextChar = skipSpaces (this_file);

  /* node number prefix */
  if (isdigit (nextChar))
    {
      started = pk_true;
      if ((fscanf (this_file, "%i", &ordinal)) == 1)
	{			/* this_file=PREFIX ? space* digit+ ! :NODE SUFFIX1 EOF */
	  nextChar = skipSpaces (this_file);
	  if (nextChar == ':')
	    {
	      curChar = fgetc (this_file);
	      nextChar = skipSpaces (this_file);
	    }
	  else
	    set_charErrored ("after a node number", nextChar);
	}
      else
	set_errored ("node number syntax");
    }

  /* node content reading */
  curChar = fgetc (this_file);
  if (curChar == '~')
    {				/* this_file=PREFIX < space* digit+ : ~  ! SUFFIX1 EOF */
      (*aTree) = (ntrees) makeTM_hole ();
      return pk_true;
    }
  if (curChar == '[')
    {				/* this_file=PREFIX ? space* digit+ : [ ! FIX-NODE ] SUFFIX1 EOF */
      return fixSize_formatLoad (this_file, aTree);
    }
  if (curChar == '<')		/* this_file=PREFIX ? space* digit+ : < ! LIST-NODE > SUFFIX1 EOF */
    return list_formatLoad (this_file, aTree);

  /* node not read */
  if (started)
    {
      (*aTree) = (ntrees) makeTM_hole ();
      set_charErrored ("node discriminant", nextChar);
    }
  if ((curChar == '>') || (curChar == ']'))
    {
      ungetc (curChar, this_file);
      if (!inner)
	set_charErrored ("node start", nextChar);
      return started;
    }

  /* error recovering */
  ungetc (curChar, this_file);
  if (skipToEndOfFormatNode (this_file));
  return pk_false;
}
