/*
 *
 * Input/output routines routines for the program pKPrototypeExtract
 * It is derived from Aam Bryant io.main (see below)
 * Copyright (c) 2010, 2011 by Loc Besnard, Paul Le Guernic for modifications
 *
 */
/*
 *
 * cextract input/output routines.
 *
 * Copyright (c) 1992 by Adam Bryant
 *s
 * See the Copyright notice in the file parse.c or in the manual page.
 *
 */

#include "HD_STAT.h"		/* system dependent */

/* ========================= DATA STRUCTURES ========================= */
static int line_count = 1;
static FILE *fpin = NULL;
static char *in_filePath = NULL;

/** return the current line number in input file */
int
get_lineCount ()
{
  return line_count;
}

/* ========================= DEBUG FUNCTIONS ========================= */
void
DEBUG_lineScale (char *prefix)
{
  fprintf (stderr, "%s          1         2         3         4         5         6         7         8         9\n", prefix);
  fprintf (stderr, "%s0123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789\n", prefix);
}

void
DEBUG_string (char *prefix1, char *prefix2, char *str_name, char *str_val)
{
  if (str_val == NULL)
    fprintf (stderr, "%s %s %s is NULL\n", prefix1, prefix2, str_name);
  else
    fprintf (stderr, "%s %s %s is <<%s>>\n", prefix1, prefix2, str_name, str_val);
}

/* ========================= ERROR FUNCTIONS ========================= */
void
exitERR (const char *mess)
{
  fprintf (stderr, "Exit on ERROR: %s\n", mess);
  exit (1);
}

static void
exitERR_format (const char *format, const char *str)
{
  fprintf (stderr, "Exit on ERROR: ");
  fprintf (stderr, format, str);
  exit (1);
}

void
exitOn_memoryAllocationErr ()
{
  exitERR ("memory allocation failure");
}

static void
exitERR_onEof (const char *req_str)
{
  exitERR_format ("unexpected end of file, <<%s>> was required", req_str);
}

extern void
exitERR_line (const char *mess)
{
  fprintf (stderr, "line %i, ", line_count);
  exitERR (mess);
}

void
warningCommentLost ()
{
  fprintf (stderr, "Warning line %i, a comment is lost\n", line_count);
}

/* ========= ========= ========= ========= ========= ========= ========= =========
 *                     AUXILIARY                     FUNCTIONS 
 * ========= ========= ========= ========= ========= ========= ========= =========
 */

/** 
 * input: FPIN = [Character*][eol]~[Character*][EOF]
 * output: FPIN  = [Character*][eol]#~[Character*][EOF]
 */
static void
register_physicalEol ()
{
  line_count++;
}

/** 
 * input: FPIN = [Character*][eol]~[Character*][EOF]
 * output: FPIN  = [Character*][eol]#~[Character*][EOF]
 */
void
register_logicalEol ()
{
  line_count++;
}

/** skip terminal blanks in str
 * success: <STR> = [Blank]*[Character_but_Blank]+ [Blank]*[eol]
 *              returns the index of the last [Character_but_Blank] in <str>
 *          <STR> = [Blank]*[eol]
 *              returns -1
 * failure: <STR> is empty
 *          <STR>  = [Character]* [Character_but_eol]
 *          exit
 */
static int
find_last_nonBlank (char *str)
{
  int index = strlen (str) - 2;
  if (index < -1)
    exitERR ("empty line in find_last_nonBlank");
  if (str[index + 1] != '\n')
    exitERR ("not eol terminating line in find_last_nonBlank");
  while (index >= 0)
    {
      if (isblank (str[index]))
	index--;
      else
	return index;
    }
  return -1;
}

/* ========= ========= ========= ========= ========= ========= ========= =========
 *                      PHYSICAL                      READING 
 * ========= ========= ========= ========= ========= ========= ========= =========
 */

/** FPIN = PREFIX # STRING ~ [Character*][EOF] is used to comment actions;
         it means that PREFIX (previously to the action) followed by STRING(dureing the action) has been read in FPIN;
         thus '~' denotes the "current position",
         at the end of an action # is the current position before the action execution
     <parameter> denotes the value of a char or char * parameter
     '['']' are brackeks
     { STR } denotes that STR is optional
    DIRECTIVE = '#' LOGICALLINE
    a LOGICALSTRING is a sequence of characters in which the sequence '\'[eol] does not occur
    a LOGICALLINE is a LOGICALSTRING not terminated by '\', and followed by [eol]
    BLANKS = {[Blank*]'\'[eol]}* [Blank*]
*/

void
char_backward (char ch)
{
  ungetc (ch, fpin);
}

/** if <ch> is the first non read character in FPIN, getc it
 * global input: FPIN = PREFIX ~ [Character*][EOF]
 * parameter: <ch> is not EOF
 * success: FPIN = PREFIX # <ch> ~ [Character*][EOF]
 * failure: FPIN = PREFIX #~ {[Character_but_<ch>][Character*]}[EOF]
 */
static boolean
getc_if_physicalNext_char_is (char ch)
{
  char next_char;
  if ((next_char = getc (fpin)) == ch)
    return pk_true;
  if (next_char != EOF)
    ungetc (next_char, fpin);
  return pk_false;
}

static boolean
get_physicalLine (char res_str[], int res_size, char *errMess)
{
  if (fgets (res_str, res_size, fpin) == NULL)
    {
      exitERR_onEof (errMess);
      return pk_false;
    }
  return pk_true;
}

/** skip preprocessor statement taking into account line numbering changes
 * global input: FPIN = PREFIX [eol] SPACES '#' ~ [Character*][EOF]
 *               PREFIX does not end with continuation line character
 * success: FPIN = PREFIX [eol] SPACES '#' # [Character_but_eol]* [eol] ~ [Character*][EOF]
 *               [Character_but_eol]* does not end with continuation line character
 * failure: FPIN = PREFIX [eol] SPACES '#' # ~ [EOF]
 */
static void
skip_directive ()
{
  int index;
  char temp_str[MID_SIZE];
  /* get the whole line */
  while (get_physicalLine (temp_str, MID_SIZE, "in preprocesseur"))
    {
      index = find_last_nonBlank (temp_str);
      line_count++;
      if (index >= 0)
	if (temp_str[index] != '\\')
	  return;
    }
}

/** check preprocessor statement for line numbering changes
 * global input: FPIN = PREFIX [eol] SPACES '#' ~ [Character*][EOF]
 *
 * success: FPIN = PREFIX [eol] SPACES '#' # [Character_but_eol]* [eol] ~ [Character*][EOF]
 * failure: FPIN = PREFIX [eol] SPACES '#' # ~ [EOF]
 */
static void
preprocessor_check ()
{
  int iii, lngth, line_value;
  char temp_str[MID_SIZE], t2_str[MID_SIZE];
  /* get the whole line */
  if (get_physicalLine (temp_str, MID_SIZE, "in preprocesseur"))
    {
      /* PLG: temp_str[lngth - 1] should be \n CHECK ? */
      /* now parse things */
      lngth = strlen (temp_str);
      for (iii = 0; iii < lngth; iii++)
	{
	  if (isdigit (temp_str[iii]))
	    {
	      if (sscanf (&(temp_str[iii]), "%d %s", &line_value, t2_str) == 2)
		{
		  line_count = line_value;
		  lngth = strlen (t2_str);
		  if (t2_str[lngth - 1] == '"')
		    t2_str[lngth - 1] = '\0';
		  if (t2_str[0] == '"')
		    iii = 1;
		  else
		    iii = 0;
		}
	      return;
	    }
	  if (!isspace (temp_str[iii]))
	    {
	      if (strncmp (&(temp_str[iii]), "line", 4) != 0)
		return;
	      iii += 3;
	    }
	}			/* for */
    }
}

/** check preprocessor statement for line numbering changes
 * global input: FPIN = PREFIX [eol] SPACES '#' ~ [Character*][EOF]
 *
 * success: FPIN = PREFIX [eol] SPACES '#' # [Character_but_eol]* [eol] ~ [Character*][EOF]
 * failure: FPIN = PREFIX [eol] SPACES '#' # ~ [EOF]
 */
static void
parse_directive ()
{
  if (macro_expansion)
    preprocessor_check ();
  else
    skip_directive ();
}

/* ========= ========= ========= ========= ========= ========= ========= =========
 *                      LOGICAL                       READING 
 * ========= ========= ========= ========= ========= ========= ========= =========
 */

/** returns the next logical character
 *  FPIN = [Character*]#{'\'[eol]}*[Char1]~[Character*][EOF]
 *  [Char1] is neither '\' followed by [eol] nor [eol] preceded by '\'
 */
char
get_logicalChar ()
{
  char next_char;
  while ((next_char = getc (fpin)) != EOF)
    {
      if (next_char == '\\')
	{
	  if ((next_char = getc (fpin)) == EOF)
	    return '\\';
	  if (next_char == '\n')
	    line_count++;
	  else
	    {
	      ungetc (next_char, fpin);
	      return '\\';
	    }
	  /* else new step in while */ }
      else
	return next_char;
    }
  return EOF;
}

/** 
 * reads new characters in fpin until an eol or a non space character is reached
 * global input: FPIN = PREFIX ~ [Character*][EOF]
 * parameter: NONE
 * success: FPIN = PREFIX # [SPACE_but_eol]+ ~ [Character*] [EOF], where
 *           first character in [Character*] [EOF] is eol or a non space character
 * failure: FPIN = PREFIX # ~ [Character*] [EOF], where
 *           first character in [Character*] [EOF] is eol or a non space character
 *          exit on reading errors
 */
boolean
skip_logicalSpaces_but_eol ()
{
  boolean spaced = pk_false;
  char curr_char = get_logicalChar ();
  while (isspace (curr_char))
    {
      if (curr_char == '\n')
	{
	  ungetc (curr_char, fpin);
	  return spaced;
	}
      spaced = pk_true;
      curr_char = get_logicalChar ();
    }
  if (curr_char != '\n')
    ungetc (curr_char, fpin);
  return spaced;
}

/** 
 * reads spaces (including '\n' in fpin until a non space character is reached
 * global input: FPIN = PREFIX ~ [Character*][EOF]
 * parameter: NONE
 * success: FPIN = PREFIX # [SPACE_but_eol]+ ~ [Character*] [EOF], where
 *           first character in [Character*] [EOF] is eol or a non space character
 * failure: FPIN = PREFIX # ~ [Character*] [EOF], where
 *           first character in [Character*] [EOF] is eol or a non space character
 *          exit on reading errors
 */
boolean
skip_logicalSpaces ()
{
  boolean spaced = pk_false;
  char curr_char = get_logicalChar ();
  while (isspace (curr_char))
    {
      if (curr_char == '\n')
	line_count++;
      spaced = pk_true;
      curr_char = get_logicalChar ();
    }
  ungetc (curr_char, fpin);
  return spaced;
}

/** if <ch> is the first non read logical character in FPIN, getc it
 * global input: FPIN = PREFIX ~ [Character*][EOF]
 * parameter: <ch> is neither backslash nor eol or blank
 *
 * success: FPIN = PREFIX #~ <ch> [Character*][EOF]
 * failure: FPIN = PREFIX #~ [EOF]
 * failure: FPIN = PREFIX #~ [Character_0][Character*][EOF], [Character_0] is not <ch>
 */
static boolean
logicalNext_char_is (char ch)
{
  char next_char = get_logicalChar ();
  if (next_char == ch)
    {
      ungetc (next_char, fpin);
      return pk_true;
    }
  ungetc (next_char, fpin);
  return pk_false;
}

/** if <ch> is the first non read character in FPIN, getc it
 * global input: FPIN = PREFIX ~ [Character*][EOF]
 * parameter: <ch> is not EOF
 * success: FPIN = PREFIX # <ch> ~ [Character*][EOF]
 * failure: FPIN = PREFIX #~ {[Character_but_<ch>][Character*]}[EOF]
 */
boolean
getc_if_logicalNext_char_is (char ch)
{
  char next_char = get_logicalChar ();
  if (next_char == ch)
    return pk_true;
  ungetc (next_char, fpin);
  return pk_false;
}

/** test if <prefix> is the prefix of non read character in FPIN
 * global input: FPIN = PREFIX ~ [Character*][EOF]
 * parameter: <prefix>
 *
 * success: FPIN = PREFIX #~ <prefix> [Character*][EOF]
 * failure: FPIN = PREFIX #~ [EOF]
 * failure: FPIN = PREFIX #~ [Character_0][Character*][EOF], [Character_0] is not <prefix>
 */
boolean
logicalNext_string_is (char *prefix)
{
  int prefix_len = strlen (prefix);
  int jjj = 0;
  if (prefix_len == 1)
    return logicalNext_char_is (prefix[0]);
  while (jjj < prefix_len)
    {
      if (logicalNext_char_is (prefix[jjj]))
	jjj = jjj + 1;
      else
	{
	  while (jjj > 0)
	    {
	      jjj = jjj - 1;
	      ungetc (prefix[jjj], fpin);
	    }
	  return pk_false;
	}
    }
  for (jjj = prefix_len - 1; jjj >= 0; jjj--)
    ungetc (prefix[jjj], fpin);
  return pk_true;
}

/** skip spaces including '\n', after a '\n' occurrence. line count is incremented for this occurrence and the following ones
 * FPIN = [Character*][eol]~[Character*][EOF]
 * completed: 
 *          if <testDirective> is false FPIN  = [Character*][eol]#~[Character*][EOF]
 *          if <testDirective> is pk_true 
 *                  FPIN  = [Character*][eol] # { SPACES DIRECTIVE }* SPACES ~ {[Char1][Character*]}[EOF]
 *          [Char1] is not a Blank, not '\' preceding [eol], not [eol] following '\',  not '#' following Spaces
 * return pk_true (pk_false) if a directive occurs (does not occur)
 * 
 */
boolean
skip_spacesAndDirectives_inNewLogiLine (boolean testDirective)
{
  boolean directive_occurred = pk_false;
  line_count++;
  while (pk_true)
    {
      if (skip_logicalSpaces ());
      if (getc_if_logicalNext_char_is ('#'))
	{
	  if (testDirective)
	    {
	      directive_occurred = pk_true;
	      parse_directive ();
	    }			/* parse_directive returns FPIN = [Character*][eol] . [Character*][EOF] or FPIN = [Character*][EOF]. */
	  else
	    skip_directive ();
	}
      else
	return directive_occurred;
    }
  return directive_occurred;
}

/** 
 * reads new characters in fpin until [eol] is reached
 * global input: FPIN = PREFIX '/' '/' ~ [Character*][EOF]
 * success: FPIN = PREFIX '/' '/' # STRING [eol] ~ [Character*] [EOF], '*' '/' does not occur in STRING
 * failure: FPIN = PREFIX '/' '/' # STRING [EOF]~, [eol] does not occur in STRING
 */
void
skip_C_lineComment_completion ()
{
  char curr_char;
  while ((curr_char = get_logicalChar ()) != EOF)
    {
      /* test for end of comment */
      if (curr_char == '\n')
	{
	  line_count++;
	  return;
	}
    }
}

/** 
 * reads new characters in fpin until the sequence '*' '/' is reached
 * global input: FPIN = PREFIX '/' '*' ~ [Character*][EOF]
 * success: FPIN = PREFIX '/' '*' # STRING '*' '/' ~ [Character*] [EOF], '*' '/' does not occur in STRING
 * failure: FPIN = PREFIX '/' '*' # STRING [EOF]~, '*' '/' does not occur in STRING
 */
void
skip_C_fullComment_completion ()
{
  char curr_char;
  while ((curr_char = get_logicalChar ()) != EOF)
    {
      /* test for end of comment */
      if (curr_char == '\n')
	line_count++;
      else if (curr_char == '*')
	{
	  if (getc_if_physicalNext_char_is ('/'))	/* PLG WARNING SPECIAL CHARACTERS */
	    return;
	}
    }
  exitERR_onEof ("end of comment mark");
}

/** 
 * FPIN = [Character*]['/']~[Character*][EOF]
 * success: FPIN = [Character*]['/'] # ['*'][Character*]['*''/'] ~ [Character*][EOF]
 *          FPIN = [Character*]['/'] # ['/'][Character_but_eol*][eol] ~ [Character*][EOF]
 * failure: FPIN = [Character*]['/'] #~ [Character*][EOF]
 */
boolean
skipIf_comment_completion ()
{
  if (getc_if_logicalNext_char_is ('*'))
    {
      skip_C_fullComment_completion ();	/* exit on error; otherwise, '/' is the last read character in fpin */
      return pk_true;
    }
  if (getc_if_logicalNext_char_is ('/'))
    {
      skip_C_lineComment_completion ();	/* EOF is reached or [eol] is the last read character */
      return pk_true;
    }
  return pk_false;
}

/** 
 * reads new characters in fpin until a non space character is reached
 * global input: FPIN = PREFIX ~ [Character*][EOF]
 * parameter <testDirective>: if pk_true check preprocessor lines
 * success: FPIN = PREFIX # [SPACE]+ ~ [Character*] [EOF], where
 *           SPACE can be a space character, a comment or a [eol]
 *          [Character*] [EOF] does not start with a SPACE
 * failure: FPIN = PREFIX # ~ [Character*] [EOF], where
 *          [Character*] [EOF] does not start with a SPACE or a COMMENT
 */
boolean
skipIf_comments_and_spaces (boolean testDirective)
{
  char curr_char;
  boolean stepped = pk_false;
  while ((curr_char = get_logicalChar ()) != EOF)
    {
      if (curr_char == '\n')
	{
	  if (skip_spacesAndDirectives_inNewLogiLine (testDirective));	/* FPIN  = [Character*][eol] # { SPACES DIRECTIVE }* SPACES ~ 
									   {[Character_but_'#'][Character*]}[EOF]  
									   line count is incremented for [eol] occurrence preceding initial position */
	  stepped = pk_true;
	}
      else if (curr_char == '/')
	{
	  if (skipIf_comment_completion ())	/* FPIN = [Character*]['/'] # ['*'][Character*]['*''/'] ~ [Character*][EOF]
						 * FPIN = [Character*]['/'] # ['/'][Character_but_eol*][eol] ~ [Character*][EOF]*/
	    stepped = pk_true;
	  else
	    {			/* fail: FPIN = [Character*]['/'] #~ [Character*][EOF]; exit on error; */
	      ungetc ('/', fpin);
	      return stepped;
	    }
	}
      else if (!(isspace (curr_char)))
	{
	  ungetc (curr_char, fpin);
	  return stepped;
	}
    }
  return stepped;
}


/** 
 * reads characters in fpin until '"' is reached
 * global input: FPIN = PREFIX '"'~ [Character*][EOF]
 * in expression "[eol] does not occur in STRING" the sequence  '\\'[eol] may occur
 * success: FPIN = PREFIX '"' # STRING '"' ~ [Character*] [EOF], '"', '\' and [eol] do not occur in STRING
 * failure: FPIN = PREFIX '"' # STRING  [EOF]~, '"' and [eol] do not occur in STRING
 * failure: FPIN = PREFIX  '"' # STRING ~ [eol] [Character*] [EOF], '"' and [eol] do not occur in STRING
 */
void
skip_C_string_completion ()
{
  char curr_char;
  while ((curr_char = getc (fpin)) != EOF)
    {
      if (curr_char == '\n')
	{
	  exitERR_line ("unexpected newline in string");
	}
      if (curr_char == '"')
	return;
      else if (curr_char == '\\')	/* escape sequence */
	curr_char = getc (fpin);
    }
  exitERR_onEof ("string mark");
}

/** 
 * reads new characters in fpin until the ''' is reached
 * global input: FPIN = PREFIX '''~ [Character*][EOF]
 * in expression "[eol] does not occur in STRING" the sequence  '\\'[eol] may occur
 * success: FPIN = PREFIX ''' # STRING ''' ~ [Character*] [EOF], ''' and [eol] do not occur in STRING
 * failure: FPIN = PREFIX ''' # STRING  [EOF]~, ''' and [eol] do not occur in STRING
 * failure: FPIN = PREFIX  ''' # STRING ~ [eol] [Character*] [EOF], ''' and [eol] do not occur in STRING
 */
void
skip_C_characterConstant_completion ()
{
  char curr_char;
  while ((curr_char = getc (fpin)) != EOF)
    {
      if (curr_char == '\n')
	exitERR_line ("unexpected newline in character constant");
      if (curr_char == '\'')
	return;
      if (curr_char == '\\')
	curr_char = getc (fpin);
    }
  exitERR_onEof ("character mark");
}

/** 
 * parameter  <testDirective>: if pk_true check preprocessor lines
 */
void
skipTo_endOfUnit (boolean testDirective)
{
  char curr_char;
  while ((curr_char = getc (fpin)) != EOF)
    {
      /* test for end of comment */
      if (curr_char == ';')
	{
	  ungetc (';', fpin);
	  return;
	}
      else if (curr_char == '\n')
	{
	  if (skip_spacesAndDirectives_inNewLogiLine (testDirective));
	}			/* FPIN  = [Character*][eol] # { SPACES DIRECTIVE }* SPACES ~ 
				   {[Character_but_'#'][Character*]}[EOF]  
				   line count is incremented for [eol] occurrence preceding initial position */
      else if (curr_char == '/')
	{
	  if (skipIf_comment_completion ());
	}			/* fail: FPIN = [Character*]['/'] #~ [Character*][EOF] ; exit on error; */
      else if (curr_char == '"')
	skip_C_string_completion ();	/* exit on failure  '"' is the last read character */
      else if (curr_char == '\'')
	skip_C_characterConstant_completion ();	/* exit on error ''' is the last read character  */
      else if (curr_char == '{')
	{			/* PLG IS IT CORRECT ??? */
	  skip_braceInner ();
	  return;
	}
      else if (curr_char == '}')
	exitERR_line ("unexpected '}' character");	/* exit */
      else if (curr_char == '(')
	skip_parenthesisInner ();
      else if (curr_char == ')')
	exitERR_line ("unexpected ')' character");	/* exit */
      else if (curr_char == '[')
	skip_squareBracketInner ();	/* exit on error */
      else if (curr_char == ']')
	exitERR_line ("unexpected ']' character");
    }				/* exit */
  exitERR_onEof ("; or  { ...}");
}

/** reach '{' or ';' first occurrence at level 0
 * parameter  <testDirective>: if pk_true check preprocessor lines
 * skip inner nesting such as strings, comments, C blocks,...
 * return pk_true if '{' is the last read character
 * return false if ';' or EOF is reached
 * exit on inner syntactic error
 */
boolean
reach_blockStart (boolean testDirective)
{
  char curr_char;
  while ((curr_char = getc (fpin)) != EOF)
    {
      if (curr_char == ';')
	{
	  ungetc (';', fpin);
	  return pk_false;
	}
      if (curr_char == '{')
	return pk_true;
      if (curr_char == '\n')
	{
	  if (skip_spacesAndDirectives_inNewLogiLine (testDirective));	/* FPIN  = [Character*][eol] # { SPACES DIRECTIVE }* SPACES ~ 
									   {[Character_but_'#'][Character*]}[EOF]  
									   line count is incremented for [eol] occurrence preceding initial position */
	  else if (isspace (curr_char))
	    {
	      if (skip_logicalSpaces_but_eol ());
	    }
	}
      else if (curr_char == '/')
	{
	  if (!skipIf_comment_completion ());
	}
    }				/* fail: FPIN = [Character*]['/'] #~ [Character*][EOF]; exit on error;  */
  return pk_false;
}

/** INPUT:  FPIN = PREFIX ~ [Character*] EOF
 *  OUTPUT: FPIN = PREFIX # INNER ']' ~ EOF where INNER is a well balanced (wrt '('-')' '['-']' '{'-'}' Character, String Comments  bracketing) sequence of Characters
 */
void
skip_squareBracketInner ()
{
  char curr_char;
  int depth = 1;
  while ((curr_char = getc (fpin)) != EOF)
    {
      if (curr_char == '\n')
	register_physicalEol ();	/* FPIN = [Character*] { [[eol]~[Character*][EOF]] + [[EOF]~] } */
      else if (curr_char == '/')
	{
	  if (skipIf_comment_completion ());
	}			/* fail: FPIN = [Character*]['/'] #~ [Character*][EOF] ; exit on error; */
      else if (curr_char == '"')
	skip_C_string_completion ();	/* exit on failure, '"' is the last read character  */
      else if (curr_char == '\'')
	skip_C_characterConstant_completion ();	/* exit on error ''' is the last read character  */
      else if (curr_char == '{')
	skip_braceInner ();
      else if (curr_char == '}')
	exitERR_line ("unexpected '}' character");
      else if (curr_char == '(')
	skip_parenthesisInner ();
      else if (curr_char == ')')
	exitERR_line ("unexpected ')' character");
      else if (curr_char == '[')
	depth++;
      else if (curr_char == ']')
	{
	  --depth;
	  if (depth == 0)
	    return;
	}
    }
}

/** INPUT:  FPIN = PREFIX ~ [Character*] EOF
 *  OUTPUT: FPIN = PREFIX # INNER ')' ~ EOF where INNER is a well balanced (wrt '('-')' '['-']' '{'-'}' Character, String Comments  bracketing) sequence of Characters
 */
void
skip_parenthesisInner ()
{
  char curr_char;
  int depth = 1;
  while ((curr_char = getc (fpin)) != EOF)
    {
      if (curr_char == '\n')
	register_physicalEol ();	/* FPIN = [Character*] { [[eol]~[Character*][EOF]] + [[EOF]~] } */
      else if (curr_char == '/')
	{
	  if (skipIf_comment_completion ());
	}			/* fail: FPIN = [Character*]['/'] #~ [Character*][EOF] ; exit on error; */
      else if (curr_char == '"')
	skip_C_string_completion ();	/* exit on failure, '"' is the last read character  */
      else if (curr_char == '\'')
	skip_C_characterConstant_completion ();	/* exit on error ''' is the last read character  */
      else if (curr_char == '{')
	skip_braceInner ();
      else if (curr_char == '}')
	exitERR_line ("unexpected '}' character");
      else if (curr_char == '(')
	depth++;
      else if (curr_char == ')')
	{
	  --depth;
	  if (depth == 0)
	    return;
	}
      else if (curr_char == ']')
	exitERR_line ("unexpected ']' character");
      else if (curr_char == '[')
	skip_squareBracketInner ();
    }
}

/** INPUT:  FPIN = PREFIX ~ [Character*] EOF
 *  OUTPUT: FPIN = PREFIX # INNER '}' ~ EOF where INNER is a well balanced (wrt '('-')' '['-']' '{'-'}' Character, String Comments  bracketing) sequence of Characters
 */
void
skip_braceInner ()
{
  char curr_char;
  int depth = 1;
  while ((curr_char = getc (fpin)) != EOF)
    {
      if (curr_char == '\n')
	register_physicalEol ();	/* FPIN = [Character*] { [[eol]~[Character*][EOF]] + [[EOF]~] } */
      else if (curr_char == '/')
	{
	  if (skipIf_comment_completion ());
	}			/* fail: FPIN = [Character*]['/'] #~ [Character*][EOF] ; exit on error; */
      else if (curr_char == '"')
	skip_C_string_completion ();	/* exit on failure, '"' is the last read character  */
      else if (curr_char == '\'')
	skip_C_characterConstant_completion ();	/* exit on error ''' is the last read character  */
      else if (curr_char == '{')
	depth++;
      else if (curr_char == '}')
	{
	  --depth;
	  if (depth == 0)
	    return;
	}
      else if (curr_char == '(')
	skip_parenthesisInner ();
      else if (curr_char == ')')
	exitERR_line ("unexpected ')' character");
      else if (curr_char == ']')
	exitERR_line ("unexpected ']' character");
      else if (curr_char == '[')
	skip_squareBracketInner ();
    }
}

/* if <head_str> is not empty all characters in <head_str> are first copied to <cmt_str>[<comment_len> ...]
 * read and append to <cmt_str> all characters until end comment indicator included 
 * when entering, last read character is "*" following '/'  
 * when leaving, last read character is '/' following '*'  
 */
int
get_comment (char cmt_str[])
{
  int curr_char;
  int comment_len;

  /* keep going until scan is complete */
  comment_len = 0;
  while ((curr_char = get_logicalChar ()) != EOF)
    {
      /* test for end of comment */
      if (curr_char == '\n')
	line_count++;
      if ((curr_char == '*') && getc_if_physicalNext_char_is ('/'))
	{
	  break;
	}
      /* store it */
      cmt_str[comment_len++] = curr_char;
    }
  /* close storage */
  cmt_str[comment_len] = '\0';
  return comment_len;
}

/** in:  <curr_char> is an identifier header char
 *       FPIN = PREFIX <curr_char> ~ [Character*][EOF]
 *       <res_index> is an integer that satisfaies -1 <= <res_index> < sizeof(res_str)
 *       <res_str> is an array large enough to contain an identifier after <res_index> position
 *       if  <res_index> > 0, <res_str>[<res_index>] == '\0'
 *  out: FPIN = PREFIX <curr_char> # IDENT ~ [Character*][EOF] 
 *             the first character in [Character*][EOF] is not an identifier character
 *       <curr_char>IDENT is stored in <res_str>[(res_index+1)..(res_index+strlen(IDENT))]
 *       <res_str>[res_index+strlen(IDENT)+1] == '\0'
 * returns <res_index>+strlen(<curr_char>IDENT)
 */
int
get_identifier_completion (char curr_char, char res_str[], int res_index)
{
  int ii = res_index;
  do
    {
      ii++;
      res_str[ii] = curr_char;
      if ((curr_char = getc (fpin)) == EOF)
	{
	  ii++;
	  res_str[ii] = '\0';
	  return ii;
	}
    }
  while ((isalnum (curr_char)) || (curr_char == '_'));
  ungetc (curr_char, fpin);
  ii++;
  res_str[ii] = '\0';
  //  curr_char, res_index, ii, (char*)(res_str+res_index+1));
  return ii;
}

void
check_inputFile (char *filePath)
{
  STRUCT_STAT fst;

  char tmp_str[SMALL_SIZE];
  if (filePath == NULL)
    exitERR ("input file path is empty");
  if (FUNC_STAT (filePath, &fst) != 0)
    {
      sprintf (tmp_str, "error: could not access file <%s>", filePath);
      exitERR (tmp_str);
    }

  if ((fpin = fopen (filePath, "r")) == NULL)
    {
      sprintf (tmp_str, "error: could not read file <%s>", filePath);
      exitERR (tmp_str);
    }
  fclose (fpin);
}

void
init_reading ()
{
  while (getc_if_physicalNext_char_is ('#'))
    parse_directive ();		/* parse_directive returns FPIN = [Character*][eol] . [Character*][EOF] or FPIN = [Character*][EOF]. */
}

void
open_inputFile (char *filePath)
{
  char tmp_str[SMALL_SIZE];
  if ((fpin = fopen (filePath, "r")) == NULL)
    {
      sprintf (tmp_str, "unable to open input file <%s>", filePath);
      exitERR (tmp_str);
    }
  in_filePath = filePath;
}

void
close_inputFile ()
{
  char tmp_str[SMALL_SIZE];
  if (fclose (fpin) == EOF)
    {
      sprintf (tmp_str, "unable to close input file <%s>", in_filePath);
      exitERR (tmp_str);
    }
}

/* ========================= STANDARD WRITING ========================= */

/* output file */
static FILE *fpout;

#define PUT_CHAR(CHAR) fputc(CHAR, fpout)
#define PUT_STRING(STR) fputs(STR, fpout)

#define PUT_SPACE() PUT_CHAR(' ')
#define PUT_EOL() PUT_CHAR('\n')
#define PUT_CHAR_EOL(CHAR) PUT_CHAR(CHAR); PUT_EOL()
#define PUT_STRING_SP(STR) PUT_STRING(STR); PUT_SPACE()
#define PUT_STRING_EOL(STR) PUT_STRING(STR); PUT_EOL()
#define PUT_COMMENTSTART() PUT_STRING("/*")
#define PUT_COMMENTSTART_SP() PUT_COMMENTSTART(); PUT_SPACE()
#define PUT_COMMENTSTART_EOL() PUT_STRING("/*"); PUT_EOL()
#define PUT_COMMENTEND() PUT_STRING("*/")
#define PUT_COMMENTEND_EOL() PUT_COMMENTEND(); PUT_EOL()

void
put_space ()
{
  PUT_SPACE ();
}

static void
put_nSpace (int length)
{
  int ii;
  for (ii = 0; ii < length; ii++)
    {
      PUT_SPACE ();
    }
}

void
put_char (char CC)
{
  PUT_CHAR (CC);
}

void
put_eol ()
{
  PUT_EOL ();
}

void
put_abstractProtoEnd ()
{
  PUT_CHAR_EOL (';');
  PUT_EOL ();
}


void
put_string (const char *str)
{
  PUT_STRING (str);
}

void
put_publicAccess ()
{
  PUT_STRING_EOL ("public:");
  PUT_EOL ();
}

void
put_protectedAccess ()
{
  PUT_STRING_EOL ("protected:");
  PUT_EOL ();
}

void
put_privateAccess ()
{
  PUT_STRING_EOL ("private:");
  PUT_EOL ();
}

void
put_comment (const char *str)
{
  PUT_COMMENTSTART ();
  PUT_STRING (str);
  PUT_COMMENTEND ();
}

void
put_comment_eol (const char *str)
{
  PUT_COMMENTSTART ();
  PUT_STRING_SP (str);
  PUT_COMMENTEND_EOL ();
}

static void
put_definedVariable ()
{
  PUT_STRING ("__");
  PUT_STRING (header_string);
  PUT_STRING ("__");
}

/** if header_string == NULL or is empty nothing is done
 */
static void
put_conditionalDef_opening ()
{
  if (header_string != NULL)
    if (header_string[0] != '\0')
      {				// Added June 2011 lbesnard
	PUT_STRING ("#ifndef ");
	put_definedVariable ();
	PUT_EOL ();
	PUT_STRING ("#define ");
	put_definedVariable ();
	PUT_EOL ();
	PUT_EOL ();
      }
}

/** if header_string == NULL or is empty nothing is done
 */
static void
put_conditionalDef_closing ()
{
  if (header_string != NULL)	// Added June 2011 lbesnard
    if (header_string[0] != '\0')
      {				// Added June 2011 lbesnard
	PUT_STRING ("#endif ");
	PUT_COMMENTSTART ();
	put_definedVariable ();
	PUT_COMMENTEND ();
	PUT_EOL ();
      }
}

static void
put_fileInformation (void)
{
  char tmp_str[SMALL_SIZE];
  PUT_COMMENTSTART_EOL ();
  PUT_STRING_SP (" *   This file was automatically generated by version");
  sprintf (tmp_str, "%d.%d of %s.", VERSION, PATCHLEVEL, PROGRAM);
  PUT_STRING_EOL (tmp_str);
  PUT_STRING_EOL (" *   Manual editing not recommended.");
  PUT_SPACE ();
  PUT_COMMENTEND_EOL ();
}

static char *
current_local_time (void)
{
  time_t timeval;

  timeval = time ((time_t *) NULL);
  return (ctime (&timeval));
}

static void
put_creationDate (void)
{
  char tmp_str[SMALL_SIZE];
  int tmp_lgth;
  PUT_COMMENTSTART_SP ();
  put_nSpace (20);
  PUT_STRING_SP ("Created:");
  strcpy (tmp_str, current_local_time ());
  tmp_lgth = strlen (tmp_str);
  if ((tmp_lgth > 0) && (tmp_str[tmp_lgth - 1] == '\n'))
    tmp_str[tmp_lgth - 1] = '\0';
  PUT_STRING (tmp_str);
  put_nSpace (20);
  PUT_COMMENTEND_EOL ();
}

void
put_filePrelude ()
{
  put_fileInformation ();
  if (registerTime)
    put_creationDate ();
  PUT_EOL ();
  put_conditionalDef_opening ();
}

void
put_filePostlude ()
{
  /* finish header */
  PUT_EOL ();
  put_conditionalDef_closing ();
}

void
put_forced_comment_eol (const char *orAnd_str)
{
  PUT_COMMENTSTART ();
  PUT_STRING_SP (commentPrefix);
  PUT_STRING_SP ("\\todo{");
  PUT_STRING_SP ("MISSING GLOBAL FILE");
  PUT_STRING_SP (orAnd_str);
  PUT_STRING_SP ("FIRST FUNCTION, API DOCUMENTATION");
  PUT_CHAR ('}');
  PUT_SPACE ();
  PUT_COMMENTEND_EOL ();
  PUT_EOL ();
}

void
put_headingCommentMissing ()
{
  PUT_COMMENTSTART ();
  PUT_STRING_SP (commentPrefix);
  PUT_STRING_SP ("\\todo{");
  PUT_STRING_SP ("MISSING GLOBAL DOCUMENTATION");
  PUT_CHAR ('}');
  PUT_SPACE ();
  PUT_COMMENTEND_EOL ();
}

void
put_forced_functionComment ()
{
  PUT_COMMENTSTART ();
  PUT_STRING_SP (commentPrefix);
  PUT_STRING_SP ("\\todo{");
  PUT_STRING_SP ("MISSING FUNCTION API DOCUMENTATION");
  PUT_CHAR ('}');
  PUT_SPACE ();
  PUT_COMMENTEND_EOL ();
}


void
open_outputFile (const char *filePath)
{
  char tmp_str[SMALL_SIZE];
  if (filePath == NULL)
    fpout = stdout;
  else if ((fpout = fopen (filePath, "w")) == NULL)
    {
      sprintf (tmp_str, "unable to open output file <%s>", filePath);
      exitERR (tmp_str);
    }
}

void
close_outputFile ()
{
  if (fpout != stdout)
    if (fclose (fpout) == EOF)
      exitERR ("unable to close output file");
}

void
remove_file (char *filePath)
{
  if (remove (filePath) != 0)
    fprintf (stderr, "WARNING: unable to remove the temporary file <%s>\n", filePath);
}
