/*
 *
 * The purpose of this program is to extract the function descriptions
 * (prototypes) from C and C++ source code.  It also provides for the creation
 * of documentation based on those prototypes with doxygen style
 * 
 * It is derived from cextract (see below)
 * Copyright (c) 2010, 2011 by Loc Besnard, Paul Le Guernic for modifications
 *
 */

/* the command to perform the C preprocessing (full path best) with option to keep all comments but in directive
 * is OS dependent
 */
#define PREPROCESSING "/lib/cpp -C -D_PUBLIC_=public: -D_PROTECTED_=protected:  -D_PRIVATE_=private: "

/* note:  need 3 %s entries:  first = command, second = infile, third = outfile */
#define PIPEOUT_FMT "%s %s %s"

#include <stdlib.h>

int commentMode_option;
boolean forceComment;
char *commentPrefix;
char *header_string;
boolean warnComment = pk_false;
int staticMode_option;
boolean macro_expansion = pk_false;
boolean registerTime = pk_false;
boolean exportOnlyCommentedFunctions;

static char *outPrototype_filePath;
static char *inSrc_filePath;	/* not required */
static char *expandedSrc_filePath;	/* not required */
static boolean setImplicitHeader = pk_false;

static int is_staticStatusDefined;

static int is_commentStatusDefined;

/* show the version of the program */
static void
show_version (void)
{
  fprintf (stderr, "%s: version %d.%d, Copyright 2011 by Loc Besnard, Paul Le Guernic\n", PROGRAM, VERSION, PATCHLEVEL);
}

/* show the usage for the program */
static void
show_usage (void)
{
  fprintf (stderr, "Usage: %s [?] [-c[str]] or [-C[str]], [-H[str]] [-o ofile] [-p] [-t] [-s[!]] [-v] [-A] file \n", PROGRAM);
  fprintf (stderr, "\t arguments have free order \n");
  fprintf (stderr, "\t C/C++ language options --------------------------------------\n");
  fprintf (stderr, "\t -c            keep the last comment related to each prototype\n");
  fprintf (stderr, "\t -cstring      keep last comment related to each prototype when this comment starts with string\n");
  fprintf (stderr, "\t -c*           does not work with some  shell: use -c\"*\"\n");
  fprintf (stderr, "\t -C, -Cstring, as -c, -cstring , adding a specific comment when comment is missing \n");
  fprintf (stderr, "\t               comment option missing: comments are ignored\n");
  fprintf (stderr, "\t -Hstr         use header string 'str' in #define command and protype comments\n");
  fprintf (stderr, "\t -H            use unsuffixed source file name as header string in #define command and protype comments\n");
  fprintf (stderr, "\t               header option missing: text generation requiring header string are omitted\n");
  fprintf (stderr, "\t -o ofile      send output to 'ofile'\n");
  fprintf (stderr, "\t               output option missing: send output to standard output\n");
  fprintf (stderr, "\t -p            preprocessing is applied to src file before prototype extraction\n");
  fprintf (stderr, "\t -t            register the date of the file generation \n");
  fprintf (stderr, "\t               date option missing: the building date is not registered \n");
  fprintf (stderr, "\t -v            display version information then exit\n");
  fprintf (stderr, "\t -w            print a warning when a candidate comment is overwritten\n");
  fprintf (stderr, "\t ?             display help then exit\n");
  fprintf (stderr, "\t specific C language options ------------------------------------------\n");
  fprintf (stderr, "\t -s            prototypes of static functions are registered\n");
  fprintf (stderr, "\t -s!           prototypes of extern functions are not registered\n");
  fprintf (stderr, "\t               static option missing: prototypes of static functions are not registered\n");
  fprintf (stderr, "\t Wizzard options ------------------------------------------\n");
  fprintf (stderr, "\t -A            Only the commented functions are exported\n");
}

/* show general usage information */
static void
show_help (void)
{
  fprintf (stderr, "Action: %s extracts C/C++ prototype and documentation for functions defined in argument 'file'\n", PROGRAM);
  fprintf (stderr, "\t    'file' contains ANSI text that may include substitutes for 'extern', 'static', ...\n");
  fprintf (stderr, "\t           (see language_kw_extract.h in %s directory)\n", PROGRAM);
  fprintf (stderr, "\t     struct, union and enum specifiers cannot be used in function declarations\n");
  fprintf (stderr, "\tprototypes:\n");
  fprintf (stderr, "\t    output contains the list of ANSI prototypes that does not contain some keywords replaced by their substitutes \n");
  fprintf (stderr, "\t    implicit extern storage class specifier is made explicit as 'EXTERN' for each C function \n");
  fprintf (stderr, "\tComments\n");
  fprintf (stderr, "\t    include first file comment wrt the given options\n");
  fprintf (stderr, "\t    include comments immediately preceding functions wrt the given options\n");
  fprintf (stderr, "\t    may fail when 'file' contains include to files not in the execution directory\n");
  fprintf (stderr, "\t    abstract declarations are extracted\n");
  show_usage ();
  exit (0);
}

static void
exitERR_command (const char *mess)
{
  fprintf (stderr, "### ERROR in ### %s ### : Exit on %s\n", PROGRAM, mess);
  show_usage ();
  exit (1);
}

/* preprocessor */
static void
expand_macros (char *inSrc_filePath, char *expandedSrc_filePath, char *cpp_prog)
{
  static char newcommand[SMALL_SIZE];

  sprintf (newcommand, PIPEOUT_FMT, cpp_prog, inSrc_filePath, expandedSrc_filePath);
  if (system (newcommand) == -1)
    exitERR ("file expansion failed");
}

static void
create_input_file (char *filePath)
{
  char tmp_str[SMALL_SIZE];
  if (inSrc_filePath != NULL)
    {
      sprintf (tmp_str, "double definition of input file name <%s>", inSrc_filePath);
      exitERR (tmp_str);
    }
  if ((inSrc_filePath = (char *) malloc ((sizeof (char)) * (strlen (filePath) + 1))) == NULL)
    exitOn_memoryAllocationErr ();
  strcpy (inSrc_filePath, filePath);
  /* check if the file exits and if it is readable */
}

static void
create_expandedFile ()
{
  /* remove expanded_file before the program termination */
  if ((expandedSrc_filePath = (char *) malloc ((sizeof (char)) * (L_tmpnam + 1))) == NULL)
    exitOn_memoryAllocationErr ();
  if (tmpnam (expandedSrc_filePath) == NULL)
    exitERR ("temporary file creation failed");
}

static void
create_outputFile (char *filePath)
{
  if (outPrototype_filePath != NULL)
    exitERR_command ("double definition of output file name");
  if ((outPrototype_filePath = (char *) malloc ((sizeof (char)) * (strlen (filePath) + 1))) == NULL)
    exitOn_memoryAllocationErr ();
  strcpy (outPrototype_filePath, filePath);
}

static void
set_commentStatus (char *str)
{
  int str_len;
  if (is_commentStatusDefined)
    exitERR_command ("too much occurrences of -C options");
  str_len = strlen (str);
  if (str_len == 0)
    {				/* -C: include all comments */
      if ((commentPrefix = (char *) malloc (sizeof (char))) == NULL)
	exitOn_memoryAllocationErr ();
      commentPrefix[0] = '\0';
      commentMode_option = ANY_COMMENTS;
    }
  else
    {
      if ((commentPrefix = (char *) malloc ((sizeof (char)) * (str_len + 1))) == NULL)
	exitOn_memoryAllocationErr ();
      strcpy (commentPrefix, str);
      commentMode_option = SELECTED_COMMENTS;
    }
  is_commentStatusDefined = pk_true;
}

static void
set_headerString (char *str)
{
  int str_len;
  if (setImplicitHeader || (header_string != NULL))
    exitERR_command ("double definition of header string");
  str_len = strlen (str);
  if (str_len == 0)
    setImplicitHeader = pk_true;
  else
    {
      if ((header_string = (char *) malloc ((sizeof (char)) * (str_len + 1))) == NULL)
	exitOn_memoryAllocationErr ();
      strcpy (header_string, str);
    }
}

static void
set_outputFile (int count, int argc, char **argv)
{
  if (count == argc)
    exitERR_command ("file name missing in -o option");
  create_outputFile (argv[count]);
}

static void
set_staticStatus (char *str)
{				/* no static function if -s does not occur */
  int str_len = strlen (str);
  if (is_staticStatusDefined)
    exitERR_command ("too much occurrences of -s options");
  if (str_len == 0)		/* -s: include static function definitions */
    staticMode_option = ANY_STATICS;
  else if (str_len <= 1)
    {
      if (str[0] == '!')	/* -so: include only static function definitions */
	staticMode_option = ONLY_STATICS;
      else
	exitERR_command ("unknown attribute to -s option");
    }
  else
    exitERR_command ("unknown attribute to -s option");
  is_staticStatusDefined = pk_true;
}

/* determine the configuration command, exit if invalid */
static void
parse_commandArguments (int argc, char **argv)
{
  int count;

  for (count = 1; count < argc; count++)
    {
      if ((argv[count][0]) == '-')
	{
	  char actionType = argv[count][1];
	  if (actionType == 'c')
	    {
	      set_commentStatus ((char *) (argv[count] + 2));	/* exit on error */
	      forceComment = pk_false;
	    }
	  else if (actionType == 'C')
	    {
	      set_commentStatus ((char *) (argv[count] + 2));	/* exit on error */
	      forceComment = pk_true;
	    }
	  else if (actionType == 'H')
	    set_headerString ((char *) (argv[count] + 2));	/* exit on error */
	  else if (actionType == 'o')
	    {
	      count = count + 1;
	      set_outputFile (count, argc, argv);
	    }			/* exit on error */
	  else if (actionType == 'p')
	    macro_expansion = pk_true;	/* exit on error */
	  else if (actionType == 's')
	    set_staticStatus ((char *) (argv[count] + 2));	/* exit on error */
	  else if (actionType == 't')
	    registerTime = pk_true;	/* exit on error */
	  else if (actionType == 'v')
	    {
	      show_version ();
	      exit (0);
	    }
	  else if (actionType == 'w')
	    warnComment = pk_true;
	  else if (actionType == 'A') // Added LBesnard(export only the commented functions)
	      exportOnlyCommentedFunctions = pk_true;
	    else
	      exitERR_command ("unknown action");
	}
      else if ((argv[count][0]) == '?')
	{
	  show_help ();
	  exit (0);
	}
      else
	{			/* treat it as an input file name */
	  create_input_file (argv[count]);	/* set inSrc_filePat; exit on error */
	  check_inputFile (inSrc_filePath);
	}			/* exit on error */
    }
}


/* control program flow and command line options */
int
main (int argc, char **argv)
{
  /* Initialisatiosn */
  header_string = NULL;
  outPrototype_filePath = NULL;
  inSrc_filePath = NULL;
  is_staticStatusDefined = pk_false;
  is_commentStatusDefined = pk_false;
  exportOnlyCommentedFunctions=pk_false;

  // strcpy(PROGRAM, &(argv[0][jjj]));

  parse_commandArguments (argc, argv);

  if (!is_staticStatusDefined)
    staticMode_option = NO_STATICS;

  if (!is_commentStatusDefined)
    commentMode_option = NO_COMMENTS;

  if (inSrc_filePath == NULL)
    exitERR_command ("parsing file argument missing\n");

  if (setImplicitHeader)
    header_string = extractCFileName (inSrc_filePath);

  /* set option for non macro expansion */
  if (macro_expansion)
    {
      /* expand macro */
      create_expandedFile ();	/* set expandedSrc_filePath; exit on ERROR */
      expand_macros (inSrc_filePath, expandedSrc_filePath, PREPROCESSING);
      file_proceed (expandedSrc_filePath, outPrototype_filePath);
      remove_file (expandedSrc_filePath);
    }
  else
    file_proceed (inSrc_filePath, outPrototype_filePath);
  exit (0);
}
